/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <Xpa/Primitive.h>
#include <Xpa/PrimitiveP.h>
#include <Xpa/CanvasP.h>
#include <Xpa/Canvas.h>
  
#define W	w -> canvas
#define P	w -> primitive
#define C	w -> core

static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Redisplay (Widget, XEvent*, Region);
static void		Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static void		Resize (Widget);

CanvasClassRec canvasClassRec = {
	/* CoreCLassPart */
	{
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"Canvas",						/* class_name */
		sizeof (CanvasRec),				/* widget_size */
		NULL,							/* class_initialize */
		NULL,							/* class_part_init */
		FALSE,							/* class_inited */
		Initialize,						/* initialize */
		NULL,							/* initialize_hook */
		Realize,						/* realize */
		NULL,							/* actions */
		0,								/* num_actions */
		NULL,							/* resources */
		0,								/* num_resources */
		NULLQUARK,						/* xrm_class */
		False,							/* compress_motion */
		XtExposeCompressMaximal,		/* compress_exposure */
		False,							/* compress_enterlv */
		False,							/* visible_interest */
		NULL,							/* destroy */
		Resize,							/* resize */
		Redisplay,						/* expose */
		NULL,							/* set_values */
		NULL,							/* set_values_hook */
		XtInheritSetValuesAlmost,		/* set_values_almost */
		NULL,							/* get_values_hook */
		NULL,							/* accept_focus */
		XtVersion,						/* version */
		NULL,							/* callback_private */
		"",								/* tm_table */
		NULL,							/* query_geometry */
	},
	/* PrimitiveCLassPart */
	{
		NULL,
	},
	/* CanvasClassPart */
	{
		NULL,
	},
};

WidgetClass canvasWidgetClass = (WidgetClass)&canvasClassRec;

/*
 *	Cre la fentre.
 */

static void Realize (Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	CanvasWidget	w = (CanvasWidget) widget;

	*value_mask |= CWBitGravity;
	attributes -> bit_gravity = NorthWestGravity;
	XtCreateWindow (widget, InputOutput, (Visual *)CopyFromParent, *value_mask, attributes);
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	CanvasWidget	w = (CanvasWidget) new;

	W.width = C.width;
	W.height = C.height;
}

/*
 *	Traite le changement de taille (efface les ombres)
 */

static void Resize (Widget widget)
{
	CanvasWidget	w = (CanvasWidget) widget;
	Display*		dpy = XtDisplay (widget);
	Window			win = XtWindow (widget); 
	int				st = P.shadow_thickness;

	if (!XtIsRealized (widget) || P.shadow_thickness == 0) {
		W.width = C.width;		/* mettre  jour les */
		W.height = C.height;	/* les dimensions */
		return;
	}

	/* rafficher les ombres */
	if (C.width >= W.width) {
		if (C.height >= W.height) {
			if (C.width != W.width)		/*  droite */
				XClearArea (dpy, win, W.width-st, 0, st, W.height, True);
			if (C.height != W.height)	/* en bas */
				XClearArea (dpy, win, 0, W.height-st, W.width, st, True);
		} else {
			if (C.width != W.width)		/*  droite */
				XClearArea (dpy, win, C.width-st, 0, st, C.height, True);
			/* en bas */
			XClearArea (dpy, win, 0, C.height-st, W.width, st, True);
		}
	} else {
		if (C.height >= W.height) {
			/*  droite */
			XClearArea (dpy, win, C.width-st, 0, st, C.height, True);
			if (C.height != W.height)	/* en bas */
				XClearArea (dpy, win, 0, C.height-st, C.width, st, True);
		} else {
			/*  droite */
			XClearArea (dpy, win, C.width-st, 0, st, C.height, True);
			/* en bas */
			XClearArea (dpy, win, 0, C.height-st, C.width, st, True);
		}
	}
	W.width = C.width;		/* mettre  jour les */
	W.height = C.height;	/* les dimensions (APRS avoir affich les ombres) */
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* e, Region region)
{
	CanvasWidget	w = (CanvasWidget) widget;
	Display*		dpy = XtDisplay (widget);
	Window			window = XtWindow (widget); 

	if (P.shadow_thickness > 0) {
		XSetRegion (dpy, P.top_GC, region);
		XSetRegion (dpy, P.bot_GC, region);
		_DrawShadow (dpy, window,
						P.bot_GC, P.top_GC, P.shadow_thickness,
						C.width, C.height, 0, 0);
		XSetClipMask (dpy, P.top_GC, None);
		XSetClipMask (dpy, P.bot_GC, None);
	}
}
