/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/CompositeP.h>
#include <X11/StringDefs.h>
#include <Xpa/ClipP.h>

static XtGeometryResult
			ClipGeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void	InsertChild (Widget);
static void	ClipResize (Widget);
static void	ClipChangeManaged (Widget);
static void	ClipInitialize (Widget, Widget, Arg*, Cardinal*);

#define	C	((Widget)cw) -> core
#define	CP	((ClipWidget)cw) -> composite
#define	W	((ClipWidget)cw) -> clip

/*
 *	La classe composite mais avec un geometry manager.
 */

ClipClassRec clipClassRec = {
	/* core_class fields */
	{
		/* superclass */			(WidgetClass) &compositeClassRec,
		/* class_name */			"ClipWindow",
		/* widget_size */			sizeof(ClipRec),
		/* class_initialize */		NULL,
		/* class_partinit */			NULL,
		/* class_inited */			False,
		/* initialize */				ClipInitialize,
		/* Init hook */				NULL,
		/* realize */				XtInheritRealize,
		/* actions */				NULL,
		/* num_actions */			0,
		/* resources */			NULL,
		/* num_resources */		0,
		/* xrm_class */			NULLQUARK,
		/* compress_motion */		False,
		/* compress_exposure */	False,
		/* compress_enterleave */	False,
		/* visible_interest */		False,
		/* destroy */				NULL,
		/* resize */				ClipResize,
		/* expose */				NULL,
		/* set_values */			NULL,
		/* set values hook */		NULL,
		/* set values almost */		XtInheritSetValuesAlmost,
		/* get values hook */		NULL,
		/* accept_focus */			NULL,
		/* Version */				XtVersion,
		/* PRIVATE cb list */		NULL,
		/* TM Table */			"",
		/* query_geom */			NULL,
	},
	/* composite_class fields */
	{
		/* geometry_manager */	ClipGeometryManager,
		/* change_managed */		ClipChangeManaged,
		/* insert_child */			InsertChild,
		/* delete_child */			XtInheritDeleteChild,
		/* extension */			NULL,
  }
};

WidgetClass	clipWidgetClass = (WidgetClass) & clipClassRec;

/*
 *	Droule le clipper.
 */

static void wait_for_expose (Widget widget)
{
	XtAppContext	app = XtWidgetToApplicationContext (widget);
	XEvent		e;
	Display*	dpy = XtDisplay (widget);
	Window		win = XtWindow (widget);

#if 0
	do {
		XtAppNextEvent (app, &e);
		if (e.type == GraphicsExpose)
			e.type = Expose;
		XtDispatchEvent (&e);
		} while (e.type != Expose && e.type != NoExpose);

	while (XCheckTypedWindowEvent (dpy, win, Expose, &e) ||
	       XCheckTypedWindowEvent (dpy, win, GraphicsExpose, &e))
	{
		e.type = Expose;
		XtDispatchEvent (&e);
	}
#else
	XSync (dpy, False);
	while (XtAppPending (app))
		XtAppProcessEvent (app, -1);
#endif
}

static void Layout (Widget widget, Boolean set_size)
{
	ClipWidget	cw = (ClipWidget) widget;
	Widget		w;
	int			y;
	int			i;

	if (set_size) {
		for (i = 0; i < (int) CP.num_children; i++) {
			w = CP.children [i];
			if (!w -> core.managed)
				continue;
			/* largeur du plus large */
			if (C.width < w -> core.width)	C.width = w -> core.width;
			/* hauteur du dernier */
			C.height = w -> core.height;
		}
	}
	if (W.rolled) {
		if (W.top)
			XtConfigureWidget (W.top, 0, 0, C.width, W.top -> core.height, 0);
		if (W.bot)
			XtConfigureWidget (W.bot, 0, W.top -> core.height + 2, C.width, C.height - W.top -> core.height - 2, 0);
	} else {
		for (i = (int) CP.num_children; --i >= 0; ) {
			w = CP.children [i];
			if (w -> core.managed) {
				W.bot = w;
				XtConfigureWidget (w, 0, 0, C.width, C.height, 0);
				break;
			}
		}
		for (y = 0; --i >= 0; ) {
			w = CP.children [i];
			if (!w -> core.managed)
				break;

			W.top = w;
			y -= w -> core.height + 2;	/* 2 pixels de sparation */
			XtConfigureWidget (w, 0, y, C.width, w -> core.height, 0);
		}
	}
}

void ClipRoll (Widget widget, Boolean show)
{
#define	offset	4
	ClipWidget	cw = (ClipWidget) widget;

	if (CP.num_children < 2 || W.rolled == show)
		return;

	W.rolled = show;
	if (show) {
		while (W.top -> core.y + offset < 0) {
			XtMoveWidget (W.top, 0, W.top -> core.y + offset);
			XtMoveWidget (W.bot, 0, W.bot -> core.y + offset);
			wait_for_expose (W.top);
		}
		XtMoveWidget (W.bot, 0, W.bot -> core.y - W.top -> core.y);
		XtMoveWidget (W.top, 0, 0);
	} else {
		XtConfigureWidget (W.bot, 0, W.bot -> core.y, C.width, C.height, 0);
		while (W.bot -> core.y - offset > 0) {
			XtMoveWidget (W.top, 0, W.top -> core.y - offset);
			XtMoveWidget (W.bot, 0, W.bot -> core.y - offset);
			wait_for_expose (W.bot);
		}
		XtMoveWidget (W.top, 0, W.top -> core.y - W.bot -> core.y);
		XtMoveWidget (W.bot, 0, 0);
	}
	Layout (widget, False);
}

static void ClipResize (Widget widget)
{
	Layout (widget, False);
}

static void InsertChild (Widget w)
{
	ClipWidget				cw = (ClipWidget) XtParent (w);
	CompositeWidgetClass	superclass = (CompositeWidgetClass) compositeWidgetClass;
    
	if (CP.num_children < 2) {
		(*superclass->composite_class.insert_child) (w);
		return;
	}
	XtWarning ("Roller: No more than two widgets in window\n");
}

static void ClipChangeManaged (Widget widget)
{
	Layout (widget, True);
}

static void ClipInitialize (Widget request, Widget cw, Arg* args, Cardinal* nargs)
{
	if (C.width == 0)	C.width = 10;
	if (C.height == 0)	C.height = 10;
	W.rolled = False;
	W.top = 0;
	W.bot = 0;
}

static XtGeometryResult ClipGeometryManager (Widget ch, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	reply -> request_mode = 0;
	return XtGeometryYes;
}
