/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <Xpa/Primitive.h>
#include <Xpa/PrimitiveP.h>
#include <Xpa/ComboP.h>
#include <Xpa/Combo.h>
  
#define W	w -> combo
#define P	w -> primitive
#define C	w -> core

static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Redisplay (Widget, XEvent*, Region);

/*  Resource definitions for Combo */

static XtResource resources[] =
{
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct*),
	  XtOffset (ComboWidget, combo.fp), XtRString, "fixed"
	},
	{ XtNlabel, XtCLabel, XtRString, sizeof (char*),
	  XtOffset (ComboWidget, combo.label), XtRImmediate, NULL
	},
	{ XtNarmColor, XtCArmColor, XtRPixel, sizeof (Pixel),
	  XtOffset (ComboWidget, combo.arm_color),
	  XtRCallProc, (XtPointer) _GetArmColor
	},
};

static void		Arm (Widget, XEvent*, String*, Cardinal*);
static void		Disarm (Widget, XEvent*, String*, Cardinal*);

static char defaultTranslations [] =
	"<Btn1Down>:	Arm()\n"
	"<Btn1Up>:		Disarm()";

static XtActionsRec actionsList [] =
{
	{"Arm",	Arm},
	{"Disarm",	Disarm},
};

ComboClassRec comboClassRec = {
	/* CoreCLassPart */
	{
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"Combo",						/* class_name */
		sizeof (ComboRec),				/* widget_size */
		NULL,							/* class_initialize */
		NULL,							/* class_part_init */
		FALSE,							/* class_inited */
		Initialize,						/* initialize */
		NULL,							/* initialize_hook */
		XtInheritRealize,				/* realize */
		actionsList,					/* actions */
		XtNumber (actionsList),			/* num_actions */
		resources,						/* resources */
		XtNumber (resources),			/* num_resources */
		NULLQUARK,						/* xrm_class */
		False,							/* compress_motion */
		XtExposeCompressMaximal,		/* compress_exposure */
		False,							/* compress_enterlv */
		False,							/* visible_interest */
		NULL,							/* destroy */
		NULL,							/* resize */
		Redisplay,						/* expose */
		NULL,							/* set_values */
		NULL,							/* set_values_hook */
		XtInheritSetValuesAlmost,		/* set_values_almost */
		NULL,							/* get_values_hook */
		NULL,							/* accept_focus */
		XtVersion,						/* version */
		NULL,							/* callback_private */
		defaultTranslations,			/* tm_table */
		NULL,							/* query_geometry */
	},
	/* PrimitiveCLassPart */
	{
		NULL,
	},
	/* ComboClassPart */
	{
		NULL,
	},
};

WidgetClass comboWidgetClass = (WidgetClass)&comboClassRec;

/*
 *	le GC pour le label.
 */

static void get_labelGC (ComboWidget w)
{
	XGCValues		values;
	unsigned long	valuemask;
	Screen*			screen = XtScreen ((Widget)w);
	Display*		dpy = XtDisplay ((Widget)w);

	valuemask = GCForeground | GCFont;
	values.foreground = P.foreground;
	values.font = W.fp -> fid;

	W.label_GC = XtGetGC ((Widget) w, valuemask, &values);
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	ComboWidget	w = (ComboWidget) new;

	get_labelGC (w);
	W.left_set = False;
	W.right_set = False;
	if (C.width == 0 || C.height == 0) {
		int	fh = W.fp -> ascent + W.fp -> descent;
		int	len = W.label != 0 ? strlen (W.label) : 0;

		C.width = 2 + (fh + 2) +
				  XTextWidth (W.fp, W.label, len) + (fh + 2) + 2;
		C.height = fh + 2 + 2;
	}
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* e, Region region)
{
	ComboWidget	w = (ComboWidget) widget;
	Display*	dpy = XtDisplay (widget);
	Window		window = XtWindow (widget);
	int			fh = W.fp -> ascent + W.fp -> descent;
	int			len = W.label != 0 ? strlen (W.label) : 0;

	if (P.shadow_thickness > 0) {
		if (region != 0) {
			XSetRegion (dpy, P.top_GC, region);
			XSetRegion (dpy, P.bot_GC, region);
		}
		/* la flche de gauche */
		if (W.left_set) {
			XSetForeground (dpy, W.label_GC, W.arm_color);
			_DrawArrow (dpy, window,
						P.bot_GC, P.top_GC, W.label_GC,
						2, 2, fh, fh,
						P.shadow_thickness, XpARROW_LEFT);
		} else {
			XSetForeground (dpy, W.label_GC, C.background_pixel);
			_DrawArrow (dpy, window,
						P.top_GC, P.bot_GC, W.label_GC,
						2, 2, fh, fh,
						P.shadow_thickness, XpARROW_LEFT);
		}
		/* la flche de droite */
		if (W.right_set) {
			XSetForeground (dpy, W.label_GC, W.arm_color);
			_DrawArrow (dpy, window,
						P.bot_GC, P.top_GC, W.label_GC,
						2 + XTextWidth (W.fp, W.label, len) + 2 + fh + 2, 2, fh, fh,
						P.shadow_thickness, XpARROW_RIGHT);
		} else {
			XSetForeground (dpy, W.label_GC, C.background_pixel);
			_DrawArrow (dpy, window,
						P.top_GC, P.bot_GC, W.label_GC,
						2 + XTextWidth (W.fp, W.label, len) + 2 + fh + 2, 2, fh, fh,
						P.shadow_thickness, XpARROW_RIGHT);
		}
		/* virer les zones de clipping */
		XSetClipMask (dpy, P.top_GC, None);
		XSetClipMask (dpy, P.bot_GC, None);
	}
	/* le label */
	XSetForeground (dpy, W.label_GC, P.foreground);
	if (W.label != 0) {
		if (region != 0)
			XSetRegion (dpy, W.label_GC, region);
		XDrawString (dpy, window, W.label_GC,
					 2 + 2 + fh, 2 + W.fp -> ascent,
					 W.label, len);
		XSetClipMask (dpy, W.label_GC, None);
	}
}

static void Arm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ComboWidget		w = (ComboWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				mx = e -> x;
	int				my = e -> y;
	int				fh = W.fp -> ascent + W.fp -> descent;
	int				ll = XTextWidth (W.fp, W.label, strlen (W.label));

	if (2 < mx && mx < 2 + fh && 2 < my && my < 2 + fh)
		W.left_set = True;
	else
	if (2 + ll + 2 + fh < mx && mx < 2 + ll + 2 + fh + fh && 2 < my && my < 2 + fh)
		W.right_set = True;
	else
		return;		/* pas cliqu dans une flche */

	Redisplay (widget, NULL, 0);
	XtCallCallbackList (widget, P.select, (XtPointer) (long) W.left_set);
}

static void Disarm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ComboWidget	w = (ComboWidget) widget;

	W.left_set = False;
	W.right_set = False;
	Redisplay (widget, NULL, 0);
}
