/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <Xpa/Container.h>
#include <Xpa/ContainerP.h>


#define	C	pw -> core
#define	W	pw -> container

static void	Initialize (Widget, Widget, ArgList, Cardinal*);
static void	Redisplay (Widget, XEvent*, Region);
static void	Realize (Widget, XtValueMask*, XSetWindowAttributes*);

static XtResource resources [] = {
	{ XtNcursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(ContainerWidget, container.cursor), XtRString, "top_left_arrow"
	},
	{ XtNtopShadowColor, XtCBackground, XtRPixel, sizeof (Pixel),
	  XtOffset(ContainerWidget, container.top_pixel), XtRCallProc,
	  (XtPointer) _GetTopShadowColor
	},
	{ XtNbottomShadowColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset(ContainerWidget, container.bot_pixel), XtRCallProc,
	  (XtPointer) _GetBottomShadowColor
	},
	{ XtNshadowThickness, XtCShadowThickness, XtRInt, sizeof (int),
	  XtOffset(ContainerWidget, container.shadow_thickness), XtRImmediate, (XtPointer) 2
	},
	{ XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset(ContainerWidget, container.foreground), XtRString, "black"
	},
	{ XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	  XtOffset(Widget, core.background_pixel), XtRString, "grey"
	},
	{ XtNuserData, XtCUserData, XtRPointer, sizeof (XtPointer),
	  XtOffset(ContainerWidget, container.user_data), XtRImmediate, (XtPointer) NULL
	},
};

ContainerClassRec containerClassRec = {
  /* core_class fields */
  {
    /* superclass         */    (WidgetClass) &constraintClassRec,
    /* class_name         */    "Container",
    /* widget_size        */    sizeof(ContainerRec),
    /* class_initialize   */    NULL,
    /* class_partinit     */    NULL,
    /* class_inited       */	False,
    /* initialize         */    Initialize,
    /* Init hook	  */    NULL,
    /* realize            */    Realize,
    /* actions		  */	NULL,
    /* num_actions	  */	0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber (resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	False,
    /* compress_exposure  */	True,
    /* compress_enterleave*/	False,
    /* visible_interest   */    False,
    /* destroy            */    NULL,
    /* resize             */    NULL,
    /* expose             */    Redisplay,
    /* set_values         */    NULL,
    /* set values hook    */    NULL,
    /* set values almost  */    XtInheritSetValuesAlmost,
    /* get values hook    */    NULL,
    /* accept_focus       */    NULL,
    /* Version            */    XtVersion,
    /* PRIVATE cb list    */    NULL,
    /* TM Table           */    "",
    /* query_geom         */    NULL,
    /* display acccelerator */	XtInheritDisplayAccelerator,
    /* extension */	NULL,
  },{
/* composite_class fields */
    /* geometry_manager   */    NULL,
    /* change_managed     */    XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,	/* Inherit from superclass */
    /* delete_child	  */	XtInheritDeleteChild,	/* Inherit from superclass */
    /* extension */		NULL,
  },{
/* constraint_class fields */
    /*  constraint resource set     */	NULL,
    /*  num_resources               */	0,
    /*  size of the constraint data */	0,
    /*  contraint initilize proc    */	(XtInitProc) NULL,
    /*  contraint destroy proc      */	(XtWidgetProc) NULL,
    /*  contraint set values proc   */	(XtSetValuesFunc) NULL,
  },{
/* container class fields */     
     /* extension */               NULL
 }	
};

WidgetClass containerWidgetClass = (WidgetClass)&containerClassRec;

/*
 *	La procdure de cration de la fentre X.
 */

static void Realize	(Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	ContainerWidget	pw = (ContainerWidget) widget;

	*value_mask |= CWCursor;
	attributes -> cursor = W.cursor;
	XtCreateWindow (widget, InputOutput, CopyFromParent, *value_mask, attributes);
}

/*
 *	Les GC pour les ombres et le grip.
 */

static void getGC (ContainerWidget pw)
{
	XGCValues		values;
	unsigned long	valuemask;
	/* Ombres du haut */
	valuemask = GCForeground /* | GCBackground */ ;
	values.foreground = W.top_pixel;
/*	values.background = W.foreground; */
	W.top_GC = XtGetGC ((Widget)pw, valuemask, &values);
	/* Ombres du bas */
	values.foreground = W.bot_pixel;
	W.bot_GC = XtGetGC ((Widget)pw, valuemask, &values);
}

static void Initialize (Widget request, Widget w, ArgList al, Cardinal* n)
{
	ContainerWidget	pw = (ContainerWidget)w;

	if (request->core.height == 0)	C.height = 300;
	if (request->core.width==0)	C.width = 400;

	getGC (pw);
}

/*
 *	Afficher les separateurs et les grips.
 */

static void Redisplay (Widget w, XEvent* event, Region region)
{
	ContainerWidget	pw = (ContainerWidget) w;
	Display*	dpy = XtDisplay (w);
	Window		win = XtWindow (w);

	XSetRegion (dpy, W.top_GC, region);
	XSetRegion (dpy, W.bot_GC, region);
	if (W.shadow_thickness > 0)
		_DrawShadow (dpy, win,
			W.top_GC, W.bot_GC, W.shadow_thickness,
			(int)C.width, (int)C.height, 0, 0);
	XSetClipMask (dpy, W.top_GC, None);
	XSetClipMask (dpy, W.bot_GC, None);
}
