/*****************************************************************

  Copyright (c) 1989, 1999 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/xpm.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "DialP.h"
#include "Dial.h"
#include "dials4.xpm"
  
#define W	w -> dial
#define P	w -> primitive
#define C	w -> core

static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Destroy (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

static void		Move (Widget, XEvent*, String*, Cardinal*);
static void		Key (Widget, XEvent*, String*, Cardinal*);

static char defaultTranslations [] = "\
	<Key>Up:		Key()\n\
	<Key>Down:		Key()\n\
	<Btn1Down>:		Move()\n\
	<Btn1Motion>:	Move()\n\
	<Btn1Up>:		Move()";

static XtActionsRec actionsList [] =
{
	{"Move",	Move},
	{"Key",	Key},
};

static XtResource resources [] =
{     
	{ XtNposition, XtCPosition, XtRShort, sizeof (int),
	  XtOffset (DialWidget, dial.position), XtRImmediate, (XtPointer) 0
	},
	{ XtNminimum, XtCMinimum, XtRShort, sizeof (int),
	  XtOffset (DialWidget, dial.minimum), XtRImmediate, (XtPointer) 0
	},
	{ XtNmaximum, XtCMaximum, XtRShort, sizeof (int),
	  XtOffset (DialWidget, dial.maximum), XtRImmediate, (XtPointer) 102
	},
};

DialClassRec dialClassRec = {
	/* CoreCLassPart */
	{
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"Dial",							/* class_name */
		sizeof (DialRec),				/* widget_size */
		NULL,							/* class_initialize */
		NULL,							/* class_part_init */
		FALSE,							/* class_inited */
		Initialize,						/* initialize */
		NULL,							/* initialize_hook */
		XtInheritRealize,				/* realize */
		actionsList,					/* actions */
		XtNumber (actionsList),			/* num_actions */
		resources,						/* resources */
		XtNumber (resources),			/* num_resources */
		NULLQUARK,						/* xrm_class */
		False,							/* compress_motion */
		True,							/* compress_exposure */
		True,							/* compress_enterlv */
		False,							/* visible_interest */
		Destroy,						/* destroy */
		NULL,							/* resize */
		Redisplay,						/* expose */
		SetValues,						/* set_values */
		NULL,							/* set_values_hook */
		XtInheritSetValuesAlmost,		/* set_values_almost */
		NULL,							/* get_values_hook */
		NULL,							/* accept_focus */
		XtVersion,						/* version */
		NULL,							/* callback_private */
		defaultTranslations,			/* tm_table */
		NULL,							/* query_geometry */
	},
	/* PrimitiveCLassPart */
	{
		NULL,
	},
	/* DialClassPart */
	{
		None,
	},
};

WidgetClass dialWidgetClass = (WidgetClass)&dialClassRec;

/*
 *	Fonction d'initialisation de chaque instance.
 */

static XpmAttributes	attr;

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	DialWidget	w = (DialWidget) new;

	if (dialClassRec.dial_class.pixmaps == None) {
		if (XpmSuccess != XpmCreatePixmapFromData (XtDisplay (new),
									RootWindowOfScreen (XtScreen (new)),
									dials4,& dialClassRec.dial_class.pixmaps,
									NULL, &attr))
			fprintf (stderr, "Cannot create dial ???\n");
	}

	C.width = attr.width / 6 + 2 * P.shadow_thickness;
	C.height = attr.height + 2 * P.shadow_thickness;
	W.dialGC = XtGetGC ((Widget) w, 0, NULL);
}

/*
 *	Dessiner la roue.
 */

static void show_wheel (DialWidget w)
{
	Pixmap			pixmap = dialClassRec.dial_class.pixmaps;
	Display*		dpy = XtDisplay ((Widget)w);
	Window			window = XtWindow ((Widget)w); 
	int				np;

	np = (((W.position - W.minimum) * 60) / (W.maximum - W.minimum)) % 6;
	XCopyArea (dpy, pixmap, window, W.dialGC,
				np * attr.width / 6, 0, attr.width / 6, attr.height,
				P.shadow_thickness, P.shadow_thickness);
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* e, Region region)
{
	DialWidget		w = (DialWidget) widget;
	int				which_one;
	Display*		dpy = XtDisplay (widget);
	Window			window = XtWindow (widget); 

	XSetRegion (dpy, W.dialGC, region);
	if (P.shadow_thickness > 0) {
		XSetRegion (dpy, P.top_GC, region);
		XSetRegion (dpy, P.bot_GC, region);
		_DrawShadow (dpy, window,
						P.top_GC, P.bot_GC, P.shadow_thickness,
						C.width, C.height, 0, 0);
		XSetClipMask (dpy, P.top_GC, None);
		XSetClipMask (dpy, P.bot_GC, None);
	}
	show_wheel (w);	/* dessiner la roue */
	XSetClipMask (dpy, W.dialGC, None);
}

/*
 *	SetValues.
 */

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	DialWidget	cu = (DialWidget) current;
	DialWidget	w = (DialWidget) new;
	Boolean 	do_redisplay = False;

	if (cu -> primitive.foreground != P.foreground) {
		do_redisplay = True;
	}

	return do_redisplay;
}

/*
 *	La procedure de destruction d'une instance.
 */

static void Destroy (Widget widget)
{
	DialWidget	w = (DialWidget) widget;

	XtDestroyGC (W.dialGC);
}

/*
 *	Deplacement de la souris
 */


static void Key (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	DialWidget		w = (DialWidget) widget;
	char			c;
	KeySym			ks;

	(void) XLookupString ((XKeyEvent*)event, &c, 1, &ks, 0);
	switch (ks) {
		case XK_Up:
			W.position += (W.maximum - W.minimum) / 60 + 1;
			if (W.position > W.maximum)
				W.position = W.maximum;
			break;
		case XK_Down:
			W.position -= (W.maximum - W.minimum) / 60 + 1;
			if (W.position < W.minimum)
				W.position = W.minimum;
			break;
		default:
			return;
	}
	show_wheel (w);
}

static void Move (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	DialWidget		w = (DialWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				np;
	static int		start_y;
	static int		old_pos;

	switch (e -> type) {
	case ButtonPress:
		start_y = e -> y;
		old_pos = W.position;
		return;

	case MotionNotify:
		np = (start_y  - e -> y) / 2;
		W.position = np + old_pos;
		if (W.position > W.maximum)
			W.position = W.maximum;
		else if (W.position < W.minimum)
			W.position = W.minimum;
		show_wheel (w);
		return;

	case ButtonRelease:
		return;
	}
}
