/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/*************************************************************************
*
*	La boite de dialogue generique.
*
*************************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include "Primitive.h"
#include "PrimitiveP.h"
#include "Container.h"
#include "ContainerP.h"
#include "PushButton.h"
#include "PushButtonP.h"
#include "Wix.h"
#include "Dialog.h"
#include "DialogP.h"

#define MARGINS (w->manager.shadow_thickness)
#define OFF_TITRE 4
#define OFF_TOGGLE -3
#define OFF_VERT 2
#define OFF_HORIZ 2
#define sup(x,y) ((x)>(y) ? (x) : (y))

static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		_resize (DialogWidget, Boolean);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static XtGeometryResult	GeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void		ChangeManaged (Widget);

static XtResource resources [] = {
	{ XtNtext, XtCText, XtRWidgetList, sizeof (Widget*),
	  XtOffset (DialogWidget, dialog.edit_widgets), XtRPointer, NULL
	},
	{ XtNtextCount, XtCTextCount, XtRInt, sizeof (int),
	  XtOffset (DialogWidget, dialog.num_fields), XtRImmediate, NULL
	},
	{ XtNdefaultToggle, XtCDefaultToggle, XtRInt, sizeof (int),
	  XtOffset (DialogWidget, dialog.def_toggle), XtRImmediate, (XtPointer)-1
	},
	{ XtNdefaultButton, XtCDefaultButton, XtRInt, sizeof (int),
	  XtOffset (DialogWidget, dialog.def_button), XtRImmediate, (XtPointer)-1
	},
	{ XtNautoOk, XtCAutoOk, XtRBoolean, sizeof (Boolean),
	  XtOffset (DialogWidget, dialog.auto_ok), XtRImmediate, (XtPointer)False
	},
	{ XtNtoggle, XtCToggle, XtRStringArray, sizeof (String*),
	  XtOffset (DialogWidget, dialog.toggle), XtRPointer, NULL
	},
	{ XtNfields, XtCFields, XtRStringArray, sizeof (String*),
	  XtOffset (DialogWidget, dialog.fields), XtRPointer, NULL
	},
	{ XtNokLabel, XtCLabel, XtRString, sizeof (String),
	  XtOffset (DialogWidget, dialog.ok_label), XtRPointer, NULL
	},
	{ XtNapplyLabel, XtCLabel, XtRString, sizeof (String),
	  XtOffset (DialogWidget, dialog.apply_label), XtRPointer, NULL
	},
	{ XtNcancelLabel, XtCLabel, XtRString, sizeof (String),
	  XtOffset (DialogWidget, dialog.cancel_label), XtRPointer, NULL
	},
	{ XtNokCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (DialogWidget, dialog.ok_callback), XtRPointer, NULL
	},
	{ XtNapplyCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (DialogWidget, dialog.apply_callback), XtRPointer, NULL
	},
	{ XtNcancelCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (DialogWidget, dialog.cancel_callback), XtRPointer, NULL
	},
	{ XtNtitle, XtCTitle, XtRString, sizeof (String),
	  XtOffset (DialogWidget, dialog.title), XtRImmediate, "Title"
	},
};

DialogClassRec dialogClassRec = {
  /* core_class fields */
  {
    /* superclass         */    (WidgetClass) &containerClassRec,
    /* class_name         */    "Dialog",
    /* widget_size        */    sizeof(DialogRec),
    /* class_initialize   */    NULL,
    /* class_partinit     */    NULL,
    /* class_inited       */	False,
    /* initialize         */    Initialize,
    /* Init hook	  */    NULL,
    /* realize            */    XtInheritRealize,
    /* actions		  */	NULL,
    /* num_actions	  */	0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber (resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	False,
    /* compress_exposure  */	True,
    /* compress_enterleave*/	False,
    /* visible_interest   */    False,
    /* destroy            */    NULL,
    /* resize             */    Resize,
    /* expose             */    Redisplay,
    /* set_values         */    SetValues,
    /* set values hook    */    NULL,
    /* set values almost  */    XtInheritSetValuesAlmost,
    /* get values hook    */    NULL,
    /* accept_focus       */    NULL,
    /* Version            */    XtVersion,
    /* PRIVATE cb list    */    NULL,
    /* TM Table           */    "",
    /* dialog_geom        */    NULL,
  },{
/* composite_class fields */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension */		NULL,
  },{
/* Constraint class Init */
    NULL,
    0,
    0,
    NULL,
    NULL,
    NULL
  },{
/* manager_class fields */
	NULL,
  },{
/* Dialog class - none */     
   0
 }	
};

WidgetClass dialogWidgetClass = (WidgetClass)&dialogClassRec;

/*
 *	La calback de confirmation.
 */

static void okCallback (Widget button, XtPointer client, XtPointer call)
{
	DialogWidget	w = (DialogWidget) client;
	long				n;

	/* recuperer le bouton appuye */
	for (n = 0; n < W.num_toggle; n++)
		if (W.toggle_widgets [n] == W.active)
			break;
	XtCallCallbackList ((Widget) w, W.ok_callback, (XtPointer)n);
}

/*
 *	Pour rendre les boutons exclusifs.
 */

static void exclusive (Widget b, XtPointer client, XtPointer call)
{
	DialogWidget	w = (DialogWidget) client;

	if (W.active == b) {
		XtVaSetValues (W.active,
			XtNset, True,
			NULL);
		return;
	}
	if (W.active)
		XtVaSetValues (W.active,
			XtNset, False,
			NULL);
	W.active = b;
	if (W.auto_ok)
		okCallback (b, w, NULL);
}

static void cancelCallback (Widget button, XtPointer client, XtPointer call)
{
	DialogWidget	w = (DialogWidget) client;

	XtCallCallbackList ((Widget) w, W.cancel_callback, NULL);
}

static void applyCallback (Widget button, XtPointer client, XtPointer call)
{
	DialogWidget	w = (DialogWidget) client;

	XtCallCallbackList ((Widget) w, W.apply_callback, NULL);
}

/*
 *	Gere le clavier.
 */

static void keyboard (Widget wix, XtPointer client, XtPointer call)
{
	DialogWidget	w = (DialogWidget) client;
	XKeyEvent*		e = (XKeyEvent*) call;
	int				n;
	char			c;
	KeySym			ks;

	if (XLookupString (e, &c, 1, &ks, 0)) {
		if (c == '\r') {
			switch (W.def_button) {
				case OK_BUTTON:
					FakeClick (OK);
					/* okCallback (NULL, w, NULL); */
					return;
				case CANCEL_BUTTON:
					FakeClick (CAN);
					/* cancelCallback (NULL, w, NULL); */
					return;
				case APPLY_BUTTON:
					FakeClick (APPLY);
					/* applyCallback (NULL, w, NULL); */
					return;
			}
		}
		if (c == 033) {
			FakeClick (CAN);
			/* cancelCallback (NULL, w, NULL); */
			return;
		}
	} else {
		if (ks == XK_Down) {
			/* recuperer le texte suivant */
			for (n = 0; n < W.num_fields; )
				if (W.edit_widgets [n++] == wix) {
					if (n == W.num_fields)
						wix = W.edit_widgets [0];
					else
						wix = W.edit_widgets [n];
					break;
				}
			XtSetKeyboardFocus ((Widget) w, wix);
		} else
		if (ks == XK_Up) {
			/* recuperer le texte precedent */
			for (n = 0; n < W.num_fields; n++)
				if (W.edit_widgets [n] == wix) {
					if (n == 0)
						wix = W.edit_widgets [W.num_fields-1];
					else
						wix = W.edit_widgets [n-1];
					break;
				}
			XtSetKeyboardFocus ((Widget) w, wix);
		}
	}
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	static XtTranslations	keyTranslations = NULL;
	DialogWidget	w = (DialogWidget) new;
	XFontStruct*	fp;
	Boolean			oblong;
	int				n;
	int				l;

	if (request -> core.width == 0)		C.width = DEFAULT_WIDTH;
	if (request -> core.height == 0)	C.height = DEFAULT_HEIGHT;

	if (!keyTranslations)
		keyTranslations = XtParseTranslationTable (
			"<Key>Up:	activate()\n\
			 <Key>Escape:	activate()\n\
			 <Key>Down:	activate()");

	OK = XtVaCreateWidget ("ok", pushButtonWidgetClass, new,
					XtNlabel, W.ok_label,
					XtNborderWidth,	0,
					NULL);
	XtAddCallback (OK, XtNrelease, okCallback, (XtPointer) w);
	fp = ((PushButtonWidget)OK) -> pushbutton.fp;

	/* Compter et creer les toggle buttons */
	if (W.toggle) {
		for (n = 0; W.toggle [n] != NULL; n++)
			;
		W.num_toggle = n;
		W.toggle_widgets = (Widget*) XtMalloc (n * sizeof (Widget));
		W.max_toggle = 0;
		for (n = 0; W.toggle [n] != NULL; n++) {
			W.toggle_widgets [n] = XtVaCreateWidget (W.toggle [n], pushButtonWidgetClass, new,
						XtNborderWidth,	0,
						XtNmargin,			0,
						XtNtoggle,			True,
						NULL);
			XtAddCallback (W.toggle_widgets [n], XtNrelease, exclusive, (XtPointer) w);
			l = (W.toggle_widgets [n])->core.width;
			if (W.max_toggle < l)
				W.max_toggle = l;
		}
		if (W.def_toggle >= 0 && W.def_toggle < W.num_toggle) {
			W.active = W.toggle_widgets [W.def_toggle];
			XtVaSetValues (W.active, XtNset, True,	NULL);
		} else
			W.active = NULL;
	} else {
		W.num_toggle = 0;
		W.max_toggle = 0;
	}
	/* Compter et creer les champs de saisie */
	if (W.fields) {
		for (n = 0; W.fields [n] != NULL; n++)
			;
		W.num_fields = n;
		W.max_label = 0;
		W.edit_widgets = (Widget*) XtMalloc (n * sizeof (Widget));
		for (n = 0; W.fields [n] != NULL; n++) {
			W.edit_widgets [n] = XtVaCreateWidget ("saisie",wixWidgetClass, new,
						XtNmonoLine, True,
						XtNkillStack, False,
						XtNborderWidth,	0,
						NULL);
			XtOverrideTranslations (W.edit_widgets [n], keyTranslations);
			XtAddCallback (W.edit_widgets [n], XtNactivateCallback, keyboard, (XtPointer) w);
			l = XTextWidth (fp, W.fields [n], strlen (W.fields[n]));
			if (W.max_label < l)
				W.max_label = l;
		}
	} else {
		W.max_label = 0;
		W.num_fields = 0;
	}

	CAN = XtVaCreateWidget ("cancel", pushButtonWidgetClass, new,
					XtNlabel, W.cancel_label,
					XtNborderWidth,	0,
					NULL);
	XtAddCallback (CAN, XtNrelease, cancelCallback, (XtPointer) w);

	W.apply_button = XtVaCreateWidget ("Apply", pushButtonWidgetClass, new,
				XtNlabel, W.apply_label,
				XtNborderWidth,	0,
				NULL);
	XtAddCallback (W.apply_button, XtNrelease, applyCallback, (XtPointer) w);

	W.title_height = 0;
	W.title_width = 0;

	if (W.title) {
		W.title_width = XTextWidth (fp, W.title, strlen (W.title));
		W.title_height = fp -> ascent + fp -> descent;
	}

	if (W.toggle) XtManageChildren (W.toggle_widgets, W.num_toggle);
	if (W.fields) {
		XtManageChildren (W.edit_widgets, W.num_fields);
		XtSetKeyboardFocus (new, W.edit_widgets [0]);
	}
	XtVaGetValues (OK, XtNoblong, &oblong, NULL);
	switch (W.def_button) {
		case OK_BUTTON:
			XtVaSetValues (OK,
				XtNshowAsDefault, oblong ? 3 : 3 * w->manager.shadow_thickness,
				NULL);
				break;

		case CANCEL_BUTTON:
			XtVaSetValues (CAN,
				XtNshowAsDefault, oblong ? 3 : 3 * w->manager.shadow_thickness,
				NULL);
				break;

		case APPLY_BUTTON:
			XtVaSetValues (APPLY,
				XtNshowAsDefault, oblong ? 3 : 3 * w->manager.shadow_thickness,
				NULL);
	}
	XtManageChild (OK);
	XtManageChild (CAN);
	_resize (w, True);
}

static	Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	DialogWidget	cu = (DialogWidget) current;
	DialogWidget	w = (DialogWidget) new;
	Boolean			do_redisplay = False;

	W.num_fields = cu -> dialog.num_fields;
	W.edit_widgets = cu -> dialog.edit_widgets;

	switch (W.def_toggle) {
		case -2:
			break;
		case -1:
			if (W.active)
				XtVaSetValues (W.active, XtNset, False, NULL);
			W.active = NULL;
			break;
		default:
			if (W.def_toggle >= 0 && W.def_toggle < W.num_toggle) {
				if (W.active == W.toggle_widgets [W.def_toggle])
					break;
				if (W.active)
					XtVaSetValues (W.active, XtNset, False, NULL);
				W.active = W.toggle_widgets [W.def_toggle];
				XtVaSetValues (W.active, XtNset, True, NULL);
			}
	}

	if (cu -> dialog.title != W.title) {
		W.title_width = XTextWidth (((PushButtonWidget)OK) -> pushbutton.fp, W.title, strlen (W.title));
		_resize (w, True);
		do_redisplay = True;
	}

	if (cu -> dialog.apply_label != W.apply_label) {
		XtVaSetValues (W.apply_button, XtNlabel, W.apply_label, NULL);
		do_redisplay = True;
	}

	if (cu -> dialog.ok_label != W.ok_label) {
		XtVaSetValues (W.ok_button, XtNlabel, W.ok_label, NULL);
		do_redisplay = True;
	}

	if (cu -> dialog.cancel_label != W.cancel_label) {
		XtVaSetValues (W.cancel_button, XtNlabel, W.cancel_label, NULL);
		do_redisplay = True;
	}

	return do_redisplay;
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	DialogWidget	w = (DialogWidget) widget;
	Display*		dpy = XtDisplay (widget);
	Window			win = XtWindow (widget);
	PushButtonWidget ok = (PushButtonWidget)OK;
	XFontStruct*	fp = ok -> pushbutton.fp;
	int				st = ok -> primitive.shadow_thickness;
	int				y;
	int				n;

	(*containerClassRec.core_class.expose) (widget, event, region);

	/* afficher le titre */
	if (W.title) {
		XDrawString (dpy, win, ok -> pushbutton.label_GC,
			(C.width - W.title_width) / 2 ,
			MARGINS + OFF_TITRE + fp -> ascent,
			W.title, strlen (W.title));
	}
	/* les labels */
	y = MARGINS + OFF_TITRE + W.title_height + OFF_TITRE;
	/* sauter les toggle buttons */
	if (W.num_toggle > 0) {
		for (n = 0; n < W.num_toggle; n++) {
			y += (W.toggle_widgets [n]) -> core.height + OFF_TOGGLE;
		}
		y = y - OFF_TOGGLE + OFF_VERT;
	}
	/* afficher les labels */
	if (W.num_fields > 0) {
		int	ib, st;

		XtVaGetValues (W.edit_widgets [0],
			XtNinnerBorder,	&ib,
			XtNshadowThickness,	&st,
			NULL);

		for (n = 0; n < W.num_fields; n++) {
			XDrawString (dpy, win, ok -> pushbutton.label_GC,
							MARGINS + OFF_HORIZ, y + fp -> ascent + ib + st,
							W.fields [n], strlen (W.fields [n]));
			y += (W.edit_widgets [n]) -> core.height + OFF_VERT;
		}
	}
	/* afficher le separateur */
	if (st > 0)
		_DrawShadow (dpy, win, ok -> primitive.bot_GC, ok -> primitive.top_GC, st,
						C.width - 2*(MARGINS+OFF_HORIZ), 2,
						MARGINS+OFF_HORIZ, y - 1);
}

static void _resize (DialogWidget w, Boolean flag)
{
	Dimension	width;
	int			y;
	int			n;
	Dimension	def_shadow = 0;
	Widget		d_button = NULL;

	if (flag) {
		C.height = MARGINS + OFF_TITRE + W.title_height + OFF_TITRE
			   + (W.num_toggle > 0
			   	? OFF_TITRE + W.num_toggle * (OFF_TOGGLE + W.toggle_widgets [0] -> core.height)
				: 0)
			   + (W.num_fields > 0
			   	? OFF_TITRE + W.num_fields * (OFF_VERT + W.edit_widgets [0] -> core.height)
				: 0)
	 /* separateur */  + 2
	 		   + OFF_VERT + (OK) -> core.height
			   + OFF_TITRE + MARGINS;
		y = W.num_fields > 0
			  ? (W.max_label + W.edit_widgets [0] -> core.width)
			  : 0;
		width = MARGINS + OFF_HORIZ
			+ sup (W.title_width,
				sup ((OK)->core.width + (CAN)->core.width + (APPLY)->core.width,
				  sup (W.max_toggle, y)))
			+ MARGINS + OFF_HORIZ;
		if (C.width < width)
			C.width = width;
	}

	y = MARGINS + OFF_TITRE + W.title_height + OFF_TITRE;
	/* placer les toggle buttons */
	if (W.num_toggle > 0) {
		for (n = 0; n < W.num_toggle; n++) {
			XtMoveWidget (W.toggle_widgets [n], OFF_HORIZ, y);
			y += (W.toggle_widgets [n]) -> core.height + OFF_TOGGLE;
		}
		y = y - OFF_TOGGLE + OFF_VERT;
	}
	/* placer les zones de saisie */
	if (W.num_fields > 0) {
		for (n = 0; n < W.num_fields; n++) {
			XtConfigureWidget (W.edit_widgets [n],
				MARGINS + OFF_HORIZ + W.max_label + OFF_HORIZ, y,
				C.width - (MARGINS + OFF_HORIZ + MARGINS + OFF_HORIZ + W.max_label + OFF_HORIZ),
				(W.edit_widgets [n]) -> core.height,
				0);
			y += (W.edit_widgets [n]) -> core.height + OFF_VERT;
		}
	}
	/* le separateur */
	y += 2 + OFF_VERT;
	/* les boutons */
	if (W.def_button >= 0) {
		d_button = DialogGetChild ((Widget) w, W.def_button);
		def_shadow = ((PushButtonWidget)d_button) -> pushbutton.def_shadow;
	}
	XtMoveWidget (OK, MARGINS + OFF_HORIZ, OK == d_button ? y : y+def_shadow);
	XtMoveWidget (CAN, C.width - (MARGINS + OFF_HORIZ + (CAN)->core.width), CAN == d_button ? y : y+def_shadow);
	XtMoveWidget (APPLY, (C.width - (APPLY)->core.width)/2, APPLY == d_button ? y : y+def_shadow);
}

static void Resize (Widget widget)
{
	_resize ((DialogWidget) widget, False);
}

static XtGeometryResult GeometryManager (Widget w, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	if ((request -> request_mode & CWX) || (request -> request_mode & CWY))
		return (XtGeometryResult) XtGeometryNo;

	if (request -> request_mode & CWWidth) {
		w -> core.width = request -> width;
		reply -> width = request -> width;
	}
	if (request -> request_mode & CWHeight) {
		w -> core.height = request -> height;
		reply -> height = request -> height;
	}
	reply -> request_mode = request -> request_mode;
	return (XtGeometryResult) XtGeometryYes;
}

static void ChangeManaged (Widget w)
{
	_resize ((DialogWidget) w, True);
}

/*
 *	Les fonctions publiques
 */

Widget DialogGetChild (Widget widget, int n)
{
	DialogWidget	w = (DialogWidget) widget;

	switch (n) {
		case OK_BUTTON:		return OK;
		case CANCEL_BUTTON:	return CAN;
		case APPLY_BUTTON:	return APPLY;
	}
	return NULL;
}
