/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/*************************************************************************
*
*	Le selecteur de fichiers.
*
*************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <unistd.h>
#include <stdlib.h>
#if SYS_DIR
#	include <sys/dir.h>
#else
#	include <dirent.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/xpm.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "Container.h"
#include "ContainerP.h"
#include "Wix.h"
#include "PushButton.h"
#include "PushButtonP.h"
#include "Icon.h"
#include "IconP.h"
#include "Valuator.h"
#include "ScrolledWin.h"
#include "ScrolledWinP.h"
#include "FSelector.h"
#include "FSelectorP.h"

#include "FRegIc.bm"
#include "FSmallIc.bm"
#include "FRegIc.xpm"
#include "FSmallIc.xpm"

static Pixmap	regularPixmap [N_PIXMAP];
static Pixmap	smallPixmap [N_PIXMAP];

#define MARGINS (w->manager.shadow_thickness)
#define DEFAULT_HEIGHT 300
#define DEFAULT_WIDTH 500

#define	C	w -> core
#define	CP	w -> composite
#define	W	w -> fileSelector
#define	IC	w -> fileSelector.fs_icon
#define	CL	w -> fileSelector.clip
#define	WIX	w -> fileSelector.fs_wix
#define	OK	w -> fileSelector.ok_button
#define	CAN	w -> fileSelector.cancel_button
#define	EXP	w -> fileSelector.expand_button
#define	FILTER	w -> fileSelector.filter_button

/*
 *	Fabrique un tableau de strings avec les noms des
 *	fichiers du directory specifie.
 */

static int list_files (char* directory, char* prefix, char* max_prefix, FileSelectorWidget w)
{
	XtWidgetGeometry	request, reply;
	DIR*		dirp;
#if SYS_DIR
	struct direct*	item;
#else
	struct dirent*	item;
#endif
	char		name [128];
	char		tc [256];
	Pixmap		pix;
	Pixmap		prev_pix = (Pixmap) -1;
	register char*	p1;
	register char*	p2;
	int		len;
	int		nb;
	int		type;
	int		prev_type = -1;
	struct stat	b;
	Boolean		small = ((IconWidget)IC) -> icon.small;

	len = strlen (prefix);		/* pour utiliser bcmp */
					/* pour chercher une chaine prefixe */
	if (-1 == access (directory, R_OK | X_OK) || !(dirp = opendir (directory)))
	{
		if (max_prefix)
			*max_prefix = 0;/* prefixe vide */
		return -1;		/* drole de directory! */
	}
	/*
	 *	Faire poireauter.
	 */
	if (XtIsRealized (IC)) {
		XDefineCursor (XtDisplay (w), XtWindow (IC),
				((IconWidget)IC) -> icon.wait_cursor);

		XtSetSensitive ((Widget) w, False);
		XSync (XtDisplay ((Widget) w), False);
	}

	if (max_prefix)
		*max_prefix = 3;	/* prefixe vide */

	nb = 0;	/* au debut */
	(void) readdir (dirp);		/* sauter . */
	for (item = readdir (dirp); item != NULL; item = readdir (dirp)) {
		if (len && bcmp (item -> d_name, prefix, len))
			continue;

		p1 = item -> d_name;
#ifndef NO_CHECK_DIR
		sprintf (tc, "%s/%s", directory, p1);
		stat (tc, &b);	/* est-ce un repertoire */
		if (S_ISDIR (b.st_mode))
			type = F_DIRECTORY;
		else if (b.st_mode & 0111)
			type = F_EXECUTABLE;
		else
#endif
		if (*p1 == '.') {
			if (p1 [1] == '.' && p1 [2] == 0)
				type = F_DIRECTORY;
			else {
				if (W.no_dotfile)
					continue;
				type = F_DOTFILE;
			}
		} else {
			type = F_PLAINFILE;
#if SYS_DIR
			p1 = & item -> d_name [item -> d_namlen-1];
#else
			p1 = & item -> d_name [strlen (item -> d_name)-1];
#endif
			switch (*p1) {
			case 's':
				if (*--p1 == 'p' && *--p1 == '.') {
					type = F_POSTSCRIPT;
					break;
				}
				goto def;
/* .xbm ou .xpm */	case 'm':
				if ((*--p1 == 'b' || *p1 == 'p') && *--p1 == 'x' && *--p1 == '.') {
					type = F_BITMAP;
					break;
				}
				goto def;
			case 'x':
				if (*--p1 == 'e' && *--p1 == 't' && *--p1 == '.') {
					type = F_TEX;
					break;
				}
				goto def;
			case 'i':
				if (*--p1 == 'v' && *--p1 == 'd' && *--p1 == '.') {
					type = F_DVI;
					break;
				}
				goto def;
			case 'Z':
			case 'z':
				if (*--p1 == '.' || *p1 == 'g' ) {
					type = F_COMPRESSED;
					break;
				}
				goto def;
			case 'a':
				if (*--p1 == '.') {
					type = F_ARCHIVE;
					break;
				}
				if (*p1 == 'v' && *--p1 == 'a' && *--p1 == 'j' && *--p1 == '.') {
					type = F_JAVA;
					break;
				}
				goto def;
			case 'C':
				if (*--p1 == '.') {
					type = F_CPLUSPLUS;
					break;
				}
				goto def;
			case 'c':
				if (*--p1 == '.') {
					type = F_CTEXT;
					break;
				}
				if (*p1 == 'c' && *--p1 == '.') {
					type = F_CPLUSPLUS;
					break;
				}
				goto def;
			case 'h':
				if (*--p1 == '.') {
					type = F_HEADER;
					break;
				}
				goto def;
			case 'o':
				if (*--p1 == '.') {
					if (W.no_object)
						continue;
					type = F_OBJECT;
					break;
				}
				/* fall to ... */
			default:
			def:
#ifdef NO_CHECK_DIR
				sprintf (tc, "%s/%s", directory, item -> d_name);
				stat (tc, &b);	/* est-ce un repertoire */
				if ((b.st_mode & S_IFMT) == S_IFDIR)
					type = F_DIRECTORY;
				else if (b.st_mode & 0111)
					type = F_EXECUTABLE;
#else
				break;
#endif
			}
		}
		if (W.dir_only && type != F_DIRECTORY)
			continue;

		if (type != F_DIRECTORY && !FilenameMatch (W.filter, item -> d_name))
			continue;

		if (small)
			pix = W.small_pixmap [type];
		else
			pix = W.pixmap [type];

		if (nb == 1) {
			IcDestroyAllIpix (IC);
			IcCreateIpix (IC, name, prev_pix, prev_type, prev_type == F_DIRECTORY);
		}

		prev_pix = pix;
		prev_type = type;
		strcpy (name, item -> d_name);

		if (++nb > 1)
			IcCreateIpix (IC, name, pix, type, type == F_DIRECTORY);
		/*
		 *	Chercher le plus grand prefixe.
		 */
		if (max_prefix) {
			if (*max_prefix == 3)
				strcpy (max_prefix, name);
			else {
				for (p1 = name, p2 = max_prefix; *p1++ == *p2++;);
				p2 [-1] = 0;
			}
		}
	}
	closedir (dirp);
	if (max_prefix && (*max_prefix == 3))
		max_prefix [0] = 0;
	/*
	 *	Si rien ne correspond, tant pis.
	 */
	if (nb == 0)
		goto home;	/* regarder dans max_prefix */
	/*
	 *	Si yen a qu'un, c'est celui la!
	 */
	if (nb == 1)
		if (W.dir_only || prev_type == F_DIRECTORY) {
			IcDestroyAllIpix (IC);
			IcCreateIpix (IC, name, prev_pix, prev_type, prev_type == F_DIRECTORY);
		} else
			goto home;	/* rentrer */

	IconAlign (IC, True, True);
	request.request_mode = CWX|CWY;
	request.x = 0;
	request.y = 0;
	XtMakeGeometryRequest (IC, &request, &reply);
	IcRedraw (IC);
home:
	if (XtIsRealized (IC)) {
		XDefineCursor (XtDisplay (w), XtWindow (IC),
				((IconWidget)IC) -> primitive.cursor);
		XtSetSensitive ((Widget) w, True);
	}
	return nb;
}

/*
 *	Les callbacks pour sortir du file selector.
 */

static void fs_cancel (Widget button, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) client;

	XtCallCallbackList ((Widget) w, W.cancel_callback, NULL);
}

static void fs_done (Widget button, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) client;
	char*				p;
	char*				rep;
	char				tc [256];

	rep = AbsPath (W.directory,
			p = WixCopyBuffer (W.fs_wix));
	XtFree (p);

	if (W.dir_only && button == OK) {
		strcpy (W.directory, rep);
		WixReplace (WIX, 0, -1, "", 0);
		WixRedraw (WIX);
		XtCallCallbackList ((Widget) w, W.ok_callback, rep);
		return;
	}
	if (-1 != list_files (rep, "", NULL, w)) {
		strcpy (W.directory, rep);
		sprintf (tc, "%s/", rep);
		WixReplace (W.fs_wix, 0, -1, tc, -1);
		WixRedraw (W.fs_wix);
		SwSetTitle (CL, W.directory);
		return;
	}

	WixReplace (WIX, 0, -1, rep, rindex (rep, '/') - rep + 1);
	WixRedraw (WIX);

	XtCallCallbackList ((Widget) w, W.ok_callback, rep);
}

static void fs_expand (Widget list, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) client;
	char*				dir;
	char*				prefix;
	char*				p;
	char				max_prefix [256];
	char				tc [256];
	int					n;

	p = WixCopyBuffer (W.fs_wix);
	if ((prefix = rindex (p, '/')) != 0) {
		*prefix++ = 0;
		dir = AbsPath (W.directory, p);
	} else if (*p == '~') {
		prefix = "";
		dir = AbsPath (W.directory, p);
	} else {
		dir = W.directory;
		prefix = p;
	}

	n = list_files (dir, prefix, max_prefix, w);
	XtFree (p);
	if (n >= 0) {
		sprintf (tc, "%s/%s", dir, max_prefix);
		if (n == 1 && (-1 != list_files (tc, "", NULL, w)))
			strcat (tc, "/");

		WixReplace (W.fs_wix, 0, -1, tc, -1);
		WixRedraw (W.fs_wix);
	} else
		XBell (XtDisplay (W.fs_wix), 100);
}

static void fs_select (Widget list, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) client;
	Ipix				ic = (Ipix) call;
	char*				dir;
	char*				p;
	char				tc [256];

	p = WixCopyBuffer (W.fs_wix);
	if ((dir = rindex (p, '/')) != 0)
		*dir = 0;
	dir = AbsPath (W.directory, p);
	XtFree (p);

	sprintf (tc, "%s/%s", dir, ic -> label);
	WixReplace (W.fs_wix, 0, -1, tc, -1);
	WixRedraw (W.fs_wix);
}

/*
 *	Rediriger le drag & drop.
 */

static void fs_drop (Widget ic, XtPointer client, XtPointer call)
{
	FileSelectorWidget		w = (FileSelectorWidget) client;
	DropCallbackStruct*		dropcb = (DropCallbackStruct*) call;
	FileSelectorWidget		dest_fmgr;
	FSDropCallbackStruct	cb;

	strcpy (cb.from, AbsPath (W.directory, dropcb -> source -> label));
	cb.ip_to = dropcb -> dest;
	cb.ip_from = dropcb -> source;
	cb.dest_fmgr = NULL;
	if (XtClass (dropcb -> target) == iconWidgetClass) {
		cb.target = NULL;
		dest_fmgr = (FileSelectorWidget) XtParent (XtParent (XtParent (dropcb -> target)));
		cb.dest_fmgr = (Widget) dest_fmgr;
		if (!dest_fmgr || XtClass (dest_fmgr) != fileSelectorWidgetClass) {
			cb.target = dropcb -> target;
			XtCallCallbackList ((Widget) w, W.drop_callback, (XtPointer) &cb);
			return;
		}
		if (dropcb -> dest) {		/* jete dans une icone */
			if (dropcb -> dest -> pixmap == dest_fmgr -> fileSelector.small_pixmap [F_DIRECTORY] ||
			    dropcb -> dest -> pixmap == dest_fmgr -> fileSelector.pixmap [F_DIRECTORY])
				strcpy (cb.to, AbsPath (dest_fmgr -> fileSelector.directory, dropcb -> dest -> label));
			else
				return;		/* pas jete dans une icone de directory */
		} else if (dest_fmgr == w)	/* jete dans le meme mais pas dans une icone */
			return;
		else				/* jete dans un autre fileSelector */
			strcpy (cb.to, dest_fmgr -> fileSelector.directory);
	} else {
		cb.target = dropcb -> target;
		strcpy (cb.to, W.directory);	/* ca peut toujours servir */
	}

	XtCallCallbackList ((Widget) w, W.drop_callback, (XtPointer) &cb);
}

static void fs_kbd (Widget wix, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) client;
	XKeyEvent*			e = (XKeyEvent*) call;
	char				c;
	KeySym				ks;

	if (XLookupString (e, &c, 1, &ks, 0)) {
		if (c == '\r') {
			FakeClick (OK);
			return;
		}
		if (c == ('g'&037)) {
			FakeClick (CAN);
			return;
		}
	}
	fs_expand (wix, (XtPointer) w, NULL);
}

static void recompute_icons (FileSelectorWidget	w)
{
	char*	p;
	char*	dir;

	p = WixCopyBuffer (W.fs_wix);
	if ((dir = rindex (p, '/')) != 0)
		*dir = 0;
	dir = AbsPath (W.directory, p);

	(void) list_files (dir, "", NULL, w);
	XtFree (p);
}

static void fs_big_small (Widget button, XtPointer client, XtPointer call)
{
	FileSelectorSmallIcon (XtParent (button), (Boolean)(int) (long) call);
}

static int cmp_name (Ipix* s1, Ipix* s2)
{
	if ((*s1) -> rank != (*s2) -> rank) {
		if ((*s1) -> rank == F_DIRECTORY)
			return -1;
		if ((*s2) -> rank == F_DIRECTORY)
			return 1;
	}
	return strcmp ((*s1) -> label, (*s2) -> label);
}

static int cmp_type (Ipix* s1, Ipix* s2)
{
	if ((*s1) -> rank != (*s2) -> rank)
		return ((*s1) -> rank < (*s2) -> rank) ? -1 : 1;

	return strcmp ((*s1) -> label, (*s2) -> label);
}

static void fs_sort (Widget button, XtPointer client, XtPointer call)
{
	FileSelectorTypeSort (XtParent (button), call == NULL);
}

/*
 *	Recuperer les composants.
 */

Widget FileSelectorGetChild (Widget widget, int i)
{
	FileSelectorWidget	w = (FileSelectorWidget) widget;

	switch (i) {
		case FS_OK:	return OK;
		case FS_CANCEL:	return CAN;
		case FS_FILTER:	return FILTER;
		case FS_EXPAND:	return EXP;
		case FS_FSPEC:	return WIX;
		case FS_ICON:	return IC;
		default:
			return (Widget) 0;
	}
}

void FileSelectorSmallIcon (Widget widget, Boolean small)
{
	FileSelectorWidget	w = (FileSelectorWidget) widget;

	((IconWidget)IC) -> icon.small = small;
	recompute_icons (w);
}

void FileSelectorTypeSort (Widget widget, Boolean type)
{
	FileSelectorWidget	w = (FileSelectorWidget) widget;

	W.sort = type;
	XtVaSetValues (IC,
		XtNsortFunction, type ? cmp_type : cmp_name,
		NULL);
	recompute_icons (w);
}

void FileSelectorUpdate (Widget widget, String dir)
{
	FileSelectorWidget	w = (FileSelectorWidget) widget;
	char				tc [256];

	if (dir)
		strcpy (W.directory, AbsPath (W.directory, dir));

	(void) list_files (W.directory, "", tc, w);
	WixReplace (WIX, 0, -1, tc, -1);
}

/*
 *	Fait apparaitre le file selector.
 */

void PopFileSelector (Widget widget, char* title, char* name, char* dir, Boolean dirOnly)
{
	FileSelectorWidget	w = (FileSelectorWidget) widget;
	Arg					arg;
	String				s;
	char				tc [256];

	W.dir_only = dirOnly;
	if (title != W.title) {
		W.title = title;
		XtSetArg (arg, XtNlabel, W.title);
		XtSetValues (OK, &arg, 1);
	}
	if ((s = rindex (name, '/')) != 0) {
		strcpy (tc, name);
		tc [s - name + 1] = 0;
	} else
		sprintf (tc, "%s/", dir);

	strcpy (W.directory, dir);
	WixReplace (WIX, 0, -1, tc, -1);
	(void) list_files (tc, "", NULL, w);
	XtManageChild (widget);
	XtSetKeyboardFocus (widget, WIX);
}

static void createPixmaps (Display* dpy, Pixel pix)
{
    Window		root = DefaultRootWindow (dpy);
    XpmAttributes	xpmatt;
    XpmColorSymbol	xpmcol;
    int			i;
    int			ok;

    xpmcol.name = "None";
    xpmcol.value = NULL;
    xpmcol.pixel = pix;
    xpmatt.colorsymbols = &xpmcol;
    xpmatt.numsymbols = 1;
    xpmatt.valuemask = XpmColorSymbols;

    if (DefaultDepth (dpy, DefaultScreen (dpy)) > 1) {
	ok = XpmCreatePixmapFromData (dpy, root, bm_file_xpm, &regularPixmap [F_BITMAP], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, z_file_xpm, &regularPixmap [F_COMPRESSED], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, a_file_xpm, &regularPixmap [F_ARCHIVE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, dot_xpm, &regularPixmap [F_DOTFILE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, dir_xpm, &regularPixmap [F_DIRECTORY], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, c_file_xpm, &regularPixmap [F_CTEXT], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, cc_file_xpm, &regularPixmap [F_CPLUSPLUS], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, o_file_xpm, &regularPixmap [F_OBJECT], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, h_file_xpm, &regularPixmap [F_HEADER], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, file_xpm, &regularPixmap [F_PLAINFILE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, exec_xpm, &regularPixmap [F_EXECUTABLE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, tex_file_xpm, &regularPixmap [F_TEX], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, dvi_file_xpm, &regularPixmap [F_DVI], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, ps_file_xpm, &regularPixmap [F_POSTSCRIPT], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, java_file_xpm, &regularPixmap [F_JAVA], NULL, &xpmatt);

	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_bm_file_xpm, &smallPixmap [F_BITMAP], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_z_file_xpm, &smallPixmap [F_COMPRESSED], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_a_file_xpm, &smallPixmap [F_ARCHIVE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_dot_xpm, &smallPixmap [F_DOTFILE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_dir_xpm, &smallPixmap [F_DIRECTORY], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_c_file_xpm, &smallPixmap [F_CTEXT], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_cc_file_xpm, &smallPixmap [F_CPLUSPLUS], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_o_file_xpm, &smallPixmap [F_OBJECT], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_h_file_xpm, &smallPixmap [F_HEADER], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_file_xpm, &smallPixmap [F_PLAINFILE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_exec_xpm, &smallPixmap [F_EXECUTABLE], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_tex_file_xpm, &smallPixmap [F_TEX], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_dvi_file_xpm, &smallPixmap [F_DVI], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_ps_file_xpm, &smallPixmap [F_POSTSCRIPT], NULL, &xpmatt);
	if (ok == XpmSuccess) ok = XpmCreatePixmapFromData (dpy, root, small_java_file_xpm, &smallPixmap [F_JAVA], NULL, &xpmatt);

	if (ok == XpmSuccess)
		return;

	for (i = 0; i < (int) XtNumber (regularPixmap); i++)
		if (regularPixmap [i])
			XFreePixmap (dpy, regularPixmap [i]);
	for (i = 0; i < (int) XtNumber (smallPixmap); i++)
		if (smallPixmap [i])
			XFreePixmap (dpy, smallPixmap [i]);
	/* fall thru the monochrome case ... */
    }
    regularPixmap [F_BITMAP] =		XCreateBitmapFromData (dpy, root, bm_file_bits, bm_file_width, bm_file_height);
    regularPixmap [F_COMPRESSED] =	XCreateBitmapFromData (dpy, root, z_file_bits, z_file_width, z_file_height);
    regularPixmap [F_ARCHIVE] =		XCreateBitmapFromData (dpy, root, a_file_bits, a_file_width, a_file_height);
    regularPixmap [F_DOTFILE] =		XCreateBitmapFromData (dpy, root, dot_bits, dot_width, dot_height);
    regularPixmap [F_DIRECTORY] =	XCreateBitmapFromData (dpy, root, dir_bits, dir_width, dir_height);
    regularPixmap [F_CTEXT] =		XCreateBitmapFromData (dpy, root, c_file_bits, c_file_width, c_file_height);
    regularPixmap [F_CPLUSPLUS] =	XCreateBitmapFromData (dpy, root, cc_file_bits, cc_file_width, cc_file_height);
    regularPixmap [F_OBJECT] =		XCreateBitmapFromData (dpy, root, o_file_bits, o_file_width, o_file_height);
    regularPixmap [F_HEADER] =		XCreateBitmapFromData (dpy, root, h_file_bits, h_file_width, h_file_height);
    regularPixmap [F_PLAINFILE] =	XCreateBitmapFromData (dpy, root, file_bits, file_width, file_height);
    regularPixmap [F_EXECUTABLE] =	XCreateBitmapFromData (dpy, root, exec_bits, exec_width, exec_height);
    regularPixmap [F_TEX] =		XCreateBitmapFromData (dpy, root, tex_file_bits, tex_file_width, tex_file_height);
    regularPixmap [F_DVI] =		XCreateBitmapFromData (dpy, root, dvi_file_bits, dvi_file_width, dvi_file_height);
    regularPixmap [F_POSTSCRIPT] =	XCreateBitmapFromData (dpy, root, ps_file_bits, ps_file_width, ps_file_height);
    regularPixmap [F_JAVA] =	XCreateBitmapFromData (dpy, root, java_file_bits, java_file_width, java_file_height);

    smallPixmap [F_BITMAP] =		XCreateBitmapFromData (dpy, root, bm_file_sbm_bits, bm_file_sbm_width, bm_file_sbm_height);
    smallPixmap [F_COMPRESSED] =	XCreateBitmapFromData (dpy, root, z_file_sbm_bits, z_file_sbm_width, z_file_sbm_height);
    smallPixmap [F_ARCHIVE] =		XCreateBitmapFromData (dpy, root, a_file_sbm_bits, a_file_sbm_width, a_file_sbm_height);
    smallPixmap [F_DOTFILE] =		XCreateBitmapFromData (dpy, root, dot_file_sbm_bits, dot_file_sbm_width, dot_file_sbm_height);
    smallPixmap [F_DIRECTORY] =		XCreateBitmapFromData (dpy, root, dir_sbm_bits, dir_sbm_width, dir_sbm_height);
    smallPixmap [F_CTEXT] =		XCreateBitmapFromData (dpy, root, c_file_sbm_bits, c_file_sbm_width, c_file_sbm_height);
    smallPixmap [F_CPLUSPLUS] =		XCreateBitmapFromData (dpy, root, cc_file_sbm_bits, cc_file_sbm_width, cc_file_sbm_height);
    smallPixmap [F_OBJECT] =		XCreateBitmapFromData (dpy, root, o_file_sbm_bits, o_file_sbm_width, o_file_sbm_height);
    smallPixmap [F_HEADER] =		XCreateBitmapFromData (dpy, root, h_file_sbm_bits, h_file_sbm_width, h_file_sbm_height);
    smallPixmap [F_PLAINFILE] =		XCreateBitmapFromData (dpy, root, file_sbm_bits, file_sbm_width, file_sbm_height);
    smallPixmap [F_EXECUTABLE] =	XCreateBitmapFromData (dpy, root, exec_sbm_bits, exec_sbm_width, exec_sbm_height);
    smallPixmap [F_TEX] =		XCreateBitmapFromData (dpy, root, tex_file_sbm_bits, tex_file_sbm_width, tex_file_sbm_height);
    smallPixmap [F_DVI] =		XCreateBitmapFromData (dpy, root, dvi_file_sbm_bits, dvi_file_sbm_width, dvi_file_sbm_height);
    smallPixmap [F_POSTSCRIPT] =	XCreateBitmapFromData (dpy, root, ps_file_sbm_bits, ps_file_sbm_width, ps_file_sbm_height);
    smallPixmap [F_JAVA] =	XCreateBitmapFromData (dpy, root, java_file_sbm_bits, java_file_sbm_width, java_file_sbm_height);
}

/*
 *	Declaration des methodes.
 */

static void		ClassInitialize (void);
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static XtGeometryResult	GeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void		ChangeManaged (Widget);

static XtGeometryResult	ClipGeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void	ClipInitialize (Widget, Widget, Arg*, Cardinal*);
static void ClipResize (Widget);

/*
 *	La classe composite mais avec un geometry manager.
 */

static CompositeClassRec clipClassRec = {
  /* core_class fields */
  {
    /* superclass         */    (WidgetClass) &compositeClassRec,
    /* class_name         */    "Clip",
    /* widget_size        */    sizeof(CompositeRec),
    /* class_initialize   */    NULL,
    /* class_partinit     */    NULL,
    /* class_inited       */	False,
    /* initialize         */    ClipInitialize,
    /* Init hook	  */    NULL,
    /* realize            */    XtInheritRealize,
    /* actions		  */	NULL,
    /* num_actions	  */	0,
    /* resources          */    NULL,
    /* num_resources      */    0,
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	False,
    /* compress_exposure  */	False,
    /* compress_enterleave*/	False,
    /* visible_interest   */    False,
    /* destroy            */    NULL,
    /* resize             */    ClipResize,
    /* expose             */    NULL,
    /* set_values         */    NULL,
    /* set values hook    */    NULL,
    /* set values almost  */    XtInheritSetValuesAlmost,
    /* get values hook    */    NULL,
    /* accept_focus       */    NULL,
    /* Version            */    XtVersion,
    /* PRIVATE cb list    */    NULL,
    /* TM Table           */    "",
    /* query_geom         */    NULL,
  },{
/* composite_class fields */
    /* geometry_manager   */    (XtGeometryHandler)ClipGeometryManager,
    /* change_managed     */    XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension */		NULL,
  }
};

static XtResource resources [] = {
	{ XtNokCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (FileSelectorWidget, fileSelector.ok_callback), XtRPointer, NULL
	},
	{ XtNcancelCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (FileSelectorWidget, fileSelector.cancel_callback), XtRPointer, NULL
	},
	{ XtNdropCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (FileSelectorWidget, fileSelector.drop_callback), XtRPointer, NULL
	},
	{ XtNrenameCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (FileSelectorWidget, fileSelector.rename_callback), XtRPointer, NULL
	},
	{ XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (FileSelectorWidget, manager.foreground), XtRString, "white"
	},
	{ XtNfilter, XtCFilter, XtRString, sizeof (String),
	  XtOffset (FileSelectorWidget, fileSelector.filter), XtRImmediate, "*"
	},
	{ XtNtitle, XtCTitle, XtRString, sizeof (String),
	  XtOffset (FileSelectorWidget, fileSelector.title), XtRImmediate, NULL
	},
	{ XtNdirectory, XtCDirectory, XtRString, sizeof (String),
	  XtOffset (FileSelectorWidget, fileSelector.directory), XtRImmediate, NULL
	},
	{ XtNdirOnly, XtCDirOnly, XtRBoolean, sizeof (Boolean),
	  XtOffset (FileSelectorWidget, fileSelector.dir_only), XtRImmediate, (XtPointer) False
	},
	{ XtNnoObject, XtCNoObject, XtRBoolean, sizeof (Boolean),
	  XtOffset (FileSelectorWidget, fileSelector.no_object), XtRImmediate, (XtPointer) False
	},
	{ XtNnoEdit, XtCNoEdit, XtRBoolean, sizeof (Boolean),
	  XtOffset (FileSelectorWidget, fileSelector.no_edit), XtRImmediate, (XtPointer) True
	},
	{ XtNnoDotfile, XtCNoDotfile, XtRBoolean, sizeof (Boolean),
	  XtOffset (FileSelectorWidget, fileSelector.no_dotfile), XtRImmediate, (XtPointer) False
	},
	{ XtNoffset, XtCOffset, XtRInt, sizeof (int),
	  XtOffset (FileSelectorWidget, fileSelector.offset), XtRImmediate, (XtPointer) 2
	},
	{ XtNzPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_COMPRESSED]), XtRImmediate, (XtPointer) -1
	},
	{ XtNaPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_ARCHIVE]), XtRImmediate, (XtPointer) -1
	},
	{ XtNexecPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_EXECUTABLE]), XtRImmediate, (XtPointer) -1
	},
	{ XtNcPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_CTEXT]), XtRImmediate, (XtPointer) -1
	},
	{ XtNccPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_CPLUSPLUS]), XtRImmediate, (XtPointer) -1
	},
	{ XtNtexPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_TEX]), XtRImmediate, (XtPointer) -1
	},
	{ XtNdviPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_DVI]), XtRImmediate, (XtPointer) -1
	},
	{ XtNpsPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_POSTSCRIPT]), XtRImmediate, (XtPointer) -1
	},
	{ XtNbmPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_BITMAP]), XtRImmediate, (XtPointer) -1
	},
	{ XtNoPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_OBJECT]), XtRImmediate, (XtPointer) -1
	},
	{ XtNjavaPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_JAVA]), XtRImmediate, (XtPointer) -1
	},
	{ XtNfilePixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_PLAINFILE]), XtRImmediate, (XtPointer) -1
	},
	{ XtNdirPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_DIRECTORY]), XtRImmediate, (XtPointer) -1
	},
	{ XtNdotPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_DOTFILE]), XtRImmediate, (XtPointer) -1
	},
	{ XtNhPixmap, XtCIconPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (FileSelectorWidget, fileSelector.pixmap [F_HEADER]), XtRImmediate, (XtPointer) -1
	},
};

FileSelectorClassRec fileSelectorClassRec = {
  /* core_class fields */
  {
    /* superclass         */    (WidgetClass) &containerClassRec,
    /* class_name         */    "FileSelector",
    /* widget_size        */    sizeof(FileSelectorRec),
    /* class_initialize   */    ClassInitialize,
    /* class_partinit     */    NULL,
    /* class_inited       */	False,
    /* initialize         */    Initialize,
    /* Init hook	  */    NULL,
    /* realize            */    XtInheritRealize,
    /* actions		  */	NULL,
    /* num_actions	  */	0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber (resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	False,
    /* compress_exposure  */	True,
    /* compress_enterleave*/	False,
    /* visible_interest   */    False,
    /* destroy            */    NULL,
    /* resize             */    Resize,
    /* expose             */    Redisplay,
    /* set_values         */    SetValues,
    /* set values hook    */    NULL,
    /* set values almost  */    XtInheritSetValuesAlmost,
    /* get values hook    */    NULL,
    /* accept_focus       */    NULL,
    /* Version            */    XtVersion,
    /* PRIVATE cb list    */    NULL,
    /* TM Table           */    "",
    /* query_geom         */    NULL,
  },{
/* composite_class fields */
    /* geometry_manager   */    (XtGeometryHandler)GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension */		NULL,
  },{
/* Constraint class Init */
    NULL,	/* resources */
    0,		/* num_resources */
    0,		/* constraint_size */
    NULL,	/* initialize */
    NULL,	/* destroy */
    NULL,	/* set_values */
    NULL	/* extension */
      
  },{
/* manager_class fields */
	NULL,
  },{
/* File Selector class - none */     
   0
 }	
};

WidgetClass fileSelectorWidgetClass = (WidgetClass)&fileSelectorClassRec;

static void ClassInitialize (void)
{
	chdir (GetCurrentDir ());
}

/*
 *	Fonction de scroll ...
 */

static void h_scroll (Widget sb, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) client;
	int					pos = (int) (long) call;
	int	offset = W.edit->core.x - IC -> core.x;

	XtMoveWidget (IC, -pos, IC -> core.y);
	XtMoveWidget (W.edit, offset - pos, W.edit->core.y);
}

static void v_scroll (Widget sb, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) client;
	int					pos = (int) (long) call;
	int	offset = W.edit->core.y - IC -> core.y;

	XtMoveWidget (IC, IC -> core.x, -pos);
	XtMoveWidget (W.edit, W.edit->core.x, offset - pos);
}

/*
 *	Edition du nom de fichier.
 */

static void fs_rename (Widget wix, XtPointer client, XtPointer call)
{
	FileSelectorWidget	w = (FileSelectorWidget) XtParent (XtParent (XtParent (wix)));
	Ipix				ip = (Ipix) client;
	FSRenameCallbackStruct	cb;

	XtRemoveCallback (wix, XtNactivateCallback, fs_rename, ip);
	cb.old_name = ip -> label;
	cb.new_name = WixCopyBuffer (wix);
	cb.doit = True;
	XtCallCallbackList ((Widget) w, W.rename_callback, (XtPointer) &cb);
	if (cb.doit) {
		XtFree (ip -> label);
		ip -> label = cb.new_name;
	}
	XtUnmanageChild (wix);
	XtSetKeyboardFocus ((Widget) w, WIX);
	if (cb.doit)
		(void) list_files (W.directory, "", NULL, w);
}

static void fs_edit (Widget widget, XtPointer client, XtPointer call)
{
	IconWidget			ic = (IconWidget) widget;
	FileSelectorWidget	w = (FileSelectorWidget) client;
	Ipix				ip = (Ipix) call;
	Widget				ww = W.edit;
	int					wib;
	int					wst;

	XtVaGetValues (ww,
		XtNinnerBorder,		&wib,
		XtNshadowThickness,	&wst,
		NULL);
	if (ic->icon.small)
		XtMoveWidget (ww,
			ip -> x + ip -> p_width + ic->core.x
				+ 4 + ic->primitive.shadow_thickness
				- wib - wst,
			ip -> y + ic->core.y
				- ic->icon.fp->ascent - ic->icon.fp->descent
				+ ip -> p_height + (ip -> height - ip -> p_height) / 2
				- wib - wst);
	else
		XtMoveWidget (ww,
			ip -> x + ic->core.x
				+ (ip -> width -  XTextWidth (ic->icon.fp, ip -> label, strlen (ip->label)))/2
				- wib - wst,
			ip -> y + ip -> p_height + ic->core.y
				+ 2 + ic->primitive.shadow_thickness
				- wib - wst);

	WixReplace (ww, 0, -1, ip -> label, -1);
	WixRedraw (ww);
	XRaiseWindow (XtDisplay (widget), XtWindow (ww));
	XtAddCallback (ww, XtNactivateCallback, fs_rename, ip);
	XtManageChild (ww);
	XtSetKeyboardFocus (widget, CL);
	XtSetKeyboardFocus (CL, XtParent (ww));
	XtSetKeyboardFocus (XtParent (ww), ww);
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	FileSelectorWidget		w = (FileSelectorWidget) new;
	static XtTranslations	trans = NULL;
	Widget					components [10];
	static Boolean			first = True;
	int						nc = 0;
	int						n;
	Widget					wid;
	Boolean					small;
	Boolean					oblong;
	Pixel					bg_pix;

	if (!trans) {
		trans = XtParseTranslationTable (
			"<Key>Tab:	activate()\n\
			 <Key>Escape:	activate()");
	}

	if (request -> core.width==0)	C.width = DEFAULT_WIDTH;
	if (request -> core.height==0)	C.height = DEFAULT_HEIGHT;

	W.directory = XtMalloc (256);
	strcpy (W.directory, GetCurrentDir ());
	components [nc++] = CL = XtVaCreateWidget ("sw", scrolledWindowWidgetClass, new,
								XtNforeground,	w->manager.foreground,
								XtNbackground, C.background_pixel,
#ifndef NO_PATH
								XtNwideScrollBar,	False,
								XtNtitle,		W.directory,
#endif
								NULL);
	wid = XtVaCreateManagedWidget ("clip", (WidgetClass) &clipClassRec, CL, NULL);
	W.sort = False;	/* initialement trie par type */
	IC = XtVaCreateManagedWidget ("icon", iconWidgetClass, wid,
					XtNsortFunction, W.sort ? cmp_name : cmp_type,
					NULL);
	/*
	 *	Creer les pixmaps en remplacant la couleur None par le fond.
	 */
	if (first) {
		first = False;
		XtVaGetValues (IC, XtNbackground, &bg_pix, NULL);
		createPixmaps (XtDisplay (w), bg_pix);
	}
	/*
	 *	Initialiser les pixmaps des icones.
	 */
	for (n = N_PIXMAP; --n >= 0;) {
		if (W.pixmap [n] == (Pixmap)-1)
			W.pixmap [n] = regularPixmap [n];
		W.small_pixmap [n] = smallPixmap [n];
	}

	W.edit = XtVaCreateWidget ("icon_edit", wixWidgetClass, wid,
						XtNmonoLine,	True,
						XtNkillStack,	False,
						XtNcolumnNumber,16,
						XtNbackground, ((IconWidget)IC) -> core.background_pixel,
						NULL);
	XtAddCallback (((ScrolledWindowWidget)CL) -> swindow.hScrollBar,
						XtNsliderMoved, h_scroll, (XtPointer) w);
	XtAddCallback (((ScrolledWindowWidget)CL) -> swindow.vScrollBar,
						XtNsliderMoved, v_scroll, (XtPointer) w);

	small = ((IconWidget)IC) -> icon.small;
	if (!W.no_edit)
		XtAddCallback (IC, XtNeditCallback, fs_edit, (XtPointer) w);
	XtAddCallback (IC, XtNselect, fs_select, (XtPointer) w);
	XtAddCallback (IC, XtNdblClickCallback, fs_done, (XtPointer) w); /* double click */
	XtAddCallback (IC, XtNdropCallback, fs_drop, (XtPointer) w); /* drag & drop */

	components [nc++] = WIX = XtVaCreateWidget ("file", wixWidgetClass, new,
								XtNmonoLine, 	True,
								XtNkillStack,	False,
								NULL);
	XtOverrideTranslations (WIX, trans);
	XtAddCallback (WIX, XtNactivateCallback, fs_kbd, (XtPointer) w);

	components [nc++] = OK = XtVaCreateWidget ("Ok", pushButtonWidgetClass, new,
									XtNlabel,	W.title,
									XtNforeground,	w->manager.foreground,
									XtNbackground, C.background_pixel,
									NULL);
	XtVaGetValues (OK, XtNoblong, &oblong, NULL);
	XtVaSetValues (OK,
			XtNshowAsDefault, oblong ? 3 : 3 * w->manager.shadow_thickness,
			NULL);
	XtAddCallback (OK, XtNrelease, fs_done, (XtPointer) w);

	components [nc++] = EXP = XtVaCreateWidget ("Expand", pushButtonWidgetClass, new,
								XtNforeground,	w->manager.foreground,
								XtNbackground, C.background_pixel,
								XtNtoggle,	True,
								XtNset,		small,
								XtNlabel,	"Small Icons",
								NULL);
	XtAddCallback (EXP, XtNrelease, fs_big_small, (XtPointer) IC);

	components [nc++] = FILTER = XtVaCreateWidget ("Filter", pushButtonWidgetClass, new,
									XtNforeground,	w->manager.foreground,
									XtNbackground, C.background_pixel,
									XtNtoggle,	True,
									XtNset,		W.sort,
									XtNlabel,	"Sort by name",
									NULL);
	XtAddCallback (FILTER, XtNrelease, fs_sort, (XtPointer) IC);

	components [nc++] = CAN = XtVaCreateWidget ("Cancel", pushButtonWidgetClass, new,
									XtNforeground,	w->manager.foreground,
									XtNbackground, C.background_pixel,
									NULL);
	XtAddCallback (CAN, XtNrelease, fs_cancel, (XtPointer) w);
	XtManageChildren (components, nc);
}

static	Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	FileSelectorWidget	cu = (FileSelectorWidget) current;
	FileSelectorWidget	w = (FileSelectorWidget) new;

	if (cu -> fileSelector.title != W.title) {
		XtVaSetValues (OK,
				XtNlabel, W.title,
				NULL);
	}
	return False;
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	(*containerClassRec.core_class.expose) (widget, event, region);
}

static void Resize (Widget widget)
{
	FileSelectorWidget	w = (FileSelectorWidget) widget;
	int					offset = W.offset;
	Dimension			icHeight;
	Dimension			wixHeight = WIX -> core.managed ? WIX -> core.height : 0;
	Dimension			okHeight = OK -> core.managed ? OK -> core.height : 0;
	Dimension			okWidth = OK -> core.managed ? OK -> core.width : 0;
	Dimension			ok_sha = ((PushButtonWidget)OK) -> pushbutton.def_shadow;
	Dimension			buttonHeight = CAN -> core.height;
	Dimension			cancelWidth = CAN -> core.managed ? CAN -> core.width : 0;
	Dimension			expandWidth = EXP -> core.managed ? EXP -> core.width : 0;
	Dimension			filterWidth = FILTER -> core.managed ? FILTER -> core.width : 0;

	W.sep_y = 0;
	if (CL -> core.managed) {
		Dimension wi, he;

		icHeight = C.height - 2 * MARGINS - okHeight - wixHeight - 2 - 5 * offset;
		if (!OK -> core.managed &&
		    !EXP -> core.managed &&
		    !CAN -> core.managed &&
		    !FILTER -> core.managed)
		{
			icHeight += offset;
			if (!WIX -> core.managed)
				W.sep_y = -1;
		}
		if (!WIX -> core.managed)
			icHeight += 2 * offset;

		XtConfigureWidget (CL, offset + MARGINS, offset + MARGINS,
			C.width - 2 * (offset + MARGINS), icHeight, 0);
		wi = IC -> core.width;
		he = IC -> core.height;
		if (wi < CL -> core.width) wi = CL -> core.width;
		if (he < icHeight) he = icHeight;
		XtConfigureWidget (IC, 0, 0, wi, he, 0);
	} else
		icHeight = 0;

	if (!W.sep_y)
		W.sep_y = MARGINS /* + 2 */ + icHeight + 2 * offset;

	if (WIX -> core.managed)
		XtConfigureWidget (WIX, offset + MARGINS, W.sep_y + offset, C.width - 2 * (offset + MARGINS), wixHeight, 0);

	if (OK -> core.managed)
		XtConfigureWidget (OK, offset + MARGINS, C.height - MARGINS - okHeight - offset,
			okWidth, okHeight, 0);

	if (EXP -> core.managed)
		XtConfigureWidget (EXP,
			(C.width - expandWidth) / 3,
			C.height - MARGINS - ok_sha - buttonHeight - offset,
			expandWidth, buttonHeight, 0);

	if (FILTER -> core.managed)
		XtConfigureWidget (FILTER,
			(2 * (C.width - filterWidth)) / 3,
			C.height - MARGINS - ok_sha - buttonHeight - offset,
			filterWidth, buttonHeight, 0);

	if (CAN -> core.managed)
		XtConfigureWidget (CAN, C.width - MARGINS - offset - cancelWidth, C.height - ok_sha - buttonHeight - offset,
			cancelWidth, buttonHeight, 0);

	if (XtIsRealized (widget))
		IconAlign (IC, True, True);
}

static void ClipInitialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	if (new -> core.width == 0)		new -> core.width = 700;
	if (new -> core.height == 0)	new -> core.height = 350;
}

static void ClipResize (Widget cw)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) XtParent (cw);

	XtVaSetValues (sw -> swindow.hScrollBar,
		XtNsliderExtent, cw -> core.width,
		NULL);
	XtVaSetValues (sw -> swindow.vScrollBar,
		XtNsliderExtent, cw -> core.height,
		NULL);
}

static XtGeometryResult ClipGeometryManager (Widget ch, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) XtParent (XtParent (ch));
	Widget					hsb = sw -> swindow.hScrollBar;
	Widget					vsb = sw -> swindow.vScrollBar;
	Widget					edit = ((FileSelectorWidget)XtParent (sw)) -> fileSelector.edit;
	
	if (request -> request_mode & CWX) {
		int	offset = edit->core.x - ch -> core.x;

		ch -> core.x = request -> x;
		XtMoveWidget (edit, offset + request -> x, edit->core.y);
		XtVaSetValues (hsb, XtNsliderOrigin, - request -> x, NULL);
	}

	if (request -> request_mode & CWY) {
		int	offset = edit->core.y - ch -> core.y;

		ch -> core.y = request -> y;
		XtMoveWidget (edit, edit -> core.x, offset + request -> y);
		XtVaSetValues (vsb, XtNsliderOrigin, - request -> y, NULL);
	}

	if (request -> request_mode & CWWidth) {
		ch -> core.width = request -> width;
		XtVaSetValues (hsb,
			XtNsliderMax, ch -> core.width,
			XtNsliderExtent, (ch -> core.parent) -> core.width,
			NULL);
	}
	if (request -> request_mode & CWHeight) {
		ch -> core.height = request -> height;
		XtVaSetValues (vsb,
			XtNsliderMax, ch -> core.height,
			XtNsliderExtent, (ch -> core.parent) -> core.height,
			NULL);
	}
	reply -> request_mode = 0;
	return XtGeometryYes;
}

static XtGeometryResult GeometryManager (Widget w, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	if ((request -> request_mode & CWX) || (request -> request_mode & CWY))
		return XtGeometryNo;

	if (request -> request_mode & CWWidth) {
		w -> core.width = request -> width;
		reply -> width = request -> width;
	}
	if (request -> request_mode & CWHeight) {
		w -> core.height = request -> height;
		reply -> height = request -> height;
	}
	reply -> request_mode = request -> request_mode;
	return XtGeometryYes;
}

static void ChangeManaged (Widget w)
{
	Resize (w);
}
