/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/CompositeP.h>
#include <X11/StringDefs.h>
#include <X11/xpm.h>
#include <X11/cursorfont.h>

#include <Xpa/FrameP.h>
#include <Xpa/Frame.h>
#include <Xpa/close.xbm>
#include <Xpa/iconify.xbm>
#include <Xpa/maximize.xbm>

#define	C	((Widget)fw) -> core
#define	CP	fw -> composite
#define	W	fw -> frame

static XtGeometryResult
			GeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static Boolean	SetValues (Widget, Widget, Widget, ArgList, Cardinal*);
static void	InsertChild (Widget);
static void	ChangeManaged (Widget);
static void	Initialize (Widget, Widget, Arg*, Cardinal*);
static void	Redisplay (Widget, XEvent*, Region);
static void	Resize (Widget);
static void	Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static void Layout (Widget, Boolean);

static void Drag (Widget, XEvent*, String*, Cardinal*);
static void Focus (Widget, XEvent*, String*, Cardinal*);

static char defaultTranslations [] =
	"<FocusIn>:		Focus()\n"
	"<FocusOut>:		Focus()\n"
	"<Btn1Down>:		Drag()\n"
	"<Btn1Motion>:	Drag()\n"
	"<Btn1Up>:		Drag()";

static XtActionsRec actions [] =
{
	{"Focus",(XtActionProc) Focus},
	{"Drag",	(XtActionProc) Drag},
};

static XtResource resources [] = {
	{ XtNborderSize, XtCBorderSize, XtRInt, sizeof(int),
	  XtOffset (FrameWidget, frame.border_size), XtRImmediate, (XtPointer) 6
	},
	{ XtNuserData, XtCUserData, XtRPointer, sizeof (XtPointer),
	  XtOffset(FrameWidget, frame.user_data), XtRImmediate, (XtPointer) NULL
	},
	{ XtNcursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(FrameWidget, frame.cursor), XtRString, "top_left_arrow"
	},
	{ XtNmotionCursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(FrameWidget, frame.motion_cursor), XtRString, "fleur"
	},
	{ XtNtitle, XtCTitle, XtRString, sizeof (String),
	  XtOffset(FrameWidget, frame.title), XtRImmediate, (XtPointer) NULL
	},
	{ XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (FrameWidget, frame.foreground), XtRString, "grey"
	},
	{ XtNhilite, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (FrameWidget, frame.hilite), XtRString, "white"
	},
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct*),
	  XtOffset (FrameWidget, frame.font), XtRString, "fixed"
	},
	{ XtNcloseCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (FrameWidget, frame.close_callback), XtRPointer, (XtPointer) NULL
	},
	{ XtNmaxCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (FrameWidget, frame.max_callback), XtRPointer, (XtPointer) NULL
	},
	{ XtNminCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (FrameWidget, frame.min_callback), XtRPointer, (XtPointer) NULL
	},
};

FrameClassRec frameClassRec = {
	{
		/* superclass */			(WidgetClass) &compositeClassRec,
		/* class_name */			"Frame",
		/* size */					sizeof(FrameRec),
		/* Class Initializer */		NULL,
		/* class_part_initialize */	NULL,
		/* Class init'ed ? */			FALSE,
		/* initialize */				Initialize,
		/* initialize_notify */		NULL,		
		/* realize */				Realize,
		/* actions */				actions,
		/* num_actions */			XtNumber (actions),
		/* resources */				resources,
		/* resource_count */		XtNumber (resources),
		/* xrm_class */				NULLQUARK,
		/* compress_motion */		FALSE,
		/* compress_exposure */	FALSE,
		/* compress_enterleave */	FALSE,
		/* visible_interest */		FALSE,
		/* destroy */				NULL,
		/* resize */				Resize,
		/* expose */				Redisplay,
		/* set_values */			SetValues,
		/* set_values_hook */		NULL,			
		/* set_values_almost */		XtInheritSetValuesAlmost,  
		/* get_values_hook */		NULL,
		/* accept_focus */			NULL,
		/* intrinsics version */		XtVersion,
		/* callback offsets */		NULL,
		/* tm_table */				defaultTranslations,
		/* query_geometry */		NULL,
		/* display_accelerator */		NULL,
		/* extension */				NULL
	},{
/* composite_class fields */
		/* geometry_manager */		GeometryManager,
		/* change_managed */		ChangeManaged,
		/* insert_child */			InsertChild,
		/* delete_child */			XtInheritDeleteChild,
		/* extension */				NULL,
	},{
/* frame class fields */
		/* extension */				NULL,
	}
};

WidgetClass frameWidgetClass = (WidgetClass) &frameClassRec;

/*
 *	Raffichage.
 */

/*
 *	Transforme le repre RVB en HSV
 */

static void rgb2hsv (XColor *xc, double* hr, double* sr, double* vr)
{
	double rd, gd, bd, h, s, v, max, min, del, rc, gc, bc;

	rd = xc -> red / 65535.0;            /* rd,gd,bd range 0-1 instead of 0-65535 */
	gd = xc -> green / 65535.0;
	bd = xc -> blue / 65535.0;

	/* compute maximum of rd,gd,bd */
	if (rd>=gd) { if (rd>=bd) max = rd;  else max = bd; }
	else { if (gd>=bd) max = gd;  else max = bd; }

	/* compute minimum of rd,gd,bd */
	if (rd<=gd) { if (rd<=bd) min = rd;  else min = bd; }
	else { if (gd<=bd) min = gd;  else min = bd; }

	del = max - min;

	v = max;
	s = (max == 0.0) ? 0.0 : del / max;

	h = 0;
	if (s != 0.0) {
		rc = (max - rd) / del;
		gc = (max - gd) / del;
		bc = (max - bd) / del;

		if      (rd==max) h = bc - gc;
		else if (gd==max) h = 2 + rc - bc;
		else if (bd==max) h = 4 + gc - rc;

		h = h * 60;
		if (h < 0)
			h += 360;
	}
	*hr = h/60.;  *sr = s;  *vr = v;
}

/*
 *	Transforme le repere HSV en RVB
 */

#define floor(x) (x)

static Pixel hsv2rgb (Display* dpy, Colormap cmap, double h, double s, double v)
	/* h: 0 -> 6 s: 0 -> 1 v: 0 -> 1 */
{
	XColor	xc;
	int		j;
	double	rd, gd, bd;
	double	f, p, q, t;

	if (s == 0.0) { rd = v;  gd = v;  bd = v; }
	else {
		if (h == 6.0)
			h = 0.0;
		j = (int) floor (h);
		f = h - j;
		p = v * (1 - s);
		q = v * (1 - (s * f));
		t = v * (1 - (s * (1 - f)));

		switch (j) {
			/* j: 0 -> 5 mais on ne sait jamais!! */
			default:
			case 0:  rd = v;  gd = t;  bd = p; break;
			case 1:  rd = q;  gd = v;  bd = p; break;
			case 2:  rd = p;  gd = v;  bd = t; break;
			case 3:  rd = p;  gd = q;  bd = v; break;
			case 4:  rd = t;  gd = p;  bd = v; break;
			case 5:  rd = v;  gd = p;  bd = q; break;
		}
	}
	/* stocker la couleur */
	xc.flags = DoRed|DoGreen|DoBlue;
	xc.red = ((int) floor ((rd * 255.0) + 0.5))<<8;
	xc.green = ((int) floor ((gd * 255.0) + 0.5))<<8;
	xc.blue = ((int) floor ((bd * 255.0) + 0.5))<<8;
	XAllocColor (dpy, cmap, &xc);
	return xc.pixel;
}

static void draw_title (FrameWidget fw)
{
	Display*	dpy = XtDisplay ((Widget) fw);
	Window		win = XtWindow ((Widget) fw);

	if (W.title) {
		XSetFont (dpy, W.gc, W.font -> fid);
		/* un petite ombre */
		XSetForeground (dpy, W.gc, !W.focused ? W.hilite : W.foreground);
		XDrawString (dpy, win, W.gc, 21, W.font -> ascent, W.title, strlen (W.title));
		XSetForeground (dpy, W.gc, W.focused ? W.hilite : W.foreground);
		XDrawString (dpy, win, W.gc, 20, W.font -> ascent - 1, W.title, strlen (W.title));
	}
}

static void draw_icon (FrameWidget fw, Boolean on, Pixmap pixmap, int x, int y, int width, int height)
{
	Display*	dpy = XtDisplay ((Widget) fw);
	Window		win = XtWindow ((Widget) fw);

	XSetFillStyle (dpy, W.gc, FillStippled);
	XSetStipple (dpy, W.gc, pixmap);

	XSetForeground (dpy, W.gc, on ? W.foreground : W.hilite);
	XSetTSOrigin (dpy, W.gc, x+2, y+2);
	XFillRectangle (dpy, win, W.gc, x+2, y+2, width, height);

	XSetForeground (dpy, W.gc, on ? W.hilite : W.foreground);
	XSetTSOrigin (dpy, W.gc, x, y);
	XFillRectangle (dpy, win, W.gc, x, y, width, height);
	XSetFillStyle (dpy, W.gc, FillSolid);
}

static Pixmap	closepix;
static Pixmap	iconpix;
static Pixmap	maxpix;

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	FrameWidget	fw = (FrameWidget) widget;
	XColor		xc;
	Display*	dpy = XtDisplay (widget);
	Window		win = XtWindow (widget);
	Colormap	cmap = C.colormap;
	int			i, j;

	if (W.border_size == 0)
		return;		/* cadre sans bord ! */

	if (W.npixels == 0) {
		static double grad [MAX_SIZE] = {
			3/30., 5/30., 6/30., 8/30., 10/30.,
			11/30., 12/30., 14/30., 15/30., 16/30.,
			17/30., 18/30., 19/30., 20/30., 20/30.,
			20/30., 20/30., 20/30., 20/30., 20/30.,
		};
		double	h, s, v;

		W.npixels = W.border_size*2;
		xc.flags = DoRed|DoGreen|DoBlue;
		xc.pixel = C.background_pixel;
		XQueryColor (dpy, cmap, &xc);
		rgb2hsv (&xc, &h, &s, &v);
		for (i = 0; i < W.npixels; ++i)
			W.pixels [i] = hsv2rgb (dpy, cmap, h, s, v - 16/30. + grad [i]);
		/* Acclrer les dplacements sur les machines lentes. */
		if (0 == (XtBuildEventMask (widget) & PointerMotionHintMask))
			XSelectInput (dpy, win, XtBuildEventMask (widget) | PointerMotionHintMask);
		/* charger les petits pixmaps */
		closepix = XCreateBitmapFromData (dpy, win, close_bits, close_width, close_height);
		iconpix = XCreateBitmapFromData (dpy, win, iconify_bits, iconify_width, iconify_height);
		maxpix = XCreateBitmapFromData (dpy, win, maximize_bits, maximize_width, maximize_height);
	}
	/* bord de gauche */
	for (i = 0, j = 0; i < W.npixels; i += 2, ++j) {
		XSetForeground (dpy, W.gc, W.pixels [W.npixels - 1 - i]);
		XFillRectangle (dpy, win, W.gc, j, j, 1, C.height - 2*j);
	}
	/* bord de droite */
	for (i = 0, j = 0; i < W.npixels; i += 2, ++j) {
		XSetForeground (dpy, W.gc, W.pixels [W.npixels - 1 - i]);
		XFillRectangle (dpy, win, W.gc, C.width-j-1, j, 1, C.height - 2*j);
	}
	/* bord du haut */
	for (i = 0; i < W.npixels/2; ++i) {
		XSetForeground (dpy, W.gc, W.pixels [W.npixels - 1 - i]);
		XFillRectangle (dpy, win, W.gc, i, i, C.width-2*i, 1);
	}
	for (j = i; i < W.npixels; ++i) {
		XSetForeground (dpy, W.gc, W.pixels [W.npixels - 1 - i]);
		XFillRectangle (dpy, win, W.gc, j, i, C.width-2*j, 1);
	}
	/* bord du bas */
	for (i = 0, j = 0; i < W.npixels; i += 2, ++j) {
		XSetForeground (dpy, W.gc, W.pixels [W.npixels - 1 - i]);
		XFillRectangle (dpy, win, W.gc, j, C.height-j-1, C.width-2*j, 1);
	}
	if (W.border_size >= 8) {
		draw_title (fw);	/* titre */

		draw_icon (fw, True, closepix, C.width - 30, 1, close_width, close_height);
		draw_icon (fw, True, maxpix, C.width - 30 - 18, 1, maximize_width, maximize_height);
		draw_icon (fw, True, iconpix, C.width - 30 - 18 - 18, 1, iconify_width, iconify_height);
	}
}

static void Realize	(Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	FrameWidget	fw = (FrameWidget) widget;
	Display*	dpy = XtDisplay (widget);

	*value_mask |= CWCursor;
	attributes -> cursor = W.cursor;
	XtCreateWindow (widget, InputOutput, CopyFromParent, *value_mask, attributes);
	W.gc = XCreateGC (dpy, XtWindow (widget), 0, 0);
}

static Boolean SetValues (Widget cu, Widget req, Widget w, ArgList al, Cardinal* cn)
{
	FrameWidget	fw = (FrameWidget) w;
	FrameWidget	cw = (FrameWidget) cu;
	Boolean		do_redisplay = False;

	if (cw -> frame.title != W.title) {
#if 0
		if (cw -> frame.title != NULL)
			XtFree (cw -> frame.title);	/* on frit un truc pas malloqu ... */
#endif
		W.title = XtNewString (W.title);
		do_redisplay = True;
	}
	return do_redisplay;
}

/*
 *	N'autoriser qu'un sous widget
 */

static void InsertChild (Widget w)
{
	FrameWidget	fw = (FrameWidget) XtParent (w);
    
	if (CP.num_children == 0)
		(*compositeClassRec.composite_class.insert_child) (w);
	else
		XtWarning ("No more than one widget in a frame.\n");
}

/*
 *	On est d'accord avec tout.
 */

static XtGeometryResult GeometryManager
			(Widget wid, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	FrameWidget			fw = (FrameWidget)(wid -> core.parent);
	XtWidgetGeometry	my_request;

	if (request -> request_mode & (CWX | CWY))
		return XtGeometryNo;

	my_request.request_mode = (request -> request_mode & XtCWQueryOnly);
	if (request -> request_mode & CWWidth) {
	    my_request.width = request -> width + 2*W.border_size;
	    my_request.request_mode |= CWWidth;
	}
	if (request -> request_mode & CWHeight) {
	    my_request.height = request -> height + 3*W.border_size;
	    my_request.request_mode |= CWHeight;
	}
	if (request -> request_mode & CWBorderWidth) {
	    my_request.border_width = request -> border_width;
	    my_request.request_mode |= CWBorderWidth;
	}
	if (XtMakeGeometryRequest((Widget)fw, &my_request, NULL) == XtGeometryYes) {
		if (!(request -> request_mode & XtCWQueryOnly)) {
			wid -> core.width = C.width - 2*W.border_size;
			wid -> core.height = C.height - 3*W.border_size;
			if (request -> request_mode & CWBorderWidth) {
				wid -> core.x = wid -> core.y = -request -> border_width;
			}
		}
		return XtGeometryYes;
	}
	return XtGeometryNo;
}

/*
 *	La procdure de traitement du changement de taille.
 */

static void Layout (Widget widget, Boolean set_size)
{
	FrameWidget	fw = (FrameWidget) widget;
	Widget		child = CP.children [0];

	if (child == NULL || ! child -> core.managed)
		return;		/* rien  placer */

	if (set_size) {
		int	width = 2*W.border_size + child -> core.width;
		int	height = 3*W.border_size + child -> core.height;

		if (C.width < width)	C.width = width;
		if (C.height < height)	C.height = height;
	}
	if (C.width > 2*W.border_size && C.height > 3*W.border_size)
		XtConfigureWidget (child, W.border_size, 2*W.border_size,
								  C.width-2*W.border_size,
								  C.height-3*W.border_size, 0);
}

static void ChangeManaged (Widget widget)
{
	Layout (widget, True);
}

static void Resize (Widget widget)
{
	Layout (widget, False);
}

/*
 *	La procdure d'initilisation.
 */

static void Initialize (Widget request, Widget widget, Arg* args, Cardinal* nargs)
{
	FrameWidget	fw = (FrameWidget) widget;

	if (C.width == 0)	C.width = 100;
	if (C.height == 0)	C.height = 100;
	W.npixels = 0;
	if (W.border_size < 0 || W.border_size > MAX_SIZE)
		W.border_size = 6;	/* par exemple */
}

static void Focus (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	FrameWidget	fw = (FrameWidget) widget;

	W.focused = (event -> type == FocusIn);
	if (W.border_size >= 8)
		draw_title (fw);
}

/*
 *	Dplace la fentre.
 */

#define NOWHERE	-1
#define ICON	0
#define ZOOM	1
#define CLOSE	2

static void Drag (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	extern WidgetClass	shellWidgetClass;
	FrameWidget			fw = (FrameWidget) widget;
	XMotionEvent*		e = (XMotionEvent*) event;
	Display*			dpy = XtDisplay (widget);
	Window				win = XtWindow (widget);
	static int			x;
	static int			y;
	static Boolean		do_move;
	static int			where = NOWHERE;
	int					xx;
	Window				wjunk;
	int					junk;
	unsigned int		ujunk;
	int					xjunk, yjunk;

	if (!XtIsSubclass (XtParent (widget), shellWidgetClass))
		return;		/* on ne dplace que les shells! */

	switch (e -> type) {
		case ButtonPress:
			/* pas dans la zone des boutons */
			do_move = (e -> x < C.width-30-18-18) || e -> y > W.npixels+2;
			/* mettre la fentre devant */
			XRaiseWindow (dpy, XtWindow (widget -> core.parent));
			/* rcuprer l'id de la fentre de dcor */
			XTranslateCoordinates (dpy, win, e -> root, e -> x, e -> y, &x, &y, &wjunk);
			/*
			 * pour en avoir sa position relle, car Xt ne met pas  jour la position
			 * si la dimension a change en mme temps ('agrandir en tirant le coin
			 * suprieur gauche')
			*/
			XGetGeometry (dpy, wjunk, &wjunk, &x, &y,
						  &ujunk, &ujunk, &ujunk, &ujunk);
			x = e -> x_root - x - 2;
			y = e -> y_root - y - 2;
			where = NOWHERE;
			if (do_move) {
				XDefineCursor (dpy, win, W.motion_cursor);
				XSync (dpy, False);
				return;
			}
			/* on est dans un bouton */
			xx = x - C.width;
			if (-30-18-18+iconify_width > xx && xx > -30-18-18) {
				/* iconification */
				where = ICON;
				draw_icon (fw, False, iconpix, C.width - 30 - 18 - 18, 1, iconify_width, iconify_height);
			} else
			if (-30-18+maximize_width > xx && xx > -30-18) {
				/* zoom */
				where = ZOOM;
				draw_icon (fw, False, maxpix, C.width - 30 - 18, 1, maximize_width, maximize_height);
			} else
			if (-30+close_width > xx && xx > -30) {
				/* fermeture */
				where = CLOSE;
				draw_icon (fw, False, closepix, C.width - 30, 1, close_width, close_height);
			}
			return;

		case ButtonRelease:
			if (do_move) {
				XDefineCursor (dpy, win, W.cursor);
				XSync (dpy, False);
				return;
			}
			/* click: rallumer le bon bouton */
			switch (where) {
				case NOWHERE:
							return;	/* aucun? */
				case ZOOM:	draw_icon (fw, True, maxpix, C.width - 30 - 18, 1, maximize_width, maximize_height);
							break;
				case ICON:	draw_icon (fw, True, iconpix, C.width - 30 - 18 - 18, 1, iconify_width, iconify_height);
							break;
				case CLOSE:	draw_icon (fw, True, closepix, C.width - 30, 1, close_width, close_height);
							break;
			}
			if (e -> x-x > 4 || e -> x-x < -4 || e -> y-y > 4 || e -> y-y < -4)
				return;

			x = e -> x - C.width;
			if (-30-18-18+iconify_width > x && x > -30-18-18)
				/* iconification */
				XtCallCallbackList (widget, W.min_callback, NULL);
			else
			if (-30-18+maximize_width > x && x > -30-18)
				/* zoom */
				XtCallCallbackList (widget, W.max_callback, NULL);
			else
			if (-30+close_width > x && x > -30)
				/* fermeture */
				XtCallCallbackList (widget, W.close_callback, NULL);
			return;

		case MotionNotify:
			if (do_move) {
				XQueryPointer (dpy, RootWindowOfScreen (XtScreen (widget)),
								&wjunk, &wjunk, &junk, &junk, &xjunk, &yjunk, &ujunk);
				XMoveWindow (dpy, XtWindow (XtParent (widget)),
							 e -> x_root-x, e -> y_root-y);
			}
			return;
	}
}
