/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "Icon.h"
#include "IconP.h"
  
#define W	w -> icon
#define P	w -> primitive
#define C	w -> core

#define MARGINS	(P.shadow_thickness + W.margin)
#define InsideIpix(ip,a,b) \
	((a) >= ip->x && (b) >= ip->y && (a) < ip->x + ip->width && (b) < ip->y + ip->height)


static	void	Initialize (Widget, Widget, Arg*, Cardinal*);
static	void	Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static	void	Destroy (Widget);
static	void	Redisplay (Widget, XEvent*, Region);
static	Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

static	void	Arm (Widget, XEvent*, String*, Cardinal*);
static	void	Reset (Widget, XEvent*, String*, Cardinal*);
static	void	Edit (Widget, XEvent*, String*, Cardinal*);
static	void	move_icon (Widget, XEvent*, String*, Cardinal*);
static	void	window_drag (Widget, XEvent*, String*, Cardinal*);

static	void	ipix_draw (IconWidget, Window, Ipix, Boolean, Boolean);
static	void	recomputeSize (IconWidget);
static	int		cmp (Ipix*, Ipix*);

static char defaultTranslations [] =
#ifdef MOVE_ICON
	"~Shift<Btn1Down>:	Arm()\n\
	 Shift<Btn1Down>:	Arm(Multiple)\n\
	 <Btn1Motion>:	Arm()\n\
	 <Btn1Up>:	Reset()\n\
	 <Btn2Up>:	Edit()\n\
	 <Btn3Down>:	Drag()\n\
	 <Btn3Motion>:	Drag()\n\
	 <Btn3Up>:	Drag()";
#else
	"<Btn1Down>:	Arm()\n\
	 <Btn1Up>:	Reset()\n\
	 <Btn2Up>:	Edit()\n\
	 <Btn3Down>:	Drag()\n\
	 <Btn3Motion>:	Drag()\n\
	 <Btn3Up>:	Drag()";
#endif

static XtActionsRec actionsList [] =
{
	{"Edit",	(XtActionProc) Edit},
	{"Arm",		(XtActionProc) Arm},
	{"Reset",	(XtActionProc) Reset},
	{"Drag",	(XtActionProc) window_drag},
};

/*  The resource list for Push Button  */

static XtResource resources [] =
{     
	{ XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	  XtOffset (IconWidget, core.background_pixel), XtRString, "grey95"
	},
	{ XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (IconWidget, primitive.foreground), XtRString, "maroon"
	},
	{ XtNsortFunction, XtCFunction, XtRFunction, sizeof (CompareFunction),
	  XtOffset (IconWidget, icon.sort_function), XtRImmediate, (XtPointer) cmp
	},
	{ XtNdblClickCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (IconWidget, icon.dbl_click_callback), XtRPointer, NULL
	},
	{ XtNdropCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (IconWidget, icon.drop_callback), XtRPointer, NULL
	},
	{ XtNeditCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (IconWidget, icon.edit_callback), XtRPointer, NULL
	},
	{ XtNhotColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (IconWidget, icon.hot_color), XtRString, "red"
	},
	{ XtNsmallIcons, XtCSmallIcons, XtRBoolean, sizeof (Boolean),
	  XtOffset (IconWidget, icon.small), XtRImmediate, (XtPointer) False
	},
	{ XtNnoDrag, XtCNoDrag, XtRBoolean, sizeof (Boolean),
	  XtOffset (IconWidget, icon.noDrag), XtRImmediate, (XtPointer) False
	},
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct*),
	  XtOffset (IconWidget, icon.fp), XtRString, "fixed"
	},
	{ XtNgridWidth, XtCWidth, XtRInt, sizeof (int),
	  XtOffset (IconWidget, icon.grid_width), XtRImmediate, (XtPointer) 0
	},
	{ XtNgridHeight, XtCWidth, XtRInt, sizeof (int),
	  XtOffset (IconWidget, icon.grid_height), XtRImmediate, (XtPointer) 0
	},
	{ XtNcursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(IconWidget, primitive.cursor), XtRString, "top_left_arrow"
	},
	{ XtNmotionCursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(IconWidget, icon.motion_cursor), XtRString, "fleur"
	},
	{ XtNwaitCursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(IconWidget, icon.wait_cursor), XtRString, "watch"
	},
};

IconClassRec iconClassRec = {
    {
	(WidgetClass) &primitiveClassRec,	/* superclass */
	"Icon",				/* class_name */
	sizeof (IconRec),		/* widget_size */
	NULL,				/* class_initialize */
	NULL,				/* class_part_init */
	False,				/* class_inited */
	Initialize,			/* initialize */
	NULL,				/* initialize_hook */
	Realize,			/* realize */
	actionsList,			/* actions */
	XtNumber (actionsList),		/* num_actions */
	resources,			/* resources */
	XtNumber (resources),		/* num_resources */
	NULLQUARK,			/* xrm_class */
	False,				/* compress_motion */
	True,				/* compress_exposure */
	True,				/* compress_enterlv */
	False,				/* visible_interest */
	Destroy,			/* destroy */
	NULL,				/* resize */
	Redisplay,			/* expose */
	SetValues,			/* set_values */
	NULL,				/* set_values_hook */
	XtInheritSetValuesAlmost,	/* set_values_almost */
	NULL,				/* get_values_hook */
	NULL,				/* accept_focus */
	XtVersion,			/* version */
	NULL,				/* callback_private */
	defaultTranslations,		/* tm_table */
	NULL,				/* query_geometry */
    },
    {
	NULL,
    },
    {
	0,
    },
};

WidgetClass iconWidgetClass = (WidgetClass)&iconClassRec;

/*
 *	Compte le nombre de clicks.
 */

static long	lastButtonUpTime = 0;
static int	numberOfClicks;

static void countClicks (XButtonEvent* e)
{
	if (((long) e -> time - lastButtonUpTime) > XtGetMultiClickTime (e -> display))
		numberOfClicks = 1;
	else
		numberOfClicks += 1;
}

/*
 *	le GC pour le label.
 */

static void get_labelGC (IconWidget w)
{
	XGCValues		values;
	unsigned long	valuemask;

	valuemask = GCForeground | GCBackground | GCFont;
	values.font = W.fp -> fid;

	values.foreground = P.foreground;
	values.background = C.background_pixel;
	W.label_GC = XtGetGC ((Widget) w, valuemask, &values);

	values.foreground = W.hot_color;
	W.inverse_GC = XtGetGC ((Widget) w, valuemask, &values);
}

static void Realize (Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	IconWidget	w = (IconWidget) widget;

	*value_mask |= CWCursor;
	attributes -> cursor = P.cursor;
	attributes -> event_mask |= PointerMotionHintMask;
#ifdef MOVE_ICON
	attributes -> event_mask |= OwnerGrabButtonMask|EnterWindowMask|LeaveWindowMask;
#endif
	XtCreateWindow (widget, InputOutput, (Visual *)CopyFromParent, *value_mask, attributes);
}

/*
 *	Fonction d'initialisation de chaque instance.
 */

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	IconWidget	w = (IconWidget) new;
	Dimension	parentWidth = XtParent (new) -> core.width;
	Dimension	parentHeight = XtParent (new) -> core.height;

	if (C.width < parentWidth)		C.width = parentWidth;
	if (C.height < parentHeight)	C.height = parentHeight;

	if (C.width == 0)	C.width = 300;
	if (C.height == 0)	C.height = 300;

	get_labelGC (w);
	W.ipix = NULL;
	W.last = NULL;
	W.free = NULL;
	W.selected = NULL;
	W.margin = 2;
}

/*
 *	SetValues.
 */

static	Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	IconWidget	cu = (IconWidget) current;
	IconWidget	w = (IconWidget) new;
	Boolean		do_redisplay = False;

	W.grid_width = cu -> icon.grid_width;
	W.grid_height = cu -> icon.grid_height;

	if (cu -> icon.fp != W.fp ||
	    cu -> primitive.foreground != P.foreground ||
	    cu -> core.background_pixel != C.background_pixel) {
		XtDestroyGC (cu -> icon.label_GC);
		XtDestroyGC (cu -> icon.inverse_GC);
		get_labelGC (w);
		do_redisplay = True;
	}

	if (cu -> icon.fp != W.fp || cu -> icon.small != W.small) {
		recomputeSize (w);
		do_redisplay = True;
	}

	return do_redisplay;
}

/*
 *	La procedure de destruction d'une instance.
 */

static void Destroy (Widget widget)
{
	IconWidget	w = (IconWidget) widget;
	Ipix		ip;

	XtDestroyGC (W.label_GC);
	XtDestroyGC (W.inverse_GC);
	for (ip = W.ipix; ip; ip = ip -> next)
		XtFree ((char*) ip);
	for (ip = W.free; ip; ip = ip -> next)
		XtFree ((char*) ip);
}

/*
 *	Indique si on est dans une icone.
 */

static Ipix InsideIcon (IconWidget w, int x, int y)
{
	Ipix	ip;

	for (ip = W.ipix; ip; ip = ip -> next)
		if (InsideIpix (ip, x, y))
			return ip;
	return NULL;
}

/*
 *	L'action de selection.
 */

static Ipix	prev_ip = NULL;
static Ipix	activate_ip = NULL;
static Ipix	prev_dest = NULL;

static void Arm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	IconWidget		w = (IconWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	Ipix			ip;
	int				d;
	static int		px = -1;
	static int		py = -1;
	static XButtonEvent	be;

	switch (e -> type) {
	case ButtonPress:
		px = e -> x;
		py = e -> y;
		be = *e;
		countClicks (e);
		if ((ip = InsideIcon (w, e -> x, e -> y)) != 0) {
			if (numberOfClicks == 1) {
				prev_ip = ip;
				IcSelectIpix (widget, ip, *num_params == 1);	/* Selection multiple */
			} else if (numberOfClicks == 2) {
				if (prev_ip == ip)		/* double click sur */
					activate_ip = ip;	/* bouton relache */
				prev_ip = NULL;
			}
			XtCallCallbackList (widget, P.select, (XtPointer) ip);
		} else
			IcSelectIpix (widget, NULL, False);	/* les deselectionner */
		return;

	case MotionNotify:
		if (W.noDrag)
			return;

		d = (e->x - px) * (e->x - px);
		d += (e->y - py) * (e->y - py);
		if (d > 100 && prev_ip) {
			XEvent			ev;
			XtAppContext	app = XtWidgetToApplicationContext (widget);

			prev_dest = NULL;
			move_icon (widget, (XEvent*) &be, NULL, NULL);	/* le ButtonPress */
			move_icon (widget, (XEvent*) e, NULL, NULL);	/* le MotionNotify */
			for (;;) {
				XtAppNextEvent (app, &ev);
				move_icon (widget, &ev, NULL, NULL);
				if (ev.type == ButtonRelease &&
				    ev.xbutton.button == be.button)
					break;
			}
		}
		return;
	}
}

static void Reset (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	IconWidget		w = (IconWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	lastButtonUpTime = e -> time;

	if (activate_ip) {
		if (W.selected && !W.selected -> next_selected)
			XtCallCallbackList (widget, W.dbl_click_callback, (XtPointer) activate_ip);
		IcSelectIpix (widget, NULL, False);
		activate_ip = NULL;
	}
}

/*
 *	Edite le nom d'une icone.
 */

static void Edit (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	IconWidget		w = (IconWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	Ipix			ip = InsideIcon (w, e -> x, e -> y);

	if (ip)
		XtCallCallbackList (widget, W.edit_callback, (XtPointer) ip);
}

/*
 *	Deplace une icone.
 */

#define MAX_PIX	64

static void move_icon (Widget wid, XEvent* event, String* params, Cardinal* num_params)
{
	IconWidget		w = (IconWidget) wid;
	XButtonEvent*	e = (XButtonEvent*) event;
	static Ipix		ip;
	Ipix			dest;
	Ipix			p;
	static Widget	widget;
	static int		prev_x;
	static int		prev_y;
	static Pixmap	pix [MAX_PIX];
	static Pixmap	save_pix [MAX_PIX];
	static Ipix		ipix [MAX_PIX];
	static int		ni;
	static GC		gc = (GC) 0;
	static int		x_offset = 0;
	static int		y_offset = 0;
	int				wx, wy;
	int				x = e -> x_root - x_offset;
	int				y = e -> y_root - y_offset;
	int				i;
	XGCValues		values;
	Display*		dpy = XtDisplay (wid);
	Window			root = RootWindowOfScreen (XtScreen (wid));
	int				depth = DefaultDepthOfScreen (XtScreen (wid));
	Window			wfoo;
	Window			target;
	Widget			wtarget;
	int				ifoo;
	int				ipx, ipy;
	DropCallbackStruct	dropcb;

	switch (event -> type) {
	default:
		XtDispatchEvent (event);
		return;

	case EnterNotify:
		widget = XtWindowToWidget (dpy, e -> window);
		if ((XtClass (widget) != iconWidgetClass))
			widget = NULL;
		return;

	case LeaveNotify:
		/*
		 *	Simuler un MotionNotify pour eteindre l'eventuel
		 *	precedent icone allume quand on sort de la fenetre.
		 */
		if (widget && prev_dest)
		{
			e -> type = MotionNotify;
			e -> x = -100;
			e -> y = -100;
			e -> x_root = -100;
			e -> y_root = -100;
			move_icon ((Widget) w, (XEvent*) e, NULL, NULL);
		}
		widget = NULL;
		return;

	case MotionNotify:
		if (!ip) return;
		/*
		 *	Remettre la rootwindow comme il faut.
		 */
		for (i = 0; i < ni; i++) {
			XCopyArea (dpy, save_pix [i], root, gc,
						0, 0, ipix [i] -> width, ipix [i] -> height,
						prev_x + ipix [i] -> x - ip -> x,
						prev_y + ipix [i] -> y - ip -> y);
		}
		/*
		 *	Allumer les icones energisables.
		 */
		if (widget) {
			if (widget == (Widget) w) {
				wx = e -> x;
				wy = e -> y;
			} else {
				Position	rx, ry;
				XtTranslateCoords (widget, 0, 0, &rx, &ry);
				wx = e -> x_root - rx;
				wy = e -> y_root - ry;
			}
			/*
			 *	Eteindre le precedent.
			 */
			if (prev_dest) {
				if (InsideIpix (prev_dest, wx, wy))
					goto suite;
				ipix_draw ((IconWidget) widget, XtWindow (widget), prev_dest, True, False);
				prev_dest = NULL;
			}
			/*
			 *	Allumer le bon.
			 */
			if ((dest = InsideIcon ((IconWidget) widget, wx, wy))
			    && dest != ip
			    && dest -> sensible)
			{
				ipix_draw ((IconWidget) widget, XtWindow (widget), dest, False, True);
				prev_dest = dest;
			}
		}
		goto suite;

	case ButtonPress:
		ip = InsideIcon (w, e -> x, e -> y);
		if (!ip)
			return;

		widget = (Widget) w;	/* on demarre chez nous */

		if (!gc) {
			values.subwindow_mode = IncludeInferiors;
			values.graphics_exposures = False;
			values.function = GXcopy;
			gc = XtGetGC (widget, GCSubwindowMode|GCFunction|GCGraphicsExposures,&values);
		}
		for (ni = 0, p = W.selected; p && ni < MAX_PIX; p = p -> next_selected, ni++) {
			ipix [ni] = p;
			save_pix [ni] = XCreatePixmap (dpy, root, p -> width, p -> height, depth);
			pix [ni] = XCreatePixmap (dpy, root, p -> width, p -> height, depth);
		}
		/*
		 *	Capturer l'image des icones.
		 */
		XSetForeground (dpy, gc, C.background_pixel);
		for (i = 0; i < ni; i++) {
			ipx = ipix [i] -> x; ipy = ipix [i] -> y;
			XFillRectangle (dpy, pix [i], gc, 0, 0, ipix [i] -> width, ipix [i] -> height);
			ipix [i] -> x = ipix [i] -> y = 0;
			ipix_draw (w, pix [i], ipix [i], False, False);
			ipix [i] -> x = ipx; ipix [i] -> y = ipy;
		/*
		 *	Effacer l'icone de la fenetre.
		 */
			XClearArea (dpy, XtWindow ((Widget) w), ipix [i] -> x, ipix [i] -> y, ipix [i] -> width, ipix [i] -> height, False);
		}
		x_offset = e -> x - ip -> x;
		y_offset = e -> y - ip -> y;

	suite:
		for (i = 0; i < ni; i++) {
			XCopyArea (dpy, root, save_pix [i], gc,
						x + ipix [i] -> x - ip -> x,
						y + ipix [i] -> y - ip -> y,
						ipix [i] -> width, ipix [i] -> height, 0, 0);
			XCopyArea (dpy, pix [i], root, gc,
						0, 0, ipix [i] -> width, ipix [i] -> height,
						x + ipix [i] -> x - ip -> x,
						y + ipix [i] -> y - ip -> y);
		}
		prev_x = x;
		prev_y = y;
		return;

	case ButtonRelease:
		if (!ip) return;
		for (i = 0; i < ni; i++) {
			XCopyArea (dpy, save_pix [i], root, gc,
						0, 0, ipix [i] -> width, ipix [i] -> height,
						x + ipix [i] -> x - ip -> x,
						y + ipix [i] -> y - ip -> y);
			XFreePixmap (dpy, pix [i]);
			XFreePixmap (dpy, save_pix [i]);
		}
		if (widget && prev_dest)
			ipix_draw ((IconWidget)widget, XtWindow (widget), prev_dest, True, False);

		for (i = 0; i < ni; i++)
			ipix_draw (w, XtWindow ((Widget) w), ipix [i], False, False);
		/*
		 *	Si c'est dans une fenetre d'icones ca roule.
		 */
		if (widget) {
			dropcb.target = widget;
			dropcb.dest = NULL;
			dropcb.source = W.selected;
			if (prev_dest)
				dropcb.dest = prev_dest;
			if (dropcb.dest != ip)
				XtCallCallbackList ((Widget) w, W.drop_callback, (XtPointer) &dropcb);
			prev_dest = NULL;
			ip = NULL;
			return;
		}
		x_offset = 0;
		y_offset = 0;
		for (target = root;;) {
			XQueryPointer (dpy, target, &wfoo, &target, &ifoo, &ifoo, &wx, &wy, &ifoo);
			if (target == root || target == None) {
				ip = NULL;
				return;
			}
			if (! (wtarget = XtWindowToWidget (dpy, target)))
				continue;

			dropcb.dest = NULL;
			dropcb.target = NULL;
			dropcb.shell = wtarget;
			dropcb.source = W.selected;
			for (widget = wtarget;;) {
				XQueryPointer (dpy, target, &wfoo, &target, &ifoo, &ifoo, &wx, &wy, &ifoo);
				if (target == None)
					break;
				dropcb.target = widget = XtWindowToWidget (dpy, target);
			}
			XtCallCallbackList ((Widget) w, W.drop_callback, (XtPointer) &dropcb);
			ip = NULL;
			return;
		}
	}
}

/*
 *	Drag dans le parent de la fentre d'icones.
 */

static void window_drag (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	IconWidget		w = (IconWidget) widget;
	XMotionEvent*	e = (XMotionEvent*) event;
	static int		drag_x, drag_y;
	int				x, y;
	int				min_x, min_y;
	Window			wjunk;
	int				junk;
	unsigned int	ujunk;
	XtWidgetGeometry	request, reply;

	switch (e -> type) {
	case ButtonRelease:
		XDefineCursor (XtDisplay (widget), XtWindow (widget), P.cursor);
		return;

	case ButtonPress:
		XDefineCursor (XtDisplay (widget), XtWindow (widget), W.motion_cursor);
		drag_x = e -> x;
		drag_y = e -> y;
		return;

	case MotionNotify:
		XQueryPointer (XtDisplay (widget), XtWindow (XtParent (widget)),
						&wjunk, &wjunk, &junk, &junk, &x, &y, &ujunk);
		x -= drag_x; y -= drag_y;

		if (x > 0)
			x = 0;
		else {
			min_x =  (int) (XtParent (widget) -> core.width) - (int) C.width;
			if (x < min_x)
				x = min_x;
		}

		if (y > 0)
			y = 0;
		else {
			min_y =  (int) (XtParent (widget) -> core.height) - (int) C.height;
			if (y < min_y)
				y = min_y;
		}
		request.request_mode = CWX|CWY;
		request.x = x;
		request.y = y;
		(void) XtMakeGeometryRequest (widget, &request, &reply);
	}
}

/*
 *	Calcule une racine carre entire.
 */

static int int_sqrt (int n)
{
	int	s;

	for (s = 1; s*s <= n ; s++);
	return s - 1;
}

/*
 *	Aligne les icones.
 */

static int cmp (Ipix* s1, Ipix* s2)
{
	if ((*s1) -> rank != (*s2) -> rank)
		return ((*s1) -> rank < (*s2) -> rank) ? -1 : 1;

	return strcmp ((*s1) -> label, (*s2) -> label);
}

void IconAlign (Widget widget, Boolean update_grid, Boolean do_sort)
{
	IconWidget	w = (IconWidget) widget;
	Ipix		ip, *pi, *tmp;
	int			i, maxw, maxh, x, y;
	int			parentWidth, parentHeight;
	int			curcol, numcols;
	XtWidgetGeometry	request, reply;
	/*
	 *	Calcule la taille de la grille et
	 *	le nombre d'icones.
	 */
	if (!(ip = W.ipix))
		return;		/* ya rien a aligner */

	maxw = maxh = 0;
	for (i = 0; ip; ip = ip -> next, i++) {
		if (maxw < ip -> width)		maxw = ip -> width;
		if (maxh < ip -> height)	maxh = ip -> height;
	}

	if (update_grid) {
		W.grid_width = maxw;
		W.grid_height = maxh;
	} else {
		maxw = W.grid_width;
		maxh = W.grid_height;
	}
	x = y = 0;
	parentWidth = (XtParent ((Widget) w)) -> core.width;
	parentHeight = (XtParent ((Widget) w)) -> core.height;
	curcol = -1;
	numcols = W.small ? parentHeight / maxh : int_sqrt (i);
	/*
	 *	Trier les icones.
	 */
	if (do_sort) {
		pi = tmp = (Ipix*) XtMalloc (i * sizeof (Ipix));
		for (ip = W.ipix; ip; ip = ip -> next)
			*pi++ = ip;
		qsort (tmp, i, sizeof (Ipix), (int (*)(const void*, const void*)) W.sort_function);
		/*
		 *	Aligner le tout.
		 */
		if (W.small)	maxw = 0;
		for (pi = tmp; --i >= 0;) {
			ip = *pi++;
			if (++curcol >= numcols) {
				y = 0;
				x += maxw;
				if (W.small) maxw = 0;
				curcol = 0;
			}
			ip -> x = x;
			if (!W.small)	/* centrer dans la maille */
				ip -> x += (maxw - ip -> width) / 2;
			ip -> y = y + (maxh - ip -> height) / 2;
			y += maxh;
			if (W.small && maxw < ip -> width) maxw = ip -> width;
		}
		XtFree ((char*) tmp);	/* degager le tableau */
	} else {
		for (ip = W.ipix; ip; ip = ip -> next) {
			if (++curcol >= numcols) {
				y = 0;
				x += maxw;
				curcol = 0;
			}
			ip -> x = x;
			if (!W.small)	/* centrer dans la maille */
				ip -> x += (maxw - ip -> width) / 2;
			ip -> y = y + (maxh - ip -> height) / 2;
			y += maxh;
		}
		
	}
	y = numcols * maxh;
	if (y < parentHeight)
		y = parentHeight;

	x = x + maxw;
	if (x < parentWidth)
		x = parentWidth;
	request.request_mode = CWWidth|CWHeight;
	request.width = x;
	request.height = y;
	(void) XtMakeGeometryRequest ((Widget) w, &request, &reply);
}

/*
 *	Affiche une icone.
 */

static void ipix_draw (IconWidget w, Window window, Ipix ip, Boolean erase, Boolean energized)
{
	Display*	dpy = XtDisplay ((Widget) w);
	GC			gc;
	int			len;
	int			pl;
	/*
	 *	Dessiner les ombres.
	 */
	if (energized) {
		gc = W.inverse_GC;
		_DrawShadow (dpy, window,
						P.bot_GC, P.top_GC, P.shadow_thickness,
						ip -> width, ip -> height,
						ip -> x, ip -> y);
	} else {
		gc = W.label_GC;
		if (ip -> selected && P.shadow_thickness > 0)
			_DrawShadow (dpy, window,
							P.top_GC, P.bot_GC, P.shadow_thickness,
							ip -> width, ip -> height,
							ip -> x, ip -> y);

		if (erase && P.shadow_thickness > 0)
			_EraseShadow (dpy, window,
							P.shadow_thickness,
							ip -> width, ip -> height,
							ip -> x, ip -> y);
	}
	/*
	 *	dessiner le pixmap
	 */
	if (ip -> pixmap) {
		if (W.small) {
		    if (ip -> p_depth == 1)
			XCopyPlane (dpy, ip -> pixmap, window, gc,
				0, 0, ip -> p_width, ip -> p_height,
				ip -> x + MARGINS,
				ip -> y + (ip -> height - ip -> p_height) / 2,
				1);
		    else
			XCopyArea (dpy, ip -> pixmap, window, gc,
				0, 0, ip -> p_width, ip -> p_height,
				ip -> x + MARGINS,
				ip -> y + (ip -> height - ip -> p_height) / 2);
		} else {
		    if (ip -> p_depth == 1)
				XCopyPlane (dpy, ip -> pixmap, window, gc,
						0, 0, ip -> p_width, ip -> p_height,
						ip -> x + (ip -> width - ip -> p_width) / 2,
						ip -> y + MARGINS,
						1);
		    else
				XCopyArea (dpy, ip -> pixmap, window, gc,
						0, 0, ip -> p_width, ip -> p_height,
						ip -> x + (ip -> width - ip -> p_width) / 2,
						ip -> y + MARGINS);
		}
	}
	/*
	 *	dessiner le nom
	 */
	if (ip -> label) {
		len = strlen (ip -> label);
		if (W.small) {
			XDrawString (dpy, window, gc,
							ip -> x + ip -> p_width + MARGINS + W.margin,
							ip -> y - W.fp -> descent + ip -> p_height + (ip -> height - ip -> p_height) / 2,
							ip -> label, len);
		} else {
			pl = XTextWidth (W.fp, ip -> label, len);
			XDrawString (dpy, window, gc,
							ip -> x + (ip -> width - pl) / 2,
							ip -> y + W.fp -> ascent + ip -> p_height + MARGINS,
							ip -> label, len);
		}
	}
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	IconWidget	w = (IconWidget) widget;
	Display*	dpy = XtDisplay (widget);
	Ipix		ip;

	XSetRegion (dpy, W.label_GC, region);
	XSetRegion (dpy, W.inverse_GC, region);

	for (ip = W.ipix; ip; ip = ip -> next) {
		if (RectangleOut != XRectInRegion (region, ip -> x, ip -> y, ip -> width, ip -> height))
			ipix_draw (w, XtWindow (widget), ip, False, False);
	}
	/*
	 *	Virer la region de clipping.
	 */
	XSetClipMask (dpy, W.inverse_GC, None);
	XSetClipMask (dpy, W.label_GC, None);
}

/*
 *	Recalcule la taille globale.
 */

static void recomputeSize (IconWidget w)
{
	Ipix	ip;

	for (ip = W.ipix; ip; ip = ip -> next) {
		ip -> width = ip -> p_width;
		ip -> height = ip -> p_height;
		if (ip -> label) {
			int	fh = W.fp -> ascent + W.fp -> descent;
			int	pl = XTextWidth (W.fp, ip -> label, strlen (ip -> label));

			if (W.small) {
				ip -> width += pl + W.margin;
				if (ip -> height < fh)	ip -> height = fh;
			} else {
				if (ip -> width < pl)	ip -> width = pl;
				ip -> height += W.fp -> ascent + W.fp -> descent;
			}
		}
		ip -> width += 2 * MARGINS;
		ip -> height += 2 * MARGINS;
	}
}

/*
 *	Fonctions publiques.
 */

void IcRedraw (Widget w)
{
	if (XtIsRealized (w))
		XClearArea (XtDisplay (w), XtWindow (w), 0, 0, 0, 0, True);
}

Ipix IcCreateIpix (Widget widget, char* label, Pixmap pixmap, short rank, Boolean sensible)
{
	IconWidget	w = (IconWidget) widget;
	int			raf;
	Window		wraf;
	int			width;
	int			height;
	int			depth;
	Ipix		ip;
	/*
	 *	Rcuprer un prcdent icone s'il en reste
	 *	ou alors s'en allouer un neuf.
	 */
	if (W.free) {
		ip = W.free;
		W.free = ip -> next;
		if (ip -> label)
			XtFree (ip -> label);
	} else {	/* inserer au bout */
		ip = (Ipix) XtMalloc (sizeof (IpixRec));
		bzero ((char*)ip, sizeof (IpixRec));
	}
	ip -> next = NULL;
	if (W.last)
		W.last -> next = ip;
	W.last = ip;
	if (!W.ipix)
		W.ipix = ip;
	/*
	 *	Recupere la geometrie du pixmap.
	 */
	if (ip -> pixmap != pixmap)
		if ((ip -> pixmap = pixmap) != 0) {
			XGetGeometry (XtDisplay (widget), pixmap, &wraf, &raf, &raf, &width, &height, &raf, &depth);
			ip -> p_depth = depth;
			ip -> p_width = width;
			ip -> p_height = height;
		} else {
			ip -> p_depth = 0;
			ip -> p_width = 0;
			ip -> p_height = 0;
		}
	/*
	 *	Mettre a jour la taille globale.
	 */
	ip -> width = ip -> p_width;
	ip -> height = ip -> p_height;
	if (label) {
		int	fh = W.fp -> ascent + W.fp -> descent;
		int	pl = XTextWidth (W.fp, label, strlen (label));

		ip -> label = XtNewString (label);
		if (W.small) {
			ip -> width += pl + W.margin;
			if (ip -> height < fh)	ip -> height = fh;
		} else {
			if (ip -> width < pl)	ip -> width = pl;
			ip -> height += W.fp -> ascent + W.fp -> descent;
		}
	}
	ip -> width += 2 * MARGINS;
	ip -> height += 2 * MARGINS;
	ip -> x = 0;
	ip -> y = 0;
	ip -> selected = False;
	ip -> rank = rank;
	ip -> sensible = sensible;
	ip -> next_selected = NULL;
	return ip;
}

Boolean IcDestroyIpix (Widget widget, Ipix ic, Boolean show)
{
	IconWidget	w = (IconWidget) widget;
	Ipix		ip;

	if (ic == W.ipix) {
		W.ipix = ic -> next;	/* c'est le premier */
		if (W.last == ic)
			W.last = W.ipix;	/* c'est le seul */
	} else {
		for (ip = W.ipix; ip; ip = ip -> next)
			if (ip -> next == ic) {
				if (W.last == ic)
					W.last = ip;
				ip -> next = ic -> next;
				break;
			}
		if (!ip)
			return False;	/* pas trouve */
	}
	ic -> next = W.free;
	W.free = ic;		/* le mettre a gauche */
	if (W.selected == ic)
		W.selected = NULL;
	/*
	 *	Aligner et reafficher.
	 */
	if (show) {
		IconAlign (widget, True, True);
		IcRedraw (widget);
	}
	return True;
}

void IcDestroyAllIpix (Widget widget)
{
	IconWidget	w = (IconWidget) widget;

	if (W.ipix) {
		W.last -> next = W.free;
		W.free = W.ipix;
		W.ipix = NULL;
		W.last = NULL;
		W.selected = NULL;
	}
}

Ipix IcGetIconList (Widget widget)
{
	IconWidget	w = (IconWidget) widget;

	return W.ipix;
}

void IcSelectIpix (Widget widget, Ipix ip, Boolean multiple)
{
	IconWidget	w = (IconWidget) widget;
	Ipix		p, q;

	if (W.selected) {
		if (!multiple || !ip) {
			for (p = W.selected; p; p = q) {
				q = p -> next_selected;
				p -> selected = False;
				p -> next_selected = NULL;
				ipix_draw (w, XtWindow (w), p, True, False);
			}
			W.selected = ip;
		} else if (ip) {
			/* ne rajouter 'ip' que s'il n'y est pas deja */
			for (p = W.selected; p && p != ip; p = p -> next_selected)
				;
			if (!p) {
				ip -> next_selected = W.selected;
				W.selected = ip;
			}
		} else
			W.selected = ip;
		XtCallCallbackList (widget, P.release, (XtPointer) W.selected);
	} else
		W.selected = ip;
	/* ip == NULL pour juste deselectionner l'icone */
	if (ip) {
		ip -> selected = True;
		ipix_draw (w, XtWindow (widget), ip, False, False);
	}
}
