/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#if SYS_DIR
#	include <sys/dir.h>
#else
#	include <dirent.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Shell.h>

#include <Xpa/Primitive.h>
#include <Xpa/PrimitiveP.h>
#include <Xpa/MenuBar.h>
#include <Xpa/MenuBarP.h>
#include <Xpa/Wix.h>
#include <Xpa/PushButton.h>
#include <Xpa/PanedWin.h>		/* pour XtNresizable !! */
#include <Xpa/Row.h>
#include <Xpa/abspath.h>

#define _PAD_	2

#define TITLE_OFFSET		10
#define ACCELERATOR_OFFSET	30
#define MARGINS	(P.shadow_thickness + W.margin)

#define W	w -> menuBar
#define P	w -> primitive
#define C	w -> core

#define monochrome (((Widget)w)->core.depth == 1)

/*
 *	Les ressources de la barre de menuBar.
 */

static String	defItem [] = { "Title", "Item", NULL };
static String*	defMenu [] = { defItem, NULL };

static XtResource resources [] =
{
	{ XtNhidden, XtCHidden, XtRBoolean, sizeof (Boolean),
	  XtOffset(MenuBarWidget, menuBar.hidden), XtRImmediate, (XtPointer) False
	},
	{ XtNfilter, XtCFilter, XtRBoolean, sizeof (String),
	  XtOffset(MenuBarWidget, menuBar.filter), XtRImmediate, (XtPointer) NULL
	},
	{ XtNsuffixes, XtCSuffixes, XtRString, sizeof (String),
	  XtOffset(MenuBarWidget, menuBar.suffixes), XtRImmediate, (XtPointer) NULL
	},
	{ XtNitem, XtCItem, XtRMenuBar, sizeof (String **),
	  XtOffset(MenuBarWidget, menuBar.item), XtRImmediate, (XtPointer) defMenu
	},
	{ XtNreloadItem, XtCReloadItem, XtRBoolean, sizeof (Boolean),
	  XtOffset(MenuBarWidget, menuBar.reload_item), XtRImmediate, (XtPointer) False
	},
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
	  XtOffset(MenuBarWidget, menuBar.font), XtRString, "Fixed"
	},
	{ XtNhelp, XtCHelp, XtRBoolean, sizeof (Boolean),
	  XtOffset(MenuBarWidget, menuBar.help), XtRImmediate, (XtPointer) False
	},
	{ XtNcursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(MenuBarWidget, menuBar.cursor), XtRString, "arrow"
	},
	{ XtNmnemonicOn, XtCMnemonicOn, XtRBoolean, sizeof (Boolean),
	  XtOffset(MenuBarWidget, menuBar.mnemonic_on), XtRImmediate, (XtPointer) True
	},
	{ XtNmargin, XtCMargin, XtRInt, sizeof (int),
	  XtOffset(MenuBarWidget, menuBar.margin), XtRImmediate, (XtPointer) 2
	},
	{ XtNoblong, XtCOblong, XtRBoolean, sizeof (Boolean),
	  XtOffset (MenuBarWidget, menuBar.oblong), XtRImmediate, (XtPointer) False
	},
	{ XtNarmColor, XtCArmColor, XtRPixel, sizeof (Pixel),
	  XtOffset (MenuBarWidget, menuBar.arm_color),
	  XtRCallProc, (XtPointer) _GetArmColor
	},
	{ XtNacceleratorColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (MenuBarWidget, menuBar.accelerator_color),
	  XtRString, "black"
	},
};

static	Display*	dpy;
static	Window		root;
static	Screen*		screen;

static	void	createMenubarDialog (MenuBarWidget);
static	void	Initialize (Widget, Widget, Arg*, Cardinal*);
static	void	Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static	void	Destroy (Widget);
static	void	Redisplay (Widget, XEvent*, Region);
static	Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

static	void	shellExpose (Widget, XtPointer, XEvent*, Boolean*);
static	void	mnemonic_handler (Widget, XtPointer, XEvent*, Boolean*);
static	void	pullMenu (MenuBarWidget, int, int);
static	void	PaneRedraw (MenuBarWidget, int);


static	void	ptr_moved (Widget, XEvent*, String*, Cardinal*);
static	void	k_up (Widget, XEvent*, String*, Cardinal*);
static	void	k_down (Widget, XEvent*, String*, Cardinal*);
static	void	k_left (Widget, XEvent*, String*, Cardinal*);
static	void	k_right (Widget, XEvent*, String*, Cardinal*);
static	void	k_return (Widget, XEvent*, String*, Cardinal*);
static	void	k_unpost (Widget, XEvent*, String*, Cardinal*);

static XtActionsRec actionsList[] =
{
	{"drag",	ptr_moved },
	{"left",	k_left },
	{"right",	k_right },
	{"up",		k_up },
	{"down",	k_down },
	{"confirm",	k_return },
	{"unpost",	k_unpost },
};

static char defaultTranslations [] = "\
	<Key>Up:		up()\n\
	<Key>Down:		down()\n\
	<Key>Left:		left()\n\
	<Key>Right:		right()\n\
	<Key>Return:	confirm()\n\
	<Key>:			unpost()\n\
	<Btn1Down>:		drag()\n\
	Button1<PtrMoved>:	drag()\n\
	<Btn1Up>:		drag()";

/*
 *	La structure classe.
 */

MenuBarClassRec menuBarClassRec = {
	/* CoreClassPart */
	{
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"MenuBar",					/* class_name */
		sizeof (MenuBarRec),		/* widget_size */
		NULL,						/* class_initialize */
		NULL,						/* class_part_init */
		False,						/* class_inited */
		Initialize,					/* initialize */
		NULL,						/* initialize_hook */
		Realize,					/* realize */
		actionsList,				/* actions */
		XtNumber (actionsList),		/* num_actions */
		resources,					/* resources */
		XtNumber (resources),		/* num_resources */
		NULLQUARK,					/* xrm_class */
		True,						/* compress_motion */
		True,						/* compress_exposure */
		True,						/* compress_enterlv */
		False,						/* visible_interest */
		Destroy,					/* destroy */
		NULL,						/* resize */
		Redisplay,					/* expose */
		SetValues,					/* set_values */
		NULL,						/* set_values_hook */
		XtInheritSetValuesAlmost,	/* set_values_almost */
		NULL,						/* get_values_hook */
		NULL,						/* accept_focus */
		XtVersion,					/* version */
		NULL,						/* callback_private */
		defaultTranslations,		/* tm_table */
		XtInheritQueryGeometry,		/* query_geometry */
		XtInheritDisplayAccelerator,/* display accelerator */
		NULL,						/* extension */
	},
	/* PrimitiveClassPart */
	{
		NULL,						/* extension */
	},
	/* MenuBarClassPart */
	{
		NULL,						/* extension */
	},
};

WidgetClass menuBarWidgetClass = (WidgetClass)&menuBarClassRec;

/*
 *	le GC pour le label.
 */

static void get_labelGC (MenuBarWidget w)
{
	static Pixmap	greyStipple = (Pixmap)0;
	static Pixmap	greyPixmap = (Pixmap)0;
	static char		grey_bits [] = {
		0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55
	};
	XGCValues		values;
	unsigned long	valuemask;

	if (!greyStipple) {
		greyStipple = XCreateBitmapFromData (dpy, root, grey_bits, 8, 8);
		greyPixmap = XCreatePixmapFromBitmapData (dpy, root, grey_bits, 8, 8,
													BlackPixelOfScreen (screen),
													WhitePixelOfScreen (screen),
													DefaultDepthOfScreen (screen));
	}

	valuemask = GCForeground | GCBackground | GCFont | GCTile;
	values.tile = greyPixmap;
	values.foreground = P.foreground;
	values.background = C.background_pixel;
	values.font = W.font -> fid;
	W.label_GC = XtGetGC ((Widget) w, valuemask, &values);

	values.fill_style = FillStippled;
	values.stipple = greyStipple;
	valuemask |= GCFillStyle | GCStipple;
	W.label_insensitive_GC = XtGetGC ((Widget) w, valuemask, &values);
}

/*
 *	Trouve la ligne en pixels de l'item 'n' du panneau selecte.
 */

static int get_line_from_number (MenuBarWidget w, int n)
{
	String*	p;
	int		i;
	int		fh = P.shadow_thickness + W.bar_height;
	int		pixline = MARGINS;

	if (W.selected < 0)
		return 0;

	for (i = 1, p = W.item [W.selected] + 1; *p && i <= n; i++, p++)
		if (**p == '-' || **p == '=')
			pixline += fh/2 - 2 - (P.shadow_thickness % 2);
		else
			pixline += fh;
	return pixline;
}

/*
 *	Trouve l'item de la ligne en pixels 'line' du panneau selecte.
 */

static int get_number_from_line (MenuBarWidget w, int line)
{
	String*	p;
	int		i;
	int		fh = P.shadow_thickness + W.bar_height;
	int		pixline = MARGINS;

	if (W.selected < 0)
		return 0;

	for (i = 0, p = W.item [W.selected] + 1; *p && pixline < line; i++, p++)
		if (**p == '-' || **p == '=')
			pixline += fh/2 - 2;
		else
			pixline += fh;
	return i-1;
}

/*
 *	Calcule la taille de la barre.
 */

static void bar_resize (MenuBarWidget w)
{
	int				width;
	int				st2 = 2 * P.shadow_thickness;
	int				marg2 = 2 * W.margin;
	int				height = W.font_height + marg2;
	int				n;
	int*			ip = W.i_width;
	String**		p;
	XFontStruct*	fp = W.font;

	for (p = W.item, width = W.margin; *p; p++) {
		n = st2 + TITLE_OFFSET + XTextWidth (fp, **p, strlen (**p));
		*ip++ = n;
		width += n;
	}
	width -= TITLE_OFFSET;
	*ip = -1;

	if (C.width < 2 * MARGINS + width)
		C.width = 2 * MARGINS + width;
	C.height = 4 * st2 + height;
	W.bar_height = height;
}

/*
 *	Calcule la taille du menuBar.
 */

static void paneResize (MenuBarWidget w, int col, int mn)
{
	XFontStruct*	fp = W.font;
	String*			p;
	String			cp;
	int				num_item = 0;
	int				acc_width = 0;
	int				label_width = 0;
	int				aw, lw;
	int				fh = W.bar_height + P.shadow_thickness;
	int				ind = 0;
	int				h = MARGINS;

	for (p = W.item [mn] + 1; *p; p++, num_item++) {
		switch (**p) {
			case '-':
			case '=':
				h += fh/2 - 3;
				continue;

			case 1:		/* togglebutton diamond off */
			case 2:		/* togglebutton diamond on */
			case 3:		/* togglebutton square off */
			case 4:		/* togglebutton square on */
			case 5:		/* insensitive togglebutton */
				ind = W.font_height;
			case 6: 	/* insensitive button */

			default:
				if ((cp = index (*p, '\t')) != 0) {
					lw = ind + XTextWidth (fp, *p, cp - *p);
					aw = XTextWidth (fp, cp+1, strlen (cp+1));
				} else {
					lw = ind + XTextWidth (fp, *p, strlen (*p));
					aw = 0;
				}
				h += fh;
		}
		if (label_width < lw)	label_width = lw;
		if (acc_width < aw)	acc_width = aw;
	}

	if (acc_width)
		acc_width += ACCELERATOR_OFFSET;

	W.m_width [mn] = 4 * MARGINS + label_width + acc_width;
	W.m_acc_pos [mn] = label_width + ACCELERATOR_OFFSET;
	W.m_height [mn] = 2 * MARGINS + h;
	W.m_selected = -1;

	if (!XtIsRealized ((Widget) w))
		W.prev_pos = -1;
}

/*
 *	Initialise les menus deroulants.
 */

static void set_menu (MenuBarWidget w)
{
	int*	ip;
	int		col = MARGINS;
	int		n = 0;

	for (ip = W.i_width; *ip != -1; ip++) {
		paneResize (w, col, n++);
		col += *ip;
	}
}

static void mnemonic_handler (Widget p, XtPointer client, XEvent* e, Boolean* cont)
{
	MenuBarWidget	w = (MenuBarWidget) client;
	XKeyEvent*		event = (XKeyEvent*) e;
	int				i;
	Window			wjunk;
	int				xx,yy;

	for (i = 0; i < W.num_item; i++) {
		if ((event -> state & Mod1Mask) && event -> keycode == W.mnemonic_kc [i]) {
			XtGrabKeyboard ((Widget) w, False, GrabModeAsync, GrabModeAsync, event -> time);
			XtGrabPointer ((Widget) w, False,
							ButtonPressMask|ButtonReleaseMask|Button1MotionMask,
							GrabModeAsync, GrabModeAsync, None, W.cursor,
							event -> time);
			XTranslateCoordinates (dpy, XtWindow ((Widget) w), root, 0, 0, &xx, &yy, &wjunk);
			if (xx * 2 + yy * 3 != W.prev_pos) {
				set_menu (w);		/* (re)configurer le menu */
				W.prev_pos = xx * 2 + yy * 3;
			}
			pullMenu (w, i, 0);
			return;
		}
	}
}

/*
 *	Installe les mnemoniques
 */

static void set_mnemonic (MenuBarWidget w)
{
	char*		cp;
	String**	p;
	int			i;
	KeyCode		kc;

	W.mnemonic [0] = 0;
	for (i = 0, p = W.item; *p; p++, i++) {
		for (cp = **p; *cp; cp++) {
			if (!index (W.mnemonic, *cp)) {
				W.mnemonic [i] = *cp;
				W.mnemonic [i+1] = 0;
				W.mnemonic_pos [i] = cp - **p;
				kc = XKeysymToKeycode (dpy,  *cp);
				W.mnemonic_kc [i] = kc;
				if (C.managed)
					XtGrabKey (XtParent ((Widget) w), kc, Mod1Mask, False, GrabModeAsync, GrabModeAsync);
				break;
			}
		}
	}
	if (C.managed)
		XtAddEventHandler (XtParent ((Widget) w), KeyReleaseMask, False, mnemonic_handler, (XtPointer) w);
}

/*
 *	Retire les mnemoniques
 */

static void remove_mnemonic (MenuBarWidget w)
{
	char*	cp;
	KeyCode	kc;

	for (cp = W.mnemonic; *cp; cp++) {
		kc = XKeysymToKeycode (dpy, *cp);
		XtUngrabKey (XtParent ((Widget) w), kc, Mod1Mask);
	}
	XtRemoveEventHandler (XtParent ((Widget) w), KeyReleaseMask, False, mnemonic_handler, (XtPointer) w);
}

/*
 *	Fonction de creation de fenetre.
 */

static void Realize (Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	MenuBarWidget	w = (MenuBarWidget) widget;

	*value_mask |= CWCursor;
	attributes -> cursor = P.cursor;
	XtCreateWindow (widget, InputOutput, (Visual*) CopyFromParent, *value_mask, attributes);
	if (W.mnemonic_on)
		set_mnemonic ((MenuBarWidget) widget);
}

/*
 *	Fonction d'initialisation de chaque instance.
 */

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	static Boolean	first = True;
	MenuBarWidget	w = (MenuBarWidget) new;
	String**		mp;

	if (first) {
		first = False;
		dpy = XtDisplay (new);
		screen = XtScreen (new);
		root = RootWindowOfScreen (screen);
	}
	W.reload_item = False;
	/*
	 *	Calculer le nombre d'items de la barre.
	 */
	for (mp = W.item; *mp; mp++) ;
	W.num_item = mp - W.item;

	W.prev_pos = -1;
	W.mnemonic = XtMalloc (W.num_item + _PAD_);
	W.mnemonic_kc = (KeyCode*) XtMalloc ((W.num_item + _PAD_) * sizeof (KeyCode));
	W.mnemonic_pos = XtMalloc (W.num_item + _PAD_);
	W.i_width = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));
	W.m_width = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));
	W.m_acc_pos = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));
	W.m_height = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));

	W.font_height = W.font -> ascent + W.font -> descent + 1;
	W.selected = -1;
	W.m_selected = -1;
	/*
	 *	Creer la fenetre pour les menus deroulants.
	 */
	W.thePopupShell = XtVaCreatePopupShell ("panel", overrideShellWidgetClass, new,
										XtNbackgroundPixmap, 	C.background_pixmap,
										XtNbackground,			C.background_pixel,
										NULL);
	XtAddEventHandler (W.thePopupShell, ExposureMask, False, shellExpose, (XtPointer) w);
	get_labelGC (w);
	bar_resize (w);
	createMenubarDialog (w);
}

/*
 *	Affiche un item de menubar.
 */

static void MenuBarRedraw (MenuBarWidget w, int n)
{
	String**	p = W.item;
	int			i;
	int			st = P.shadow_thickness;
	int			col =  2 * MARGINS;
	int			line = MARGINS + 3 * st + W.font -> ascent;
	int*		ip = W.i_width;

	for (i = 0; i < n; i++) {
		p ++;
		col += *ip++;
	}

	if (W.help && !p [1])
		col = C.width - *ip + MARGINS;
	XDrawString (dpy, XtWindow ((Widget) w), W.label_GC,col, line, **p, strlen (**p));
	if (W.mnemonic_on)
		XFillRectangle (dpy, XtWindow ((Widget) w), W.label_GC,
						col + XTextWidth (W.font, **p, (int)W.mnemonic_pos [i]), line+2,
						XTextWidth (W.font, &W.mnemonic [i], 1), 1);
}

/*
 *	Allume une ligne du menuBar.
 */

static void MenuBarArm (MenuBarWidget w)
{
	int	st = P.shadow_thickness;
	int	item;
	int	x;

	if (W.selected != -1 && st > 0) {
		for (item = x = 0; item < W.selected; item++)
			x += W.i_width [item];

		if (W.help && W.selected == W.num_item - 1)
			x = C.width - st - W.i_width [item];

		if (!W.oblong) {
			_DrawShadow (dpy, XtWindow ((Widget) w),
							P.top_GC, P.bot_GC, st,
							W.i_width [item] - TITLE_OFFSET + 2 * W.margin,
							W.font_height + 2 * st,
							x + MARGINS,
							MARGINS + 2*st);
		} else {
			if (monochrome || W.arm_color != 12345) {
				if (monochrome)
					XSetFillStyle (dpy, W.label_GC, FillTiled);
				else
					XSetForeground (dpy, W.label_GC, W.arm_color);

				_FillOvalShadow (dpy, XtWindow ((Widget) w),
									W.label_GC,
									W.i_width [item] - TITLE_OFFSET + 2 * W.margin,
									W.font_height + 2 * st,
									x + MARGINS,
									MARGINS + 2*st);
				if (monochrome)
					XSetFillStyle (dpy, W.label_GC, FillSolid);
				else
					XSetForeground (dpy, W.label_GC, P.foreground);
			}
			_DrawOvalShadow	(dpy, XtWindow ((Widget) w),
								P.bot_GC, P.top_GC, 1,
								W.i_width [item] - TITLE_OFFSET + 2 * W.margin,
								W.font_height + 2 * st,
								x + MARGINS,
								MARGINS + 2*st);
			MenuBarRedraw (w, W.selected);
		}
	}
}

static void PaneArm (MenuBarWidget w)
{
	int	st = P.shadow_thickness;

	if (W.m_selected != -1 && st > 0) {
		if (!W.oblong) {
			_DrawShadow (dpy, XtWindow (W.thePopupShell),
							P.top_GC, P.bot_GC, st,
							W.m_width [W.selected] - 2 * MARGINS,
							W.font_height + 2 * MARGINS,
							MARGINS,
							get_line_from_number (w, W.m_selected));
		} else {
			if (monochrome || W.arm_color != 12345) {
				if (monochrome)
					XSetFillStyle (dpy, W.label_GC, FillTiled);
				else
					XSetForeground (dpy, W.label_GC, W.arm_color);

				_FillOvalShadow (dpy, XtWindow (W.thePopupShell),
									W.label_GC,
									W.m_width [W.selected] - 2 * MARGINS,
									W.font_height + MARGINS,
									MARGINS,
									1 + get_line_from_number (w, W.m_selected));
				if (monochrome)
					XSetFillStyle (dpy, W.label_GC, FillSolid);
				else
					XSetForeground (dpy, W.label_GC, P.foreground);
			}
			_DrawOvalShadow (dpy, XtWindow (W.thePopupShell),
								P.bot_GC, P.top_GC, 1,
								W.m_width [W.selected] - 2 * MARGINS,
								W.font_height + MARGINS,
								MARGINS,
								1 + get_line_from_number (w, W.m_selected));
			PaneRedraw (w, W.m_selected);
		}
	}
}

/*
 *	Eteint une ligne du menuBar.
 */

static void MenuBarDisarm (MenuBarWidget w)
{
	int	st = P.shadow_thickness;
	int	item;
	int	x;

	if (W.selected != -1 && st > 0) {
		for (item = x = 0; item < W.selected; item++)
			x += W.i_width [item];

		if (W.help && W.selected == W.num_item - 1)
			x = C.width - st - W.i_width [item];

		if (!W.oblong) {
			_EraseShadow (dpy, XtWindow ((Widget) w), st,
							W.i_width [item] - TITLE_OFFSET + 2 * W.margin,
							W.font_height + 2 * st,
							x + MARGINS,
							MARGINS + 2*st);
		} else {
			XClearArea (dpy, XtWindow ((Widget) w),
							x + MARGINS,
							MARGINS + 2*st,
							W.i_width [item] - TITLE_OFFSET + 2 * W.margin + W.margin,
							W.font_height + 2 * st + W.margin, False);
			MenuBarRedraw (w, W.selected);
		}
	}
}

static void PaneDisarm (MenuBarWidget w)
{
	int	st = P.shadow_thickness;

	if (W.m_selected != -1 && st > 0) {
		if (!W.oblong) {
			_EraseShadow (dpy, XtWindow (W.thePopupShell), st,
							W.m_width [W.selected] - 2 * MARGINS,
							W.font_height + 2 * MARGINS,
							MARGINS,
							get_line_from_number (w, W.m_selected));
		} else {
			XClearArea (dpy, XtWindow (W.thePopupShell),
							MARGINS,
							1 + get_line_from_number (w, W.m_selected),
							W.m_width [W.selected] - MARGINS - W.margin,
							W.font_height + MARGINS + 1, False);
			PaneRedraw (w, W.m_selected);
		}
	}
}

/*
 *	Reaffiche tout le menuBar.
 */

static void bar_refresh (MenuBarWidget w)
{
	String**	p;
	int			i;
	int			col, line;
	int*		ip = W.i_width;
	int			st = P.shadow_thickness;
	Window		window = XtWindow ((Widget) w);

	col = 2 * MARGINS;
	line = MARGINS + 3 * st + W.font -> ascent;

	for (i = 0, p = W.item; *p; p++, col += *ip++, i++) {
		if (W.help && !p [1])
			col = C.width - *ip + MARGINS;

		XDrawString (dpy, window, W.label_GC, col, line, **p, strlen (**p));
		if (W.mnemonic_on)
			XFillRectangle (dpy, window, W.label_GC,
							col + XTextWidth (W.font, **p, (int)W.mnemonic_pos [i]), line+2,
							XTextWidth (W.font, &W.mnemonic [i], 1), 1);
	}
	MenuBarArm (w);
}

/*
 *	Reaffiche les mnemoniques.
 */

static void show_mnemonics (MenuBarWidget w)
{
	String**	p;
	int			i;
	int*		ip = W.i_width;
	int			st = P.shadow_thickness;
	int			col = 2 * MARGINS;
	int			line = MARGINS + 3 * st + W.font -> ascent;

	if (!XtIsRealized ((Widget) w))
		return;

	for (i = 0, p = W.item; *p; p++, col += *ip++, i++) {
		if (W.help && !p [1])
			col = C.width - *ip + MARGINS;
		if (W.mnemonic_on)
			XFillRectangle (dpy, XtWindow ((Widget) w), W.label_GC,
							col + XTextWidth (W.font, **p, (int)W.mnemonic_pos [i]), line+2,
							XTextWidth (W.font, &W.mnemonic [i], 1), 1);
		else
			XClearArea (dpy, XtWindow ((Widget) w),
							col + XTextWidth (W.font, **p, (int)W.mnemonic_pos [i]), line+2,
							XTextWidth (W.font, &W.mnemonic [i], 1), 1, False);
	}
}

/*
 *	La fonction de reaffichage du panneau de menu.
 */

static void PaneRedraw (MenuBarWidget w, int n)
{
	String*	p;
	String	cp;
	String	pp;
	int		st = P.shadow_thickness;
	int		height = W.bar_height;
	int		fh = height + st;
	int		pixline = st + MARGINS + height / 2;
	int		line = st + MARGINS + W.font -> ascent + (height - W.font_height) / 2;
	int		ind = 0;
	int		i;
	Window	win = XtWindow (W.thePopupShell);
	GC		gc;

	if (W.selected == -1)
		return;

	for (i = 1, p = W.item [W.selected] + 1; *p ; i++, p++) {
		switch (**p) {
		case 1: case 2: case 3: case 4:
		case 5:
			ind = W.font_height;
			break;
		case '-':
		case '=':
			if (i <= n) {
				line -= fh/2 + 3;
				pixline -= fh/2 + 3;
			}
			break;
		}
		if (i <= n) {
			line += fh;
			pixline += fh;
		}
	}
	pp = * (W.item [W.selected] + 1 + n);
	gc = W.label_GC;
	/*
	 *	Aligner  droite le dernier mot (premiere tabulation)
	 *	et aligner  gauche le debut de l'item.
	 */
	switch (*pp) {
	case 4:
		_DrawShadow (dpy, win,
						P.top_GC, P.bot_GC, st,
						(ind * 2) / 3 - 2, (ind * 2) / 3 - 2,
						2 * MARGINS,
						line - W.font -> ascent + (ind / 6));
		pp++;
		break;

	case 2:
		XSetForeground (dpy, W.label_GC,
			W.arm_color != 12345 ? W.arm_color : C.background_pixel);
		_DrawDiamondShadow (dpy, win,
				/* x */		2 * MARGINS,
				/* y */		line - W.font -> ascent + (ind / 6),
				/* size */		(ind * 2) / 3,
				/* gc's */		P.top_GC, P.bot_GC, W.label_GC,
				/* fill */		True);
		XSetForeground (dpy, W.label_GC, P.foreground);
		/* fall thru */
	case 1: case 3:
		pp++;
		break;
	case 5: case 6:
		gc = W.label_insensitive_GC;
		pp++;
		break;
	}
	cp = index (pp, '\t');
	if (cp == NULL || cp [1] == 0)
		XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, strlen (pp));
	else {
		XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, cp - pp);
		XSetForeground (dpy, gc, W.accelerator_color);
		XDrawString (dpy, win, gc,
			2 * MARGINS + W.m_acc_pos [W.selected],
			line, cp+1, strlen (cp+1));
		XSetForeground (dpy, gc, P.foreground);
	}
}

/*
 *	Raffichage du panneau de menu.
 */

static void shellExpose (Widget shell, XtPointer client, XEvent* event, Boolean* cont)
{
	MenuBarWidget	w = (MenuBarWidget) client;
	XExposeEvent*	e = (XExposeEvent*) event;
	XRectangle		rect;
	String*			p;
	String			cp;
	String			pp;
	int				st = P.shadow_thickness;
	int				pixline, line, fh;
	int				height;
	int				ind;
	Window			win = XtWindow (shell);
	GC				gc;

	if (W.selected == -1)
		return;

	height = W.bar_height;

	rect.x = e -> x;
	rect.y = e -> y;
	rect.width = e -> width;
	rect.height = e -> height;
	XSetClipRectangles (dpy, P.bot_GC, 0, 0, &rect, 1, Unsorted);
	XSetClipRectangles (dpy, P.top_GC, 0, 0, &rect, 1, Unsorted);
	XSetClipRectangles (dpy, W.label_GC, 0, 0, &rect, 1, Unsorted);
	XSetClipRectangles (dpy, W.label_insensitive_GC, 0, 0, &rect, 1, Unsorted);

	_DrawShadow (dpy, win,
					P.top_GC, P.bot_GC, st,
					W.m_width [W.selected], W.m_height [W.selected], 0, 0);

	fh = height + st;
	line = st + MARGINS + W.font -> ascent + (height - W.font_height) / 2;
	pixline = st + MARGINS + height / 2;

	for (ind = 0, p = W.item [W.selected] + 1; *p ; p++)
		if (**p >= 1 && **p <= 5) {
			ind = W.font_height;
			break;
		}

	for (p = W.item [W.selected] + 1; (pp = *p) != 0 ; line += fh, pixline += fh, p++) {
		gc = W.label_GC;
		switch (*pp) {
		case '=':
			_DrawShadow (dpy, win, 
							P.bot_GC, P.top_GC, st,
							W.m_width [W.selected] - 2 * MARGINS, 2,
							MARGINS, pixline - fh/2 + 3);
			_DrawShadow (dpy, win, 
							P.bot_GC, P.top_GC, st,
							W.m_width [W.selected] - 2 * MARGINS, 2,
							MARGINS, pixline - fh/2 + 5);
			line -= fh/2 + 3;
			pixline -= fh/2 + 3;
			continue;

		case '-':
			_DrawShadow (dpy, win, 
							P.bot_GC, P.top_GC, st,
							W.m_width [W.selected] - 2 * MARGINS, 2,
							MARGINS, pixline - fh/2 + 4);
			line -= fh/2 + 3;
			pixline -= fh/2 + 3;
			continue;

		case 4:
			_DrawShadow (dpy, win,
							P.top_GC, P.bot_GC, st,
							(ind * 2) / 3 - 2, (ind * 2) / 3 - 2,
							2 * MARGINS,
							line - W.font -> ascent + (ind / 6));
			pp++;
			break;

		case 2:
			XSetForeground (dpy, W.label_GC,
				W.arm_color != 12345 ? W.arm_color : C.background_pixel);
			_DrawDiamondShadow (dpy, win,
					/* x */		2 * MARGINS,
					/* y */		line - W.font -> ascent + (ind / 6),
					/* size */		(ind * 2) / 3,
					/* gc's */		P.top_GC, P.bot_GC, W.label_GC,
					/* fill */		True);
			XSetForeground (dpy, W.label_GC, P.foreground);
			pp ++;
			break;

		case 1: case 3:
			pp ++;
			break;

		case 5: case 6:
			gc = W.label_insensitive_GC;
			pp++;
			break;

		default:
			break;
		}
		/*
		 *	Aligner  droite le dernier mot (premiere tabulation)
		 *	et aligner  gauche le debut de l'item.
		 */
		cp = index (pp, '\t');
		if (cp == NULL || cp [1] == 0)
			XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, strlen (pp));
		else {
			XDrawString (dpy, win, gc, ind + 2 * MARGINS, line, pp, cp - pp);
			XSetForeground (dpy, gc, W.accelerator_color);
			XDrawString (dpy, win, gc,
				2 * MARGINS + W.m_acc_pos [W.selected],
				line, cp+1, strlen (cp+1));
			XSetForeground (dpy, gc, P.foreground);
		}
	}
	PaneArm (w);
	/*
	 *	Virer la region de clipping.
	 */
	XSetClipMask (dpy, P.top_GC, None);
	XSetClipMask (dpy, P.bot_GC, None);
	XSetClipMask (dpy, W.label_GC, None);
	XSetClipMask (dpy, W.label_insensitive_GC, None);
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	MenuBarWidget	w = (MenuBarWidget) widget;

	XSetRegion (dpy, P.top_GC, region);
	XSetRegion (dpy, P.bot_GC, region);
	XSetRegion (dpy, W.label_GC, region);

	_DrawShadow (dpy, XtWindow (widget),
					P.top_GC, P.bot_GC, P.shadow_thickness,
					C.width, C.height, 0, 0);

	bar_refresh (w);
	/*
	 *	Virer la region de clipping.
	 */
	XSetClipMask (dpy, P.top_GC, None);
	XSetClipMask (dpy, P.bot_GC, None);
	XSetClipMask (dpy, W.label_GC, None);
}

/*
 *	SetValues.
 */

static	Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	MenuBarWidget	cu = (MenuBarWidget) current;
	MenuBarWidget	w = (MenuBarWidget) new;
	Boolean			do_redisplay = False;
	Boolean			mnemonic_done = False;
	Boolean			bar_resize_done = False;
	Boolean			set_menu_done = False;

	if (cu -> core.background_pixel != C.background_pixel) {
		XrmValue	value;

		_GetArmColor (new, 0, &value);
		W.arm_color = *(Pixel*)value.addr;
		XtVaSetValues (W.thePopupShell, XtNbackground, C.background_pixel, NULL);
	}

	if (cu -> menuBar.item != W.item || W.reload_item) {
		String**	mp;

		W.reload_item = False;

		XtFree (W.mnemonic_pos);
		XtFree ((char*) W.mnemonic_kc);
		XtFree ((char*) W.i_width);
		XtFree ((char*) W.m_width);
		XtFree ((char*) W.m_acc_pos);
		XtFree ((char*) W.m_height);

		for (mp = W.item; *mp; mp++)
			;
		W.num_item = mp - W.item;

		W.prev_pos = -1;
		W.selected = -1;
		W.m_selected = -1;

		W.mnemonic_kc = (KeyCode*) XtMalloc ((W.num_item + _PAD_) * sizeof (KeyCode));
		W.mnemonic_pos = XtMalloc (W.num_item + _PAD_);
		W.i_width = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));
		W.m_width = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));
		W.m_acc_pos = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));
		W.m_height = (int*) XtMalloc ((W.num_item + _PAD_) * sizeof (int));

		bar_resize (w);

		if (W.mnemonic_on) {
			remove_mnemonic (w);
			XtFree (W.mnemonic);
			W.mnemonic = XtMalloc (W.num_item + _PAD_);
			set_mnemonic (w);
		} else {
			XtFree (W.mnemonic);
			W.mnemonic = XtMalloc (W.num_item + _PAD_);
		}

		mnemonic_done = True;	/* on a deja teste 'mnemonic_on' */
		bar_resize_done = True;	/* on a dj fait le resize */
		do_redisplay = True;
	}

	if (cu -> primitive.shadow_thickness != P.shadow_thickness) {
		if (! bar_resize_done) {
			bar_resize_done = True;
			bar_resize (w);
		}
		if (! set_menu_done) {
			set_menu_done = True;
			set_menu (w);
		}
		do_redisplay = True;
	}

	if (cu -> menuBar.font != W.font) {
		XtDestroyGC (cu -> menuBar.label_GC);
		get_labelGC (w);
		if (! bar_resize_done) {
			bar_resize_done = True;
			bar_resize (w);
		}
		if (! set_menu_done) {
			set_menu_done = True;
			set_menu (w);
		}
		do_redisplay = True;
	}

	if (! mnemonic_done && cu -> menuBar.mnemonic_on != W.mnemonic_on) {
		if (W.mnemonic_on)
			set_mnemonic (w);
		else
			remove_mnemonic (w);

		if (!do_redisplay)
			show_mnemonics (w);
	}

	return do_redisplay;
}

/*
 *	La procedure de destruction d'une instance.
 */

static void Destroy (Widget widget)
{
	MenuBarWidget	w = (MenuBarWidget) widget;

	XtDestroyGC (W.label_GC);
	XtFree ((char*) W.i_width);
	XtFree ((char*) W.m_width);
	XtFree ((char*) W.m_height);
}

/*
 *	Traite les evenements.
 */

static int whichTitle (MenuBarWidget w, XButtonEvent* e)
{
	int		cur_item;
	int*	ip;
	int		st = P.shadow_thickness + W.margin;
	int		mx = e -> x;
	int		my = e -> y;
	/*
	 *	Dans la barre de menu ?
	 */
	if (my < st || my >= C.height - 2*st || mx < st || mx >= C.width - 2*st)
		return -1;
	/*
	 *	Oui, dans quel item ?
	 */
	ip = W.i_width;
	cur_item = W.num_item - 1;

	if (W.help && (C.width - ip [cur_item]) <= mx && mx < (C.width - st))
		return cur_item;

	for (cur_item = 0; mx > *ip; cur_item ++) {
		if ((W.help && cur_item == W.num_item - 2) ||
		    (cur_item >= W.num_item))
			return -1;
		mx -= *ip++;
	}
	return cur_item;
}

static void pullMenu (MenuBarWidget w, int cur_item, int m)
{
	int		col, i;
	Window	wjunk;
	int		xx, yy;

	if (cur_item == W.selected)
		return;

	if (W.selected != -1)
		MenuBarDisarm (w);	/* Eteindre l'ancien titre */

	W.selected = cur_item;
	MenuBarArm (w);			/* Allumer le titre choisi */
	XTranslateCoordinates (dpy, XtWindow ((Widget) w), root, 0, 0, &xx, &yy, &wjunk);
	if (XtIsRealized (W.thePopupShell))
		XtPopdown (W.thePopupShell);	/* Pour que le save_under fonctionne */
	/*
	 *	La colonne ou se trouve le titre.
	 */
	if (W.help && cur_item == W.num_item - 1)
		/* ne pas faire dborder le panneau  droite de la fentre */
		col = C.width - W.m_width [cur_item];
	else
		for (col = i = 0; i < cur_item; i++)
			col += W.i_width [i];

	XtConfigureWidget (W.thePopupShell,
						col + xx, C.height + yy,
						W.m_width [cur_item], W.m_height [cur_item],
						0);
	W.m_selected = -1;
	if (m != -1) {
		char	c;

		while (W.item [cur_item][m+1]) {
			if ((c = W.item [cur_item] [m+1] [0]) == '-' ||
				 c == '=' ||
				 c == 5 ||
				 c == 6)
			{
				m = m + 1;
				continue;
			}
			W.m_selected = m;
			break;
		}
	}
	/*
	 *	Appeler la callback de prepostage du menu.
	 */
	XtCallCallbackList ((Widget) w, P.release, (XtPointer) (long) cur_item);
	XtPopup (W.thePopupShell, XtGrabNone);
}

static void armMenu (MenuBarWidget w, int cur_item)
{
	char*	label = W.item [W.selected] [cur_item + 1];
	char	c;

	if (*label == '-' || *label == '=')
		return;

	if (cur_item != W.m_selected) {
		if (W.m_selected != -1) {
			c = W.item [W.selected] [W.m_selected+1][0];
			if (c <= 4 || c >= 8)
				PaneDisarm (w);	/* effacer l'ancien */
		}

		W.m_selected = cur_item;
		c = *label;
		if (c <= 4 || c >= 8)
			PaneArm (w);	/* allumer le nouveau */
	}
}

static void k_left (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	MenuBarWidget	w = (MenuBarWidget) widget;
	int				cur_item;

	cur_item = W.selected;
	if (cur_item == -1)
		return;		/* pas de menu affiche */

	if (cur_item == 0)
		cur_item = W.num_item - 1;
	else
		cur_item --;
	pullMenu (w, cur_item, 0);
}

static void k_right (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	MenuBarWidget	w = (MenuBarWidget) widget;
	int				cur_item;

	cur_item = W.selected;
	if (cur_item == -1)
		return;		/* pas de menu affiche */

	if (cur_item == W.num_item - 1)
		cur_item = 0;
	else
		cur_item ++;
	pullMenu (w, cur_item, 0);
}

static void k_down (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	MenuBarWidget	w = (MenuBarWidget) widget;
	int				cur_item;
	int				n;
	int				i;
	String*			p;
	char			c;

	cur_item = W.m_selected;
	if (cur_item == -1)
		return;		/* pas d'item selectionne */

	for (n = 0, p = W.item [W.selected] + 1; *p; p++)
		n++;
	if (n == 0)
		return;		/* menu vide !! */
	cur_item = (cur_item + 1) % n;
	for (i = n; --i >=0;) {
		if ((c = W.item [W.selected] [cur_item + 1] [0]) == '-' ||
			 c == '=' ||
			 c == 5 ||
			 c == 6)
			cur_item = (cur_item + 1) % n;
		else {
			armMenu (w, cur_item);
			return;
		}
	}
}

static void k_up (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	MenuBarWidget	w = (MenuBarWidget) widget;
	int				cur_item;
	int				n;
	int				i;
	char			c;
	String*			p;

	cur_item = W.m_selected;
	if (cur_item == -1)
		return;		/* pas d'item selectionne */

	for (n = 0, p = W.item [W.selected] + 1; *p; p++)
		n++;
	if (n == 0)
		return;		/* menu vide !! */
	cur_item = (cur_item + n - 1) % n;
	for (i = n; --i >=0;) {
		if ((c = W.item [W.selected] [cur_item + 1] [0]) == '-' ||
			 c == '=' ||
			 c == 5 ||
			 c == 6)
			cur_item = (cur_item + n - 1) % n;
		else {
			armMenu (w, cur_item);
			return;
		}
	}
}

static void k_return (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	MenuBarWidget	w = (MenuBarWidget) widget;
	XKeyEvent*		event = (XKeyEvent*) e;

	if (W.selected == -1 || W.m_selected == -1)
		return;

	MenuBarDisarm (w);		/* eteindre l'item */
	XtPopdown (W.thePopupShell);	/* retirer le menu */
	XtUngrabKeyboard (widget, event -> time);
	XtUngrabPointer (widget, event -> time);
	while (XtAppPending (XtWidgetToApplicationContext (widget)))
		XtAppProcessEvent (XtWidgetToApplicationContext (widget), XtIMAll);
	if (W.m_selected != -1 && W.item [W.selected][W.m_selected+1]) {
		switch (W.item [W.selected][W.m_selected+1][0]) {
		case 1:
			W.item [W.selected][W.m_selected+1][0] = 2;
			break;
		case 2:
			W.item [W.selected][W.m_selected+1][0] = 1;
			break;
		case 3:
			W.item [W.selected][W.m_selected+1][0] = 4;
			break;
		case 4:
			W.item [W.selected][W.m_selected+1][0] = 3;
			break;
		case 5:
		case 6:
			W.selected = -1;
			return;
		}
		XtCallCallbackList (widget, P.select, (XtPointer) (long) ((W.selected << 16) + W.m_selected + 1));
	}
	W.selected = -1;
}

static void k_unpost (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	MenuBarWidget	w = (MenuBarWidget) widget;
	XKeyEvent*		event = (XKeyEvent*) e;
	char			c;

	if (W.selected == -1)
		return;

	if ((W.mnemonic_on &&
	     event -> keycode == W.mnemonic_kc [W.selected] &&
	     (event -> state & Mod1Mask))
	    || (XLookupString (event, &c, 1, NULL, NULL) && c == 033))
	{
		MenuBarDisarm (w);		/* eteindre l'item */
		XtPopdown (W.thePopupShell);	/* retirer le menu */
		XtUngrabKeyboard (widget, event -> time);
		XtUngrabPointer (widget, event -> time);
		W.selected = -1;
	}
}

static void ptr_moved (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	MenuBarWidget	w = (MenuBarWidget) widget;
	XButtonEvent*	event = (XButtonEvent*) e;
	int				st = P.shadow_thickness + W.margin;
	int				cur_item;
	static Boolean	pressed = False;
	Window			wjunk;
	int				junk;
	unsigned int	ujunk;
	int				xx,yy;
	int				mx, my;

	switch (event -> type) {
	case ButtonRelease:
		if (W.selected == -1) {
			pressed = False;
			return;
		}

		if (W.selected != whichTitle (w, event)) {
			MenuBarDisarm (w);		/* eteindre l'item */
			XtPopdown (W.thePopupShell);	/* retirer le menu */
			XtUngrabKeyboard (widget, event -> time);
			XtUngrabPointer (widget, event -> time);
			XSync (XtDisplay (widget), False);
			while (XtAppPending (XtWidgetToApplicationContext (widget)))
				XtAppProcessEvent (XtWidgetToApplicationContext (widget), XtIMAll);
			if (W.m_selected != -1) {
				switch (W.item [W.selected][W.m_selected+1][0]) {
				case 1:
					W.item [W.selected][W.m_selected+1][0] = 2;
					break;
				case 2:
					W.item [W.selected][W.m_selected+1][0] = 1;
					break;
				case 3:
					W.item [W.selected][W.m_selected+1][0] = 4;
					break;
				case 4:
					W.item [W.selected][W.m_selected+1][0] = 3;
					break;
				case 5:
				case 6:
					W.selected = -1;
					pressed = False;
					return;
				}
				XtCallCallbackList (widget, P.select, (XtPointer) (long)((W.selected << 16) + W.m_selected + 1));
			}
			W.selected = -1;	/* degager la selection */
		} else {
			char	c;
			int		m = 0;

			W.m_selected = -1;
			while (W.item [W.selected][m+1]) {
				if ((c = W.item [W.selected] [m+1] [0]) == '-' ||
					 c == '=' ||
					 c == 5 ||
					 c == 6)
				{
					m = m + 1;
					continue;
				}
				W.m_selected = m;
				break;
			}
			PaneArm (w);
			XtGrabPointer (widget, False,
							ButtonPressMask|ButtonReleaseMask|Button1MotionMask,
							GrabModeAsync, GrabModeAsync, None, W.cursor,
							event -> time);
			XtGrabKeyboard (widget, False, GrabModeAsync, GrabModeAsync, event -> time);
		}
		pressed = False;
		return;

	case ButtonPress:
		XTranslateCoordinates (dpy, XtWindow (widget), root, 0, 0, &xx, &yy, &wjunk);
		if (xx * 2 + yy * 3 != W.prev_pos) {
			set_menu (w);		/* reconfigurer le menu */
			W.prev_pos = xx * 2 + yy * 3;
		}

		pressed = True;
		cur_item = whichTitle (w, event);
		if (W.selected != -1 && W.selected != cur_item) {
			if (W.m_selected == -1)
				MenuBarDisarm (w);
		} else
			W.selected = -1;

		if (W.m_selected != -1) {
			PaneDisarm (w);
			W.m_selected = -1;
		}

		if (cur_item != -1) {
			pullMenu (w, cur_item, -1);
			return;
		}

		if (XtIsRealized (W.thePopupShell))
			goto bpress;
		return;

	case MotionNotify:
		if (!pressed) return;	/* pas le bon */

		cur_item = whichTitle (w, event);
		if (cur_item != -1) {
			pullMenu (w, cur_item, -1);
			return;
		}

		if (W.selected == -1)
			return;

	bpress:	XQueryPointer (dpy, XtWindow (W.thePopupShell), &wjunk, &wjunk, &junk, &junk, &mx, &my, &ujunk);
		if (mx < st || my <= st || mx > W.m_width [W.selected] || my >= W.m_height [W.selected] - 2 * st) {
			PaneDisarm (w);	/* virer la selection */
			W.m_selected = -1;
			return;		/* en dehors du menuBar */
		}
		/*
		 *	Dans un panneau.
		 */
		cur_item = get_number_from_line (w, my - P.shadow_thickness);
		armMenu (w, cur_item);
		return;
	}
}

/*
 *	La bote de dialogue associe  la barre de menus.
 */

#define T_EXECUTABLE	0
#define T_DIRECTORY		1
#define T_HIGHLIGHT		2
#define T_FILE			3

#define MAXLEN		40		/* taille max d'un nom de fichier */
#define MAXNAMES	256		/* nombre max de noms affichs */

/*
 *	Fabrique un tableau de strings avec les noms des fichiers du directory spcifi.
 */

static int compare (const void* a, const void* b)
{
	return strcmp ((*(char**) a)+1, (*(char**) b)+1);
}

static void show_window (MenuBarWidget w, char** list, int nb, int maxlen)
{
	int				height;
	int				ph, rh;
	static Pixel	colors [T_FILE+1];
	int				width;
	int				c;
	int				l;
	int				ncol;
	int				wcol;
	int				nl;
	static GC		gc;
	FontInfo*		f;

	if (nb < 2) {
		if (W.list_win != 0)
			XUnmapWindow (dpy, W.clip_win);
		return;
	}
	/*
	 *	Crer la fentre la premire fois (et le gc aussi).
	 */
	if (!W.list_win) {
		XSetWindowAttributes	xswa;

		if (!gc) {
			XColor	xcs, xce;

			gc = XCreateGC (dpy, root, 0, NULL);
			(void) XAllocNamedColor (dpy, DefaultColormapOfScreen (screen),
										 "#00c", &xcs, &xce);
			colors [T_DIRECTORY] = xcs.pixel;
			(void) XAllocNamedColor (dpy, DefaultColormapOfScreen (screen),
										 "#0a0", &xcs, &xce);
			colors [T_EXECUTABLE] = xcs.pixel;
			(void) XAllocNamedColor (dpy, DefaultColormapOfScreen (screen),
										 "maroon", &xcs, &xce);
			colors [T_HIGHLIGHT] = xcs.pixel;
			colors [T_FILE] = BlackPixelOfScreen (screen);
		}

		xswa.border_pixel = BlackPixelOfScreen (screen);
		xswa.background_pixel = WhitePixelOfScreen (screen);
		xswa.backing_store = WhenMapped;
		W.clip_win = XCreateWindow (dpy,
								XtWindow (C.parent),
								70, 5, 10, 10, 1,
								CopyFromParent, CopyFromParent, CopyFromParent,
								CWBackPixel|CWBorderPixel,
								&xswa);
		xswa.border_pixel = colors [T_DIRECTORY];
		W.list_win = XCreateWindow (dpy,
								W.clip_win,
								70, 5, 10, 10, 3,
								CopyFromParent, CopyFromParent, CopyFromParent,
								CWBorderPixel|CWBackPixel|CWBackingStore,
								&xswa);
		XMapWindow (dpy, W.list_win);
	}
	/* trier la liste par ordre alphabtique */
	qsort (list, nb, sizeof (char*), compare);
	/* rcuprer la font du wix (probablement  chasse fixe) */
	XtVaGetValues (W.exp_wix, XtNfont, &f, NULL);
	/* calculer la taille et le nombre de colonnes */
	wcol = (maxlen+1) * XTextWidth (f -> f_info, "n", 1);
	/* utiliser la largeur du parent (panedWindow) car la barre n'a peut tre pas sa taille adulte */
	width = C.parent -> core.width;
	ncol = (width - 10) / wcol;
	if (ncol < 1)
		ncol = 1;
	/* ajuster la largeur en fonction du nombre de colonnes */
	width = 10 + ncol * wcol;
	/* le nombre de lignes */
	nl = (nb+ncol-1) / ncol;
	/* retailler les fentres */
	rh = 2 + nl * f -> f_height;		/* hauteur du parent */
	ph = C.parent -> core.height - 40;	/* hauteur ncessaire */

	if (rh > ph) {
		height = 2 + (((ph-2) / f -> f_height) * f -> f_height);
		W.min_clip_y = height - rh + 1;	/* jusqu'o on peut scroller */
		height += 3;
	} else {
		height = rh + 6;
		W.min_clip_y = 0;		/* dans ce cas on ne scrolle pas! */
	}
	W.clip_y = 0;

	XMoveResizeWindow (dpy, W.clip_win,
					(C.parent -> core.width - width) / 2, C.y + C.height + 5,
					width - 10, height);
	XMoveResizeWindow (dpy, W.list_win, -3, W.clip_y, width - 10, rh);
	XClearWindow (dpy, W.list_win);
	XMapWindow (dpy, W.clip_win);
	XSetFont (dpy, gc, f -> f_info -> fid);
	for (l = 0; l < nl; l++) {
		int 	x = 5;
		int		prev = -1;

		for (c = 0; c+l < nb; c += nl) {
			if (list [c+l][0] != prev) {
				prev = list [c+l][0];
				XSetForeground (dpy, gc, colors [prev]);
			}
			XDrawString (dpy, W.list_win, gc,
						 x, 1 + l * f -> f_height + f -> f_ascent,
						 list [c+l]+1, strlen (list [c+l]+1));
			x += wcol;
		}
	}
}

/*
 *	Cherche si 's' matche l'une des regexprs (spares par des espaces) de 're'
 */

static Boolean match_suffixes (char* re, char* s)
{
	char	tc [256];
	char*	p1 = tc;
	char*	p2 = tc;

	for (;;) {
		switch (*p1++ = *re++) {
			case 0:
				return FilenameMatch (p2, s);

			case ' ':
				/* sauter les espaces contigus */
				for (p1 [-1] = 0; *p1 == ' '; ++p1)
					;
				if (FilenameMatch (p2, s))
					return True;
				p2 = p1;
		}
	}
}


static int list_files (MenuBarWidget w, char* directory, char* prefix, Boolean dotdot)
{
#if SYS_DIR
	struct direct*	item;
#else
	struct dirent*	item;
#endif
	int			len = strlen (prefix);
	char		pref [256];
	char*		list [MAXNAMES];
	char		buffer [256*MAXLEN];	/* comme on ne retient que les MAXLEN premiers */
	char*		buf_end = buffer;		/* caractres des noms, a suffit amplement. */
	DIR*		dirp;
	char*		name;
	char*		p;
	char*		mp;
	int			nb;
	int			length;
	int			maxlen;
	struct stat	b;
	char		tc [256];

	if (directory [0] == 0)
		directory = "/";			/* chemin absolu */
	else
		directory = AbsPath (W.directory, directory);
	if (-1 == access (directory, R_OK | X_OK) || !(dirp = opendir (directory)))
		return -1;			/* drle de directory! */

	strcpy (pref, prefix);		/* sauver le prfixe initial */
#if 0
	/* anciens filesystems: . et .. taient les deux premires entres */
	(void) readdir (dirp);		/* sauter . */
	if (!dotdot)
		(void) readdir (dirp);	/* sauter .. */
#endif
	maxlen = 0;
	for (nb = 0, item = readdir (dirp); item != NULL; item = readdir (dirp)) {
		name = item -> d_name;
		length = strlen (name);
#if 1
		if (name [0] == '.' && name [1] == 0)
			continue;
		
		if (!dotdot && name [0] == '.' && name [1] == '.' && name [2] == 0)
			continue;
#endif
		if (name [length-1] == 'o' && name [length-2] == '.')
			continue;	/* sauter les .o */
		if (len > 0 && bcmp (name, pref, len))
			continue;
		/*
		 *	Accumuler les noms dans une liste.
		 */
		if (nb < MAXNAMES && buf_end + length < & buffer [sizeof (buffer)]) {
			list [nb++] = buf_end;
			sprintf (tc, "%s/%s", directory, name);
			stat (tc, &b);	/* est-ce un repertoire */
			if (S_ISDIR (b.st_mode))
				*buf_end++ = T_DIRECTORY;		/* directory */
			else {
				/* s'il ne matche pas le filtre... */
				if (W.filter != 0 && !FilenameMatch (W.filter, name)) {
					--nb;		/* ...ne pas le mettre dans la liste */
					continue;
				}
				if (b.st_mode & 0111)
					*buf_end++ = T_EXECUTABLE;		/* executable */
				else {
					*buf_end++ = T_FILE;			/* plain file */
					/* voir si le nom matche l'un des filtres pour le colorer */
					if (W.suffixes != 0 && match_suffixes (W.suffixes, name))
						buf_end [-1] = T_HIGHLIGHT;
				}
			}
			p = name;
			/* MAXLEN caractres au maximum par nom */
			for (length = 0; (*buf_end++ = *p++) != 0 && length < MAXLEN; ++length)
				;
			buf_end [-1] = 0;
			/* calculer la taille du plus grand nom de fichier */
			if (maxlen < length)
				maxlen = length;
		}
		/* chercher le plus grand prfixe */
		if (nb == 1)
			strcpy (prefix, name);
		else {
			for (p = name, mp = prefix; *p == *mp; p++)
				mp++;
			*mp = 0;
		}
	}
	closedir (dirp);
	show_window (w, list, nb, maxlen);
	return nb;
}

/*
 *	La callback pour valider la rponse (sans faire disparatre la bote)
 */

static void defaultActionCallback (Widget widget, XtPointer client, XtPointer call)
{
	MenuBarWidget	w = (MenuBarWidget) client;
	XEvent*			ev = (XEvent*) call;
	char			tc [16];
	KeySym			ks;
	FontInfo*		f;

	switch (W.dialog_mode) {
		case DM_FileSelectionRead:
		case DM_FileSelectionCreate:
			XtVaGetValues (widget, XtNfont, &f, NULL);

			if (ev -> type == KeyPress) {
				if (0 != XLookupString ((XKeyEvent*)ev, tc, 1, &ks, 0))
					switch (tc [0]) {
					/* scrolle vers le haut au clavier */
						case 'p'&037:
							if (W.clip_y > W.min_clip_y) {
								if (W.clip_y == 0)
									W.clip_y = -3;
								W.clip_y -= f -> f_height;
								XMoveWindow (dpy, W.list_win, -3, W.clip_y);
							}
							return;

					/* scrolle vers le bas au clavier */
						case 'n'&037:
							if (W.clip_y <= -3) {
								W.clip_y += f -> f_height;
								if (W.clip_y == - 3)
									W.clip_y  = 0;
								XMoveWindow (dpy, W.list_win, -3, W.clip_y);
							}
					}
				return;
			}
			if (ev -> type == ButtonPress) {
				switch (((XButtonEvent*)ev) -> button) {
					/* scrolle vers le haut  la souris  roulette */
					case Button4:
						if (W.clip_y > W.min_clip_y) {
							if (W.clip_y == 0)
								W.clip_y = -3;
							W.clip_y -= f -> f_height;
							XMoveWindow (dpy, W.list_win, -3, W.clip_y);
						}
						return;

					/* scrolle vers le bas  la souris  roulette */
					case Button5:
						if (W.clip_y <= -3) {
							W.clip_y += f -> f_height;
							if (W.clip_y == - 3)
								W.clip_y  = 0;
							XMoveWindow (dpy, W.list_win, -3, W.clip_y);
						}
						return;
				}
			}
			return;

		case DM_GetStringIncremental:
			if (W.default_action != 0)
				(*W.default_action) (widget, call);
	}
}

/*
 *	La callback pour valider la rponse.
 */

static void activateCallback (Widget widget, XtPointer client, XtPointer call)
{
	XFocusChangeEvent	focus_in;
	MenuBarWidget		w = (MenuBarWidget) client;
	XKeyEvent*			e = (XKeyEvent*) call;
	void				(*cont)();
	KeySym				ks;
	static char			tc [256];
	char*				p;
	/*
	 *	Fait disparaitre la fentre des noms de fichiers.
	 */
	if ((W.dialog_mode == DM_FileSelectionRead ||
		 W.dialog_mode == DM_FileSelectionCreate) && W.list_win != 0)
	{
		XUnmapWindow (dpy, W.clip_win);
	}
	/*
	 *	Fait disparaitre la boite.
	 */
	if (W.hidden)
		XtUnmanageChild (W.exp_row);
	else {
		XUnmapWindow (dpy, XtWindow (W.exp_row));
		SetManaged (W.exp_row, False);
		XtManageChild ((Widget) w);
		set_mnemonic (w);
	}
	if (W.active != 0) {
		XtSetSensitive (W.active, True);
		XtSetKeyboardFocus (C.parent, W.active);
		/* un petit FocusIn car a semble manquer quelques fois */
		focus_in.type = FocusIn;
		focus_in.window = XtWindow (W.active);
		focus_in.display = dpy;
		focus_in.detail = 0;
		XtDispatchEvent ((XEvent*) &focus_in);
	} else
		XtSetKeyboardFocus (C.parent, None);

	if (e == NULL	/* la rponse est positive par construction */
					/* (un seul choix auto-valid) */
		|| (0 != XLookupString (e, tc, 1, &ks, 0) && tc [0] == '\r'))
	{
		/* mettre la rponse dans un buffer statique */
		W.locked = False;
		p = WixCopyBuffer (W.exp_wix);
		strcpy (tc, p);
		XtFree (p);
		W.fs_answer = tc;
	} else {
		W.fs_answer = NULL;	/* abandon: reponse nulle */
#ifdef RIEN_APRES
		return;		/* pire: on n'appelle mme pas la callback */
#endif
	}
	/* valider la rponse */
	if (W.confirm_action != 0)
		(*W.confirm_action) (widget, W.fs_answer);
}

/*
 *	Auto expansion des fichiers. Si un seul nom corespond, la confirmation
 *	est automatique.
 */

static void autoExpandCallback (Widget widget, XtPointer cd, XtPointer call)
{
	VerifyCallbackStruct*	cb = (VerifyCallbackStruct*) call;
	MenuBarWidget			w = (MenuBarWidget) cd;
	Boolean					dotdot;
	char					tc [256];
	char*					p;
	char*					dir;
	char*					file;
	char					prefix [256];
	int						nb;

	if (W.locked)
		return;

	switch (W.dialog_mode) {
		case DM_GetString:
			return;

		case DM_GetStringIncremental:
			if (W.auto_expand != 0)
				(*W.auto_expand) (widget, call);
			return;

		case DM_FileSelectionRead:
		case DM_FileSelectionCreate:
			if (cb -> reason == INSERT_STRING) {
				W.locked = True;
				p = WixCopyBuffer (widget);
				strncpy (tc, p, cb -> pos);
				strncpy (tc+cb -> pos, cb -> string, cb -> length);
				strcpy (tc+cb -> pos + cb -> length, p+cb -> pos);
				XtFree (p);

				dir = tc;
				file = rindex (dir, '/');
				if (file)
					*file++ = 0;
				else {
					file = dir;
					dir = ".";
				}
				strcpy (prefix, file);
				/*
				 *	Expanser tant que qu'un seul nom est trouv et que c'est un directory.
				 */
				dotdot = True;
				while ((nb = list_files (w, dir, prefix, dotdot)) == 1) {
					struct stat	bufstat;
					char		tt [256];
					char*		s;
					char*		d;

					for (s = dir, d = tt; *d++ = *s++;)
						;
					d [-1] = '/';
					for (s = prefix; *d++ = *s++;)
						;
					s = AbsPath (W.directory, tt);
					/* sortir aussi si on veut crer un fichier */
					if (W.dialog_mode == DM_FileSelectionCreate ||
						stat (s, &bufstat) == -1 ||
						!S_ISDIR (bufstat.st_mode))
						break;

					dotdot = False;
					strcpy (tc, tt);
					dir = tc;
					prefix [0] = 0;
				}
				/*
				 *	Afficher le rsultat dans le widget.
				 */
				if (nb >= 1) {
					char slash [2];
					slash [0] = '/';
					slash [1] = 0;
					WixReplace (widget, 0, -1, dir, -1);
					WixInsert (widget, slash);
					WixInsert (widget, prefix);
					WixRedraw (widget);
					cb -> doit = False;
				}
				W.locked = False;
				/* auto confirmer l'unique rponse en cas de chargement de fichier */
				if (nb == 1 && W.dialog_mode == DM_FileSelectionRead)
					activateCallback (widget, cd, NULL);
			}
			return;
		}
}

/*
 *	Cre la boite de dialogue.
 */

static void createMenubarDialog (MenuBarWidget w)
{
	static char defaultTranslations [] =
						"<Key>Return:	activate()\n"
						"Ctrl <Key>G:	activate()\n"
						"<Btn4Down>:	default-activate()\n"
						"<Btn5Down>:	default-activate()\n"
						"Ctrl <Key>N:	default-activate()\n"
						"Ctrl <Key>P:	default-activate()";

	W.list_win = 0;
	W.clip_win = 0;
	/* le conteneur */
	W.exp_row = XtVaCreateWidget ("expert_row", rowWidgetClass, C.parent,
							XtNshadowThickness,	1,
							XtNresizable,		False,
							XtNpadding,			1,
							XtNheight,			C.height,
							NULL);
	/* le label questionnant */
	W.exp_label = XtVaCreateManagedWidget (
							"exp-label", pushButtonWidgetClass, W.exp_row,
							XtNshadowThickness,	0,
							XtNlabel,			"mmmmmmmmmm",
							XtNmargin,			0,
							NULL);
	/* le wix rpondant */
	W.exp_wix = XtVaCreateManagedWidget (
							"exp-wix", wixWidgetClass, W.exp_row,
							XtNmonoLine,			True,
							XtNkillStack,			False,
							XtNtransientSelection,	True,
							XtNlineNum,				1,
							NULL);
	/* y mettre les translations idiones */
	XtOverrideTranslations (W.exp_wix,
						XtParseTranslationTable (defaultTranslations));
	XtAddCallback (W.exp_wix, XtNverifyModifyCallback, autoExpandCallback, (XtPointer) w);
	XtAddCallback (W.exp_wix, XtNactivateCallback, activateCallback, (XtPointer) w);
	XtAddCallback (W.exp_wix, XtNdefaultActionCallback, defaultActionCallback, (XtPointer) w);
	/* retailler le wix */
	XtVaSetValues (W.exp_row, XtNtoResize, W.exp_wix, NULL);
	/* et lui donner le clavier */
	XtSetKeyboardFocus (W.exp_row, W.exp_wix);
}

/*
 *	Fait apparaitre une boite de rponse rapide.
 */

void PopMenubarDialog (Widget mb, Widget active, char* label, DialogMode mode, char* reponse)
{
	MenuBarWidget	w = (MenuBarWidget) mb;

	W.dialog_mode = mode;
	W.active = active;
	/* Mettre le contenu par dfaut */
	W.locked = True;
	switch (W.dialog_mode) {
		case DM_FileSelectionRead:
		case DM_FileSelectionCreate:
			W.default_action = 0;	/* utilises que par GetStringIncremental */
			W.auto_expand = 0;
			WixReplace (W.exp_wix, 0, -1, "", -1);
			break;

		case DM_GetString:
			W.default_action = 0;	/* utilises que par GetStringIncremental */
			W.auto_expand = 0;
		case DM_GetStringIncremental:
		default:
			WixReplace (W.exp_wix, 0, -1, reponse, -1);
			SelectBuffer (W.exp_wix);
			break;
	}
	W.locked = False;
	/* Montrer la boite */
	XtVaSetValues (W.exp_label, XtNlabel, label, NULL);
	remove_mnemonic (w);
	XUnmapWindow (dpy, XtWindow (mb));
	C.managed = False;			/* pour que la disparition de la barre et */
	XtManageChild (W.exp_row);	/* l'apparition de la boite soient atomiques */
	XtSetKeyboardFocus (C.parent, W.exp_row);
	if (W.active != 0)
		XtSetSensitive (W.active, False);
	/*
	 *	Si on slecte un fichier, afficher le directory courant.
	 */
	if (W.dialog_mode == DM_FileSelectionRead ||
		W.dialog_mode == DM_FileSelectionCreate)
	{
		static Boolean	first = True;
		char 			prefix [256];
		/* initialiser le directory courant */
		if (first) {
			char*	curwd = GetCurrentDir ();

			first = False;
			(void) getcwd (curwd, 256);
		}
		prefix [0] = 0;
		if (reponse [0] == 0)
			strcpy (W.directory, GetCurrentDir ());
		else
			strcpy (W.directory, reponse);
		(void) list_files (w, W.directory, prefix, True);
	}
}

/*
 *	pour affecter les callbacks.
 */

void SetAutoExpand (Widget mb, void (*auto_expand)(Widget, XtPointer))
{
	MenuBarWidget	w = (MenuBarWidget) mb;

	W.auto_expand = auto_expand;
}

void SetDefaultAction (Widget mb, void (*default_action)(Widget, XtPointer))
{
	MenuBarWidget	w = (MenuBarWidget) mb;

	W.default_action = default_action;
}

void SetConfirmAction (Widget mb, void (*confirm_action)(Widget, char*))
{
	MenuBarWidget	w = (MenuBarWidget) mb;

	W.confirm_action = confirm_action;
}

/*
 *	Dtruit la fentre du slecteur de fichiers.
 */

void DestroyFSWindow (Widget mb)
{
	MenuBarWidget	w = (MenuBarWidget) mb;

	if (W.list_win != 0) {
		XDestroyWindow (dpy, W.list_win);
		XDestroyWindow (dpy, W.clip_win);
		W.list_win = (Window) 0;
		W.clip_win = (Window) 0;
	}
}
