/************************************************************
*										*
*	Tout pour les boites de dialogues rapides	*
*										*
*************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <unistd.h>
#if SYS_DIR
#	include <sys/dir.h>
#else
#	include <dirent.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <Xpa/Primitive.h>
#include <Xpa/PrimitiveP.h>
#include <Xpa/Wix.h>
#include <Xpa/PushButton.h>
#include <Xpa/MenuBar.h>
#include <Xpa/PanedWin.h>		/* pour XtNresizable !! */
#include <Xpa/Row.h>

/*
 *	Modes de dialogue pour la boite d'experts.
 */

typedef enum {
	DM_FileSelectionRead,
	DM_FileSelectionCreate,
	DM_GetString,
	DM_IncrementalSearch
} DialogMode;

extern char*	AbsPath (char*, char*);

typedef struct _DialogStruct {
	Widget			menubar;		/* la barre de menu associe */
	char			directory [256];/* le directory courant */
	char*			fs_answer;
	void			(*what_after) (char*);
	Widget			exp_row;		/* la boite des dialogues d'experts */
	Widget			exp_label;		/* le bouton de confirmation */
	Widget			exp_wix;		/* la zone de saisie de texte */
	Window			list_win;		/* la fentre pour afficher les noms des fichiers */
	Boolean			locked;			/* verrou pour la verifyCallback */
	DialogMode		dialog_mode;	/* type d'utilisation de la boite */
} DialogStruct;

#define T_DIRECTORY	2
#define T_FILE		3

/*
 *	Fabrique un tableau de strings avec les noms des fichiers du directory spcifi.
 */

static int compare (const void* a, const void* b)
{
	return strcmp ((*(char**) a)+1, (*(char**) b)+1);
}

static int list_files (DialogStruct* data, char* directory, char* prefix, Boolean dotdot)
{
#if SYS_DIR
	struct direct*	item;
#else
	struct dirent*	item;
#endif
	int			len = strlen (prefix);
	char		pref [256];
	char*		list [256];
	char		buffer [256*32];	/* comme on ne retient que les 15 premiers */
	char*		buf_end = buffer;	/* caractres des noms, a suffit amplement. */
	DIR*		dirp;
	char*		name;
	char*		p;
	char*		mp;
	int			nb;
	int			length;
	Display*	dpy = XtDisplay (data -> menubar);
	struct stat	b;
	char		tc [32];

	if (directory [0] == 0)
		directory = "/";			/* chemin absolu */
	else
		directory = AbsPath (data -> directory, directory);
	if (-1 == access (directory, R_OK | X_OK) || !(dirp = opendir (directory)))
		return -1;			/* drle de directory! */

	strcpy (pref, prefix);		/* sauver le prfixe initial */
	(void) readdir (dirp);		/* sauter . */
	if (!dotdot)
		(void) readdir (dirp);	/* sauter .. */
	for (nb = 0, item = readdir (dirp); item != NULL; item = readdir (dirp)) {
		name = item -> d_name;
		length = strlen (name);
		if (name [length-1] == 'o' && name [length-2] == '.')
			continue;	/* sauter les .o */
		if (len > 0 && bcmp (name, pref, len))
			continue;
		/*
		 *	Accumuler les noms dans une liste.
		 */
		if (nb < 256 && buf_end + length < & buffer [sizeof (buffer)]) {
			list [nb++] = buf_end;
			sprintf (tc, "%s/%s", directory, name);
			stat (tc, &b);	/* est-ce un repertoire */
			if (S_ISDIR (b.st_mode))
				*buf_end++ = T_DIRECTORY;		/* 2 => directory */
			else
				*buf_end++ = T_FILE;			/* 3 => plain file */
			p = name;
			length = 15;	/* 15 caractres au maximum par nom */
			while ((*buf_end++ = *p++) && --length >= 0)
				;
			buf_end [-1] = 0;
		}
		/* chercher le plus grand prfixe */
		if (nb == 1)
			strcpy (prefix, name);
		else {
			for (p = name, mp = prefix; *p == *mp; p++)
				mp++;
			*mp = 0;
		}
	}
	closedir (dirp);
	if (nb >= 2) {
#define	COL 5
		Dimension	width;
		Pixel		fg_color;
		Pixel		cm_color;
		char*		cp;
		int			c;
		int			l;
		int			n16;
		int			nl = (nb+COL-1)/COL;
		static GC	gc;
		FontInfo*	fp;
		/*
		 *	Crer la fentre la premire fois (et le gc aussi).
		 */
		if (!data -> list_win) {
			XSetWindowAttributes	xswa;

			xswa.border_pixel = BlackPixelOfScreen (XtScreen (data -> menubar));
			xswa.background_pixel = WhitePixelOfScreen (XtScreen (data -> menubar));
			xswa.backing_store = WhenMapped;
			data -> list_win = XCreateWindow (dpy,
									XtWindow (XtParent (data -> menubar)),
									70, 5, 10, 10, 1,
									CopyFromParent, CopyFromParent, CopyFromParent,
									CWBackPixel|CWBorderPixel|CWBackingStore,
									&xswa);
			if (!gc)
				gc = XCreateGC (dpy, data -> list_win, 0, NULL);
		}
		/* trier la liste par ordre alphabtique */
		qsort (list, nb, sizeof (char*), compare);
		XtVaGetValues (data -> menubar,
						XtNfont,		&fp,
						XtNwidth,		&width,
						XtNforeground,	&fg_color,
						XtNbackground,	&cm_color,
						NULL);
		n16 = 16 * XTextWidth (fp -> f_info, "n", 1);
		l = 5 + COL * n16;
		XMoveResizeWindow (dpy, data -> list_win,
						(width - l) / 2, 5,
						l, 6 + nl * fp -> f_height);
		XSetFont (dpy, gc, fp -> f_info -> fid);
		XClearWindow (dpy, data -> list_win);
		XMapWindow (dpy, data -> list_win);
		for (l = 0; l < nl; l++) {
			int 	x = 5;
			int		prev = -1;
			Pixel	color;

			for (c = 0; c+l < nb; c += nl) {
				if (list [c+l][0] != prev) {
					prev = list [c+l][0];
					XSetForeground (dpy, gc,
						prev == T_DIRECTORY ? cm_color : fg_color);
				}
				XDrawString (dpy, data -> list_win, gc,
							 x, 3 + l * fp -> f_height + fp -> f_ascent,
							 list [c+l]+1, strlen (list [c+l]+1));
				x += n16;
			}
		}
	} else
		if (data -> list_win)
			XUnmapWindow (dpy, data -> list_win);
	return nb;
}

/*
 *	La callback pour valider la rponse (en cours de route)
 */

static void defaultActionCallback (Widget w, XtPointer client, XtPointer call)
{
	DialogStruct*	data = (DialogStruct*) client;
	XKeyEvent*	e = (XKeyEvent*) call;

	switch (data -> dialog_mode) {
		case DM_FileSelectionRead:
		case DM_FileSelectionCreate:
		case DM_GetString:
			return;
		case DM_IncrementalSearch:
#if 0
			IncrementalSearchNavigate (w, data, e);
#endif
			return;
	}
}

/*
 *	La callback pour valider la rponse.
 */

static void activateCallback (Widget w, XtPointer client, XtPointer call)
{
	XFocusChangeEvent	focus_in;
	DialogStruct*		data = (DialogStruct*) client;
	XKeyEvent*			e = (XKeyEvent*) call;
	void				(*cont)();
	KeySym				ks;
	static char			tc [256];
	char*				p;
	/*
	 *	Fait disparaitre la fentre des noms de fichiers.
	 */
	if (data -> dialog_mode == DM_FileSelectionRead && data -> list_win)
		XUnmapWindow (XtDisplay (w), data -> list_win);
	/*
	 *	Fait disparaitre la boite.
	 */
#if 0
	if (data -> stripped)
		XtUnmanageChild (data -> exp_row);
	else
#endif
	{
		XUnmapWindow (XtDisplay (data -> exp_row), XtWindow (data -> exp_row));
		SetManaged (data -> exp_row, False);
		XtManageChild (data -> menubar);
		XtVaSetValues (data -> menubar, XtNmnemonicOn, True, NULL);
	}
#if 0
	XtSetSensitive (data -> active, True);
	XtSetKeyboardFocus (data -> pane, XtParent (data -> active));
	/* un petit FocusIn car a semble manquer quelques fois */
	focus_in.type = FocusIn;
	focus_in.window = XtWindow (data -> active);
	focus_in.display = XtDisplay (data -> active);
	focus_in.detail = 0;
	XtDispatchEvent ((XEvent*) &focus_in);
#endif
	if (e != NULL)	/* sinon la rponse est positive par construction */
					/* (un seul choix auto-valid) */
		if (XLookupString (e, tc, 1, &ks, 0))
			switch (tc [0]) {
				/* abandon */
				default:
					data -> what_after = NULL;
					return;

				case '\r':
				case 's'&037:
				case 'r'&037:
					break;
			}
		else
			return;		/* abandon aussi */

	/* mettre la rponse dans un buffer statique */
	data -> locked = False;
	p = WixCopyBuffer (data -> exp_wix);
	strcpy (tc, p);
	XtFree (p);
	data -> fs_answer = tc;
	/* appeller la suite */
	if (data -> what_after) {
		cont = data -> what_after;
		data -> what_after = NULL;
		(*cont) (data -> fs_answer);
	}
}

/*
 *	Auto expansion des fichiers. Si un seul nom corespond, la confirmation
 *	est automatique.
 */

static void autoExpandCallback (Widget w, XtPointer cd, XtPointer call)
{
	VerifyCallbackStruct*	cb = (VerifyCallbackStruct*) call;
	DialogStruct*			data = (DialogStruct*) cd;
	Boolean					dotdot;
	char					tc [256];
	char*					p;
	char*					dir;
	char*					file;
	char					prefix [256];
	int						nb;

	if (data -> locked)
		return;

	switch (data -> dialog_mode) {
		case DM_FileSelectionCreate:
		case DM_GetString:
			return;

		case DM_IncrementalSearch:
#if 0
			IncrementalSearch (w, data, cb);
#endif
			return;

		case DM_FileSelectionRead:
			if (cb -> reason == INSERT_STRING) {
				data -> locked = True;
				p = WixCopyBuffer (w);
				strncpy (tc, p, cb -> pos);
				strncpy (tc+cb -> pos, cb -> string, cb -> length);
				strcpy (tc+cb -> pos + cb -> length, p+cb -> pos);
				XtFree (p);

				dir = tc;
				file = rindex (dir, '/');
				if (file)
					*file++ = 0;
				else {
					file = dir;
					dir = ".";
				}
				strcpy (prefix, file);
				/*
				 *	Expanser tant que qu'un seul nom est trouv et que c'est un directory.
				 */
				dotdot = True;
				while ((nb = list_files (data, dir, prefix, dotdot)) == 1) {
					struct stat	bufstat;
					char		tt [256];
					char*		s;
					char*		d;

					for (s = dir, d = tt; *d++ = *s++;)
						;
					d [-1] = '/';
					for (s = prefix; *d++ = *s++;)
						;
					s = AbsPath (data -> directory, tt);
					if (stat (s, &bufstat) == -1 || !S_ISDIR (bufstat.st_mode))
						break;

					dotdot = False;
					strcpy (tc, tt);
					dir = tc;
					prefix [0] = 0;
				}
				/*
				 *	Afficher le rsultat dans le widget.
				 */
				if (nb >= 1) {
					WixReplace (w, 0, -1, dir, -1);
					WixInsert (w, "/");
					WixInsert (w, prefix);
					WixRedraw (w);
					cb -> doit = False;
				}
				data -> locked = False;
				if (nb == 1)
					activateCallback (w, cd, NULL);	/* auto confirmer l'unique rponse */
			}
			return;
		}
}

/*
 *	Fait apparaitre une boite de rponse rapide.
 */

void PopMenubarDialog (	Widget menubar,
						char* label,
						DialogMode mode,
						void (*what_after) (char*),
						char* reponse)
{
	static char defaultTranslations [] =
						"<Key>Return:	activate()\n"
						"Ctrl <Key>G:	activate()\n"
						"Ctrl <Key>N:	default-activate()\n"
						"Ctrl <Key>P:	default-activate()";
	DialogStruct*	data;
	Dimension		height;

	XtVaGetValues (menubar,
		XtNheight,	&height,
		XtNdialog,	&data,
		NULL);

	/* crer la boite de dialogue  la premire utilisation */
	if (data == NULL) {
		/* la barre de menu */
		data -> menubar = menubar;
		/* le conteneur */
		data -> exp_row = XtVaCreateWidget ("expert_row", rowWidgetClass, XtParent (menubar),
											XtNshadowThickness,	1,
											XtNresizable,		False,
											XtNpadding,			1,
											XtNheight,			height,
											NULL);
		/* le label questionnant */
		data -> exp_label = XtVaCreateManagedWidget (
								"exp-label", pushButtonWidgetClass, data -> exp_row,
								XtNshadowThickness,	0,
								XtNlabel,			"mmmmmmmmmm",
								XtNmargin,			0,
								XtNuserData,		data,
								NULL);
		/* le wix rpondant */
		data -> exp_wix = XtVaCreateManagedWidget (
								"exp-wix", wixWidgetClass, data -> exp_row,
								XtNuserData,			data,
								XtNmonoLine,			True,
								XtNkillStack,			False,
								XtNtransientSelection,	True,
								XtNlineNum,				1,
								NULL);
		/* y mettre les translations idiones */
		XtOverrideTranslations (data -> exp_wix,
							XtParseTranslationTable (defaultTranslations));
		XtAddCallback (data -> exp_wix, XtNverifyModifyCallback, autoExpandCallback, (XtPointer) data);
		XtAddCallback (data -> exp_wix, XtNactivateCallback, activateCallback, (XtPointer) data);
		XtAddCallback (data -> exp_wix, XtNdefaultActionCallback, defaultActionCallback, (XtPointer) data);
		/* retailler le wix */
		XtVaSetValues (data -> exp_row, XtNtoResize, data -> exp_wix, NULL);
		/* et lui donner le clavier */
		XtSetKeyboardFocus (data -> exp_row, data -> exp_wix);
	}

	/* Mettre le contenu par dfaut */
	data -> locked = True;
	data -> dialog_mode = mode;
	data -> what_after = what_after;
	WixReplace (data -> exp_wix, 0, -1, reponse, -1);
	SelectBuffer (data -> exp_wix);
	data -> locked = False;
	/* Montrer la boite */
	XtVaSetValues (data -> menubar, XtNmnemonicOn, False, NULL);
	XtVaSetValues (data -> exp_label, XtNlabel, label, NULL);
	XUnmapWindow (XtDisplay (data -> menubar), XtWindow (data -> menubar));
	SetManaged (data -> menubar, False);	/* pour que la disparition de la barre et */
	XtManageChild (data -> exp_row);		/* l'apparition de la boite soient atomiques */
	XtSetKeyboardFocus (XtParent (data -> menubar), data -> exp_row);
/*	XtSetSensitive (data -> active, False);
*/
	/*
	 *	Si on slecte un fichier, afficher le directory courant.
	 */
	if (mode == DM_FileSelectionRead) {
		char 	tc [1256];
		char*	p;

		strcpy (data -> directory, reponse);
		(void) list_files (data, data -> directory, tc, True);
	}
}
