/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <Xpa/Container.h>
#include <Xpa/ContainerP.h>
#include <Xpa/PanedWin.h>
#include <Xpa/PanedWinP.h>


#define RESIZE_OPAQUE

#define	C	pw -> core
#define	P	pw -> composite
#define	M	pw -> container
#define	W	pw -> panedWindow

#define CR	PanedWindowConstraintRec

static XtGeometryResult
				GeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static void		ChangeManaged (Widget);

static void		ConstraintInitialize (Widget, Widget, Arg*, Cardinal*);
static Boolean	ConstraintSetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static void		ConstraintDestroy (Widget);

static	void	DragAction (Widget, XEvent*, String*, Cardinal*);

#define offset 1	/* marges pour le grip */

static XtResource resources [] = {
	{ XtNpadding, XtCPadding, XtRInt, sizeof (int),
	  XtOffset (PanedWindowWidget, panedWindow.padding), XtRImmediate, (XtPointer) 10
	},
	{ XtNmargin, XtCMargin, XtRInt, sizeof (int),
	  XtOffset (PanedWindowWidget, panedWindow.margin), XtRImmediate, (XtPointer) 0
	},
};

/*  Constraint resource list for the PanedWindow  */

static XtResource constraintResources [] = 
{
	{ XtNresizable, XtCResizable, XtRBoolean, sizeof (Boolean),
	  XtOffset (PanedWindowConstraints, resizable), XtRImmediate, (XtPointer) True
	},
};

static char defaultTranslations [] = "\
	<Btn1Down>:		drag()\n\
	Button1<PtrMoved>:	drag()\n\
	<Btn1Up>:		drag()";

static XtActionsRec actionsList[] =
{
	{"drag", (XtActionProc) DragAction },
};

PanedWindowClassRec panedWindowClassRec = {
  /* core_class fields */
  {
    /* superclass         */		(WidgetClass) &containerClassRec,
    /* class_name         */		"PanedWindow",
    /* widget_size        */		sizeof(PanedWindowRec),
    /* class_initialize   */		NULL,
    /* class_partinit     */		NULL,
    /* class_inited       */		False,
    /* initialize         */		Initialize,
    /* Init hook	  */		NULL,
    /* realize            */		XtInheritRealize,
    /* actions		  */		actionsList,
    /* num_actions	  */		XtNumber (actionsList),
    /* resources          */		resources,
    /* num_resources      */	XtNumber (resources),
    /* xrm_class          */		NULLQUARK,
    /* compress_motion	  */	False,
    /* compress_exposure  */	True,
    /* compress_enterleave*/	True,
    /* visible_interest   */		False,
    /* destroy            */		NULL,
    /* resize             */		Resize,
    /* expose             */		Redisplay,
    /* set_values         */		SetValues,
    /* set values hook    */	NULL,
    /* set values almost  */	XtInheritSetValuesAlmost,
    /* get values hook    */	NULL,
    /* accept_focus       */		NULL,
    /* Version            */		XtVersion,
    /* PRIVATE cb list    */	NULL,
    /* TM Table           */		defaultTranslations,
    /* query_geom         */	NULL,
    /* display acccelerator */	XtInheritDisplayAccelerator,
    /* extension */			NULL,
  },{
/* composite_class fields */
    /* geometry_manager   */	GeometryManager,
    /* change_managed     */	ChangeManaged,
    /* insert_child	  */		XtInheritInsertChild,
    /* delete_child	  */		XtInheritDeleteChild,
    /* extension */			NULL,
  },{
/* constraint_class fields */
    /*  constraint resource set  */	constraintResources,
    /*  num_resources */			XtNumber (constraintResources),
    /*  size of the constraint data */	sizeof (PanedWindowConstraintRec),
    /*  contraint initilize proc */	ConstraintInitialize,
    /*  contraint destroy proc */	ConstraintDestroy,
    /*  contraint set values proc */	ConstraintSetValues,
  },{
/* Container class fields */     
     /* extension */			NULL
  },{
/* panedWindow class fields */     
     /* extension */			NULL
 }	
};

WidgetClass panedWindowWidgetClass = (WidgetClass)&panedWindowClassRec;

static void ConstraintInitialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	/* nothing to do */
	return;
}

static Boolean ConstraintSetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
#ifdef DEBUG
	if (newConstraint -> resizable != curConstraint -> resizable)
		printf ("%s: resizable: %d\n", XtName (new), newConstraint -> resizable);
#endif

	return False;
}

static void ConstraintDestroy (Widget w)
{
	/* nothing to do */
	return;
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	PanedWindowWidget	pw = (PanedWindowWidget) new;
	XGCValues			values;
	unsigned long		valuemask;

	if (request -> core.height == 0)	C.height = 300;
	if (request -> core.width == 0)		C.width = 400;

	W.num_managed_children = 0;
	W.managed_children = NULL;
	W.sep_number = 0;
	W.sep_y = NULL;
	/* le GC pour les separateurs */
	valuemask = GCLineWidth|GCSubwindowMode|GCFunction;
	values.function = GXinvert;
	values.line_width = 2;
	values.subwindow_mode = IncludeInferiors;
	W.sep_GC = XtGetGC (new, valuemask, &values);
}

static void _resize (PanedWindowWidget pw, Boolean flag)
{
	XtWidgetGeometry	req;
	Dimension			width = 0;
	Dimension			height = 0;
	int					i, j;
	int					y;
	Widget				w;
	Boolean				resizable;
	int					nm;		/* nombre de carreaux managed */
	/*
	 *	Calculer la largeur maximale.
	 */
	for (i = 0; i < W.num_managed_children; i++) {
		w = W.managed_children [i];
		if (width < w -> core.width)
			width = w -> core.width;	/* largeur maximale */
		height += w -> core.height;
	}
	switch (nm = W.num_managed_children) {
	/*
	 *	il n'y a aucun sous widget de managed.
	 */
	case 0:
		return;
	/*
	 *	Il y a exactement un sous widget de managed: on se superpose.
	 */
	case 1:
		req.request_mode = CWWidth|CWHeight;
		req.width = width + 2*W.margin;
		req.height = height + 2*W.margin;
		/* on s'adapte */
		if (flag &&
		    XtGeometryYes == XtMakeGeometryRequest ((Widget) pw, &req, NULL))
		{
			C.width = width + 2*W.margin;
			C.height = height + 2*W.margin;
		}
		/* on s'impose */
		w = W.managed_children [0];
		XtConfigureWidget (w, W.margin,
					W.margin,
					C.width - 2*W.margin,
					C.height - 2*W.margin,
					0);
		return;
	/*
	 *	Il y a au moins deux sous widgets de managed.
	 */
	default:
		/* y a-t-il un separateur de plus? */
		if (--nm > W.sep_number) {
			W.sep_number = nm;
			W.sep_y = (int*) XtRealloc ((char*)W.sep_y, nm * sizeof (int));
		}
		if (flag) {
			height += nm * (offset+W.padding);	/* inclure les separateurs */
			height += 2*W.margin;
			width += 2*W.margin;
			req.request_mode = CWWidth|CWHeight;
			req.width = width;
			req.height = height;
			if (XtGeometryYes == XtMakeGeometryRequest ((Widget) pw, &req, NULL)) {
				C.width = width;
				C.height = height;
			}
		}
		y = W.margin;
		j = 0;
		for (i = 0; i < W.num_managed_children-2; i++) {
			w = W.managed_children [i];
			XtConfigureWidget (w, W.margin, y, C.width - 2*W.margin, w -> core.height, 0);
			W.sep_y [j++] = y + w -> core.height;
			y += w -> core.height + offset;
			if (((CR*) w -> core.constraints) -> resizable)
				y += W.padding;
		}
		/* L'avant dernier */
		w = W.managed_children [nm-1];
		resizable = ((CR*) w -> core.constraints) -> resizable;
		if (flag && height > C.height && resizable) {
			height = (C.height - y - offset - W.padding) / 2;
			XtConfigureWidget (w, W.margin, y, C.width - 2*W.margin, height, 0);
			W.sep_y [j++] = y + height;
			y += height + offset + W.padding;
		} else {
			XtConfigureWidget (w, W.margin, y, C.width - 2*W.margin, w -> core.height, 0);
			W.sep_y [j++] = y + w -> core.height;
			y += w -> core.height + offset;
			if (resizable)
				y += W.padding;
		}
		/* Le dernier */
		height = C.height - y - W.margin;
		if (height < 10)
			height = 10;
		XtConfigureWidget (W.managed_children [nm], W.margin, y,
							C.width - 2*W.margin, height, 0);
	}
}

static void Resize (Widget widget)
{
	_resize ((PanedWindowWidget) widget, False);
}

/*
 *	Afficher les separateurs et les grips.
 */

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	PanedWindowWidget	pw = (PanedWindowWidget) widget;
	Display*			dpy = XtDisplay (widget);
	Window				win = XtWindow (widget);
	Widget				prev, next;
	int					i;
	int					mg = W.margin;
	int					y = 0;
	static Boolean		first = True;

#ifdef RESIZE_OPAQUE
	if (first) {
		first = False;
		/* Acclrer les dplacements sur les machines lentes. */
		XSelectInput (dpy, win,
						XtBuildEventMask (widget) | PointerMotionHintMask);
	}
#endif
	if (W.sep_number == 0 && mg == 0)
		return;		/* il n'y a ni contour, ni sparateur  afficher */

	XSetRegion (dpy, M.top_GC, region);
	XSetRegion (dpy, M.bot_GC, region);
	/* Afficher le contour */
#if 0
	if (mg > 0) {
		for (i = 0; i < mg; ++i) {
			XSetForeground (dpy, gc, pix [i]);
			_DrawOval (dpy, win, gc,
							mg-i, mg-i,
							C.width-2*(mg-i), C.height-2*(mg-i),
							i+20-mg);
		}
	}
#endif
	/* afficher les sparateurs */
	for (i = 0; i < W.sep_number; i++) {
		y = W.sep_y [i];
		prev = (i >= W.num_managed_children) ? NULL : W.managed_children [i];
		next = (i >= W.num_managed_children-1) ? NULL : W.managed_children [i+1];
		if (M.shadow_thickness > 0 &&
		    prev && ((CR*) prev -> core.constraints) -> resizable &&
		    next && ((CR*) next -> core.constraints) -> resizable)
			_DrawShadow (dpy, win,
							M.bot_GC, M.top_GC, M.shadow_thickness,
							C.width - 2*mg, 2,
							mg, y + (offset + W.padding)/2 - 1);
	}
	XSetClipMask (dpy, M.top_GC, None);
	XSetClipMask (dpy, M.bot_GC, None);
}

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	return False;
}

static XtGeometryResult GeometryManager (Widget w, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	if (request -> request_mode & CWHeight) {
		XtResizeWidget (w, w -> core.width, request -> height, w -> core.border_width);
		_resize ((PanedWindowWidget) XtParent (w), False);
		return (XtGeometryResult) XtGeometryDone;
	}
	return (XtGeometryResult) XtGeometryNo;
}

static void ChangeManaged (Widget widget)
{
	PanedWindowWidget	pw = (PanedWindowWidget) widget;
	int					i, j;
	Widget				child;

	if ((int) P.num_slots > W.num_managed_children)
		W.managed_children = (Widget*) XtRealloc (
											(char*)W.managed_children,
											P.num_slots * sizeof (Widget*));

	for (i = j = 0; i < (int) P.num_children; i++) {
		child = P.children [i];
		if (!child -> core.managed)
			continue;

		W.managed_children [j++] = child;
	}
	W.num_managed_children = j;
	_resize (pw, True);
}

/*
 *	Action de retaillage.
 */

static void DragAction (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	PanedWindowWidget	pw = (PanedWindowWidget) widget;
	XButtonEvent*		e = (XButtonEvent*) event;
	static int			gn = -1;
	static int			old_y;
	static int			y_offset;
	static int			upper_bound;
	static int			lower_bound;
	int					y;
	static Widget		w, prev;
	Display*			dpy = XtDisplay (widget);
	Window				win = XtWindow (widget);
	int					hsep = offset+W.padding;

	switch (e -> type) {
	case ButtonPress:
		if (gn != -1)
			return;	/* bizarre */

		for (gn = 0; gn < W.sep_number; gn++) {
			y = W.sep_y [gn] + offset/2;
			if (e -> y < y)
				break;	/* trop haut */

			if (e -> y < y + W.padding) {
				if (gn == 0) {
					upper_bound = W.margin + hsep;
					if (W.sep_number == 1)
						lower_bound = C.height - W.margin - hsep;
					else
						lower_bound = W.sep_y [gn+1] - hsep;
				} else
				if (gn == (W.sep_number-1)) {
					upper_bound = W.sep_y [gn-1] + 2 * hsep;
					lower_bound = C.height - W.margin - hsep;
				} else {
					upper_bound = W.sep_y [gn-1] + 2 * hsep;
					lower_bound = W.sep_y [gn+1] - hsep;
				}
				/* Trouver les widgets precedent et suivant */
				prev = W.managed_children [gn];
				w = W.managed_children [gn+1];
				/* Tester les contraintes */
				if (! ((CR*)prev -> core.constraints) -> resizable ||
				    ! ((CR*)w -> core.constraints) -> resizable)
				{
					gn = -1;	/* ne rien faire */
					return;
				}
				/* Tracer la ligne */
				old_y = y + W.padding/2;
#ifndef RESIZE_OPAQUE
				XDrawLine (dpy, win, W.sep_GC,
							W.margin, old_y, C.width - W.margin, old_y);
#endif
				y_offset = e -> y - old_y;
				return;
			}
		}
		gn = -1;
		return;

	case MotionNotify:
		if (gn == -1)
			return;

		y = e -> y - y_offset;
		if (y <= upper_bound)
			y = upper_bound;
		else if (y >= lower_bound)
			y = lower_bound;
#ifdef RESIZE_OPAQUE
		/* Retailler les widgets */
		old_y = y;
		y = old_y - hsep/2;
		XtConfigureWidget (prev, prev -> core.x, prev -> core.y,
							prev -> core.width,
							prev -> core.height - (W.sep_y[gn] - y),
							0);
		XtConfigureWidget (w, w -> core.x,
							w -> core.y - (W.sep_y[gn] - y),
							w -> core.width,
							w -> core.height + (W.sep_y[gn] - y),
							0);
		W.sep_y [gn] = y;
		XClearArea (dpy, win, 0, old_y-2, C.width, 4, True);
		{
			Window			window;
			unsigned int	ui;

			XQueryPointer (dpy, win, &window, &window, &y, &y, &y, &y, &ui);
		}
#else
		/* Effacer l'ancienne ligne */
		XDrawLine (dpy, win, W.sep_GC, W.margin, old_y, C.width - W.margin, old_y);
		/* Afficher la nouvelle ligne */
		old_y = y;
		XDrawLine (dpy, win, W.sep_GC, W.margin, old_y, C.width - W.margin, old_y);
#endif
		return;

	case ButtonRelease:
		if (gn == -1)
			return;	/* bizarre */
#ifndef RESIZE_OPAQUE
		/* Effacer l'ancienne ligne */
		XDrawLine (dpy, win, W.sep_GC, W.margin, old_y, C.width - W.margin, old_y);
#endif
		/* Retailler les widgets */
		y = old_y - hsep/2;
		XtConfigureWidget (prev, prev -> core.x, prev -> core.y,
							prev -> core.width,
							prev -> core.height - (W.sep_y[gn] - y),
							0);
		XtConfigureWidget (w, w -> core.x,
							w -> core.y - (W.sep_y[gn] - y),
							w -> core.width,
							w -> core.height + (W.sep_y[gn] - y),
							0);
		W.sep_y [gn] = y;
		gn = -1;
	}
}
