/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <Xpa/Primitive.h>
#include <Xpa/PrimitiveP.h>
#include <X11/Shell.h>

#define PCR	primitiveClassRec.primitive_class

#define P	w -> primitive
#define C	w -> core

/*  Resource definitions for Subclasses of Primitive */

static XtResource resources[] =
{
	{ XtNtipFont, XtCFont, XtRFontStruct, sizeof (XFontStruct*),
	  XtOffset (PrimitiveWidget, primitive.tip_fp), XtRString, "fixed"
	},
	{ XtNtipBackground, XtCBackground, XtRPixel, sizeof (Pixel),
	  XtOffset (PrimitiveWidget, primitive.tip_background), XtRString, "#ffd"
	},
	{ XtNtipMessage, XtCTipMessage, XtRString, sizeof (String),
	  XtOffset (PrimitiveWidget, primitive.tip_message), XtRImmediate, (XtPointer) 0
	},
	{ XtNgravity, XtCGravity, XtRCursor, sizeof (int),
	  XtOffset(PrimitiveWidget, primitive.gravity),
	  XtRImmediate, (XtPointer) ForgetGravity
	},
	{ XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (PrimitiveWidget, primitive.foreground), XtRString, "black"
	},
	{ XtNcursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(PrimitiveWidget, primitive.cursor), XtRString, "top_left_arrow"
	},
	{ XtNselect, XtCCallback, XtRCallback, sizeof(XtCallbackList),
	  XtOffset (PrimitiveWidget, primitive.select), XtRPointer, NULL
	},
	{ XtNrelease, XtCCallback, XtRCallback, sizeof(XtCallbackList),
	  XtOffset (PrimitiveWidget, primitive.release), XtRPointer, NULL
	},
	{ XtNtopShadowColor, XtCBackground, XtRPixel, sizeof (Pixel),
	  XtOffset(PrimitiveWidget, primitive.top_pixel), XtRCallProc,
	  (XtPointer) _GetTopShadowColor
	},
	{ XtNbottomShadowColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset(PrimitiveWidget, primitive.bot_pixel), XtRCallProc,
	  (XtPointer) _GetBottomShadowColor
	},
	{ XtNshadowThickness, XtCShadowThickness, XtRInt, sizeof (int),
	  XtOffset(PrimitiveWidget, primitive.shadow_thickness), XtRImmediate, (XtPointer) 2
	},
	{ XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	  XtOffset(Widget, core.background_pixel), XtRString, "grey"
	},
	{ XtNuserData, XtCUserData, XtRPointer, sizeof (XtPointer),
	  XtOffset(PrimitiveWidget, primitive.user_data), XtRImmediate, (XtPointer) NULL
	},
};



/*  Routine definiton used for initializing the class record.  */

static void		ClassInitialize (void);
static void		Initialize (Widget, Widget, ArgList, Cardinal*);
static Boolean	SetValues (Widget, Widget, Widget, ArgList, Cardinal*);
static void		Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static void Redisplay (Widget widget, XEvent* event, Region region);

/*  The primitive class record definition  */

PrimitiveClassRec primitiveClassRec =
{
   {
      (WidgetClass) &coreClassRec,		/* superclass */	
      "Primitive",						/* class_name */	
      sizeof(PrimitiveRec),				/* widget_size */	
	  ClassInitialize,					/* class_initialize */    
      NULL,								/* class_part_initialize */
      False,                            /* class_inited */	
	  Initialize,						/* initialize */	
      NULL,                             /* initialize_hook */
	  Realize,							/* realize */	
      NULL,								/* actions */
      0,								/* num_actions */
      resources,                        /* resources	         */	
      XtNumber(resources),              /* num_resources         */	
      NULLQUARK,                        /* xrm_class	         */	
      TRUE,                             /* compress_motion       */
      TRUE,                             /* compress_exposure     */	
      TRUE,                             /* compress_enterleave   */
      False,                            /* visible_interest      */
	  NULL,								/* destroy               */	
      NULL,                             /* resize                */	
      Redisplay,                        /* expose                */	
      SetValues,						/* set_values	         */	
      NULL,                             /* set_values_hook       */
      XtInheritSetValuesAlmost,         /* set_values_almost     */
      NULL,                             /* get_values_hook       */
      NULL,                             /* accept_focus	         */	
      XtVersion,                        /* version               */
      NULL,                             /* callback private      */
      NULL,                             /* tm_table              */
      NULL,                             /* query_geometry        */
   },

   {
      NULL,				/* extension */
   }
};

WidgetClass primitiveWidgetClass = (WidgetClass) &primitiveClassRec;
   
/*
 *	Initialisation de la classe.
 */

static WidgetClass	wc;						/* pour forcer le chargement de */
extern WidgetClass	vendorShellWidgetClass;	/* MON VendorShell */

static void ClassInitialize (void)
{
	wc = vendorShellWidgetClass;
	PCR.tipWidget = NULL;
	PCR.the_widget = NULL;
	PCR.tipTimer = 0;
}

/************************************************************************
 *
 *  GetTopShadowGC
 *     Get the graphics context used for drawing the shadows
 *
 ************************************************************************/

static void GetTopShadowGC (PrimitiveWidget	pw)
{
	XGCValues  values;
	unsigned long  valuemask;

	valuemask = GCForeground | GCBackground;
	values.foreground = pw -> primitive.top_pixel;
	values.background = pw -> primitive.foreground;

	pw -> primitive.top_GC = XtGetGC ((Widget)pw, valuemask, &values);
}

/************************************************************************
 *
 *  GetBottomShadowGC
 *     Get the graphics context used for drawing the shadows
 *
 ************************************************************************/

static void GetBottomShadowGC (PrimitiveWidget pw) {
	XGCValues  values;
	unsigned long  valuemask;

	valuemask = GCForeground | GCBackground;
	values.foreground = pw -> primitive.bot_pixel;
	values.background = pw -> primitive.foreground;

	pw -> primitive.bot_GC = XtGetGC ((Widget)pw, valuemask, &values);
}

/*
 *	Tout pour les hints & tips.
 */

static void showTip (XtPointer client, XtIntervalId* id)
{
	PrimitiveWidget	w = PCR.the_widget;
	int				width;
	Position		x, y;
	Window			child;
	int				fh = P.tip_fp -> ascent + P.tip_fp -> descent + 4;

	width = XTextWidth (P.tip_fp, P.tip_message, strlen (P.tip_message));
	/* calculer les coordones absolues */
	XtTranslateCoords ((Widget)w, 0, 0, &x, &y);
	/* ne pas sortir de l'cran */
	if (y + C.height + fh >= HeightOfScreen (XtScreen ((Widget)w)))
		y -= fh+1;
	else
		y += C.height;
	XtConfigureWidget (PCR.tipWidget, x+2, y, width+6, fh, 1);
	XtVaSetValues (PCR.tipWidget,
					XtNbackground, P.tip_background,
					NULL);
	XtPopup (PCR.tipWidget, XtGrabNone);
	PCR.tipTimer = 0;
}

static void tipDisplay (Widget shell, XtPointer client, XEvent* event, Boolean* cont)
{
	PrimitiveWidget	w = PCR.the_widget;
	Display*		dpy = XtDisplay (PCR.tipWidget);
	Window			win = XtWindow (PCR.tipWidget);

	XSetFont (dpy, PCR.tipGC, P.tip_fp -> fid);
	XDrawString (dpy, win, PCR.tipGC, 3, P.tip_fp -> ascent + 2, P.tip_message, strlen (P.tip_message));
}

/*
 *	Fait apparatre / disparatre le message d'aide.
 */

static void EnterLeaveHandler (Widget widget, XtPointer cd, XEvent* event, Boolean* cont)
{
	XtAppContext	app = XtWidgetToApplicationContext (widget);
	PrimitiveWidget	w = (PrimitiveWidget) widget;

	*cont = True;	/* appliquer les autres traitements ensuite */
	if (P.tip_message == NULL)
		return;

	switch (event -> type) {
	case EnterNotify:
		if (PCR.tipTimer == 0) {
			PCR.the_widget = w;
			PCR.tipTimer = XtAppAddTimeOut (app, 750, showTip, NULL);
		}
		return;

	case LeaveNotify:
		if (PCR.tipTimer) {
			XtRemoveTimeOut (PCR.tipTimer);
			PCR.tipTimer = 0;
		} else
			XtPopdown (PCR.tipWidget);
		return;
	}
}

/*
 *	La procdure de cration de la fentre X.
 */

static void Realize (Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	PrimitiveWidget	w = (PrimitiveWidget) widget;

	*value_mask |= CWBitGravity | CWCursor;
	attributes -> bit_gravity = P.gravity;
	attributes -> cursor = P.cursor;
	XtCreateWindow (widget, InputOutput, (Visual *)CopyFromParent, *value_mask, attributes);
	XtAddEventHandler (widget, EnterWindowMask | LeaveWindowMask, False, EnterLeaveHandler, NULL);
	if (P.tip_message != NULL) {
		if (PCR.tipWidget == NULL) {
			PCR.tipWidget = XtCreatePopupShell ("tip", overrideShellWidgetClass, widget, NULL, 0);
			XtAddEventHandler (PCR.tipWidget, ExposureMask, False, tipDisplay, NULL);
			PCR.tipGC = XCreateGC (XtDisplay (widget), XtWindow (widget), 0, NULL);
			XSetForeground (XtDisplay (widget), PCR.tipGC, BlackPixelOfScreen (XtScreen (widget)));
		}
	}
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	PrimitiveWidget	w = (PrimitiveWidget) widget;
	Display*		dpy = XtDisplay (widget);

	if (P.shadow_thickness > 0) {
		XSetRegion (dpy, P.top_GC, region);
		XSetRegion (dpy, P.bot_GC, region);
		_DrawShadow (dpy , XtWindow (widget), P.bot_GC, P.top_GC,
						P.shadow_thickness, C.width, C.height,
						0, 0);
		XSetClipMask (dpy, P.top_GC, None);
		XSetClipMask (dpy, P.bot_GC, None);
	}
}

/************************************************************************
 *
 *  Initialize
 *     The main widget instance initialization routine.
 *
 ************************************************************************/

static void Initialize (Widget request, Widget w, ArgList al, Cardinal* n)
{
	PrimitiveWidget pw = (PrimitiveWidget) w;

	/*  Get the graphics contexts  */

	GetTopShadowGC (pw);
	GetBottomShadowGC (pw);
}

/************************************************************************
 *
 *  SetValues
 *     Perform and updating necessary for a set values call.
 *
 ************************************************************************/

static Boolean SetValues (Widget cu, Widget req, Widget w, ArgList al, Cardinal* cn)
{
	PrimitiveWidget	curpw = (PrimitiveWidget) cu;
	PrimitiveWidget	newpw = (PrimitiveWidget) w;
	Boolean 		do_redisplay = False;

	if (curpw -> primitive.tip_message != newpw -> primitive.tip_message &&
		newpw -> primitive.tip_message != NULL && XtIsRealized (w))
	{
		if (PCR.tipWidget == NULL) {
			PCR.tipWidget = XtCreatePopupShell ("tip", overrideShellWidgetClass, w, NULL, 0);
			XtAddEventHandler (PCR.tipWidget, ExposureMask, False, tipDisplay, NULL);
			PCR.tipGC = XCreateGC (XtDisplay (w), XtWindow (w), 0, NULL);
			XSetForeground (XtDisplay (w), PCR.tipGC, BlackPixelOfScreen (XtScreen (w)));
		}
	}
	if (newpw -> primitive.shadow_thickness < 0)
		newpw -> primitive.shadow_thickness = curpw -> primitive.shadow_thickness;

	if (newpw -> core.background_pixel != curpw -> core.background_pixel) {
		XrmValue	value;

		_GetTopShadowColor (w, 0, &value);
		newpw -> primitive.top_pixel = *(Pixel*)value.addr;
		_GetBottomShadowColor (w, 0, &value);
		newpw -> primitive.bot_pixel = *(Pixel*)value.addr;
        do_redisplay = TRUE;
	}

	if (curpw->primitive.top_pixel != newpw->primitive.top_pixel ||
		curpw->primitive.shadow_thickness != newpw->primitive.shadow_thickness ||
		curpw->primitive.bot_pixel != newpw->primitive.bot_pixel)
	{
		XtDestroyGC (newpw -> primitive.top_GC);
		XtDestroyGC (newpw -> primitive.bot_GC);

		GetTopShadowGC (newpw);
		GetBottomShadowGC (newpw);
		do_redisplay = True;
	}

	return do_redisplay;
}

/*
 *	Trace l'ombre pour donner une apparence de relief.
 */

void _EraseShadow (Display* display, Drawable d,
				   int size, int width, int height, int x, int y)
{
	if (width > 0 && size > 0)
	{
		XClearArea (display, d, x, y, width, size, False);
		XClearArea (display, d, x, y + height - size, width, size, False);
	}

	if (size > 0 && height - (2 * size) > 0)
	{
		XClearArea (display, d, x, y + size, size, height - (2 * size), False);
		XClearArea (display, d, x + width - size, y + size, size, 
				height - (2 * size), False);
	}
}

void _DrawShadow (Display* display, Drawable d, GC top_GC, GC bottom_GC,
				  int size, int width, int height, int x, int y)
{
	static XRectangle*	rects = NULL;
	static int		rect_count = 0;
	register XRectangle*	pr;
	register int		i;

	if (size > width / 2) size = width / 2;
	if (size > height / 2) size = height / 2;
	if (size <= 0) return;

	if (rect_count == 0) {
		rects = (XRectangle *) XtMalloc (sizeof (XRectangle) * size * 4);
		rect_count = size;
	}
	if (rect_count < size) {
		rects = (XRectangle *) XtRealloc ((char*)rects, sizeof (XRectangle) * size * 4);
		rect_count = size;
	}

	for (i = 0; i < size; i++) {
		register int	width_i = width - i;
		register int	height_i = height - i;
		/*
		 * ombre du haut.
		 */
		pr = rects + i;
		pr -> x = x;
		pr -> y = y + i;
		pr -> width = width_i;
		pr -> height = 1;
		/*
		 * ombre de gauche.
		 */
		pr += size;
		pr -> x = x + i;
		pr -> y = y;
		pr -> width = 1;
		pr -> height = height_i;
		/*
		 * ombre du bas.
		 */
		pr += size;
		pr -> x = x + i + 1;
		pr -> y = y + height_i - 1;
		pr -> width = width_i - 1;
		pr -> height = 1;
		/*
		 * ombre de droite.
		 */
		pr += size;
		pr -> x = x + width_i - 1;
		pr -> y = y + i + 1;
		pr -> width = 1;
		pr -> height = height_i - 1;
	}
	size += size;
	XFillRectangles (display, d, top_GC, rects, size);
	XFillRectangles (display, d, bottom_GC, rects + size, size);
}

void _DrawDiamondShadow (Display* display, Drawable d, int x, int y, int size,
						 GC topGC, GC bottomGC, GC centerGC, Boolean fill)
{
	XSegment seg[12];
	XPoint   pt[5];
	int midX, midY;

	if (size % 2 == 0)
		size--;

	midX = x + (size + 1) / 2;
	midY = y + (size + 1) / 2;

	/* COUNTER REVERSE DRAWING EFFECT ON TINY ToggleButtonS */
	if (size <= 3) {
		/*  The top shadow segments  */

		seg[0].x1 = x + size - 1;	/*  1  */
		seg[0].y1 = midY - 1;
		seg[0].x2 = midX - 1;		/*  2  */
		seg[0].y2 = y + size - 1;

		seg[1].x1 = x + size - 2;	/*  3  */
		seg[1].y1 = midY - 1;
		seg[1].x2 = midX - 1;		/*  4  */
		seg[1].y2 = y + size - 2;

		seg[2].x1 = x + size - 3;	/*  3  */
		seg[2].y1 = midY - 1;
		seg[2].x2 = midX - 1;		/*  4  */
		seg[2].y2 = y + size - 3;

		/*--*/
		seg[3].x1 = midX - 1;		/*  5  */
		seg[3].y1 = y + size - 1;
		seg[3].x2 = x;			/*  6  */
		seg[3].y2 = midY - 1;

		seg[4].x1 = midX - 1;		/*  7  */
		seg[4].y1 = y + size - 2;
		seg[4].x2 = x + 1;		/*  8  */
		seg[4].y2 = midY - 1;

		seg[5].x1 = midX - 1;		/*  7  */
		seg[5].y1 = y + size - 3;
		seg[5].x2 = x + 2;		/*  8  */
		seg[5].y2 = midY - 1;

		/*  The bottom shadow segments  */

		seg[6].x1 = x + size - 1;	/*  9  */
		seg[6].y1 = midY - 1;
		seg[6].x2 = midX - 1;		/*  10  */
		seg[6].y2 = y;

		seg[7].x1 = x + size - 2;	/*  11  */
		seg[7].y1 = midY - 1;
		seg[7].x2 = midX - 1;		/*  12  */
		seg[7].y2 = y + 1;

		seg[8].x1 = x + size - 3;	/*  11  */
		seg[8].y1 = midY - 1;
		seg[8].x2 = midX - 1;		/*  12  */
		seg[8].y2 = y + 2;

		/*--*/

		seg[9].x1 = midX - 1;		/*  13  */
		seg[9].y1 = y;
		seg[9].x2 = x;			/*  14  */
		seg[9].y2 = midY - 1;

		seg[10].x1 = midX - 1;		/*  15  */
		seg[10].y1 = y + 1;
		seg[10].x2 = x + 1;		/*  16  */
		seg[10].y2 = midY - 1;

		seg[11].x1 = midX - 1;		/*  15  */
		seg[11].y1 = y + 2;
		seg[11].x2 = x + 2;		/*  16  */
		seg[11].y2 = midY - 1;
	} else {
		/* NORMAL SIZED ToggleButtonS */
		/*  The top shadow segments  */

		seg[0].x1 = x;			/*  1  */
		seg[0].y1 = midY - 1;
		seg[0].x2 = midX - 1;		/*  2  */
		seg[0].y2 = y;

		seg[1].x1 = x + 1;		/*  3  */
		seg[1].y1 = midY - 1;
		seg[1].x2 = midX - 1;		/*  4  */
		seg[1].y2 = y + 1;

		seg[2].x1 = x + 2;		/*  3  */
		seg[2].y1 = midY - 1;
		seg[2].x2 = midX - 1;		/*  4  */
		seg[2].y2 = y + 2;

		/*--*/

		seg[3].x1 = midX - 1;		/*  5  */
		seg[3].y1 = y;
		seg[3].x2 = x + size - 1;	/*  6  */
		seg[3].y2 = midY - 1;

		seg[4].x1 = midX - 1;		/*  7  */
		seg[4].y1 = y + 1;
		seg[4].x2 = x + size - 2;	/*  8  */
		seg[4].y2 = midY - 1;

		seg[5].x1 = midX - 1;		/*  7  */
		seg[5].y1 = y + 2;
		seg[5].x2 = x + size - 3;	/*  8  */
		seg[5].y2 = midY - 1;

		/*  The bottom shadow segments  */
    
		seg[6].x1 = x;			/*  9  */
		seg[6].y1 = midY - 1;
		seg[6].x2 = midX - 1;		/*  10  */
		seg[6].y2 = y + size - 1;

		seg[7].x1 = x + 1;		/*  11  */
		seg[7].y1 = midY - 1;
		seg[7].x2 = midX - 1;		/*  12  */
		seg[7].y2 = y + size - 2;

		seg[8].x1 = x + 2;		/*  11  */
		seg[8].y1 = midY - 1;
		seg[8].x2 = midX - 1;		/*  12  */
		seg[8].y2 = y + size - 3;

		/*--*/

		seg[9].x1 = midX - 1;		/*  13  */
		seg[9].y1 = y + size - 1;
		seg[9].x2 = x + size - 1;	/*  14  */
		seg[9].y2 = midY - 1;

		seg[10].x1 = midX - 1;		/*  15  */
		seg[10].y1 = y + size - 2;
		seg[10].x2 = x + size - 2;	/*  16  */
		seg[10].y2 = midY - 1;

		seg[11].x1 = midX - 1;		/*  15  */
		seg[11].y1 = y + size - 3;
		seg[11].x2 = x + size - 3;	/*  16  */
		seg[11].y2 = midY - 1;
	}

	XDrawSegments (display, d, topGC, &seg[3], 3);
	XDrawSegments (display, d, bottomGC, &seg[6], 6);
	XDrawSegments (display, d, topGC, &seg[0], 3);
  
/* For Fill */
	if (fill) {
		pt[0].x = x + 3;
		pt[0].y = midY - 1;
		pt[1].x = midX - 1 ;
		pt[1].y = y + 2;
		pt[2].x = x + size - 3;
		pt[2].y = midY - 1;
		pt[3].x = midX - 1 ;
		pt[3].y = y + size - 3;
	} else {
		pt[0].x = x + 4;
		pt[0].y = midY - 1;
		pt[1].x = midX - 1;
		pt[1].y = y + 3;
		pt[2].x = x + size - 4;
		pt[2].y = midY - 1;
		pt[3].x = midX - 1;
		pt[3].y = y + size - 4;
	}

	if (pt[0].x > pt[1].x) {
		pt[1].x = pt[0].x;
		pt[2].x = pt[0].x;
		pt[3].x = pt[0].x;
	}

	if (pt[0].y < pt[1].y) {
		pt[1].x = pt[0].x;
		pt[2].x = pt[0].x;
		pt[3].x = pt[0].x;
	}
	XFillPolygon (display, d, centerGC, pt, 4, Convex, CoordModeOrigin);
}

/*
 *	Calcul des ombres du haut.
 */

typedef struct {
	Pixel	background;
	Pixel	top_pixel;
	Pixel	bottom_pixel;
	Pixel	arm_pixel;
} *ShadowCache;
static ShadowCache	cache;		/* cache pour les ombres */
static int		num_slots = 0;
static int		num_colors = 0;

/*
 *	Cherche l'entree dans le cache pour bg.
 *	retourne l'indice ou background == bg, ou bien
 *	num_colors sinon.
 */

static int lookupShadow (Pixel bg)
{
	int	i;

	for (i = 0; i < num_colors; i++)
		if (cache [i].background == bg)
			break;
	return i;
}

/*
 *	Remplit une entree dans le cache.
 */

static void storeShadow (int i, Pixel bg, Pixel top_pixel, Pixel bottom_pixel,
						 Pixel arm_pixel)
{
	/*
	 *	On a deja trouve l'entree: on remplit la case.
	 */
	if (i < num_colors) {
		if (top_pixel != (Pixel) ~0)	cache [i].top_pixel = top_pixel;
		if (bottom_pixel != (Pixel) ~0)	cache [i].bottom_pixel = bottom_pixel;
		if (arm_pixel != (Pixel) ~0)	cache [i].arm_pixel = arm_pixel;
		return;
	}
	/*
	 *	l'entree suivante fera l'affaire.
	 */
	if (num_colors >= num_slots) {
		num_slots += 10;
		cache = (ShadowCache) XtRealloc (
				(char*)cache,
				num_slots * sizeof (cache[0]));
		for (i = num_colors; i < num_slots; i++) {
			cache [i].top_pixel = ~0;
			cache [i].bottom_pixel = ~0;
			cache [i].arm_pixel = ~0;
		}
	}
	i = num_colors ++;
	cache [i].background = bg;
	if (top_pixel != (Pixel) ~0)	cache [i].top_pixel = top_pixel;
	if (bottom_pixel != (Pixel) ~0)	cache [i].bottom_pixel = bottom_pixel;
	if (arm_pixel != (Pixel) ~0)	cache [i].arm_pixel = arm_pixel;
}

void _GetTopShadowColor (Widget widget, int offset, XrmValue* value)
{
	static Pixel	pix;
	Screen*			screen = XtScreen (widget);
	XColor			bg_color;
	int				v;
	int				i;

	value -> size = sizeof (Pixel);
	value -> addr = (XtPointer) &pix;

	pix = WhitePixelOfScreen (screen);
	if (DefaultDepthOfScreen (screen) == 1)
		return;

	i = lookupShadow (widget -> core.background_pixel);
	if (i < num_colors && cache [i].top_pixel != (Pixel) ~0) {
		pix = cache [i].top_pixel;
		return;
	}

	bg_color.pixel = widget -> core.background_pixel;
	XQueryColor (DisplayOfScreen (screen), widget -> core.colormap, &bg_color);

	v = bg_color.red; v += (v * 2) / 5;
	bg_color.red = (v > 65535) ? 65535 : v;

	v = bg_color.green; v += (v * 2) / 5;
	bg_color.green = (v > 65535) ? 65535 : v;

	v = bg_color.blue; v += (v * 2) / 5;
	bg_color.blue = (v > 65535) ? 65535 : v;


	if (XAllocColor (DisplayOfScreen (screen), widget -> core.colormap, &bg_color)) {
		pix = bg_color.pixel;
		storeShadow (i, widget -> core.background_pixel, pix, ~0, ~0);
	} else
		XtWarning ("Can't allocate top shadow color.");
}

/*
 *	Calcul des ombres du bas.
 */

void _GetBottomShadowColor (Widget widget, int offset, XrmValue* value)
{
	static Pixel	pix;
	Screen*			screen = XtScreen (widget);
	XColor			bg_color;
	int				i;

	value -> size = sizeof (Pixel);
	value -> addr = (XtPointer) &pix;

	pix = BlackPixelOfScreen (screen);
	if (DefaultDepthOfScreen (screen) == 1)
		return;

	i = lookupShadow (widget -> core.background_pixel);
	if (i < num_colors && cache [i].bottom_pixel != (Pixel) ~0) {
		pix = cache [i].bottom_pixel;
		return;
	}

	bg_color.pixel = widget -> core.background_pixel;
	XQueryColor (DisplayOfScreen (screen), widget -> core.colormap, &bg_color);

	bg_color.red -= (bg_color.red * 9) / 20;
	bg_color.green -= (bg_color.green * 9) / 20;
	bg_color.blue -= (bg_color.blue * 9) / 20;

	if (XAllocColor (DisplayOfScreen (screen), widget -> core.colormap, &bg_color)) {
		pix = bg_color.pixel;
		storeShadow (i, widget -> core.background_pixel, ~0, pix, ~0);
	} else
		XtWarning ("Can't allocate bottom shadow color.");
}

/*
 *	Calcul la couleur de fond des boutons armes.
 */

void _GetArmColor (Widget widget, int offset, XrmValue* value)
{
	static Pixel	pix;
	Screen*			screen = XtScreen (widget);
	XColor			bg_color;
	int				i;

	value -> size = sizeof (Pixel);
	value -> addr = (XtPointer) &pix;

	pix = (Pixel) 12345;
	if (DefaultDepthOfScreen (screen) == 1)
		return;

	i = lookupShadow (widget -> core.background_pixel);
	if (i < num_colors && cache [i].arm_pixel != (Pixel) ~0) {
		pix = cache [i].arm_pixel;
		return;
	}

	bg_color.pixel = widget -> core.background_pixel;
	XQueryColor (DisplayOfScreen (screen), widget -> core.colormap, &bg_color);

	bg_color.red -= (bg_color.red * 3) / 20;
	bg_color.green -= (bg_color.green * 3) / 20;
	bg_color.blue -= (bg_color.blue * 3) / 20;

	if (XAllocColor (DisplayOfScreen (screen), widget -> core.colormap, &bg_color)) {
		pix = bg_color.pixel;
		storeShadow (i, widget -> core.background_pixel, ~0, ~0, pix);
	} else
		XtWarning ("Can't allocate bottom shadow color.");
}

void _FillOvalShadow (Display* dpy, Drawable d, GC gc,
					  int width, int height, int x, int y)
{
	XFillArc (dpy, d, gc, x, y, height, height, 90*64, 180*64);
	XFillRectangle (dpy, d, gc, x+(height-1)/2, y, width-height+1, height);
	XFillArc (dpy, d, gc, x+width-height, y, height, height, 270*64, 180*64);
}

void _DrawOvalShadow (Display* dpy, Drawable d, GC top_GC, GC bot_GC,
					  int st, int width, int height, int x, int y)
{
	XGCValues	gcv;

	if (st == 1)	st = 0;

	if (XGetGCValues (dpy, top_GC, GCLineWidth, &gcv) &&
	    gcv.line_width != st)
	{
		gcv.line_width = st;
		XChangeGC (dpy, top_GC, GCLineWidth, &gcv);
	}

	if (XGetGCValues (dpy, bot_GC, GCLineWidth, &gcv) &&
	    gcv.line_width != st)
	{
		gcv.line_width = st;
		XChangeGC (dpy, bot_GC, GCLineWidth, &gcv);
	}

	XDrawArc (dpy, d, top_GC, x, y, height, height, 90*64, 120*64);
	XDrawArc (dpy, d, bot_GC, x, y, height, height, 210*64, 60*64);
	XDrawLine (dpy, d, top_GC, x+height/2, y, x+width-height/2, y);

	XDrawArc (dpy, d, top_GC, x+width-height, y, height, height, 30*64, 60*64);
	XDrawArc (dpy, d, bot_GC, x+width-height, y, height, height, 270*64, 120*64);
	XDrawLine (dpy, d, bot_GC, x+height/2, y+height, x+width-height/2, y+height);
}

void _EraseOvalShadow (Display* dpy, Drawable d, GC gc,
					   int st, int width, int height, int x, int y)
{
	XGCValues	gcv;

	if (st == 1)	st = 0;

	if (XGetGCValues (dpy, gc, GCLineWidth, &gcv) &&
	    gcv.line_width != st)
	{
		gcv.line_width = st;
		XChangeGC (dpy, gc, GCLineWidth, &gcv);
	}

	XDrawArc (dpy, d, gc, x, y, height, height, 90*64, 180*64);
	XDrawLine (dpy, d, gc, x+height/2, y, x+width-height/2, y);
	XDrawArc (dpy, d, gc, x+width-height, y, height, height, 270*64, 180*64);
	XDrawLine (dpy, d, gc, x+height/2, y+height, x+width-height/2, y+height);
}

/*
 *	Pour poireauter un peu.
 */

static void _wakeup_ (XtPointer p_done, XtIntervalId* id)
{
	*(Boolean*) p_done = True;
}

void Wait (Widget widget, int msec)
{
	Boolean			_done;
	XtAppContext	app = XtWidgetToApplicationContext (widget);

	(void) XtAppAddTimeOut (app, msec, _wakeup_, &_done);
	for (_done = False; !_done;)
		XtAppProcessEvent (app, XtIMAll);
}

/*
 *	Simule un clic sur un bouton.
 */

void FakeClick (Widget w)
{
	String	rien;
	XEvent	e;

	XtCallActionProc (w, "Arm", &e, &rien, 0);
	Wait (w, 150);
	XtCallActionProc (w, "Disarm", &e, &rien, 0);
}

/*
 *	Pour modifier le champ 'managed' d'un widget.
 */

void SetManaged (Widget w, Boolean managed) {
	w -> core.managed = managed;
}

/*
 *	Dessine un rectangle avec les coins arrondis
 */

void _DrawOval (Display* dpy, Drawable drawable, GC gc, int x, int y, int w, int h, int r)
{
	int				d = 2*r;
	static XSegment	segments [4];

	segments [0].x1 = x+r; segments [0].y1 = y; segments [0].x2 = x+w-r; segments [0].y2 = y;
	segments [1].x1 = x+w; segments [1].y1 = y+r; segments [1].x2 = x+w; segments [1].y2 = y+h-r;
	segments [2].x1 = x+r; segments [2].y1 = y+h; segments [2].x2 = x+w-r; segments [2].y2 = y+h;
	segments [3].x1 = x; segments [3].y1 = y+r; segments [3].x2 = x; segments [3].y2 = y+h-r;
	XDrawSegments (dpy, drawable, gc, segments, 4);
	XDrawArc (dpy, drawable, gc, x, y, d, d, 90*64, 90*64);
	XDrawArc (dpy, drawable, gc, x+w-d, y, d, d, 0, 90*64);
	XDrawArc (dpy, drawable, gc, x+w-d, y+h-d, d, d, 270*64, 90*64);
	XDrawArc (dpy, drawable, gc, x, y+h-d, d, d, 180*64, 90*64);
}

/*
 *	Remplit un rectangle avec les coins arrondis
 */

void _FillOval (Display* dpy, Drawable drawable, GC gc, int x, int y, int w, int h, int r)
{
	int				d = 2*r;

	XFillRectangle (dpy, drawable, gc, x+r, y, w-d, r);
	XFillRectangle (dpy, drawable, gc, x, y+r, w, h-d);
	XFillRectangle (dpy, drawable, gc, x+r, y+h-r, w-d, r);

	XFillArc (dpy, drawable, gc, x, y, d, d, 90*64, 90*64);
	XFillArc (dpy, drawable, gc, x+w-d, y, d, d, 0, 90*64);
	XFillArc (dpy, drawable, gc, x+w-d, y+h-d, d, d, 270*64, 90*64);
	XFillArc (dpy, drawable, gc, x, y+h-d, d, d, 180*64, 90*64);
}

/*
 *	Dessine (et remplit si cent_gc est non nul) une pointe de flche.
 */

#define SWAP(x,y) temp = x; x = y; y = temp

void _DrawArrow (Display * dpy, Drawable d,
				 GC top_gc, GC bot_gc, GC cent_gc,
				 int x, int y,
				 int width, int height, int shadow_thick,
				 unsigned int direction)
{
	/* cent_gc might be NULL, which means don't draw anything on the center */
	/* in the current implementation, on shadow_thick = 2 or 0 are supported */

	static unsigned int	allocated = 0;
	static XRectangle*	top = 0;
	static XRectangle*	cent = 0;
	static XRectangle*	bot = 0;
	XRectangle*			rect_tmp;
	int					size;
	int					xOffset = 0;
	int					yOffset = 0;
	int					wwidth;
	int					start;
	int					temp, yy, i, h, w;
	short				t = 0, b = 0, c = 0;

	if (d == 0)
		return;

	/*  Get the size and the position and allocate the rectangle lists  */

	if (width > height) {
		size = height - 2;
		xOffset = (width - height) / 2;
	} else {
		size = width - 2;
		yOffset = (height - width) / 2;
	}
	if (size < 1)
		return;
	if (allocated < size) {
		top = (XRectangle *) XtRealloc ((char *) top,
									sizeof (XRectangle) * (size / 2 + 6));
		cent = (XRectangle *) XtRealloc ((char *) cent,
									sizeof (XRectangle) * (size / 2 + 6));
		bot = (XRectangle *) XtRealloc ((char *) bot,
									sizeof (XRectangle) * (size / 2 + 6));
		allocated = size;
	}
	if (direction == XpARROW_RIGHT || direction == XpARROW_LEFT) {
		SWAP (xOffset, yOffset);
	}
	/*  Set up a loop to generate the segments.  */

	wwidth = size;
	yy = size - 1 + yOffset;
	start = 1 + xOffset;

	while (wwidth > 0) {
		if (wwidth == 1) {
			top [t].x = start;
			top [t].y = yy + 1;
			top [t].width = 1;
			top [t].height = 1;
			t++;
		} else if (wwidth == 2) {
			if (size == 2 || (direction == XpARROW_UP ||
							  direction == XpARROW_LEFT)) {
				top [t].x = start;
				top [t].y = yy;
				top [t].width = 2;
				top [t].height = 1;
				t++;
				top [t].x = start;
				top [t].y = yy + 1;
				top [t].width = 1;
				top [t].height = 1;
				t++;
				bot [b].x = start + 1;
				bot [b].y = yy + 1;
				bot [b].width = 1;
				bot [b].height = 1;
				b++;
			} else if (direction == XpARROW_UP || direction == XpARROW_LEFT) {
				top [t].x = start;
				top [t].y = yy;
				top [t].width = 2;
				top [t].height = 1;
				t++;
				bot [b].x = start;
				bot [b].y = yy + 1;
				bot [b].width = 2;
				bot [b].height = 1;
				b++;
			}
		} else {
			if (start == 1 + xOffset) {
				if (direction == XpARROW_UP || direction == XpARROW_LEFT) {
					top [t].x = start;
					top [t].y = yy;
					top [t].width = 2;
					top [t].height = 1;
					t++;
					top [t].x = start;
					top [t].y = yy + 1;
					top [t].width = 1;
					top [t].height = 1;
					t++;
					bot [b].x = start + 1;
					bot [b].y = yy + 1;
					bot [b].width = 1;
					bot [b].height = 1;
					b++;
					bot [b].x = start + 2;
					bot [b].y = yy;
					bot [b].width = wwidth - 2;
					bot [b].height = 2;
					b++;
				} else {
					top [t].x = start;
					top [t].y = yy;
					top [t].width = 2;
					top [t].height = 1;
					t++;
					bot [b].x = start;
					bot [b].y = yy + 1;
					bot [b].width = 2;
					bot [b].height = 1;
					b++;
					bot [b].x = start + 2;
					bot [b].y = yy;
					bot [b].width = wwidth - 2;
					bot [b].height = 2;
					b++;
				}
			} else {
				top [t].x = start;
				top [t].y = yy;
				top [t].width = 2;
				top [t].height = 2;
				t++;
				bot [b].x = start + wwidth - 2;
				bot [b].y = yy;
				bot [b].width = 2;
				bot [b].height = 2;
				if (wwidth == 3) {
					bot [b].width = 1;
					bot [b].x += 1;
				}
				b++;
				if (wwidth > 4) {
					cent [c].x = start + 2;
					cent [c].y = yy;
					cent [c].width = wwidth - 4;
					cent [c].height = 2;
					c++;
				}
			}
		}
		start++;
		wwidth -= 2;
		yy -= 2;
	}

	if (direction == XpARROW_DOWN || direction == XpARROW_RIGHT) {
		rect_tmp = top;
		top = bot;
		bot = rect_tmp;
		SWAP (t, b);
	}
	/*  Transform the "up" pointing arrow to the correct direction  */

	switch (direction) {
	case XpARROW_LEFT:
		i = -1;
		do {
			i++;
			if (i < t) {
				SWAP (top [i].y, top [i].x);
				SWAP (top [i].width, top [i].height);
			}
			if (i < b) {
				SWAP (bot [i].y, bot [i].x);
				SWAP (bot [i].width, bot [i].height);
			}
			if (i < c) {
				SWAP (cent [i].y, cent [i].x);
				SWAP (cent [i].width, cent [i].height);
			}
		} while (i < t || i < b || i < c);
		break;

	case XpARROW_RIGHT:
		h = height - 2;
		w = width - 2;
		i = -1;
		do {
			i++;
			if (i < t) {
				SWAP (top [i].y, top [i].x);
				SWAP (top [i].width, top [i].height);
				top [i].x = w - top [i].x - top [i].width + 2;
				top [i].y = h - top [i].y - top [i].height + 2;
			}
			if (i < b) {
				SWAP (bot [i].y, bot [i].x);
				SWAP (bot [i].width, bot [i].height);
				bot [i].x = w - bot [i].x - bot [i].width + 2;
				bot [i].y = h - bot [i].y - bot [i].height + 2;
			}
			if (i < c) {
				SWAP (cent [i].y, cent [i].x);
				SWAP (cent [i].width, cent [i].height);
				cent [i].x = w - cent [i].x - cent [i].width + 2;
				cent [i].y = h - cent [i].y - cent [i].height + 2;
			}
		} while (i < t || i < b || i < c);
		break;

	case XpARROW_DOWN:
		w = width - 2;
		h = height - 2;
		i = -1;
		do {
			i++;
			if (i < t) {
				top [i].x = w - top [i].x - top [i].width + 2;
				top [i].y = h - top [i].y - top [i].height + 2;
			}
			if (i < b) {
				bot [i].x = w - bot [i].x - bot [i].width + 2;
				bot [i].y = h - bot [i].y - bot [i].height + 2;
			}
			if (i < c) {
				cent [i].x = w - cent [i].x - cent [i].width + 2;
				cent [i].y = h - cent [i].y - cent [i].height + 2;
			}
		} while (i < t || i < b || i < c);
		break;
	}

	if (x != 0 || y != 0) {
		for (i = 0; i < t; i++) {
			top [i].x += x;
			top [i].y += y;
		}
		for (i = 0; i < c; i++) {
			cent [i].x += x;
			cent [i].y += y;
		}
		for (i = 0; i < b; i++) {
			bot [i].x += x;
			bot [i].y += y;
		}
	}
	if (shadow_thick > 0) {
		XFillRectangles (dpy, d, top_gc, top, t);
		XFillRectangles (dpy, d, bot_gc, bot, b);
	} else {
		/* handle the case where arrow shadow_thickness = 0, which give
		   a flat arrow */
		if (cent_gc!= 0) {
			XFillRectangles (dpy, d, cent_gc, top, t);
			XFillRectangles (dpy, d, cent_gc, bot, b);
		}
	}
	if (cent_gc != 0)
		XFillRectangles (dpy, d, cent_gc, cent, c);
}

#undef SWAP
