/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Shell.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "PushButtonP.h"
#include "PushButton.h"

#define PCR	primitiveClassRec.primitive_class

#define W	w -> pushbutton
#define P	w -> primitive
#define C	w -> core

#define MARGINS	(P.shadow_thickness + W.margin + W.def_shadow)

#define monochrome (((Widget)w)->core.depth == 1)

static void		Destroy (Widget);
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);

static void		Enter (Widget, XEvent*, String*, Cardinal*);
static void		Leave (Widget, XEvent*, String*, Cardinal*);
static void		Arm (Widget, XEvent*, String*, Cardinal*);
static void		Disarm (Widget, XEvent*, String*, Cardinal*);
static void		Select (Widget, XEvent*, String*, Cardinal*);

static char defaultTranslations [] = "\
	<Key>Return:	Select()\n\
	<Key>space:	Select()\n\
	<Enter>:		Enter()\n\
	<Leave>:		Leave()\n\
	<Btn1Down>:	Arm()\n\
	<Btn1Up>:	Disarm()";

static	int	square_offset = 8;

static XtActionsRec actionsList [] =
{
	{"Arm",		Arm},
	{"Disarm",	Disarm},
	{"Enter",	Enter},
	{"Leave",	Leave},
	{"Select",	Select},
};

/*  The resource list for Push Button  */

static XtResource resources [] =
{     
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct*),
	  XtOffset (PushButtonWidget, pushbutton.fp), XtRString, "fixed"
	},
	{ XtNlabel, XtCLabel, XtRString, sizeof (char*),
	  XtOffset (PushButtonWidget, pushbutton.label), XtRImmediate, NULL
	},
	{ XtNpixmap, XtCPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (PushButtonWidget, pushbutton.pixmap), XtRPointer, NULL
	},
	{ XtNoblong, XtCOblong, XtRBoolean, sizeof (Boolean),
	  XtOffset (PushButtonWidget, pushbutton.oblong), XtRImmediate, (XtPointer) False
	},
	{ XtNinvertOnSelect, XtCInvertOnSelect, XtRBoolean, sizeof (Boolean),
	  XtOffset (PushButtonWidget, pushbutton.invert_on_select), XtRImmediate, (XtPointer) False
	},
	{ XtNtoggle, XtCToggle, XtRBoolean, sizeof (Boolean),
	  XtOffset (PushButtonWidget, pushbutton.toggle), XtRImmediate, (XtPointer) False
	},
	{ XtNset, XtCSet, XtRBoolean, sizeof (Boolean),
	  XtOffset (PushButtonWidget, pushbutton.set), XtRImmediate, (XtPointer) False
	},
	{ XtNmargin, XtCMargin, XtRShort, sizeof (short),
	  XtOffset (PushButtonWidget, pushbutton.margin), XtRImmediate, (XtPointer) 2
	},
	{ XtNshowAsDefault, XtCShowAsDefault, XtRShort, sizeof (short),
	  XtOffset (PushButtonWidget, pushbutton.def_shadow), XtRImmediate, (XtPointer) 0
	},
	{ XtNarmColor, XtCArmColor, XtRPixel, sizeof (Pixel),
	  XtOffset (PushButtonWidget, pushbutton.arm_color),
	  XtRCallProc, (XtPointer) _GetArmColor
	},
};

PushButtonClassRec pushButtonClassRec = {
    {
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"PushButton",						/* class_name */
		sizeof (PushButtonRec),				/* widget_size */
		NULL,								/* class_initialize */
		NULL,								/* class_part_init */
		FALSE,								/* class_inited */
		Initialize,							/* initialize */
		NULL,								/* initialize_hook */
		XtInheritRealize,					/* realize */
		actionsList,						/* actions */
		XtNumber (actionsList),				/* num_actions */
		resources,							/* resources */
		XtNumber (resources),				/* num_resources */
		NULLQUARK,							/* xrm_class */
		False,								/* compress_motion */
		True,								/* compress_exposure */
		True,								/* compress_enterlv */
		False,								/* visible_interest */
		Destroy,							/* destroy */
		Resize,								/* resize */
		Redisplay,							/* expose */
		SetValues,							/* set_values */
		NULL,								/* set_values_hook */
		XtInheritSetValuesAlmost,			/* set_values_almost */
		NULL,								/* get_values_hook */
		NULL,								/* accept_focus */
		XtVersion,							/* version */
		NULL,								/* callback_private */
		defaultTranslations,				/* tm_table */
		NULL,								/* query_geometry */
	},
	{
		NULL,
	},
	{
		NULL,
	},
};

WidgetClass pushButtonWidgetClass = (WidgetClass)&pushButtonClassRec;

/*
 *	le GC pour le label.
 */

static void get_labelGC (PushButtonWidget w)
{
	static Pixmap	greyPixmap = (Pixmap)0;
	static char		grey_bits [] = {
						0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55
					};
	XGCValues		values;
	unsigned long	valuemask;

	if (!greyPixmap)
		greyPixmap = XCreatePixmapFromBitmapData (XtDisplay ((Widget) w),
							RootWindowOfScreen (XtScreen ((Widget) w)),
							grey_bits, 8, 8,
							BlackPixelOfScreen (XtScreen ((Widget) w)),
							WhitePixelOfScreen (XtScreen ((Widget) w)),
							DefaultDepthOfScreen (XtScreen ((Widget) w)));

	valuemask = GCForeground | GCBackground | GCFont | GCTile;
	values.tile = greyPixmap;
	values.foreground = P.foreground;
	values.background = C.background_pixel;
	values.font = W.fp -> fid;

	W.label_GC = XtGetGC ((Widget) w, valuemask, &values);
}

/*
 *	Rcupre le pixmap.
 */

static void get_pixmap (PushButtonWidget w)
{
	Window	wraf;
	int		raf;
	int		width;
	int		height;
	int		depth;

	if (W.pixmap != 0) {
		XGetGeometry (XtDisplay (w), W.pixmap, &wraf, &raf, &raf,
						&width, &height, &raf, &depth);
		W.p_width = width;
		W.p_height = height;
		W.p_depth = depth;
	} else {
		W.p_width = 0;
		W.p_height = 0;
		W.p_depth = 0;
	}
}

static void setSize (PushButtonWidget w)
{
	if (W.pixmap) {
		C.width = W.p_width + 2 * MARGINS;
		C.height = W.p_height + 2 * MARGINS;
	} else {
		C.width = XTextWidth (W.fp, W.label, strlen (W.label)) + 2 * MARGINS;
		C.height = W.fp -> ascent + W.fp -> descent + 2 * MARGINS;
	}
	if (W.toggle)
		C.width += ((W.fp -> ascent + W.fp -> descent)*2)/3 + square_offset;
	else if (W.oblong)
		C.width += MARGINS;
}

/*
 *	Fonction d'initialisation de chaque instance.
 */

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	PushButtonWidget	w = (PushButtonWidget) new;

	if (W.label == NULL)
		W.label = XtNewString (C.name);
	else
		W.label = XtNewString (W.label);	/* s'assurer que c'est friable */
	get_pixmap (w);
	if (request -> core.width == 0 || request -> core.height == 0)
		setSize (w);
	get_labelGC (w);
	P.gravity = NorthWestGravity;
	W.width = C.width;
	W.height = C.height;
}

/*
 *	Traite le changement de taille (efface les ombres)
 */

static void Resize (Widget widget)
{
	PushButtonWidget	w = (PushButtonWidget) widget;
	Display*			dpy = XtDisplay (widget);
	Window				win = XtWindow (widget); 

	if (!XtIsRealized (widget)) {
		W.width = C.width;		/* mettre  jour les */
		W.height = C.height;	/* les dimensions */
		return;
	}
	if (W.toggle) {
		XClearArea (dpy, win, MARGINS, MARGINS, W.width - 2*MARGINS, W.height - 2*MARGINS, True);
		W.width = C.width;		/* mettre  jour les */
		W.height = C.height;	/* les dimensions */
		return;
	}
	if (W.pixmap) {
		XClearArea (dpy, win,
					(W.width - W.p_width) / 2,
					(W.height - W.p_height) / 2,
					W.p_width, W.p_height, False);
		XCopyArea (dpy, W.pixmap, win, W.label_GC,
					0, 0, W.p_width, W.p_height,
					(C.width - W.p_width) / 2,
					(C.height - W.p_height) / 2);
	} else
	/* rafficher le label */
	if (W.label) {
		int	length = strlen (W.label);
		int	pix_length = XTextWidth (W.fp, W.label, length);
		int	fh = W.fp -> ascent + W.fp -> descent;

		XClearArea (dpy, win,
					(W.width - pix_length) / 2, (W.height - fh) / 2,
					pix_length, fh, False);
		XDrawString (dpy, win, W.label_GC,
					 (C.width - pix_length) / 2, (C.height - fh) / 2 + W.fp -> ascent,
					 W.label, length);
	}
	/* rafficher les ombres */
	if (MARGINS > 0) {
		int	offset =  W.def_shadow;

		if (W.oblong) {
		/* les ombres oblongues */
			offset += W.margin;

			XSetForeground (dpy, W.label_GC, C.background_pixel);
			/* la super ombre des boutons 'par dfaut' */
			if (W.def_shadow > 0)
				_EraseOvalShadow (dpy, win, W.label_GC,
								  1, W.width - 2, W.height - 2, 1, 1);
			_EraseOvalShadow (dpy, win, W.label_GC,
							  1, W.width - 2*offset, W.height - 2*offset, offset, offset);
			XSetForeground (dpy, W.label_GC, P.foreground);
			/* le bouton est probablement relach */
			_DrawOvalShadow (dpy, win, P.top_GC, P.bot_GC, 1, C.width - 2*offset, C.height - 2*offset, offset, offset);
			if (W.def_shadow > 0)
				_DrawOvalShadow (dpy, win, P.bot_GC, P.top_GC, 1, C.width - 2, C.height - 2, 1, 1);
		} else {
		/* les ombres rectangulaires */
			/* la super ombre des boutons 'par dfaut' */
			if (offset > 0 && P.shadow_thickness > 0) {
				_EraseShadow (dpy, win, P.shadow_thickness, W.width, W.height, 0, 0);
				_DrawShadow (dpy, win, P.bot_GC, P.top_GC,
							 P.shadow_thickness, C.width, C.height, 0, 0);
			}
			/* les ombre standard */
			_EraseShadow (dpy, win, P.shadow_thickness,
							W.width - 2*offset, W.height - 2*offset, offset, offset);
			/* le bouton est probablement relach */
			_DrawShadow (dpy, win, P.top_GC, P.bot_GC, P.shadow_thickness,
							C.width - 2 * offset, C.height - 2 * offset, offset, offset);
		}
	}
	W.width = C.width;
	W.height = C.height;
}

/*
 *	La fonction de raffichage.
 */

static void draw_label (PushButtonWidget w)
{
	int	l;
	int	pl;

	if (W.label && *W.label) {
		l = strlen (W.label);
		pl = XTextWidth (W.fp, W.label, l);
		XDrawString (XtDisplay ((Widget)w), XtWindow ((Widget)w), W.label_GC,
						(C.width - pl) / 2,
						(C.height - W.fp -> ascent - W.fp -> descent) / 2 + W.fp -> ascent,
						W.label, l);
	}
}

static void redraw (PushButtonWidget w, Region region, Boolean all)
{
	Display*	dpy = XtDisplay ((Widget) w);
	Window		window = XtWindow ((Widget) w); 
	int			offset;
	int			x_label;
	int			square_size;
	int			l;
	int			pl;
	Pixel		fg = P.foreground;
	Pixel		bg = C.background_pixel;
	/*
	 *	Bouton oblong.
	 */
	if (W.oblong && !W.toggle) {
		offset =  W.margin + W.def_shadow;
		if (region) {
			XSetRegion (dpy, P.top_GC, region);
			XSetRegion (dpy, P.bot_GC, region);
			XSetRegion (dpy, W.label_GC, region);
		}
		if (monochrome || W.arm_color != 12345) {
			XSetForeground (dpy, W.label_GC, W.set ? W.arm_color : bg);
			if (monochrome && W.set)
				XSetFillStyle (dpy, W.label_GC, FillTiled);
			_FillOvalShadow (dpy, window, W.label_GC, C.width - 2*offset, C.height - 2*offset, offset, offset);
			XSetForeground (dpy, W.label_GC, fg);
			XSetFillStyle (dpy, W.label_GC, FillSolid);
		}
		if (W.set)
			_DrawOvalShadow (dpy, window, P.bot_GC, P.top_GC, 1, C.width - 2*offset, C.height - 2*offset, offset, offset);
		else
			_DrawOvalShadow (dpy, window, P.top_GC, P.bot_GC, 1, C.width - 2*offset, C.height - 2*offset, offset, offset);

		if (W.def_shadow > 0)
			_DrawOvalShadow (dpy, window, P.bot_GC, P.top_GC, 1, C.width - 2, C.height - 2, 1, 1);

		draw_label (w);	/* affiche le label */

		if (region) {
			XSetClipMask (dpy, P.top_GC, None);
			XSetClipMask (dpy, P.bot_GC, None);
			XSetClipMask (dpy, W.label_GC, None);
		}
		return;
	}
	/*
	 *	Affiche les ombres.
	 */
	if (P.shadow_thickness > 0) {
		int	x =  W.def_shadow;

		if (region) {
			XSetRegion (dpy, P.top_GC, region);
			XSetRegion (dpy, P.bot_GC, region);
		}

		if (!W.toggle && W.arm_color != 12345 && !all) {
			if (W.set)
				XSetForeground (dpy, W.label_GC, W.arm_color);
			else
				XSetForeground (dpy, W.label_GC, bg);

			XFillRectangle (dpy, window, W.label_GC,
								P.shadow_thickness,
								P.shadow_thickness,
								C.width - 2 * P.shadow_thickness,
								C.height - 2 * P.shadow_thickness);
			XSetForeground (dpy, W.label_GC, fg);
			all = True;
		}
		if (!W.toggle) {
			_DrawShadow (dpy, window,
							W.set ? P.bot_GC : P.top_GC,
							W.set ? P.top_GC : P.bot_GC,
							P.shadow_thickness,
							C.width - 2 * x, C.height - 2 * x,
							x, x);
			if (W.def_shadow > 0) {
				_DrawShadow (dpy, window, P.bot_GC, P.top_GC,
								P.shadow_thickness, C.width, C.height,
								0, 0);
			}
		}
		if (region) {
			XSetClipMask (dpy, P.top_GC, None);
			XSetClipMask (dpy, P.bot_GC, None);
		}
	}
	/*
	 *	dessiner le pixmap
	 */

	if (W.label && *W.label) {
		l = strlen (W.label);
		pl = XTextWidth (W.fp, W.label, l);
	} else
		l = 0;

	if (W.toggle) {
		if (W.pixmap) {
			square_size = W.p_height;
			offset = square_size + square_offset;
			x_label = offset + (C.width - offset - W.p_width) / 2;
		} else {
			square_size = ((W.fp -> ascent + W.fp -> descent)*2)/3;
			offset = square_size + square_offset;
			x_label = offset + (C.width - offset - pl) / 2;
		}
	} else {
		if (W.pixmap)
			x_label = (C.width - W.p_width) / 2;
		else
			x_label = (C.width - pl) / 2;
	}

	if (all || W.invert_on_select) {
		if (region)
			XSetRegion (dpy, W.label_GC, region);

		if (W.pixmap) {
			if (W.p_depth == 1)
				XCopyPlane (dpy, W.pixmap, window, W.label_GC, 0, 0,
								W.p_width, W.p_height,
								x_label,
								(C.height - W.p_height) / 2, 1);
			else
				XCopyArea (dpy, W.pixmap, window, W.label_GC, 0, 0,
								W.p_width, W.p_height,
								x_label,
								(C.height - W.p_height) / 2);
		} else if (l > 0) {
			XDrawString (dpy, window, W.label_GC, x_label,
							(C.height - W.fp -> ascent - W.fp -> descent) / 2 + W.fp -> ascent,
							W.label, l);
		}

		if (!W.toggle && W.invert_on_select && W.set) {
			XSetState (dpy, W.label_GC, fg ^ bg, bg, GXxor, AllPlanes);
			XFillRectangle (dpy, window, W.label_GC,
								P.shadow_thickness, P.shadow_thickness,
								C.width - 2 * P.shadow_thickness,
								C.height - 2 * P.shadow_thickness);
			XSetState (dpy, W.label_GC, fg, bg, GXcopy, AllPlanes);
		}

		if (region)
			XSetClipMask (dpy, W.label_GC, None);
	}
	if (W.toggle) {
		if (P.shadow_thickness > 0)
			if (W.oblong)
				_DrawOvalShadow (dpy, window,
									W.set ? P.bot_GC : P.top_GC,
									W.set ? P.top_GC : P.bot_GC,
									1,
									square_size,
									square_size,
									x_label - offset,
									(C.height - square_size) / 2);
			else
				_DrawShadow (dpy, window,
								W.set ? P.bot_GC : P.top_GC,
								W.set ? P.top_GC : P.bot_GC,
								P.shadow_thickness,
								square_size,
								square_size,
								x_label - offset,
								(C.height - square_size) / 2);

		if (W.invert_on_select) {
			if (!W.set)
				XSetForeground (dpy, W.label_GC, bg);
			XFillRectangle (dpy, window, W.label_GC,
							x_label - offset + P.shadow_thickness,
							(C.height - square_size) / 2 + P.shadow_thickness,
							square_size - 2 * P.shadow_thickness,
							square_size - 2 * P.shadow_thickness);
			if (!W.set)
				XSetForeground (dpy, W.label_GC, fg);
		} else
			if (W.arm_color != 12345) {
				XSetForeground (dpy, W.label_GC, W.set ? W.arm_color : bg);
				XFillRectangle (dpy, window, W.label_GC,
								x_label - offset + P.shadow_thickness,
								(C.height - square_size) / 2 + P.shadow_thickness,
								square_size - 2 * P.shadow_thickness,
								square_size - 2 * P.shadow_thickness);
				XSetForeground (dpy, W.label_GC, fg);
			}
	}
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	redraw ((PushButtonWidget) widget, region, True);
}

/*
 *	SetValues.
 */

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	PushButtonWidget	cu = (PushButtonWidget) current;
	PushButtonWidget	w = (PushButtonWidget) new;
	Boolean				do_redisplay = False;

	if (cu -> core.background_pixel != C.background_pixel) {
		XrmValue	value;

		_GetArmColor (new, 0, &value);
		W.arm_color = *(Pixel*)value.addr;
	}

	if (cu -> pushbutton.fp != W.fp ||
	    cu -> primitive.foreground != P.foreground ||
	    cu -> core.background_pixel != C.background_pixel) {
		XtDestroyGC (cu -> pushbutton.label_GC);
		get_labelGC (w);
		do_redisplay = True;
	}

	if (cu -> pushbutton.pixmap != W.pixmap) {
		get_pixmap (w);
		setSize (w);
		do_redisplay = True;
	}

	if (cu -> pushbutton.label != W.label) {
		XtFree (cu -> pushbutton.label);	/* librer l'ancien */
		W.label = XtNewString (W.label);	/* allouer le nouveau */
		setSize (w);
		do_redisplay = True;
	}

	if (cu -> pushbutton.toggle != W.toggle ||
		cu -> pushbutton.oblong != W.oblong ||
		cu -> pushbutton.def_shadow != W.def_shadow)
	{
		setSize (w);
		do_redisplay = True;
	}

	if (cu -> pushbutton.set != W.set && W.toggle)
		do_redisplay = True;

	return do_redisplay;
}

/*
 *	La procedure de destruction d'une instance.
 */

static void Destroy (Widget widget)
{
	PushButtonWidget	w = (PushButtonWidget) widget;

	XtDestroyGC (W.label_GC);
}

static Boolean	_armed = False;

static void Enter (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	XtAppContext		app = XtWidgetToApplicationContext (widget);
	PushButtonWidget	w = (PushButtonWidget) widget;

	if (W.toggle || !_armed)
		return;			/* ne concerne que les poussoirs arms */

	W.set = True;
	redraw (w, NULL, False);
}

static void Leave (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	PushButtonWidget	w = (PushButtonWidget) widget;

	if (W.toggle || !_armed)
		return;			/* ne concerne que les poussoirs arms */

	W.set = False;
	redraw (w, NULL, False);
}

/*
 *	L'action de selection.
 */

static void release (XtPointer cd, XtIntervalId* id)
{
	Disarm ((Widget) cd, NULL, NULL, NULL);
}

static void Select (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	Arm (widget, event, params, num_params);
	(void) XtAppAddTimeOut (XtWidgetToApplicationContext (widget), 200, release, (XtPointer) widget);
}

static void Arm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	PushButtonWidget	w = (PushButtonWidget) widget;

	if (!W.toggle) {
		_armed = True;
		W.set = True;
		redraw (w, NULL, False);
		if (PCR.tipTimer){
			XtRemoveTimeOut (PCR.tipTimer);
			PCR.tipTimer = 0;
		}
		XtCallCallbacks (widget, XtNselect, NULL);
	}
}

static void Disarm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	PushButtonWidget	w = (PushButtonWidget) widget;

	_armed = W.set;		/* le bouton tait-il appuy ? */
	W.set = W.toggle ? !W.set : False;
	redraw (w, NULL, False);
	if (PCR.tipTimer){
		XtRemoveTimeOut (PCR.tipTimer);
		PCR.tipTimer = 0;
	}
	if (W.toggle || _armed)
		XtCallCallbacks (widget, XtNrelease, (XtPointer) (long) W.set);
	_armed = False;
}

/*
 *	Dessine une ombre dans une fenetre quelconque en utilisant
 *	les GC du pushButton passe en argument.
 */

void DrawWindowShadow (Widget widget, Widget dest, XRectangle* prect, Boolean mode, int st, int x, int y, int width, int height)
{
	PushButtonWidget	w = (PushButtonWidget) widget;
	Display*			dpy = XtDisplay (dest);

	if (st <= 0)
		return;

	if (prect) {
		XSetClipRectangles (dpy, P.top_GC, 0, 0, prect, 1, Unsorted);
		XSetClipRectangles (dpy, P.bot_GC, 0, 0, prect, 1, Unsorted);
	}

	_DrawShadow (dpy, XtWindow (dest),
					mode ? P.top_GC : P.bot_GC,
					mode ? P.bot_GC : P.top_GC,
					st, width, height, x, y);
	if (prect) {
		XSetClipMask (dpy, P.top_GC, None);
		XSetClipMask (dpy, P.bot_GC, None);
	}
}


/*
 *	Change le label du bouton.
 */

void PushButtonSetLabel (Widget widget, char* label)
{
	PushButtonWidget	w = (PushButtonWidget) widget;

	XSetForeground (XtDisplay (widget), W.label_GC, C.background_pixel);
	draw_label (w);
	XSetForeground (XtDisplay (widget), W.label_GC, P.foreground);
	XtFree (W.label);				/* frire l'ancien */
	W.label = XtNewString (label);	/* pour allouer le nouveau */
	draw_label (w);
}
