/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/*************************************************************************
*
*	La boite de dialogue gnrique.
*
*************************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "Container.h"
#include "ContainerP.h"
#include "PushButton.h"
#include "PushButtonP.h"
#include "Wix.h"
#include "Query.h"
#include "QueryP.h"

#define MARGINS (w->manager.shadow_thickness)

static XtGeometryResult
				GeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void		ClassInitialize (void);
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static void		ChangeManaged (Widget);

static	void	OkAction (Widget, XEvent*, String*, Cardinal*);
static	void	ConfirmAction (Widget, XEvent*, String*, Cardinal*);
static	void	CancelAction (Widget, XEvent*, String*, Cardinal*);

static void		_resize (QueryWidget, Boolean);

static XtTranslations	traverseTranslations = NULL;

static XtResource resources [] = {
	{ XtNlabel1, XtCLabel, XtRString, sizeof (String),
	  XtOffset (QueryWidget, query.label1), XtRPointer, NULL
	},
	{ XtNlabel2, XtCLabel, XtRString, sizeof (String),
	  XtOffset (QueryWidget, query.label2), XtRPointer, NULL
	},
	{ XtNlabel3, XtCLabel, XtRString, sizeof (String),
	  XtOffset (QueryWidget, query.label3), XtRPointer, NULL
	},
	{ XtNokLabel, XtCLabel, XtRString, sizeof (String),
	  XtOffset (QueryWidget, query.ok_label), XtRPointer, NULL
	},
	{ XtNapplyLabel, XtCLabel, XtRString, sizeof (String),
	  XtOffset (QueryWidget, query.apply_label), XtRPointer, NULL
	},
	{ XtNcancelLabel, XtCLabel, XtRString, sizeof (String),
	  XtOffset (QueryWidget, query.cancel_label), XtRPointer, NULL
	},
	{ XtNokCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (QueryWidget, query.ok_callback), XtRPointer, NULL
	},
	{ XtNapplyCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (QueryWidget, query.apply_callback), XtRPointer, NULL
	},
	{ XtNcancelCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (QueryWidget, query.cancel_callback), XtRPointer, NULL
	},
	{ XtNtitle, XtCTitle, XtRString, sizeof (String),
	  XtOffset (QueryWidget, query.title), XtRImmediate, NULL
	},
	{ XtNoffset, XtCOffset, XtRInt, sizeof (int),
	  XtOffset (QueryWidget, query.offset), XtRImmediate, (XtPointer) 10
	},
	{ XtNpixmap, XtCPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (QueryWidget, query.pixmap), XtRImmediate, NULL
	},
	{ XtNprompt, XtCPrompt, XtRInt, sizeof (int),
	  XtOffset (QueryWidget, query.prompt), XtRImmediate, (XtPointer) NULL
	},
	{ XtNdefaultButton, XtCDefaultButton, XtRInt, sizeof (int),
	  XtOffset (QueryWidget, query.def_button), XtRImmediate, (XtPointer) -1
	},
};

static char defaultTranslations [] =
	"<Key>Return:	confirm()\n\
	 <Key>Y:		ok()\n\
	 <Key>N:		cancel()\n\
	 Ctrl <Key>G:	cancel()";

static XtActionsRec actionsList[] =
{
	{"ok",		OkAction },
	{"confirm",	ConfirmAction },
	{"cancel",	CancelAction },
};

QueryClassRec queryClassRec = {
  /* core_class fields */
  {
    /* superclass         */    (WidgetClass) &containerClassRec,
    /* class_name         */    "Query",
    /* widget_size        */    sizeof(QueryRec),
    /* class_initialize   */    ClassInitialize,
    /* class_partinit     */    NULL,
    /* class_inited       */	False,
    /* initialize         */    Initialize,
    /* Init hook	  */    NULL,
    /* realize            */    XtInheritRealize,
    /* actions		  */	actionsList,
    /* num_actions	  */	XtNumber(actionsList),
    /* resources          */    resources,
    /* num_resources      */    XtNumber (resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	False,
    /* compress_exposure  */	True,
    /* compress_enterleave*/	False,
    /* visible_interest   */    False,
    /* destroy            */    NULL,
    /* resize             */    Resize,
    /* expose             */    Redisplay,
    /* set_values         */    SetValues,
    /* set values hook    */    NULL,
    /* set values almost  */    XtInheritSetValuesAlmost,
    /* get values hook    */    NULL,
    /* accept_focus       */    NULL,
    /* Version            */    XtVersion,
    /* PRIVATE cb list    */    NULL,
    /* TM Table           */    defaultTranslations,
    /* query_geom         */    NULL,
  },{
/* composite_class fields */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension */		NULL,
  },{
/* Constraint class Init */
    NULL,
    0,
    0,
    NULL,
    NULL,
    NULL
  },{
/* manager_class fields */
	NULL,
  },{
/* Query class - none */     
   0
 }	
};

WidgetClass queryWidgetClass = (WidgetClass)&queryClassRec;

/*
 *	Les callbacks pour sortir de la boite.
 */

static void ConfirmAction (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	QueryWidget	w = (QueryWidget) widget;

	if (W.prompt == 0)
		switch (W.def_button) {
			case CANCEL_BUTTON:
				FakeClick (CAN);
				return;

			case OK_BUTTON:
				FakeClick (OK);
				return;
		}
}

static void CancelAction (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	QueryWidget	w = (QueryWidget) widget;

	if (W.prompt == 0)
		FakeClick (CAN);
}

static void OkAction (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	QueryWidget	w = (QueryWidget) widget;

	if (W.prompt == 0)
		FakeClick (OK);
}

static void de_activate (Widget wix)
{
	StopBlink ();
	DeselectBuffer (wix);
	RemoveCursor (wix);
	XtVaSetValues (wix, XtNnoCursor, True, NULL);
}

static void re_activate (Widget wix)
{
	XtVaSetValues (wix, XtNnoCursor, False, NULL);
	ShowCursor (wix);
	StartBlink (wix);
}

static void applyCallback (Widget widget, XtPointer client, XtPointer call)
{
	QueryWidget	w = (QueryWidget) client;

	XtCallCallbackList ((Widget) w, W.apply_callback, NULL);
}

static void cancelCallback (Widget widget, XtPointer client, XtPointer call)
{
	QueryWidget	w = (QueryWidget) client;

	if (W.prompt >= 2)
		de_activate (WIX2);
	if (W.prompt == 3)
		de_activate (WIX3);
	if (W.prompt >= 1)
		re_activate (WIX1);
	XtCallCallbackList ((Widget) w, W.cancel_callback, NULL);
}

static void doneCallback (Widget widget, XtPointer client, XtPointer call)
{
	QueryWidget	w = (QueryWidget) client;
	char*		p = NULL;
	static char	rep [256];

	if (W.prompt == 1 && WIX1 && WIX1 -> core.managed) {
		p = WixCopyBuffer (WIX1);
		strncpy (rep, p, 255);
		rep [255] = 0;
		XtFree (p);
		p = rep;
	}
	if (W.prompt >= 2)
		de_activate (WIX2);
	if (W.prompt == 3)
		de_activate (WIX3);
	if (W.prompt >= 1)
		re_activate (WIX1);

	XtCallCallbackList ((Widget) w, W.ok_callback, (XtPointer) p);
}

static void kbdCallback (Widget wix, XtPointer client, XtPointer call)
{
	QueryWidget	w = (QueryWidget) client;
	XKeyEvent*	e = (XKeyEvent*) call;
	char		c;
	KeySym		ks;

	if (XLookupString (e, &c, 1, &ks, 0)) {
		if (c == '\r') {
			if (W.prompt == 1 || wix == WIX2)
				FakeClick (OK);
			else if (W.prompt == 2 && wix == WIX1) {
				de_activate (wix);
				XtSetKeyboardFocus ((Widget) w, WIX2);
				re_activate (WIX2);
			} else if (W.prompt == 3 && wix == WIX2) {
				de_activate (wix);
				XtSetKeyboardFocus ((Widget) w, WIX3);
				re_activate (WIX3);
			}
			return;
		}
		if (c == ('g'&037)) {
			FakeClick (CAN);
			return;
		}
	} else if ((ks == XK_Up || ks == XK_Down) && W.prompt == 2) {
		if (wix == WIX2) {
			de_activate (wix);
			XtSetKeyboardFocus ((Widget) w, WIX1);
			re_activate (WIX1);
		} else {
			de_activate (WIX1);
			XtSetKeyboardFocus ((Widget) w, WIX2);
			re_activate (WIX2);
		}
	} else if (W.prompt == 3) {
		if (ks == XK_Up) {
			if (wix == WIX1) {
				de_activate (wix);
				XtSetKeyboardFocus ((Widget) w, WIX3);
				re_activate (WIX3);
			} else if (wix == WIX2) {
				de_activate (wix);
				XtSetKeyboardFocus ((Widget) w, WIX1);
				re_activate (WIX1);
			} else {
				de_activate (WIX3);
				XtSetKeyboardFocus ((Widget) w, WIX2);
				re_activate (WIX2);
			}
		} else
		if (ks == XK_Down) {
			if (wix == WIX1) {
				de_activate (wix);
				XtSetKeyboardFocus ((Widget) w, WIX2);
				re_activate (WIX2);
			} else if (wix == WIX2) {
				de_activate (wix);
				XtSetKeyboardFocus ((Widget) w, WIX3);
				re_activate (WIX3);
			} else {
				de_activate (WIX3);
				XtSetKeyboardFocus ((Widget) w, WIX1);
				re_activate (WIX1);
			}
		}
	}
}

/*
 *	Fait apparaitre la boite.
 */

void PopQuery (Widget widget)
{
	QueryWidget	w = (QueryWidget) widget;

	XtManageChild (widget);
	if (XtIsRealized (widget))
		XRaiseWindow (XtDisplay (widget), XtWindow (widget));
	if (WIX1 && WIX1 -> core.managed) {
		XtSetKeyboardFocus (widget, WIX1);
		re_activate (WIX1);
	} else
		XtSetKeyboardFocus (widget, None);
}

static void ClassInitialize (void)
{
	traverseTranslations = XtParseTranslationTable (
									"<Key>Up:	activate()\n\
									 <Key>Down:	activate()");
}

static void get_dimemsions (QueryWidget w)
{
	int				raf;
	Window			wraf;
	int				wi;
	int				he;
	int				depth;
	XFontStruct*	fp = ((PushButtonWidget)OK) -> pushbutton.fp;

	/*
	 *	Le pixmap.
	 */
	W.height = 0;
	if (W.pixmap) {
		XGetGeometry (XtDisplay ((Widget) w), W.pixmap, &wraf, &raf, &raf,
						&wi, &he, &raf, &depth);
		W.p_width = wi;
		W.p_height = he;
		W.p_depth = depth;
		W.height = he;
	} else {
		W.p_width = 0;
		W.p_height = 0;
	}
	/*
	 *	Le titre.
	 */
	W.title_width = 0;
	W.title_nl = 0;
	if (W.title) {
		char*	p = W.title;
		char*	e = p;
		int		lw;				/* largeur de la ligne courante */
		int		h;				/* hauteur totale du title */

		for (;;) {
			switch (*e++) {
				case 0:
				case '\n':
					W.title_nl ++;
					lw = XTextWidth (fp, p, e-p-1);
					if (W.title_width < lw)
						W.title_width = lw;
					if (e [-1] == 0)
						break;
					p = e;
					continue;

				default:
					continue;
			}
			break;
		}
		h = W.title_nl * (fp -> ascent + fp -> descent);
		if (W.height < h)
			W.height = h;
	}
	/*
	 *	Calculer l'offset des wix pour laisser la place aux labels.
	 */
	W.wix_offset = 0;
	if (W.prompt > 0 && W.label1)
		W.wix_offset = XTextWidth (fp, W.label1, strlen (W.label1));

	if (W.prompt > 1 && W.label2) {
		raf = XTextWidth (fp, W.label2, strlen (W.label2));
		if (W.wix_offset < raf)
			W.wix_offset = raf;
	}

	if (W.prompt > 2 && W.label3) {
		raf = XTextWidth (fp, W.label3, strlen (W.label3));
		if (W.wix_offset < raf)
			W.wix_offset = raf;
	}

	if (W.wix_offset)
		W.wix_offset += W.offset;
}

static void Initialize (Widget request, Widget new, Arg* arglist, Cardinal* nargs)
{
	QueryWidget	w = (QueryWidget) new;
	Boolean		oblong;
	Arg			args [10];
	int			n;

	if (request -> core.width == 0)
		C.width = DEFAULT_WIDTH;

	if (request -> core.height == 0)
		C.height = DEFAULT_HEIGHT;

	WORK = NULL;
	WIX1 = NULL;
	WIX2 = NULL;
	WIX3 = NULL;
	if (W.prompt > 0) {
		n = 0;
		XtSetArg (args [n], XtNmonoLine, True), n++;
		XtSetArg (args [n], XtNkillStack, False), n++;
		XtSetArg (args [n], XtNtransientSelection, True), n++;
		WIX1 = XtCreateManagedWidget ("edit1", wixWidgetClass, new, args, n);
		XtAddCallback (WIX1, XtNactivateCallback, kbdCallback, (XtPointer) w);
		if (W.prompt >= 2) {
			WIX2 = XtCreateManagedWidget ("edit2", wixWidgetClass, new, args, n);
			XtOverrideTranslations (WIX1, traverseTranslations);
			XtOverrideTranslations (WIX2, traverseTranslations);
			XtAddCallback (WIX2, XtNactivateCallback, kbdCallback, (XtPointer) w);
		}
		if (W.prompt == 3) {
			WIX3 = XtCreateManagedWidget ("edit3", wixWidgetClass, new, args, n);
			XtOverrideTranslations (WIX3, traverseTranslations);
			XtAddCallback (WIX3, XtNactivateCallback, kbdCallback, (XtPointer) w);
		}
	}

	OK = XtVaCreateManagedWidget ("Ok", pushButtonWidgetClass, new,
									XtNlabel, W.ok_label,
									NULL);
	XtVaGetValues (OK, XtNoblong, &oblong, NULL);
	if (W.def_button == OK_BUTTON)
		XtVaSetValues (OK,
			XtNshowAsDefault, oblong ? 3 : 3 * w->manager.shadow_thickness,
			NULL);
	XtAddCallback (OK, XtNrelease, doneCallback, (XtPointer) w);

	n = 0;
	if (W.def_button == CANCEL_BUTTON)
		XtSetArg (args [n], XtNshowAsDefault, oblong ? 3 : 3 * w->manager.shadow_thickness), n++;
	XtSetArg (args [n], XtNlabel, W.cancel_label), n++;
	CAN = XtCreateManagedWidget ("Cancel", pushButtonWidgetClass, new, args, n);
	XtAddCallback (CAN, XtNrelease, cancelCallback, (XtPointer) w);

	APPLY = XtVaCreateWidget ("Apply", pushButtonWidgetClass, new,
									XtNlabel,	W.apply_label,
									NULL);
	XtAddCallback (APPLY, XtNrelease, applyCallback, (XtPointer) w);
	get_dimemsions (w);
	_resize (w, True);
}

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* arglist, Cardinal* nargs)
{
	QueryWidget	cu = (QueryWidget) current;
	QueryWidget	w = (QueryWidget) new;
	Boolean		do_redisplay = False;
	Arg			args [4];
	int			n;

	if (cu -> query.pixmap != W.pixmap ||
	    cu -> query.title != W.title ||
	    cu -> query.label1 != W.label1 ||
	    cu -> query.label2 != W.label2 ||
	    cu -> query.label3 != W.label3)
	{
		get_dimemsions (w);
		if (cu -> query.def_button == W.def_button && cu -> query.prompt == W.prompt)
			_resize (w, True);
		do_redisplay = True;
	}

	if (cu -> query.apply_label != W.apply_label) {
		XtVaSetValues (APPLY, XtNlabel, W.apply_label, NULL);
		do_redisplay = True;
	}

	if (cu -> query.ok_label != W.ok_label) {
		XtVaSetValues (OK, XtNlabel, W.ok_label, NULL);
		do_redisplay = True;
	}

	if (cu -> query.cancel_label != W.cancel_label) {
		XtVaSetValues (CAN, XtNlabel, W.cancel_label, NULL);
		do_redisplay = True;
	}

	if (cu -> query.def_button != W.def_button) {
		Boolean	oblong;

		XtVaGetValues (OK, XtNoblong, &oblong, NULL);
		XtSetArg (args [0], XtNshowAsDefault, 0);
		XtSetArg (args [1], XtNshowAsDefault, oblong ? 3 : 3 * w->manager.shadow_thickness);
		if (W.def_button == CANCEL_BUTTON) {
			XtSetValues (CAN, args+1, 1);
			XtSetValues (OK, args, 1);
		}
		else if (W.def_button == OK_BUTTON) {
			XtSetValues (CAN, args, 1);
			XtSetValues (OK, args+1, 1);
		}
		if (cu -> query.prompt == W.prompt)	/* pas la peine de le */
			_resize (w, True);		/* faire deux fois */
		do_redisplay = True;
	}

	if (cu -> query.prompt != W.prompt) {
		n = 0;
		XtSetArg (args [n], XtNmonoLine, True), n++;
		XtSetArg (args [n], XtNkillStack, False), n++;

		switch (W.prompt) {
		case 0:
			XtSetKeyboardFocus (new, None);
			if (WIX1) XtUnmanageChild (WIX1);
			if (WIX2) XtUnmanageChild (WIX2);
			break;

		case 1:
			if (WIX2) XtUnmanageChild (WIX2);
			if (WIX1) XtManageChild (WIX1);
			else {
				WIX1 = XtCreateManagedWidget ("edit1", wixWidgetClass, new, args, n);
				XtAddCallback (WIX1, XtNactivateCallback, kbdCallback, (XtPointer) w);
				XtOverrideTranslations (WIX1, traverseTranslations);
			}
			XtSetKeyboardFocus (new, WIX1);
			break;

		case 2:
			if (WIX1) XtManageChild (WIX1);
			else {
				WIX1 = XtCreateManagedWidget ("edit1", wixWidgetClass, new, args, n);
				XtAddCallback (WIX1, XtNactivateCallback, kbdCallback, (XtPointer) w);
				XtOverrideTranslations (WIX1, traverseTranslations);
			}
			if (WIX2) XtManageChild (WIX2);
			else {
				WIX2 = XtCreateManagedWidget ("edit2", wixWidgetClass, new, args, n);
				XtAddCallback (WIX2, XtNactivateCallback, kbdCallback, (XtPointer) w);
				XtOverrideTranslations (WIX2, traverseTranslations);
			}
			XtSetKeyboardFocus (new, WIX1);
			break;

		case 3:
			if (WIX1) XtManageChild (WIX1);
			else {
				WIX1 = XtCreateManagedWidget ("edit1", wixWidgetClass, new, args, n);
				XtAddCallback (WIX1, XtNactivateCallback, kbdCallback, (XtPointer) w);
				XtOverrideTranslations (WIX1, traverseTranslations);
			}
			if (WIX2) XtManageChild (WIX2);
			else {
				WIX2 = XtCreateManagedWidget ("edit2", wixWidgetClass, new, args, n);
				XtAddCallback (WIX2, XtNactivateCallback, kbdCallback, (XtPointer) w);
				XtOverrideTranslations (WIX2, traverseTranslations);
			}
			if (WIX3) XtManageChild (WIX3);
			else {
				WIX3 = XtCreateManagedWidget ("edit3", wixWidgetClass, new, args, n);
				XtAddCallback (WIX3, XtNactivateCallback, kbdCallback, (XtPointer) w);
				XtOverrideTranslations (WIX3, traverseTranslations);
			}
			XtSetKeyboardFocus (new, WIX1);
			break;
		}
		_resize (w, True);
		do_redisplay = True;
	}
	return do_redisplay;
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	QueryWidget	w = (QueryWidget) widget;
	Display*	dpy = XtDisplay (w);
	PushButtonWidget
				ok = (PushButtonWidget)OK;
	XFontStruct*
				fp = ok -> pushbutton.fp;
	int		 	st = ok -> primitive.shadow_thickness;
	int			wixHeight1 = (WIX1 && WIX1 -> core.managed) ? WIX1 -> core.height + W.offset : 0;
	int			wixHeight2 = (WIX2 && WIX2 -> core.managed) ? WIX2 -> core.height + W.offset : 0;
	int			wixHeight3 = (WIX3 && WIX3 -> core.managed) ? WIX3 -> core.height + W.offset : 0;
	int			workHeight =  (WORK && WORK -> core.managed) ? WORK -> core.height + W.offset/2 : 0;

	(*containerClassRec.core_class.expose) (widget, event, region);
	if (st > 0)
		_DrawShadow (dpy, XtWindow (w),
						ok -> primitive.bot_GC, ok -> primitive.top_GC, st,
						C.width - 4, 2,
						2, W.sep_y - 1);
	if (W.pixmap)
	    if (W.p_depth == 1)
			XCopyPlane (dpy, W.pixmap, XtWindow (widget), ok -> pushbutton.label_GC,
							0, 0, W.p_width, W.p_height,
							W.offset + MARGINS,
							(W.sep_y - workHeight - wixHeight1 - wixHeight2 - wixHeight3 - W.height) / 2, 1);
	    else
			XCopyArea (dpy, W.pixmap, XtWindow (widget), ok -> pushbutton.label_GC,
					0, 0, W.p_width, W.p_height,
					W.offset + MARGINS,
					(W.sep_y - workHeight - wixHeight1 - wixHeight2 - wixHeight3 - W.height) / 2);
	if (W.title) {
		char*	p = W.title;
		char*	e = p;
		int		x;
		int		y = (W.sep_y - workHeight - wixHeight1 - wixHeight2 - wixHeight3 - W.height) / 2 + fp -> ascent;
		/*
		 *	Le message est centr s'il est sur une seule ligne, sinon il est align  gauche.
		 */
		x = (W.title_nl == 1)
			? (C.width - W.p_width - W.offset - W.title_width) / 2 + W.p_width + W.offset
			: W.p_width + 2*W.offset;
		for (;;) {
			switch (*e++) {
				case 0:
				case '\n':
					XDrawString (dpy, XtWindow ((Widget) w),
									ok -> pushbutton.label_GC,
									x, y, p, e-p-1);
					if (e [-1] == 0)
						break;
					y += fp -> ascent + fp -> descent;		/* une ligne de plus */
					p = e;
					continue;

				default:
					continue;
			}
			break;
		}
	}
	if (W.prompt > 0 && W.label1)
		XDrawString (dpy, XtWindow ((Widget) w), ok -> pushbutton.label_GC,
						st + W.offset/2, W.sep_y + 2 * st - wixHeight1 - wixHeight2 - wixHeight3 + fp -> ascent,
						W.label1, strlen (W.label1));
	if (W.prompt > 1 && W.label2)
		XDrawString (dpy, XtWindow ((Widget) w), ok -> pushbutton.label_GC,
						st + W.offset/2, W.sep_y + 2 * st - wixHeight3 - wixHeight2 + fp -> ascent,
						W.label2, strlen (W.label2));
	if (W.prompt > 2 && W.label3)
		XDrawString (dpy, XtWindow ((Widget) w), ok -> pushbutton.label_GC,
						st + W.offset/2, W.sep_y + 2 * st - wixHeight3 + fp -> ascent,
						W.label3, strlen (W.label3));
}

#define sup(x,y) ((x)>(y) ? (x) : (y))

static void _resize (QueryWidget w, Boolean flag)
{
	int			offset = W.offset;
	Dimension	workWidth =  (WORK && WORK -> core.managed) ? WORK -> core.width : 0;
	Dimension	workHeight =  (WORK && WORK -> core.managed) ? WORK -> core.height + offset : 0;
	Dimension	h;
	Dimension	wixHeight1 = (WIX1 && WIX1 -> core.managed) ? WIX1 -> core.height + offset : 0;
	Dimension	wixHeight2 = (WIX2 && WIX2 -> core.managed) ? WIX2 -> core.height + offset : 0;
	Dimension	wixHeight3 = (WIX3 && WIX3 -> core.managed) ? WIX3 -> core.height + offset : 0;
	Dimension	wixWidth = wixHeight1 ? WIX1 -> core.width : 0;
	Dimension	okHeight = OK -> core.managed ? OK -> core.height : 0;
	Dimension	cancelHeight = CAN -> core.managed ? CAN -> core.height : 0;
	Dimension	applyHeight = APPLY -> core.managed ? APPLY -> core.height : 0;
	Dimension	okWidth = OK -> core.managed ? OK -> core.width : 0;
	Dimension	applyWidth = APPLY -> core.managed ? APPLY -> core.width : 0;
	Dimension	cancelWidth = CAN -> core.managed ? CAN -> core.width : 0;
	Dimension	can_sha = cancelWidth ? ((PushButtonWidget)CAN) -> pushbutton.def_shadow : 0;
	Dimension	ok_sha = okWidth ? ((PushButtonWidget)OK) -> pushbutton.def_shadow : 0;
	Dimension	apply_sha = can_sha > ok_sha ? can_sha : ok_sha;
	Dimension	buttonHeight = okHeight > cancelHeight ? okHeight : cancelHeight;

	if (okWidth < cancelWidth)
		okWidth = cancelWidth;

	if (flag) {
		C.width = sup (okWidth + applyWidth + cancelWidth + 3*offset,
						sup (W.title_width + W.p_width,
							sup (wixWidth, workWidth)))
					+ W.wix_offset + 2 * (offset + MARGINS);
		C.height = workHeight + 2 * MARGINS + 4 * offset + buttonHeight + 2 + wixHeight1 + wixHeight2 + wixHeight3 + W.height;
	}

	W.sep_y = C.height - MARGINS - buttonHeight - 2 * offset;

	if (wixHeight3)
		XtConfigureWidget (WIX3, W.wix_offset + offset + MARGINS, W.sep_y - wixHeight3, C.width - 2 * (offset + MARGINS) - W.wix_offset, wixHeight3 - offset, 0);

	if (wixHeight2)
		XtConfigureWidget (WIX2, W.wix_offset + offset + MARGINS, W.sep_y - wixHeight3 - wixHeight2, C.width - 2 * (offset + MARGINS) - W.wix_offset, wixHeight2 - offset, 0);

	if (wixHeight1)
		XtConfigureWidget (WIX1, W.wix_offset + offset + MARGINS, W.sep_y - wixHeight1 - wixHeight2 - wixHeight3, C.width - 2 * (offset + MARGINS) - W.wix_offset, wixHeight1 - offset, 0);

	h = 2 * MARGINS + 3 * offset + buttonHeight + 2 + wixHeight1 + wixHeight2 + wixHeight3 + W.height;
	if (workHeight)
		XtConfigureWidget (WORK, W.wix_offset + offset + MARGINS, W.sep_y - offset/2 - (C.height - h) - wixHeight1 - wixHeight2 - wixHeight3, C.width - 2 * (offset + MARGINS) - W.wix_offset, C.height - h, 0);

	if (applyWidth)
		XtConfigureWidget (APPLY, (C.width - applyWidth)/2, C.height - apply_sha - applyHeight - offset,
							applyWidth, applyHeight, 0);

	if (cancelWidth)
		XtConfigureWidget (CAN, C.width - MARGINS - offset - okWidth, C.height - ok_sha - cancelHeight - offset,
							okWidth, cancelHeight, 0);

	if (okWidth) {
		if (applyWidth || cancelWidth)
			XtConfigureWidget (OK, offset + MARGINS, C.height - can_sha - okHeight - offset,
								okWidth, okHeight, 0);
		else
			XtConfigureWidget (OK, (C.width - okWidth)/2, C.height - can_sha - okHeight - offset,
								okWidth, okHeight, 0);
	}
}

static void Resize (Widget w)
{
	_resize ((QueryWidget) w, False);
}

static void preferred_size (QueryWidget w, Dimension* pwidth, Dimension* pheight)
{
	int			offset = W.offset;
	Dimension	workWidth =  (WORK && WORK -> core.managed) ? WORK -> core.width : 0;
	Dimension	workHeight =  (WORK && WORK -> core.managed) ? WORK -> core.height + offset : 0;
	Dimension	wixHeight1 = (WIX1 && WIX1 -> core.managed) ? WIX1 -> core.height + offset : 0;
	Dimension	wixHeight2 = (WIX2 && WIX2 -> core.managed) ? WIX2 -> core.height + offset : 0;
	Dimension	wixHeight3 = (WIX3 && WIX3 -> core.managed) ? WIX3 -> core.height + offset : 0;
	Dimension	wixWidth = wixHeight1 ? WIX1 -> core.width : 0;
	Dimension	okHeight = OK -> core.managed ? OK -> core.height : 0;
	Dimension	cancelHeight = CAN -> core.managed ? CAN -> core.height : 0;
	Dimension	applyHeight = APPLY -> core.managed ? APPLY -> core.height : 0;
	Dimension	okWidth = OK -> core.managed ? OK -> core.width : 0;
	Dimension	applyWidth = APPLY -> core.managed ? APPLY -> core.width : 0;
	Dimension	cancelWidth = CAN -> core.managed ? CAN -> core.width : 0;
	Dimension	buttonHeight = okHeight > cancelHeight ? okHeight : cancelHeight;
	Dimension the_width;

	if (okWidth < cancelWidth)
		okWidth = cancelWidth;

	the_width = sup (okWidth + applyWidth + cancelWidth + 3*offset,
					sup (W.title_width + W.p_width,
						sup (wixWidth, workWidth)))
				+ W.wix_offset + 2 * (offset + MARGINS);
	*pwidth = the_width;
	*pheight = workHeight + 2 * MARGINS + 4 * offset + buttonHeight + 2 + wixHeight1 + wixHeight2 + wixHeight3 + W.height;
}

static XtGeometryResult GeometryManager (Widget widget, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	XtWidgetGeometry	req, rep;
	QueryWidget			w = (QueryWidget) XtParent (widget);
	Dimension			swidth;
	Dimension			sheight;

	if (request -> request_mode & (CWX|CWY))
		return XtGeometryNo;

	if (request -> request_mode & (CWWidth|CWHeight)) {
		req.request_mode = request -> request_mode & (CWWidth|CWHeight);
		/* sauver la taille actuelle du widget qui veut changer de taille */
		swidth = widget -> core.width;
		sheight = widget -> core.height;
		/* tenter la taille requise */
		if (request -> request_mode & CWWidth)
			widget -> core.width = request -> width;
		if (request -> request_mode & CWHeight)
			widget -> core.height = request -> height;
		/* calculer la taille prfere */
		preferred_size (w, &req.width, &req.height);
		/* voir si a va  ou si on peut retailler la boite */
		if (XtGeometryYes == XtMakeGeometryRequest ((Widget) w, &req, &rep)
			|| (C.width >= req.width && C.height >= req.height))
		{
			reply -> request_mode = request -> request_mode;
			reply -> width = request -> width;
			reply -> height = request -> height;
			_resize (w, False);
			return XtGeometryYes;
		}
		/* rian  faire, remettre sa taille au widget */
		widget -> core.width = swidth;
		widget -> core.height = sheight;
	}
	return XtGeometryNo;
}

static void ChangeManaged (Widget w)
{
	_resize ((QueryWidget) w, True);
}

/*
 *	Les fonctions publiques
 */

Widget QueryGetChild (Widget widget, int n)
{
	QueryWidget	w = (QueryWidget) widget;

	switch (n) {
		case OK_BUTTON:		return OK;
		case CANCEL_BUTTON:	return CAN;
		case APPLY_BUTTON:	return APPLY;
		case EDIT:			return WIX1;
		case AUX_EDIT:		return WIX2;
		case AUX2_EDIT:		return WIX3;
	}
	return NULL;
}

void QuerySetWorkArea (Widget widget, Widget work) {
	QueryWidget	w = (QueryWidget) widget;

	WORK = work;
	_resize (w, True);
}
