/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "Container.h"
#include "ContainerP.h"
#include "Row.h"
#include "RowP.h"


#define	C	pw -> core
#define	P	pw -> composite
#define	M	pw -> container
#define	W	pw -> row

#define MARGINS (M.shadow_thickness+W.padding)

static XtGeometryResult
				GeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static void		ChangeManaged (Widget);

static XtResource resources [] = {
	{ XtNpadding, XtCPadding, XtRInt, sizeof (int),
	  XtOffset (RowWidget, row.padding), XtRImmediate, (XtPointer) 2
	},
	{ XtNtoResize, XtCToResize, XtRWidget, sizeof (Widget),
	  XtOffset (RowWidget, row.to_resize), XtRImmediate, (XtPointer) NULL
	},
	{ XtNorientation, XtCOrientation, XtRInt, sizeof (int),
	  XtOffset (RowWidget, row.orientation), XtRImmediate, (XtPointer) XwHORIZONTAL
	},
	{ XtNshadowThickness, XtCShadowThickness, XtRInt, sizeof (int),
	  XtOffset(RowWidget, container.shadow_thickness), XtRImmediate, (XtPointer) 0
	},
};

RowClassRec rowClassRec = {
  /* core_class fields */
  {
	/* superclass  */			(WidgetClass) &containerClassRec,
	/* class_name */			"Row",
	/* widget_size */			sizeof(RowRec),
	/* class_initialize */		NULL,
	/* class_partinit */			NULL,
	/* class_inited */			False,
	/* initialize */				Initialize,
	/* Init hook */				NULL,
	/* realize */				XtInheritRealize,
	/* actions */				NULL,
	/* num_actions */			0,
	/* resources */			resources,
	/* num_resources */		XtNumber (resources),
	/* xrm_class */			NULLQUARK,
	/* compress_motion */		False,
	/* compress_exposure */	True,
	/* compress_enterleave */	False,
	/* visible_interest */		False,
	/* destroy */				NULL,
	/* resize */				Resize,
	/* expose */				Redisplay,
	/* set_values */			NULL,
	/* set values hook */		NULL,
	/* set values almost */		XtInheritSetValuesAlmost,
	/* get values hook */		NULL,
	/* accept_focus */			NULL,
	/* Version */				XtVersion,
	/* PRIVATE cb list */		NULL,
	/* TM Table */			"",
	/* query_geom */			NULL,
	/* display acccelerator */	XtInheritDisplayAccelerator,
	/* extension */			NULL,
  },{
  /* composite_class fields */
	/* geometry_manager */	GeometryManager,
	/* change_managed */		ChangeManaged,
	/* insert_child */			XtInheritInsertChild,	/* Inherit from superclass */
	/* delete_child */			XtInheritDeleteChild,	/* Inherit from superclass */
	/* extension */			NULL,
  },{
  /* constraint_class fields */
	/*  constraint resource set */		NULL,
	/*  num_resources */			0,
	/*  size of the constraint data */	0,
	/*  contraint initilize proc */		(XtInitProc) NULL,
	/*  contraint destroy proc */		(XtWidgetProc) NULL,
	/*  contraint set values proc */	(XtSetValuesFunc) NULL,
  },{
  /* Container class fields */     
 	/* extension */			NULL
  },{
  /* row class fields */     
 	/* extension */			NULL
 }	
};

WidgetClass rowWidgetClass = (WidgetClass)&rowClassRec;

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	(*containerClassRec.core_class.expose) (widget, event, region);
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	RowWidget	pw = (RowWidget) new;

	if (request -> core.height == 0)	C.height = 10;
	if (request -> core.width == 0)		C.width = 10;
}

static void _resize (RowWidget pw, Boolean flag)
{
	Dimension			width = 0;
	Dimension			height = 0;
	int					i, x;
	Widget				w;
	XtWidgetGeometry	req, rep;
	/*
	 *	Calculer la taille optimale.
	 */
	if (W.orientation == XwHORIZONTAL) {
		if (flag) {
			for (i = 0; i < (int) P.num_children; i++) {
				w = P.children [i];
				if (!w -> core.managed)
					continue;

				if (height < w -> core.height)
					height = w -> core.height;	/* hauteur  maximale */
				width += w -> core.width + W.padding;
			}
			width += 2 * MARGINS - W.padding;
			height += 2 * M.shadow_thickness;
			req.width = width;
			req.height = height;
			req.request_mode = CWWidth|CWHeight;
			if (XtGeometryNo != XtMakeGeometryRequest ((Widget) pw, &req, &rep)) {
				C.width = width;
				C.height = height;
			}
		}
		x = MARGINS;
		for (i = 0; i < (int) P.num_children; i++) {
			w = P.children [i];
			if (!w -> core.managed)
				continue;
			/*
			 *	Retailler horizontalement celui choisi (tous sont retailles verticalement).
			 */
			if (w == W.to_resize) {
				Widget	ww;
				int		xx = x;
				int		j;

				for (j = i+1; j < (int) P.num_children; j++) {
					ww = P.children [j];
					if (!ww -> core.managed)
						continue;
					xx += ww -> core.width + W.padding;
				}
				XtConfigureWidget (w, x, M.shadow_thickness, C.width - MARGINS - xx, C.height - 2*M.shadow_thickness, 0);
				x += C.width - MARGINS - xx + W.padding;
			} else {
				XtConfigureWidget (w, x, M.shadow_thickness, w -> core.width, C.height - 2*M.shadow_thickness, 0);
				x += w -> core.width + W.padding;
			}
		}
	} else {
		if (flag) {
			for (i = 0; i < (int) P.num_children; i++) {
				w = P.children [i];
				if (!w -> core.managed)
					continue;

				if (width < w -> core.width)
					width = w -> core.width;	/* largeur  maximale */
				height += w -> core.height + W.padding;
			}
			height += 2 * MARGINS - W.padding;
			width += 2 * MARGINS;
			req.width = width;
			req.height = height;
			req.request_mode = CWWidth|CWHeight;
			if (XtGeometryNo != XtMakeGeometryRequest ((Widget) pw, &req, &rep)) {
				C.width = width;
				C.height = height;
			}
		}
		x = MARGINS;
		for (i = 0; i < (int) P.num_children; i++) {
			w = P.children [i];
			if (!w -> core.managed)
				continue;
			/*
			 *	Retailler verticalement celui choisi (tous sont retailles horizontalement).
			 */
			if (w == W.to_resize) {
				Widget	ww;
				int		xx = x;
				int		j;

				for (j = i+1; j < (int) P.num_children; j++) {
					ww = P.children [j];
					if (!ww -> core.managed)
						continue;
					xx += ww -> core.height + W.padding;
				}
				XtConfigureWidget (w, MARGINS, x, C.width - 2*MARGINS, C.height - MARGINS -xx, 0);
				x += C.height - MARGINS - xx + W.padding;
			} else {
				XtConfigureWidget (w, MARGINS, x, C.width - 2*MARGINS, w -> core.height, 0);
				x += w -> core.height + W.padding;
			}
		}
	}
}

static void Resize (Widget widget)
{
	_resize ((RowWidget) widget, False);
}

static void preferred_size (RowWidget pw, Dimension* pwidth, Dimension* pheight)
{
	Dimension	width = 0;
	Dimension	height = 0;
	int			i;
	Widget		w;

	if (W.orientation == XwHORIZONTAL) {
		for (i = 0; i < (int) P.num_children; i++) {
			w = P.children [i];
			if (!w -> core.managed)
				continue;

			if (height < w -> core.height)
				height = w -> core.height;	/* hauteur  maximale */
			width += w -> core.width + W.padding;
		}
		*pwidth = width + 2 * M.shadow_thickness - W.padding;
		*pheight = height + 2 * M.shadow_thickness;
	} else {
		for (i = 0; i < (int) P.num_children; i++) {
			w = P.children [i];
			if (!w -> core.managed)
				continue;

			if (width < w -> core.width)
				width = w -> core.width;	/* largeur  maximale */
			height += w -> core.height + W.padding;
		}
		*pheight = height + 2 * MARGINS - W.padding;
		*pwidth = width + 2 * MARGINS;
	}
}

static XtGeometryResult GeometryManager (Widget w, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	XtWidgetGeometry	req, rep;
	RowWidget			pw = (RowWidget) XtParent (w);
	Dimension			swidth;
	Dimension			sheight;

	if (request -> request_mode & (CWX|CWY))
		return XtGeometryNo;	/* pas de changement de position */

	if (request -> request_mode & (CWWidth|CWHeight)) {
		req.request_mode = request -> request_mode & (CWWidth|CWHeight);
		/* sauver la taille actuelle du widget qui veut changer de taille */
		swidth = w -> core.width;
		sheight = w -> core.height;
		/* tenter la taille requise */
		if (request -> request_mode & CWWidth)
			w -> core.width = request -> width;
		if (request -> request_mode & CWHeight)
			w -> core.height = request -> height;
		/* calculer la taille prfere du Row */
		preferred_size (pw, &req.width, &req.height);
		/* voir si a va  ou si on peut retailler le Row */
		if (XtGeometryYes == XtMakeGeometryRequest ((Widget) pw, &req, &rep)
			|| (C.width >= req.width && C.height >= req.height))
		{
			reply -> request_mode = request -> request_mode;
			reply -> width = request -> width;
			reply -> height = request -> height;
			_resize (pw, False);
			return XtGeometryYes;
		}
		/* rian  faire, remettre sa taille au widget */
		w -> core.width = swidth;
		w -> core.height = sheight;
	}
	return XtGeometryNo;
}

static void ChangeManaged (Widget widget)
{
	_resize ((RowWidget) widget, True);
}
