/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "RulerP.h"
#include "Ruler.h"
  
#define W	w -> ruler
#define P	w -> primitive
#define C	w -> core

#define MARGINS	(P.shadow_thickness + W.margin)

static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Destroy (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

static void		Move (Widget, XEvent*, String*, Cardinal*);
static void		Create (Widget, XEvent*, String*, Cardinal*);
static void		Remove (Widget, XEvent*, String*, Cardinal*);

static int	theOne = -1;	/* le taquet qui est en train de bouger */

static char defaultTranslations [] = "\
	<Btn1Down>:		Move()\n\
	<Btn1Motion>:	Move()\n\
	<Btn1Up>:		Move()\n\
	<Btn2Down>:		Create()\n\
	<Btn2Motion>:	Move()\n\
	<Btn2Up>:		Move()\n\
	<Btn3Down>:		Remove()\n\
	<Btn3Up>:		Remove()";

static XtActionsRec actionsList [] =
{
	{"Move",	Move},
	{"Create",	Create},
	{"Remove",	Remove},
};

static XtResource resources [] =
{     
	{ XtNleftOffset, XtCMargin, XtRShort, sizeof (short),
	  XtOffset (RulerWidget, ruler.left_offset), XtRImmediate, (XtPointer) 0
	},
	{ XtNmargin, XtCMargin, XtRShort, sizeof (short),
	  XtOffset (RulerWidget, ruler.margin), XtRImmediate, (XtPointer) 2
	},
	{ XtNscale, XtCScale, XtRShort, sizeof (short),
	  XtOffset (RulerWidget, ruler.scale), XtRImmediate, (XtPointer) 10
	},
	{ XtNhotColor, XtCHotColor, XtRPixel, sizeof (Pixel),
	  XtOffset (RulerWidget, ruler.hot_color), XtRString, "black"
	},
	{ XtNstopSize, XtCStopSize, XtRShort, sizeof (short),
	  XtOffset (RulerWidget, ruler.stop_size), XtRImmediate, (XtPointer) 6
	},
};

RulerClassRec rulerClassRec = {
    {
	(WidgetClass) &primitiveClassRec,	/* superclass */
	"Ruler",			/* class_name */
	sizeof (RulerRec),		/* widget_size */
	NULL,				/* class_initialize */
	NULL,				/* class_part_init */
	FALSE,				/* class_inited */
	Initialize,			/* initialize */
	NULL,				/* initialize_hook */
	XtInheritRealize,			/* realize */
	actionsList,			/* actions */
	XtNumber (actionsList),		/* num_actions */
	resources,			/* resources */
	XtNumber (resources),		/* num_resources */
	NULLQUARK,			/* xrm_class */
	False,				/* compress_motion */
	True,				/* compress_exposure */
	True,				/* compress_enterlv */
	False,				/* visible_interest */
	Destroy,			/* destroy */
	NULL,				/* resize */
	Redisplay,			/* expose */
	SetValues,			/* set_values */
	NULL,				/* set_values_hook */
	XtInheritSetValuesAlmost,	/* set_values_almost */
	NULL,				/* get_values_hook */
	NULL,				/* accept_focus */
	XtVersion,			/* version */
	NULL,				/* callback_private */
	defaultTranslations,		/* tm_table */
	NULL,				/* query_geometry */
    },
    {
	NULL,
    },
    {
	NULL,
    },
};

WidgetClass rulerWidgetClass = (WidgetClass)&rulerClassRec;

/*
 *	le GC pour le ruler.
 */

static void get_rulerGC (RulerWidget w)
{
	XGCValues  values;

	values.foreground = P.foreground;
	W.rulerGC = XtGetGC ((Widget) w, GCForeground, &values);
}

/*
 *	Fonction d'initialisation de chaque instance.
 */

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	RulerWidget	w = (RulerWidget) new;

	if (request -> core.width == 0)		C.width = 800;
	if (request -> core.height == 0)	C.height = 2*MARGINS + 2*W.stop_size + 15;
	get_rulerGC (w);
	W.tabStops [0] = -1;
	W.scroll = 0;
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* e, Region region)
{
	RulerWidget		w = (RulerWidget) widget;
	register int	i;
	register int	s = W.stop_size;
	Display*		dpy = XtDisplay (widget);
	Window			window = XtWindow (widget); 
	XPoint			triangle [3];

	XSetRegion (dpy, W.rulerGC, region);
	if (P.shadow_thickness > 0) {
		XSetRegion (dpy, P.top_GC, region);
		XSetRegion (dpy, P.bot_GC, region);
		_DrawShadow (dpy, window,
						P.top_GC, P.bot_GC, P.shadow_thickness,
						C.width, C.height, 0, 0);
		XSetClipMask (dpy, P.top_GC, None);
		XSetClipMask (dpy, P.bot_GC, None);
	}
	/*
	 *	Dessiner la graduation.
	 */
	for (i = - W.scroll; i < (int) (C.width - 2*MARGINS - s - W.left_offset); i += W.scale) {
		if (i < 0)
			continue;
		XFillRectangle (dpy, window, W.rulerGC,
							MARGINS + s + W.left_offset + i, MARGINS,
							2, 6 + (((i+W.scroll)%(5*W.scale))==0 ? 4 : 0));
	}
	XFillRectangle (dpy, window, W.rulerGC,
						MARGINS + s + W.left_offset, MARGINS,
						i + 2 - W.scale, 2);
	/*
	 *	Dessiner les taquets.
	 */
	triangle [0].y = C.height - MARGINS;
	triangle [1].x = 2*s;
	triangle [1].y = 0;
	triangle [2].x = - s;
	triangle [2].y = - 2*s + 1;
	for (i = 0; W.tabStops [i] >= 0; i++) {
		triangle [0].x = W.tabStops [i] + MARGINS + W.left_offset - W.scroll;
		if (W.tabStops [i] < W.scroll)
			continue;
		if (triangle [0].x > C.width)
			break;	/* inutile de patiner dans la semoule */
		if (i == theOne)
			XSetForeground (dpy, W.rulerGC, W.hot_color);
		XFillPolygon (dpy, window, W.rulerGC,
						triangle, XtNumber (triangle),
						Convex, CoordModePrevious);
		XDrawLine (dpy, window, W.rulerGC,
						triangle [0].x + s, triangle [0].y - 2*s + 1,
						triangle [0].x + s, triangle [0].y - 2*s - 4);
		if (i == theOne)
			XSetForeground (dpy, W.rulerGC, P.foreground);
	}
	XSetClipMask (dpy, W.rulerGC, None);
}

/*
 *	SetValues.
 */

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	RulerWidget	cu = (RulerWidget) current;
	RulerWidget	w = (RulerWidget) new;
	Boolean 	do_redisplay = False;

	if (cu -> primitive.foreground != P.foreground) {
		XtDestroyGC (cu -> ruler.rulerGC);
		get_rulerGC (w);
		do_redisplay = True;
	}

	if (cu -> ruler.margin != W.margin)
		do_redisplay = True;

	if (cu -> ruler.scale != W.scale)
		do_redisplay = True;

	if (cu -> ruler.left_offset != W.left_offset)
		do_redisplay = True;

	if (cu -> ruler.stop_size != W.stop_size)
		do_redisplay = True;

	return do_redisplay;
}

/*
 *	La procedure de destruction d'une instance.
 */

static void Destroy (Widget widget)
{
	RulerWidget	w = (RulerWidget) widget;

	XtDestroyGC (W.rulerGC);
	XFreePixmap (XtDisplay (widget), W.pixmap);
}

/*
 *	Reaffiche un taquet.
 */

static void drawTriangle (RulerWidget w, int x)
{
	int	s = W.stop_size;

	XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
				x + MARGINS + W.left_offset - W.scroll,
				C.height - MARGINS - 2*s - 5,
				2*s, 2*s + 5, True);
}

/*
 *	Indique dans quel taquet on a tape.
 */

static int insideTab (RulerWidget w, XButtonEvent* e)
{
	register int	i;
	register int	s = W.stop_size;
	register int	x = e -> x - MARGINS - s - W.left_offset + W.scroll;

	if (e -> y < C.height - MARGINS - s - 5)
		return -1;		/* trop haut */
	if (e -> y > C.height - MARGINS)
		return -1;		/* trop bas */
	for (i = 0; i < 100 && W.tabStops [i] >= 0; i++)
		if (W.tabStops [i] - s < x && W.tabStops [i] + s > x)
			return i;
	return -1;
}

/*
 *	Deplacement d'un taquet.
 */


static void Move (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	RulerWidget		w = (RulerWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				i;
	int				s = W.stop_size;
	int				x = e -> x - MARGINS - s - W.left_offset + W.scroll;

	switch (e -> type) {
	case ButtonPress:
		if (theOne != -1)
			return;		/* bizarre */
		i = insideTab (w, e);
		if (i >= 0) {
			theOne = i;
			drawTriangle (w, W.tabStops [i]);
		}
		return;

	case MotionNotify:
		if (theOne < 0)
			return;
		if (x <= 0)
			return;
		if (theOne >= 1 && x < W.tabStops [theOne-1] + 5)
			return;
		if (W.tabStops [theOne+1] > 0 && x > W.tabStops [theOne+1] - 5)
			return;

		x = (x / W.scale) * W.scale;
		if (x == W.tabStops [theOne])
			return;

		drawTriangle (w, W.tabStops [theOne]);
		W.tabStops [theOne] = x;
		drawTriangle (w, x);
		return;

	case ButtonRelease:
		if (theOne != -1) {
			drawTriangle (w, W.tabStops [theOne]);
			XtCallCallbacks (widget, XtNselect, (XtPointer) W.tabStops);
			theOne = -1;
		}
		return;
	}
}

/*
 *	Creation d'un nouveau taquet.
 */

static void Create (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	RulerWidget		w = (RulerWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				i;
	int				s = W.stop_size;
	int				x = e -> x - MARGINS - s - W.left_offset + W.scroll;

	switch (e -> type) {
	case ButtonPress:
		if (theOne != -1)
			return;		/* bizarre */
		i = insideTab (w, e);
		if (i >= 0)
			return;
		x = (x / W.scale) * W.scale;
		for (i = 0;; i++) {
			if (i >= (int) (XtNumber (W.tabStops) - 2))
				return;
			if (W.tabStops [i] < 0) {
				theOne = i;
				W.tabStops [i] = x;
				W.tabStops [i+1] = -1;
				break;
			}
			if (W.tabStops [i] > x) {
				theOne = i;
				while (W.tabStops [i] >= 0) i++;
				for (i++; i > theOne; --i)
					W.tabStops [i] = W.tabStops [i-1];
				W.tabStops [i] = x;
				break;
			}
		}
		drawTriangle (w, W.tabStops [i]);
	}
	return;
}

/*
 *	Destruction d'un taquet.
 */

static void Remove (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	RulerWidget		w = (RulerWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				i, x;
	static Boolean	pressed = False;

	if (theOne != -1)
		return;

	switch (e -> type) {
	case ButtonPress:
		if ((i = insideTab (w, e)) < 0)
			return;

		for (x = W.tabStops [i]; W.tabStops [i] >= 0; i++)
			W.tabStops [i] = W.tabStops [i+1];
		drawTriangle (w, x);
		pressed = True;
		return;

	case ButtonRelease:
		if (pressed) {
			XtCallCallbacks (widget, XtNselect, (XtPointer) W.tabStops);
			pressed = False;
		}
		theOne = -1;
		return;
	}
}

/*
 *	Les Fonctions Publiques.
 */

void RulerScroll (Widget widget, int n)
{
	RulerWidget	w = (RulerWidget) widget;
	int			how;
	int			delta;
	int			margins;
	int			s = W.stop_size;
	Window		window = XtWindow (widget);
	Display*	dpy = XtDisplay (widget);
	XEvent		e;

	delta = MARGINS + s + W.left_offset;
	margins = MARGINS;
	how = n - W.scroll;
	W.scroll = n;
	if (how == 0 || !XtIsRealized (widget) || ! C.managed)
		return;
	if (how > 0) {
		XCopyArea (dpy, window, window, W.rulerGC, 
					delta + how, margins,
					C.width - delta - margins - how, C.height - 2 * margins,
					delta, margins);
		XClearArea (dpy, window,
					C.width - margins - how - 20, margins,
					how + 20 + margins, C.height - 2 * margins, True);
	} else {
		how = -how;
		XCopyArea (dpy, window, window, W.rulerGC,
					delta, margins,
					C.width - delta - margins - how, C.height - 2 * margins,
					delta + how, margins);
		XClearArea (dpy, window,
					delta, margins,
					how + 20, C.height - 2 * margins, True);
	}
	XSync (dpy, False);
	while (XCheckTypedEvent (dpy, Expose, &e))
		XtDispatchEvent (&e);
	while (XCheckTypedEvent (dpy, GraphicsExpose, &e)) {
		e.type = Expose;
		XtDispatchEvent (&e);
	}
}

void RulerSetTabstops (Widget widget, int* t, int n)
{
	RulerWidget	w = (RulerWidget) widget;
	int			i;

	for (i = 0; i < 100-1 && i < n && t [i] >= 0; i++)
		W.tabStops [i] = t [i];
	W.tabStops [i-1] = -1;
	if (XtIsRealized (widget) && C.managed)
		XClearArea (XtDisplay (widget), XtWindow (widget), 0, 0, 0, 0, True);
}
