/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "Primitive.h"
#include "Container.h"
#include "ContainerP.h"
#include "Scrollbar.h"
#include "ScrolledWin.h"
#include "ScrolledWinP.h"


#define	C	sw -> core
#define	M	sw -> manager
#define	W	sw -> swindow
#define	VSB	sw -> swindow.vScrollBar
#define	HSB	sw -> swindow.hScrollBar

#define DEFAULT_HEIGHT 300
#define DEFAULT_WIDTH 300

/*
 *	Dclaration des mthodes.
 */

static XtGeometryResult
				GeometryManager (Widget, XtWidgetGeometry*, XtWidgetGeometry*);
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static void		Resize (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static void		ChangeManaged (Widget);
static void		InsertChild (Widget);

static void		_resize (ScrolledWindowWidget, Boolean);

/*
 *	Les actions.
 */

static void	SelectAction (Widget, XEvent*, String*, Cardinal*);

/*
 *	Les ressources.
 */

static XtResource resources [] = {
	{ XtNselect, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (ScrolledWindowWidget, swindow.select), XtRPointer, NULL
	},
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct*),
	  XtOffset (ScrolledWindowWidget, swindow.fp), XtRString, "fixed"
	},
	{ XtNlabelColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset (ScrolledWindowWidget, swindow.label_color), XtRString, "black"
	},
	{ XtNtitle, XtCTitle, XtRString, sizeof (char*),
	  XtOffset (ScrolledWindowWidget, swindow.title), XtRImmediate, NULL
	},
	{ XtNlabel, XtCLabel, XtRString, sizeof (char*),
	  XtOffset (ScrolledWindowWidget, swindow.label), XtRImmediate, NULL
	},
	{ XtNpixmap, XtCPixmap, XtRPixmap, sizeof (Pixmap),
	  XtOffset (ScrolledWindowWidget, swindow.pixmap), XtRPointer, NULL
	},
	{ XtNhScrollBar, XtCHScrollBar, XtRWidget, sizeof (Widget),
	  XtOffset (ScrolledWindowWidget, swindow.hScrollBar), XtRImmediate, NULL
	},
	{ XtNvScrollBar, XtCVScrollBar, XtRWidget, sizeof (Widget),
	  XtOffset (ScrolledWindowWidget, swindow.vScrollBar), XtRImmediate, NULL
	},
	{ XtNoffset, XtCOffset, XtRInt, sizeof (int),
	  XtOffset (ScrolledWindowWidget, swindow.offset), XtRImmediate, (XtPointer) 1
	},
	{ XtNwideScrollBar, XtCWideScrollBar, XtRBoolean, sizeof (Boolean),
	  XtOffset (ScrolledWindowWidget, swindow.wide_scrollbar), XtRImmediate, (XtPointer) False
	},
};

static char			defaultTranslations [] = "<Btn1Up>:	Select()";
static XtActionsRec	actionsList [] = { {"Select", SelectAction}, };

ScrolledWindowClassRec scrolledWindowClassRec = {
  /* core_class fields */
  {
    /* superclass         */    (WidgetClass) &containerClassRec,
    /* class_name         */    "ScrolledWindow",
    /* widget_size        */    sizeof(ScrolledWindowRec),
    /* class_initialize   */	NULL,
    /* class_partinit     */    NULL,
    /* class_inited       */	FALSE,
    /* initialize         */    Initialize,
    /* Init hook	  */    NULL,
    /* realize            */    Realize,
    /* actions		  */	actionsList,
    /* num_actions	  */	XtNumber(actionsList),
    /* resources          */    resources,
    /* num_resources      */    XtNumber (resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion	  */	TRUE,
    /* compress_exposure  */	TRUE,
    /* compress_enterleave*/	TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    Resize,
    /* expose             */    Redisplay,
    /* set_values         */    SetValues,
    /* set values hook    */    NULL,
    /* set values almost  */    XtInheritSetValuesAlmost,
    /* get values hook    */    NULL,
    /* accept_focus       */    NULL,
    /* Version            */    XtVersion,
    /* PRIVATE cb list    */    NULL,
    /* TM Table           */    defaultTranslations,
    /* query_geom         */    NULL,
  },{
/* composite_class fields */
    /* geometry_manager   */    GeometryManager,
    /* change_managed     */    ChangeManaged,
    /* insert_child	  */	InsertChild,	/* Basically Inherited */
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension */		NULL,
  },{
/* Constraint class Init */
    NULL,
    0,
    0,
    NULL,
    NULL,
    NULL,
    NULL
      
  },{
/* manager_class fields */
   /* extension   */    NULL
  },{
/* ScrolledWindow class */     
		/* extension */	NULL
 }	
};

WidgetClass scrolledWindowWidgetClass = (WidgetClass) &scrolledWindowClassRec;

/*
 *	Cration de la fentre.
 */

static void Realize	(Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	*value_mask |= CWBitGravity;
	attributes -> bit_gravity = SouthEastGravity;	/* pour le label en bas  droite */
	XtCreateWindow (widget, InputOutput, CopyFromParent, *value_mask, attributes);
}

/*
 *	le GC pour le label.
 */

static void get_labelGC (ScrolledWindowWidget sw)
{
	XGCValues		values;
	unsigned long 	valuemask;

	valuemask = GCForeground | GCBackground | GCFont;
	values.foreground = M.foreground;
	values.background = C.background_pixel;
	values.font = W.fp -> fid;

	W.label_GC = XtGetGC ((Widget) sw, valuemask, &values);
}

/*
 *	Recupere le pixmap.
 */

static void get_pixmap (ScrolledWindowWidget sw)
{
	Window	wraf;
	int		raf;

	if (W.pixmap != 0)
		XGetGeometry (XtDisplay ((Widget) sw), W.pixmap, &wraf, &raf, &raf,
						&W.p_width, &W.p_height, &raf, &W.p_depth);
	else {
		W.p_width = 0;
		W.p_height = 0;
		W.p_depth = 0;
	}
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) new;

	VSB = XtVaCreateWidget ("vsb", scrollbarWidgetClass, new,
								XtNbackground,	C.background_pixel,
								XtNforeground,	M.foreground,
								NULL);
	HSB = XtVaCreateWidget ("vsb", scrollbarWidgetClass, new,
								XtNbackground,	C.background_pixel,
								XtNforeground,	M.foreground,
								XtNslideOrientation, XpaHORIZONTAL,
								NULL);
	W.bboard = NULL;

	if (request -> core.height == 0)	C.height = HSB -> core.height;
	if (request -> core.width == 0)		C.width = VSB -> core.width;

	get_pixmap (sw);
	get_labelGC (sw);
}

static void InsertChild (Widget	w)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) XtParent (w);
	ContainerWidgetClass	superclass = (ContainerWidgetClass) containerWidgetClass;
    
	if (W.hScrollBar == NULL || W.vScrollBar == NULL ||
	    (w == (Widget) W.hScrollBar) || (w == (Widget) W.vScrollBar))
	{
		(*superclass->composite_class.insert_child)(w);
		XtManageChild (w);
		return;
	}

	if (W.bboard == NULL) {
		W.bboard = w;
		(*superclass->composite_class.insert_child) (w);
	} else
		XtWarning ("ScrolledWindow: No more than one widget in window\n");
}

static void _resize (ScrolledWindowWidget sw, Boolean flag)
{
	Dimension vwidth = VSB -> core.managed ? VSB -> core.width : 0;
	Dimension hheight = HSB -> core.managed ? HSB -> core.height : 0;
	Dimension bheight = W.bboard && W.bboard -> core.managed ? W.bboard -> core.height : 0;
	Dimension bwidth = bheight ? W.bboard -> core.width : 0;
	Dimension base_height = 0;

	if (W.title || W.label)
		base_height = W.fp -> ascent + W.fp -> descent;
	if (HSB && HSB -> core.managed)
		if (base_height < hheight)
			base_height = hheight;

	if (W.bboard && W.bboard -> core.managed) {
		Dimension h, ww;

		if (flag) {
			if (C.height < bheight + base_height + W.offset + 2 * M.shadow_thickness)
				C.height = bheight + base_height + W.offset + 2 * M.shadow_thickness;
			if (C.width < bwidth + vwidth + W.offset + 2 * M.shadow_thickness)
				C.width = bwidth + vwidth + W.offset + 2 * M.shadow_thickness;
		}

		h = C.height - base_height - W.offset - 2 * M.shadow_thickness;
		if (h < 4)
			h = 4;

		ww = C.width - vwidth - W.offset - 2 * M.shadow_thickness;
		if (ww < 6)
			ww = 6;
		XtConfigureWidget (W.bboard,
			vwidth + W.offset + M.shadow_thickness,
			M.shadow_thickness, ww, h, 0);
	}
	if (VSB && VSB -> core.managed) {
		Dimension h = C.height - base_height - W.offset;

		if (h < 4)
			h = 4;
		XtConfigureWidget (VSB,
			0, 0, vwidth, h, 0);
	}
	if (HSB && HSB -> core.managed) {
		Dimension ww = (W.wide_scrollbar
							? C.width - vwidth - W.offset
							: C.width / 2 - vwidth - W.offset);

		if (ww < 6)
			ww = 6;
		XtConfigureWidget (HSB,
			vwidth + W.offset, C.height - base_height, ww, hheight, 0);
	}
}

static void Resize (Widget widget)
{
	_resize ((ScrolledWindowWidget) widget, False);
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) widget;
	Display*				dpy = XtDisplay (widget);
	Window					win = XtWindow (widget);
	int						vwidth = VSB -> core.managed ? VSB -> core.width : 0;
	int						hheight = HSB -> core.managed ? HSB -> core.height : 0;
	int						hwidth = HSB -> core.managed ? HSB -> core.width : 0;
	int						base_height = 0;
	int						tl = 0;

	XSetRegion (dpy, W.label_GC, region);

	if (W.title || W.label)
		base_height = W.fp -> ascent + W.fp -> descent;
	if (HSB && HSB -> core.managed)
		if (base_height < hheight)
			base_height = hheight;

	if (W.pixmap)
		if (W.p_depth == 1)
			XCopyPlane (dpy, W.pixmap, win, W.label_GC,
							0, 0,
							vwidth < W.p_width ? vwidth : W.p_width,
							base_height < W.p_height ? base_height : W.p_height,
							0, C.height - base_height, 1);
		else
			XCopyArea (dpy, W.pixmap, win, W.label_GC,
							0, 0,
#if 1
							W.p_width,
#else
							vwidth < W.p_width ? vwidth : W.p_width,
#endif
							base_height < W.p_height ? base_height : W.p_height,
							0, C.height - base_height);
	if (W.title) {
		int	l = strlen (W.title);

		tl = XTextWidth (W.fp, W.title, l);
		XDrawString (dpy, win, W.label_GC,
						C.width - W.offset - tl,
						C.height - W.offset/2 - base_height + W.fp -> ascent,
						W.title, l);
	}
	if (W.label) {
		int	l = strlen (W.label);
		int	pl = XTextWidth (W.fp, W.label, l);

		XSetForeground (dpy, W.label_GC, W.label_color);
		XDrawString (dpy, win, W.label_GC,
						(C.width - hwidth - tl - pl) / 2 + hwidth,
						C.height - W.offset/2 - base_height + W.fp -> ascent,
						W.label, l);
		XSetForeground (dpy, W.label_GC, M.foreground);
	}
	_DrawShadow (dpy, win,
					M.bot_GC, M.top_GC, M.shadow_thickness,
					C.width - vwidth - W.offset,
					C.height - base_height - W.offset,
					vwidth + W.offset, 0);

	XSetClipMask (dpy, W.label_GC, None);
}

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	ScrolledWindowWidget	cu = (ScrolledWindowWidget) current;
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) new;
	Boolean					do_redisplay = False;

	if (cu -> swindow.pixmap != W.pixmap) {
		get_pixmap (sw);
		do_redisplay = True;
	}

	if (cu -> swindow.title != W.title)
		do_redisplay = True;

	if (cu -> swindow.label != W.label)
		do_redisplay = True;

	return do_redisplay;
}

static void preferred_size (Widget widget, Dimension bwidth, Dimension bheight,
							Dimension* width, Dimension* height)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) widget;
	Dimension				vwidth = VSB -> core.managed ? VSB -> core.width : 0;
	Dimension				hheight = HSB -> core.managed ? HSB -> core.height : 0;
	Dimension				base_height = W.fp -> ascent + W.fp -> descent;

	if (!hheight && (W.title || W.label))	hheight = VSB -> core.width;
	if (base_height < hheight)		base_height = hheight;

	*height = bheight + base_height + W.offset + 2 * M.shadow_thickness;
	*width = bwidth + vwidth + W.offset + 2 * M.shadow_thickness;
}

static XtGeometryResult GeometryManager (Widget w, XtWidgetGeometry* request, XtWidgetGeometry* reply)
{
	XtWidgetGeometry	req, rep;

	if ((request -> request_mode & CWX) || (request -> request_mode & CWY))
		return XtGeometryNo;

	if (request -> request_mode & (CWWidth|CWHeight)) {
		req.request_mode = request -> request_mode & (CWWidth|CWHeight);
		preferred_size (XtParent (w),
						request -> width, request -> height,
						&req.width, &req.height);
		if (XtGeometryYes == XtMakeGeometryRequest (XtParent (w), &req, &rep)) {
			reply -> request_mode = request -> request_mode;
			reply -> width = request -> width;
			reply -> height = request -> height;
			_resize ((ScrolledWindowWidget) XtParent (w), False);
			return XtGeometryYes;
		}
	}
	return XtGeometryNo;
}

static void ChangeManaged (Widget widget)
{
	_resize ((ScrolledWindowWidget) widget, True);
}

/*
 *	Les actions.
 */

static void SelectAction (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) widget;
	XButtonEvent*			e = (XButtonEvent*) event;
	int						vwidth = VSB -> core.managed ? VSB -> core.width : 0;
	int 					hheight = HSB -> core.managed ? HSB -> core.height : 0;

	if (e -> x < vwidth && e -> y > C.height - hheight && e -> y < C.height)
		XtCallCallbacks (widget, XtNselect, NULL);
}

/*
 *	Fonctions publiques.
 */

Widget SwHSB (Widget w)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) w;

	return HSB;
}

Widget SwVSB (Widget w)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) w;

	return VSB;
}

void SwSetTitle (Widget w, String title)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) w;

	int	hheight = HSB -> core.managed ? HSB -> core.height : 0;
	int	hwidth = HSB -> core.managed ? HSB -> core.width : 0;
	int	base_height = W.fp -> ascent + W.fp -> descent;

	W.title = title;
	if (base_height < hheight)
		base_height = hheight;

	if (!XtIsRealized (w))
		return;

	XClearArea (XtDisplay (w), XtWindow (w),
		W.offset + hwidth, C.height - W.offset/2 - base_height,
		C.width, W.fp -> ascent + W.fp -> descent,
		True);

}

void SwSetLabel (Widget w, String label)
{
	ScrolledWindowWidget	sw = (ScrolledWindowWidget) w;

	Display*	dpy = XtDisplay (w);
	Window		win = XtWindow (w);
	int		hheight = HSB -> core.managed ? HSB -> core.height : 0;
	int		hwidth = HSB -> core.managed ? HSB -> core.width : 0;
	int		base_height = W.fp -> ascent + W.fp -> descent;
	int		tl = 0;

	if (!hheight && (W.title || W.label))	hheight = VSB -> core.width;
	if (base_height < hheight)		base_height = hheight;

	W.label = label;

	if (!XtIsRealized (w))
		return;

	if (W.title)
		tl = XTextWidth (W.fp, W.title, strlen (W.title));

	XClearArea (dpy, win, 
		W.offset + hheight, C.height - W.offset/2 - base_height,
		C.width - tl - hheight - 2 * W.offset, W.fp -> ascent + W.fp -> descent, False);

	if (label) {
		int	l = strlen (label);
		int	pl = XTextWidth (W.fp, label, l);

		XSetForeground (dpy, W.label_GC, W.label_color);
		XDrawString (dpy, win, W.label_GC,
			(C.width - hwidth - tl - pl) / 2 + hwidth,
			C.height - W.offset/2 - base_height + W.fp -> ascent,
			W.label, l);
		XSetForeground (dpy, W.label_GC, M.foreground);
	}
}
