/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "TextListP.h"
#include "TextList.h"
#include "Valuator.h"
#include "ScrolledWin.h"

#define W	w -> textList
#define P	w -> primitive
#define C	w -> core

/*
 *	Les ressources de la barre de textList.
 */

static XtResource resources [] =
{
	{ XtNvScrollSizeCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (TextListWidget, textList.vscroll_size), XtRPointer, (XtPointer) NULL
	},
	{ XtNvScrollSize, XtCScroll, XtRInt, sizeof (int),
	  XtOffset (TextListWidget, textList.sb_size), XtRImmediate, 0
	},
	{ XtNvScrollPositionCallback, XtCCallback, XtRCallback, sizeof (XtCallbackList),
	  XtOffset (TextListWidget, textList.vscroll_pos), XtRPointer, (XtPointer) NULL
	},
	{ XtNvScrollPosition, XtCScroll, XtRInt, sizeof (int),
	  XtOffset (TextListWidget, textList.sb_pos), XtRImmediate, 0
	},
	{ XtNfirstItem, XtCFirstItem, XtRInt, sizeof (int),
	  XtOffset (TextListWidget, textList.first_item), XtRImmediate, 0
	},
	{ XtNsingleSelection, XtCSingleSelection, XtRBoolean, sizeof (Boolean),
	  XtOffset(TextListWidget, textList.single_selection), XtRImmediate, (XtPointer) True
	},
	{ XtNlineNum, XtCLineNum, XtRInt, sizeof(int),
	  XtOffset(TextListWidget, textList.num_lines), XtRImmediate, (XtPointer) 5
	},
	{ XtNoffset, XtCOffset, XtRInt, sizeof(int),
	  XtOffset(TextListWidget, textList.space), XtRImmediate, (XtPointer) 2
	},
	{ XtNselection, XtCSelection, XtRString, sizeof (String),
	  XtOffset(TextListWidget, textList.selection), XtRString, NULL
	},
	{ XtNitemList, XtCItemList, XtRStringArray, sizeof (String*),
	  XtOffset(TextListWidget, textList.items), XtRImmediate, NULL
	},
	{ XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
	  XtOffset(TextListWidget, textList.font), XtRString, "Fixed"
	},
	{ XtNselectionStyle, XtCSelectionStyle, XtRInt, sizeof (int),
	  XtOffset(TextListWidget, textList.selection_style), XtRImmediate, 0
	},
	{ XtNselectionColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset(TextListWidget, textList.selection_color), XtRString, "red"
	},
	{ XtNselectionNumber, XtCSelectionNumber, XtRInt, sizeof (int),
	  XtOffset(TextListWidget, textList.selection_number), XtRImmediate, (XtPointer) -1
	},
};

static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		Destroy (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

/*
 *	Procdures d'action.
 */

static void	Arm (Widget, XEvent*, String*, Cardinal*);
static void	Activate (Widget, XEvent*, String*, Cardinal*);
static void	SelectItem (Widget, XEvent*, String*, Cardinal*);
static void	SelectPrevious (Widget, XEvent*, String*, Cardinal*);
static void	SelectNext (Widget, XEvent*, String*, Cardinal*);
static void	Release (Widget, XEvent*, String*, Cardinal*);
static void	SearchItem (Widget, XEvent*, String*, Cardinal*);
static void	PreviousPage (Widget, XEvent*, String*, Cardinal*);
static void	NextPage (Widget, XEvent*, String*, Cardinal*);
static void	ScrollDown (Widget, XEvent*, String*, Cardinal*);
static void	ScrollUp (Widget, XEvent*, String*, Cardinal*);

static XtActionsRec actionsList [] =
{
	{"Arm",			Arm },
	{"Activate",		Activate },
	{"release",		Release },
	{"search-item",	SearchItem },
	{"select-item",	SelectItem },
	{"next-item",	SelectNext },
	{"previous-item",	SelectPrevious },
	{"next-page",	NextPage },
	{"previous-page",	PreviousPage },
	{"scroll-down",	ScrollDown },
	{"scroll-up",		ScrollUp },
};

#if 1
static char defaultTranslations [] = "\
	<Btn1Down>:			Arm()\n\
	<Btn1Up>(2):			Activate()\n\
	<Key>Return:			select-item()\n\
	<Key>Up:			previous-item()\n\
	<Key>Down:			next-item()\n\
	Ctrl<Key>Prior:		scroll-up()\n\
	Ctrl<Key>Next:		scroll-down()\n\
	<Key>Prior:			previous-page()\n\
	<Key>Next:			next-page()\n\
	<Key>:				search-item()";
#else
static char defaultTranslations [] = "\
	<Btn1Down>:			Arm()\n\
	Button1<PtrMoved>:	Arm()\n\
	<Btn1Up>:			Activate()\n\
	<Btn2Down>:			release()\n\
	Button2<PtrMoved>:	release()\n\
	<Key>Return:			select-item()\n\
	<Key>Up:			previous-item()\n\
	<Key>Down:			next-item()\n\
	Ctrl<Key>Prior:		scroll-up()\n\
	Ctrl<Key>Next:		scroll-down()\n\
	<Key>Prior:			previous-page()\n\
	<Key>Next:			next-page()\n\
	<Key>:				search-item()";
#endif

TextListClassRec textListClassRec = {
	/* CoreClassPart */
	{
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"TextList",							/* class_name */
		sizeof (TextListRec),				/* widget_size */
		NULL,								/* class_initialize */
		NULL,								/* class_part_init */
		FALSE,								/* class_inited */
		Initialize,							/* initialize */
		NULL,								/* initialize_hook */
		XtInheritRealize,					/* realize */
		actionsList,						/* actions */
		XtNumber (actionsList),				/* num_actions */
		resources,							/* resources */
		XtNumber (resources),				/* num_resources */
		NULLQUARK,							/* xrm_class */
		True,								/* compress_motion */
		True,								/* compress_exposure */
		False,								/* compress_enterlv */
		False,								/* visible_interest */
		Destroy,							/* destroy */
		Resize,								/* resize */
		Redisplay,							/* expose */
		SetValues,							/* set_values */
		NULL,								/* set_values_hook */
		XtInheritSetValuesAlmost,			/* set_values_almost */
		NULL,								/* get_values_hook */
		NULL,								/* accept_focus */
		XtVersion,							/* version */
		NULL,								/* callback_private */
		defaultTranslations,				/* tm_table */
		NULL,								/* query_geometry */
	},
	/* PrimitiveClassPart */
	{
		NULL,								/* Extension */
	},
	/* TextListClassPart */
	{
		NULL,								/* Extension */
	},
};

WidgetClass textListWidgetClass = (WidgetClass)&textListClassRec;

/*
 *	le GC pour le label.
 */

static void get_labelGC (TextListWidget w)
{
	XGCValues		values;
	unsigned long	valuemask;

	valuemask = GCForeground | GCLineWidth | GCBackground | GCFont | GCGraphicsExposures;
	values.graphics_exposures = True;
	values.foreground = P.foreground;
	values.background = C.background_pixel;
	values.line_width = 2;
	values.font = W.font -> fid;

	W.label_GC = XtGetGC ((Widget) w, valuemask, &values);
}

/*
 *	recalcule tout si on change les items.
 */

static void scrollbar_size (TextListWidget w)
{
	W.sb_size = W.num_items ? (W.num_lines * 100) / W.num_items : 100;
	if (W.sb_size > 100)
		W.sb_size = 100;
	XtCallCallbackList ((Widget) w, W.vscroll_size, (XtPointer) (long) W.sb_size);
}

static void scrollbar_position (TextListWidget w)
{
	W.sb_pos = W.num_items ? (W.first_item * 100) / W.num_items : 0;
	if (W.sb_pos + W.sb_size > 100)
		scrollbar_size (w);
	XtCallCallbackList ((Widget) w, W.vscroll_pos, (XtPointer) (long) W.sb_pos);
}

static int change_item (TextListWidget w)
{
	int		width = 0;
	String*	p = W.items;

	/*
	 *	liste vide.
	 */
	if (p == NULL) {
		W.num_items = 0;
		W.selected = NULL;
		W.first_item = 0;
		return 100;
	}

	for (; *p != NULL; ++p) {
		int iw = XTextWidth (W.font, *p, strlen (*p));
		if (iw > width)
			width = iw;
	}
	W.num_items = p - W.items;		/* calculer le nombre d'items */
	if (W.selected)
		XtFree (W.selected);

	W.selected = XtMalloc (W.num_items);
	bzero (W.selected, W.num_items);
	W.first_item = 0;
	return width;
}

/*
 *	Fonction d'initialisation de chaque instance.
 */

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	TextListWidget	w = (TextListWidget) new;
	int				width;

	W.font_height = W.font -> ascent + W.font -> descent + 1;
	W.selected = NULL;

	width = change_item (w);
	if (W.single_selection && W.selection_number != -1)
		W.selected [W.selection_number] = 1;

	if (request -> core.width == 0) {
		C.width = width + 2 * P.shadow_thickness + 2 * W.space + 1;
		if (W.selection_style == 2)
			C.width += W.font_height;
	}

	if (request -> core.height == 0)
		C.height = W.num_lines * W.font_height + 2 * P.shadow_thickness + 2 * W.space - 1;
	else
		W.num_lines = (C.height - (2 * P.shadow_thickness + 2 * W.space - 1)) / W.font_height;

	get_labelGC (w);
}

/*
 *	Allume ou teint la slection.
 */

static void hilite_item (TextListWidget w, int num, Boolean on_off)
{
	int			st = P.shadow_thickness + W.space;
	int			line = st + num * W.font_height;
	Display*	dpy = XtDisplay ((Widget) w);
	Window		win = XtWindow ((Widget) w);

	switch (W.selection_style) {
	default:
	case 0:
		if (on_off) {
			XSetForeground (dpy, W.label_GC, W.selection_color);
			XFillRectangle (dpy, win, W.label_GC, st, line, C.width - 2 * st, W.font_height - 1);
			XSetForeground (dpy, W.label_GC, P.foreground);
			XDrawString (dpy, win, W.label_GC, st+1, line + W.font -> ascent,
							W.items [num+W.first_item], strlen (W.items [num+W.first_item]));
		} else {
			XSetForeground (dpy, W.label_GC, C.background_pixel);
			XFillRectangle (dpy, win, W.label_GC, st, line, C.width - 2 * st, W.font_height - 1);
			XSetForeground (dpy, W.label_GC, P.foreground);
			XDrawString (dpy, win, W.label_GC, st+1, line + W.font -> ascent,
							W.items [num+W.first_item], strlen (W.items [num+W.first_item]));
		}
		return;

	case 1:
		/*
		 *	Entourer la slection par un rectangle.
		 */
		if (on_off) {
			XSetForeground (dpy, W.label_GC, W.selection_color);
			XDrawRectangle (dpy, XtWindow ((Widget) w), W.label_GC, st, line, C.width - 2 * st, W.font_height - 1);
		} else {
			XSetForeground (dpy, W.label_GC, C.background_pixel);
			XDrawRectangle (dpy, XtWindow ((Widget) w), W.label_GC, st, line, C.width - 2 * st, W.font_height - 1);
		}
		XSetForeground (dpy, W.label_GC, P.foreground);
		return;

	case 2:
		/*
		 *	Dessiner un `check mark'.
		 */
		if (on_off) {
			XSetForeground (dpy, W.label_GC, W.selection_color);
			XDrawLine (dpy, XtWindow ((Widget) w), W.label_GC,
						st, line + W.font_height/2,
						st + W.font_height/3, line - 2 + W.font_height);

			XDrawLine (dpy, XtWindow ((Widget) w), W.label_GC,
						st + W.font_height/3, line - 2 + W.font_height,
						st + (2*W.font_height)/3, line - 3);
		} else {
			XSetForeground (dpy, W.label_GC, C.background_pixel);
			XDrawLine (dpy, XtWindow ((Widget) w), W.label_GC,
						st, line + W.font_height/2,
						st + W.font_height/3, line - 2 + W.font_height);

			XDrawLine (dpy, XtWindow ((Widget) w), W.label_GC,
						st + W.font_height/3, line - 2 + W.font_height,
						st + (2*W.font_height)/3, line - 3);
		}
		XSetForeground (dpy, W.label_GC, P.foreground);
		return;
	}
}

/*
 *	Reaffiche tout le textList.
 */

static void repaint (TextListWidget w, int from, int to, int clear)
{
	int			fh, line;
	int			st = P.shadow_thickness + W.space;
	int			n = from;
	int			i = W.first_item + from;
	int			left;
	Display*	dpy = XtDisplay ((Widget) w);

	fh = W.font_height;
	line = st + from * W.font_height;

	if (clear)
		XClearArea (dpy, XtWindow ((Widget) w), st, line, C.width - 2 * st, fh * (to - from), False);

	if (W.first_item + from >= W.num_items)
		return;		/* plus rien a afficher */

	line += W.font -> ascent;
	left = st + 1;
	if (W.selection_style == 2)
		left += W.font_height;
	for (; i < W.num_items && n < to; i++, n++) {
		XDrawString (dpy, XtWindow ((Widget) w), W.label_GC, left, line, W.items [i], strlen (W.items [i]));
		if (W.selected [i])
			hilite_item (w, n, True);
		line += fh;
	}
}

/*
 *	Scrolle de telle facon que "first" soit le
 *	premier item visible.
 */

static void set_first (TextListWidget w, int first)
{
	int			st = P.shadow_thickness;
	int			num;
	Display*	dpy = XtDisplay ((Widget) w);
	Window		win = XtWindow ((Widget) w);
	XEvent		ev;

	if (first < 0)	first = 0;
	else if (first > W.num_items - W.num_lines)
		if ((first = W.num_items - W.num_lines) < 0)
			first = 0;

	num = first - W.first_item;
	if (num == 0)	return;		/* facile! */
	W.first_item = first;
	if (num > 0) {
		int	from = 0;

		if (num < W.num_lines) {
			XCopyArea (dpy, win, win, W.label_GC,
						st, st + W.space + num * W.font_height,
						C.width - 2 * st, (W.num_lines - num) * W.font_height,
						st, st + W.space);
			from = W.num_lines - num;
		}
		repaint (w, from, W.num_lines, True);
	} else {
		int	to = W.num_lines;

		num = -num;
		if (num < W.num_lines) {
			XCopyArea (dpy, win, win, W.label_GC,
						st, st + W.space,
						C.width - 2 * st, (W.num_lines - num) * W.font_height,
						st, st + W.space + num * W.font_height);
			to = num;
		}
		repaint (w, 0, to, True);
	}
	XSync (dpy, False);
	while (XCheckTypedWindowEvent (dpy, win, Expose, &ev) ||
	       XCheckTypedWindowEvent (dpy, win, GraphicsExpose, &ev))
	{
		ev.type = Expose;
		XtDispatchEvent (&ev);
	}
}

void TextListScroll (Widget widget, XtPointer client, XtPointer call)
{
	TextListWidget	w = (TextListWidget) client;
	int				first = (int) (long) call;
	int				num;

	if (W.num_items <= W.num_lines)
		return;

	num = 100 - W.sb_size;
	if (num <= 0) num = 1;
	set_first (w, (first * (W.num_items - W.num_lines)) / num);
}

/*
 *	La fonction de reaffichage.
 */

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	TextListWidget	w = (TextListWidget) widget;
	Display*		dpy = XtDisplay (widget);

	XSetRegion (dpy, P.top_GC, region);
	XSetRegion (dpy, P.bot_GC, region);
	XSetRegion (dpy, W.label_GC, region);

	_DrawShadow (dpy, XtWindow (widget), P.top_GC, P.bot_GC,
					P.shadow_thickness,
					C.width, C.height, 0, 0);
	/*
	 *	Peut arriver a la suite d'un changement de taille
	 *	refuse par le widget parent; donc remettre les
	 *	pendules a l'heure.
	 */
	if (W.num_lines > (C.height - 2 * P.shadow_thickness - 2 * W.space + 1) / W.font_height)
		Resize (widget);

	repaint (w, 0, W.num_lines, False);
	/*
	 *	Virer la region de clipping.
	 */
	XSetClipMask (dpy, W.label_GC, None);
	XSetClipMask (dpy, P.top_GC, None);
	XSetClipMask (dpy, P.bot_GC, None);
}

/*
 *	SetValues.
 */

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	TextListWidget	cu = (TextListWidget) current;
	TextListWidget	w = (TextListWidget) new;
	int				i;
	Boolean			do_redisplay = False;
	/*
	 *	On n'a pas le droit de changer ces ressources.
	 */
	W.sb_size = cu -> textList.sb_size;
	W.sb_pos = cu -> textList.sb_pos;
	W.selection = cu -> textList.selection;

	if (cu -> textList.space != W.space)
		do_redisplay = True;

	if (cu -> textList.font != W.font) {
		XtDestroyGC (W.label_GC);
		get_labelGC (w);
		W.font_height = W.font -> ascent + W.font -> descent + 1;
		Resize (new);
		do_redisplay = True;
	}

	if (W.first_item != cu -> textList.first_item) {
		if (W.num_items > W.num_lines && W.first_item >= W.num_lines && W.first_item < W.num_items)
		{
			scrollbar_position (w);
			do_redisplay = True;
		} else
			W.first_item = cu -> textList.first_item;
	}
	/*
	 *	Changement du contenu de la liste.
	 */
	if (cu -> textList.items != W.items)
	{
		int		width;

		width = change_item (w);
		cu -> textList.selection_number = -1;	/* pour forcer le passage en dessous */
		scrollbar_position (w);
		scrollbar_size (w);
		C.width = width + 2 * P.shadow_thickness + 2 * W.space + 1;
		if (W.selection_style == 2)
			C.width += W.font_height;
		do_redisplay = True;
	}
	/*
	 *	Changement du nombre de lignes visibles.
	 */
	if (cu -> textList.num_lines != W.num_lines) {
		C.height = W.num_lines * W.font_height + 2 * P.shadow_thickness + 2 * W.space - 1;
		do_redisplay = True;
	}
	/*
	 *	Prslection d'un item.
	 */
	if (W.selection_number != cu -> textList.selection_number) {
		if (W.selection_number >= 0 && W.selection_number < W.num_items && W.selected [W.selection_number] == 0) {
			if (W.single_selection)
				for (i = 0; i < W.num_items; i++)
					W.selected [i] = 0;
			W.selected [W.selection_number] = 1;
			do_redisplay = True;
		}
	}
	return do_redisplay;
}

/*
 *	Changement de taille.
 */

static void Resize (Widget widget)
{
	TextListWidget	w = (TextListWidget) widget;

	W.num_lines = (C.height - 2 * P.shadow_thickness - 2 * W.space + 1) / W.font_height;
	scrollbar_size (w);
}

/*
 *	La procedure de destruction d'une instance.
 */

static void Destroy (Widget widget)
{
	TextListWidget	w = (TextListWidget) widget;

	XtDestroyGC (W.label_GC);
	XtFree (W.selected);
}

/*
 *	les actions:
 */

#define where(w, y) ((y - P.shadow_thickness - W.space) / W.font_height)

/*
 *	Selection d'une ligne.
 */

static void Activate (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				y = where (w, e -> y);

	if (W.single_selection && e -> x > 0 && e -> x < C.width && y >= 0 && y < W.num_lines && y < W.num_items - W.first_item && W.selected [y + W.first_item]) {
		W.selection_number = y + W.first_item;
		XtCallCallbackList (widget, P.select, (XtPointer) (long) (y + W.first_item));
	}
}

static void Arm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				y = where (w, e -> y);
	int				i;

	if (y >= 0 && y < W.num_lines && y < W.num_items - W.first_item && !W.selected [y + W.first_item]) {
		if (W.single_selection) {
			for (i = 0; i < W.num_items; i++)
				if (W.selected [i]) {
					W.selected [i] = 0;
					if (i >= W.first_item)
						hilite_item (w, i - W.first_item, False);
					break;
				}
			W.selected [y + W.first_item] = 1;
			W.selection_number = y + W.first_item;
		} else {
			W.selected [y + W.first_item] = 1;
			W.selection_number = y + W.first_item;
			XtCallCallbackList (widget, P.select, (XtPointer) (long) (y + W.first_item));
		}
		hilite_item (w, y, True);	/* met en video inverse */
	}
}

/*
 *	Page suivante
 */

static void NextPage (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;

	if (W.first_item + W.num_lines > W.num_items)
		return;

	set_first (w, W.first_item + W.num_lines);
	scrollbar_position (w);
}

/*
 *	Page prcdente
 */

static void PreviousPage (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;

	if (W.first_item == 0)
		return;

	set_first (w, W.first_item - W.num_lines);
	scrollbar_position (w);
}

/*
 *	Scroll vers le haut.
 */

static void ScrollUp (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;

	if (W.first_item + 1 > W.num_items)
		return;

	set_first (w, W.first_item + 1);
	scrollbar_position (w);
}

/*
 *	Scroll vers le bas.
 */

static void ScrollDown (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;

	if (W.first_item - 1 == 0)
		return;

	set_first (w, W.first_item - 1);
	scrollbar_position (w);
}

/*
 *	Deselection d'une ligne.
 */

static void Release (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				y = where (w, e -> y);

	if (!W.single_selection && y >= 0 && y < W.num_lines && y < W.num_items - W.first_item && W.selected [y + W.first_item]) {
		W.selected [y + W.first_item] = 0;
		hilite_item (w, y, False);	/* remet en video normale */
		XtCallCallbackList (widget, P.release, (XtPointer) (long) (y + W.first_item));
		W.selection_number = -1;
	}
}

/*  Hilite le prochain item qui commence par le caractre tap.
 *  Ne marche qu'en mode XwSINGLE.
 */

static void SearchItem (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;
	XKeyEvent*		e = (XKeyEvent*) event;
	char 			first_char;
	unsigned int	i, n;
	
	if (!W.single_selection)
		return;
	
	/* Quel caractre ? (s'il n'est pas imprimable, on oublie)
	 */
	if (XLookupString(e, &first_char, 1, 0, 0) != 1)
		return;
	
	/* Liste vide : on laisse bton.
	 */
	if (W.num_items == 0)
		return;

	/* On cherche l'item  partir de la slection.
	 */
#define next(i) (((i)+1)%W.num_items)

	for (n = 0, i = next(W.selection_number); n < W.num_items; n++, i = next(i))
		if (W.items[i] && (W.items[i][0] | 0x20) == (first_char | 0x20))
			break;

	/* On d-slectionne le prcdent.
	 */
	if (W.single_selection && W.selection_number >= 0) {
		if (W.selection_number >= W.first_item && W.selection_number < W.first_item + W.num_lines)
			hilite_item (w, W.selection_number - W.first_item, False);
		W.selected [W.selection_number] = 0;
		W.selection_number = -1;
	}

	/* Pas trouv : on gueule !
	 */
	if (n >= W.num_items)
		return;		/* pas trouv */
	/*
	 * On scrolle la liste si ncessaire.
	 */
	if (i < W.first_item || i >= W.first_item + W.num_lines) {
		if (i < W.first_item)
			set_first (w, i);
		else
			set_first (w, i - (W.num_lines - 1));
		scrollbar_position(w);
	}
	/* On slectionne le nouveau.
	 */
	W.selection_number = i;
	W.selected [i] = 1;
	hilite_item (w, i - W.first_item, True);
}

/*
 *	Slectionne l'item suivant.
 */

static void SelectNext (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;

	if (W.selection_number != -1 && W.selection_number < W.num_items - 1) {
		/*
		 * Eteindre le precedent.
		 */
		if (W.selection_number >= W.first_item && W.selection_number < W.first_item + W.num_lines)
			hilite_item (w, W.selection_number - W.first_item, False);
		W.selected [W.selection_number] = 0;
		W.selection_number ++;
		if (W.selection_number >= W.first_item + W.num_lines)
			set_first (w, W.selection_number - (W.num_lines - 1));

		W.selected [W.selection_number] = 1;
		hilite_item (w, W.selection_number - W.first_item, True);

		scrollbar_position (w);
	}
}

/*
 *	Selectionne l'item precedent.
 */

static void SelectPrevious (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;

	if (W.selection_number > 0) {
		/*
		 * Eteindre le precedent.
		 */
		if (W.selection_number >= W.first_item && W.selection_number < W.first_item + W.num_lines)
			hilite_item (w, W.selection_number - W.first_item, False);
		W.selected [W.selection_number] = 0;
		W.selection_number --;
		/*
		 *	Scroller si necessaire.
		 */
		if (W.selection_number < W.first_item)
			set_first (w, W.selection_number);
		/*
		 *	Selectonner le nouveau.
		 */
		W.selected [W.selection_number] = 1;
		hilite_item (w, W.selection_number - W.first_item, True);
		scrollbar_position (w);
	}
}

/*
 *	Appelle le callback avec l'item selectionn.
 */

static void SelectItem (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	TextListWidget	w = (TextListWidget) widget;

	if (W.single_selection && W.selection_number >= 0)
		XtCallCallbackList (widget, P.select, (XtPointer) (long) W.selection_number);
}

/*
 *	La liste scrollante.
 */

static void positionCallback (Widget w, XtPointer client, XtPointer call)
{
	XtVaSetValues ((Widget) client, XtNsliderOrigin, (long) call, NULL);
}

static void sizeCallback (Widget w, XtPointer client, XtPointer call)
{
	XtVaSetValues ((Widget) client, XtNsliderExtent, (long) call, NULL);
}

Widget XpaCreateScrolledList (Widget parent, char* nom, Arg* arg, Cardinal n_arg)
{
	Widget	tl;
	Widget	sw;
	Widget	sb;

	sw = XtCreateWidget ("sw", scrolledWindowWidgetClass, parent, NULL, 0);
	tl = XtCreateManagedWidget (nom, textListWidgetClass, sw, arg, n_arg);
	XtVaGetValues (sw, XtNhScrollBar, &sb, NULL);
	XtUnmanageChild (sb);
	XtVaGetValues (sw, XtNvScrollBar, &sb, NULL);

	XtAddCallback (tl, XtNvScrollPositionCallback, positionCallback, (XtPointer) sb);
	XtAddCallback (tl, XtNvScrollSizeCallback, sizeCallback, (XtPointer) sb);
	XtAddCallback (sb, XtNsliderMoved, TextListScroll, (XtPointer) tl);

	scrollbar_size ((TextListWidget) tl);
	scrollbar_position ((TextListWidget) tl);

	return tl;
}
