/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/keysymdef.h>

#include "Primitive.h"
#include "PrimitiveP.h"
#include "ValuatorP.h"
#include "Valuator.h"

#define monochrome (((Widget)w)->core.depth == 1)

/*
 *	Procdures d'action.
 */

static void	Arm (Widget, XEvent*, String*, Cardinal*);
static void	Disarm (Widget, XEvent*, String*, Cardinal*);
static void	Drag (Widget, XEvent*, String*, Cardinal*);

/* Default translation table and action list  */

static char defaultTranslations [] = "\
	<Btn1Down>:		Arm()\n\
    <Btn1Up>:		Disarm()\n\
    <Btn1Motion>:	Drag()";

static XtActionsRec actionsList [] =
{
	{"Arm",		Arm},
	{"Disarm",	Disarm},
	{"Drag",	Drag},
};

/* Resource list for Valuator  */
static XtResource resources[] =
{
	{ XtNsliderMin, XtCSliderMin, XtRInt, sizeof (int),
	  XtOffset (ValuatorWidget, valuator.min), XtRImmediate, (XtPointer) 0
	},
	{ XtNsliderMax, XtCSliderMax, XtRInt, sizeof (int),
	  XtOffset (ValuatorWidget, valuator.max), XtRImmediate, (XtPointer) 100
	},
	{ XtNsliderOrigin, XtCSliderOrigin, XtRInt, sizeof (int),
	  XtOffset (ValuatorWidget, valuator.origin), XtRImmediate, (XtPointer) 0
	},
	{ XtNsliderExtent, XtCSliderExtent, XtRInt, sizeof (int),
	  XtOffset (ValuatorWidget, valuator.extent), XtRImmediate, (XtPointer) 10
	},
	{ XtNslideOrientation, XtCSlideOrientation, XtRSlideOrientation, sizeof (int),
	  XtOffset (ValuatorWidget, valuator.orientation), XtRImmediate, (XtPointer) XwVERTICAL
	},
	{ XtNsliderMoved, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset (ValuatorWidget, valuator.slider_moved), XtRPointer, (XtPointer) NULL
	},
	{ XtNsliderReleased, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset (ValuatorWidget, valuator.slider_released), XtRPointer, (XtPointer) NULL
	},
	{ XtNareaSelected, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset (ValuatorWidget, valuator.area_selected), XtRPointer, (XtPointer) NULL
	},
	{ XtNsliderColor, XtCArmColor, XtRPixel, sizeof (Pixel),
	  XtOffset (ValuatorWidget, valuator.slider_color),
	  XtRCallProc, (XtPointer) _GetArmColor
	},
};

/* Non-transitory variables used for interactive sliding  */
static int	initialX;
static int	initialY;
static int	separationX;
static int	separationY;

/* Static routine definitions  */
static void		Initialize (Widget, Widget, Arg*, Cardinal*);
static void		Resize (Widget);
static void		Destroy (Widget);
static void		Redisplay (Widget, XEvent*, Region);
static Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);

/* The Valuator class record definition  */
ValuatorClassRec XwvaluatorClassRec =
{
	{
		(WidgetClass) & primitiveClassRec,	/* superclass */
		"Valuator",							/* class_name */
		sizeof (ValuatorRec),				/* widget_size */
		NULL,								/* class_initialize */
		NULL,								/* class_part_initialize */
		FALSE,								/* class_inited */
		Initialize,							/* initialize */
		NULL,								/* initialize_hook */
		XtInheritRealize,					/* realize */
		actionsList,						/* actions */
		XtNumber (actionsList),				/* num_actions */
		resources,							/* resources */
		XtNumber (resources),				/* num_resources */
		NULLQUARK,							/* xrm_class */
		TRUE,								/* compress_motion */
		TRUE,								/* compress_exposure */
		TRUE,								/* compress_enterleave */
		FALSE,								/* visible_interest */
		Destroy,							/* destroy */
		Resize,								/* resize */
		Redisplay,							/* expose */
		SetValues,							/* set_values */
		NULL,								/* set_values_hook */
		XtInheritSetValuesAlmost,			/* set_values_almost */
		NULL,								/* get_values_hook */
		NULL,								/* accept_focus */
		XtVersion,							/* version */
		NULL,								/* callback private */
		defaultTranslations,				/* tm_table */
		NULL,								/* query_geometry */
	},
	{
		NULL,								/* Extension */
	},
	{
		NULL,								/* Extension */
	}
};

WidgetClass XwvaluatorWidgetClass = (WidgetClass) & XwvaluatorClassRec;


static void GetSliderGC (ValuatorWidget w)
{
	Widget			widget = (Widget) w;
	static Pixmap	greyPixmap = (Pixmap)0;
	static char		grey_bits [] = {
		0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55
	};
	XGCValues       values;
	XtGCMask        valueMask;

	valueMask = GCForeground | GCGraphicsExposures;
	if (monochrome) {
		if (!greyPixmap)
			greyPixmap = XCreatePixmapFromBitmapData (XtDisplay (widget),
							RootWindowOfScreen (XtScreen (widget)),
							grey_bits, 8, 8,
							BlackPixelOfScreen (XtScreen (widget)),
							WhitePixelOfScreen (XtScreen (widget)),
							DefaultDepthOfScreen (XtScreen (widget)));
		valueMask |=  GCTile | GCFillStyle;
		values.fill_style = FillTiled;
		values.tile = greyPixmap;
	}
	values.foreground = W.slider_color != 12345
							? W.slider_color
							: P.foreground;
	values.graphics_exposures = False;
	W.slider_GC = XtGetGC (widget, valueMask, &values);
}

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	ValuatorWidget	w = (ValuatorWidget) new;
	Pixel			tmp;

	/* Set the internally used variables.  */
	W.sliding_on = FALSE;

	/* Check the data put into the new widget from .Xdefaults  */
	/* or through the arg list.                                */
	if (W.min >= W.max) {	W.min = 0; W.max = 100; }

	if (W.extent < 1 || W.extent > W.max - W.min)
		W.extent = 1;

	if (W.origin < W.min || W.origin > W.max - W.extent)
		W.origin = W.min;

	if (W.orientation != XwHORIZONTAL && W.orientation != XwVERTICAL)
		W.orientation = XwVERTICAL;

	/* Set up a geometry for the widget if it is currently 0.  */
	if (request -> core.width == 0) {
		if (W.orientation == XwHORIZONTAL)
			C.width += 100;
		else
			C.width = 15;
	}
	if (request->core.height == 0) {
		if (W.orientation == XwHORIZONTAL)
			C.height = 15;
		else
			C.height += 100;
	}
	/* Get the drawing graphics contexts.  */
	tmp = C.background_pixel;
	C.background_pixel = W.slider_color;
	W.slider_color = tmp;
	GetSliderGC (w);
}

/*
 *	Calculate the slider location and size in pixels so that it can be drawn.
 */

static void CalcSliderRect (ValuatorWidget w)
{
	float	userSize;
	float	trueSize;
	float	factor;
	float	slideStart;
	float	slideSize;
	int		borderWidth = P.shadow_thickness;

	userSize = W.max - W.min;
	trueSize = ((W.orientation == XwHORIZONTAL) ? C.width : C.height) - 2.0 * borderWidth;

	factor = trueSize / userSize;
	slideStart = (float) (W.origin - W.min) * factor + borderWidth;
	slideSize = (float) (W.extent) * factor;

	if (W.orientation == XwHORIZONTAL) {
		W.sliderX = (int) (slideStart + 0.5);
		W.sliderY = borderWidth;
		W.sliderWidth = (int) (slideSize + 0.5);
		W.sliderHeight = C.height - 2 * borderWidth;
	} else {
		W.sliderX = borderWidth;
		W.sliderY = (int) (slideStart + 0.5);
		W.sliderWidth = C.width - 2 * borderWidth;
		W.sliderHeight = (int) (slideSize + 0.5);
	}
	if (W.sliderWidth < MIN_SLIDER_SIZE) {
		W.sliderWidth = MIN_SLIDER_SIZE;
		if (W.orientation == XwHORIZONTAL && W.sliderX + W.sliderWidth + 1 > C.width)
			W.sliderX = C.width - W.sliderWidth - 1;
	}
	if (W.sliderHeight < MIN_SLIDER_SIZE) {
		W.sliderHeight = MIN_SLIDER_SIZE;
		if (W.orientation == XwVERTICAL && W.sliderY + W.sliderHeight + 1 > C.height)
			W.sliderY = C.height - W.sliderHeight - 1;
	}
}

static void DrawSlider (ValuatorWidget w, int x, int y)
{
	XFillRectangle (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					W.slider_GC,
					x, y, W.sliderWidth, W.sliderHeight);

	if (P.shadow_thickness > 0)
		_DrawShadow (XtDisplay ((Widget) w), XtWindow ((Widget) w),
						W.armed ? P.bot_GC : P.top_GC,
						W.armed ? P.top_GC : P.bot_GC,
						P.shadow_thickness, W.sliderWidth, W.sliderHeight, x, y);
}

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	ValuatorWidget	w = (ValuatorWidget) widget;

	XSetRegion (XtDisplay (widget), P.top_GC, region);
	XSetRegion (XtDisplay (widget), P.bot_GC, region);
	if (P.shadow_thickness > 0)
		_DrawShadow (XtDisplay (widget), XtWindow (widget),
						P.bot_GC, P.top_GC, P.shadow_thickness,
						C.width, C.height, 0, 0);
	XSetClipMask (XtDisplay (widget), P.top_GC, None);
	XSetClipMask (XtDisplay (widget), P.bot_GC, None);

	DrawSlider (w, W.sliderX, W.sliderY);
}

static void Resize (Widget widget)
{
	CalcSliderRect ((ValuatorWidget) widget);
}

static void Destroy (Widget widget)
{
	ValuatorWidget	w = (ValuatorWidget) widget;

	XtDestroyGC (W.slider_GC);
	XtRemoveAllCallbacks (widget, XtNsliderMoved);
	XtRemoveAllCallbacks (widget, XtNareaSelected);
	XtRemoveAllCallbacks (widget, XtNareaSelected);
}

static void RedrawSlider (ValuatorWidget w)
{
	if (XtIsRealized ((Widget) w)) {
		XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					W.sliderX, W.sliderY,
					W.sliderWidth, W.sliderHeight,
					False);
		CalcSliderRect (w);
		DrawSlider (w, W.sliderX, W.sliderY);
	}
}

static Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	ValuatorWidget	cu  = (ValuatorWidget) current;
	ValuatorWidget	w = (ValuatorWidget) new;
	Boolean			returnFlag = False;

	if (W.min >= W.max) {
		W.min = cu->valuator.min;
		W.max = cu->valuator.max;
	}
	if (W.extent < 1 || W.extent > W.max - W.min) {
		if (cu->valuator.extent <= W.max - W.min)
			W.extent = cu->valuator.extent;
		else
			W.extent = 1;
	}

	if (W.origin < W.min)
		W.origin = W.min;

	if (W.origin > W.max - W.extent)
		W.origin = W.max - W.extent;

	if (W.orientation != XwHORIZONTAL && W.orientation != XwVERTICAL)
		W.orientation = cu->valuator.orientation;

	/* See if the GC's need to be regenerated and widget redrawn.  */
	if (P.foreground != cu->primitive.foreground) {
		returnFlag = True;
		XtDestroyGC (W.slider_GC);
		GetSliderGC (w);
	}
	if (W.extent != cu->valuator.extent || W.min != cu->valuator.min ||
		W.max != cu->valuator.max || W.orientation != cu->valuator.orientation)
	{
		returnFlag = True;
	}
	/* If the return flag is False "no redraw" and the slider  */
	/* origin has changed, erase and redraw the slider.        */
	if (returnFlag == False && W.origin != cu->valuator.origin)
		RedrawSlider (w);

	CalcSliderRect (w);
	return returnFlag;
}


static int CalcSliderVal (ValuatorWidget w, int x, 	int y)
{
	int		st = P.shadow_thickness;
	int		borderWidth = st;
	float	userSize;
	float	trueSize;
	float	referencePoint;
	float	temp;

	if (W.orientation == XwHORIZONTAL) {
		referencePoint = (float) x - separationX;
		trueSize = C.width - 2 * borderWidth - 4;
	} else {
		referencePoint = (float) y - separationY;
		trueSize = C.height - 2 * borderWidth - 4;
	}
	temp = (referencePoint - borderWidth - 1) / trueSize;
	userSize = W.max - W.min;
	temp = temp * userSize + W.min + 0.5;
	if ((int) temp < W.min)
		temp = W.min;
	else if ((int) temp > W.max - W.extent)
		temp = W.max - W.extent;
	return (int) temp;
}

static void Arm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ValuatorWidget	w = (ValuatorWidget) widget;
	XButtonEvent*	buttonEvent = (XButtonEvent*) event;
	int				borderWidth = 0;
	int				selectionPos;
	int				sliderX = W.sliderX;
	int				sliderY = W.sliderY;
	int				sliderWidth = W.sliderWidth;
	int				sliderHeight = W.sliderHeight;

	if (C.width - 2 * borderWidth > 3 && C.height - 2 * borderWidth > 3) {
		/* Calculate whether the selection point is in the slider  */
		if (buttonEvent->x >= sliderX && buttonEvent->x <= sliderX + sliderWidth - 1 &&
		    buttonEvent->y >= sliderY && buttonEvent->y <= sliderY + sliderHeight - 1)
		{
			separationX = buttonEvent->x - sliderX;
			separationY = buttonEvent->y - sliderY;
			initialX = sliderX;
			initialY = sliderY;
			W.sliding_on = True;
			W.armed = True;
			DrawSlider (w, sliderX, sliderY);
		} else {
			separationX = 0;
			separationY = 0;
			selectionPos = CalcSliderVal (w, buttonEvent->x, buttonEvent->y);
			XtCallCallbacks (widget, XtNareaSelected, (XtPointer) (long) selectionPos);
		}
	}
}

static void Disarm (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ValuatorWidget	w = (ValuatorWidget) widget;

	W.armed = False;
	DrawSlider (w, W.sliderX, W.sliderY);
	if (W.sliding_on == True) {
		W.sliding_on = False;
		XtCallCallbacks (widget, XtNsliderReleased, (XtPointer) (long) W.origin);
	}
}

static void MoveSlider (ValuatorWidget w, int currentX, int currentY)
{
	int			oldX = W.sliderX;
	int			oldY = W.sliderY;
	int			width = W.sliderWidth;
	int			height = W.sliderHeight;
	int			st = P.shadow_thickness;
	XSegment	seg [2];

	if (W.orientation == XwHORIZONTAL) {
		W.sliderX = currentX;
		seg[0].y1 = seg[0].y2 = oldY + 2 * st;
		seg[1].y1 = seg[1].y2 = oldY + height - 3 * st;
		if (oldX < currentX) {
			seg[0].x1 = seg[1].x1 = oldX;
			seg[0].x2 = seg[1].x2 = oldX + currentX - oldX - st;
		} else {
			seg[0].x1 = seg[1].x1 = currentX + width;
			seg[0].x2 = seg[1].x2 = seg[0].x1 + oldX - currentX - st;
		}

		DrawSlider (w, currentX, currentY);
		XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					seg[0].x1, oldY,
					seg[0].x2 - seg[0].x1 + st, height, False);
	} else {
		W.sliderY = currentY;
		seg[0].x1 = seg[0].x2 = oldX + 2 * st;
		seg[1].x1 = seg[1].x2 = oldX + width - 3 * st;
		if (oldY < currentY) {
			seg[0].y1 = seg[1].y1 = oldY;
			seg[0].y2 = seg[1].y2 = oldY + currentY - oldY - st;
		} else {
			seg[0].y1 = seg[1].y1 = currentY + height;
			seg[0].y2 = seg[1].y2 = seg[0].y1 + oldY - currentY - st;
		}
		DrawSlider (w, currentX, currentY);
		XClearArea (XtDisplay ((Widget) w), XtWindow ((Widget) w),
					oldX, seg[0].y1,
					width, seg[0].y2 - seg[0].y1 + st, False);
	}
}

static void Drag (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	ValuatorWidget	w = (ValuatorWidget) widget;
	XButtonEvent*	buttonEvent = (XButtonEvent*) event;
	int				newX, newY;
	int				slideVal;
	int             st = P.shadow_thickness;

	if (W.sliding_on != True)
		return;

	if (W.orientation == XwHORIZONTAL) {
		newX = buttonEvent->x - separationX;
		newY = W.sliderY;
		if (newX < st)
			newX = st;
		if (newX + W.sliderWidth > C.width - st)
			newX = C.width - W.sliderWidth - st;
	} else {
		newX = W.sliderX;
		newY = buttonEvent->y - separationY;
		if (newY < st)
			newY = st;
		if (newY + W.sliderHeight > C.height - st)
			newY = C.height - W.sliderHeight - st;
	}

	if ((W.orientation == XwHORIZONTAL && newX != initialX) ||
	    (W.orientation == XwVERTICAL && newY != initialY))
	{
		slideVal = CalcSliderVal (w, buttonEvent->x, buttonEvent->y);
		if (slideVal == W.origin)
			return;

		W.origin = slideVal;
		MoveSlider (w, newX, newY);
		XtCallCallbacks (widget, XtNsliderMoved, (XtPointer) (long) slideVal);
		initialX = newX;
		initialY = newY;
	}
}
