/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/ShellP.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>

static void	Initialize (Widget, Widget, Arg*, Cardinal*);
static void	Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static void	Resize (Widget);
static void	Destroy (Widget);
static void ClassInitialize (void);

static Boolean	_WithShapeExtension = False;

#ifndef XtRMenuBar
#define	XtRMenuBar	"MenuBar"
#endif

#ifndef XtRWixFont
#define XtRWixFont	"WixFont"
#endif

#ifndef XtRAliasTable
#define XtRAliasTable	"AliasTable"
#endif

#ifndef XtRKeyTable
#define XtRKeyTable	"KeyTable"
#endif

typedef struct {
	void		(*_delete_response) (Widget);
	void		(*_shape_resize) (Widget, Pixmap, int, int);
	Pixmap		_shape_mask;
} VendorExtRec;

#define delete_response	((VendorExtRec*)(widget -> core.constraints)) -> _delete_response
#define shape_resize	((VendorExtRec*)(widget -> core.constraints)) -> _shape_resize
#define shape_mask		(((VendorExtRec*)(widget -> core.constraints)) -> _shape_mask)

static XtResource resources [] = {
	{ XtNinput, XtCInput, XtRBool, sizeof(Bool),
	  XtOffset (VendorShellWidget, wm.wm_hints.input), XtRImmediate, (XtPointer)True
	},
};

VendorShellClassRec vendorShellClassRec = {
	{
		/* superclass */			(WidgetClass) &wmShellClassRec,
		/* class_name */			"VendorShell",
		/* size */					sizeof(VendorShellRec),
		/* Class Initializer */		ClassInitialize,
		/* class_part_initialize */	NULL,
		/* Class init'ed ? */			FALSE,
		/* initialize */				Initialize,
		/* initialize_notify */		NULL,		
		/* realize */				Realize,
		/* actions */				NULL,
		/* num_actions */			0,
		/* resources */				resources,
		/* resource_count */		XtNumber (resources),
		/* xrm_class */				NULLQUARK,
		/* compress_motion */		FALSE,
		/* compress_exposure */	TRUE,
		/* compress_enterleave */	FALSE,
		/* visible_interest */		FALSE,
		/* destroy */				Destroy,
		/* resize */				Resize,
		/* expose */				NULL,
		/* set_values */			NULL,
		/* set_values_hook */		NULL,			
		/* set_values_almost */		XtInheritSetValuesAlmost,  
		/* get_values_hook */		NULL,
		/* accept_focus */			NULL,
		/* intrinsics version */		XtVersion,
		/* callback offsets */		NULL,
		/* tm_table */				NULL,
		/* query_geometry */		NULL,
		/* display_accelerator */		NULL,
		/* extension */				NULL
	},{
		/* geometry_manager */		XtInheritGeometryManager,
		/* change_managed */		XtInheritChangeManaged,
		/* insert_child */			XtInheritInsertChild,
		/* delete_child */			XtInheritDeleteChild,
		/* extension */				NULL
	},{
		/* extension */				NULL
	},{
		/* extension */				NULL
	},{
		/* extension */				NULL
	}
};

WidgetClass vendorShellWidgetClass = (WidgetClass) &vendorShellClassRec;

/*
 *	Affecte les handlers de delete_window et de re_shape.
 */

void SetDeleteWindow (Widget widget, void (*proc) (Widget))
{
	delete_response = proc;
}

void SetReshape (Widget widget, void (*proc) (Widget, Pixmap, int, int))
{
	shape_resize = proc;
}

/*
 *	WM_DELETE_WINDOW handler.
 */

static Atom	wm_protocols;
static Atom	wm_delete_window;

static void wm_delete_handler (Widget widget, XtPointer cd, XEvent* event, Boolean* cont)
{
	VendorShellWidget	w = (VendorShellWidget) widget;

	if (event -> type == ClientMessage &&
	    event -> xclient.message_type == wm_protocols &&
	    event -> xclient.data.l[0] == (long) wm_delete_window)
	{
		if (delete_response == 0) {
			if (XtIsSubclass (widget, transientShellWidgetClass))
				XtUnmapWidget (widget);
			else
			if (XtIsSubclass (widget, applicationShellWidgetClass))
				exit (0);
			else
				XtUnmapWidget (widget);
		} else
			(*delete_response) (widget);
	}
}

static void Destroy (Widget widget) {
	if (shape_mask)
		XFreePixmap (XtDisplay (widget), shape_mask);
}

static void call_reshape (Widget widget)
{
	if (shape_mask)
		XFreePixmap (XtDisplay (widget), shape_mask);
	shape_mask = XCreatePixmap (XtDisplay (widget), XtWindow (widget),
													widget -> core.width,
													widget -> core.height, 1);
	(*shape_resize) (widget, shape_mask,
						(int) widget -> core.width,
						(int) widget -> core.height);
	XShapeCombineMask (XtDisplay (widget), XtWindow (widget),
						ShapeBounding, 0, 0, shape_mask, ShapeSet);
}

/*
 *	La procdure de traitement du changement de taille.
 */

static void Resize (Widget widget)
{
	(*wmShellClassRec.core_class.resize) (widget);
	if (_WithShapeExtension && (shape_resize != 0) && XtIsRealized (widget))
		call_reshape (widget);
}

/*
 *	La procdure de cration de la fentre X.
 */

static void Realize	(Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	Display*		dpy = XtDisplay (widget);
	static Boolean	first = True;
	int				shape_event_base;
	int				shape_error_base;

	if (first) {
		first = False;
		wm_delete_window = XInternAtom (dpy, "WM_DELETE_WINDOW", False);
		wm_protocols = XInternAtom (dpy, "WM_PROTOCOLS", False);
		_WithShapeExtension = XShapeQueryExtension (dpy, &shape_event_base, &shape_error_base);
	}

	(*wmShellClassRec.core_class.realize) (widget, value_mask, attributes);
	(void) XSetWMProtocols (dpy, XtWindow (widget), &wm_delete_window, 1);
	XtAddEventHandler (widget, 0, True, wm_delete_handler, NULL);
	if (_WithShapeExtension && (shape_resize != 0))
		call_reshape (widget);
}

/*
 *	La procdure d'initilisation.
 */

static void Initialize (Widget request, Widget widget, Arg* args, Cardinal* nargs)
{
	VendorExtRec*	vp = (VendorExtRec*) XtMalloc (sizeof (VendorExtRec));

	vp -> _delete_response = 0;	/* pas de delete_response */
	vp -> _shape_resize = 0;	/* pas de shape_resize */
	vp -> _shape_mask = 0;		/* car pas de shape_mask */
	widget -> core.constraints = (XtPointer) vp;	/* planqu dans les contraintes */
}

/*
 *	Les divers convertisseurs de ressources.
 */

/*
 *	String to Pixmap conversion.
 */

static XtConvertArgRec strToPix [] = {
	{ XtBaseOffset, (XtPointer) XtOffset (Widget, core.parent), sizeof (Widget) },
};

static void CvtStringToPixmap (XrmValue* args, Cardinal* num_args, XrmValue* from, XrmValue* to)
{
	int				raf;
	static Pixmap	pix;
	Widget			w = * ((Widget*) args [0].addr);
	Display*		dpy = XtDisplay (w);
	Window			root = RootWindowOfScreen (XtScreen (w));

	to -> size = sizeof(Pixmap);
	to -> addr = (XtPointer)&pix;

	if(strcmp (from -> addr, "none") == 0) {
		pix = None;
		return;
	}
	if (BitmapSuccess != XReadBitmapFile (dpy, root, (char*)from -> addr,
										      &raf, &raf, &pix, &raf, &raf) &&
	    XpmSuccess != XpmReadPixmapFile (dpy, root, (char*)from -> addr,
										     &pix, NULL, NULL))
	{
		fprintf (stderr, "Cannot convert the file \033[4m%s\033[m to a pixmap.\n", (char*)from -> addr);
		pix = None;
	}
}

/*
 *	String to StringArray conversion.
 */

static void CvtStringToStringArray (XrmValue* args, Cardinal* num_args, XrmValue* from, XrmValue* to)
{
	static String*	result;
	String			string;
	String			s;
	int				n;

	to -> size = sizeof(String*);
	to -> addr = (XtPointer)&result;

	string = s = XtNewString ((String) from -> addr);
	for (n = 0; *s;)
		if (*s++ == ',')
			n++;
	/* il ne faut pas de virgule a la fin du dernier  (meme si c'est le seul) */
	if (n == 0) {
		result = (String*) XtMalloc (2 * sizeof (String));
		result [0] = string;
		result [1] = NULL;
		return;
	}
	/* il y en a plusieurs */
	n++;	/* comme on compte les virgule il y a une chaine de plus */
	result = (String*) XtMalloc ((n+1) * sizeof (String));
	n = 0;
	s = string;
	result [0] = s;
	while (*s)
		if (*s++ == ',') {
			s [-1] = 0;
			result [++n] = s;
		}
	result [++n] = NULL;
}

/*
 *	Convertisseur pour la dfinition des menuBars.
 */

static void StringToMenuBar (XrmValue* args, Cardinal* nargs, XrmValue* from, XrmValue* to)
{
#define	IN_MENU	1
	static String**	result;
	String*			p;
	String**		mp;
	char*			cp;
	char*			np;
	int				state = 0;
	int				bar, menuBar;

	to -> size = sizeof (result);
	cp = (String) from -> addr;
	if (cp == NULL) {
		to -> addr = NULL;
		return;
	}
/*
 *	Calculer la place memoire necessaire.
 */
	bar = 0;
	menuBar = 0;
	for (;;) {
		switch (*cp++) {
		default:	continue;
		case '[':	bar++; menuBar++; continue;
		case ',':	menuBar++; continue;
		case '\0':	break;
		}
		break;
	}
	menuBar += bar;	/* un null de plus au bout de chaque menuBar */
	bar ++;		/* un null de plus au bout de la barre */
/*
 *	Allouer la place.
 */
	mp = (String**)	XtMalloc (bar * sizeof (String**));
	p = (String*)	XtMalloc (menuBar * sizeof (String*));
	result = mp;
	to -> addr = (XtPointer) &result;
/*
 *	Analyser la chaine.
 */
	np = XtMalloc (strlen ((char*) from -> addr));
	cp = (char*) from -> addr;
	for (;;) switch (*np++ = *cp++) {
		case '\t':
			while (*cp == '\t')
				cp++;
			continue;

		case '[':		/* nouveau menu */
			if (state != IN_MENU) {
				state = IN_MENU;
				*mp++ = p;
				*p++ = np;
			}
			continue;
		case ']':
			if (state == IN_MENU) {
				np [-1] = '\0';	/* remplacer le crochet par un null */
				*p++ = NULL;	/* fin du menu */
				state = 0;
			}
			continue;
		case ',':
			if (state == IN_MENU) {
				np [-1] = '\0';	/* remplacer la virgule par un null */
				*p ++ = np;	/* item suivant */
			} else
				XtWarning ("Bad menuBar conversion: ] , [");
			continue;
		case '\0':
			if (state == IN_MENU) {
				*p = NULL;	/* fin du menu */
				XtWarning ("Bad menuBar conversion: missing ]");
			}
			*mp = NULL;	/* fin de la barre */
			return;
	}
}

static void ClassInitialize (void)
{
	extern void	_WixRegisterConverters ();

	XtAddConverter (XtRString, XtRPixmap, CvtStringToPixmap, strToPix, XtNumber (strToPix));
	XtAddConverter (XtRString, XtRStringArray, CvtStringToStringArray, NULL, 0);
	XtAddConverter (XtRString, XtRMenuBar, StringToMenuBar, NULL, 0);
	_WixRegisterConverters ();
}
