/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#ifndef	WIX_H
#define	WIX_H

#include <Xpa/ScrolledWin.h>

extern WidgetClass wixWidgetClass;

/*
 *	les codes d'erreurs et de messages.
 */

#define	T_OK		0
#define	T_ERROR		1
#define	T_REMOVED	2
#define	T_COPIED	3
#define	T_PREFIX	4
#define	T_ARGCNT	5
#define	T_OCTAL		6
#define	T_WRITTEN	7

#define	E_EXEC_MACRO		0
#define	E_BEG_BUFFER		1
#define	E_END_BUFFER		2
#define	E_OPENING			4
#define	E_CLOSING			5
#define	E_SET_MARK			6
#define	E_UNDEF_CMD			7
#define	E_BAD_OCTAL			8
#define	E_BAD_FONT			9
#define E_EXPAND_ALIAS		10
#define E_EXPAND_BUFFER		11
#define E_EXPAND_FILE		12
#define E_EXPAND_KILL_STACK	13

#define INSERT_STRING	0
#define DELETE_STRING	1
#define COMMAND_MARK	2
#define UNMODIFIED_MARK	3

typedef struct {
	unsigned char*	a_src;		/* ce qu'on remplace */
	unsigned char*	a_dst;		/* ce qui est remplace */
} alias;

/*
 *	Traitement des mots clefs.
 */

#define	MAX_KEYWORDS	512
#define MAX_REGEX		64

#define HILITED	16

#define STANDARD_COLOR		0
#define COMMENT_COLOR		1
#define CPP_COLOR			2
#define KEYWORD_COLOR		3
#define LIBRARY_COLOR		4
#define STRING_COLOR		5
#define CONSTANT_COLOR		6
#define CLASS_COLOR			7

typedef struct {
	char*	keyword;		/* le nom */
	int		length;			/* sa longueur */
	int		color;			/* sa couleur */
	int		class_col;		/* celle du prcdent / suivant */
	int		regexp;			/* chaine / regexpr / temporaire */
} KeyWord;

#define CurrentCharType(w) ((int)((*_GetAttCursorEnd(w)) & ~HILITED))

/*
 *	Tout pour les fontes.
 */

typedef struct {
	XFontStruct*	f_info;		/* structure X */
	char*			f_width;	/* table des largeurs des caracteres */
	int				f_height;	/* hauteur de la fonte */
	int				f_ascent;	/* ligne de base */
	char*			f_name;		/* le nom complet de la fonte */
} FontInfo;

/*
 *	La table des clefs.
 */

typedef struct _Trans {
	struct _Trans*	prefix;	/* est-ce un prefixe: 0: non, sinon @ table */
	XtActionProc	proc;	/* l'action associee */
	char*			arg;	/* son eventuel argument */
} Trans;

#define	K_UP		128
#define	K_DOWN		129
#define	K_LEFT		130
#define	K_RIGHT		131
#define	K_NUMBER	(K_RIGHT+1)	/* nombre total de codes utilises */

#define Width(c,fp) ((fp) -> f_width [(unsigned)(c)])

/*
 *	Les informations passes  la VerifyCallback
 */

typedef struct {
	int		reason;
	int		pos;
	char*	string;
	int		length;
	int		doit;
} VerifyCallbackStruct;

/*
 *	Les informations passes  la undoRedoCallback
 */

typedef struct {
	int				reason;
	int				pos;
	unsigned char*	string;
	int				length;
} UndoRedoCallbackStruct;

typedef void (*MessageProc) (Widget, int, int);

/*
 *	Format des commandes (pour les macros)
 */

#ifndef	_PRIVATE_
typedef int (*CommandProc) (Widget, int, void*);

typedef struct {
	CommandProc function;
	int			count;
	void*		arg;
} Command;

typedef struct {
	char*			string;
	XtActionProc	proc;
	CommandProc		function;
	char*			help_string;
} WixCommandRec;

extern Command	DefaultMacroBuffer [];
extern _CPP_ void		nlws (Widget, int, unsigned char**);
extern _CPP_ void		WixSetHorizontalScroll (Widget, int);
extern _CPP_ int		WixGetHorizontalScroll (Widget);
extern _CPP_ int		WixGetY (Widget);
extern _CPP_ void		WixSetRefreshArea (Widget, int, int);
extern _CPP_ int		WixGetNumLines (Widget);
extern _CPP_ void		SaveCommand (Widget, CommandProc, void*, Boolean);
extern _CPP_ int		InsertLineNumber (Widget w, int, void*);
extern _CPP_ int		InsertChar (Widget, int, void*);
extern _CPP_ int		BeginningOfBuffer (Widget, int, void*);
extern _CPP_ int		BeginningOfLine (Widget, int, void*);
extern _CPP_ int		EndOfLine (Widget, int, void*);
extern _CPP_ int		EndOfBuffer (Widget, int, void*);
extern _CPP_ void		KInsertString (Widget, unsigned char*, int);
extern _CPP_ void		WixHighlight (Widget, int);
extern _CPP_ void		WixUnhighlight (Widget, int);
extern _CPP_ void		WixUnhighlightAll (Widget);
extern _CPP_ void		WixReplace (Widget, int, int, char*, int);
extern _CPP_ int		WixBackwardSearch (Widget, unsigned char*, int);
extern _CPP_ unsigned char*
						_WixForwardSearch (Widget w, unsigned char*, unsigned char*, int);
extern _CPP_ unsigned char*
						_WixBackwardSearch (Widget w, unsigned char*, unsigned char*, int);
extern _CPP_ int		WixForwardSearch (Widget, unsigned char*, int);
extern _CPP_ int		BackwardChar (Widget, int, void*);
extern _CPP_ int		ForwardChar (Widget, int, void*);
extern _CPP_ int		ForwardLine (Widget, int, void*);
extern _CPP_ int		NewLine (Widget, int, void*);
extern _CPP_ int		find_open (Widget, int, int, Boolean);
extern _CPP_ int		find_close (Widget, int, int, Boolean);
extern _CPP_ int		CBeginningOfFunction (Widget, int, void*);
extern _CPP_ int		CEndOfFunction (Widget, int, void*);
extern _CPP_ int		removeChar (Widget, int, Boolean);
extern _CPP_ void		RemoveCursor (Widget);
extern _CPP_ void		ShowCursor (Widget);
extern _CPP_ void		StopBlink (void);
extern _CPP_ void		StartBlink (Widget);
#endif

extern long	EndOfKillStack;
extern int	FdKillFile;
extern int	num_arg;

/*
 *	Les fonctions exportes
 */

extern _CPP_ void		start_macro_learn (Widget);
extern _CPP_ void		stop_macro_learn (Widget);

extern _CPP_ void	insert_char (Widget, XEvent*, String*, Cardinal*);

extern _CPP_ unsigned char*	_GetBufferBeg (Widget);
extern _CPP_ unsigned char*	_GetBufferEnd (Widget);
extern _CPP_ unsigned char*	_GetCursorBeg (Widget);
extern _CPP_ unsigned char*	_GetCursorEnd (Widget);

extern _CPP_ unsigned char*	_GetAttCursorBeg (Widget);
extern _CPP_ unsigned char*	_GetAttCursorEnd (Widget);
extern _CPP_ unsigned char*	_GetAttributeBeg (Widget);
extern _CPP_ unsigned char*	_GetAttributeEnd (Widget);

extern _CPP_ char*		GetKilled (int, long*);
extern _CPP_ int		WixGotoLine (Widget, int);
extern _CPP_ void		WixSetFont (Widget, int);
extern _CPP_ int		WixGetColumn (Widget);
extern _CPP_ int		WixGetWidth (Widget);
extern _CPP_ char*		WixCopyBuffer (Widget);
extern _CPP_ void		WixRedraw (Widget);
extern _CPP_ int		EnableRedisplay (Widget, Boolean, Boolean);
extern _CPP_ void		ClearRegion (Widget);
extern _CPP_ int		WixSaveFile (Widget, char*);
extern _CPP_ Boolean	WixInsertFile (Widget, char*);
extern _CPP_ Boolean	WixLoadFile (Widget, char*);
extern _CPP_ void		WixRegisterMessage (Widget, MessageProc);
extern _CPP_ void		WixSetSelection (Widget, XButtonEvent*, char*, int);
extern _CPP_ void		WixClearBuffer (Widget);
extern _CPP_ void		SelectBuffer (Widget);
extern _CPP_ void		DeselectBuffer (Widget);
extern _CPP_ void		WixInsert (Widget, char*);
extern _CPP_ Boolean	WixSetMark (Widget, void*);
extern _CPP_ Boolean	WixGotoMark (Widget, void*);
extern _CPP_ Boolean	WixDeleteMark (Widget, void*);
extern _CPP_ Boolean	WixCreateMark (Widget, void*);
extern _CPP_ void		WixSetInsertPosition (Widget, int);
extern _CPP_ int		WixGetInsertPosition (Widget);
extern _CPP_ int		WixGetHorScroll (Widget);
extern _CPP_ void		WixSetTabstops (Widget, int*);
extern _CPP_ void		WixGetTabstops (Widget, int*);
extern _CPP_ void		WixSetActionProc (XtActionProc (* ) ());

extern _CPP_ KeyWord*	CvtStringToKeyWordTable (char*, char*);
extern _CPP_ Trans*		CvtStringToKeyTable (char*);
extern _CPP_ Trans*		OverrideTranslation (Trans*, char*);
extern _CPP_ Trans*		CopyKeyTable (Trans*);

extern _CPP_ FontInfo*	GetFont (Display*, char*, Boolean);
extern _CPP_ alias*		CvtStringToAlias ();
extern _CPP_ Widget		XtCreateScrolledWix ();
extern _CPP_ Widget		XpaCreateScrolledWix (Widget, char*, Arg*, int);

/*
 *	Les noms des ressources.
 */

#ifndef	XtNlineNumberColor
#define	XtNlineNumberColor	"lineNumberColor"
#endif

#if 1
#ifndef	XtNbeginKeyword
#define	XtNbeginKeyword	"beginKeyword"
#endif

#ifndef	XtNendKeyword
#define	XtNendKeyword	"endKeyword"
#endif

#endif

#ifndef	XtNstartComment
#define	XtNstartComment	"startComment"
#endif

#ifndef	XtNstopComment
#define	XtNstopComment	"stopComment"
#endif

#ifndef	XtNlineComment
#define	XtNlineComment	"lineComment"
#endif

#ifndef XtNstringDelimitor
#define XtNstringDelimitor	"stringDelimitor"
#define XtCStringDelimitor	"StringDelimitor"
#endif

#ifndef	XtCComment
#define	XtCComment	"Comment"
#endif

#ifndef	XtNcolumnNumber
#define	XtNcolumnNumber		"columnNumber"
#define	XtCColumnNumber		"ColumnNumber"
#endif

#ifndef	XtNkeyTable
#define	XtNkeyTable		"keyTable"
#define	XtCKeyTable		"KeyTable"
#endif

#ifndef	XtNvalue
#define	XtNvalue		"value"
#define	XtCValue		"Value"
#endif

#ifndef XtNfullCursor
#define XtNfullCursor "fullCursor"
#define XtCFullCursor "FullCursor"
#endif

#ifndef	XtNnoCursor
#define	XtNnoCursor		"noCursor"
#define	XtCNoCursor		"NoCursor"
#endif

#ifndef	XtNblockCursor
#define	XtNblockCursor		"blockCursor"
#define	XtCBlockCursor		"BlockCursor"
#endif

#ifndef	XtNmotionCursor
#define	XtNmotionCursor		"motionCursor"
#endif

#ifndef	XtNhorizontalCursor
#define	XtNhorizontalCursor	"horizontalCursor"
#endif

#ifndef	XtNverticalCursor
#define	XtNverticalCursor	"verticalCursor"
#endif

#ifndef	XtNbinaryEdit
#define	XtNbinaryEdit		"binaryEdit"
#define	XtCBinaryEdit		"BinaryEdit"
#endif

#ifndef	XtNmonoLine
#define	XtNmonoLine		"monoLine"
#define	XtCMonoLine		"MonoLine"
#endif

#ifndef	XtNreadOnly
#define	XtNreadOnly		"readOnly"
#endif

#ifndef	XtCReadOnly
#define	XtCReadOnly		"ReadOnly"
#endif

#ifndef	XtNwrap
#define	XtNwrap			"wrap"
#define	XtCWrap			"Wrap"
#endif

#ifndef	XtNwrapColumn
#define	XtNwrapColumn		"wrapColumn"
#define	XtCWrapColumn		"WrapColumn"
#endif

#ifndef	XtNcurrentLineNumber
#define	XtNcurrentLineNumber	"currentLineNumber"
#endif

#ifndef	XtNbufferLineNumber
#define	XtNbufferLineNumber	"bufferLineNumber"
#endif

#ifndef	XtNfont
#define	XtNfont			"font"
#endif

#ifndef	XtNcommentFont
#define	XtNcommentFont		"commentFont"
#endif

#ifndef	XtNcommentColor
#define	XtNcommentColor		"commentColor"
#endif

#ifndef	XtNkeywordFont
#define	XtNkeywordFont		"keywordFont"
#endif

#ifndef	XtNkeywordColor
#define	XtNkeywordColor		"keywordColor"
#endif

#ifndef	XtNcppFont
#define	XtNcppFont		"cppFont"
#endif

#ifndef	XtNcppColor
#define	XtNcppColor		"cppColor"
#endif

#ifndef	XtNstringFont
#define	XtNstringFont		"stringFont"
#endif

#ifndef	XtNstringColor
#define	XtNstringColor		"stringColor"
#endif

#ifndef	XtNlibraryFont
#define	XtNlibraryFont		"libraryFont"
#endif

#ifndef	XtNlibraryColor
#define	XtNlibraryColor		"libraryColor"
#endif

#ifndef	XtNconstantFont
#define	XtNconstantFont		"constantFont"
#endif

#ifndef	XtNconstantColor
#define	XtNconstantColor	"constantColor"
#endif

#ifndef	XtNclassFont
#define	XtNclassFont		"classFont"
#endif

#ifndef	XtNclassColor
#define	XtNclassColor		"classColor"
#endif

#ifndef	XtNsliderMax
#define	XtNsliderMax		"sliderMax"
#define	XtCSliderMax		"SliderMax"
#endif

#ifndef	XtNverifyModifyCallback
#define	XtNverifyModifyCallback	"verifyModifyCallback"
#endif

#ifndef	XtNundoRedoCallback
#define	XtNundoRedoCallback	"undoRedoCallback"
#endif

#ifndef	XtNcontextCallback
#define	XtNcontextCallback	"contextCallback"
#endif

#ifndef	XtNbufferModifiedCallback
#define	XtNbufferModifiedCallback	"bufferModifiedCallback"
#endif

#ifndef	XtNaliasTable
#define	XtNaliasTable		"aliasTable"
#define	XtCAliasTable		"AliasTable"
#endif

#ifndef	XtNaliasBuffer
#define	XtNaliasBuffer		"aliasBuffer"
#define	XtCAliasBuffer		"AliasBuffer"
#endif

#ifndef	XtNaliasBufferLength
#define	XtNaliasBufferLength	"aliasBufferLength"
#define	XtCAliasBufferLength	"AliasBufferLength"
#endif

#ifndef	XtNhScrollPage
#define	XtNhScrollPage		"hScrollPage"
#define	XtCHScrollPage		"HScrollPage"
#endif

#ifndef XtNtransientSelection
#define	XtNtransientSelection	"transientSelection"
#define	XtCTransientSelection	"TransientSelection"
#endif

#ifndef	XtNbufferModified
#define	XtNbufferModified	"bufferModified"
#define	XtCBufferModified	"BufferModified"
#endif

#ifndef	XtCScroll
#define	XtCScroll		"Scroll"
#endif

#ifndef	XtNvScrollSize
#define	XtNvScrollSize		"vScrollSize"
#endif

#ifndef	XtNvScrollPosition
#define	XtNvScrollPosition	"vScrollPosition"
#endif

#ifndef	XtNvScrollSizeCallback
#define	XtNvScrollSizeCallback	"vScrollSizeCallback"
#endif

#ifndef	XtNvScrollPositionCallback
#define	XtNvScrollPositionCallback	"vScrollPositionCallback"
#endif

#ifndef	XtNhScrollSizeCallback
#define	XtNhScrollSizeCallback	"hScrollSizeCallback"
#endif

#ifndef	XtNhScrollPositionCallback
#define	XtNhScrollPositionCallback	"hScrollPositionCallback"
#endif

#ifndef	XtNinnerBorder
#define	XtNinnerBorder		"innerBorder"
#define	XtCInnerBorder		"InnerBorder"
#endif

#ifndef	XtNlineNum
#define	XtNlineNum		"lineNum"
#define	XtCLineNum		"LineNum"
#endif

#ifndef	XtNtabsize
#define	XtNtabsize		"tabsize"
#define	XtCTabsize		"Tabsize"
#endif

#ifndef	XtNactivateCallback
#define	XtNactivateCallback	"activateCallback"
#endif

#ifndef	XtNdefaultActionCallback
#define	XtNdefaultActionCallback	"defaultActionCallback"
#endif

#ifndef	XtNdelimitors
#define	XtNdelimitors		"delimitors"
#define	XtCDelimitors		"Delimitors"
#endif

#ifndef	XtNshowLineNumber
#define	XtNshowLineNumber	"showLineNumber"
#define	XtCShowLineNumber	"ShowLineNumber"
#endif

#ifndef	XtNlineNumberWidth
#define	XtNlineNumberWidth	"lineNumberWidth"
#define	XtCLineNumberWidth	"LineNumberWidth"
#endif

#ifndef	XtNkillStack
#define	XtNkillStack		"killStack"
#define	XtCKillStack		"KillStack"
#endif

#ifndef	XtNkeywordTable
#define	XtNkeywordTable	"keywordTable"
#define	XtCKeywordTable	"KeywordTable"
#define	XtRKeywordTable	"KeywordTable"
#endif

#ifndef	XtNcaseSensitive
#define	XtNcaseSensitive	"caseSensitive"
#define	XtCCaseSensitive	"CaseSensitive"
#endif

#ifndef	XtNselectColor
#define	XtNselectColor		"selectColor"
#endif

#ifndef XtNframeThickness
#define	XtNframeThickness	"frameThickness"
#define	XtNframeColor		"frameColor"
#endif

#ifdef STANDALONE
#define	XtNshadowThickness      "shadowThickness"
#define	XtCShadowThickness      "ShadowThickness"

#define	XtNtopShadowColor       "topShadowColor"
#define	XtNbottomShadowColor    "bottomShadowColor"

#define	XtNuserData             "userData"
#define	XtCUserData             "UserData"
#endif

#endif	/* WIX_H */
