/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/keysym.h>

#include "Primitive.h"
#include "PrimitiveP.h"

#define	_PRIVATE_
#include "Wix.h"
#include "WixP.h"

/*
 *	Une action generique se deroule de la facon suivante:
 *
 *		1. Traiter l'argument numerique
 *		2. Sauvegarder la commande (pour les macros)
 *		3. Appeler la fonction implementant la commande
 *		4. Reafficher la fenetre.
 */

int			num_arg = 1;

static Command	PreviousCommand;
static Command*	PCommandPtr = &PreviousCommand;
Command			DefaultMacroBuffer [128];
Time			_SelectTime, _KillTime;

static int	change_font_region (WixWidget, int, void*);

/*
 *	Enregistre la fonction d'affichage des messages.
 */

void WixRegisterMessage (Widget widget, MessageProc fun)
{
	WixWidget	w = (WixWidget) widget;

	W.messageFunction = fun;
}

void stop_macro_learn (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	PCommandPtr -> function = 0;	/* pour arreter l'execution */
	PCommandPtr = &PreviousCommand;	/* buffer fixe pour RepeatCommand */
	W.macro_learn = False;			/* on a appris */
}

void start_macro_learn (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	PCommandPtr = DefaultMacroBuffer;	/* viser le buffer */
	W.macro_learn = True;				/* aller  l'cole */
}

/*
 *	Engrange la commande pour les macros.
 */

static void overrideCommand (WixWidget w, CommandProc cmd, void* arg, Boolean sv)
{
	PCommandPtr -> function = cmd;	/* engranger la commande */
	PCommandPtr -> count = num_arg;	/* en cas de repetition */
	PCommandPtr -> arg = (sv && W.macro_learn) ? XtNewString (arg) : arg;
}

void SaveCommand (Widget widget, CommandProc cmd, void* arg, Boolean sv)
{
	WixWidget	w = (WixWidget) widget;

	overrideCommand (w, cmd, arg, sv);
	if (W.macro_learn)
		PCommandPtr++;
}

/*
 *	Execute une macro.
 */

int ExecuteMacro (WixWidget w, int count, void* args)
{
	Command*	mb = (Command*) args;
	Command*	p;

	if (W.macro_learn) {
		XBell (XtDisplay (w), 100);
		return 0;
	}
	PCommandPtr -> function = ExecuteMacro;
	PCommandPtr -> count = count;
	PCommandPtr -> arg = (void*) mb;

	while (--count >= 0)
		for (p = mb; p -> function; p++) {
			if (!(*p -> function) (w, p -> count, p -> arg))
				return 0;
			if (W.need_hpos) {
				GetHpos (w);
				W.cur_col = W.cur_x;
				W.need_hpos = False;
			}
		}
	return 1;
}

static int process_action (WixWidget w, CommandProc cmd, void* arg)
{
	int	status;

	PCommandPtr -> function = cmd;		/* engranger la commande */
	PCommandPtr -> count = num_arg;		/* en cas de repetition */
	PCommandPtr -> arg = arg;

	if (W.fresh_selection) {
		if (cmd == WipeRegion && W.b_marks [2].addr != W.b_marks [1].addr) {
			WipeSelection (w);		/* dgage la slection */
			return 1;
		}
		if (cmd == InsertChar && W.b_marks [2].addr != W.b_marks [1].addr)
			WipeSelection (w);		/* dgage la slection */
		else if (W.transient_selection)
			DeselectBuffer ((Widget) w);
		W.fresh_selection = False;			/* dfrachit la rgion */
	}

	status = (*cmd) (w, num_arg, arg);
	if (W.macro_learn) {
		if (!status)
			stop_macro_learn ((Widget) w);
		else
			PCommandPtr++;
	}
	num_arg = 1;			/* remettre le compte a un */
	W.hscroll = False;		/* montrer le curseur */
	W.focus_in = True;		/* en plein */
	WixUnhighlightAll (w);	/* effacer la prcdente chaine cherche */
	UpdateWindow (w);
	return status;
}


void exec_macro (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, ExecuteMacro, DefaultMacroBuffer);
	if (W.messageFunction)
		W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_EXEC_MACRO);
}

void repeat_command (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	if (W.messageFunction)
		W.messageFunction (widget, T_OK, 0);

	if (PCommandPtr -> function) {
		W.fresh_selection = False;	/* invalide la region */
		(*PCommandPtr -> function) (w, PCommandPtr -> count, PCommandPtr -> arg);
		UpdateWindow (w);
	}
}

/*
 *	Insere un caractere soit tel quel, soit a
 *	partir de son code octal.
 */

void insert_octal (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	char			c;
	KeySym			ks;
	int				octal = 0;
	int				ndigit = 3;
	void			rubout_char ();
	XtAppContext	app = XtWidgetToApplicationContext (widget);

	StopBlink ();
	if (W.messageFunction)
		W.messageFunction (widget, T_OCTAL, octal);
	for (;;) {
		XtAppPeekEvent (app, event);
		switch (event -> type) {
		case KeyPress:
			XtAppNextEvent (app, event);
			if (XLookupString ((XKeyEvent*)event, &c, 1, &ks, 0)) {
				if (c <= '7' && c >= '0') {
					octal = octal * 8 + c - '0';
					if (W.messageFunction)
						W.messageFunction (widget, T_OCTAL, octal);
					if (--ndigit == 0) {
						process_action (w, InsertChar, (void*) (long) octal);
						break;
					}
					continue;
				} else if (c == '\b') {
					if (++ndigit > 3)
						break;
					octal = octal / 8;
					if (W.messageFunction)
						W.messageFunction (widget, T_OCTAL, octal);
					continue;
				} else if (octal == 0) {
					process_action (w, InsertChar, (void*) (long) c);
					break;
				} else if (W.messageFunction) {
						W.messageFunction (widget, T_ERROR, E_BAD_OCTAL);
						break;
				}
			}
			continue;
		case KeyRelease:
			XtAppNextEvent (app, event);
			continue;

		case LeaveNotify:
		case EnterNotify:
			break;

		default:
			XtAppNextEvent (app, event);
			XtDispatchEvent (event);
			continue;
		}
		StartBlink (w);
		return;
	}
}

void multiplier (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	char		c, ch;
	KeySym		ks;
	XtAppContext	app = XtWidgetToApplicationContext (widget);

	StopBlink ();
	if (XLookupString ((XKeyEvent*)event, &c, 1, &ks, 0)) {
		if (c <= '9' && c >= '1')
			num_arg = c - '0';
		else
			num_arg = 4;
	}
	ch = c;
	if (W.messageFunction)
		W.messageFunction (widget, T_ARGCNT, num_arg);
	for (;;) {
		XtAppPeekEvent (app, event);
		switch (event -> type) {
		case KeyPress:
			if (XLookupString ((XKeyEvent*)event, &c, 1, &ks, 0)) {
				if (c <= '9' && c >= '0') {
					num_arg = num_arg * 10 + c - '0';
					if (W.messageFunction)
						W.messageFunction (widget, T_ARGCNT, num_arg);
					XtAppNextEvent (app, event);
					continue;
				} else if (c == '\b') {
					num_arg /= 10;
					if (num_arg <= 0)
						num_arg = 1;
					if (W.messageFunction)
						W.messageFunction (widget, T_ARGCNT, num_arg);
					XtAppNextEvent (app, event);
					continue;
				} else if (c == ch) {
					num_arg *= 4;
					if (W.messageFunction)
						W.messageFunction (widget, T_ARGCNT, num_arg);
					XtAppNextEvent (app, event);
					continue;
				}
			}
			break;
		case KeyRelease:
			XtAppNextEvent (app, event);
			continue;

		case LeaveNotify:
		case EnterNotify:
			break;

		default:
			XtAppNextEvent (app, event);
			XtDispatchEvent (event);
			continue;
		}
		StartBlink (w);
		return;
	}
}

void yank (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	/*
	 *	Inserer le plus recent entre la selection (locale) et
	 *	la destruction (ou copie de la region).
	 */
	if (_SelectTime >= _KillTime && num_arg == 1 && !W.macro_learn) {
		stuff (widget, event, params, num_params);
		return;
	}

	process_action (w, Yank, NULL);
	if (W.messageFunction)
		W.messageFunction (widget, T_OK, 0);
}

void twiddle_chars (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, TwiddleChars, NULL);
	if (W.messageFunction)
		W.messageFunction (widget, T_OK, 0);
}

void rubout_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, RuboutWord, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
}

void lower_case_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, LowerCaseWord, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void upper_case_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, UpperCaseWord, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}


void capitalize_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, CapitalizeWord, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void delete_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, DeleteWord, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
}

void backward_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, BackwardWord, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_BEG_BUFFER);
}

void end_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, ForwardWord, 0 /* pour aller  la fin du mot */);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_END_BUFFER);
}

void forward_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, ForwardWord, (void*) 1 /* pour aller au dbut du mot suivant */);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_END_BUFFER);
}

void flasher (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	char 		c;
	KeySym		ks;
	int			cmd_status;

	if (XLookupString ((XKeyEvent*) event, &c, 1, &ks, 0)) {
		cmd_status = process_action (w, Flasher, (void*) (long) c);
		if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
	}
}

void match_opening_brace (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, FindOpeningBrace, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_OPENING);
}

void match_closing_brace (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, FindClosingBrace, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_CLOSING);
}

void match_opening_par (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, FindOpeningParenthesis, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_OPENING);
}

void match_closing_par (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, FindClosingParenthesis, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_CLOSING);
}

void backward_char (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, BackwardChar, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_BEG_BUFFER);
}

void backward_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, BackwardLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_BEG_BUFFER);
}

void backward_page (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, BackwardPage, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_BEG_BUFFER);
}

void beginning_of_buffer (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, BeginningOfBuffer, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void beginning_of_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, BeginningOfLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void copy_end_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, CopyEndLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_COPIED, cmd_status);
}

void copy_region (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, CopyRegion, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_COPIED, cmd_status);
}

void wipe_region (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, WipeRegion, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
}

void curline_top (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, CurlineTop, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void curline_center (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, CurlineCenter, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void delete_char (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	int			cmd_status;

	if (*num_params == 1)
		num_arg *= atoi (params [0]);

	cmd_status = process_action (w, DeleteChar, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
}

void delete_beg_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, DeleteBeginningOfLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
}

void delete_end_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, DeleteEndLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
}

void delete_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, DeleteLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
}

void end_of_buffer (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, EndOfBuffer, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void end_of_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, EndOfLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void exchange_point_mark (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, ExchangePointMark, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void forward_char (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, ForwardChar, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_END_BUFFER);
}

void forward_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, ForwardLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_END_BUFFER);
}

void forward_page (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	int	cmd_status = process_action (w, ForwardPage, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_END_BUFFER);
}

void goto_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	(void) process_action (w, GotoLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

/*
 *	si "chaine1" est a gauche du curseur, la remplace par "chaine2".
 */

void replace_string (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	char		c;
	KeySym		ks;
	int			n = *num_params;
	char*		p;
	int			len;
	int			i;
	int			h;

	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);

	if (n < 2)
		return;

	for (i = 0; i < n; i += 2) {
		len = strlen (params [i]);	/* longueur de la chane  remplacer */
		if (strncmp (params [i], W.b_cursor_b - len, len))
			continue;		/* pas trouve */

		h = W.cur_line -> height;	/* hauteur actuelle de la ligne */
		KBackwardChar (w, len);		/* dtruire  gauche du curseur */
		KDeleteChar (w, len, 0);	/* ne pas sauver dans la kill stack */
		for (p = params [i+1]; *p; p++)
			KInsertChar (w, *p);	/* insrer la nouvelle */
		*W.b_cursor_b = 0;
		RemoveCursor (w);
		if (W.refresh_from > W.cur_y)
			W.refresh_from = W.cur_y;

		if (h != W.cur_line -> height)	/* la hauteur a-t-elle change ? */
			W.refresh_to = 1000;		/* oui: rafficher le bas de l'cran */

		if (W.refresh_to < W.cur_y)
			W.refresh_to = W.cur_y;
		W.need_hpos = True;
		UpdateWindow (w);
		return;
	}
	if (XLookupString ((XKeyEvent*) event, &c, 1, &ks, 0))
		process_action (w, InsertChar, (void*) (long) c);
}

void insert_char (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	int			n;
	char		c;
	KeySym		ks;

	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
	if ((n = *num_params) > 0) {
		while (--n >= 0)
			WixInsert (widget, *params++);
		num_arg = 1;			/* remettre le compte a un */
		W.fresh_selection = False;	/* invalide la region */
		UpdateWindow (w);
		W.hscroll = False;		/* scroller pour montrer le curseur */
		return;
	}
	if (XLookupString ((XKeyEvent*) event, &c, 1, &ks, 0))
		process_action (w, InsertChar, (void*) (long) c);
}

void expand_alias (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
	(void) process_action (w, ExpandAlias, event);
}

void expand_from_alias (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
	process_action (w, ExpandFromAlias, event);
}

void insert_line_number (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, InsertLineNumber, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void new_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, NewLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void new_line_indent (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, NewLineIndent, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void new_line_indent_c (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, NewLineIndentC, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void c_begin_function_action (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget w = (WixWidget) widget;

	process_action (w, CBeginningOfFunction, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void c_end_function_action (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget w = (WixWidget) widget;

	process_action (w, CEndOfFunction, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void c_indent_line_action (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget w = (WixWidget) widget;

	process_action (w, CIndentLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void open_space (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, OpenSpace, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void open_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, OpenLine, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void pop_kill_stack (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	process_action (w, PopKillStack, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void rubout_char (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	if (W.fresh_selection && W.b_marks [1].addr != W.b_marks [2].addr)
		WipeSelection (w);
	else {
		int	cmd_status = process_action (w, RuboutChar, NULL);
		if (W.messageFunction) W.messageFunction (widget, T_REMOVED, cmd_status);
	}
}

void scroll_down (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	int			cmd_status;

	if (*num_params > 0)
		num_arg = atoi (params [0]);
	cmd_status = process_action (w, ScrollDown, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_BEG_BUFFER);
}

void scroll_left (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	W.hscroll = True;
	(void) process_action (w, ScrollLeft, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void scroll_right (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	W.hscroll = True;
	(void) process_action (w, ScrollRight, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

void scroll_up (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;
	int			cmd_status;

	if (*num_params > 0)
		num_arg = atoi (params [0]);

	cmd_status = process_action (w, ScrollUp, NULL);
	if (W.messageFunction) W.messageFunction (widget, cmd_status ? T_OK : T_ERROR, E_END_BUFFER);
}

void set_mark (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	(void) process_action (w, SetMark, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_ERROR, E_SET_MARK);
}

void insert_tabulation (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	(void) process_action (w, Tabulation, NULL);
	if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
}

/*
 *	Action pour commuter l'affichage des numros de ligne.
 */

void switch_line_numbering (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	W.show_line_number = !W.show_line_number;
	W.line_number_width = W.show_line_number /* 6 chiffres de la fonte 0 */
			? 6 * Width ('x', FONT (0)) + 2 + W.margin
			: 0;
	/* Rafficher la fentre */
	XClearArea (XtDisplay (widget), XtWindow (widget), 0, 0, 0, 0, True);
}

/*
 *	Tout pour le clavier et les clefs composees.
 */

#define findSpace(cp)	while (*cp != ' ' && *cp != '\t') cp ++
#define skipSpace(cp)	while (*cp == ' ' || *cp == '\t') cp ++
#define endLine(cp)	while (*cp != '\n') cp++;

/*
 *	Commande indefinie
 */

static void undefinedAction (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	if (W.messageFunction)
		W.messageFunction (widget, T_ERROR, E_UNDEF_CMD);
}

/*
 *	Traitement des clefs multi touches.
 */

/*
 *	^X^R	read-file\n\
 *	^S	forward-search ...
 */

static char* nextChar (char* string, int* pch)
{
	int	mask = ~0;

	switch (*string) {
		case '^':
			mask = 037;
			string++;
			break;
		case '$':
			switch (*++string) {
				case 'U': *pch = K_UP; break;
				case 'D': *pch = K_DOWN; break;
				case 'L': *pch = K_LEFT; break;
				case 'R': *pch = K_RIGHT; break;
				case 'E': *pch = 033; break;
			}
			return ++string;
	}
	if (*string == '\\') string++;
	*pch = (*string == '?' && mask == 037)
		? 0177
		: *string & mask;
	return ++string;
}

static void delete_keytable (Trans* tt)
{
	int	i;

	/* attention: ne recupere pas les chaines d'argument d'actions */
	for (i = 0; i < K_NUMBER; i++)
		if (tt [i].prefix)
			delete_keytable (tt [i].prefix);
	XtFree ((char*)tt);
}

static Trans* stringToTrans (Trans* tt, char* string)
{
	int	ch;

	for (;;) {
		string = nextChar (string, &ch);
		if (*string) {
			/* c'est un prfixe */
			if (! tt [ch].prefix) {
				tt [ch].prefix = (Trans*) XtMalloc (K_NUMBER * sizeof (Trans));
				bzero ((char*)tt [ch].prefix, K_NUMBER * sizeof (Trans));
			}
			tt = tt [ch].prefix;
			continue;
		}
		if (tt [ch].prefix) {
			/* c'etait un prefix alors tant pis pour lui */
			delete_keytable (tt [ch].prefix);
			tt [ch].prefix = NULL;
		}
		return tt + ch;
	}
}

/*
 *	l'action liee a toutes les clefs:
 *	<Key>:	keyboard-action ()
 */

void keyboard_action (Widget widget, XEvent* event, String* s, Cardinal* ns)
{
	WixWidget	w = (WixWidget) widget;
	int			ch;
	char		c;
	KeySym		ks;
	Boolean		metaflag = False;
	XKeyEvent*	ev = (XKeyEvent*) event;

	if (ev -> state & Mod1Mask) {
		metaflag = True;			/* on a vu un meta */
		ev -> state &= ~Mod1Mask;	/* on le planque! */
	}
	/* attraper le prochain caractere */
	if (XLookupString (ev, &c, 1, &ks, 0))
		ch = c & 0177;
	else switch (ks) {
		default:	return;
		case XK_Left:	ch = K_LEFT; break;
		case XK_Right:	ch = K_RIGHT; break;
		case XK_Up:	ch = K_UP; break;
		case XK_Down:	ch = K_DOWN; break;
	}
	/* mapper Meta X en Esc-X */
	if (metaflag && W.prev_trans [033].prefix)		/* Meta -> Esc- */
		W.prev_trans = W.prev_trans [033].prefix;

	if (W.prev_trans [ch].prefix)		/* c'est un prfixe */
		W.prev_trans = W.prev_trans [ch].prefix;
	else {
		if (W.prev_trans [ch].proc) {
			/* appeller l'action */
			static Cardinal	one = 1;

			PushCommand (w);			/* empiler la commande */
			if (W.prev_trans [ch].arg)
				(*W.prev_trans [ch].proc) (widget, event, &W.prev_trans [ch].arg, &one);
			else
				(*W.prev_trans [ch].proc) (widget, event, s, ns);
		} else
			undefinedAction (widget, event, s, ns);
		W.prev_trans = W.trans_table;	/* remettre l'automate */
										/* dans l'etat initial */
	}
}

Trans* OverrideTranslation (Trans* tt, char* str)
{
	extern WixClassRec	wixClassRec;
	XtActionProc	proc;
	Trans*		tr;
	char*		skey;
	char*		act_name;
	char*		act_arg = NULL;
	char		ktab [10240];
	int		i;

	for (i = 0; i < (int) sizeof (ktab) - 1; i++)
		if (! (ktab [i] = str [i])) {
			if (ktab [i-1] != '\n')
				ktab [i++] = '\n';
			break;
		}
	ktab [i] = 0;

	for (str = ktab; *str;) {
		skipSpace (str);	/* sauter les espaces initiaux */
		skey = str; findSpace (str); *str ++ = 0;
		skipSpace (str);
		act_name = str;		/* le nom de l'action */
		for (;;str++) {
			switch (*str) {
			case '\n':
				act_arg = NULL;
				break;
			case ' ':
				*str++ = 0;
				continue;
			case '(':
				*str++ = 0;	/* terminer le nom */
				act_arg = str;	/* debut d'argument */
				continue;
			case ')':
				*str++ = 0;	/* fin d'argument */
				while (*str != '\n') str++;
				break;
			default:
				continue;
			}
			break;
		}
		*str++ = 0;
		if (! (proc = (*wixClassRec.wix_class.get_action_proc) (act_name)))
			fprintf (stderr, "Unknown action: %s for key: %s\n", act_name, skey);
		else {
			tr = stringToTrans (tt, skey);
			tr -> proc = proc;
			tr -> arg = act_arg ? XtNewString (act_arg) : NULL;
		}
	}
	return tt;
}

Trans* CvtStringToKeyTable (char* str)
{
	int		i;
	Trans*	tt;

	tt = (Trans*) XtMalloc (K_NUMBER * sizeof (Trans));
	bzero ((char*)tt, K_NUMBER * sizeof (Trans));
	for (i = ' '; i <= '~'; i++)
		tt [i].proc = insert_char;	/* le prfixe est nullifie */
						/* par le bzero precedent */
	return OverrideTranslation (tt, str);
}

Trans* CopyKeyTable (Trans* tt)
{
	int		i;
	Trans*	r;

	r = (Trans*) XtMalloc (K_NUMBER * sizeof (Trans));
	bcopy ((char*)tt, (char*)r, K_NUMBER * sizeof (Trans));
	for (i = 0; i < K_NUMBER; i++)
		if (r [i].prefix)
			r [i].prefix = CopyKeyTable (r [i].prefix);
	return r;
}

/*
 *	Tout pour les aliases.
 */

#define	isalpha(c) (((c)>='A' && (c)<='Z') || ((c)>='a' && (c)<='z') || ((c)>='0' && (c)<='9') || (c)=='_')

/*
 *	Insre une ligne et une indentation.
 */

void nlws (WixWidget w, int indent, unsigned char** pp)
{
	int	n;

	if (pp) *(*pp)++ = '\n';
	KInsertChar (w, '\n');			/* aller a la ligne */
	for (n = indent / W.tab_equiv; --n >= 0;) {
		if (pp) *(*pp)++ = '\t';
		KInsertChar (w, '\t');		/* d'abord les tabs */
	}
	for (n = indent % W.tab_equiv; --n >= 0;) {
		if (pp) *(*pp)++ = ' ';
		KInsertChar (w, ' ');		/* puis les espaces */
	}
}

/*
 *	Cherche la string dans la table d'alias.
 */

static alias *search (alias* ap, unsigned char* string, int len, Boolean prefix_only)
{
	int	l;

	for (; ap && ap -> a_src; ap ++)
		if ((prefix_only || len == strlen (ap -> a_src)) && !bcmp (string, ap -> a_src, len))
		{
			l = strlen (ap -> a_dst);
			if (l == len && !bcmp (string, ap -> a_dst, l))
				return 0;	/* dj expans */
			else
				return ap;
		}
	return 0;	/* pas trouv */
}

/*
 *	Retourne la valeur de l'indentation de la ligne courante.
 */

int get_indent (WixWidget w)
{
	unsigned char*	cp;
	unsigned char*	end;
	int				col;

	for (end = cp = W.b_cursor_b; *--cp != '\n';);
	for (cp++, col = 0; cp < end; cp++)
		if (*cp == '\t')
			col += W.tab_equiv - (col % W.tab_equiv);
		else if (*cp == ' ')
			col += 1;
		else
			break;
	return col;
}

/*
 *	Attend autre chose que le caractre de l'vnement pass en argument.
 */

static Boolean wait_for_char (WixWidget w, XEvent* event)
{
	XEvent			e;
	XtAppContext	app = XtWidgetToApplicationContext ((Widget) w);

	StopBlink ();
	for (;;) {
	    switch (XtAppPeekEvent (app, &e), e.type) {
		case KeyPress:
			if (e.xkey.keycode == ((XKeyEvent*) event) -> keycode &&
			    e.xkey.state == ((XKeyEvent*) event) -> state) {
				XtAppNextEvent (app, &e);
				StartBlink (w);
				return True;
			}
		case ButtonPress:
		case ButtonRelease:
		case MotionNotify:
			StartBlink (w);
			return False;
		case KeyRelease:
			XtAppNextEvent (app, &e);
			continue;
		default:
			XtAppNextEvent (app, &e);
			XtDispatchEvent (&e);
			continue;
	    }
	}
}

static unsigned char* insert_spaces (WixWidget w, int indent, unsigned char* p)
{
	int	n;

	*p++ = '\n';			/* aller a la ligne */
	for (n = indent / W.tab_equiv; --n >= 0;)
		*p++ = '\t';		/* insrer d'abord des tabulations */
	for (n = indent % W.tab_equiv; --n >= 0;)
		*p++ = ' ';			/* puis des espaces */
	return p;
}

/*
 *	Tente une expansion  l'aide de la table d'alias du widget.
 */

static int _expand_alias (WixWidget w, XEvent* event, Boolean do_loop, Boolean prefix_only)
{
	alias*			ap;
	unsigned char*	cp;
	unsigned char*	mk;
	unsigned char*	sv;
	int				indent;
	int				n;
	int				rfrom, rto;
	int				to_delete;
	int				to_moveback;
	unsigned char	prefixe [256];

	int				current_pos;
	unsigned char	expanded [1024];
	unsigned char*	ep;

	if (W.read_only)
		return 0;

	if (!(ap = W.alias_table))		/* pas de table d'alias */
		return 0;
	/*
	 *	Reculer jusqu'au precedent delimiteur.
	 */
	for (cp = W.b_cursor_b, *cp = 0; ! index (W.delim, *--cp); );
	cp ++;
	if (cp == W.b_cursor_b)
		return 0;

	n = W.b_cursor_b - cp;	/* longueur du prfixe */
	bcopy (cp, prefixe, n+1);	/* le prfixe lui meme */

	if (!(ap = search (ap, cp, n, prefix_only)))
		return 0;

	W.do_undo = False;		/* dsactiver l'undo/redo */
	KBackwardChar (w, n);
	KDeleteChar (w, n, 0);	/* ne pas sauver dans la kill stack */
	sv = W.b_cursor_b;
	indent = get_indent (w);

	current_pos = W.b_cursor_b - W.beginning;
retry:
	mk = NULL;
	ep = expanded;

	RemoveCursor (w);
	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;

	for (cp = ap -> a_dst; *cp; cp ++) {
		if (*cp == '\\')
			switch (*++cp) {
			default:
				*ep++ = *cp;
				continue;

			case '*':
				mk = W.b_cursor_b + (ep - expanded);
				continue;

			case 'n':
				*ep++ = '\n';
				continue;

			case 't':
				*ep++ = '\t';
				continue;

			case 'i':
				ep = insert_spaces (w, indent, ep);
				continue;

			case '+':
				ep = insert_spaces (w, indent += W.tab_equiv, ep);
				continue;

			case '-':
				ep = insert_spaces (w, indent -= W.tab_equiv, ep);
				continue;
			}
		else {
			*ep++ = *cp;
		}
	}
	if (ep > expanded)
		KInsertString (w, expanded, ep - expanded);
	to_delete = W.b_cursor_b - sv;
	if (mk != NULL)
		KBackwardChar (w, W.b_cursor_b - mk);
	to_moveback = W.b_cursor_b - sv;

	rfrom = W.refresh_from;
	rto = W.refresh_to;
	W.need_hpos = True;
	W.is_modified = True;
	UpdateWindow (w);
	if (!do_loop || !wait_for_char (w, event)) {
		W.do_undo = True;
		if (ep > expanded) {
			PushDelete (w, current_pos, prefixe, n);
			PushInsert (w, current_pos, expanded, ep - expanded);
		}
		return 1;
	}

	KBackwardChar (w, to_moveback);
	KDeleteChar (w, to_delete, 0);
	W.refresh_from = rfrom;
	W.refresh_to = rto;
	W.need_hpos = True;
	W.is_modified = True;
	UpdateWindow (w);
	if (! (ap+1) -> a_src || !(ap = search (ap+1, prefixe, n, prefix_only))) {
		KInsertString (w, prefixe, n);
		W.refresh_from = rfrom;
		W.refresh_to = rfrom;
		W.need_hpos = True;
		W.is_modified = True;
		UpdateWindow (w);
		W.do_undo = True;
		return -1;	/* yen a plus mais yen a eu */
	}
	goto retry;
}

/*
 *	Tente qu'une expansion depuis la table d'alias.
 */

int ExpandFromAlias (WixWidget w, int count, void* arg)
{
	XEvent*	event = (XEvent*) arg;
	char	c;

	(void) _expand_alias (w, event, False, False);
	if (XLookupString ((XKeyEvent*) event, &c, 1, 0, 0))
		InsertChar (w, count, (void*) (long) c);
	return 1;
}

/*
 *	Tente une expansion  l'aide du contenu du buffer.
 */

extern unsigned char*	_WixBackwardSearch ();

static int expand_from_buffer (WixWidget w, XEvent* event)
{
	unsigned char*	cp;
	int				maxloop;
	unsigned char	prefixe [256];
	unsigned char*	cp_beg;
	unsigned char*	sav_cp;
	unsigned char*	sav_beg;
	int				n, size;
	int				rfrom;
	Boolean			ok = False;
	int				current_pos;
	unsigned char	expanded [1024];
	unsigned char*	ep;

	if (W.read_only || W.b_cursor_b == W.beginning)
		return 0;
	/*
	 *	Reculer jusqu'au precedent delimiteur.
	 */
	cp = W.b_cursor_b;
	for (*cp = 0; ! index (W.delim, *--cp); );
	cp ++;
	if (cp == W.b_cursor_b)
		return 0;

	n = W.b_cursor_b - cp;		/* longueur du prfixe */
	bcopy (cp, prefixe, n+1);	/* le prfixe lui meme */
	/*
	 *	Chercher avant le curseur, sinon a partir de la fin.
	 */
	cp_beg = cp;	/* pour ne pas compter le prfixe lui meme */
	for (;;) {
		if (! (cp = _WixBackwardSearch (w, cp - 1, prefixe, 1)) &&
		    ! (cp = _WixBackwardSearch (w, W.end, prefixe, 1)))
			return 0;
		if (index (W.delim, cp [-1]))
			break;
	}
	if (cp == cp_beg)
		return 0;	/* on n'a trouve QUE le prfixe */

	RemoveCursor (w);
	rfrom = W.cur_y;
	cp_beg = cp;	/* pour ne faire qu'un tour */
	W.do_undo = False;		/* dsactiver l'undo/redo */
	current_pos = W.b_cursor_b - W.beginning;
	for (;;) {
		int	h;

		sav_cp = cp;
		sav_beg = W.b_cursor_b;
		h = W.cur_line -> height;
		for (ep = expanded, cp += n; !index (W.delim, (*cp)); cp++) {
			*ep++ = *cp;
			KInsertChar (w, *cp);
		}
		*ep = 0;
		*W.b_cursor_b = 0;

		if ((size = cp - sav_cp - n) != 0) {
			if (W.refresh_from > rfrom) W.refresh_from = rfrom;
			calc_height (w, W.cur_line);
			if (h != W.cur_line -> height)
				W.refresh_to = 1000;
			else if (W.refresh_to < rfrom)   W.refresh_to = rfrom;
		}
		UpdateWindow (w);
		if (!wait_for_char (w, event)) {
			W.do_undo = True;
			if (ep > expanded)
				PushInsert (w, current_pos, expanded, ep - expanded);
			return 1;
		}

		ok = False;
		cp = sav_cp;
		for (maxloop = 1000; --maxloop >=0;) {
			if (! (cp = _WixBackwardSearch (w, cp - 1, prefixe, 1)) &&
			    ! (cp = _WixBackwardSearch (w, W.end, prefixe, 1)))
			{
				ok = True;
				break;
			}
			if (cp == sav_beg || cp == sav_cp || cp == cp_beg) {
				ok = True;
				break;
			}
			if (index (W.delim, cp [-1])) {
				if (size == 0 || (!bcmp (cp+n, sav_beg, size) && index (W.delim, cp [n+size])))
					continue;
				break;
			}
		}
		if (maxloop < 0) {
			fprintf (stderr, "Warning: loop in expand_from_buffer.\n");
			W.do_undo = True;
			return -1;
		}
		if (size > 0) {
			h = W.cur_line -> height;	/* hauteur actuelle de la ligne */
			KBackwardChar (w, size);
			KDeleteChar (w, size, 0);
			*W.b_cursor_b = 0;
			if (W.refresh_from > rfrom) W.refresh_from = rfrom;
			calc_height (w, W.cur_line);	/* nouvelle hauteur de ligne */
			if (h != W.cur_line -> height)
				W.refresh_to = 1000;
			else if (W.refresh_to < rfrom)   W.refresh_to = rfrom;
			W.need_hpos = True;
			UpdateWindow (w);
		}
		if (ok) {
			W.do_undo = True;
			return -1;
		}
	}
}

/*
 *	Recherche une chane en avant.
 *	algorithme de Boyer & Moore.
 */

static unsigned char* forward_search (
	unsigned char*	text,
	unsigned char*	text_end,
	unsigned char*	string,
	int				len	/* strlen (string) - 1 */)
{
	int						dec, n;
	unsigned char*			s;
	static unsigned char	pattern [256];
	static short			skip [256] = {
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	};

	if (strcmp (pattern, string)) {
		for (s = pattern; *s; )
			skip [*s++] = -1;
		for (s = string; *s; s++)
			skip [*s] = len - (s - string);
		strcpy (pattern, string);
	}

	s = string + len;
	text += len;
	n = 1;
	while (text < text_end && s >= string) {
		if (*text == *s) {
			text --;
			s --;
			n ++;
		} else {
			if ((dec = skip [*text]) == -1)
				dec = len+1;
			else if (dec < n)
				dec = n;
			text += dec;
			s = string + len;
			n = 1;
		}
	}
	if (s >= string)
		return 0;	/* echec */

	return text+1;	/* se mettre au debut du pattern */
}

static int expand_from_file (WixWidget w, XEvent* event)
{
	unsigned char*	cp;
	unsigned char*	text_end;
	int				maxloop;
	unsigned char	prefixe [256];
	unsigned char*	sav_cp;
	unsigned char*	sav_beg;
	int				n, size;
	int				rfrom;
	int				filelen;
	Boolean			ok = False;

	int				current_pos;
	unsigned char	expanded [1024];
	unsigned char*	ep;

	if (W.read_only || W.b_cursor_b == W.beginning || W.alias_buffer_length == 0)
		return 0;
	/*
	 *	Reculer jusqu'au prcdent dlimiteur.
	 */
	cp = W.b_cursor_b;
	for (*cp = 0; ! index (W.delim, *--cp); );
	cp ++;
	if (cp == W.b_cursor_b)
		return 0;

	n = W.b_cursor_b - cp;		/* longueur du prfixe */
	bcopy (cp, prefixe, n+1);	/* le prfixe lui meme */
	/*
	 *	Chercher dans le fichier.
	 */
	cp = W.alias_buffer;
	filelen = W.alias_buffer_length;
	text_end = cp + filelen - 1;
	W.do_undo = False;		/* dsactiver l'undo/redo */
	current_pos = W.b_cursor_b - W.beginning;
	for (;;) {
		cp = forward_search (cp+1, text_end, prefixe, n-1);
		if (!cp)
			return 0;
		if (!isalpha (cp [-1]))	/* delimiteurs basiques */
			break;
	}
	RemoveCursor (w);
	rfrom = W.cur_y;
	for (;;) {
		sav_cp = cp;
		sav_beg = W.b_cursor_b;
		for (ep = expanded, cp += n; isalpha (*cp); cp++) {	/* delimiteurs basiques */
			*ep++ = *cp;
			KInsertChar (w, *cp);
		}
		*ep = 0;
		*W.b_cursor_b = 0;

		if ((size = cp - sav_cp - n) != 0) {
			if (W.refresh_from > rfrom) W.refresh_from = rfrom;
			if (W.refresh_to < rfrom)   W.refresh_to = rfrom;
		}
		UpdateWindow (w);
		if (!wait_for_char (w, event)) {
			W.do_undo = True;
			if (ep > expanded)
				PushInsert (w, current_pos, expanded, ep - expanded);
			return 1;
		}

		ok = False;
		cp = sav_cp;
		for (maxloop = 1000; --maxloop >=0;) {
			if (! (cp = forward_search (cp+1, text_end, prefixe, n-1)))
			{
				ok = True;
				break;
			}
			if (cp == sav_beg || cp == sav_cp) {
				ok = True;
				break;
			}
			if (!isalpha (cp [-1])) {
				if (size == 0 || (!bcmp (cp+n, sav_beg, size) && !isalpha (cp [n+size])))
					continue;
				break;
			}
		}
		if (maxloop < 0) {
			fprintf (stderr, "Warning: loop in expand_from_file.\n");
			W.do_undo = True;
			return -1;
		}
		if (size > 0) {
			KBackwardChar (w, size);
			KDeleteChar (w, size, 0);
			*W.b_cursor_b = 0;
			if (W.refresh_from > rfrom) W.refresh_from = rfrom;
			if (W.refresh_to < rfrom)   W.refresh_to = rfrom;
			W.need_hpos = True;
			UpdateWindow (w);
		}
		if (ok) {
			W.do_undo = True;
			return -1;
		}
	}
}

int ExpandAlias (WixWidget w, int count, void* arg)
{
	XEvent*	event = (XEvent*) arg;
	int		reta;
	int		retb;
	int		retc;
	char	c;

	for (;;) {
		/* Tente une expansion  partir de la table d'alias du mode courant */
		if (W.messageFunction)
			W.messageFunction ((Widget) w, T_ERROR, E_EXPAND_ALIAS);
		if ((reta = _expand_alias (w, event, True, True)) == 1) {
			if (W.messageFunction) W.messageFunction ((Widget) w, T_OK, 0);
			break;
		}
		/* Tente une expansion  partir du buffer */
		if (W.messageFunction)
			W.messageFunction ((Widget) w, T_ERROR, E_EXPAND_BUFFER);
		if ((retb = expand_from_buffer (w, event)) == 1) {
			if (W.messageFunction) W.messageFunction ((Widget) w, T_OK, 0);
			break;
		}
		/* Tente une expansion  partir du fichier d'abbrviations du mode courant */
		if (W.messageFunction)
			W.messageFunction ((Widget) w, T_ERROR, E_EXPAND_FILE);
		if ((retc = expand_from_file (w, event)) == 1) {
			if (W.messageFunction) W.messageFunction ((Widget) w, T_OK, 0);
			break;
		}
		/* Reboucle si l'une des expansions  russie */
		if (reta == 0 && retb == 0 && retc == 0) {
			if (XLookupString ((XKeyEvent*) event, &c, 1, 0, 0)) {
				(void) InsertChar (w, count, (void*) (long) c);
				overrideCommand (w, InsertChar, (void*) (long) c, False);
			}
			if (W.messageFunction) W.messageFunction ((Widget) w, T_OK, 0);
			break;
		}
	}
	return 1;
}

/*
 *	Remet les chanes dtruites  la faon d'expand-alias.
 */

void IncrementalYankAction (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	int				current_pos;
	int				y_pos;
	int				n;
	int				nlines;
	unsigned char*	cp;
	long			length;

	if (W.read_only)
		return;

	W.do_undo = False;							/* dsactiver l'undo/redo */
	current_pos = W.b_cursor_b - W.beginning;	/* le point d'insertion */
	if (W.messageFunction)
		W.messageFunction (widget, T_ERROR, E_EXPAND_KILL_STACK);
	RemoveCursor (w);
	for (n = 1;; n++) {
		cp = GetKilled (n, &length);			/* rcuprer le n-ime item */
		if (cp != NULL) {
			y_pos = W.cur_y;
			KInsertString (w, cp, length);		/* l'insrer */
			nlines = W.cur_y - y_pos;			/* nombre de lignes insres */
			W.refresh_from = y_pos;
			W.refresh_to = nlines > 0 ? 1000 : y_pos;
			UpdateWindow (w);					/* et le montrer */
		}
		if (!wait_for_char (w, event)) {
			W.do_undo = True;
			if (W.messageFunction) W.messageFunction (widget, T_OK, 0);
			if (cp != NULL) {
				PushInsert (w, current_pos, cp, length);
				XtFree ((char*) cp);
			}
			return;
		}
		if (cp != NULL) {
			XtFree ((char*) cp);
			KBackwardChar (w, length);
			KDeleteChar (w, length, 0);
			W.need_hpos = True;
		} else
			n = 0;								/* repartir au dbut */

		W.cur_y = y_pos;
		if (nlines > 0) {
			W.refresh_from = 0;
			W.refresh_to = 1000;
		} else {
			W.refresh_from = y_pos;
			W.refresh_to = y_pos;
		}
		UpdateWindow (w);
	}
}

/*
 *	Insre une nouvelle ligne avec la mme indentation que la prcdente.
 */

int NewLineIndent (WixWidget w, int count, void* arg)
{
	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;
	W.refresh_to = 1000;
	W.is_modified = True;

	RemoveCursor (w);
	nlws (w, get_indent (w), NULL);
	return 1;
}

/*
 *	Retourne vrai si est seulement si on est dans un commentaire multi-ligne.
 */

Boolean InsideMultilineComment (WixWidget w, unsigned char* cp)
{
    unsigned char*	beg = W.beginning;
	int				start_length;
	int				stop_length;

	if (! W.start_comment || ! W.stop_comment)
		return False;	/* pas de dlimiteurs de commentaires */

	start_length = strlen (W.start_comment);
	stop_length = strlen (W.stop_comment);

	for (; cp > beg; --cp) {
		/* avril 2009: ne pas dpasser du buffer! */
		if (cp-start_length < beg)
			break;

		if (!bcmp (cp-start_length, W.start_comment, start_length)) {
			if (! W.line_comment)
				return True;	/* pas de commentaires sur une ligne */

			stop_length = strlen (W.line_comment);
			while (*--cp != '\n') {
				if (cp-stop_length < beg)
					break;

				if (!bcmp (cp-stop_length, W.line_comment, stop_length))
					return False;	/* on n'est pas dans un commentaires  */
			}						/* multi-ligne */
			return True;	/* on est dans un commentaires multi-ligne */
		}
		if (cp-stop_length < beg)
			break;

		if (!bcmp (cp-stop_length, W.stop_comment, stop_length))
			return False;
	}
	return False;
}

/*
 *	Insre une nouvelle ligne avec la mme indentation que la prcdente plus un cran
 *	si la ligne courante se termine avec une { ou commence par if, for, do while, ou case.
 */

#define	isdelim(c) ((c)==' ' || (c)=='\t' || (c)=='(')

int NewLineIndentC (WixWidget w, int count, void* arg)
{
	int				i;
	unsigned char*	p;
	unsigned char*	cp = W.b_cursor_b;

	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;
	W.refresh_to = 1000;
	W.is_modified = True;

	RemoveCursor (w);
	if (InsideMultilineComment (w, cp)) {
		while (*--cp != '\n');	/* aller au dbut de la ligne */
		p = cp++;				/* inclure le \n */
		while (*cp == ' ' || *cp == '\t')
			cp ++;				/* trouver les espaces  insrer */

		KInsertString (w, p, cp - p);
		if (*cp == '*') {
			/* n'insrer l'* que dans un commentaire avec * */
			for (p = cp+1; *p == ' ' || *p == '\t'; )
				++p;
			KInsertString (w, cp, p - cp);
		}
	} else {
		i = get_indent (w);
		while (*--cp != '\n')
			if (*cp == '{') {
				i += W.tab_equiv;
				break;
			}
		if (*cp++ == '\n') {
			while (*cp == ' ' || *cp == '\t')
				cp++;
			if ( (isdelim (cp [2]) && !bcmp (cp, "if", 2)) ||
				 (isdelim (cp [3]) && !bcmp (cp, "for", 3)) ||
				 (isdelim (cp [4]) && !bcmp (cp, "case", 4)) ||
				 (isdelim (cp [5]) && !bcmp (cp, "while", 5)) ||
				 (isdelim (cp [5]) && !bcmp (cp, "default", 7)) ||
				 (isdelim (cp [2]) && !bcmp (cp, "do", 2)))
				i += W.tab_equiv;
			else {
				cp = W.b_cursor_b;
				if (isdelim (cp [-5]) && !bcmp (cp-4, "else", 4))
					i += W.tab_equiv;
			}
		}
		nlws (w, i, NULL);
	}
	return 1;
}
