/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "Primitive.h"
#include "PrimitiveP.h"

#define	_PRIVATE_
#include "Wix.h"
#include "WixP.h"

#undef NUM_MARKS
#define NUM_MARKS 16

#define DEF_LINES	128		/* nombre de lignes par defaut */
#define LINE_INCR	64		/* increment */

long	EndOfKillStack = 0;	/* la fin du cimetiere (pour storeInKillStack) */
int		FdKillFile;
extern Time	_LastActionTime;/* heure de la precedente action */
extern Time	_KillTime;		/* heure de la precedente destruction */

/*
 *	Le mcanisme d'UNDO/REDO
 */

void PushCommand (WixWidget w)
{
	UndoRedoCallbackStruct	urcb;

	if (W.undo_redo_callback != NULL) {
		urcb.reason = COMMAND_MARK;
		XtCallCallbackList ((Widget)w, W.undo_redo_callback, (XtPointer) &urcb);
	}
}

void PushInsert (WixWidget w, int position, unsigned char* text, int length)
{
	UndoRedoCallbackStruct	urcb;

	if (W.undo_redo_callback != NULL && W.do_undo) {
		urcb.reason = INSERT_STRING;
		urcb.pos = position;
		urcb.string = text;
		urcb.length = length;
		XtCallCallbackList ((Widget)w, W.undo_redo_callback, (XtPointer) &urcb);
	}
}

void PushDelete (WixWidget w, int position, unsigned char* text, int length)
{
	UndoRedoCallbackStruct	urcb;

	if (W.undo_redo_callback != NULL && W.do_undo) {
		urcb.reason = DELETE_STRING;
		urcb.pos = position;
		urcb.string = text;
		urcb.length = length;
		XtCallCallbackList ((Widget)w, W.undo_redo_callback, (XtPointer) &urcb);
	}
}

/*
 *	Si on veut encrypter le buffer.
 */

#ifdef CRYPT

#define ROTORSZ 256
#define MASK 0377

static char	t1 [ROTORSZ];
static char	t2 [ROTORSZ];
static char	t3 [ROTORSZ];

CryptSetup (char* pw) {
	long 			seed;
	int				ic, i, k, temp;
	unsigned int	random;
	char			buf [13];

	strncpy (buf, pw, 8);

	buf[8] = buf[0];
	buf[9] = buf[1];

	strncpy (buf, crypt (buf, buf+8), 13);

	seed = 123;
	for (i=0; i<13; i++)
		seed = seed*buf[i] + i;
	for(i=0;i<ROTORSZ;i++) {
		t1[i] = i;
		t3[i] = 0;
	}
	for(i=0;i<ROTORSZ;i++) {
		seed = 5*seed + buf[i%13];
		random = seed % 65521;
		k = ROTORSZ-1 - i;
		ic = (random&MASK)%(k+1);
		random >>= 8;
		temp = t1[k];
		t1[k] = t1[ic];
		t1[ic] = temp;
		if(t3[k]!=0)
			continue;
		ic = (random&MASK) % k;
		while(t3[ic]!=0) ic = (ic+1) % k;
		t3[k] = ic;
		t3[ic] = k;
	}
	for(i=0;i<ROTORSZ;i++)
		t2[t1[i]&MASK] = i;
}

CryptBuffer (char* cp, char* end, char* to)
{
	int	n1, n2;

	for (n1 = n2 = 0; cp < end;) {
		*to++ = t2[(t3[(t1[(*cp++ +n1)&MASK]+n2)&MASK]-n2)&MASK]-n1;
		n1++;
		if(n1==ROTORSZ) {
			n1 = 0;
			n2++;
			if(n2==ROTORSZ) n2 = 0;
		}
	}
}

#endif

/*
 *	Insre des lignes dans le tableau de lignes.
 */

static void insertLines (WixWidget w, int n)
{
	register Line*	p1 = W.last_line + 3;
	register Line*	p2 = p1+n;
	register Line*	p3 = W.cur_line + 1;

	while (p1 > p3) *--p2 = *--p1;
	W.last_line += n;
	p1 = W.last_line + 1;
	p1++ -> chars = NULL;
	p1++ -> chars = NULL;
	p1 -> chars = NULL;
}

/*
 *	dtruit des lignes dans le tableau de lignes.
 */

static void deleteLines (WixWidget w, int n)
{
	register Line*	p1 = W.cur_line + 1;
	register Line*	p2 = p1 + n;

	while ((*p1++ = *p2++), (p1-1) -> chars != NULL);
	W.last_line -= n;
	p1 = W.last_line+1;
	p1++ -> chars = NULL;
	p1++ -> chars = NULL;
	p1 -> chars = NULL;
}

/*
 *	Agrandit le tableau de lignes.
 */

static void reallocLines (WixWidget w, int n)
{
	register Line*	l;

	l = (Line*) XtRealloc ((char*)W.first_line, n * sizeof (Line));

	W.cur_line = l + (W.cur_line - W.first_line);
	W.top_line = l + (W.top_line - W.first_line);
	W.last_line = l + (W.last_line - W.first_line);
	W.first_line = l;
	W.num_slots = n;
}

/*
 *	Recalcule la hauteur de la ligne.
 */

#define CAP_ACCENTED(c) \
	if ((unsigned char) '' <= (c) && (c) <= (unsigned char) '')\
		{ offset = 2; }

void calc_height (WixWidget w, Line* l)
{
	int				ft = 0;
	int				offset = 0;
	unsigned char*	a;
	unsigned char*	p;

	if ((p = l -> chars) != 0)
		for (a = l -> fn; *p != '\n'; a++, p++) {
			CAP_ACCENTED (*p);
			if (ft != *a && FONT (ft) -> f_height < FONT (*a) -> f_height)
				ft = *a;
		}
	l -> height = FONT (ft) -> f_height + offset;
	l -> base_line = FONT (ft) -> f_ascent + offset;
}

/*
 *	Remet la fonte 0 a tout le buffer. (pour les modes sans commentaires)
 *	Doit tre suivi d'un MakeLines pour recalculer les hauteurs.
 */

void ResetComments (WixWidget w)
{
	bzero (W.attr_beg, W.a_cursor_b - W.attr_beg);
	bzero (W.a_cursor_e, W.attr_end - W.a_cursor_e);
	W.hilited = False;
}

/*
 *	Traitement des mots clefs.
 */

/*
 *	Reconnait une expression rgulire restreinte:
 *
 *	.			matche un caractre quelconque
 *	[x-yX-YZ]	matche un caractre de x  y ou de X  Y ou Z
 *	[^x-y]		matche un caractre qui n'est pas de x  y
 *	[-ab]		matche - ou a ou b
 *
 * retourne le nombre de caractres atteints si "s" matche "re", ou 0 sinon.
 * l'expression rgulire a t teste avant et n'est pas mal forme.
 */

#define INVALID_RE	-2
#define PLUSED_RE	-1

static int re_match (char* re, char* s)
{
	char*	text = s;
	char*	sv;
	char*	cp;
	int		ok;
	int		neg;

	for (;;) switch (*re) {
		/* rptition du caractre prcdent */
		case '+':
			*re = 0;		/* simuler un regexp constitue d'uniquement le */
							/* caractre prcdent */
			while (1 == re_match (sv, s))
				++s;		/* et essayer de la matcher  mort */
			*re++ = '+';	/* remettre le plus */
			sv = re;		/* et continuer ppre */
			continue;
		/* on est arriv au bout ! */
		case 0:				/* comme on ne cherche en fait qu'un prfixe qui matche */
			return s-text;	/* on ne vrifie pas la fin de 's' */
		/* caractre spcial */
		case '\\':
			sv = re;
			++re;	/* considrer le caractre suivant tel quel */
			if (*re++ != *s++)
				return 0;
			continue;

		default:
			sv = re;
			if (*re++ != *s++)
				return 0;
			continue;
		/* caractre quelconque */
		case '.':
			sv = re;
			re++;
			s++;
			continue;
		/* intervalles ou caractres spcifiques */
		case '[':
			sv = re;
			++re;
			ok = 0;
			neg = 0;
			if (re [0] == '^')
				++re, neg = 1;	/* ngation */
			while (re [0] != ']') {
				/* caractre spcial */
				if (re [0] == '\\')
					++re;
				/* intervalle */
				if (re [1] == '-') {
					/* borne sup spciale */
					if (re [2] == '\\') {
						if (*s >= *re && *s <= re [3]) {
							ok = 1;
							break;
						}
						re += 4;
						continue;
					}
					/* borne sup standard */
					if (*s >= *re && *s <= re [2]) {
						ok = 1;
						break;
					}
					re += 3;
					continue;
				}
				/* caractre spcifique */
				if (*s == *re) {
					ok = 1;
					break;
				}
				re++;
			}
			if (neg+ok != 1)	/* 0: affirmatif et pas trouv */
				return 0;		/* 2: ngatif mais trouv */
			/* on a trouv entre les crochets; aller aprs le fermant */
			if (ok)
				++re;
			while (*re != ']')
				if (*re++ == '\\')
					++re;
			++re;	/* s'arrter aprs le ] */
			++s;	/* pour regarder la suite */
			continue;
	}
	return 0;	/* non atteint */
}

/*
 *	Retourne la longueur d'une expression rgulire restreinte,
 *	-2 s'il y a un plus mais que des alphanumriques, ou -1 si elle est mal forme.
 *	hashable sera faux s'il s'agit d'une  expression rgulire.
 *	S'il s'agit d'une string, hashable sera vrai. Dans ce dernier cas,
 *	les ventuels backslashes devenus inutiles seront retirs de la chane.
 */

static int re_length (char* re, Boolean* hashable, char* delim)
{
	int		length;
	Boolean	backslash = False;
	Boolean	star = False;
	Boolean alpha = True;
	char*	regex = re;

	*hashable = True;
	for (length = 0;; ++length, ++re)
		switch (*re) {
			default:
			/* est-ce un alphanumrique ? */
				if (alpha && index (delim, *re))
					alpha = False;
				continue;
			/* on est arriv au bout ! */
			case 0:
				if (*hashable && backslash) {
					/* comme c'est une string, on peut virer les backslashes */
					char	tmp [256];	/* largement suffisant */
					char*	p = tmp;

					for (re = regex; (*p++ = *re++) != 0; )
						if (p [-1] == '\\')
							--p;
					/* remettre la string dans l'argument */
					strcpy (regex, tmp);
				}
				else if (star) {
					if (alpha)
						return PLUSED_RE;	/* regexpr avec + mais sans dlimiteur */
					return INVALID_RE;		/* + interdit dans une regexpr gnrale */
				}
				return length;
			/* caractre spcial */
			case '\\':
				backslash = True;
				++re;	/* considrer le caractre suivant tel quel */
				continue;
			/* avec + => ne devra contenir que des alphanumriques stricts */
			case '+':
				*hashable = False;
				star = True;
				continue;
			/* intervalle */
			case '[':
				*hashable = False;
				for (;; ++re) {
					switch (*re) {
						default:
							continue;
						case '\\':
							++re;
							continue;
						case ']':
							break;
						case 0:
							return INVALID_RE;	/* expression rgulire mal forme */
					}
					break;
				}
				continue;
			case '.':
				alpha = False;
				*hashable = False;
				continue;
		}
	return INVALID_RE;	/* non atteint */
}

/*
 *	Fonction de hash-code pour les mots clefs.
 */

static int hash (char* name, int n)
{
	int	h;

	for (h = 11; --n >= 0;)
		h = ((h * 17) ^ (int)(*name++)) % MAX_KEYWORDS;

	if (h < 0)
		h += MAX_KEYWORDS;
	return h;
}

/*
 *	Recherche un mot clef dans la table; si 'color' est >= 0 le met dans la table.
 */

static int intern_keyword (KeyWord* keyword_table, char* name, int color, int length, int regexp)
{
	char	tc [128];
	int		h = hash (name, length);
	int		i;

	for (i = MAX_KEYWORDS; --i >= 0;) {
		if (keyword_table [h].keyword == NULL) {
			if (color < 0)
				return -1;	/* pas trouve */
			/* on le met dans la table */
			if (name [length] != 0) {
				/* en faire une chane C */
				bcopy (name, tc, length);
				tc [length] = 0;
				keyword_table [h].keyword = XtNewString (tc);
			} else
				keyword_table [h].keyword = XtNewString (name);
			keyword_table [h].length = length;
			keyword_table [h].color = color;
			keyword_table [h].regexp = regexp;
			return h;
		}

		if (keyword_table [h].length == length && !strncmp (keyword_table [h].keyword, name, length))
			return h;

		h = (h + 41) % MAX_KEYWORDS;	/* collision */
	}
	return -1;	/* plus de place dans la table */
}

#ifdef DEBUG
static void print_keyword_table (KeyWord* table)
{
	int	i;

	for (i = MAX_KEYWORDS; --i >= 0;)
		if (table [i].keyword != NULL)
			fprintf (stdout, "name: %s length: %d isregexpr: %d col: %d class_col: %d\n",
					table [i].keyword,
					table [i].length,
					table [i].regexp,
					table [i].color,
					table [i].class_col);
}
#endif

/*
 *	Retire les entres temporaires de la table.
 */

static void remove_temp (KeyWord* table)
{
	int	i;

	if (table == 0)
		return;		/* des fois que la table soie vide! */

	for (i = MAX_KEYWORDS; --i >= 0;)
		if (table [i].keyword != NULL && table [i].regexp == -1) {
#ifdef DEBUG
			fprintf (stderr, "remove tmp: %s\n", table [i].keyword);
#endif
			XtFree (table [i].keyword);
			table [i].keyword = 0;
		}
}

/*
 *	Initialise la table des mots clef.
 */

#define skipSpace(cp)	while (*cp == ' ' || *cp == '\t') cp ++
#define findSpace(cp)	while (*cp != ' ' && *cp != '\t') cp ++
#define endLine(cp)	while (*cp != '\n' && *cp != '\0') cp++

/*
 *	Les premires MAX_KEYWORDS entres sont pour la table de hash code des
 *	mots clefs standard, les MAX_REGEX suivantes sont pour la table (ordonne) des
 *	mots clefs dfinis par une expression rgulire restreinte.
 */

/* 'name' contient du dlimiteur ? */

static Boolean contain_delimitor (char* name, char* delim)
{
	while (*name)
		if (index (delim, *name++))
			return True;
	return False;
}

/* couleur => numro de couleur */

static int string2color (char* color)
{
	if (!strcmp (color, "KeywordColor"))	return KEYWORD_COLOR;
	if (!strcmp (color, "CppColor"))		return CPP_COLOR;
	if (!strcmp (color, "CommentColor"))	return COMMENT_COLOR;
	if (!strcmp (color, "StandardColor"))	return STANDARD_COLOR;
	if (!strcmp (color, "LibraryColor"))		return LIBRARY_COLOR;
	if (!strcmp (color, "ConstantColor"))	return CONSTANT_COLOR;
	if (!strcmp (color, "ClassColor"))		return CLASS_COLOR;

	fprintf (stderr, "Invalid color specification: %s\n"
					 "(use: KeywordColor, CommentColor, CppColor,"
					 " LibraryColor, ConstantColor, ClassColor or StandardColor).\n",
					color);
	return STANDARD_COLOR;
}

KeyWord* CvtStringToKeyWordTable (char* s, char* delim) {
	KeyWord*	keyword_table;
	char*		name;
	char*		color;
	int			length;
	int			col;
	int			class_col;
	Boolean		with_delim;
	Boolean		hashable;
	int			i;
	int			n;
	int			last_re = MAX_KEYWORDS;	/* dbut de la table des mots clefs spciaux */

	if (s == NULL)
		return NULL;	/* table vide */

	keyword_table = (KeyWord*) XtMalloc ((MAX_KEYWORDS+MAX_REGEX) * sizeof (KeyWord));
	bzero (keyword_table, (MAX_KEYWORDS) * sizeof (KeyWord));
	/* initialiser la longueur des regexpr */
	for (col = MAX_KEYWORDS; col < MAX_KEYWORDS+MAX_REGEX; ++col)
		keyword_table [col].length = INVALID_RE;

	while (*s) {
		skipSpace (s); name = s; findSpace (s); *s++ = 0;	/* le nom */
	/*
	 *	space* nom space* color space* \n
	 *	space* nom space* color space+ class_color space+ < space* \n
	 *	space* nom space* color space+ class_color space+ > space* \n
	 */
		skipSpace (s); color = s;							/* la couleur */
		while (('A' <= *s && *s <= 'Z') || ('a' <= *s && *s <= 'z'))
			++s;
		class_col = 123456;		/* pas de couleur supplmentaire */
		switch (*s) {
			case ' ':
			case '\t':
				*s++ = 0;
				col = string2color (color);
				skipSpace (s);
				/* il y ena encore aprs */
				if (*s == '\n') {
					*s++ = 0;
					break;
				}
				/* c'est le dernier */
				if (*s == 0)
					break;

				color = s;
				findSpace (s); *s++ = 0;
				class_col = string2color (color);
				skipSpace (s);
				if (*s == '<')
					class_col = - class_col;
				else if (*s != '>')
					fprintf (stderr, "Syntax...\n");
				endLine (s);
				if (*s == '\n')
					++s;
				break;

			case '\n':
				*s++ = 0;
				/* fall thru */
			case 0:
				col = string2color (color);
				break;
		}

		length = re_length (name, &hashable, delim);
		if (length == 0) {
			fprintf (stderr, "Null length regexpr: %s\n", name);
			continue;
		}
		if (length == INVALID_RE) {
			fprintf (stderr, "Invalid regexpr: %s\n", name);
			continue;
		}

		with_delim = contain_delimitor (name, delim);	/* contient du dlimiteur ? */
		if (!hashable || with_delim) {
			/* construire notre table des mots clefs spciaux en la triant */
			/* par ordre dcroissant de longueur, car c'est la longueur */
			/* (et non pas l'encadrement par des dlimiteurs) qui sera */
			/* le critre utilis pour reprer le mot clef dans le buffer */
			for (i = MAX_KEYWORDS; i < MAX_KEYWORDS+MAX_REGEX; ++i) {
				/* s'il y a du dlimiteur dans le mot clef et qu'il est hashable */
				/* on ne s'intresse qu' sa longueur, donc on ne met qu'une */
				/* seule entre repre par un champ 'name' nul */
				if (with_delim && hashable &&
					length == keyword_table [i].length &&
					keyword_table [i].keyword == NULL)
				{
					/* on le met dans la table de hashcode des mots clefs */
					n = intern_keyword (keyword_table, name, col, length, 1);
					if (class_col != 123456)
						keyword_table [n].class_col = class_col;
					break;
				}
				/* on a enfin trouv sa place */
				if (length > keyword_table [i].length) {
					/* tout pousser d'un cran vers la fin */
					if (last_re == MAX_KEYWORDS+MAX_REGEX-1) {
						fprintf (stderr, "Too many keywords in the regexpr table (max %d).\n",
											MAX_REGEX);
						break;
					}
					++last_re;	/* une entre de plus */
					for (n = last_re; n > i; --n)
						keyword_table [n] = keyword_table [n-1];
					/* insrer notre nouvelle entre */
					if (with_delim && hashable) {
						n = intern_keyword (keyword_table, name, col, length, 1);
						if (class_col != 123456)
							keyword_table [n].class_col = class_col;
						keyword_table [i].keyword = NULL;
					} else {
						keyword_table [i].keyword = XtNewString (name);
						if (class_col != 123456)
							keyword_table [i].class_col = class_col;
					}
					keyword_table [i].length = length;
					keyword_table [i].color = col;
					break;
				}
			}
			continue;
		}
		/* hashable et ne contient pas de dlimiteur */
		if (-1 == (i = intern_keyword (keyword_table, name, col, length, 0))) {
			fprintf (stderr, "Too many keywords in the table (max %d).\n", MAX_KEYWORDS);
			break;
		} else if (class_col != 123456)
			keyword_table [i].class_col = class_col;
	}
#ifdef DEBUG
	print_keyword_table (keyword_table);
#endif
	return keyword_table;
}

/*
 *	Cherche une chane dans la table de hashcode; si regexpr est <= 0 on cherche un mot sinon
 *	une expression rgulire.
 */

static int find_keyword (WixWidget w, unsigned char* cp, int n, int regexp)
{
	if (W.keyword_table == NULL)
		return -1;
	return intern_keyword (W.keyword_table, cp, -1, n, regexp <= 0 ? 0 : regexp);
}

#define isnum(c) ((c)>='0' && (c)<='9')
#define isspace(c) ((c)==' ' || (c)=='\t' || (c)=='\n' || (c) == 0)
#define	isalpha(c) (index (W.delim, c) == 0)
#define colorize(ap,color) if ((*ap & ~HILITED) != color) *ap = color, found = 1
#define standardize(ap) colorize ((ap), STANDARD_COLOR)
#define commentify(ap) colorize (ap, COMMENT_COLOR)
#define constantify(ap) colorize (ap, CONSTANT_COLOR)
#define stringify(ap) colorize (ap, STRING_COLOR)
#define numerize(ap) colorize (ap, CPP_COLOR)

#define STANDARD_TEXT	0
#define INSIDE_COMMENT	1
#define	INSIDE_STRING	2

/*
 * Coloration du mot suivant.
 */

static void color_next (WixWidget w, unsigned char* p, int ix)
{
	unsigned char*	q;

	if (W.keyword_table [ix].class_col > 0) {
		/* sauter les espaces */
		while (isspace (*p))
			++p;
		/* ne colorier que les suites de non-dlimiteurs */
		for (q = p; isalpha (*p); ++p)
			;
		/* oui, mettre ce suivant dans la table (sous forme temporaire) */
		if (p > q)
			(void) intern_keyword (W.keyword_table, q,
								   W.keyword_table [ix].class_col,
								   p-q, -1);
	}
}

/*
 *	Explore le buffer  la recherche de commentaires, ou de mots clefs.
 */

static int processComments (WixWidget w, unsigned char* cp, unsigned char* ap, unsigned char* to, int* state)
{
	Boolean			with_regexp;
	int				n;
	int				i, j, k;
	int				color;
	int				found;
	unsigned char*	p;
	unsigned char*	q;
	int				begin_len = W.start_comment ? strlen (W.start_comment) : 0;
	int				end_len = W.stop_comment ? strlen (W.stop_comment) : 0;
	int 			line_len = W.line_comment ? strlen (W.line_comment) : 0;
	int				beg_keyword_len = W.beg_keyword ? strlen (W.beg_keyword) : 0;
	int				end_keyword_len = W.end_keyword ? strlen (W.end_keyword) : 0;

	for (found = 0; cp < to; ++cp, ++ap) {
		switch (*state) {
		case INSIDE_STRING:
			stringify (ap);
			if (*cp == '\\') {
				cp++;
				ap++;
				stringify (ap);
				continue;
			}
			if (*cp == W.string_delim) {
				*state = STANDARD_TEXT;
				standardize (ap);	/* le dlimiteur est standard */
			}
			continue;

		case STANDARD_TEXT:
			/* dbut de commentaire sur plusieurs lignes */
			if (begin_len > 0 && *cp == W.start_comment [0] &&
				!bcmp (cp, W.start_comment, begin_len))
			{
				*state = INSIDE_COMMENT;
				commentify (ap);
				continue;
			}
			/*  commentaires en une ligne */
			if (line_len > 0 && cp [-1] != '\\' && *cp == W.line_comment [0] && !bcmp (cp, W.line_comment, line_len)) {
				for (; *cp != '\n'; cp++, ap++)
					commentify (ap);
				standardize (ap);	/* le \n est en standard */
				continue;
			}
			/* chaine constante (sauf si le guillemet n'est pas dans les dlimiteurs) */
			if (*cp == W.string_delim && cp [-1] != '\\' && !isalpha (W.string_delim)) {
				standardize (ap);
				*state = INSIDE_STRING;
				continue;
			}
			/*  mot clef dlimit par deux balises (sur une ligne) */
			if (beg_keyword_len > 0 && cp [-1] != '\\' &&
				*cp == W.beg_keyword [0] &&
				!bcmp (W.beg_keyword+1, cp+1, beg_keyword_len-1))
			{
				/* avec dlimiteur de fin de mot clef */
				if (W.end_keyword) {
					while (*cp != '\n' && *cp != ' ' && *cp != '\t' &&
						   (*cp != W.end_keyword [0] ||
						    bcmp (W.end_keyword, cp+1, end_keyword_len-1)))
					{
						colorize (ap, KEYWORD_COLOR);
						++cp; ++ap;
					}
					/* colorier aussi le dlimiteur (sauf les espaces ou autres) */
					if (*cp == W.end_keyword [0]) {
						for (i = end_keyword_len; --i >= 0; ++cp, ++ap)
							colorize (ap, KEYWORD_COLOR);
						--ap; --cp;
					}
					continue;
				} else {
				/* sans dlimiteur de fin de mot clef (utiliser les dlimiteurs standard) */
					/* si le mot clef est seul, regarder dans la table (nota: si on ne met */
					/* pas le !isspace (cp [...]) on autorise le dbut de mot clef  tre */
					/* seul  MAIS colori) */
					if (!isspace (cp [beg_keyword_len]) && isalpha (cp [beg_keyword_len]))
					{
						/* colorier le dbut de mot clef */
						for (p = W.beg_keyword; *p != 0; ++p) {
							colorize (ap, KEYWORD_COLOR);
							++cp; ++ap;
						}
						while (!isspace (*cp) && isalpha (*cp))
						{
							colorize (ap, KEYWORD_COLOR);
							++cp; ++ap;
						}
						/* ne pas colorier le dlimiteur */
						--cp; --ap;
						continue;
					}
				}
			}
			/* s'il n'y a pas de table des mots clefs, c'est rgl! */
			if (W.keyword_table == NULL) {
				standardize (ap);
				continue;
			}
			/* chercher si c'est un mot clef */
			for (p = cp; isalpha (*p); p++)
				;
			/* c'est bien un mot avec un sparateur devant */
			if (p > cp && !isalpha (cp [-1])) {
				/* y a-t-il un mot clef aprs qui le ferait changer de couleur ? */
				unsigned char*	z;
				/* sauter les espaces */
				for (q = p; *q == ' ' || *q == '\t'; ++q)
					;
				/* trouver la fin du mot */
				for (z = q; isalpha (*q); ++q)
					;
				if (q > z && ((i = find_keyword (w, z, q-z, 0)) >= 0) &&
					W.keyword_table [i].class_col < 0)
					/* oui! */
					(void) intern_keyword (W.keyword_table, cp,
										   - W.keyword_table [i].class_col,
										   p-cp, -1);
				/* faut-il colorer ce mot ? */
				if ((i = find_keyword (w, cp, p-cp, 0)) >= 0) {
					int	color = W.keyword_table [i].color;
					for (n = W.keyword_table [i].length; --n >= 0; cp++, ap++)
						colorize (ap, color);
					--cp; --ap;
					/* ce mot indique-t-il une coloration pour le suivant ? */
					color_next (w, p, i);
					continue;
				}
			}
			/* regarder s'il s'agit d'une expression rgulire ou */
			/* d'un mot clef comportant des dlimiteurs non prcd d'un backslash */
			if (*cp == '\\') {
				standardize (ap);	/* ne pas colorier le backslash */
				++cp; ++ap;			/* et sauter le suivant */
				standardize (ap);	/* sans le colorier */
				continue;
			}
			with_regexp = False;
			for (i = MAX_KEYWORDS; (n = W.keyword_table [i].length) > 0; ++i) {
				/* des dlimiteurs ? */
				if (W.keyword_table [i].keyword == NULL) {
					if ((j = find_keyword (w, cp, n, 1)) >= 0 &&
						W.keyword_table [j].regexp == 1)
					{
						/* regarder si ce mot clef fait changer de couleur le prcdent */
						if (W.keyword_table [j].class_col < 0) {
							for (q = cp-1; isspace (*q); --q)
								;
							for (p = q; isalpha (*p); --p)
								;
							/* oui, mettre ce prcdent dans la table (sous forme temporaire) */
							if (q > p) {
								unsigned char*	at;

								color = - W.keyword_table [j].class_col;
								(void) intern_keyword (W.keyword_table, p+1,
													   color, q-p, -1);
								/* le colorer s'il ne l'est pas dj (le premier a priori) */
								at = ap - (cp - (p+1));
								if (*at == 0) {
									for (; q > p; ++p, ++at)
										colorize (at, color);
								}
							}
						}
						/* colorer le mot clef */
						color = W.keyword_table [j].color;
						for (k = W.keyword_table [j].length; --k >= 0; cp++, ap++)
							colorize (ap, color);
						--cp; --ap;
						/* ce mot indique-t-il une coloration pour le suivant ? */
						color_next (w, cp+1, j);
						goto continuer;	/* c'est colori, passer  la suite */
					}
					continue;
				}
				with_regexp = True;	/* il y en a de possibles */
				/* une expression rgulire ? */
				if ((k = re_match (W.keyword_table [i].keyword, cp)) != 0) {
					color = W.keyword_table [i].color;
					for (; --k >= 0; cp++, ap++)
						colorize (ap, color);
					--cp; --ap;
					/* ce mot indique-t-il une coloration pour le suivant ? */
					color_next (w, cp+1, i);
					goto continuer;	/* c'est colori, passer  la suite */
				}
			}
			/* regarder les expressions rgulires sur les mots */
			if (p > cp && !isalpha (cp [-1])) {
				while (n == PLUSED_RE) {
					k = re_match (W.keyword_table [i].keyword, cp);
					/* match exact sur toute la longueur du mot */
					if (k == p - cp) {
						color = W.keyword_table [i].color;
						for (; --k >= 0; cp++, ap++)
							colorize (ap, color);
						--cp; --ap;
						/* ce mot indique-t-il une coloration pour le suivant ? */
						color_next (w, cp+1, i);
						goto continuer;	/* c'est colori, passer  la suite */
					}
					n = W.keyword_table [++i].length;
				}
			}
			/* on peut standardiser le mot sans problmes */
			if (!with_regexp && p > cp) {
				for (; p > cp; cp++, ap++)
					standardize (ap);
				--cp;
				--ap;
				continue;
			}
			/* caractre  ne pas colorier! */
			standardize (ap);
continuer:	continue;

		case INSIDE_COMMENT:
			if (end_len > 0 && *cp == W.stop_comment [0] && !bcmp (cp, W.stop_comment, end_len)) {
				*state = STANDARD_TEXT;
				for (n = end_len; --n >= 0; ++cp, ++ap)
					commentify (ap);
				standardize (ap);	/* le caractre suivant est en standard */
				continue;
			}
			commentify (ap);
			continue;
		}
	}
	return found;
}

/*
 *	Filtre la ligne courante a la recherche de mots clefs ou de commentaires.
 */

static void filterCurrentLine (WixWidget w) {
	unsigned char*	ap = W.a_cursor_b;
	unsigned char*	ae = W.a_cursor_e;
	unsigned char*	cp = W.b_cursor_b;
	unsigned char*	ce = W.b_cursor_e;
	int				state;

	if (!W.do_redisplay)
		return;

	if (!W.start_comment && !W.stop_comment && !W.line_comment &&
		!W.beg_keyword && !W.keyword_table)
		return;	/* vraiment rien  colorier! */

	/* faire la ligne courante */
	while (((*--ae = *--ap), (*--ce = *--cp)) != '\n')
		;
	ce++; ae++;		/* se mettre au dbut de la ligne */
	/*
	 *	O en est-on ?
	 */
	if (InsideMultilineComment (w, cp))
		state = INSIDE_COMMENT;		/* dans un commentaire */
	else if (*ap == STRING_COLOR)	/* couleur du \n prcdent */
		state = INSIDE_STRING;		/* dans une chane */
	else
		state = STANDARD_TEXT;			/* normal ! */
	/*
	 *	Trouver la fin de la ligne.
	 */	
	for (cp = W.b_cursor_e; *cp != '\n'; cp++)
		;
	(void) processComments (w, ce, ae, cp+1, &state);
	remove_temp (W.keyword_table);
	/* refaire la ligne courante */
	ap = W.a_cursor_b;
	ae = W.a_cursor_e;
	cp = W.b_cursor_b;
	ce = W.b_cursor_e;
	while (((*--ap = *--ae), (*--cp = *--ce)) != '\n')
		;
}

/*
 *	Filtre le buffer. (traite les commentaires, les mots clefs, etc...)
 */

int FilterBuffer (WixWidget	w) {
	int				state = STANDARD_TEXT;
	int				found = 0;
	unsigned char*	ap = W.a_cursor_b;
	unsigned char*	ae = W.a_cursor_e;
	unsigned char*	cp = W.b_cursor_b;
	unsigned char*	ce = W.b_cursor_e;

	if (!W.do_redisplay)
		return found;

	if (!W.start_comment && !W.stop_comment && !W.line_comment &&
		!W.beg_keyword && !W.keyword_table)
		return found;	/* vraiment rien  colorier! */

	/* faire la ligne courante */
	while (((*--ae = *--ap), (*--ce = *--cp)) != '\n')
		;
	/* explorer le buffer avant la ligne courante */
	found = processComments (w, W.beginning, W.attr_beg, cp, &state);
	/* explorer depuis la ligne courante */
	found += processComments (w, ce, ae, W.end, &state);
	remove_temp (W.keyword_table);
	/* refaire la ligne courante */
	while (ce < W.b_cursor_e) {
		*cp++ = *ce++;	/* inutile ! */
		*ap++ = *ae++;
	}
	return found;
}

#undef STANDARD_TEXT
#undef INSIDE_COMMENT
#undef INSIDE_STRING

/*
 *	Refait la ligne courante.
 */

void MakeCurrentLine (WixWidget w)
{
	unsigned char*	ce = W.b_cursor_e;
	unsigned char*	cb = W.b_cursor_b;
	unsigned char*	ae = W.a_cursor_e;
	unsigned char*	ab = W.a_cursor_b;
	int				offset = 0;
	Line*			l = W.cur_line;
	int				ft = 0;

	/*
	 *	Aller au debut de la ligne.
	 */
	while ((*--ae = *--ab, *--ce = *--cb) != '\n')
		if (ft != *ae && FONT (ft) -> f_height < FONT (*ae) -> f_height)
			ft = *ae;
	l -> chars = ++ce;		/* les caractres */
	l -> fn = ++ae;			/* les fontes */
	ft = *ae;
	/*
	 *	Aller au bout de la ligne.
	 */
	while (ae++, *ce++ != '\n') {
		CAP_ACCENTED (ce [-1]);
		if (ft != *ae && FONT (ft) -> f_height < FONT (*ae) -> f_height)
			ft = *ae;
	}
	l -> height = FONT (ft) -> f_height + offset;
	l -> base_line = FONT (ft) -> f_ascent + offset;
}

/*
 *	Refait la liste des lignes et retourne leur nombre.
 */

void MakeLines (WixWidget w)
{
	Line*			l = W.first_line;
	unsigned char*	p = W.beginning;
	unsigned char*	a = W.attr_beg;
	unsigned char*	ce = W.b_cursor_e;
	unsigned char*	ae = W.a_cursor_e;
	int				ft;
	int				offset = 0;

	if (p < W.b_cursor_b) {
		l -> chars = p;
		l -> fn = a;
		ft = *a;
		l -> height = FONT (ft) -> f_height;
		l -> base_line = FONT (ft) -> f_ascent;
		while (p < W.b_cursor_b) {
			if (ft != *a && FONT (ft) -> f_height < FONT (*a) -> f_height)
				ft = *a;
			CAP_ACCENTED (*p);
			if ((a++, *p++) == '\n') {
				l -> height = FONT (ft) -> f_height + offset;
				l -> base_line = FONT (ft) -> f_ascent + offset;
				offset = 0;
				l++;
				if (p >= W.b_cursor_b)
					break;
				l -> chars = p;
				l -> fn = a;
				ft = *a;
			}
		}
		do {
			*--ae = *--a;
			CAP_ACCENTED (p [-1]);
		} while ((*--ce = *--p) != '\n');
		l -> chars = ce + 1;
		l -> fn = ae + 1;
	} else {
		l -> chars = ce;
		l -> fn = ae;
		ft = *ae;
	}
	l -> height = FONT (ft) -> f_height + offset;
	l -> base_line = FONT (ft) -> f_ascent + offset;
	offset = 0;
	W.cur_line = l;
	p = W.b_cursor_e;
	a = W.a_cursor_e;
	for (;;) {
		if (ft != *a && FONT (ft) -> f_height < FONT (*a) -> f_height)
			ft = *a;
		CAP_ACCENTED (*p);
		if ((a++, *p++) == '\n') {
			l -> height = FONT (ft) -> f_height + offset;
			l -> base_line = FONT (ft) -> f_ascent + offset;
			offset = 0;
			if (p > W.end)
				break;
			l++;
			l -> chars = p;
			l -> fn = a;
			ft = *a;
		}
	}
	l -> height = FONT (ft) -> f_height + offset;
	l -> base_line = FONT (ft) -> f_ascent + offset;
	W.last_line = l;
	l++; l -> chars = NULL;
	l++; l -> chars = NULL;
	l++; l -> chars = NULL;
}

/*
 *	Appelle la callback avant de modifier le bouffeur.
 */

static int verifyCallback (WixWidget w, int reason, int pos, char* string, int length)
{
	VerifyCallbackStruct	cb;

	cb.reason = reason;
	cb.pos = pos;
	cb.string = string;
	cb.length = length;
	cb.doit = 1;
	XtCallCallbackList ((Widget)w, W.verify_modify_callback, &cb);
	return cb.doit;
}

/*
 *	Retourne une marque a partir de son identifieur.
 */

static mark* getMarkFromData (WixWidget w, void* data)
{
	mark*	mp;

	for (mp = W.b_marks+3; mp -> addr; mp++) {
		if ((long) mp -> addr == -1)
			continue;
		if (mp -> data == data)
			return mp;
	}
	return NULL;	/* pas trouvee */
}

/*
 *	Met une marque a l'endroit courant.
 */

Boolean WixSetMark (Widget widget, void* data)
{
	WixWidget	w = (WixWidget) widget;
	mark*		mp;

	if ((mp = getMarkFromData (w, data)) != 0) {
		mp -> addr = W.b_cursor_e;
		return True;
	}
	return False;
}

/*
 *	Va a la marque.
 */

Boolean WixGotoMark (Widget widget, void* data)
{
	WixWidget	w = (WixWidget) widget;
	mark*		mp;

	if ((mp = getMarkFromData (w, data)) != 0) {
		GotoMark (w, mp -> addr);
		return True;
	}
	return False;
}

/*
 *	Detruit une marque.
 */

Boolean WixDeleteMark (Widget widget, void* data)
{
	WixWidget	w = (WixWidget) widget;
	mark*		mp;

	if ((mp = getMarkFromData (w, data)) != 0) {
		if (mp [1].addr == NULL)
			mp -> addr = NULL;
		else
			mp -> addr = (unsigned char*) -1;
		return True;
	}
	return False;
}

/*
 *	Cree une nouvelle marque.
 */

Boolean WixCreateMark (Widget widget, void* data)
{
	WixWidget	w = (WixWidget) widget;
	mark*		mp;

	for (mp = W.b_marks+3; mp -> addr != 0; mp++)
		;

	if (mp == & W.b_marks [W.num_marks-1]) {
		W.b_marks = (mark*) XtRealloc ((char*)W.b_marks, (W.num_marks+NUM_MARKS) * sizeof (mark));
		mp = & W.b_marks [W.num_marks-1];
		bzero ((char*)mp, NUM_MARKS * sizeof (mark));
		W.num_marks += NUM_MARKS;
	}

	mp -> addr = W.b_cursor_e;
	mp -> data = data;
	return True;
}

static void adj_mark_del (WixWidget w, int nc)
{
	unsigned char*	e = W.b_cursor_e;
	mark*			mp;

	for (mp = W.b_marks; mp -> addr; mp++)
		if ((long) mp -> addr != -1 && mp -> addr >= e) {
			if (mp -> addr < e + nc)
				mp -> addr = e + nc;
		}
}

static void adj_mark_forw (WixWidget w, int n)
{
	unsigned char*	e = W.b_cursor_e;
	mark*			mp;

	for (mp = W.b_marks; mp -> addr; mp++)
		if ((long) mp -> addr != -1 && mp -> addr >= e && mp -> addr <= e + n)
			mp -> addr -= e - W.b_cursor_b;
}

static void adj_mark_back (WixWidget w, int n)
{
	unsigned char*	b = W.b_cursor_b;
	mark*			mp;

	for (mp = W.b_marks; mp -> addr; mp++)
		if ((long) mp -> addr != -1 && mp -> addr >= b - n && mp -> addr <= b)
			mp -> addr += W.b_cursor_e - b;
}

/*
 *	Initialise le mecanisme de sauvegarde.
 */

int InitKillers (void) {
	char*	tmp = "..ktmp..";

	FdKillFile = creat (tmp, 0600);	/* creer le fichier */
	if (FdKillFile < 0)				/* ca craint: tant pis */
		return 0;
	close (FdKillFile);
	FdKillFile = open (tmp, 2);		/* lecture & ecriture */
	unlink (tmp);					/* pour ne pas laisser */
	if (FdKillFile < 0)				/* trainer le fichier  la sortie */
		return 0;
	return 1;
}

/*
 *	Sauve un bloc de texte dans le cimetiere.
 */

static int storeInKillStack (unsigned char* string, long n)
{
	_KillTime = _LastActionTime;			/* noter l'heure */
	lseek (FdKillFile, EndOfKillStack, 0);	/* crire au bout */
	EndOfKillStack += write (FdKillFile, (char*) string, (int) n);
	EndOfKillStack += write (FdKillFile, (char *)&n, sizeof n);
	return n;				/* pointer vers le bloc */
							/* prcdent */
}

/*
 *	Copie la region entre le curseur et la marque `mk'
 *	dans la kill-stack.
 */

int KCopyRegion (WixWidget w, unsigned char* mk)
{
	register int n;

	if (mk == W.b_cursor_e)	/* region vide */
		return 0;		/* 0 caractres copies */

	if (mk < W.b_cursor_e)
		storeInKillStack (mk, n = W.b_cursor_b - mk);
	else
		storeInKillStack (W.b_cursor_e, n = mk - W.b_cursor_e);

	return n;
}

/*
 *	Insere un fichier dans la fentre courante.
 */

static void insert_fd (WixWidget w, int fd, int len)
{
	char*	string = XtMalloc (len);

	read (fd, string, len);
	KInsertString (w, string, len);
	XtFree (string);
}

/*
 *	Restaure un bloc precedemment sauv.
 */

char* GetKilled (int n, long* len)
{
	char*	buffer;
	long	faddr;

	lseek (FdKillFile, EndOfKillStack, 0);		/* aller au bout */
	while (--n >= 0) {
		lseek (FdKillFile, (long) -sizeof faddr, 1);
		if (sizeof faddr != read (FdKillFile, (char *)&faddr, sizeof faddr))
			return 0;
		if (-1 == lseek (FdKillFile, (long) -(faddr + sizeof faddr), 1))
			return 0;
	}
	buffer = XtMalloc (faddr+1);
	(void) read (FdKillFile, buffer, (int) faddr);
	buffer [faddr] = 0;
	*len = faddr;
	return buffer;
}

int Yank (WixWidget w, int n, void* arg)
{
	long faddr;

	lseek (FdKillFile, EndOfKillStack, 0);		/* aller au bout */
	while (--n >= 0) {
		lseek (FdKillFile, (long) -sizeof faddr, 1);
		if (sizeof faddr != read (FdKillFile, (char *)&faddr, sizeof faddr))
			return 0;

		if (-1 == lseek (FdKillFile, (long) -(faddr + sizeof faddr), 1))
			return 0;
	}
	insert_fd (w, FdKillFile, (int) faddr);
	return 1;
}

/*
 *	Cree un nouveau buffer.
 */

int BufferCreate (WixWidget w)
{
	register unsigned char*	cp;
	register Line*		l;
	/*
	 *	le buffer des fontes.
	 */
	cp = (unsigned char *) XtMalloc (DEF_BUFSIZE);
	if (cp == NULL)
		return 0;

	cp [0] = 0;
	cp [1] = 0;
	cp [DEF_BUFSIZE - 1] = 0;
	W.a_cursor_b = W.attr_beg = cp + 1;
	W.a_cursor_e = W.attr_end = cp + DEF_BUFSIZE - 1;
	/*
	 *	le buffer des caractres.
	 */
	cp = (unsigned char *) XtMalloc (DEF_BUFSIZE);
	if (cp == NULL) {
		XtFree (W.attr_beg - 1);
		return 0;
	}
	cp [0] = '\n';				/* un LF au debut */
	cp [DEF_BUFSIZE - 1] = '\n';		/* et un autre a la fin */

	W.beginning = cp + 1;		/* debut du buffer */
	W.end = cp + DEF_BUFSIZE - 1;	/* fin du buffer */
	/*
	 *	la liste des lignes.
	 */
	l = (Line*) XtMalloc (DEF_LINES * sizeof (Line));
	W.cur_line = l;
	W.first_line = l;
	W.last_line = l;
	W.top_line = l;
	W.num_slots = DEF_LINES;
	l -> chars = W.end;
	l -> fn = W.attr_end;
	l -> height = FONT (0) -> f_height;
	l -> base_line = FONT (0) -> f_ascent;
	l++; l -> chars = NULL;
	l++; l -> chars = NULL;
	l++; l -> chars = NULL;
	/*
	 *	les marques.
	 */
	W.num_marks = NUM_MARKS;
	W.b_marks = (mark *) XtMalloc (W.num_marks * sizeof (mark));
	bzero ((char*)W.b_marks, NUM_MARKS * sizeof (mark));
	W.b_marks [0].addr = W.end;
	W.b_marks [1].addr = W.beginning - 1; W.b_marks [1].data = (XtPointer) -1;
	W.b_marks [2].addr = W.beginning - 1; W.b_marks [2].data = (XtPointer) -1;

	*W.crypt_key = 0;
	W.cur_x = 0;
	W.cur_y = 0;
	W.hor_scroll = 0;
	W.b_cursor_b = W.beginning;
	W.b_cursor_e = W.end;
	W.refresh_from = 1000;
	W.refresh_to = 0;
	W.need_hpos = False;

	W.lineno = 0;
	W.nolines = 0;

	W.was_modified = False;
	W.is_modified = False;
	W.hscroll = False;
	W.focus_in = False;
	return 1;
}

/*
 *	Efface le contenu d'un buffer.
 */

void WixClearBuffer (Widget widget)
{
	WixWidget	w = (WixWidget) widget;
	Line*		l;

	bzero ((char*)W.b_marks, W.num_marks * sizeof (mark));
	W.b_marks [0].addr = W.end;
	W.b_marks [1].addr = W.beginning - 1; W.b_marks [1].data = (XtPointer) -1;
	W.b_marks [2].addr = W.beginning - 1; W.b_marks [2].data = (XtPointer) -1;

	W.a_cursor_b = W.attr_beg;	/* effacer les fontes */
	W.a_cursor_e = W.attr_end;

	l = W.first_line;
	W.cur_line = l;
	W.last_line = l;
	W.top_line = l;
	l -> chars = W.end;
	l -> fn = W.attr_end;
	l -> height = FONT (0) -> f_height;
	l -> base_line = FONT (0) -> f_ascent;
	l++; l -> chars = NULL;
	l++; l -> chars = NULL;
	l++; l -> chars = NULL;

	W.is_modified = False;
	W.b_cursor_b = W.beginning;
	W.b_cursor_e = W.end;
	*W.crypt_key = 0;
	W.fresh_selection = False;
	W.cur_x = 0;
	W.cur_col = 0;
	W.cur_y = 0;
	W.nolines = 0;
	W.lineno = 0;
	W.refresh_from = 0;
	W.refresh_to = 1000;
	W.need_hpos = False;
	W.hscroll = False;
}

/*
 *	Change la taille d'un buffer.
 */

static int bufferResize (WixWidget w, int new_size /* increment de taille */)
{
	unsigned char*	cp;
	unsigned char*	fp;
	unsigned char*	new_b;
	unsigned char*	new_cb;
	unsigned char*	new_ce;
	unsigned char*	new_e;
	mark*			mp;
	int				size_a;
	int				size_b;

	new_size += (W.end - W.beginning) + 2;	/* la nouvelle taille */

	if (! (fp = (unsigned char *) XtMalloc (new_size)))
		return 0;
	if (! (cp = (unsigned char *) XtMalloc (new_size))) {
		XtFree (fp);
		return 0;
	}
	/*
	 *	Recopier le buffer.
	 */
	size_b = W.b_cursor_b - W.beginning;
	size_a = W.end - W.b_cursor_e;

	bcopy (W.beginning - 1, cp, size_b + 1);
	bcopy (W.b_cursor_e, cp + new_size - size_a - 1, size_a + 1);
	XtFree (W.beginning - 1);
	/*
	 *	Recopier les fontes.
	 */
	bcopy (W.attr_beg - 1, fp, size_b + 1);
	bcopy (W.a_cursor_e, fp + new_size - size_a - 1, size_a + 1);
	XtFree (W.attr_beg - 1);
	/*
	 *	Corriger la structure buffer.
	 */
	new_b = cp + 1;			/* nouveau debut de buffer */
	new_e = cp + new_size - 1;	/* nouvelle fin de buffer */
	new_cb = new_b + size_b;
	new_ce = new_e - size_a;
	/*
	 *	Corriger la structure fontes.
	 */
	W.attr_beg = fp + 1;
	W.attr_end = fp + new_size - 1;
	W.a_cursor_b = W.attr_beg + size_b;
	W.a_cursor_e = W.attr_end - size_a;
	/*
	 *	Replacer les marques.
	 */
	for (mp = W.b_marks; mp -> addr; mp++)	/* pousser les marques */
		if (mp -> addr >= W.beginning && mp -> addr <= W.b_cursor_b)
			mp -> addr = new_b + (mp -> addr - W.beginning);
		else
		if (mp -> addr >= W.b_cursor_e && mp -> addr <= W.end)
			mp -> addr = new_ce + (W.b_cursor_e - mp -> addr);

	W.beginning = new_b;
	W.end = new_e;
	W.b_cursor_b = new_cb;
	W.b_cursor_e = new_ce;
	/*
	 *	refaire le tableau de lignes.
	 */
	MakeLines (w);	/* a optimiser */
	return 1;
}

/*
 *	Insere un caractre dans le buffer associe a la fentre t.
 */

int KInsertChar (WixWidget w, unsigned char c)
{
	unsigned char	tab [2];
	unsigned char*	cp;
	Line*			l;
	int				x;
	int				ret = 0;

	if (W.read_only)
		return ret;

	tab [0] = c;
	tab [1] = 0;
	if (!verifyCallback (w, INSERT_STRING, W.b_cursor_b - W.beginning, tab, 1))
		return ret;

	if (Bfree (w) < C.width * W.num_lines / Width (' ', FONT (0)))
		bufferResize (w, DEF_INCR);

	if (W.a_cursor_b == W.attr_beg)
		*W.a_cursor_b ++ = 0;					/* fonte par dfaut au dbut */
	else {
		W.a_cursor_b [0] = W.a_cursor_b [-1];	/* fonte du caractre prcdent */
		W.a_cursor_b ++;
	}
	PushInsert (w, W.b_cursor_b - W.beginning, &c, 1);
	*W.b_cursor_b ++ = c;
	*W.b_cursor_b = 0;	/* inutile ?? */
	W.is_modified = True;
	/*
	 *	Calculer la nouvelle colonne.
	 */
	switch (c) {
	case '\n':
		W.cur_x = 0;
		W.cur_col = 0;
		W.cur_y ++;
		W.nolines ++;
		W.lineno ++;
		if (W.b_marks [1].addr > W.b_cursor_b) W.b_marks [1].data = (char*) W.b_marks [1].data + 1;
		if (W.b_marks [2].addr > W.b_cursor_b) W.b_marks [2].data = (char*) W.b_marks [2].data + 1;
		l = W.cur_line;
		if (l -> chars > W.beginning) {
			int	trou = W.b_cursor_e - W.b_cursor_b + 1;

			l -> chars -= trou;
			l -> fn -= trou;
		}
		calc_height (w, l);
		insertLines (w, 1);
		l++;
		l -> chars = W.b_cursor_e;
		l -> fn = W.a_cursor_e;
		W.cur_line = l;
		/* des fois qu'on ait coup des mots clefs en deux ! */
		*--W.b_cursor_e = *--W.b_cursor_b;	/* remonter  la ligne prcdente vite fait */
		*--W.a_cursor_e = *--W.a_cursor_b;
		filterCurrentLine (w);				/* la colorier */
		++W.b_cursor_e; ++W.b_cursor_b;		/* retourner sur la nouvelle ligne */
		++W.a_cursor_e; ++W.a_cursor_b;
		filterCurrentLine (w);				/* la colorier aussi */
		calc_height (w, l);
		if (W.num_slots - W.nolines < 16)
			reallocLines (w, W.num_slots + LINE_INCR);
		return ret;

	default:
		if ((W.start_comment && index (W.start_comment, c)) ||
			(W.stop_comment && index (W.stop_comment, c)) ||
			c == W.string_delim)
		/* on a pu fait une modification qui a un effet sur tout le buffer */
			ret = FilterBuffer (w);
		else
		/* ou alors que sur la ligne courante */
			filterCurrentLine (w);
		MakeCurrentLine (w);	/* refait la ligne courante */
		if (W.wrap == False) {
			W.need_hpos = True;
			return ret;
		}
		GetHpos (w);
		W.cur_col = W.cur_x;
		W.need_hpos = False;
		x = C.width - (MARGINS + W.line_number_width) - Width ('m', FONT (0));
		if (W.cur_x < x)
			return ret;
		/* chercher le precedent separateur */
		cp = W.b_cursor_b - 1;
		for (;;) {
			if (*--cp == '\n')
				return ret;		/* ligne sans espaces */
			if (index (W.delim, *cp))
				break;
		}
		x = W.b_cursor_b - cp - 1;
		KBackwardChar (w, x);
		nlws (w, get_indent (w), NULL);
		KForwardChar (w, x);
		W.need_hpos = True;
		return ret;
	}
}

/*
 *	Insre une chane caractres dans le buffer.
 */

void KInsertString (WixWidget w, unsigned char* s, int len)
{
	unsigned char*	cp;
	unsigned char*	end;
	int				nb;

	if (W.read_only)
		return;

	if (W.wrap) {
		Boolean red = W.do_redisplay;	/* a va invalider les FilterBuffer */

		W.do_redisplay = False;
		nb = W.cur_y;
		if (W.refresh_from > nb)
			W.refresh_from = nb;

		while (--len >= 0)
			KInsertChar (w, *s++);

		W.refresh_to = (W.cur_y == nb) ? W.refresh_from : 1000;
		W.do_redisplay = red;
		return;
	}

	if (!verifyCallback (w, INSERT_STRING, W.b_cursor_b - W.beginning, s, len))
		return;

	if (Bfree (w) < (C.width * W.num_lines / Width (' ', FONT (0))) + len)
		bufferResize (w, DEF_INCR + len);
	/*
	 *	Recopier les fontes.
	 */
	cp = W.a_cursor_b;
	end = cp + len;
	while (cp < end)
		*cp++ = W.a_cursor_b [-1];
	W.a_cursor_b = end;
	/*
	 *	Insrer les caractres.
	 */
	PushInsert (w, W.b_cursor_b - W.beginning, s, len);
	cp = W.b_cursor_b;
	end = cp + len;

	for (nb = 0; cp < end;)
		if ((*cp++ = *s++) == '\n')
			nb ++;

	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;

	if (nb) {
		if (W.b_marks [1].addr > W.b_cursor_b) W.b_marks [1].data = (char*) W.b_marks [1].data + nb;
		if (W.b_marks [2].addr > W.b_cursor_b) W.b_marks [2].data = (char*) W.b_marks [2].data + nb;
		W.nolines += nb;	/* nb lignes de plus */
		W.lineno += nb;		/* curseur a la fin */
		W.cur_y += nb;		/* et donc plus bas dans la fentre */
		W.refresh_to = 1000;
		W.b_cursor_b = end;	/* mettre a jour le curseur */
		if (W.num_slots - W.nolines < 16)
			reallocLines (w, W.num_slots + nb + LINE_INCR);
		MakeLines (w);		/*  optimiser */
		(void) FilterBuffer (w);
		MakeLines (w);		/* pour refaire les hauteurs */
	} else {
		W.b_cursor_b = end;	/* mettre a jour le curseur */
		MakeCurrentLine (w);
		if (FilterBuffer (w))
			W.refresh_to = 1000;	/* a a pu terminer une chane ou un commentaire */
		else if (W.refresh_to < W.cur_y)
			W.refresh_to = W.cur_y;		/* ne rafficher que la ligne courante */
		calc_height (w, W.cur_line);	/* pour refaire les hauteurs */
	}
	W.need_hpos = True;
	W.is_modified = True;
}

/*
 *	Detruit `n' caractres de la fentre `t'.
 */

int KDeleteChar (WixWidget w, int n, int sv_flag)
{
	int				i;
	unsigned char*	cp;
	unsigned char*	end;

	if (W.read_only)
		return -1;

	if (!verifyCallback (w, DELETE_STRING, W.b_cursor_b - W.beginning, "", n))
		return -1;

	/*
	 *	Corriger la taille au besoin.
	 */
	i = W.end - W.b_cursor_e;
	if (n > i)
		n = i;
	if (n <= 0)
		return -1;	/* rien a detruire */
	/*
	 *	Detruire les informations de fontes.
	 */
	W.a_cursor_e += n;
	/*
	 *	Verifier si ca detruit au moins une ligne.
	 */
	cp = W.b_cursor_e;
	end = cp + n;
	PushDelete (w, W.b_cursor_b - W.beginning, cp, n);

	for (i = 0; cp < end;)
		if (*--end == '\n')
			i ++;

	adj_mark_del (w, n);
	W.b_cursor_e = cp + n;
	W.nolines -= i;	/* i lignes de moins */
	if (W.b_marks [1].addr > W.b_cursor_b) W.b_marks [1].data = (char*) W.b_marks [1].data - i;
	if (W.b_marks [2].addr > W.b_cursor_b) W.b_marks [2].data = (char*) W.b_marks [2].data - i;
	/*
	 *	Sauver au besoin dans le kill stack.
	 */
	if (sv_flag && W.kill_stack)
		storeInKillStack (cp, n);
	bzero (W.b_cursor_b, n);	/* pour voir */
	/*
	 *	Retourne 1 si on a detruit un \n 0 sinon.
	 */
	W.is_modified = True;
	if (i > 0) {
		deleteLines (w, i);
		for (end = cp+n; end > cp; ++cp)
			if ((W.start_comment && index (W.start_comment, *cp)) ||
				(W.stop_comment && index (W.stop_comment, *cp)) ||
				*cp == W.string_delim)
			{
				(void) FilterBuffer (w);
				MakeCurrentLine (w);
				return 1;
			}
		filterCurrentLine (w);
		MakeCurrentLine (w);
		return 1;
	} else {
		char	c = W.b_cursor_e [-1];	/* le caractre dtruit */
		int		ret = 0;

		if (n > 1) {
			ret = FilterBuffer (w);
			MakeCurrentLine (w);
			return ret;
		}

		if ((W.start_comment && index (W.start_comment, c)) ||
			(W.stop_comment && index (W.stop_comment, c)) ||
			c == W.string_delim)
			ret = FilterBuffer (w);
		else
			filterCurrentLine (w);
		MakeCurrentLine (w);
		return ret;
	}
}

/*
 *	Avance de `n' caractres.
 */

int KForwardChar (WixWidget w, int n)
{
	unsigned char*	cb = W.b_cursor_b;
	unsigned char*	ce = W.b_cursor_e;
	int				nb = 0;
	int				trou = ce - cb;
	Line*			l;

	if (n <= 0)
		return 0;
	if (n > W.end - ce)
		return 0;
	/*
	 *	Mettre a jour les fontes.
	 */
	if (n > 1) {
		bcopy (W.a_cursor_e, W.a_cursor_b, n);
		W.a_cursor_b += n;
		W.a_cursor_e += n;
	} else {
		*W.a_cursor_b++ = *W.a_cursor_e++;
	}
	/*
	 *	Mettre a jour les caractres.
	 */
	adj_mark_forw (w, n);

	for (l = W.cur_line; --n >= 0;)
		if ((*cb++ = *ce++) == '\n') {
			l -> chars -= trou;
			l -> fn -= trou;
			l++;
			nb ++;
		}
	W.cur_line = l;	/* nouvelle ligne courante */
	W.lineno += nb;	/* descente de nb lignes */

	W.b_cursor_b = cb;
	W.b_cursor_e = ce;
	W.cur_y += nb;
	return 1;
}

/*
 *	Recule de `n' caractres.
 */

int KBackwardChar (WixWidget w, int n)
{
	unsigned char*	cb = W.b_cursor_b;
	unsigned char*	ce = W.b_cursor_e;
	unsigned char*	ab = W.a_cursor_b;
	unsigned char*	ae = W.a_cursor_e;
	Line*			l;
	int				nb = 0;
	int				trou = ce - cb;

	if (n > cb - W.beginning)
		return 0;
	/*
	 *	Mettre a jour les caractres.
	 */
	adj_mark_back (w, n);

	for (l = W.cur_line; --n >= 0;)
		if ((*--ae = *--ab, (*--ce = *--cb)) == '\n') {
			l--;
			l -> chars += trou;
			l -> fn += trou;
			nb ++;
		}

	W.b_cursor_b = cb;
	W.b_cursor_e = ce;
	W.a_cursor_b = ab;
	W.a_cursor_e = ae;
	while ((*--ae = *--ab, (*--ce = *--cb)) != '\n')
		;	/* refaire la lignes courante */
	W.cur_line = l;	/* nouvelle ligne courante */
	W.lineno -= nb;	/* remontee de nb lignes */
	W.cur_y -= nb;
	return 1;
}

/*
 *	Charge le fichier fd.
 */

static void read_fd (WixWidget w, int fd, int len)
{
	unsigned char*	e;
	int				nb;
	mark*			mp;

	if (Bfree (w) < (C.width * W.num_lines / Width (' ', FONT (0))) + len)
		bufferResize (w, DEF_INCR + len);
	/*
	 *	Razer le buffer des fontes.
	 */
	W.a_cursor_b = W.attr_beg;
	W.a_cursor_e = W.attr_end - len;
	bzero (W.a_cursor_e, len);
	/*
	 *	Lire les caractres.
	 */
	W.b_cursor_b = W.beginning;
	e = W.end - len;
	W.b_cursor_e = e;
	if (len > 0)
		read (fd, e, len);

#ifdef CRYPT
	if (*W.crypt_key) {		/* decrypter */
		CryptSetup (W.crypt_key);
		CryptBuffer (e, e + len, e);
	}
#endif

	W.refresh_from  = 0;
	W.refresh_to = 1000;

	W.b_marks [0].addr = e;			/* remettre La marque */
	W.b_marks [1].addr = W.beginning - 1;	/* et la region */
	W.b_marks [2].addr = W.beginning - 1;	/* a vide */
	for (mp = W.b_marks + 3; mp -> addr; mp++)	/* detruire les marques */
		if ((long) mp -> addr != -1)
			mp -> addr = NULL;

	for (nb = 0; e < W.end; nb++)
		while (*e++ != '\n');	/* compter les lignes */

	if (W.num_slots - nb < 16)	/* agrandir le tableau au besoin */
		reallocLines (w, nb + LINE_INCR);

	(void) FilterBuffer (w);
	MakeLines (w);
	W.nolines = nb;
	W.lineno = 0;
	W.is_modified = False;
	W.cur_x = 0;
	W.cur_col = 0;
	W.cur_y = 0;
}

/*
 *	Sauve le fichier fd.
 */

static int write_fd (WixWidget w, int fd)
{
	int		len1;
	int		len2;
	char*	cp;
	int		ok;

	len1 = W.b_cursor_b - W.beginning;
	len2 = W.end - W.b_cursor_e;

	if (*W.crypt_key) {		/* decrypter */
		cp = XtMalloc (len1 + len2);
#ifdef CRYPT
		CryptSetup (W.crypt_key);
		CryptBuffer (W.beginning, W.b_cursor_b, cp);
		CryptBuffer (W.b_cursor_e, W.end, cp + len1);
#endif
		len1 += len2;	/* retourner la taille */
		ok = (len1 == write (fd, cp, len1)) ? len1 : -1;
		XtFree (cp);
	} else {
		len1 = write (fd, W.beginning, len1);
		ok = (len2 == write (fd, W.b_cursor_e, len2)) ? len1+len2 : -1;
	}
	if (ok >= 0)
		W.is_modified = False;
	return ok;
}

/*
 *	Charge le fichier `name'
 */

Boolean WixLoadFile (Widget widget, String name)
{
	WixWidget	w = (WixWidget) widget;
	struct stat	b;
	int			fd;

	fd = open (name, 0);
	if (fd < 0)
		return False;

	fstat (fd, &b);
	read_fd (w, fd, b.st_size);
	close (fd);
	return True;
}

/*
 *	Insere le fichier `name'
 */

Boolean WixInsertFile (Widget widget, String name)
{
	WixWidget	w = (WixWidget) widget;
	struct stat	b;
	int			fd;

	fd = open (name, 0);
	if (fd < 0)
		return False;

	if (-1 == fstat (fd, &b))	/* rcuprer la taille */
		b.st_size = 0;

	if (b.st_size > 0) {
		insert_fd (w, fd, b.st_size);
		close (fd);
	}
	return True;
}

/*
 *	Ecrit un fichier.
 */

int WixSaveFile (Widget widget, String name)
{
	WixWidget	w = (WixWidget) widget;
	int			fd;
	int			ok;

	fd = creat (name, 0666);
	if (fd > 0) {
		ok = write_fd (w, fd);
		close (fd);
		return ok;
	}
	return -1;
}

/*
 *	Rcupere les diffrents pointeurs.
 */

unsigned char* _GetBufferBeg (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.beginning;
}

unsigned char* _GetBufferEnd (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.end;
}

unsigned char* _GetCursorBeg (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.b_cursor_b;
}

unsigned char* _GetCursorEnd (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.b_cursor_e;
}

unsigned char* _GetAttCursorBeg (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.a_cursor_b;
}

unsigned char* _GetAttCursorEnd (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.a_cursor_e;
}

unsigned char* _GetAttributeBeg (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.attr_beg;
}

unsigned char* _GetAttributeEnd (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.attr_end;
}

/*
 *	Rcupere la position d'insertion.
 */

int WixGetInsertPosition (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.b_cursor_b - W.beginning;
}

/*
 *	Affecte la position d'insertion.
 */

void WixSetInsertPosition (Widget widget, int pos)
{
	WixWidget	w = (WixWidget) widget;
	int			curpos;

	curpos = W.b_cursor_b - W.beginning;
	if (pos > curpos)
		ForwardChar (w, pos - curpos, NULL);
	else
	if (pos < curpos)
		BackwardChar (w, curpos - pos, NULL);
}

/*
 *	Modifie la fonte du caractre courant.
 */

void WixSetFont (Widget widget, int fno)
{
	WixWidget	w = (WixWidget) widget;

	W.a_cursor_b [-1] = fno;
}

/*
 *	Retourne la colonne courante (en pixels).
 */

int WixGetColumn (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	if (W.need_hpos) {
		W.need_hpos = False;
		GetHpos (w);
	}
	return W.cur_x;
}

/*
 *	Retourne la largeur de la fentre.
 */

int WixGetWidth (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return C.width - (MARGINS + W.line_number_width);
}
