/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>

#include "Primitive.h"
#include "PrimitiveP.h"

#define	_PRIVATE_
#include "Wix.h"
#include "WixP.h"

static int check_bounds = 0;

/*
 *	Va a la ligne numero `count'.
 */

int WixGotoLine (Widget widget, int line)
{
	WixWidget	w = (WixWidget) widget;
	Line*		l = W.first_line + line - 1;

	if (l >= W.last_line)
		EndOfBuffer (w, 1, NULL);
	else if (l != W.cur_line)
		GotoMark (w, l -> chars);
	return 1;
}

int GotoLine (WixWidget w, int count, void* arg)
{
	if (count > 1)
		WixGotoLine ((Widget) w, count);
	return 1;
}

/*
 *	Recherche une chane en avant.
 *	algorithme de Boyer & Moore.
 */

/* pour ne tenir compte ni de la casse, ni des accents, et mapper le \n sur l'espace */

static unsigned char	_iso2ascii [256] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	'\b', '\t', ' ', '\r', 12, 13, 14, 15,
	16, 17, 18, 19, 20, 21, 22, 23,
	24, 25, 26, 27, 28, 29, 30, 31,
	' ', '!', '\"', '#', '$', '%', '&', '\'',
	'(', ')', '*', '+', ',', '-', '.', '/',
	'0', '1', '2', '3', '4', '5', '6', '7',
	'8', '9', ':', ';', '<', '=', '>', '?',
	'@', 'a', 'b', 'c', 'd', 'e', 'f', 'g',
	'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
	'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
	'x', 'y', 'z', '[', '\\', ']', '^', '_',
	'`', 'a', 'b', 'c', 'd', 'e', 'f', 'g',
	'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
	'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
	'x', 'y', 'z', '{', '|', '}', '~', '',
	128, 129, 130, 131, 132, 133, 134, 135,
	136, 137, 138, 139, 140, 141, 142, 143,
	144, 145, 146, 147, 148, 149, 150, 151,
	152, 153, 154, 155, 156, 157, 158, 159,
	160, 161, 163, 163, 164, 165, 166, 167,
	168, 169, 170, 171, 172, 173, 174, 175,
	176, 177, 178, 179, 180, 181, 182, 183,
	184, 185, 186, 187, 188, 189, 190, 191,
	'a', 'a', 'a', 'a', 'a', 'a', 'a', 'c',
	'e', 'e', 'e', 'e', 'i', 'i', 'i', 'i',
	'd', 'n', 'o', 'o', 'o', 'o', 'o', 'x',
	'o', 'u', 'u', 'u', 'u', 'y', 'b', 'b',
	'a', 'a', 'a', 'a', 'a', 'a', 'a', 'c',
	'e', 'e', 'e', 'e', 'i', 'i', 'i', 'i',
	'd', 'n', 'o', 'o', 'o', 'o', 'o', 'x',
	'o', 'u', 'u', 'u', 'u', 'y', 'b', 'b'
};

static unsigned char	_iso2iso [256] = {
	0, 1, 2, 3, 4, 5, 6, 7,
	8, 9, 10, 11, 12, 13, 14, 15,
	16, 17, 18, 19, 20, 21, 22, 23,
	24, 25, 26, 27, 28, 29, 30, 31,
	32, 33, 34, 35, 36, 37, 38, 39,
	40, 41, 42, 43, 44, 45, 46, 47,
	48, 49, 50, 51, 52, 53, 54, 55,
	56, 57, 58, 59, 60, 61, 62, 63,
	64, 65, 66, 67, 68, 69, 70, 71,
	72, 73, 74, 75, 76, 77, 78, 79,
	80, 81, 82, 83, 84, 85, 86, 87,
	88, 89, 90, 91, 92, 93, 94, 95,
	96, 97, 98, 99, 100, 101, 102, 103,
	104, 105, 106, 107, 108, 109, 110, 111,
	112, 113, 114, 115, 116, 117, 118, 119,
	120, 121, 122, 123, 124, 125, 126, 127,
	128, 129, 130, 131, 132, 133, 134, 135,
	136, 137, 138, 139, 140, 141, 142, 143,
	144, 145, 146, 147, 148, 149, 150, 151,
	152, 153, 154, 155, 156, 157, 158, 159,
	160, 161, 162, 163, 164, 165, 166, 167,
	168, 169, 170, 171, 172, 173, 174, 175,
	176, 177, 178, 179, 180, 181, 182, 183,
	184, 185, 186, 187, 188, 189, 190, 191,
	192, 193, 194, 195, 196, 197, 198, 199,
	200, 201, 202, 203, 204, 205, 206, 207,
	208, 209, 210, 211, 212, 213, 214, 215,
	216, 217, 218, 219, 220, 221, 222, 223,
	224, 225, 226, 227, 228, 229, 230, 231,
	232, 233, 234, 235, 236, 237, 238, 239,
	240, 241, 242, 243, 244, 245, 246, 247,
	248, 249, 250, 251, 252, 253, 254, 255,
};

unsigned char* _WixForwardSearch (WixWidget w, unsigned char* cp, unsigned char* string, int count)
{
	static	unsigned char	pattern [256];
	static	short			skip [256] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	};
	int				len = strlen (string) - 1;
	int				dec, n;
	unsigned char*	s;
	unsigned char*	map = W.case_sensitive ? _iso2iso : _iso2ascii;

	if (strcmp (pattern, string)) {
		for (s = pattern; *s; )
			skip [*s++] = -1;
		for (s = string; *s; s++)
			skip [map [*s]] = len - (s - string);
		strcpy (pattern, string);
	}

	while (-- count >= 0) {
		s = string + len;
		cp += len;
		n = 1;
		while (cp < W.end && s >= string) {
			if (*cp == *s || map [*cp] == map [*s]) {
				cp --;
				s --;
				n ++;
			} else {
				if ((dec = skip [map [*cp]]) == -1)
					dec = len+1;
				else if (dec < n)
					dec = n;
				cp += dec;
				s = string + len;
				n = 1;
			}
		}
		if (s >= string)
			return NULL;	/* echec */
		cp += len+2;		/* se mettre apres le pattern */
	}
	return cp;
}

int WixForwardSearch (WixWidget w, unsigned char* string, int count)
{
	unsigned char*	cp;

	cp = _WixForwardSearch (w, W.b_cursor_e, string, count);
	if (cp != NULL) {
		ForwardChar (w, cp - W.b_cursor_e, NULL);
		return 1;
	}
	return 0;
}

/*
 *	Recherche une chane en arriere.
 */

unsigned char* _WixBackwardSearch (WixWidget w, unsigned char* cp, unsigned char* string, int count)
{
	static	unsigned char	pattern [256];
	static	short		skip [256] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	};
	int			len = strlen (string) - 1;
	register int		dec, n;
	register unsigned char*	s;
	unsigned char*	map = W.case_sensitive ? _iso2iso : _iso2ascii;

	if (strcmp (pattern, string)) {
		for (s = pattern; *s; )
			skip [*s++] = -1;
		for (s = string + len; s >= string; s --)
			skip [map [*s]] = s - string;
		strcpy (pattern, string);
	}

	while (-- count >= 0) {
		s = string;
		cp -= len;
		n = 1;
		while (s <= string+len && cp > W.beginning)
		{
			if (map [*cp] == map [*s]) {
				cp ++;
				s ++;
				n ++;
			} else {
				if ((dec = skip [map [*cp]]) == -1)
					dec = len+1;
				else if (dec < n)
					dec = n;
				cp -= dec;
				if (cp < W.b_cursor_e && cp > W.b_cursor_b)
					cp -= W.b_cursor_e - W.b_cursor_b;
				s = string;
				n = 1;
			}
		}
		if (s <= string+len)
			return 0;	/* echec */
		cp -= len+1;	/* se mettre avant le pattern */
	}
	return cp;
}

int WixBackwardSearch (WixWidget w, unsigned char* string, int count)
{
	unsigned char*	cp = W.b_cursor_b - 1;

	if ((cp = _WixBackwardSearch (w, cp, string, count)) != 0) {
		BackwardChar (w, W.b_cursor_b - cp, NULL);
		return 1;
	}
	return 0;
}

/*
 *	Detruit n caracteres.
 */

int removeChar (WixWidget w, int n, Boolean save)
{
	int	err;
	int	h = W.cur_line -> height;

	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;
	err = KDeleteChar (w, n, save);
	if (err < 0)
		return 0;

	if (err > 0 || h != W.cur_line -> height)
		W.refresh_to = 1000;
	else {
		if (W.refresh_to < W.refresh_from)
			W.refresh_to = W.refresh_from;
	}
	return 1;
}

int DeleteChar (WixWidget w, int n, void* arg)
{
	int ret = removeChar (w, n, n > 1);	/* ne sauver que si on dtruit plusieurs caractres */

	W.need_hpos = True;
	return ret;
}

/*
 *	Detruit `n' caracteres a gauche du curseur.
 */

int RuboutChar (WixWidget w, int n, void* arg)
{
	if (BackwardChar (w, n, NULL)) {
		UpdateWindow (w);
		return DeleteChar (w, n, NULL);
	}
	return 0;
}

/*
 *	Insere une tabulation.
 */

int Tabulation (WixWidget w, int count, void* arg)
{
	return InsertChar (w, count, (char*)'\t');
}

/*
 *	Coupe une ligne en deux.
 */

int OpenSpace (WixWidget w, int count, void* arg)
{
	int	n;

	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;

	for (n = count; --n >=0 ;)
		KInsertChar (w, '\n');

	KBackwardChar (w, count);
	W.refresh_to = 1000;
	W.need_hpos = 1;
	return 1;
}

/*
 *	beg-line, open-space
 */

int OpenLine (WixWidget w, int count, void* arg)
{
	BeginningOfLine (w, 1, NULL);
	return OpenSpace (w, count, NULL);
}

/*
 *	Insere une ligne.
 */

int NewLine (WixWidget w, int count, void* arg)
{
	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;
	KInsertChar (w, '\n');
	W.refresh_to = 1000;
	return 1;
}

/*
 *	Insere le caractere c.
 */

int InsertChar (WixWidget w, int count, void* arg)
{
	int	c = (int) (long) arg;
	int	n = W.cur_y;
	int	h = W.cur_line -> height;
	int	ret = 0;

	if (W.refresh_from > n)	W.refresh_from = n;
	while (--count >= 0)
		ret = KInsertChar (w, c);
	if (n != W.cur_y || h != W.cur_line -> height || ret)
		W.refresh_to = 1000;
	else if (W.refresh_to < W.refresh_from)
		W.refresh_to = W.refresh_from;

	return 1;
}

/*
 *	Insere le numro de la ligne courante.
 */

int InsertLineNumber (WixWidget w, int count, void* arg)
{
	char	tc [64];

	if (W.refresh_from > W.cur_y)	W.refresh_from = W.cur_y;
	KInsertString (w, tc, sprintf (tc, "%d", W.lineno + 1));
	if (W.refresh_to < W.refresh_from)
		W.refresh_to = W.refresh_from;
	return 1;
}

/*
 *	Insere une chaine de caracteres.
 */

int InsertString (WixWidget w, int count, void* string)
{
	KInsertString (w, (char*) string, strlen ((char*)string));
	return 1;
}

void WixInsert (Widget widget, char* string)
{
	KInsertString ((WixWidget) widget, string, strlen (string));
}

/*
 *	Allume le fond d'une suite de caractres (sur la mme ligne).
 *	si 'length' est ngatif  gauche du curseur, sinon  droite.
 */

void WixHighlight (WixWidget w, int length)
{
	unsigned char*	b = W.a_cursor_b;
	unsigned char*	e = W.a_cursor_e;

	if (length > 0) {
		if (e + length > W.attr_end)
			return;			/* trop loin */
		while (--length >= 0)
			*e++ |= HILITED;
	} else {
		if (e + length < W.attr_beg)
			return;			/* trop loin aussi */
		length = -length;
		while (--length >= 0) {
			*--e |= HILITED;
			*--b |= HILITED;
		}
	}
	W.hilited = True;	/* on a allum quelques caractres */
	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;
	if (W.refresh_to < W.cur_y)
		W.refresh_to = W.cur_y;
}

/*
 *	teint le fond d'une suite de caractres (sur la mme ligne).
 *	si 'length' est ngatif  gauche du curseur, sinon  droite.
 */

void WixUnhighlight (WixWidget w, int length)
{
	unsigned char*	b = W.a_cursor_b;
	unsigned char*	e = W.a_cursor_e;

	if (!W.hilited)
		return;			/* on n'a rien  faire */

	if (length > 0) {
		if (e + length > W.attr_end)
			return;			/* trop loin */
		while (--length >= 0)
			*e++ &= ~HILITED;
	} else {
		if (e + length < W.attr_beg)
			return;			/* trop loin aussi */
		length = -length;
		while (--length >= 0) {
			*--e &= ~HILITED;
			*--b &= ~HILITED;
		}
	}
	if (W.refresh_from > W.cur_y)
		W.refresh_from = W.cur_y;
	if (W.refresh_to < W.cur_y)
		W.refresh_to = W.cur_y;
}

/*
 *	teint le fond des caractres du buffer.
 */

void WixUnhighlightAll (WixWidget w)
{
	unsigned char*	a;
	int				n = 0;
	int				y;

	if (!W.hilited)
		return;			/* on n'a rien  faire */
	/*
	 *	regarder avant le curseur.
	 */
	for (a = W.attr_beg; a < W.a_cursor_b; n++)
		if (*a++ & HILITED) {
			a [-1] &= ~HILITED;
			y = WixIsVisible (w, n);
			if (y >= 0) {
				if (W.refresh_from > y)	W.refresh_from = y;
				if (W.refresh_to < y)	W.refresh_to = y;
			}
		}
	/*
	 *	regarder la ligne courante.
	 */
	for (a = W.a_cursor_e - (W.b_cursor_e - W.cur_line -> chars); a < W.a_cursor_e; )
		*a++ &= ~HILITED;
	/*
	 *	regarder aprs le curseur.
	 */
	for (; a < W.attr_end; n++)
		if (*a++ & HILITED) {
			a [-1] &= ~HILITED;
			y = WixIsVisible (w, n);
			if (y >= 0) {
				if (W.refresh_from > y)	W.refresh_from = y;
				if (W.refresh_to < y)	W.refresh_to = y;
			}
		}

	W.hilited = False;	/* tout est teint */
}

/*
 *	Remplace une portion texte.
 *	si to < 0 remplace toute la fin du texte.
 */

void WixReplace (WixWidget w, int from, int to, char* str, int length)
{
	if (length < 0)
		length = strlen (str);

	WixSetInsertPosition ((Widget) w, from);
	if (to < 0)
		removeChar (w, W.end - W.b_cursor_e, False);
	else if (to > from)
		removeChar (w, to - from, False);
	if (length > 0)
		KInsertString (w, str, length);
}

/*
 *	Retire les n derniers blocs de la kill stack.
 */

int PopKillStack (WixWidget w, int n, void* arg)
{
	char*	buffer;
	long	faddr;

	lseek (FdKillFile, EndOfKillStack, 0);		/* aller au bout */
	while (--n >= 0) {
		lseek (FdKillFile, (long) -sizeof faddr, 1);
		if (sizeof faddr != read (FdKillFile, (char *)&faddr, sizeof faddr))
			return 0;
		if (-1 == lseek (FdKillFile, (long) -(faddr + sizeof faddr), 1))
			return 0;
	}
	EndOfKillStack = lseek (FdKillFile, 0, 1);	/* o on en est maintenant */
	return 1;
}

/*
 *	Intervertit les deux caracteres precedents.
 */

int TwiddleChars (WixWidget w, int count, void* arg)
{
	unsigned char*	p = W.b_cursor_b;
	int				pos = p - W.beginning;
	unsigned char	tc [4];

	if (p >= W.beginning + 2 && p [-1] != '\n' && p [-2] != '\n') {
		tc [0] = p [-1];
		tc [1] = p [-2];
		tc [2] = 0;
		WixReplace (w, pos - 2, pos, tc, 2);
	}
	return 1;
}

/*
 *	Avance de `n' pages.
 */

int ForwardPage (WixWidget w, int n, void* arg)
{
	register Line*	l = W.cur_line - W.cur_y;	/* W.top_line */
	register int	h;

	while (--n >= 0) {
		for (h = 0; h < C.height - MARGINS; l++) {
			if (l > W.last_line)
				return 0;
			h += l -> height;
		}
		l -= 2;
	}

	GotoMark (w, l -> chars);
	W.cur_y = 0;
	W.refresh_from = 0;
	W.refresh_to = 1000;
	return 1;
}

/*
 *	Recule de `n' pages.
 */

int BackwardPage (WixWidget w, int n, void* arg)
{
	register Line*	l = W.cur_line - W.cur_y;	/* W.top_line */
	register int	h;

	if (l == W.first_line)
		return 1;	/* on est deja au debut */

	while (--n >= 0) {
		for (h = C.height - MARGINS; h > MARGINS; --l) {
			if (l == W.first_line)
				goto ok;
			h -= l -> height;
		}
		l++;
	}
ok:
	GotoMark (w, l -> chars);
	W.cur_y = 0;
	W.refresh_from = 0;
	W.refresh_to = 1000;
	return 1;
}

/*
 *	Attend et traite les prochains Expose.
 */

void wait_for_expose (WixWidget w)
{
	XEvent		e;
	Display*	dpy = XtDisplay ((Widget)w);
	Window		win = XtWindow ((Widget)w);

	XSync (dpy, False);
	while (XCheckTypedWindowEvent (dpy, win, Expose, &e) ||
	       XCheckTypedWindowEvent (dpy, win, GraphicsExpose, &e))
	{
		e.type = Expose;
		XtDispatchEvent (&e);
	}
}

/*
 *	Calcule la position en pixels de la ligne n.
 */

static int line_to_pixel (WixWidget w, int n)
{
	int		i, h;
	Line*	l = W.cur_line - W.cur_y;	/* W.top_line */

	for (i = h = 0; i < n && l < W.last_line; i++, l++)
		h += l -> height;
	return h;
}

/*
 *	Met la ligne courante au debut de la fenetre.
 *	(special pour Marc Durocher)
 */

int CurlineTop (WixWidget w, int count, void* arg)
{
	int			h;
	Display*	dpy = XtDisplay ((Widget)w);

	if (W.cur_y == 0)
		return 1;	/* deja fait */

	h = line_to_pixel (w, W.cur_y);
	XCopyArea (dpy, XtWindow ((Widget)w), XtWindow ((Widget)w), W.text_GC, 
			MARGINS, MARGINS + h,
			C.width - 2*MARGINS, C.height - 2*MARGINS - h,
			MARGINS, MARGINS);
	XClearArea (dpy, XtWindow ((Widget)w),
			MARGINS, C.height - h - MARGINS,
			C.width - 2*MARGINS, h, True);
	W.cur_y = 0;
	wait_for_expose (w);
	return 1;
}

/*
 *	Met la ligne courante au milieu de la fentre
 *	(special pour Nicolas Roussel)
 */

int CurlineCenter (WixWidget w, int count, void* arg)
{
	if (W.cur_y != W.top_scroll && W.cur_line - W.first_line >= W.top_scroll) {
		W.cur_y = W.top_scroll;
		W.refresh_from = 0;
		W.refresh_to = 1000;
	}
	return 1;
}

/*
 *	Fait dfiler le texte vers le bas.
 */

int ScrollDown (WixWidget w, int n, void* arg)
{
	int			y;
	int			h;
	Display*	dpy = XtDisplay ((Widget)w);
	/*
	 *	y a-t-il assez de lignes devant?
	 */
	if ((W.cur_line - W.cur_y) - n < W.first_line)
		return 0;

	h = C.height - line_to_pixel (w, n) - MARGINS;	/* epaisseur restante */
	y = W.cur_y;
	BackwardLine (w, n, NULL);
	W.cur_y = y;
	/*
	 *	Scroll explicite a la souris donc
	 *	on optimise en reutilisant ce qu'il y a
	 *	deja dans la fenetre. (Ca evite de voir la
	 *	vague passer.
	 */
	if ((y = line_to_pixel (w, n)) < C.height - 2*MARGINS) {
		W.fresh_selection = False;	/* invalide la region */
		XCopyArea (dpy, XtWindow ((Widget)w), XtWindow ((Widget)w), W.text_GC, 
				MARGINS, MARGINS,
				C.width - 2*MARGINS, h+10,		/* +10 pour viterles bavures en bas */
				MARGINS, MARGINS + y);
	} else
		y = C.height - 2*MARGINS;
	XClearArea (dpy, XtWindow ((Widget)w),
			MARGINS, MARGINS,
			C.width - 2*MARGINS, y, True);
	wait_for_expose (w);
	W.refresh_from = 1000;
	W.refresh_to = 0;
	return 1;
}

/*
 *	Fait dfiler le texte vers le haut.
 */

int ScrollUp (WixWidget w, int n, void* arg)
{
	Display*	dpy = XtDisplay ((Widget)w);
	int			y = W.cur_y;
	int			h;
	int			found;
	int			chk = check_bounds;

	h = line_to_pixel (w, n);	/* hauteur o commence le bloc  scroller */

	check_bounds = 1;
	found = ForwardLine (w, n, NULL);
	check_bounds = chk;

	if (! found)
		return 0;

	W.cur_y = y;
	if (h < C.height - MARGINS) {
		W.fresh_selection = False;	/* invalide la region */
		XCopyArea (dpy, XtWindow ((Widget)w), XtWindow ((Widget)w), W.text_GC, 
				MARGINS, MARGINS + h,
				C.width - 2*MARGINS, C.height - 2*MARGINS - h,
				MARGINS, MARGINS);
	} else
		h = C.height - 2*MARGINS;

	XClearArea (dpy, XtWindow ((Widget)w),
			MARGINS, C.height - h - MARGINS,
			C.width - 2*MARGINS, h, True);
	wait_for_expose (w);
	W.refresh_from = 1000;
	W.refresh_to = 0;
	return 1;
}

/*
 *	Dcale le texte vers la droite de la largeur d'un ' '
 */

int ScrollRight (WixWidget w, int n, void* arg)
{
	int	hincr = Width (' ', FONT (0)) * n;

	Hscroll (w, W.hor_scroll - hincr);
	return 1;
}

/*
 *	Decale le texte vers la gauche de la largeur d'un ' '
 */

int ScrollLeft (WixWidget w, int n, void* arg)
{
	int	hincr = Width (' ', FONT (0)) * n;

	if (W.no_cursor || (W.cur_x >= W.hor_scroll + hincr))
		Hscroll (w, W.hor_scroll + hincr);
	return 1;
}

/*
 *	Avance de `n' lignes.
 */

int ForwardLine (WixWidget w, int n, void* arg)
{
	Line*	l = W.cur_line + n;

	if (l > W.last_line && (check_bounds || n == 1))
		return 0;

	RemoveCursor (w);
	n = (l <= W.last_line && l -> chars)
		? l -> chars - W.b_cursor_e
		: W.end - W.b_cursor_e;
	KForwardChar (w, n);
	ToHpos (w, False);
	return 1;
}

/*
 *	Recule de `n' lignes.
 */

int BackwardLine (WixWidget w, int n, void* arg)
{
	Line*	l = W.cur_line - n;

	if (l < W.first_line)
		return 0;

	RemoveCursor (w);
	KBackwardChar (w, W.b_cursor_b - l -> chars);
	ToHpos (w, False);
	return 1;
}

/*
 *	Avance de `n' caracteres.
 */

int ForwardChar (WixWidget w, int n, void* arg)
{
	RemoveCursor (w);
	if (KForwardChar (w, n)) {
		W.need_hpos = 1;
		return 1;
	}
	return 0;
}

/*
 *	Recule de `n' caracteres.
 */

int BackwardChar (WixWidget w, int n, void* arg)
{
	RemoveCursor (w);
	if (KBackwardChar (w, n)) {
		W.need_hpos = 1;
		return 1;
	}
	return 0;
}

/*
 *	Va en debut de ligne.
 */

int BeginningOfLine (WixWidget w, int count, void* arg)
{
	if (W.b_cursor_b [-1] != '\n') {
		RemoveCursor (w);
		KBackwardChar (w, W.b_cursor_e - W.cur_line -> chars);
	}
	W.cur_x = 0;
	W.cur_col = 0;
	return 1;
}

/*
 *	Va en fin de ligne.
 */

int EndOfLine (WixWidget w, int count, void* arg)
{
	register unsigned char *e = W.b_cursor_e;

	if (*e == '\n')
		return 1;

	RemoveCursor (w);
	while (*e++ != '\n')
		;
	KForwardChar (w, -- e - W.b_cursor_e);
	W.need_hpos = 1;
	return 1;
}

/*
 *	Echange le curseur et la marque.
 */

int ExchangePointMark (WixWidget w, int count, void* arg)
{
	register unsigned char	*m = W.b_marks [0].addr;

	if (m == W.b_cursor_e)
		return 1;

	W.b_marks [0].addr = W.b_cursor_e;
	if (m < W.b_cursor_e)
		BackwardChar (w, W.b_cursor_b - m, NULL);
	else
		ForwardChar (w, m - W.b_cursor_e, NULL);
	return 1;
}

/*
 *	Va sur la marque `m'
 */

int GotoMark (WixWidget w, unsigned char* m)
{
	if (m < W.b_cursor_b)
		BackwardChar (w, W.b_cursor_b - m, NULL);
	else if (m > W.b_cursor_e)
		ForwardChar (w, m - W.b_cursor_e, NULL);
	/* replacer le curseur correctement */
	GetHpos (w);
	W.cur_col = W.cur_x;
	W.need_hpos = 0;
	return 1;
}

/*
 *	Met la marque sous le curseur.
 */

int SetMark (WixWidget w, int count, void* arg)
{
	W.b_marks [0].addr = W.b_cursor_e;
	return 1;
}

/*
 *	Detruit la region.
 */

static int delete_region (WixWidget w, unsigned char* m)
{
	int n;

	if (m == W.b_cursor_e)	/* region vide */
		return 0;			/* 0 caracteres detruits */

	if (m < W.b_cursor_e)	/* si la marque avant le curseur */
							/* reculer de ce qu'il faut */
		BackwardChar (w, n = W.b_cursor_b - m, NULL);
	else
		n = m - W.b_cursor_e;

	(void) removeChar (w, n, True);	/* sauver mme un seul caractre */
	W.need_hpos = True;
	return n;
}

/*
 *	Copie la region.
 */

int CopyRegion (WixWidget w, int count, void* arg)
{
	return KCopyRegion (w, W.b_marks [0].addr);
}

/*
 *	Efface la slection.
 */

void WipeSelection (WixWidget w)
{
	int	n;

	if (W.b_marks [1].addr == W.b_cursor_e)
		removeChar (w, W.b_marks [2].addr - W.b_marks [1].addr, True);
	else if (W.b_marks [2].addr == W.b_cursor_e) {
		n = W.b_cursor_b - W.b_marks [1].addr;
		BackwardChar (w, n, NULL);
		UpdateWindow (w);
		removeChar (w, n, True);
	} else
		return;		/* TRAITER LE CAS OU LE CURSEUR N'EST PAS  UN BOUT!! */
	UpdateWindow (w);
}

/*
 *	Efface la region (ou la slection si elle est rcente).
 */

int WipeRegion (WixWidget w, int count, void* arg)
{
	return delete_region (w, W.b_marks [0].addr);
}

/*
 *	Copie la fin de la ligne.
 */

int CopyEndLine (WixWidget w, int count, void* arg)
{
	register unsigned char *cp;

	for (cp = W.b_cursor_e; *cp != '\n'; cp ++)
		;
	return KCopyRegion (w, cp);

}

/*
 *	Detruit n lignes.
 */

int DeleteLine (WixWidget w, int n, void* arg)
{
	unsigned char *mk;

	BeginningOfLine (w, 1, NULL);	/* aller en debut de ligne */
	mk = W.b_cursor_b;				/* mettre une marque */
	if (ForwardLine (w, n, NULL)) {	/* descendre de n lignes */
		BeginningOfLine (w, 1, NULL);	/* et les degager */
		return delete_region (w, mk);
	}
	return 0;
}

/*
 *	Dtruit le dbut de la ligne.
 */

int DeleteBeginningOfLine (WixWidget w, int n, void* arg)
{
	unsigned char *mk;

	mk = W.b_cursor_e;				/* mettre une marque */
	BeginningOfLine (w, 1, NULL);	/* aller au dbut de la ligne */
	return delete_region (w, mk);	/* et degager le tout */
}

/*
 *	Detruit la fin de la ligne.
 */

int DeleteEndLine (WixWidget w, int n, void* arg)
{
	unsigned char *mk;

	if (n > 1)
		return DeleteLine (w, n, NULL);

	mk = W.b_cursor_b;
	if (*W.b_cursor_e == '\n') {
		Yank (w, 1, NULL);
		KForwardChar (w, 1);
	} else
		EndOfLine (w, 1, NULL);

	return delete_region (w, mk);	/* et degager le tout */
}

/*
 *	Va au debut du buffer.
 */

int BeginningOfBuffer (WixWidget w, int count, void* arg)
{
	if (W.b_cursor_b <= W.beginning)
		return 0;

	RemoveCursor (w);
	KBackwardChar (w, W.b_cursor_b - W.beginning);
	W.cur_x = 0;
	W.cur_col = 0;
	return 1;
}

/*
 *	Va a la fin du buffer.
 */

int EndOfBuffer (WixWidget w, int count, void* arg)
{
	if (W.b_cursor_e >= W.end)
		return 0;

	RemoveCursor (w);
	KForwardChar (w, W.end - W.b_cursor_e);
	W.need_hpos = 1;
	return 1;
}

/*
 *	Positionne le curseur en (x, y) dans la fenetre.
 */

int GotoXY (WixWidget w, int x, int y)
{
	Line*			l = W.cur_line - W.cur_y;	/* W.top_line */
	unsigned char*	e;
	int				i;

	if (x < 0 || y < 0)
		return 0;

	x -= MARGINS + W.line_number_width;

	if (y >= C.height - 2*MARGINS)
		y = C.height - 2*MARGINS;

	for (i = 0; l < W.last_line; i++, l++) {
		y -= l -> height;
		if (y <= 0)
			break;
	}
	y = i - W.cur_y;

	if (y > 0) {
		/*
		 *	Avancer au debut de la bonne ligne.
		 */
		for (e = W.b_cursor_e; --y >= 0; ) {
			while (*e++ != '\n')
				;
			if (e >= W.end) {
				EndOfBuffer (w, 1, NULL);
				return 1;
			}
		}
		RemoveCursor (w);
		KForwardChar (w, e - W.b_cursor_e);
	} else if (y < 0) {
		/*
		 *	Reculer au debut de la bonne ligne.
		 */
		for (--y, e = W.b_cursor_b; ++y <= 0; ) {
			while (*--e != '\n')
				;
			if (e < W.beginning)
				break;
		}
		RemoveCursor (w);
		KBackwardChar (w, W.b_cursor_b - e - 1);
	} else {
		/*
		 *	Aller au debut de la ligne.
		 */
		RemoveCursor (w);
		BeginningOfLine (w, 1, NULL);
	}
	W.cur_col = x + W.hor_scroll;
	ToHpos (w, True);
	W.cur_col = W.cur_x;
	return 1;
}

/*
 *	Tout pour les parenthesages.
 */

#define	LP '('
#define	RP ')'
#define	LB '['
#define	RB ']'
#define	LC '{'
#define	RC '}'
#define	SLASH '/'
#define	STAR '*'
#define	DQUOTE '\"'
#define	QUOTE '\''

#define	STANDARD	0
#define	STRING		1
#define	QUOTED		2
#define	COMMENT		3
#define	COMMENT_CPP	4

/*
 *	Retourne le nombre d'accolades ouvertes et pas encore fermees
 *	jusqu'au curseur.
 */

static int block_level (WixWidget w, unsigned char** p_brace) {
	int				nb = 0;
	int				state = STANDARD;
	unsigned char*	cp;

	for (cp = W.beginning; cp < W.b_cursor_b; cp++) {
		switch (*cp) {
			case '\\':
				cp ++;
				continue;

			case '\"':
				if (state == STANDARD)
					state = STRING;			/* debut de chaine */
				else if (state == STRING)
					state = STANDARD;		/* fin de chaine */
				continue;

			case '\'':
				if (state == STANDARD)
					state = QUOTED;			/* debut de chaine */
				else if (state == QUOTED)
					state = STANDARD;		/* fin de chaine */
				continue;

			case '/':
				if (cp [-1] == '*' && state == COMMENT)
					state = STANDARD;		/* fin de commentaire C */
				else if (cp [-1] == '/' && state == STANDARD)
					state = COMMENT_CPP;	/* debut de commentaire C++ */
				continue;

			case '*':
				if (cp [-1] == '/' && state == STANDARD)
					state = COMMENT;		/* debut de commentaire C */
				continue;

			case '\n':
				if (state == COMMENT_CPP)
					state = STANDARD;		/* fin de commentaire C++ */
				continue;

			case '{':
				if (state == STANDARD) {
					if (nb == 0 && p_brace != NULL)
						* p_brace = cp;		/* o est l'ouvrante initiale */
					nb++;					/* un bloc de plus */
				}
				continue;

			case '}':
				if (state == STANDARD)
					--nb;					/* un bloc de moins */
				continue;
		}
	}
#if 1
	/* aller au dbut de la ligne */
	while (*--cp != '\n')
		;
	/* puis au dbut de la prcdente */
	while (*--cp != '\n')
		if (*cp == '{')
			return nb;	/* comme il y a une { le calcul est correct */

	for (++cp; *cp == ' ' || *cp == '\t'; ++cp)
		;	/* sauter les espaces */

	/* regarder si c'est un if, for, while, do */
	if (cp [0] == 'i' && cp [1] == 'f' &&
		(cp [2] < 'a' || cp [2] > 'z' || cp [2] < 'A' || cp [2] > 'Z'))
		return nb + 1;

	if (cp [0] == 'f' && cp [1] == 'o' && cp [2] == 'r' &&
		(cp [3] < 'a' || cp [3] > 'z' || cp [3] < 'A' || cp [3] > 'Z'))
		return nb + 1;

	if (cp [0] == 'w' && cp [1] == 'h' && cp [2] == 'i' && cp [3] == 'l' && cp [4] == 'e' &&
		(cp [5] < 'a' || cp [5] > 'z' || cp [5] < 'A' || cp [5] > 'Z'))
		return nb + 1;

	if (cp [0] == 'd' && cp [1] == 'o' &&
		(cp [2] < 'a' || cp [2] > 'z' || cp [2] < 'A' || cp [2] > 'Z'))
		return nb + 1;
#endif
	return nb;
}

/*
 *	R-indente une ligne en C, C++, JAVA
 */

#define	iskeyword(c) ((c)>='a' && (c)<='z')

int CIndentLine (WixWidget w, int count, void* arg)
{
	int	n;
	int	i;
	int	tabs, spaces;

	n = block_level (w, NULL);
	BeginningOfLine (w, 1, NULL);
	tabs = 0;
	spaces = 0;
	for (i = 0; ; i++)
		if (W.b_cursor_e [i] == ' ')
			spaces++;		/* compter les espaces */
		else
		if (W.b_cursor_e [i] == '\t')
			tabs++;			/* et les tabulations */
		else
			break;
	if (spaces == 0 && tabs == n)
		ForwardChar (w, n, NULL);	/* c'tait dj bien indent */
	else {
		(void) removeChar (w, i, False);	/* les dtruire */
		W.need_hpos = True;
		InsertChar (w, n, (char*) '\t');
	}
	UpdateWindow (w);
	return 1;
}

/*
 *	Trouve le 'right' fermant le prochain 'left'.
 *	C_or_CPP indique si on est dans un mode du genre C, C++ ou JAVA
 *	pour inclure le traitement des commentaires et des quotes.
 *
 *	Retourne 1 (et se place aprs 'right') si on a trouv,
 *	sinon 0 si pas trouv.
 */

int find_close (WixWidget w, int left, int right, Boolean C_or_CPP)
{
	int				nb;
	int				state;
	unsigned char*	cp = W.b_cursor_e;
	unsigned char*	end = W.end;

	/* chercher la prochaine chose ouvrante */
	while (*cp++ != left)
		if (cp >= end)
			return 0;	/* pas trouv */

	state = STANDARD;	/* on suppose etre en dehors d'un commentaire ou */
						/* d'une chane ou chose quote */
	for (nb = 1; cp < end; cp++) {
		switch (*cp) {
			case '\\':
				cp++;						/* sauter le suivant quel qu'il soit */
				continue;

			case '\"':
				if (state == STANDARD)
					state = STRING;			/* dbut de chane */
				else if (state == STRING)
					state = STANDARD;		/* fin de chane */
				continue;

			case '\'':
				if (!C_or_CPP)
					continue;				/* \'  est standard */

				if (state == STANDARD)
					state = QUOTED;			/* dbut de chane */
				else if (state == QUOTED)
					state = STANDARD;		/* fin de chane */
				continue;

			case '/':
				if (!C_or_CPP)
					continue;				/* ne signifie rien dans ce cas */

				if (cp [-1] == '*' && state == COMMENT)
					state = STANDARD;		/* fin de commentaire C */
				else if (cp [-1] == '/' && state == STANDARD)
					state = COMMENT_CPP;	/* dbut de commentaire C++ */
				continue;

			case '*':
				if (!C_or_CPP)
					continue;				/* ne signifie rien dans ce cas */

				if (cp [-1] == '/' && state == STANDARD)
					state = COMMENT;		/* dbut de commentaire C */
				continue;

			case '\n':
				if (state == COMMENT_CPP)
					state = STANDARD;		/* fin de commentaire C++ */
				continue;

			default:
				if (*cp == left) {
					if (state == STANDARD)
						nb++;					/* un bloc de plus */
					continue;
				}
				if (*cp == right && state == STANDARD) {
					if (--nb == 0) {
						/* Gagn: se placer aprs 'right' */
						W.need_hpos = True;
						ForwardChar (w, (cp - W.b_cursor_e) + 1, NULL);
						return 1;
					}
				}
				continue;
		}
	}
	return 0;	/* pas de match */
}

/*
 *	Trouve le 'left' ouvrant le prcdent 'right'.
 *	C_or_CPP indique si on est dans un mode du genre C, C++ ou JAVA
 *	pour inclure le traitement des commentaires et des quotes.
 *
 *	Retourne 1 (et se place sur 'left') si on a trouv,
 *	sinon 0 si pas trouv.
 */

int find_open (WixWidget w, int left, int right, Boolean C_or_CPP)
{
	int				nb;
	int				state;
	unsigned char*	cp = W.b_cursor_b;
	unsigned char*	start = W.beginning;

	/* chercher la prcdente chose fermante */
	while (*--cp != right)
		if (cp <= start)
			return 0;	/* pas trouv */

	state = STANDARD;	/* on suppose etre en dehors d'un commentaire ou */
						/* d'une chane ou chose quote */
	for (nb = 1; cp > start;) {
		switch (*--cp) {
			case '\"':
				if (cp [-1] == '\\')
					if (cp [-2] == '\\')
						cp -= 2;			/* \\" correspond a une fin de chane */
					else
						continue;			/* \"  est un " standard */

				if (state == STANDARD)
					state = STRING;			/* entre dans une chane */
				else if (state == STRING)
					state = STANDARD;		/* sortie d'une chane */
				continue;

			case '\'':
				if (!C_or_CPP)
					continue;				/* on n'est pas concern */

				if (cp [-1] == '\\')
					if (cp [-2] == '\\')
						cp -= 2;			/* \\' correspond a une fin de */
					else					/* constante caractre */
						continue;			/* \'  est un ' standard */

				if (state == STANDARD)
					state = QUOTED;			/* entre dans une chane */
				else if (state == QUOTED)
					state = STANDARD;		/* sortie d'une chane */
				continue;

			case '/':
				if (!C_or_CPP)
					continue;				/* ne signifie rien dans ce cas */

				if (cp [-1] == '*' && state == STANDARD)
					state = COMMENT;		/* entre dans un commentaire C */
				continue;

			case '*':
				if (!C_or_CPP)
					continue;				/* ne signifie rien dans ce cas */

				if (cp [-1] == '/' && state == COMMENT)
					state = STANDARD;		/* sortie d'un commentaire C */
				continue;

			case '\n':
				if (C_or_CPP && state == STANDARD) {
					/* aller au dbut de la ligne */
					unsigned char*	p = cp;
					while (*--p != '\n')
						;
					/*  la recherche d'un ventuel commentaire C++ */
					for (++p; p < cp; p++)
						if (p [0] == '/' && p [1] == '/')
							break;
					cp = p;					/* se mettre devant le commentaire */
				}
				continue;

			default:
				if (*cp == right) {
					if (state == STANDARD)
						nb++;					/* un bloc de plus */
					continue;
				}
				if (*cp == left && state == STANDARD) {
					if (--nb == 0) {
						/* Gagn: se placer aprs 'right' */
						W.need_hpos = True;
						BackwardChar (w, W.b_cursor_b - cp, NULL);
						return 1;
					}
				}
				continue;
		}
	}
	return 0;	/* pas de match */
}

int FindClosingParenthesis (WixWidget w, int count, void* arg)
{
	return find_close (w, LP, RP, True);
}

int FindClosingBrace (WixWidget w, int count, void* arg)
{
	return find_close (w, LC, RC, True);
}

int FindOpeningParenthesis (WixWidget w, int count, void* arg)
{
	return find_open (w, LP, RP, True);
}

int FindOpeningBrace (WixWidget w, int count, void* arg)
{
	return find_open (w, LC, RC, True);
}

/*
 *	Va au dbut de la fonction courante.
 */

int CBeginningOfFunction (WixWidget w, int count, void* arg)
{
	unsigned char*	bp;
	int				nb;

	nb = block_level (w, &bp);			/* retrouve l'accolade ouvrante */
	if (nb != 0) {
		EnableRedisplay ((Widget) w, False, True);
		GotoMark (w, bp);				/*  nous y place */
		find_open (w, LP, RP, True);	/* dbut des arguments */
		BackwardWord (w, 1, NULL);		/* nom de la fonction */
		EnableRedisplay ((Widget) w, True, True);
		return 1;
	}
	return 0;
}

/*
 *	Va  la fin de la fonction courante.
 */

int CEndOfFunction (WixWidget w, int count, void* arg)
{
	unsigned char*	bp;
	int				nb;

	nb = block_level (w, &bp);			/* retrouve l'accolade ouvrante */
	if (nb != 0) {
		EnableRedisplay ((Widget) w, False, True);
		GotoMark (w, bp);				/*  nous y place */
		find_close (w, LC, RC, True);	/* va  l'accolade fermante */
		EnableRedisplay ((Widget) w, True, True);
		return 1;
	}
	return 0;
}

/*
 *	Poireaute quelque temps.
 */

static void on_alarm () {}

static void _wait_ (unsigned long how)
{
	struct itimerval	value;
	struct itimerval	ovalue;

	signal (SIGALRM, on_alarm);

	how *= 1000;
	value.it_value.tv_sec = 0;	/* initialiser le timer */
	value.it_value.tv_usec = how;	/* pour alarmer le prochain */
	value.it_interval.tv_sec = 0;	/* cinqieme de seconde */
	value.it_interval.tv_usec = 0;

	setitimer (ITIMER_REAL, &value, &ovalue);
	pause ();
}

/*
 *	Fait flasher les parentheses.
 */

int Flasher (WixWidget w, int cnt, void* arg)
{
	int				ch = (int) (long) arg;
	int				c;
	unsigned char*	theMark;

	InsertChar (w, cnt, arg);	/* insrer le caractre */
	switch (ch) {
		case RP: c = LP; break;
		case RC: c = LC; break;
		case RB: c = LB; break;
		default: return 1;	/* ne flasher que les ()[]{} */
	}
	UpdateWindow (w);		/* afficher */
	theMark = W.b_cursor_e;		/* pour revenir */

	if (0 == find_open (w, c, ch, True))
		return 0;			/* ya pas de truc ouvrant */

	GetHpos (w);			/* find_open ne repositionne pas */
	W.cur_col = W.cur_x;		/* la colonne */
	W.need_hpos = False;

	if (W.cur_y < 0 || W.cur_x < W.hor_scroll || W.cur_x >= W.hor_scroll + C.width - (MARGINS + W.line_number_width)) {
		KForwardChar (w, theMark - W.b_cursor_e);
		W.need_hpos = True;
	} else {
		UpdateWindow (w);
		XSync (XtDisplay (w), False);
		_wait_ (250);	/* la belle procedure Wait traitant tous les vnements, du coup */
						/* il arrive que les caractres taps APRS le truc ouvrant */
						/* soient insr AVANT !!!! */
		GotoMark (w, theMark);
		UpdateWindow (w);
	}
	return 1;
}

/*
 *	Les commandes sur les mots.
 */

#define isupper(x) (((x)<='Z'&&(x)>='A')||((x)<=(unsigned char)'\337'&&(x)>=(unsigned char)'\300'))
#define islower(x) (((x)<='z'&&(x)>='a')||((x)<=(unsigned char)'\377'&&(x)>=(unsigned char)'\340'))
#define isword(w,x) (!index (W.delim, (x)))

static unsigned char *_forward_word (WixWidget w, unsigned char* cp, Boolean skpspc)
{
	unsigned char	c;

	while (c = *cp, !isword (w, c))	cp ++;	/* sauter les espaces */
	while (c = *cp, isword (w, c))	cp ++;	/* aller a la fin du mot */
	if (skpspc)
		while (c = *cp, !isword (w, c))	cp ++;	/* aller au dbut du mot suivant */
	return cp;
}

static unsigned char *_backward_word (WixWidget w, unsigned char* cp)
{
	unsigned char c;

	while (c = *--cp, !isword (w, c));	/* sauter les espaces */
	while (c = *--cp, isword (w, c));	/* aller au debut du mot */
	return cp;
}

int DeleteWord (WixWidget w, int cnt, void* arg)
{
	unsigned char *mk;

	mk = W.b_cursor_b;				/* mettre une marque */
	if (ForwardWord (w, cnt, NULL))	/* avancer de n mots */
		return delete_region (w, mk);	/* et tout degager */
	return 0;
}

int RuboutWord (WixWidget w, int cnt, void* arg)
{
	unsigned char *mk;

	mk = W.b_cursor_e;		/* mettre une marque */
	if (BackwardWord (w, cnt, NULL))
		return delete_region (w, mk);
	return 0;
}

int ForwardWord (WixWidget w, int cnt, void* arg)
{
	unsigned char *cp = W.b_cursor_e;
	unsigned char *end = W.end;

	while (-- cnt >= 0)
		if ((cp = _forward_word (w, cp, arg != 0)) >= end) {
			cp = end;
			break;
		}

	return ForwardChar (w, cp - W.b_cursor_e, NULL);
}

int BackwardWord (WixWidget w, int cnt, void* arg)
{
	unsigned char *cp = W.b_cursor_b;
	unsigned char *beg = W.beginning;

	while (-- cnt >= 0)
		if ((cp = _backward_word (w, cp)) <= beg) {
			cp = beg;
			break;
		}

	if (cp != beg)
		cp ++;		/* pour etre en debut de mot et pas avant */
	return BackwardChar (w, W.b_cursor_b - cp, NULL);
}

/*
 *	Minusculise un mot.
 */

int LowerCaseWord (WixWidget w, int cnt, void* arg)
{
	unsigned char	tc [1024];
	int				i = 0;
	int				nsp;
	unsigned char*	p = W.b_cursor_e;
	Boolean			flag = False;

	while (i < sizeof (tc) && isword (w, *p))
		tc [i++] = *p++;
	tc [i] = 0;
	/* compter les espaces suivant le mot */
	for (nsp = 0; !isword (w, *p); ++p)
		++nsp;

	for (p = tc; *p != 0; p++)
		if (isupper (*p)) {
			flag = True;
			*p |= 040;	/* mettre le bit 6 */
			continue;
		}

	if (flag) {
		WixReplace (w,
			W.b_cursor_b - W.beginning,
			W.b_cursor_b - W.beginning + i,
			tc, i);
		if (W.refresh_from > W.cur_y)
			W.refresh_from = W.cur_y;
		if (W.refresh_to < W.cur_y)
			W.refresh_to = W.cur_y;
		W.is_modified = True;
		if (nsp > 0)
			ForwardChar (w, nsp, NULL);	/* avancer au dbut du mot suivant */
	}
	return True;
}

/*
 *	Capitalise un mot.
 */

int UpperCaseWord (WixWidget w, int cnt, void* arg)
{
	unsigned char	tc [1024];
	int				i = 0;
	int				nsp;
	int				ch;
	unsigned char*	p = W.b_cursor_e;
	Boolean			flag = False;

	while (i < sizeof (tc) && isword (w, *p))
		tc [i++] = *p++;
	tc [i] = 0;
	/* compter les espaces suivant le mot */
	for (nsp = 0; !isword (w, *p); ++p)
		++nsp;

	for (p = tc; (ch = (int) *p) != 0; p++)
		if (islower (ch)) {
			flag = True;
			*p &= ~040;	/* retirer le bit 6 */
			continue;
		}

	if (flag) {
		WixReplace (w,
			W.b_cursor_b - W.beginning,
			W.b_cursor_b - W.beginning + i,
			tc, i);
		if (W.refresh_from > W.cur_y)
			W.refresh_from = W.cur_y;
		if (W.refresh_to < W.cur_y)
			W.refresh_to = W.cur_y;
		W.is_modified = True;
		if (nsp > 0)
			ForwardChar (w, nsp, NULL);	/* avancer au dbut du mot suivant */
	}
	return True;
}

/*
 *	Change la casse de l'initiale du mot courant.
 */

int CapitalizeWord (WixWidget w, int count, void* arg)
{
	unsigned char*	p = W.b_cursor_b;
	unsigned char*	b;
	unsigned char	c;
	int				ch;
	Boolean			flag = False;

	if (isword (w, *W.b_cursor_e)) {
		for (b = p = W.b_cursor_b; c = *--p, isword (w, c);)
			;
		++p;
		if (b == p)
			c = W.b_cursor_e [0];
		else
			c = *p;
		ch = (int) c;
		if (isupper (ch)) {
			flag = True;
			c |= 040;	/* transformer en bas de casse */
		}
		else if (islower (ch)) {
			flag = True;
			c &= ~040;	/* transformer en haut de casse */
		}
		if (flag) {
			WixReplace (w,
				p - W.beginning,
				p - W.beginning + 1,
				&c, 1);
			if (b == p)
				KBackwardChar (w, 1);	/* WixReplace laisse  la fin du texte remplac */
			else
				KForwardChar (w, b - p - 1);
			MakeCurrentLine (w);
			if (W.refresh_from > W.cur_y)
				W.refresh_from = W.cur_y;
			if (W.refresh_to < W.cur_y)
				W.refresh_to = W.cur_y;
			W.is_modified = True;
		}
	}
	return True;
}
