/*****************************************************************

  Copyright (c) 1989, 1999 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/Xatom.h>

#include "Primitive.h"
#include "PrimitiveP.h"

#define	_PRIVATE_
#include "Wix.h"
#include "WixP.h"

/*
 *	Retourne la colonne du taquet de tabulation suivant.
 */

static int getNextTabstop (WixWidget w, int column)
{
	int	i;

	for (;;) {
		for (i = 0; i < (int) XtNumber (W.tab_stops) && W.tab_stops [i] >= 0; i++)
			if (column < W.tab_stops [i]) {
				i = W.tab_stops [i];
				goto ok;
			}
		/*
		 *	Au bout du tableau, on repique le
		 *	dernier ecart.
		 */
		i = W.tab_stops [i-1] - ((i>1) ? W.tab_stops [i-2] : 0);
		i += column - (column%i);
		/*
		 *	pas de tabulation plus petite que 4 pixels.
		 */
		ok:
		if (i - column < 4) {
			column += 4;
			continue;
		}
		return i;
	}
}

/*
 *	Affiche le curseur d'une fenetre.
 *	on = 0:	efface le curseur
 *	on = 1:	affiche le curseur en creux
 *	on = 2:	affiche le curseur en plein
 */

static void curs_draw (WixWidget w, int on)
{
	Display*		dpy = XtDisplay ((Widget) w);
	Window			win = XtWindow ((Widget) w);
	int				col, line;
	Line*			l;
	int				base_line;
	int				c_width, c_fn;
	unsigned char	c;
	Pixel			fg, bg;

	if (!XtIsRealized ((Widget) w) || !W.do_redisplay)
		return;

	col = W.cur_x - W.hor_scroll + MARGINS + W.line_number_width;
	if (col < W.line_number_width)
		return;

	l = W.cur_line - W.cur_y;	/* W.top_line */
	for (line = MARGINS; l < W.cur_line; l++)
		line += l -> height;
	base_line = l -> base_line;
	/*
	 *	le caractere a droite du curseur.
	 */
	c = *W.b_cursor_e;
	c_fn = *W.a_cursor_e;
	if (c == '\t' || c == '\n')
		c = ' ';
	else if (c < ' ' || c == 0177 || ((c & 0200) && W.bin_mode))
		c = '\\';

	c_width = Width (c, FONT (c_fn));
	/*
	 *	les couleurs (basees sur le caractere a droite du curseur).
	 */
	if ((W.b_marks [1].addr <= W.b_cursor_e && W.b_cursor_e < W.b_marks [2].addr) ||
	    (W.b_marks [2].addr <= W.b_cursor_e && W.b_cursor_e < W.b_marks [1].addr))
	{
		fg = C.background_pixel;
		bg = COLOR (c_fn);
	} else {
		fg = COLOR (c_fn);
		if (c_fn & HILITED)
			bg = W.select_color;
		else
			bg = C.background_pixel;
	}
	/*
	 *	On y va.
	 */
	switch (on) {
	case 1:
	case 2:
		XSetForeground (dpy, W.text_GC, fg);
		XFillRectangle (dpy, win, W.text_GC, col, line, c_width, l -> height);
		XSetForeground (dpy, W.text_GC, bg);
		XSetFont (dpy, W.text_GC, FONT (c_fn) -> f_info -> fid);
		XDrawString (dpy, win, W.text_GC, col, line + base_line, &c, 1);
		if (on == 1) {
			XSetState (dpy, W.text_GC, fg ^ bg, bg, GXxor, AllPlanes);
			XFillRectangle (dpy, win, W.text_GC, col+1, line + 1, c_width - 2, l -> height - 2);
		}
		break;
	case 0:
		XSetForeground (dpy, W.text_GC, bg);
		XFillRectangle (dpy, win, W.text_GC, col, line, c_width, l -> height);
		XSetForeground (dpy, W.text_GC, fg);
		XSetFont (dpy, W.text_GC, FONT (c_fn) -> f_info -> fid);
		XDrawString (dpy, win, W.text_GC, col, line + base_line, &c, 1);
		break;
	}
	XSetState (dpy, W.text_GC, P.foreground, C.background_pixel, GXcopy, AllPlanes);
}

void RemoveCursor (WixWidget w)
{
	if (!W.no_cursor)
		curs_draw (w, 0);
}

void ShowCursor (WixWidget w)
{
	if (!W.hscroll && !W.no_cursor && C.sensitive && C.ancestor_sensitive) {
		if (W.need_hpos) {
			GetHpos (w);
			W.cur_col = W.cur_x;
			W.need_hpos = 0;
		}
		curs_draw (w, W.focus_in ? 2 : 1);
		/* prvenir l'application du changement de contexte (fonte/couleur) */
		if (W.state != *W.a_cursor_e) {
			W.state = (int) * W.a_cursor_e;
			XtCallCallbackList ((Widget) w, W.context_callback, (XtPointer) (long) W.state);
		}
	}
}

/*
 *	Affiche dans la fenetre `t' la chaine pointee par `line'
 *	a la ligne `lineno'.
 */

typedef struct {
	char*	chars;
	int	nchars;
	int	delta;
	int	fn;
} TextItem;

static void display_line (WixWidget w, Line* l, int lineno, Boolean withPixmap)
{
	Display*		dpy = XtDisplay ((Widget) w);
	Window			win = XtWindow ((Widget) w);
	int				fn;
	int				line_number = l - W.first_line + 1;
	unsigned char*	line = l -> chars;
	unsigned char*	ap = l -> fn;
	int				offset;
	int				fromRegion = -1;
	int				toRegion = -1;
	int				trou = W.b_cursor_e - W.b_cursor_b;
	int				charNum = line < W.b_cursor_b
								? line - W.beginning
								: line - W.beginning - trou;
	int				begSel = W.b_marks [1].addr >= W.b_cursor_e
								? W.b_marks [1].addr - W.beginning - trou
								: W.b_marks [1].addr - W.beginning;
	int				endSel = W.b_marks [2].addr >= W.b_cursor_e
								? W.b_marks [2].addr - W.beginning - trou
								: W.b_marks [2].addr - W.beginning;
	static Pixmap	pix = (Pixmap) 0;
	static int		pix_width = 0,
					pix_height = 0;
	TextItem		text_item [128];
	TextItem*		tip;
	TextItem*		tp;
	int				col, nchars;
	int				ocol, delta;
	unsigned char*	pl;
	unsigned char	tline [2048];
	unsigned char	c;
	int				tabInit = 0;

	if (pix_width < C.width || pix_height < l -> height) {
		if (pix)
			XFreePixmap (dpy, pix);
		pix_width = C.width;
		pix_height = l -> height;
		pix = XCreatePixmap (dpy, win, pix_width, pix_height, C.depth);
	}
	if (begSel > endSel) {
		nchars = endSel;
		endSel = begSel;
		begSel = nchars;
	}
	/*
	 *	Aller sur le premier caractere visible de la fenetre.
	 */
	for (col = 0; col < W.hor_scroll; line++, ap++, charNum++) {
		if (fromRegion < 0 && charNum >= begSel)
			fromRegion = 0;
		if (toRegion < 0 && charNum >= endSel)
			toRegion = 0;
		switch (*line) {
		case '\n':			/* ligne vide: rentrer */
			ocol = 0;
			tip = text_item;
			goto blank;
		case '\t':			/* tabulation */
			col = getNextTabstop (w, col);
			continue;
		default:
			c = *line;
			if (c < ' ' || c == 0177 || ((c & 0200) && W.bin_mode)) {
				tline [0] = '\\';
				tline [1] = '0' + ((c & 0300)>>6);
				tline [2] = '0' + ((c & 070)>>3);
				tline [3] = '0' + (c & 07);
				offset = CharsWidth (FONT (*ap), tline, 4);
			} else
				offset = Width (c, FONT (*ap));
			if (col + offset > W.hor_scroll)
				break;
			col += offset;
			continue;
		}
		break;
	}
	/*
	 *	Afficher la partie visible de la ligne.
	 */
	fn = *ap;
	pl = tline;
	col -= W.hor_scroll;
	ocol = col;
	tip = text_item;
	tip -> chars = (char*) pl;
	tip -> fn = *ap;
	tip -> delta = col;	/* colonne de depart */

	tabInit = (W.hor_scroll > 0 && line [-1] == '\t') /* tab scroll */
					? 0 : ocol;

	for (nchars = 0; col < (int) C.width - 2*MARGINS;) {
		switch (*pl = *line) {
		case '\n':
			if (fromRegion < 0 && charNum >= begSel)
				fromRegion = col;
			if (toRegion < 0 && charNum >= endSel)
				toRegion = col;
			break;
		case '\t':
			if (fromRegion < 0 && charNum >= begSel)
				fromRegion = col;
			if (toRegion < 0 && charNum >= endSel)
				toRegion = col;
			delta = getNextTabstop (w, col+W.hor_scroll) - (col+W.hor_scroll);
			col += delta;
			for (charNum++, line++, ap++; *line == '\t'; charNum++, line++, ap++) {
				if (fromRegion < 0 && charNum >= begSel)
					fromRegion = col;
				if (toRegion < 0 && charNum >= endSel)
					toRegion = col;
				{
					int i = getNextTabstop (w, col+W.hor_scroll) - (col+W.hor_scroll);
					delta += i;
					col += i;
				}
			}
			tip -> nchars = nchars;
			if (nchars != 0)
				tip++;	/* pas en debut de ligne */
			fn = *ap;
			tip -> chars = (char*) pl;
			tip -> fn = fn;
			tip -> delta = col;
			nchars = 0;
			continue;

		default:
			if (fn != *ap) {
				fn = *ap;
				tip++ -> nchars = nchars;
				tip -> chars = (char*) pl;
				tip -> fn = fn;
				tip -> delta = col;
				nchars = 0;
			}

			if (fromRegion < 0 && charNum >= begSel)
				fromRegion = col;
			if (toRegion < 0 && charNum >= endSel)
				toRegion = col;
			c = *pl;
			if (c < ' ' || c == 0177 || ((c & 0200) && W.bin_mode)) {
				*pl++ = '\\';
				*pl++ = '0' + ((c & 0300)>>6);
				*pl++ = '0' + ((c & 070)>>3);
				*pl++ = '0' + (c & 07);
				col += CharsWidth (FONT (*ap), pl-4, 4);
				nchars += 4;
			} else {
				col += Width (c, FONT (*ap));
				nchars++;
				pl++;
			}
			charNum++;
			line++;
			ap++;
			continue;
		}
		break;
	}
	/*
	 *	Afficher la ligne.
	 */
	tip -> nchars = nchars;		/* le dernier bout */
	tip ++;
	if (withPixmap) {
		XSetForeground (dpy, W.line_GC, C.background_pixel);
		XFillRectangle (dpy, pix, W.line_GC, 0, 0, C.width, l -> height);
		XSetForeground (dpy, W.line_GC, P.foreground);
		for (tp = text_item; tp < tip; tp++) {
			if (tp -> nchars == 0)
				continue;

			XSetForeground (dpy, W.line_GC, COLOR (tp -> fn));
			XSetFont (dpy, W.line_GC, FONT (tp -> fn) -> f_info -> fid);
			if (tp -> fn & HILITED) {
				XSetBackground (dpy, W.line_GC, W.select_color);
				XDrawImageString (dpy, pix, W.line_GC, tp -> delta,
									l -> base_line, tp -> chars, tp -> nchars);
			} else
				XDrawString (dpy, pix, W.line_GC, tp -> delta, l -> base_line,
								tp -> chars, tp -> nchars);
		}
		if (fromRegion < 0 && toRegion >= 0)
			fromRegion = 0;
		else if (fromRegion >= 0 && toRegion < 0)
			toRegion = C.width;
		if (toRegion > fromRegion && begSel > -1) {
			XSetState (dpy, W.line_GC, P.foreground ^ C.background_pixel, C.background_pixel, GXxor, AllPlanes);
			XFillRectangle (dpy, pix, W.line_GC, fromRegion, 0, toRegion - fromRegion, l -> height);
			XSetState (dpy, W.line_GC, P.foreground, C.background_pixel, GXcopy, AllPlanes);
		}
		XCopyArea (dpy, pix, win, W.text_GC, 0, 0, C.width, l -> height, MARGINS + W.line_number_width, lineno);
	} else {
	blank:
		if (fromRegion < 0 && toRegion >= 0)
			fromRegion = 0;
		else if (fromRegion >= 0 && toRegion < 0)
			toRegion = C.width;

		for (tp = text_item; tp < tip; tp++) {
			if (tp -> nchars == 0)
				continue;

			XSetForeground (dpy, W.text_GC, COLOR (tp -> fn));
			XSetFont (dpy, W.text_GC, FONT (tp -> fn) -> f_info -> fid);
			if (tp -> fn & HILITED) {
				XSetBackground (dpy, W.text_GC, W.select_color);
				XDrawImageString (dpy, win, W.text_GC,
								     tp -> delta + MARGINS + W.line_number_width,
								     lineno+l -> base_line,
								     tp -> chars,
								     tp -> nchars);
			} else
				XDrawString (dpy, win, W.text_GC,
						     tp -> delta + MARGINS + W.line_number_width,
						     lineno+l -> base_line,
				    		 tp -> chars,
				     tp -> nchars);
		}
		if (fromRegion < 0 && toRegion >= 0)
			fromRegion = 0;
		else if (fromRegion >= 0 && toRegion < 0)
			toRegion = C.width;
		if (toRegion > fromRegion && begSel > -1) {
			XSetState (dpy, W.text_GC, P.foreground ^ C.background_pixel, C.background_pixel, GXxor, AllPlanes);
			XFillRectangle (dpy, win, W.text_GC, tabInit + MARGINS + W.line_number_width + fromRegion, lineno, toRegion - fromRegion, l -> height);
			XSetState (dpy, W.text_GC, P.foreground, C.background_pixel, GXcopy, AllPlanes);
		}
	}
	/*
	 *	Afficher le numero de ligne.
	 */
	if (W.line_number_width > 0) {
		char	tc [16];

		sprintf (tc, "%d", line_number);
		XSetFont (dpy, W.text_GC, FONT (0) -> f_info -> fid);
		XSetForeground (dpy, W.text_GC, W.line_color);
		XDrawString (dpy, win, W.text_GC,
			MARGINS, lineno + l -> base_line, tc, strlen (tc));
		XSetForeground (dpy, W.text_GC, P.foreground);
	}
}

/*
 *	Calcule l'adresse (dans le buffer) du premier caractere
 *	en haut a gauche de la fenetre.
 *	Affecte cur_y avec la valeur idoine.
 */

static void get_top (WixWidget w)
{
	unsigned char*	b = W.b_cursor_b;
	int				nlines = 0;

	while (nlines ++ < W.top_scroll) {
		while (*--b != '\n')
			; /* rien */
		if (b <= W.beginning)
			break;
	}
	W.cur_y = --nlines;
}

/*
 *	Va  la position horizontale indique par cur_col, sachant
 *	que l'on est en dbut de ligne.
 */

void ToHpos (WixWidget w, Boolean before)
{
	unsigned char	tline [4];
	unsigned char*	e;
	unsigned char*	ae;
	int				col;
	int				incr;

	e = W.b_cursor_e;
	ae = W.a_cursor_e;

	for (col = 0; col < W.cur_col; e++, ae++, col += incr) {
		switch (*e) {
		case '\n':
			goto sortie;
		case '\t':			/* tabulation */
			incr = getNextTabstop (w, col) - col;
			break;
		default:
			if (*e < ' ' || *e == 0177 || ((*e & 0200) && W.bin_mode)) {
				tline [0] = '\\';
				tline [1] = '0' + ((*e & 0300)>>6);
				tline [2] = '0' + ((*e & 070)>>3);
				tline [3] = '0' + (*e & 07);
				incr = CharsWidth (FONT (*ae), tline, 4);
			} else
				incr = Width (*e, FONT (*ae));
			break;
		}
		if (before && col + incr >= W.cur_col)
			break;
	}
sortie:
	KForwardChar (w, e - W.b_cursor_e);
	W.cur_x = col;
}

/*
 *	Recalcule la position horizontale du curseur.
 */

int GetHpos (WixWidget w)
{
	unsigned char	tline [4];
	unsigned char*	p = W.cur_line -> chars;
	unsigned char*	ce = W.b_cursor_e;
	unsigned char*	a = W.cur_line -> fn;
	int				col = 0;

	if (p)
		for (;p < ce; p++, a++) {
			if (*p == '\t') {
				col = getNextTabstop (w, col);
			} else if (*p < ' ' || *p == 0177 || ((*p & 0200) && W.bin_mode)) {
				tline [0] = '\\';
				tline [1] = '0' + ((*p & 0300)>>6);
				tline [2] = '0' + ((*p & 070)>>3);
				tline [3] = '0' + (*p & 07);
				col += CharsWidth (FONT (*a), tline, 4);
			} else
				col += Width (*p, FONT (*a));
		}
	W.cur_x = col;
	return p ? p - W.cur_line -> chars : 0;
}

/*
 *	Retourne le numro de la ligne dans la fentre du 'pos-ime' caractre (-1 si en dehors)
 */

int WixIsVisible (WixWidget w, int pos)
{
	Line*			l = W.cur_line - W.cur_y;
	Line*			top = l;		/* top_line */
	unsigned char*	beg;
	int				h;

	if (pos < top -> chars - W.beginning && W.cur_y != 0)
		return -2;	/* trop haut */
	/*
	 *	chercher avant la ligne courante.
	 */
	for (h = 0; l < W.cur_line; l++, h += l -> height)
		if (pos <= l -> chars - W.beginning)
			return (l - top) - 1;
	/*
	 *	chercher aprs la ligne courante.
	 */
	beg = W.beginning + (W.b_cursor_e - W.b_cursor_b);
	for (; l < W.last_line && h < C.height; l++, h += l -> height)
		if (pos <= l -> chars - beg)
			return (l - top) - 1;

	return -1;		/* trop bas */
}

/*
 *	Calcule la position en pixels de la ligne n.
 */

static int l_line_to_pixel (WixWidget w, int n)
{
	int		i, h;
	Line*	l = W.cur_line - W.cur_y;	/* W.top_line */

	for (i = h = 0; i < n && l < W.last_line; i++, l++)
		h += l -> height;
	/* fin de buffer sans \n ? */
	if (i < n)
		h += l -> height;	/* devrait tre la hauteur de la fonte par dfaut */
	return h;
}

/*
 *	Raffiche la fentre de la ligne `from'  la ligne `to'.
 *	`withPixmap' indique si on utilise un pixmap pour viter le
 *	flash  chaque caractre.
 */

static void display_aux (WixWidget w, int from, int to, Boolean withPixmap)
{
	Display*	dpy = XtDisplay ((Widget) w);
	Window		win = XtWindow ((Widget) w);
	int			pix_from, pix_to;
	Line*		l;
	int			i;
	/*
	 *	Calculer les coordonnees en pixels.
	 */
	l = W.cur_line - W.cur_y;	/* W.top_line */
	pix_from = MARGINS;
	for (i = 0; i < from && l < W.last_line; i++, l++) {
		if (pix_from > C.height)
			return;					/* from est trop bas */
		pix_from += l -> height;
	}
	for (pix_to = pix_from; i <= to && pix_to <= C.height && l < W.last_line; i++, l++)
		pix_to += l -> height;

	if (l >= W.last_line || (pix_to + l -> height) > C.height)
		pix_to = C.height - MARGINS;
	/*
	 *	Effacer l'ecran.
	 */
	if (!withPixmap) {
		XSetForeground (dpy, W.text_GC, C.background_pixel);
		XFillRectangle (dpy, win, W.text_GC,
                        P.shadow_thickness,
						pix_from,
                        C.width - 2*P.shadow_thickness,
						pix_to - pix_from);
		XSetForeground (dpy, W.text_GC, P.foreground);
	}
	/*
	 *	Afficher la ligne de separation.
	 */
	if (W.line_number_width > 0)
		if (C.depth > 1) {
			XDrawLine (dpy, win, P.top_GC,
				W.line_number_width + MARGINS - W.margin - 2, pix_from,
				W.line_number_width + MARGINS - W.margin - 2, pix_to);
			XDrawLine (dpy, win, P.bot_GC,
				W.line_number_width + MARGINS - W.margin - 1, pix_from,
				W.line_number_width + MARGINS - W.margin - 1, pix_to);
		} else {
			XDrawLine (dpy, win, W.text_GC,
				W.line_number_width + MARGINS - W.margin - 1, pix_from,
				W.line_number_width + MARGINS - W.margin - 1, pix_to);
		}
	/*
	 *	Afficher la page.
	 */
	for (l = W.cur_line - W.cur_y + from; pix_from < pix_to && l <= W.last_line && l -> chars; l++) {
		display_line (w, l, pix_from, withPixmap);
		pix_from += l -> height;
	}
}

/*
 *	Reaffiche le contenu de la fenetre `t' de la ligne
 *	`from' incluse a la ligne `to' exclue.
 */

static Boolean check_hscroll (WixWidget w)
{
	int	x;
	int	hincr = W.tab_size;

	if (W.cur_x < W.hor_scroll) {
		while (W.cur_x < W.hor_scroll)
			W.hor_scroll -= hincr;
		if (W.hor_scroll < 0)
			W.hor_scroll = 0;
		return True;
	} else {
		x = W.hor_scroll + C.width - (MARGINS + W.line_number_width) - Width (' ', FONT (0));
		if (W.cur_x >= x) {
			while (W.cur_x >= x) {
				W.hor_scroll += hincr;
				x += hincr;
			}
			return True;
		}
	}
	return False;
}

void RefreshWindow (WixWidget w, int from, int to)
{
	int	ln = 1000;	/* plus bas que le bas de la fentre */

	if (from < 0)
		from = 0;
	if (to > ln)
		to = ln;
	/*
	 *	Scroll horizontal ?
	 */
	if (!W.hscroll && check_hscroll (w)) {
		from = 0;
		to = ln;
	}
	/*
	 *	Scroll vertical ?
	 */
	if (l_line_to_pixel (w, W.cur_y + 1) > C.height - 2*MARGINS) {
		W.cur_y = W.bot_scroll;
		from = 0;
		to = ln;
	}
	else
	if (W.cur_y < 0) {
		get_top (w);		/* calcule topchar et affecte la */
							/* valeur correcte  cur_y */
		from = 0;
		to = ln;
	}
	/*
	 *	Que reafficher ?
	 */
	if (from > to)
		return;		/* rien du tout */

	if (to < 0 || from > ln)
		return;		/* rien non plus */

	if (XtIsRealized ((Widget) w))
		display_aux (w, from, to, to == from);
}

/*
 *	Reaffiche le contenu d'une fenetre.
 */

void TextRedisplay (WixWidget w, int pix_from, int pix_to)
{
	Line*	l;
	int		p;

	if (pix_from > pix_to)
		return;

	if (W.line_number_width > 0)
		if (C.depth > 1) {
			XDrawLine (XtDisplay ((Widget) w), XtWindow ((Widget) w), P.top_GC,
				W.line_number_width + MARGINS - W.margin - 2, pix_from,
				W.line_number_width + MARGINS - W.margin - 2, pix_to);
			XDrawLine (XtDisplay ((Widget) w), XtWindow ((Widget) w), P.bot_GC,
				W.line_number_width + MARGINS - W.margin - 1, pix_from,
				W.line_number_width + MARGINS - W.margin - 1, pix_to);
		} else {
			XDrawLine (XtDisplay ((Widget) w), XtWindow ((Widget) w), W.text_GC,
				W.line_number_width + MARGINS - W.margin - 1, pix_from,
				W.line_number_width + MARGINS - W.margin - 1, pix_to);
		}
	pix_from -= MARGINS;
	pix_to -= MARGINS;

	l = W.cur_line - W.cur_y;	/* W.top_line */
	for (p = 0; p + l -> height < pix_from; l++) {
		p += l -> height;
		if (l >= W.last_line)
			return;
	}
	/*
	 *	Afficher la page.
	 */
	while (p < pix_to && l -> chars) {
		display_line (w, l, p + MARGINS, False);
		p += l -> height;
		l++;
	}
}

/*
 *	Mise a jour de la fenetre apres une commande.
 */

void WixSetHorizontalScroll (Widget widget, int h)
{
	WixWidget	w = (WixWidget) widget;

	W.hor_scroll = h;
}

int WixGetHorizontalScroll (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.hor_scroll;
}

void WixSetRefreshArea (Widget widget, int from, int to)
{
	WixWidget	w = (WixWidget) widget;

	W.refresh_from = from;
	W.refresh_to = to;
}

int WixGetNumLines (Widget widget) {
	WixWidget	w = (WixWidget) widget;

	return W.nolines;
}

int WixGetY (Widget widget) {
	WixWidget	w = (WixWidget) widget;

	return W.cur_y;
}

void WixRedraw (Widget w)
{
	UpdateWindow ((WixWidget) w);
}

int EnableRedisplay (Widget widget, Boolean b, Boolean filter)
{
	WixWidget	w = (WixWidget) widget;

	if (b == W.do_redisplay)
		return 0;

	if (b) {
		W.do_redisplay = True;	/* doit etre AVANT UpdateWindow */
		if (filter)
			FilterBuffer (w);		/* refiltrer le buffer car il est invalid sans redisplay */
		UpdateWindow (w);
	} else {
		RemoveCursor (w);
		W.do_redisplay = False;	/* doit etre APRES RemoveCursor */
	}
	return 0;
}

void UpdateWindow (WixWidget w)
{
	if (W.need_hpos) {
		GetHpos (w);
		W.cur_col = W.cur_x;
		W.need_hpos = 0;
	}
	if (!W.do_redisplay)
		return;		/* reaffichage bloque */
	/*
	 *	Rafrachir la fentre.
	 */
	RefreshWindow (w, W.refresh_from, W.refresh_to);
	/*
	 *	Regarder s'il y a eu des modifs.
	 */
	if (W.was_modified ^ W.is_modified) {
		W.was_modified = W.is_modified;
		XtCallCallbackList ((Widget) w, W.modify_callback, (XtPointer) (long) W.is_modified);
		UpdateScrollbar (w, True);
	} else
		UpdateScrollbar (w, False);
	/*
	 *	croiser from et to pour initialier une
	 *	zone de reaffichage vide.
	 */
	W.refresh_from = 1000;
	W.refresh_to = 0;
	if (XtIsRealized ((Widget) w))
		ShowCursor (w);
}

/*
 *	Tout pour la selection.
 */

static unsigned char*	prev_e;
extern Time				_SelectTime;

/*
 *	Allume la selection.
 */

static void refreshSelect (WixWidget w, int y)
{
	int	py = W.prev_y;

	if (y == py) {
		display_aux (w, py, py, True);
	}

	else if (y > py) {
		display_aux (w, py, y, False);
	}

	else {	/* if (y < py) */
		display_aux (w, y, py, False);
	}

	W.prev_y = y;
}

/*
 *	Eteint la selection.
 */

static void clearRegion (Widget widget, Atom* atom)
{
	WixWidget	w = (WixWidget) widget;
	int			from;
	int			to;

	if (W.b_marks [1].addr == W.beginning - 1 || W.b_marks [2].addr == W.beginning - 1)
		return;

	W.b_marks [1].addr = W.beginning - 1;
	W.b_marks [2].addr = W.beginning - 1;

	if (!XtIsRealized ((Widget) w))
		return;		/* arrive quand on ferme la fenetre */
					/* qui contenait la selection */

	from = (long) W.b_marks [1].data - (W.lineno - W.cur_y);
	if (from < 0)
		from = 0;

	to = (long) W.b_marks [2].data - (W.lineno - W.cur_y);
	if (to < 0)
		return;

	if (to >= from)
		display_aux (w, from, to, to == from);
}

void ClearRegion (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	W.fresh_selection = False;
}

/*
 *	Appelee par le reclameur de selection.
 */

#if 1
static char* stockage = 0;
static int   stockage_len = 0;
#endif

static Boolean set_sel (Widget widget, Atom* sel, Atom* target, Atom* type,
						XtPointer* value, unsigned long* length, int* format)
{
	if (*target == XA_STRING) {
		*type = XA_STRING;
#if 1
		*((char**)value) = stockage;
		*format = 8;
		*length = stockage_len;
#else
		*((char**)value) = XFetchBytes (XtDisplay (widget), (int*) length);
		*format = 8;
#endif
		return True;
	}
	return False;
}

/*
 *	Affecte la selection.
 */

#if 1
static void selectionDone (Widget widget, Atom* sel, Atom* target)
{
	;
}
#endif

void WixSetSelection (Widget widget, XButtonEvent* e, char* string, int length)
{
	if (length > 0 && XtOwnSelection (widget, XA_PRIMARY, e -> time, set_sel, clearRegion, selectionDone))
	{
		_SelectTime = e -> time;
#if 1
		if (stockage != 0)
			XtFree (stockage);
		stockage = XtMalloc (length+1);
		bcopy (string, stockage, length);
		stockage_len = length;
#else
		XStoreBytes (XtDisplay (widget), string, length);
#endif
	}
}

/*
 *	Compte le nombre de clicks.
 */

static long	lastButtonUpTime = 0;
static int	numberOfClicks;

static void countClicks (XButtonEvent* e)
{
	if (((long) e -> time - lastButtonUpTime) > XtGetMultiClickTime (e -> display))
		numberOfClicks = 1;
	else
		numberOfClicks += 1;
}

/*
 *	Augmente la selection.
 */

void extend_selection (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	unsigned char*	p;
	int				n;

#if 0
	if (W.b_marks [1].addr == W.beginning - 1)
		fprintf (stderr, "ya pas de debut\n");
#endif
	if (GotoXY (w, e -> x, e -> y)) {
		if (W.b_cursor_e == prev_e)
			return;
		prev_e = W.b_cursor_e;
		if (W.b_cursor_e > W.b_marks [1].addr) {
			W.prev_y = W.cur_y - (W.lineno - (long) W.b_marks [2].data);
			if (W.prev_y < 0) W.prev_y = 0;
			W.b_marks [2].addr = W.b_cursor_e;
			W.b_marks [2].data = (XtPointer) (long) W.lineno;
			p = W.b_marks [1].addr;
			n = W.b_cursor_b - p;
		} else {
			W.prev_y = W.cur_y - (W.lineno - (long) W.b_marks [1].data);
			if (W.prev_y >= W.num_lines) W.prev_y = W.num_lines -1;
			W.b_marks [1].addr = W.b_cursor_e;
			W.b_marks [1].data = (XtPointer) (long) W.lineno;
			p = W.b_marks [1].addr;
			n = W.b_marks [2].addr - p;
		}
		refreshSelect (w, W.cur_y);
		ShowCursor (w);

		WixSetSelection (widget, e, p, n);
	}
}

/*
 *	Commence une slection
 */

static int	button_number = -1;			/* pour eviter les MotionNotify sans ButtonPress */

void selection_start (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	XtAppContext	app;
	XEvent			evt;

	countClicks (e);
	button_number = e -> button;

	switch (numberOfClicks) {
	case 1:
		/*
		 *	Si des Expose suivent l'appui du bouton, les traiter AVANT le ButtonPress.
		 *	(cas de mwm ou on clique dans la fenetre pour la faire passer devant).
		 */
		app = XtWidgetToApplicationContext (widget);
		XtAppPeekEvent (app, &evt);
		while (evt.type == Expose) {
			XtAppNextEvent (app, &evt);
			XtDispatchEvent (&evt);
			XtAppPeekEvent (app, &evt);
		}
		clearRegion ((Widget) w, NULL);
		GotoXY (w, e -> x, e -> y);
		W.b_marks [1].addr = W.beginning - 1;
		W.b_marks [2].addr = W.beginning - 1;
		RemoveCursor (w);
		W.b_marks [1].addr = W.b_cursor_e;
		W.b_marks [1].data = (XtPointer) (long) W.lineno;
		W.b_marks [2].data = (XtPointer) (long) W.lineno;	/* pour le cas ou on ne fait que cliquer */
		prev_e = W.b_cursor_e;
		W.prev_y = W.cur_y;
		W.no_cursor = True;
		return;

	case 2:
		select_word (widget, event, params, num_params);
		W.no_cursor = True;
		return;

	case 3:
		select_line (widget, event, params, num_params);

	default:	/* numberOfClicks >= 3 */
		W.no_cursor = True;
		return;
	}
}

/*
 *	Etire l'elastique.
 */

static XtIntervalId	timer = 0;	/* timer pour la selection automatique */
static XMotionEvent	the_event;

static void scroll_selection (XtPointer client, XtIntervalId* id)
{
	Widget			widget = (Widget) client;
	XtAppContext	app = XtWidgetToApplicationContext (widget);

	timer = XtAppAddTimeOut (app, 100, scroll_selection, widget);
	extend_adjust (widget, (XEvent*) &the_event, NULL, NULL);
}

void extend_adjust (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XMotionEvent*	e = (XMotionEvent*) event;
	XtAppContext	app = XtWidgetToApplicationContext (widget);
	Boolean			remove = True;

	if (button_number < 0)
		return;	/* pas le bon bouton */

	if (e -> y >= (int) C.height) {
		remove = False;
		if (& the_event != e) {
			if (!timer)
				timer = XtAppAddTimeOut (app, 100, scroll_selection, (XtPointer) w);
			the_event = *e;
			e -> y = C.height;
		}
		if (ScrollUp (w, 1, NULL)) {
			W.prev_y --;
		}
	}
	else if (e -> y <= 0) {
		remove = False;
		if (& the_event != e) {
			if (!timer)
				timer = XtAppAddTimeOut (app, 100, scroll_selection, (XtPointer) w);
			e -> x = 0;
			e -> y = 0;
			the_event = *e;
		}
		if (ScrollDown (w, 1, NULL)) {
			W.prev_y ++;
		}
	}
	if (GotoXY (w, e -> x, e -> y)) {
		if (W.b_cursor_e == prev_e)
			return;
		if (remove && timer) {
			XtRemoveTimeOut (timer);
			timer = 0;
		}
		/* extension de slection par mot */
		if (numberOfClicks == 2 && !index (W.delim, *W.b_cursor_e))
			ForwardWord (w, 1, NULL);
		/* extension de slection par ligne complte */
		else if (numberOfClicks == 3)
			BeginningOfLine (w, 1, NULL);

		prev_e = W.b_cursor_e;
		W.b_marks [2].addr = W.b_cursor_e;
		W.b_marks [2].data = (XtPointer) (long) W.lineno;
		refreshSelect (w, W.cur_y);
	}
}

/*
 *	Termine l'etirage.
 */

void extend_end (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	unsigned char*	p;
	int				n;

	lastButtonUpTime = e -> time;
	if (timer) {
		XtRemoveTimeOut (timer);
		timer = 0;
	}

	if (button_number < 0)
		return;
	button_number = -1;	/* le bouton est relache */

	if (W.b_marks [1].addr > W.b_cursor_e) {
		W.b_marks [2].addr = W.b_marks [1].addr;
		W.b_marks [2].data = W.b_marks [1].data;
		W.b_marks [1].addr = W.b_cursor_e;
		W.b_marks [1].data = (XtPointer) (long) W.lineno;

		p = W.b_cursor_e;
		n = W.b_marks [2].addr - W.b_cursor_e;
	} else {
		W.b_marks [2].addr = W.b_cursor_e;
		W.b_marks [2].data = (XtPointer) (long) W.lineno;

		p = W.b_marks [1].addr;
		n = W.b_cursor_b - W.b_marks [1].addr;
	}
	/*
	 *	Selectionner le texte si yen a ou
	 *	invalider la selection sinon.
	 */
	if (n > 0) {
		W.fresh_selection = True;
		WixSetSelection (widget, e, p, n);
	} else {
		W.fresh_selection = False;
		W.b_marks [1].addr = W.beginning - 1;
		W.b_marks [2].addr = W.beginning - 1;
	}
	/*
	 *	Reafficher le curseur.
	 */
	W.no_cursor = False;
	W.hscroll = False;
	UpdateWindow (w);
}

/*
 *	Scrolle la fenetre.
 */

void scroll_window (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XMotionEvent*	e = (XMotionEvent*) event;
	static int		prev_line;
	static int		prev_col;
	static int		vh;
	int				cur_line;
	int				fh = FONT (0) -> f_height;
	int				hincr = Width (' ', FONT (0));
	/*
	 *	Debut de drag;
	 */
	if (e -> type == ButtonPress) {
		prev_line = e -> y / fh;
		prev_col = e -> x;
		vh = 0;
		XDefineCursor (XtDisplay (widget), XtWindow (widget), W.motion_cursor);
		return;
	}
	/*
	 *	Fin de drag.
	 */
	if (e -> type == ButtonRelease) {
		prev_line = -12345;
		prev_col = -12345;
		XDefineCursor (XtDisplay (widget), XtWindow (widget), P.cursor);
		ShowCursor (w);
		return;
	}
	/*
	 *	En cours de drag.
	 */
	if (prev_line == -12345 || prev_col == -12345)
		return;		/* deplacement sans bouton appuye... */
	/*
	 *	Scroll horizontal si abs (delta) > hincr
	 */
	if (vh >= 0 && e -> x != prev_col) {
		int	delta = prev_col - e -> x;

		if (vh == 1 || delta > hincr || delta < -hincr) {
			if (vh == 0) {
				XDefineCursor (XtDisplay (widget), XtWindow (widget), W.horizontal_cursor);
				vh = 1;
			}
			if (delta > hincr || delta < -hincr) {
				delta = (delta / hincr) * hincr;
				Hscroll (w, W.hor_scroll + delta);
				prev_col = e -> x;
			}
		}
	}
	/*
	 *	Scroll vertical sinon.
	 */
	cur_line = e -> y / fh;
	if (vh <= 0 && cur_line != prev_line) {
		XDefineCursor (XtDisplay (widget), XtWindow (widget), W.vertical_cursor);
		vh = -1;
		if (cur_line > prev_line)
			ScrollDown (w, cur_line - prev_line, NULL);
		else
			ScrollUp (w, prev_line - cur_line, NULL);
		prev_line = cur_line;
	}
	UpdateScrollbar (w, False);
}

/*
 *	Selectionne la ligne complete.
 */

void select_line (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				x;

	if (!GotoXY (w, e -> x, e -> y))
		return;

	clearRegion ((Widget) w, NULL);
	BeginningOfLine (w, 1, NULL);

	W.b_marks [1].addr = W.b_cursor_e;
	W.b_marks [1].data = (XtPointer) (long) W.lineno;

	EndOfLine (w, 1, NULL);
	if (W.need_hpos) {
		GetHpos (w);
		W.cur_col = W.cur_x;
		W.need_hpos = 0;
	}
	W.b_marks [2].addr = W.b_cursor_e;
	W.b_marks [2].data = (XtPointer) (long) W.lineno;

	refreshSelect (w, W.cur_y);
	W.fresh_selection = True;
	if ((x = W.b_cursor_b - W.b_marks [1].addr) != 0)
		WixSetSelection (widget, e, W.b_marks [1].addr, x);
}

/*
 *	Selectionne un mot complet.
 */

void select_word (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;
	int				x;

	clearRegion ((Widget) w, NULL);
	GotoXY (w, e -> x, e -> y);
	if (index (W.delim, *W.b_cursor_e)) {
		/* mettre quand mme les marques car sinon, ca perturbe ButtonUp */
		W.b_marks [1].addr = W.b_cursor_e;
		W.b_marks [1].data = (XtPointer) (long) W.lineno;
		W.b_marks [2].addr = W.b_cursor_e;
		W.b_marks [2].data = (XtPointer) (long) W.lineno;
		return;
	}

	if (!index (W.delim, W.b_cursor_e[-1]))
		BackwardWord (w, 1, NULL);
	UpdateWindow (w);
	W.b_marks [1].addr = W.b_cursor_e;
	W.b_marks [1].data = (XtPointer) (long) W.lineno;

	ForwardWord (w, 1, NULL);
	W.b_marks [2].addr = W.b_cursor_e;
	W.b_marks [2].data = (XtPointer) (long) W.lineno;
	UpdateWindow (w);

	refreshSelect (w, W.cur_y);
	W.fresh_selection = True;
	if ((x = W.b_cursor_b - W.b_marks [1].addr) != 0)
		WixSetSelection (widget, e, W.b_marks [1].addr, x);
}

/*
 *	Lui farcit le buffer.
 */

static void stuff_sel (Widget widget, XtPointer cd, Atom* sel, Atom* type,
						XtPointer value, unsigned long* length, int* format)
{
	if (*type == 0) {
		char* line;
		int inbytes;

		line = XFetchBuffer(XtDisplay(widget), &inbytes, 0);
		if (inbytes > 0)
			KInsertString ((WixWidget) widget, (unsigned char*) line, inbytes);
		UpdateWindow ((WixWidget) widget);
		return;
	}
	if (*type == XA_STRING) {
		if (*length != 0) {
			KInsertString ((WixWidget) widget, (unsigned char*) value, (int) *length);
			XtFree (value);
		}
		UpdateWindow ((WixWidget) widget);
	}
}

void stuff (Widget widget, XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget		w = (WixWidget) widget;
	XButtonEvent*	e = (XButtonEvent*) event;

	W.fresh_selection = False;
	XtGetSelectionValue (widget, XA_PRIMARY, XA_STRING, stuff_sel, NULL, e -> time);
}

/*
 *	Scroll horizontal.
 */

void Hscroll (WixWidget w, int h)
{
	int			how;
	Display*	dpy = XtDisplay ((Widget) w);
	Window		window = XtWindow ((Widget) w);

	if (!XtIsRealized ((Widget) w))
		return;

	if (h < 0) h = 0;
	if (h == W.hor_scroll)
		return;
	how = h - W.hor_scroll;
	W.hor_scroll = h;
	W.fresh_selection = False;	/* invalide la region */

	if (how > 0) {
		XCopyArea (dpy, window, window, W.text_GC, 
				MARGINS + W.line_number_width + how, MARGINS,
				C.width - 2 * MARGINS - W.line_number_width - how, C.height - 2 * MARGINS,
				MARGINS + W.line_number_width, MARGINS);
		XClearArea (dpy, window,
				C.width - MARGINS - how, MARGINS,
				how, C.height - 2 * MARGINS, True);
	} else {
		how = -how;
		XCopyArea (dpy, window, window, W.text_GC, 
				MARGINS + W.line_number_width, MARGINS,
				C.width - 2 * MARGINS - how, C.height - 2 * MARGINS,
				MARGINS + W.line_number_width + how, MARGINS);
		XClearArea (dpy, window,
				MARGINS + W.line_number_width, MARGINS,
				how, C.height - 2 * MARGINS, True);
	}
	wait_for_expose (w);
}

void HorizontalScroll (Widget dummy, XtPointer client, XtPointer call)
{
	WixWidget	w = (WixWidget) client;
	int			pos = (long) call;
	int			hincr = Width (' ', FONT (0));

	pos = (pos * C.width * W.nb_page) / W.slider_max;
	Hscroll (w, (pos / hincr) * hincr);
}

static void update_horizontal (WixWidget w)
{
	int	h = W.slider_max / W.nb_page;

	if (h != W.hsb_size) {
		W.hsb_size = h;
		XtCallCallbackList ((Widget) w, W.hscroll_size, (XtPointer) (long) h);
	}
	h = (W.hor_scroll * h) / C.width;
	if (h != W.hsb_pos) {
		W.hsb_pos = h;
		if (h + W.hsb_size < W.slider_max)
			XtCallCallbackList ((Widget) w, W.hscroll_pos, (XtPointer) (long) h);
	}
}

void VerticalScroll (Widget dummy, XtPointer client, XtPointer call)
{
	WixWidget	w = (WixWidget) client;
	int			pos = (int) (long) call;
	int			nb_lines;
	int			wantedfirstline;
	int			currentfirstline;

	if (W.nolines <= W.num_lines-1)
		return;

	if (pos > W.slider_max)
		pos = W.slider_max;
	if (W.sb_size >= W.slider_max)
		wantedfirstline = 0;
	else
		wantedfirstline = (pos * (W.nolines - (W.num_lines-1))) / (W.slider_max - W.sb_size);
	currentfirstline = W.lineno - W.cur_y;
	if ((nb_lines = wantedfirstline - currentfirstline) != 0) {
		Boolean	hscroll = W.hscroll;

		W.hscroll = True;
		if (nb_lines > 0)	ScrollUp (w, nb_lines, NULL);
		else				ScrollDown (w, - nb_lines, NULL);
		W.hscroll = hscroll;
		W.prev_firstline = W.lineno - W.cur_y;
		W.sb_pos = pos;
	}
}

/*
 *	Met a jour le decor de la fenetre.
 */

void UpdateScrollbar (WixWidget w, Boolean force)
{
	int	size;
	int	pos;
	int	i;

	update_horizontal (w);
	pos = W.lineno - W.cur_y;

	if (W.prev_nolines == W.nolines && pos == W.prev_firstline && !force)
		return;

	W.prev_nolines = W.nolines;
	W.prev_firstline = pos;

	size = (W.num_lines-1) >= W.nolines
		? W.slider_max
		: ((W.num_lines-1) * W.slider_max) / W.nolines;
	if (size < 10) size = 10;
	i = W.nolines - (W.num_lines-1);
	if (i <= 0)
		i = 1;
	pos = pos > 0 ? (pos * (W.slider_max - size)) / i : 0;
	if (pos > W.slider_max - size)
		pos = W.slider_max - size;

	if (force || (W.sb_size != size && W.sb_pos != pos)) {
		if (pos + W.sb_size > W.slider_max) {
			XtCallCallbackList ((Widget) w, W.vscroll_size, (XtPointer) (long) size);
			XtCallCallbackList ((Widget) w, W.vscroll_pos, (XtPointer) (long) pos);
		} else {
			XtCallCallbackList ((Widget) w, W.vscroll_pos, (XtPointer) (long) pos);
			XtCallCallbackList ((Widget) w, W.vscroll_size, (XtPointer) (long) size);
		}
		W.sb_pos = pos;
		W.sb_size = size;
		return;
	}
	if (W.sb_size != size) {
		W.sb_size = size;
		XtCallCallbackList ((Widget) w, W.vscroll_size, (XtPointer) (long) size);
	}
	else if (W.sb_pos != pos) {
		W.sb_pos = pos;
		XtCallCallbackList ((Widget)w, W.vscroll_pos, (XtPointer) (long) pos);
	}
}

/*
 *	Slectionne tout le buffer.
 */

void SelectBuffer (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	W.b_marks [1].addr = W.beginning;
	W.b_marks [1].data = (XtPointer) 0;

	W.b_marks [2].addr = W.end;
	W.b_marks [2].data = (XtPointer) (long) W.nolines;

	W.fresh_selection = True;
}

/*
 *	Deslectionne le buffer.
 */

void DeselectBuffer (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	clearRegion (widget, (Atom*) 0);
	W.fresh_selection = False;
}
