/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#ifndef	WIXP_H
#define	WIXP_H

typedef struct {
	unsigned char*	addr;
	XtPointer		data;
} mark;


#define DEF_BUFSIZE	2048		/* taille d'un buffer par defaut */
#define DEF_INCR	10240		/* increment par defaut */
#define NUM_MARKS	16			/* nombre de marques initialement */

#define Bfree(w) (w -> wix.b_cursor_e - w -> wix.b_cursor_b)

#define COLOR(fn) (W.tab_colors[fn & ~HILITED])
#define FONT(fn) (W.tab_fonts[fn & ~HILITED])

/*
 *	Definition de la classe Wix.
 */

typedef struct {
	XtActionProc	(*get_action_proc)();
} WixClassPart;

typedef struct _WixClassRec {
	CoreClassPart		core_class;
	PrimitiveClassPart	primitive_class;
	WixClassPart		wix_class;
} WixClassRec;

extern WixClassRec wixClassRec;

/*
 *	La structure decrivant chaque ligne.
 */

typedef struct _line {
	unsigned char*	chars;		/* pointeur dans le buffer */
	unsigned char*	fn;			/* la fonte de chaque caractere */
	int				height;		/* hauteur du plus haut caractere */
	int				base_line;	/* la ligne de base */
} Line;

typedef struct _WixPart {
	XtCallbackList	modify_callback;
	XtCallbackList	verify_modify_callback;
	XtCallbackList	context_callback;
	XtCallbackList	undo_redo_callback;
	XtCallbackList	execute;
	XtCallbackList	def_action;
	XtCallbackList	vscroll_pos;
	XtCallbackList	vscroll_size;
	XtCallbackList	hscroll_pos;
	XtCallbackList	hscroll_size;
	MessageProc		messageFunction;

	int				margin;					/* espace vide autour du texte */
	int				frame_thickness;		/* pour 'allumer' le widget */
	Pixel			frame_color;			/* couleur d'allumage */
	GC				text_GC;
	GC				line_GC;
	Pixel			line_color;				/* couleur du numero de ligne */
	Pixel			select_color;			/* couleur des caractres slectionns */
	Cursor			horizontal_cursor;
	Cursor			vertical_cursor;
	Cursor			motion_cursor;
	FontInfo*		tab_fonts [HILITED];	/* les fontes */
	Pixel			tab_colors [HILITED];	/* leur couleur */

	char*			alias_buffer;			/* le buffer du fichier d'alias */
	int				alias_buffer_length;	/* sa longueur */
	alias*			alias_table;			/* table d'alias du widget */
	Trans*			trans_table;			/* table de translation */
	Trans*			prev_trans;				/* l'etape precedente (pour les prefixes) */

	int				hsb_pos;				/* position de la scrollbar */
	int				hsb_size;				/* taille de la scrollbar */
	int				sb_pos;					/* position de la scrollbar */
	int				sb_size;				/* taille de la scrollbar */
	int				slider_max;				/* SliderMax */

	int				nb_page;				/* taille de scroll horizontal */
	int				num_cols;				/* nombre de colonnes */
	int				num_lines;				/* nombre de lignes */
	int				tab_size;				/* largeur d'un tab en pixels */
	int				tab_equiv;				/* largeur d'un tab en caracteres */
	int				tab_stops [32];			/* 32 taquets de tabulation */
	int				cur_x;					/* X-coord du curseur */
	int				cur_col;				/* colonne du curseur */
	int				cur_y;					/* Y-coord du curseur */

	int				num_slots;				/* taille du tableau de lignes */
	Line*			first_line;				/* la premiere ligne du buffer */
	Line*			last_line;				/* la derniere ligne du buffer */
	Line*			cur_line;				/* la ligne courante */
	Line*			top_line;				/* la premiere ligne de la fenetre */

	Boolean			need_hpos;				/* faut il recalculer cur_x ? */

	Boolean			fresh_selection;
	Boolean			transient_selection;	/* n'allume la slection que le temps d'une commande */
	int				prev_y;

	int				refresh_from;			/* a partir d'ou */
	int				refresh_to;				/* et jusqu'a ou reafficher */

	int				hor_scroll;				/* decalage horizontal */
	int				top_scroll;				/* numero de ligne pour le scroll */
	int				bot_scroll;				/* numero de ligne pour le scroll */
	int				lineno;					/* numero de la ligne courante */
	int				nolines;				/* nombre de lignes du buffer */
	int				prev_firstline;			/* scrollbar: numero de la premiere ligne de la fenetre */
	int				prev_nolines;			/* scrollbar: nombre de lignes du buffer */

	XtIntervalId	cur_timer;				/* timer du curseur */
	Boolean			blink;

	int				state;					/* couleur/fonte du curseur */
	unsigned char*	a_cursor_b;				/* curseur dans le buffer des fontes */
	unsigned char*	a_cursor_e;				/* idem */
	unsigned char*	attr_beg;				/* debut du buffer des fontes */
	unsigned char*	attr_end;				/* fin du buffer des fontes */

	unsigned char*	b_cursor_b;				/* curseur dans le buffer */
	unsigned char*	b_cursor_e;				/* curseur dans le buffer */
	unsigned char*	beginning;				/* debut du buffer */
	unsigned char*	end;					/* fin du buffer */
	mark*			b_marks;				/* les marques */
	int				num_marks;				/* leur nombre */

	char			crypt_key [16];			/* la clef d'encryptage */

	Boolean			hilited;				/* il y a des caractres allums */
	Boolean			was_modified;
	Boolean			is_modified;
	Boolean			case_sensitive;			/* recherches sensibles  la casse ? */
	Boolean			hscroll;
	Boolean			read_only;
	Boolean			no_cursor;
	Boolean			focus_in;
	Boolean			kill_stack;				/* avec ou sans sauvegarde dans le kill stack */
	Boolean			wrap;
	Boolean			show_line_number;		/* affiche les numeros de ligne */
	int				line_number_width;		/* largeur occuppee par le numero */
	Boolean			macro_learn;			/* en phase d'apprentissage de macro */
	Boolean			bin_mode;				/* edition binaire */
	Boolean			mono;					/* edition monoligne */
	Boolean			do_redisplay;			/* avec ou sans redisplay */
	Boolean			do_undo;				/* avec ou sans undo/redo */
	String			value;
	String			delim;					/* les delimiteurs pour expand_alias */
	String			start_comment;			/* debut de commentaire */
	String			stop_comment;			/* fin de commentaire */
	String			line_comment;			/* commentaire sur une ligne */
	char			string_delim;			/* dlimiteur de chaines */
	String			beg_keyword;			/* dbut de mot clef */
	String			end_keyword;			/* fin de mot clef */
	KeyWord*		keyword_table;			/* la table des mots clefs */

	Dimension		width;					/* taille avant retaillage */
	Dimension		height;					/* pour le resize incrmental */
} WixPart;

typedef struct _WixRec {
	CorePart		core;
	PrimitivePart	primitive;
	WixPart			wix;
} WixRec;

extern Display*	dpy;
extern Window	root;
extern Screen*	screen;

typedef WixRec*	WixWidget;

#ifdef	_PRIVATE_
typedef int (*CommandProc) (WixWidget, int, void*);

typedef struct {
	CommandProc function;
	int			count;
	void*		arg;
} Command;

typedef struct {
	char*			string;
	XtActionProc	proc;
	CommandProc		function;
	char*			help_string;
} WixCommandRec;
#endif

#define	C	w -> core
#define	P	w -> primitive
#define	W	w -> wix

#define MARGINS (P.shadow_thickness + W.margin)

/*
 *	Quelques fonctions.
 */

extern void PushInsert (WixWidget, int, unsigned char*, int);
extern void PushDelete (WixWidget, int, unsigned char*, int);
extern void PushCommand (WixWidget);
extern void	KInsertString (WixWidget, unsigned char*, int);
extern void	MakeCurrentLine (WixWidget);
extern void	MakeLines (WixWidget);
extern int	KCopyRegion (WixWidget, unsigned char*);
extern int	KInsertChar (WixWidget, unsigned char);
extern int	KDeleteChar (WixWidget, int, int);
extern int	KForwardChar (WixWidget, int);
extern int	KBackwardChar (WixWidget, int);
extern int	FilterBuffer (WixWidget);
extern Boolean	InsideMultilineComment (WixWidget, unsigned char*);
extern void	ResetComments (WixWidget);
extern int	BufferCreate (WixWidget);
extern int	InitKillers (void);

extern int	GetHpos (WixWidget);
extern void	RemoveCursor (WixWidget);
extern void	ShowCursor (WixWidget);
extern void	UpdateWindow (WixWidget);
extern void	ToHpos (WixWidget, Boolean);
extern void	RefreshWindow (WixWidget, int, int);
extern void	TextRedisplay (WixWidget, int, int);
extern void	Hscroll (WixWidget, int);
extern void HorizontalScroll (Widget, XtPointer, XtPointer);
extern void VerticalScroll (Widget, XtPointer, XtPointer);
extern void UpdateScrollbar (WixWidget, Boolean);

extern void wait_for_expose (WixWidget);
extern int	GotoXY (WixWidget, int, int);
extern int	CharsWidth (FontInfo*, char*, int);
extern int	get_indent (WixWidget);
extern void nlws (WixWidget, int, unsigned char**);
extern void StartBlink (WixWidget);
extern void StopBlink (void);
extern int	GotoMark (WixWidget, unsigned char*);
extern Trans*	CvtStringToKeyTable (char*);

/*
 *	Les procedures d'actions.
 */

extern void IncrementalYankAction (Widget, XEvent*, String*, Cardinal*);
extern void	c_indent_line_action (Widget, XEvent*, String*, Cardinal*);
extern void	c_begin_function_action (Widget, XEvent*, String*, Cardinal*);
extern void	c_end_function_action (Widget, XEvent*, String*, Cardinal*);
extern void	replace_string (Widget, XEvent*, String*, Cardinal*);
extern void	insert_octal (Widget, XEvent*, String*, Cardinal*);
extern void	match_opening_brace (Widget, XEvent*, String*, Cardinal*);
extern void	match_closing_brace (Widget, XEvent*, String*, Cardinal*);
extern void	match_opening_par (Widget, XEvent*, String*, Cardinal*);
extern void	match_closing_par (Widget, XEvent*, String*, Cardinal*);
extern void	new_line_indent (Widget, XEvent*, String*, Cardinal*);
extern void	new_line_indent_c (Widget, XEvent*, String*, Cardinal*);
extern void	exec_macro (Widget, XEvent*, String*, Cardinal*);
extern void	multiplier (Widget, XEvent*, String*, Cardinal*);
extern void	twiddle_chars (Widget, XEvent*, String*, Cardinal*);
extern void	expand_alias (Widget, XEvent*, String*, Cardinal*);
extern void	repeat_command (Widget, XEvent*, String*, Cardinal*);
extern void	yank (Widget, XEvent*, String*, Cardinal*);
extern void	lower_case_word (Widget, XEvent*, String*, Cardinal*);
extern void	upper_case_word (Widget, XEvent*, String*, Cardinal*);
extern void	capitalize_word (Widget, XEvent*, String*, Cardinal*);
extern void	rubout_word (Widget, XEvent*, String*, Cardinal*);
extern void	delete_word (Widget, XEvent*, String*, Cardinal*);
extern void	backward_word (Widget, XEvent*, String*, Cardinal*);
extern void	forward_word (Widget, XEvent*, String*, Cardinal*);
extern void	end_word (Widget, XEvent*, String*, Cardinal*);
extern void	flasher (Widget, XEvent*, String*, Cardinal*);
extern void	backward_char (Widget, XEvent*, String*, Cardinal*);
extern void	backward_line (Widget, XEvent*, String*, Cardinal*);
extern void	backward_page (Widget, XEvent*, String*, Cardinal*);
extern void	beginning_of_buffer (Widget, XEvent*, String*, Cardinal*);
extern void	beginning_of_line (Widget, XEvent*, String*, Cardinal*);
extern void	copy_end_line (Widget, XEvent*, String*, Cardinal*);
extern void	copy_region (Widget, XEvent*, String*, Cardinal*);
extern void	wipe_region (Widget, XEvent*, String*, Cardinal*);
extern void	curline_top (Widget, XEvent*, String*, Cardinal*);
extern void	curline_center (Widget, XEvent*, String*, Cardinal*);
extern void	delete_char (Widget, XEvent*, String*, Cardinal*);
extern void	delete_beg_line (Widget, XEvent*, String*, Cardinal*);
extern void	delete_end_line (Widget, XEvent*, String*, Cardinal*);
extern void	delete_line (Widget, XEvent*, String*, Cardinal*);
extern void	end_of_buffer (Widget, XEvent*, String*, Cardinal*);
extern void	end_of_line (Widget, XEvent*, String*, Cardinal*);
extern void	exchange_point_mark (Widget, XEvent*, String*, Cardinal*);
extern void	forward_char (Widget, XEvent*, String*, Cardinal*);
extern void	forward_line (Widget, XEvent*, String*, Cardinal*);
extern void	forward_page (Widget, XEvent*, String*, Cardinal*);
extern void	goto_line (Widget, XEvent*, String*, Cardinal*);
extern void insert_line_number (Widget, XEvent*, String*, Cardinal*);
extern void	new_line (Widget, XEvent*, String*, Cardinal*);
extern void	open_space (Widget, XEvent*, String*, Cardinal*);
extern void	open_line (Widget, XEvent*, String*, Cardinal*);
extern void	pop_kill_stack (Widget, XEvent*, String*, Cardinal*);
extern void	rubout_char (Widget, XEvent*, String*, Cardinal*);
extern void	scroll_down (Widget, XEvent*, String*, Cardinal*);
extern void	scroll_left (Widget, XEvent*, String*, Cardinal*);
extern void	scroll_right (Widget, XEvent*, String*, Cardinal*);
extern void	scroll_up (Widget, XEvent*, String*, Cardinal*);
extern void	set_mark (Widget, XEvent*, String*, Cardinal*);
extern void	insert_tabulation (Widget, XEvent*, String*, Cardinal*);

extern void	keyboard_action (Widget, XEvent*, String*, Cardinal*);
extern void	expand_from_alias (Widget, XEvent*, String*, Cardinal*);
extern void	switch_line_numbering (Widget, XEvent*, String*, Cardinal*);

extern void	extend_selection (Widget, XEvent*, String*, Cardinal*);
extern void	selection_start (Widget, XEvent*, String*, Cardinal*);
extern void	extend_adjust (Widget, XEvent*, String*, Cardinal*);
extern void	extend_end (Widget, XEvent*, String*, Cardinal*);
extern void	scroll_window (Widget, XEvent*, String*, Cardinal*);
extern void	select_line (Widget, XEvent*, String*, Cardinal*);
extern void	select_word (Widget, XEvent*, String*, Cardinal*);
extern void	stuff (Widget, XEvent*, String*, Cardinal*);

/*
 *	Les fonctions (commes les actions mais pour les macros)
 */

extern int	ExecuteMacro (WixWidget, int, void*);
extern int	NewLineIndentC (WixWidget, int, void*);
extern int	NewLineIndent (WixWidget, int, void*);
extern int	ExpandAlias (WixWidget, int, void*);
extern int	ExpandFromAlias (WixWidget, int, void*);
extern int	ForwardSearch (WixWidget, int, void*);
extern int	BackwardSearch (WixWidget, int, void*);

extern int BackwardChar (WixWidget, int, void*);
extern int BackwardLine (WixWidget, int, void*);
extern int BackwardPage (WixWidget, int, void*);
extern int BackwardWord (WixWidget, int, void*);
extern int BeginningOfBuffer (WixWidget, int, void*);
extern int BeginningOfLine (WixWidget, int, void*);
extern int CBeginningOfFunction (WixWidget, int, void*);
extern int CEndOfFunction (WixWidget, int, void*);
extern int CIndentLine (WixWidget, int, void*);
extern int CapitalizeWord (WixWidget, int, void*);
extern int CopyEndLine (WixWidget, int, void*);
extern int CopyRegion (WixWidget, int, void*);
extern int CurlineTop (WixWidget, int, void*);
extern int CurlineCenter (WixWidget, int, void*);
extern int DeleteChar (WixWidget, int, void*);
extern int DeleteBeginningOfLine (WixWidget, int, void*);
extern int DeleteEndLine (WixWidget, int, void*);
extern int DeleteLine (WixWidget, int, void*);
extern int DeleteWord (WixWidget, int, void*);
extern int EndOfBuffer (WixWidget, int, void*);
extern int EndOfLine (WixWidget, int, void*);
extern int ExchangePointMark (WixWidget, int, void*);
extern int FindClosingBrace (WixWidget, int, void*);
extern int FindClosingParenthesis (WixWidget, int, void*);
extern int FindOpeningBrace (WixWidget, int, void*);
extern int FindOpeningParenthesis (WixWidget, int, void*);
extern int Flasher (WixWidget, int, void*);
extern int ForwardChar (WixWidget, int, void*);
extern int ForwardLine (WixWidget, int, void*);
extern int ForwardPage (WixWidget, int, void*);
extern int ForwardWord (WixWidget, int, void*);
extern int GotoLine (WixWidget, int, void*);
extern int InsertChar (WixWidget, int, void*);
extern int InsertLineNumber (WixWidget, int, void*);
extern int InsertString (WixWidget, int, void*);
extern int LowerCaseWord (WixWidget, int, void*);
extern int NewLine (WixWidget, int, void*);
extern int OpenLine (WixWidget, int, void*);
extern int OpenSpace (WixWidget, int, void*);
extern int PopKillStack (WixWidget, int, void*);
extern int RuboutChar (WixWidget, int, void*);
extern int RuboutWord (WixWidget, int, void*);
extern int ScrollDown (WixWidget, int, void*);
extern int ScrollLeft (WixWidget, int, void*);
extern int ScrollRight (WixWidget, int, void*);
extern int ScrollUp (WixWidget, int, void*);
extern int SetMark (WixWidget, int, void*);
extern int Tabulation (WixWidget, int, void*);
extern int TwiddleChars (WixWidget, int, void*);
extern int UpperCaseWord (WixWidget, int, void*);
extern int WipeRegion (WixWidget, int, void*);
extern int Yank (WixWidget, int, void*);

#endif	/* WIXP_H */
