/*****************************************************************

  Copyright (c) 1989, 1999 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <Xpa/Primitive.h>
#include <Xpa/PrimitiveP.h>

#define	_PRIVATE_
#include <Xpa/Wix.h>
#include <Xpa/WixP.h>

#include <Xpa/Scrollbar.h>

static	void	ClassInitialize (void);
static	void	Initialize (Widget, Widget, Arg*, Cardinal*);
static	void	Realize (Widget, XtValueMask*, XSetWindowAttributes*);
static	void	Resize (Widget);
static	void	Destroy (Widget);
static	void	Redisplay (Widget, XEvent*, Region);
static	Boolean	SetValues (Widget, Widget, Widget, Arg*, Cardinal*);
static	void	GetValuesHook (Widget, Arg*, Cardinal*);

static	XtActionProc	getActionProc ();

static XrmQuark	valueQuark;
static Trans*	_monoKeyTable;
static Trans*	_multiKeyTable;

#define	MAX_FONTS	256
static FontInfo	font_info [MAX_FONTS];

/*
 *	Les ressources.
 */

#define XtRWixFont	"WixFont"
#define XtRAliasTable	"AliasTable"
#define XtRKeyTable	"KeyTable"

static char multiKeyTable [] =
	"^X^X	exchange-point-mark\n"
	"^X^B	scroll-left\n"
	"^X^F	scroll-right\n"
	"^X^P	scroll-up\n"
	"^X^N	scroll-down\n"
	"^X#	insert-line-number\n"
	"^Z^L	show-line-number\n"
	"^@	set-mark\n"
	"^A	beginning-of-line\n"
	"^B	backward-char\n"
	"^C	repeat-command\n"
	"^D	delete-char\n"
	"^E	end-of-line\n"
	"^F	forward-char\n"
	"^H	rubout-char\n"
	"^I	tabulation\n"
	"^K	delete-end-line\n"
	"^L	center-curline\n"
	"^M	new-line\n"
	"^N	forward-line\n"
	"^O	open-space\n"
	"^P	backward-line\n"
	"^Q	insert-octal\n"
	"^R	delete-line\n"
	"^T	twiddle-chars\n"
	"^U	multiplier\n"
	"^V	forward-page\n"
	"$E$D	forward-page\n"
	"^W	wipe-region\n"
	"^Y	yank\n"
	"^?	delete-char\n"
	"$Eb	backward-word\n"
	"$Ec	capitalize-word\n"
	"$Ed	delete-word\n"
	"$Ee	end-word\n"
	"$Ef	forward-word\n"
	"$Ek	copy-end-line\n"
	"$El	lower-case-word\n"
	"$Eo	open-line\n"
	"$Eu	upper-case-word\n"
	"$Ev	backward-page\n"
	"$E$U	backward-page\n"
	"$Ew	copy-region\n"
	"$Ey	incremental-yank\n"
	"$E^?	delete-word\n"
	"$E^H	rubout-word\n"
	"$E1	multiplier\n"
	"$E2	multiplier\n"
	"$E3	multiplier\n"
	"$E4	multiplier\n"
	"$E5	multiplier\n"
	"$E6	multiplier\n"
	"$E7	multiplier\n"
	"$E8	multiplier\n"
	"$E9	multiplier\n"
	"$E<	beginning-of-buffer\n"
	"$E>	end-of-buffer\n"
	"$L	backward-char\n"
	"$R	forward-char\n"
	"$U	backward-line\n"
	"$D	forward-line\n";

/*
 *	la table de translations pour les wix mono ligne.
 */

static char monoKeyTable [] =
	"^X^X	exchange-point-mark\n"
	"^@	set-mark\n"
	"^A	beginning-of-line\n"
	"^B	backward-char\n"
	"^D	delete-char\n"
	"^E	end-of-line\n"
	"^F	forward-char\n"
	"^G	activate\n"
	"^H	rubout-char\n"
	"^I	tabulation\n"
	"^K	delete-end-line\n"
	"^M	activate\n"
	"^Q	insert-octal\n"
	"^T	twiddle-chars\n"
	"^U	multiplier\n"
	"^W	wipe-region\n"
	"^Y	yank\n"
	"^?	delete-char\n"
	"$Eb	backward-word\n"
	"$Ec	capitalize-word\n"
	"$Ed	delete-word\n"
	"$Ee	end-word\n"
	"$Ef	forward-word\n"
	"$Ek	copy-end-line\n"
	"$El	lower-case-word\n"
	"$Eu	upper-case-word\n"
	"$Ew	copy-region\n"
	"$E^?	rubout-word\n"
	"$E^H	rubout-word\n"
	"$E1	multiplier\n"
	"$E2	multiplier\n"
	"$E3	multiplier\n"
	"$E4	multiplier\n"
	"$E5	multiplier\n"
	"$E6	multiplier\n"
	"$E7	multiplier\n"
	"$E8	multiplier\n"
	"$E9	multiplier\n"
	"$E<	beginning-of-buffer\n"
	"$E>	end-of-buffer\n"
	"$L	backward-char\n"
	"$R	forward-char\n";

static XtResource resources [] =
{
	{ XtNframeThickness, XtCShadowThickness, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.frame_thickness), XtRImmediate, 0
	},
	{ XtNframeColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.frame_color), XtRString, "red"
	},
	{ XtNlineNumberColor, XtCForeground, XtRPixel, sizeof (Pixel),
	  XtOffset(WixWidget, wix.line_color), XtRString, "black"
	},
	{ XtNcursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(PrimitiveWidget, primitive.cursor), XtRString, "xterm"
	},
	{ XtNmotionCursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(WixWidget, wix.motion_cursor), XtRString, "fleur"
	},
	{ XtNhorizontalCursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(WixWidget, wix.horizontal_cursor), XtRString, "sb_h_double_arrow"
	},
	{ XtNverticalCursor, XtCCursor, XtRCursor, sizeof (Cursor),
	  XtOffset(WixWidget, wix.vertical_cursor), XtRString, "sb_v_double_arrow"
	},
	{ XtNvalue, XtCValue, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.value), XtRString, ""
	},
	{ XtNaliasBuffer, XtCAliasBuffer, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.alias_buffer), XtRImmediate, NULL
	},
	{ XtNaliasBufferLength, XtCAliasBufferLength, XtRInt, sizeof (int),
	  XtOffset(WixWidget, wix.alias_buffer_length), XtRImmediate, 0
	},
	{ XtNdelimitors, XtCDelimitors, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.delim), XtRString,
					" \t\n;:'`~}]=+-|!@$%^&()?>.<,\""
	},
	{ XtNkillStack, XtCKillStack, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.kill_stack), XtRImmediate, (XtPointer) True
	},
	{ XtNshowLineNumber, XtCShowLineNumber, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.show_line_number), XtRImmediate, (XtPointer) False
	},
	{ XtNlineNumberWidth, XtCLineNumberWidth, XtRInt, sizeof (int),
	  XtOffset(WixWidget, wix.line_number_width), XtRImmediate, (XtPointer) 0
	},
	{ XtNwrap, XtCWrap, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.wrap), XtRImmediate, (XtPointer) False
	},
	{ XtNbinaryEdit, XtCBinaryEdit, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.bin_mode), XtRImmediate, (XtPointer) False
	},
	{ XtNmonoLine, XtCMonoLine, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.mono), XtRImmediate, (XtPointer) False
	},
	{ XtNreadOnly, XtCReadOnly, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.read_only), XtRImmediate, (XtPointer) False
	},
	{ XtNfullCursor, XtCFullCursor, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.focus_in), XtRImmediate, (XtPointer) True
	},
	{ XtNnoCursor, XtCNoCursor, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.no_cursor), XtRImmediate, (XtPointer) False
	},
	{ XtNtransientSelection, XtCTransientSelection, XtRBoolean, sizeof (Boolean),
	  XtOffset(WixWidget, wix.transient_selection), XtRImmediate, (XtPointer) False
	},
	{ XtNaliasTable, XtCAliasTable, XtRAliasTable, sizeof(alias*),
	  XtOffset(WixWidget, wix.alias_table), XtRImmediate, (XtPointer) NULL
	},
	{ XtNkeyTable, XtCKeyTable, XtRKeyTable, sizeof(Trans*),
	  XtOffset(WixWidget, wix.trans_table), XtRImmediate, NULL
	},
	{ XtNbufferModified, XtCBufferModified, XtRBoolean, sizeof(Boolean),
	  XtOffset(WixWidget, wix.is_modified), XtRImmediate, (XtPointer) False
	},
	{ XtNcaseSensitive, XtCCaseSensitive, XtRBoolean, sizeof(Boolean),
	  XtOffset(WixWidget, wix.case_sensitive), XtRImmediate, (XtPointer) True
	},
	{ XtNcontextCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset(WixWidget, wix.context_callback), XtRPointer, (XtPointer) NULL
	},
	{ XtNbufferModifiedCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset(WixWidget, wix.modify_callback), XtRPointer, (XtPointer) NULL
	},
	{ XtNverifyModifyCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset(WixWidget, wix.verify_modify_callback), XtRPointer, (XtPointer) NULL
	},
	{ XtNundoRedoCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
	  XtOffset(WixWidget, wix.undo_redo_callback), XtRPointer, (XtPointer) NULL
	},
	{ XtNlineNum, XtCLineNum, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.num_lines), XtRImmediate, (XtPointer) 4
	},
	{ XtNcolumnNumber, XtCColumnNumber, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.num_cols), XtRImmediate, (XtPointer) 30
	},
	{ XtNcurrentLineNumber, XtCLineNum, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.lineno), XtRImmediate, 0
	},
	{ XtNbufferLineNumber, XtCLineNum, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.nolines), XtRImmediate, 0
	},
	{ XtNinnerBorder, XtCInnerBorder, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.margin), XtRImmediate, (XtPointer) 2
	},
	{ XtNtabsize, XtCTabsize, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.tab_equiv), XtRImmediate, (XtPointer) 8
	},
	{ XtNsliderMax, XtCSliderMax, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.slider_max), XtRImmediate, (XtPointer) 1000
	},
	{ XtNhScrollPage, XtCHScrollPage, XtRInt, sizeof(int),
	  XtOffset(WixWidget, wix.nb_page), XtRImmediate, (XtPointer) 5
	},
	{ XtNhScrollPositionCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  XtOffset (WixWidget, wix.hscroll_pos), XtRPointer, (XtPointer) NULL
	},
	{ XtNhScrollSizeCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  XtOffset (WixWidget, wix.hscroll_size), XtRPointer, (XtPointer) NULL
	},
	{ XtNvScrollPositionCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  XtOffset (WixWidget, wix.vscroll_pos), XtRPointer, (XtPointer) NULL
	},
	{ XtNvScrollPosition, XtCScroll, XtRInt, sizeof (int),
	  XtOffset (WixWidget, wix.sb_pos), XtRImmediate, 0
	},
	{ XtNvScrollSizeCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  XtOffset (WixWidget, wix.vscroll_size), XtRPointer, (XtPointer) NULL
	},
	{ XtNvScrollSize, XtCScroll, XtRInt, sizeof (int),
	  XtOffset (WixWidget, wix.sb_size), XtRImmediate, 0
	},
	{ XtNactivateCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  XtOffset (WixWidget, wix.execute), XtRPointer, (XtPointer) NULL
	},
	{ XtNdefaultActionCallback, XtCCallback, XtRCallback, sizeof(XtPointer),
	  XtOffset (WixWidget, wix.def_action), XtRPointer, (XtPointer) NULL
	},
	{ XtNbeginKeyword, XtCComment, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.beg_keyword), XtRImmediate, NULL
	},
	{ XtNendKeyword, XtCComment, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.end_keyword), XtRImmediate, NULL
	},
	{ XtNstartComment, XtCComment, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.start_comment), XtRImmediate, NULL
	},
	{ XtNstopComment, XtCComment, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.stop_comment), XtRImmediate, NULL
	},
	{ XtNlineComment, XtCComment, XtRString, sizeof (String),
	  XtOffset(WixWidget, wix.line_comment), XtRImmediate, NULL
	},
	{ XtNstringDelimitor, XtCStringDelimitor, "Char", sizeof (char),
	  XtOffset(WixWidget, wix.string_delim), XtRImmediate, (XtPointer) (int) '\"'
	},
	{ XtNfont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [STANDARD_COLOR]), XtRString, "Fixed"
	},
	{ XtNcommentFont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [COMMENT_COLOR]), XtRString, "Fixed"
	},
	{ XtNcommentColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.tab_colors [COMMENT_COLOR]), XtRString, "blue"
	},
	{ XtNcppFont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [CPP_COLOR]), XtRString, "Fixed"
	},
	{ XtNcppColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.tab_colors [CPP_COLOR]), XtRString, "red"
	},
	{ XtNstringFont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [STRING_COLOR]), XtRString, "Fixed"
	},
	{ XtNstringColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.tab_colors [STRING_COLOR]), XtRString, "dimgrey"
	},
	{ XtNkeywordFont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [KEYWORD_COLOR]), XtRString, "Fixed"
	},
	{ XtNkeywordColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.tab_colors [KEYWORD_COLOR]), XtRString, "brown"
	},
	{ XtNlibraryFont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [LIBRARY_COLOR]), XtRString, "Fixed"
	},
	{ XtNlibraryColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.tab_colors [LIBRARY_COLOR]), XtRString, "maroon"
	},
	{ XtNconstantFont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [CONSTANT_COLOR]), XtRString, "Fixed"
	},
	{ XtNconstantColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.tab_colors [CONSTANT_COLOR]), XtRString, "green"
	},
	{ XtNclassFont, XtCFont, XtRWixFont, sizeof(FontInfo*),
	  XtOffset(WixWidget, wix.tab_fonts [CLASS_COLOR]), XtRString, "Fixed"
	},
	{ XtNclassColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.tab_colors [CLASS_COLOR]), XtRString, "black"
	},
	{ XtNkeywordTable, XtCKeywordTable, XtRKeywordTable, sizeof(KeyWord*),
	  XtOffset(WixWidget, wix.keyword_table), XtRPointer, (XtPointer) NULL
	},
	{ XtNselectColor, XtCForeground, XtRPixel, sizeof(Pixel),
	  XtOffset(WixWidget, wix.select_color), XtRString, "lightgrey"
	},
};

static char defaultTranslations [] =
	"<Btn2Down>:			stuff()\n"
	"<Btn1Down>:			selection-start()\n"
	"<Btn1Motion>:		extend-adjust()\n"
	"<Btn1Up>:			extend-end()\n"
	"~Shift<Btn3Down>:		scroll-window()\n"
	"~Shift<Btn3Motion>:	scroll-window()\n"
	"~Shift<Btn3Up>:		scroll-window()\n"
	"Ctrl<Btn4Down>:		scroll-down(8)\n"
	"Ctrl<Btn5Down>:		scroll-up(8)\n"
	"Shift<Btn4Down>:	scroll-down(4)\n"
	"Shift<Btn5Down>:	scroll-up(4)\n"
	"<Btn4Down>:		scroll-down(2)\n"
	"<Btn5Down>:		scroll-up(2)\n"
	"Ctrl<Key>Prior:		scroll-up()\n"
	"Ctrl<Key>Next:		scroll-down()\n"
	"<Key>Insert:			open-space()\n"
	"<Key>Home:			beginning-of-line()\n"
	"<Key>End:			end-of-line()\n"
	"<Key>Prior:			backward-page()\n"
	"<Key>Next:			forward-page()\n"
	"<Key>:				keyboard-action()";


static void activate (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	WixWidget w = (WixWidget) widget;

	XtCallCallbackList (widget, W.execute, (XtPointer) e);
}

static void defaultActivate (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	WixWidget w = (WixWidget) widget;

	XtCallCallbackList (widget, W.def_action, (XtPointer) e);
}

static void noop (Widget widget, XEvent* e, String* params, Cardinal* num_params)
{
	/* ne fait rien */
}

WixCommandRec WixActionsList [] =
{
	{ "indent-line-c", c_indent_line_action, CIndentLine, "Position the cursor at the correct indentation level on the current line. This is specific to the C, C++ and JAVA modes; for example: after an opening curly brace there is one more indentation level." },
	{ "c-beginning-of-function", c_begin_function_action, CBeginningOfFunction, "Position the cursor at the beginning of the current function. This is specific to the C, C++ and JAVA modes." },
	{ "c-end-of-function", c_end_function_action, CEndOfFunction, "Position the cursor at the end of the current function. This is specific to the C, C++ and JAVA modes." },

	{ "backward-char", (XtActionProc) backward_char, BackwardChar, "Move the cursor back by numarg characters." },
	{ "backward-line", (XtActionProc) backward_line, BackwardLine, "Move the cursor up by numarg lines." },
	{ "backward-page", (XtActionProc) backward_page, BackwardPage, "Go up by numarg screenfuls." },
	{ "backward-word", (XtActionProc) backward_word, BackwardWord, "Move the cursor backwards by numarg words." },
	{ "beginning-of-buffer", (XtActionProc) beginning_of_buffer, BeginningOfBuffer, "Move the cursor to the beginning of the buffer." },
	{ "beginning-of-line", (XtActionProc) beginning_of_line, BeginningOfLine, "Move the cursor to the beginning of the current line." },
	{ "capitalize-word", (XtActionProc) capitalize_word, CapitalizeWord, "Change the initial of the current word from upper case to lower case and vice versa." },
	{ "copy-end-line", (XtActionProc) copy_end_line, CopyEndLine, "Copy the end of the current line to the kill-stack." },
	{ "copy-region", (XtActionProc) copy_region, CopyRegion, "Copy the region (the text between the mark and the cursor) to the kill-stack." },
	{ "center-curline", (XtActionProc) curline_center, NULL, "Redraw the window with the current line in the middle." },
	{ "curline-top", (XtActionProc) curline_top, NULL, "Redraw the window with the current line at the top." },
	{ "default-activate", (XtActionProc) defaultActivate, NULL, NULL },
	{ "delete-char", (XtActionProc) delete_char, DeleteChar, "Delete numarg characters." },
	{ "delete-beginning-of-line", (XtActionProc) delete_beg_line, DeleteBeginningOfLine, "Delete the beginning of the current line." },
	{ "delete-end-line", (XtActionProc) delete_end_line, DeleteEndLine, "Delete the end of the current line." },
	{ "delete-line", (XtActionProc) delete_line, DeleteLine, "Delete numarg lines and push them (as a whole) onto the kill-stack." },
	{ "delete-word", (XtActionProc) delete_word, DeleteWord, "Remove the numarg following words." },
	{ "end-of-buffer", (XtActionProc) end_of_buffer, EndOfBuffer, "Move the cursor to the end of the buffer." },
	{ "end-of-line", (XtActionProc) end_of_line, EndOfLine, "Move the cursor to the end of the current line." },
	{ "end-word", (XtActionProc) end_word, ForwardWord, "Move the cursor to the end of the current word." },
	{ "exchange-point-mark", (XtActionProc) exchange_point_mark, ExchangePointMark, "Interchange the positions of the cursor and the mark." },
	{ "exec-macro", (XtActionProc) exec_macro, ExecuteMacro, "Execute the macro last defined." },
	{ "expand-alias", (XtActionProc) expand_alias, ExpandAlias, 
"Try to expand the current prefix from the alias table of the current\
 mode, then from the words of the buffer itself. the cursor stop\
 blinking to show that you have entered the\
 expand-mode: each time you hit the TAB key the next alias (or word)\
 matching the prefix is inserted. To exit this mode you must hit\
 any key except TAB." },
	{ "expand-from-alias",	(XtActionProc) expand_from_alias, NULL, 
"Try to expand the current prefix from alias table;\
 the cursor stop blinking to show that you have entered the\
 expand-mode: each time you hit the TAB key the next alias\
 matching the prefix is inserted. To exit this mode you must hit\
 any key except TAB." },
	{ "extend-adjust", (XtActionProc) extend_adjust, NULL, "Select text in the direction of the mouse cursor." },
	{ "extend-end", (XtActionProc) extend_end, NULL, "The highlighted text becomes the primary selection (the selection is committed)" },
	{ "extend-selection",	(XtActionProc) extend_selection, NULL, NULL },
	{ "flasher", (XtActionProc) flasher, InsertChar, "Move the cursor backward to the opening item, stay there for half a second then return." },
	{ "forward-char", (XtActionProc) forward_char, ForwardChar, "Move the cursor forward by numarg characters." },
	{ "forward-line", (XtActionProc) forward_line, ForwardLine, "Move the cursor down by numarg lines." },
	{ "forward-page", (XtActionProc) forward_page, ForwardPage, "Go down by numarg screenfuls." },
	{ "forward-word", (XtActionProc) forward_word, ForwardWord, "Move the cursor forward by numarg words." },
	{ "goto-line", (XtActionProc) goto_line, GotoLine, "Move the cursor to the numarg-th line of the buffer. For example: to go to the line number 25 (with the key ^G bound to goto-line):\n\t\tESC 25 ^G" },
	{ "incremental-yank", IncrementalYankAction, NULL, "Put back the text previously deleted. (See expand-alias)." },
	{ "insert-char", (XtActionProc) insert_char, InsertString, "Insert the character typed." },
	{ "insert-line-number", (XtActionProc) insert_line_number, InsertLineNumber, "Insert the number of the current line." },
	{ "insert-octal", (XtActionProc) insert_octal, NULL, "Insert the next typed character regardless of any special command meaning it may have." },
	{ "keyboard-action", (XtActionProc) keyboard_action, NULL, NULL },
	{ "lower-case-word", (XtActionProc) lower_case_word, LowerCaseWord, "Change all characters between the cursor and the end of the word to lower-case." },
	{ "match-closing-brace", (XtActionProc) match_closing_brace, FindClosingBrace, "Find the closing brace matching the next opening brace." },
	{ "match-closing-par", (XtActionProc) match_closing_par, FindClosingParenthesis, "Find the closing parenthesis matching the next opening parenthesis." },
	{ "match-opening-brace", (XtActionProc) match_opening_brace, FindOpeningBrace, "Find the opening brace matching the preceding closing brace." },
	{ "match-opening-par", (XtActionProc) match_opening_par, FindOpeningParenthesis, "Find the opening parenthesis matching the preceding closing parenthesis." },
	{ "multiplier", (XtActionProc) multiplier, NULL, "Allow a numeric argument to be entered. (mistakes can be corrected with BackSpace)" },
	{ "new-line", (XtActionProc) new_line, NewLine, "Split a line into two parts (add a line-feed character)." },
	{ "new-line-indent", (XtActionProc) new_line_indent, NewLineIndent, "Insert a line-feed character and position the cursor at the correct indentation level on the next line." },
	{ "new-line-indent-c", (XtActionProc) new_line_indent_c, NewLineIndentC, "Insert a line-feed character and position the cursor at the correct indentation level on the next line. This is specific to the C mode; for example: after an opening curly brace there is one more indentation level." },
	{ "noop", (XtActionProc) noop, NULL, "Does nothing. (bound to all the undefined keys)" },
	{ "open-line", (XtActionProc) open_line, OpenLine, "Open an empty line above the cursor." },
	{ "open-space", (XtActionProc) open_space, OpenSpace, "Open an empty line (append a line-feed character)." },
	{ "pop-kill-stack", (XtActionProc) pop_kill_stack, PopKillStack, "Remove the numarg last items from the kill stack." },
	{ "repeat-command", (XtActionProc) repeat_command, NULL, "Repeat the previous command." },
	{ "replace-string",	(XtActionProc) replace_string, NULL, "replace this with that: for example\n<Key>`:	replace-string(\"e\",\"&egrave;\",\"a\",\"&agrave;\")" },
	{ "rubout-char", (XtActionProc) rubout_char, RuboutChar, "Delete the numarg preceding characters." },
	{ "rubout-word", (XtActionProc) rubout_word, RuboutWord, "Remove the numarg preceding words." },
	{ "scroll-down", (XtActionProc) scroll_down, ScrollDown, "Scroll the current window down by numarg lines." },
	{ "scroll-left", (XtActionProc) scroll_left, ScrollLeft, "Scroll the current window left by numarg columns." },
	{ "scroll-right", (XtActionProc) scroll_right, ScrollRight, "Scroll the current window right by numarg columns." },
	{ "scroll-up", (XtActionProc) scroll_up, ScrollUp, "Scroll the current window up by numarg lines." },
	{ "scroll-window", (XtActionProc) scroll_window, NULL, "Scroll the window vertically keeping the current pointer position." },
	{ "select-line", (XtActionProc) select_line, NULL, "Select the line pointed to by the mouse cursor." },
	{ "select-word", (XtActionProc) select_word, NULL, "Select the word pointed to by the mouse cursor." },
	{ "selection-start", (XtActionProc) selection_start, NULL, "Begin text selection." },
	{ "set-mark", (XtActionProc) set_mark, SetMark, "Set the mark at the current position." },
	{ "stuff", (XtActionProc) stuff, NULL, "Copy the text from the primary selection (the text shown in reverse video) to insertion point." },
	{ "show-line-number",	(XtActionProc) switch_line_numbering, NULL, "Show (hide) the line numbers in the text window." },
	{ "tabulation", (XtActionProc) insert_tabulation, Tabulation, "Insert a tabulation character." },
	{ "twiddle-chars", (XtActionProc) twiddle_chars, TwiddleChars, "Swap the two characters preceding the cursor." },
	{ "upper-case-word", (XtActionProc) upper_case_word, UpperCaseWord, "Change all characters between the cursor and the end of the word to upper-case." },
	{ "wipe-region", (XtActionProc) wipe_region, WipeRegion, "Delete the region (the text between the mark and the cursor)." },
	{ "yank", (XtActionProc) yank, Yank, "Insert the numarg-th killed item." },

	{ "activate", (XtActionProc) activate, NULL, NULL },
};

int	WixActionsNumber = XtNumber (WixActionsList);

WixClassRec wixClassRec = {
	/* core_class fields */	
    {
		(WidgetClass) &primitiveClassRec,	/* superclass */
		"Wix",								/* class_name */
		sizeof (WixRec),					/* widget_size */
		ClassInitialize,					/* class_initialize */
		NULL,								/* class_part_init */
		False,								/* class_inited */
		Initialize,							/* initialize */
		NULL,								/* initialize_hook */
		Realize,							/* realize */
		NULL,								/* actions */
		0,									/* num_actions */
		resources,							/* resources */
		XtNumber (resources),				/* num_resources */
		NULLQUARK,							/* xrm_class */
		True,								/* compress_motion */
		XtExposeCompressMaximal,			/* compress_exposure */
		False,								/* compress_enterlv */
		False,								/* visible_interest */
		Destroy,							/* destroy */
		Resize,								/* resize */
		Redisplay,							/* expose */
		SetValues,							/* set_values */
		NULL,								/* set_values_hook */
		XtInheritSetValuesAlmost,			/* set_values_almost */
		GetValuesHook,						/* get_values_hook */
		NULL,								/* accept_focus */
		XtVersion,							/* version */
		NULL,								/* callback_private */
		defaultTranslations,				/* tm_table */
		XtInheritQueryGeometry,				/* query_geometry */
		XtInheritDisplayAccelerator,		/* display accelerator */
		NULL,								/* extension */
	},
	/* primitive class fields */
	{
		NULL,
	},
	/* wix_class fields */	
	{
		getActionProc,						/* get_action_proc */
	}
};

WidgetClass wixWidgetClass = (WidgetClass)&wixClassRec;

/*
 *	Pour retrouver la procedure associee a une action.
 */

static	XtActionProc getActionProc (char* name)
{
	int	i;
	/*
	 *	les actions du widget.
	 */
	for (i = 0; i < WixActionsNumber; i++)
		if (!strcmp (WixActionsList [i].string, name))
			return WixActionsList [i].proc;
	/*
	 *	pas trouve.
	 */
	return NULL;
}

/*
 *	La procdure de cration de la fentre X.
 */

static void Realize	(Widget widget, XtValueMask* value_mask, XSetWindowAttributes* attributes)
{
	WixWidget	w = (WixWidget) widget;

	*value_mask |= CWBitGravity | CWCursor;
	attributes -> bit_gravity = NorthWestGravity;
	attributes -> cursor = P.cursor;
	XtCreateWindow (widget, InputOutput, CopyFromParent, *value_mask, attributes);
}

/*
 *	Donne la taille d'une chaine C
 */

int StringWidth (FontInfo* fp, unsigned char* s)
{
	int		n = 0;
	char*	cp = fp -> f_width;

	while (*s)
		n += cp [*s++];
	return n;
}

/*
 *	Donne la taille d'une suite d'octets.
 */

int CharsWidth (FontInfo* fp, char* s, int l)
{
	int				n = 0;
	unsigned char*	cp = (unsigned char *) fp -> f_width;

	while (--l >= 0)
		n += cp [(int)*s++];
	return n;
}

/*
 *	Le convertisseur pour les fontes.
 */

FontInfo* GetFont (Display* dpy, char* fontname, Boolean queryp)
{
	XFontStruct*	fp;
	XCharStruct*	p;
	FontInfo*		pp;
	char*			cp;
	unsigned int	n;
	/*
	 * Chercher la fonte.
	 */
	for (pp = font_info; pp -> f_info; pp++) {
		if (pp >= &font_info [MAX_FONTS])
			return NULL;		/* trop de fontes */
		if (!strcmp (pp -> f_name, fontname)) {
			return pp;		/* deja la */
		}
	}
	if (queryp)
		return NULL;			/* fonte pas charge */
	/*
	 *	Une vraie nouvelle fonte.
	 */
	fp = XLoadQueryFont (dpy, fontname);
	if (!fp)
		return NULL;

	pp -> f_info = fp;
	pp -> f_name = XtNewString (fontname);
	pp -> f_height = fp -> ascent + fp -> descent;
	pp -> f_ascent = fp -> ascent;
	pp -> f_width = XtMalloc (256);	/* allouer la table des largeurs */
	bzero (pp -> f_width, 256);	/* zro-ifie */

	n = fp -> min_char_or_byte2;
	if ((p = fp -> per_char) != 0) {
		for (cp = &pp -> f_width [n]; n <= fp -> max_char_or_byte2; n++, p++)
			*cp++ = p -> width;
	} else {
		for (cp = &pp -> f_width [n]; n <= fp -> max_char_or_byte2; n++)
			*cp++ = fp -> max_bounds.width;
	}
	return pp;
}

static XtConvertArgRec _font_conv [] = {
	{ XtBaseOffset, 0, sizeof (Widget) },
};

static void StringToFont (XrmValue* args, Cardinal* nargs, XrmValue* from, XrmValue* to)
{
	static FontInfo*	result;
	Widget				w;

	to -> size = sizeof (result);
	to -> addr = (XtPointer) &result;
	if (from -> addr == NULL) {
		to -> addr = NULL;
		return;
	}
	w = *((Widget*)(args [0].addr));
	result = GetFont (XtDisplay (w), (char*)from -> addr, False);
	if (!result)
		to -> addr = NULL;
}

/*
 *	Convertisseur pour la dfinition des tables d'alias.
 */

alias* CvtStringToAlias (unsigned char* str)
{
	alias*			result;
	unsigned char*	cp;
	alias*			ap;
	int				n;

	/* compter le nombre d'alias */
	for (n = 1, cp = str; *cp; n ++, cp ++)
		while (*cp != '\n') cp++;
	/* sauter les espaces initiaux */
	for (cp = str; *cp == ' ' || *cp == '\t' || *cp == '\n'; cp ++) ;
	/* allouer la table */
	cp = (unsigned char*) XtNewString (cp);
	ap = (alias*) XtMalloc (n * sizeof (alias));

	for (result = ap; *cp; ap++) {
		/* sauter les espaces initiaux */
		while (*cp == ' ' || *cp == '\t' || *cp == '\n') cp ++;
		ap -> a_src = cp;	/* commencer la chane source */
		while (*cp != ' ' && *cp != '\t' && *cp != '\n')
			cp ++;
		if (*cp == '\n') {
			*cp++ = 0;
			ap -> a_dst = ap -> a_src;
			continue;
		}
		*cp ++ = 0;	/* terminer la chane source */
		while (*cp == ' ' || *cp == '\t') cp++;
		ap -> a_dst = cp;	/* commencer la chane destination */
		while (*cp && *cp != '\n')
			cp ++;
		if (*cp != 0)
			*cp ++ = 0;	/* virer le \n du bout */
	}
	ap -> a_src = NULL;	/* terminer la table */
	return result;
}

static void StringToAlias (XrmValue* args, Cardinal* nargs, XrmValue* from, XrmValue* to)
{
	static alias*	result;

	to -> size = sizeof (result);
	if (from -> addr == NULL) {
		to -> addr = NULL;
		return;
	}
	result = CvtStringToAlias (from -> addr);
	to -> addr = (XtPointer) &result;
}

static void StringToKeyTable (XrmValue* args, Cardinal* nargs, XrmValue* from, XrmValue* to)
{
	static Trans*	result;

	to -> size = sizeof (result);
	if (from -> addr == NULL) {
		to -> addr = NULL;
		return;
	}
	result = CvtStringToKeyTable (from -> addr);
	to -> addr = (XtPointer) &result;
}

/*
 *	Initialisation globale.
 */

void _WixRegisterConverters () {
	static Boolean	done = False;

	if (! done) {
		done = True;
		XtAddConverter (XtRString, XtRAliasTable, StringToAlias, NULL, 0);
		XtAddConverter (XtRString, XtRKeyTable, StringToKeyTable, NULL, 0);
		XtAddConverter (XtRString, XtRWixFont, StringToFont, _font_conv, 1);
	}
}

static void ClassInitialize (void) {
	XtActionsRec*	actions;
	register int	i;

	valueQuark = XrmStringToQuark (XtNvalue);
	/*
	 *	Dupliquer la table d'actions
	 */
	actions = (XtActionsRec*) XtMalloc (WixActionsNumber * sizeof (XtActionsRec));
	for (i = 0; i < WixActionsNumber; i++) {
		actions [i].string = WixActionsList [i].string;
		actions [i].proc = WixActionsList [i].proc;
	}

	wixClassRec.core_class.actions = actions;
	wixClassRec.core_class.num_actions = XtNumber (WixActionsList);

	_monoKeyTable = CvtStringToKeyTable (monoKeyTable);
	_multiKeyTable = CvtStringToKeyTable (multiKeyTable);
	/*
	 *	La machinerie du kill ring
	 */
	InitKillers ();
}

/*
 *	Calcule les position de scrolling vertical.
 */

static void set_scroll (WixWidget w)
{
	if ((W.num_lines / 2) > 0)
		W.top_scroll = W.bot_scroll = W.num_lines / 2;
	else
		W.top_scroll = W.bot_scroll = 0;
}

/*
 *	le GC pour le texte.
 */

static void get_textGC (WixWidget w)
{
	XGCValues		values;
	unsigned long	valuemask;

	valuemask = GCForeground | GCBackground | GCFont | GCGraphicsExposures;
	values.foreground = P.foreground;
	values.background = C.background_pixel;
	values.graphics_exposures = True;
	values.font = FONT (0) -> f_info -> fid;

	if (!W.text_GC)
		W.text_GC = XCreateGC (XtDisplay (w),
			RootWindowOfScreen (XtScreen (w)), valuemask, &values);
	else
		XChangeGC (XtDisplay (w), W.text_GC, valuemask, &values);
}

/*
 *	le GC pour la ligne de reaffichage.
 */

static void get_lineGC (WixWidget w)
{
	XGCValues		values;
	unsigned long	valuemask;

	valuemask = GCForeground | GCBackground | GCFont;
	values.line_width = 2;
	valuemask |= GCLineWidth;
	values.foreground = P.foreground;
	values.background = C.background_pixel;
	values.font = FONT (0) -> f_info -> fid;

	if (!W.line_GC)
		W.line_GC = XCreateGC (XtDisplay (w),
			RootWindowOfScreen (XtScreen (w)), valuemask, &values);
	else
		XChangeGC (XtDisplay (w), W.line_GC, valuemask, &values);
}

/*
 *	Changement de fonte.
 */

static void change_font (WixWidget w, FontInfo* f)
{
	int	fh = f -> f_height;
	int	fw = Width ('x', f);

	if (fw <= 0)
		fw = f -> f_info -> max_bounds.width;

	W.tab_size = W.tab_equiv * fw;
	W.num_cols = (C.width - 2 * MARGINS) / fw;
	W.num_lines = (C.height - 2 * MARGINS) / fh;
	if (W.num_lines <= 0)
		W.num_lines = 1;
	W.need_hpos = 1;
	MakeLines (w);		/* recalcule les hauteurs */
	set_scroll (w);
}

/*
 *	Fait clignoter le curseur.
 */

static	XtIntervalId	theTimer = (XtIntervalId) 0;

static void blinkCursor (XtPointer cd, XtIntervalId* id)
{
	WixWidget		w = (WixWidget) cd;
	XtAppContext	app = XtWidgetToApplicationContext ((Widget)w);

	if (w != NULL && W.focus_in) {
		if (W.blink) {
			RemoveCursor (w);
			W.blink = False;
		} else {
			ShowCursor (w);
			W.blink = True;
		}
	}
	theTimer = XtAppAddTimeOut (app, 250, blinkCursor, cd);
}

void StopBlink (void)
{
	if (theTimer) {
		XtRemoveTimeOut (theTimer);
		theTimer = (XtIntervalId) 0;
	}
}

void StartBlink (WixWidget w)
{
	XtAppContext	app = XtWidgetToApplicationContext ((Widget)w);

	StopBlink ();
	W.blink = True;
	theTimer = XtAppAddTimeOut (app, 250, blinkCursor, (XtPointer) w);
}

/*
 *	Pour avoir le focus comme il faut!
 */

static void handleFocusEvents (Widget widget, XtPointer cd, XEvent* ev, Boolean* cont)
{  
	WixWidget			w = (WixWidget) widget;
	XFocusChangeEvent*	e = (XFocusChangeEvent*) ev;

	switch (ev -> type) {
		case FocusIn:
			if ((!W.focus_in && e -> detail != NotifyPointer) ||
				(W.focus_in && e -> detail == NotifyPointer))
			{
				StartBlink (w);
				W.focus_in = True;
				ShowCursor (w);
			}
			return;

		case FocusOut:
			if (W.focus_in) {
				StopBlink ();
				W.focus_in = False;
				ShowCursor (w);
			}
		return;
	}
}

/*
 *	Colore les cuseurs de facon correcte.
 */

static void recolor_cursors (WixWidget w)
{
	XColor	xc [2];

	xc [0].pixel = P.foreground;
	xc [1].pixel = C.background_pixel;
	XQueryColors (XtDisplay (w), DefaultColormapOfScreen (XtScreen (w)),
		      xc, 2);
	XRecolorCursor (XtDisplay (w), P.cursor, xc, &xc [1]);
	XRecolorCursor (XtDisplay (w), W.motion_cursor, xc, &xc [1]);
	XRecolorCursor (XtDisplay (w), W.horizontal_cursor, xc, &xc [1]);
	XRecolorCursor (XtDisplay (w), W.vertical_cursor, xc, &xc [1]);
}

Time	_LastActionTime;

static void setTime (Widget widget, XtPointer client_data, String action_name,
					 XEvent* event, String* params, Cardinal* num_params)
{
	WixWidget	w = (WixWidget) widget;

	if (XtClass (widget) != wixWidgetClass)
		return;

	_LastActionTime = event -> xkey.time;
	if (bcmp (action_name, "keyboard-action", sizeof ("keyboard-action")-1))
		PushCommand (w);
}

/*
 *	Fonction d'initialisation de chaque instance.
 */

static void Initialize (Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	WixWidget		w = (WixWidget) new;
	int				fw;
	static Boolean	first = True;

	if (first) {
	/* obtenir l'heure de chaque action */
		first = False;
		XtAppAddActionHook (
			XtWidgetToApplicationContext (new),
			setTime,
			NULL);
	}
	W.do_undo = True;	/* undo/redo activ */
	W.do_redisplay = True;	/* redisplay activ */
	recolor_cursors (w);

	if (W.margin < 0) W.margin = 2;

	fw = Width ('x', FONT (0));
	if (fw <= 0)
		fw = FONT (0) -> f_info -> max_bounds.width;

	W.tab_size = W.tab_equiv * fw;
	W.line_number_width = W.show_line_number /* 6 chiffres de la fonte 0 */
				? 6 * fw + 2 + W.margin
				: 0;

	W.tab_colors [0] = P.foreground;
	WixSetTabstops (new, NULL);

	if (request -> core.width == 0)
		C.width = W.num_cols * fw + 2 * MARGINS + W.line_number_width;
	if (request -> core.height == 0)
		C.height = (W.num_lines * FONT (0) -> f_height) + 2 * MARGINS;
	else {
		W.num_lines = (C.height - 2 * MARGINS) / FONT (0) -> f_height;
		if (W.num_lines <= 0)
			W.num_lines = 1;
	}

	if (W.mono) {
		if (!W.trans_table) W.trans_table = _monoKeyTable;
		W.num_lines = 1;
		C.height =  FONT (0) -> f_height + 2 * MARGINS;
	} else
		if (!W.trans_table) W.trans_table = _multiKeyTable;

	set_scroll (w);
	BufferCreate (w);
	if (W.value && *W.value)
		WixInsert (new, W.value);

	W.state = -1;		/* ShowCursor le mettra  jour */
	W.prev_nolines = -1;
	W.prev_firstline = -1;
	W.fresh_selection = False;
	W.macro_learn = False;
	W.hilited = False;
	W.messageFunction = NULL;
	W.prev_trans = W.trans_table;
	W.text_GC = 0;
	get_textGC (w);
	W.line_GC = 0;
	get_lineGC (w);
	W.blink = False;
	XtAddEventHandler (new, FocusChangeMask, False, handleFocusEvents, NULL);
	W.width = C.width;
	W.height = C.height;
}

/*
 *	La methode de resize.
 */

static void Resize (Widget widget)
{
	WixWidget	w = (WixWidget) widget;
	Display*	dpy = XtDisplay (widget);
	Window		win = XtWindow (widget);
	int			fw;

	fw = Width ('x', FONT (0));
	if (fw <= 0)
		fw = FONT (0) -> f_info -> max_bounds.width;

	W.num_lines = (C.height - 2 * MARGINS) / FONT (0) -> f_height;
	if (W.mono || W.num_lines <= 0)
		W.num_lines = 1;

	W.num_cols = (C.width - 2 * MARGINS - W.line_number_width) / fw;
	set_scroll (w);

	if (W.cur_y >= W.num_lines) {
		W.cur_y = W.top_scroll;
		/* afficher le tout car a scrolle  cause du curseur qui doit tre visible */
		if (XtIsRealized (widget))
			XClearArea (dpy, win, 0, 0, 0, 0, True);
	} else if (MARGINS > 0 && XtIsRealized (widget)) {
		/* rafficher les ombres et la marge */
		if (C.width >= W.width) {
			if (C.height >= W.height) {
				if (C.width != W.width)		/*  droite */
					XClearArea (dpy, win, W.width-MARGINS, 0, MARGINS, W.height, True);
				if (C.height != W.height)	/* en bas */
					XClearArea (dpy, win, 0, W.height-MARGINS, W.width, MARGINS, True);
			} else {
				if (C.width != W.width)		/*  droite */
					XClearArea (dpy, win, C.width-MARGINS, 0, MARGINS, C.height, True);
				/* en bas */
				XClearArea (dpy, win, 0, C.height-MARGINS, W.width, MARGINS, True);
			}
		} else {
			if (C.height >= W.height) {
				/*  droite */
				XClearArea (dpy, win, C.width-MARGINS, 0, MARGINS, C.height, True);
				if (C.height != W.height)	/* en bas */
					XClearArea (dpy, win, 0, C.height-MARGINS, C.width, MARGINS, True);
			} else {
				/*  droite */
				XClearArea (dpy, win, C.width-MARGINS, 0, MARGINS, C.height, True);
				/* en bas */
				XClearArea (dpy, win, 0, C.height-MARGINS, C.width, MARGINS, True);
			}
		}
	}
	W.width = C.width;
	W.height = C.height;
	/*
	 *	Forcer un update de la scrollbar verticale
	 */
	UpdateScrollbar (w, True);
}

/*
 *	Le reaffichage.
 */

static void Redisplay (Widget widget, XEvent* event, Region region)
{
	WixWidget	w = (WixWidget) widget;
	Display*	dpy = XtDisplay (widget);
	XRectangle r;

	if (event -> type == NoExpose)
		return;
	/*
	 *	Rafficher les ombres.
	 */
	if (W.frame_thickness > 0) {
		XSetRegion (dpy, W.line_GC, region);
		XSetForeground (dpy, W.line_GC, W.frame_color);
		XDrawRectangle (dpy, XtWindow (widget), W.line_GC,
						1, 1, C.width-W.frame_thickness, C.height-W.frame_thickness);
		XSetClipMask (dpy, W.line_GC, None);
		XSetForeground (dpy, W.line_GC, P.foreground);
	} else
	if (P.shadow_thickness > 0) {
		XSetRegion (dpy, P.top_GC, region);
		XSetRegion (dpy, P.bot_GC, region);
		_DrawShadow (dpy, XtWindow (widget),
			P.bot_GC, P.top_GC, P.shadow_thickness,
			C.width, C.height, 0, 0);
		XSetClipMask (dpy, P.top_GC, None);
		XSetClipMask (dpy, P.bot_GC, None);
	}
	/*
	 *	Rafficher le texte en n'empitant pas sur les ombres.
	 */
	XClipBox (region, &r);
	if (r.x < MARGINS) r.x = MARGINS;
	if (r.y < MARGINS) r.y = MARGINS;

	if (r.x + r.width > C.width - MARGINS)
		r.width = C.width - MARGINS - r.x;

	if (r.y + r.height > C.height - MARGINS)
		r.height = C.height - MARGINS - r.y;

	if (r.y < C.height -MARGINS && r.height > 0 && r.width > 0) {
		XSetClipRectangles (dpy, W.text_GC, 0, 0, &r, 1, Unsorted);
		TextRedisplay (w, r.y, r.y + r.height);
		ShowCursor (w);
		/*
		 *	Virer la rgion de clipping du texte.
		 */
		r.x = r.y = MARGINS;
		r.width = C.width - 2 * MARGINS;
		r.height = C.height - 2 * MARGINS;
		XSetClipRectangles (dpy, W.text_GC, 0, 0, &r, 1, Unsorted);
	}
}

/*
 *	La mthode de GetValues.
 */

static void GetValuesHook (Widget widget, Arg* args, Cardinal* pn)
{
	int	n;

	for (n = *pn; --n >= 0; )
		if (XrmStringToQuark (args [n].name) == valueQuark) {
			* ((String*) args [n].value) = WixCopyBuffer (widget);
			return;
		}
}

/*
 *	Changement des ressources.
 */

static	Boolean SetValues (Widget current, Widget request, Widget new, Arg* args, Cardinal* nargs)
{
	WixWidget	cu = (WixWidget) current;
	WixWidget	w = (WixWidget) new;
	Display*	dpy = XtDisplay (w);
	Boolean		do_redisplay = False;
	/*
	 *	Ressources read-only. (doit etre en tte)
	 */
	W.sb_size = cu -> wix.sb_size;
	W.sb_pos = cu -> wix.sb_pos;
	W.nolines = cu -> wix.nolines;
	W.lineno = cu -> wix.lineno;
	W.line_number_width = cu -> wix.line_number_width;
	/*
	 *	Changement de table de transition.
	 */
	if (cu -> wix.trans_table != W.trans_table)	/* remettre l'automate */
		W.prev_trans = W.trans_table;			/* dans l'tat initial */
	/*
	 *	Changement du curseur
	 */
	if (cu -> primitive.cursor != P.cursor && XtIsRealized (new)) {
		XDefineCursor (dpy, XtWindow (new), P.cursor);
		XSync (dpy, False);
	}
	/*
	 *	Changement de couleurs.
	 */
	if (cu -> primitive.foreground != P.foreground ||
	    cu -> core.background_pixel != C.background_pixel)
	{
		W.tab_colors [0] = P.foreground;
		recolor_cursors (w);
	}
	/*
	 *	Changement du contenu.
	 */
	if (cu -> wix.value != W.value) {
		WixClearBuffer (new);
		WixInsert (new, W.value);
		do_redisplay = True;
	}
	/*
	 *	Changement de delimiteurs de commentaires.
	 */
	if (cu -> wix.start_comment != W.start_comment ||
	    cu -> wix.stop_comment != W.stop_comment ||
	    cu -> wix.line_comment != W.line_comment)
	{
		if (!W.start_comment && !W.stop_comment && !W.line_comment)
			ResetComments (w);
		else
			(void) FilterBuffer (w);
		MakeLines (w);
		do_redisplay = True;
	}
	/*
	 *	Changement de sensibilite.
	 */
	if (cu -> core.sensitive != C.sensitive || cu -> core.ancestor_sensitive != C.ancestor_sensitive) {
		if (C.sensitive && C.ancestor_sensitive)
			ShowCursor (w);
		else
			RemoveCursor (w);
	}
	/*
	 *	Changement de l'epaisseur de l'ombre.
	 */
	if (cu -> primitive.shadow_thickness != P.shadow_thickness)
		do_redisplay = True;
	/*
	 *	Le nombre de lignes.
	 */
	if (cu -> wix.num_lines != W.num_lines) {
		C.height = (W.num_lines * FONT (0) -> f_height) + 2 * MARGINS;
		do_redisplay = True;
	}
	/*
	 *	Le nombre de colonnes.
	 */
	if (cu -> wix.num_cols != W.num_cols) {
		C.width = (W.num_cols * Width ('x', FONT (0))) + 2 * MARGINS;
		do_redisplay = True;
	}
	/*
	 *	changement de fonte.
	 */
	if (FONT (0) != cu -> wix.tab_fonts [0]) {
		if (FONT (0)) {
			change_font (w, FONT (0));
			XSetFont (dpy, W.text_GC, FONT (0) -> f_info -> fid);
			XSetFont (dpy, W.line_GC, FONT (0) -> f_info -> fid);
			UpdateScrollbar (w, True);
			do_redisplay = True;
		} else
			FONT (0) = cu -> wix.tab_fonts [0];
	}
	/*
	 *	Avec ou sans numeros de ligne?
	 */
	if (cu -> wix.show_line_number != W.show_line_number)
		do_redisplay = True;

	W.line_number_width = W.show_line_number
				? 6 * Width ('x', FONT (0)) + 2 + W.margin
				: 0;
	/*
	 *	Buffer modifie.
	 */
	if (W.is_modified != cu -> wix.is_modified)
		W.was_modified = W.is_modified;
	/*
	 *	Taille d'une tabulation.
	 */
	if (W.tab_equiv != cu -> wix.tab_equiv) {
		int	fw = Width ('x', FONT (0));

		if (fw <= 0)
			fw = FONT (0) -> f_info -> max_bounds.width;
		W.tab_size = W.tab_equiv * fw;
		do_redisplay = True;
	}
	/*
	 *	Marge.
	 */
	if (W.margin != cu -> wix.margin) {
		if (W.margin < 0) W.margin = 2;
		do_redisplay = True;
	}
	/*
	 *	Changement de l'paisseur du cadre.
	 */
	if (cu -> wix.frame_thickness != W.frame_thickness && !do_redisplay) {
		if (W.frame_thickness > 0) {
			XSetForeground (dpy, W.line_GC, W.frame_color);
			XDrawRectangle (dpy, XtWindow (new), W.line_GC,
							1, 1, C.width-W.frame_thickness, C.height-W.frame_thickness);
			XSetForeground (dpy, W.line_GC, P.foreground);
		} else if (P.shadow_thickness > 0) {
			_DrawShadow (dpy, XtWindow (new),
				P.bot_GC, P.top_GC, P.shadow_thickness,
				C.width, C.height, 0, 0);
		}
	}
	/*
	 *	Changement de l'affichage des caractres non ascii.
	 */
	if (cu -> wix.bin_mode != W.bin_mode)
		do_redisplay = True;

	return do_redisplay;
}

static void Destroy (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	StopBlink ();	/* ne plus referencer le widget ... */
	if (W.b_marks)
		XtFree ((char*)W.b_marks);
	if (W.beginning)
		XtFree (W.beginning - 1);
	XFreeGC (XtDisplay (widget), W.text_GC);
}

/*
 *	Les fonctions publiques.
 */

int WixGetHorScroll (Widget widget)
{
	WixWidget	w = (WixWidget) widget;

	return W.hor_scroll;
}

void WixSetTabstops (Widget widget, int* t)
{
	WixWidget		w = (WixWidget) widget;
	unsigned int	i;

	if (t == NULL) {
		W.tab_stops [0] = W.tab_size;
		W.tab_stops [1] = -1;
	} else {
		for (i = 0; i < XtNumber (W.tab_stops); i++)
			W.tab_stops [i] = *t++;
	}
	if (XtIsRealized (widget))
		RefreshWindow (w, 0, 1000);
}

void WixGetTabstops (Widget widget, int* t)
{
	WixWidget		w = (WixWidget) widget;
	unsigned int	i;

	for (i = 0; i < XtNumber (W.tab_stops); i++)
		*t++ = W.tab_stops [i];
}

void WixSetActionProc (XtActionProc (*proc)())
{
	wixClassRec.wix_class.get_action_proc = proc;
}

char* WixCopyBuffer (Widget widget)
{
	WixWidget	w = (WixWidget) widget;
	int			len1;
	int			len2;
	char*		p;

	len1 = W.b_cursor_b - W.beginning;
	len2 = W.end - W.b_cursor_e;
	p = XtMalloc (len1 + len2 + 1);		/* un de plus pour un nul */

	bcopy (W.beginning, p, len1);
	bcopy (W.b_cursor_e, p + len1, len2);
	p [len1+len2] = 0;
	return p;
}

/*************************************************************************
*       Cre un wix dans une scrolled window.
*
*************************************************************************/

static void scroll_size (Widget widget, XtPointer client, XtPointer call)
{
	XtVaSetValues ((Widget) client, XtNsliderExtent, call, NULL);
}

static void scroll_pos (Widget widget, XtPointer client, XtPointer call)
{
	XtVaSetValues ((Widget) client, XtNsliderOrigin, call, NULL);
}

static void wix_hscroll (Widget sb, XtPointer client, XtPointer call)
{
	HorizontalScroll (NULL, (XtPointer) client, (XtPointer) call);
}

static void wix_vscroll (Widget sb, XtPointer client, XtPointer call)
{
	VerticalScroll (NULL, (XtPointer) client, (XtPointer) call);
}

static void wix_page (Widget sb, XtPointer client, XtPointer call)
{
	WixWidget	w = (WixWidget) client;
	Boolean		direction = (Boolean) (long) call;

	if (direction)
		(void) ForwardPage (w, 1, NULL);
	else
		(void) BackwardPage (w, 1, NULL);
	UpdateWindow (w);
}

Widget XpaCreateScrolledWix (Widget parent, char* name, Arg* _arg, int _narg)
{
	Widget	swin, vsb, hsb;
	Widget	wix;
	Arg		args [64];
	int		n = 0;
	/*
	 *      La Scrolled Window.
	 */
	XtSetArg (args [n], XtNshadowThickness, 0), n++;
	swin = XtCreateWidget ("wix-sw", scrolledWindowWidgetClass, parent, args, n);
	n = 0;
	XtSetArg (args [n], XtNhScrollBar, &hsb), n++;
	XtSetArg (args [n], XtNvScrollBar, &vsb), n++;
	XtGetValues (swin, args, n);
	/*
	 *      La scrollbar verticale.
	 */
	n = 0;
	XtSetArg (args [n], XtNsliderMax, 1000), n++;
	XtSetArg (args [n], XtNsliderExtent, 1000), n++;
	XtSetValues (vsb, args, n);
	/*
	 *      La scrollbar horizontale.
	 */
	n = 0;
	XtSetArg (args [n], XtNsliderMax, 1000), n++;
	XtSetArg (args [n], XtNsliderExtent, 200), n++;
	XtSetValues (hsb, args, n);
	/*
	 *      Le wix.
	 */
	wix = XtCreateManagedWidget (name, wixWidgetClass, swin, _arg, _narg);
	/*
	 *      Les callbacks.
	 */
	XtAddCallback (hsb, XtNsliderMoved, wix_hscroll, (XtPointer) wix);
	XtAddCallback (wix, XtNhScrollSizeCallback, scroll_size, (XtPointer) hsb);
	XtAddCallback (wix, XtNhScrollPositionCallback, scroll_pos, (XtPointer) hsb);

	XtAddCallback (vsb, XtNsliderMoved, wix_vscroll, (XtPointer) wix);
	XtAddCallback (vsb, XtNareaSelected, wix_page, (XtPointer) wix);
	XtAddCallback (wix, XtNvScrollSizeCallback, scroll_size, (XtPointer) vsb);
	XtAddCallback (wix, XtNvScrollPositionCallback, scroll_pos, (XtPointer) vsb);

	return (Widget) wix;
}

Widget XtCreateScrolledWix (Widget parent, char* name, Arg* _arg, int _narg)
{
	return XpaCreateScrolledWix (parent, name, _arg, _narg);
}
