/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include "client.h"

char*	getenv ();

static remote	r;

static void usage () {
	fprintf (stderr, "\
Usage:\n\
wx [-display display_name]  Connect to a wxd server on that display.\n\
   [-fn font]		    Use this font in the editing window.\n\
   [-wait]		    Wait (block) until the window is closed.\n\
   [-directory dirname]	    Set the 'current directory'.\n\
   [+fsel]		    Show the file selector when the window appears.\n\
   [-g / -geometry WxH+X+Y] Set the size and/or the position of the window.\n\
   [filename]	   	    The name of the file to be edited.\n");
	exit (1);
}

static void onalarm () {
	fprintf (stderr, "wxd server not responding.\n");
	exit (1);
}

static void wait_for_wxd (Display* dpy, Window root, Atom atom)
{
	XEvent	ev;

	signal (SIGALRM, onalarm);
	alarm (5);	/* attendre au maximum 5 secondes */
	XSelectInput (dpy, root, PropertyChangeMask);
	for (;;) {
		XNextEvent (dpy, &ev);
		if (ev.type == PropertyNotify &&
		    ev.xproperty.window == root &&
		    ev.xproperty.atom == atom &&
		    ev.xproperty.state == PropertyNewValue) {
			alarm (0);
			return;
		}
	}
}

main (int argc, char** argv)
{
	Display*	dpy;
	Window		root;
	XEvent		ev;
	Atom		A_FILE;
	Atom		A_HERE;
	Atom		A_CLIENT;
	char*		p;
	char*		files [32];
	char*		display = NULL;

	Atom		type;
	unsigned long	nitems, left;
	int		format;
	int		i;
	int		waitflag = 0;
	int		n_files = 0;

	for (argv ++; *argv; argv ++) {
		if (**argv == '-') {
			if (!strcmp (*argv, "-wait")) {
				waitflag = 1;
				continue;
			}
			if (!strcmp (*argv, "-fn")) {
				strcpy (r.font, *++argv);
				continue;
			}
			if (!strcmp (*argv, "-g") || !strcmp (*argv, "-geometry")) {
				strcpy (r.geometry, *++argv);
				continue;
			}
			if (!strcmp (*argv, "-directory")) {
				strcpy (r.directory, *++argv);
				continue;
			}
			if (!strcmp (*argv, "-display")) {
				display = *++argv;
				continue;
			}
			usage ();
		}
		files [n_files++] = *argv;		/* les fichiers  diter */
	}

	dpy = XOpenDisplay (display);
	if (dpy == 0) {
		fprintf (stderr, "Cannot connect to X server %s\n", display);
		exit (1);
	}

	root = DefaultRootWindow (dpy);
	/*
	 *	Y a t il un serveur ?
	 */
	A_HERE = XInternAtom (dpy, "WIX_SERVER", False);
	if (XGetWindowProperty (dpy, root, A_HERE, 0, 256,
				False, XA_STRING, &type, &format, &nitems, &left,
				(unsigned char**)&p) != Success || type != XA_STRING) {
		/*
		 *	non: le lancer sur CETTE machine.
		 */
		if (fork ())
			wait (0);
		else {
			execlp ("wxd", "wxd", "-display", DisplayString (dpy), NULL);
			fprintf (stderr, "Cannot exec wxd.\n");
			exit (1);
		}
		wait_for_wxd (dpy, root, A_HERE);
	}
	/*
	 *	Trouver le nom absolu du directory courant.
	 */
	if (!r.directory [0]) {
		if ((p = getenv ("cwd")) != NULL)
			strcpy (r.directory, p);
		else {
			if (getcwd (r.directory, 256) == NULL)
			{
				r.directory [0] = '.';
				r.directory [1] = 0;
			}
		}
	}

	A_FILE = XInternAtom (dpy, "WIX_FILE", False);
	A_CLIENT = XInternAtom (dpy, "WIX_CLIENT", False);

	strcpy (r.file, n_files > 0 ? files [0] : "");
	if (n_files > 1) {
		char*	p = r.files;

		p [0] = 0;
		for (i = 1; i < n_files && i < 20; ++i)
			p += sprintf (p, "%s\n", files [i]);
		p [-1] = 0;
	}
	sprintf (r.client_pid, "%d", getpid ());
	XChangeProperty (dpy, root, A_FILE, XA_STRING, 8, PropModeReplace,
				(unsigned char*)&r, sizeof (r));
	XSync (dpy, 0);
	if (waitflag) {
		XSelectInput (dpy, root, PropertyChangeMask);
		for (;;) {
			XNextEvent (dpy, &ev);
			if (ev.type == PropertyNotify &&
			    ev.xproperty.window == root &&
			    ev.xproperty.atom == A_CLIENT &&
			    ev.xproperty.state == PropertyNewValue)
			{
				if (XGetWindowProperty (dpy, root, A_CLIENT, 0, 256,
					False, XA_STRING, &type, &format, &nitems, &left,
					(unsigned char**)&p) == Success
				    && type == XA_STRING
				    && !strcmp (r.client_pid, p))
				{
					exit (0);
				}
			}
		}
	}
	else
		exit (0);
	return 0;
}
