/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#ifdef SYSV
#include <sys/utsname.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/Ruler.h>
#include <Xpa/Query.h>
#include <Xpa/MenuBar.h>
#include <Xpa/Wix.h>

#include "wix.h"
#include "client.h"

extern	char*	AbsPath (char*, char*);

static	char*	discardChangesMessage = "This file is modified but not saved.\nDo you want to proceed anyway ?";
static	char*	writeFailedMessage = "It is impossible to write this file\nit is probably write protected.";
static	char*	loadFailedMessage = "This file cannot be loaded.\nit may be read protected or\nit does not exist.";
static	char*	fileExistsMessage = "A file with this name already exists.\nDo you want to erase it ?";
static	char*	macroLearnMessage = "Begin macro collection";
static	char*	macroLearnedMessage = "End macro collection";
static	char*	quitMessage = "Do you really want to quit the WiX server ?";
static	char*	fileModifiedMessage = "This file has been modified by an\nother program while being edited.\nDo you still want to write it ?";
static	char*	loadFileMessage = " Load File ";
static	char*	writeFileMessage = " Save File ";
static	char*	insertFileMessage = " Insert File ";
static	char*	changeDirectoryMessage = "Change Directory";
static	char*	reloadMessage = "This file is already being edited\nin another window. Do you\nstill want to load it ?";

		int		def_width = 0;
		int		def_height = 0;

char			HomeDirectory [128];

static void setHints (WixWindow*);

WixWindow*	_head = NULL;

static void ch_curs (WixWindow* data, Cursor cur)
{
	XtVaSetValues (data -> wix,
			XtNcursor, cur,
			NULL);
	if (data -> shadow)
		XtVaSetValues (data -> shadow,
			XtNcursor, cur,
			NULL);
	if (data -> wix2)
		XtVaSetValues (data -> wix2,
			XtNcursor, cur,
			NULL);
}

void WatchCursor (WixWindow* data) { ch_curs (data, WatchCurs); }
void TextCursor (WixWindow* data) { ch_curs (data, TextCurs); }

/*
 *	Tlporte le pointeur au milieu de la fenetre.
 */

void warpPointer (WixWindow* data)
{
	Dimension	width, height;
	Widget		pane = data -> pane;

	XtVaGetValues (pane,
		XtNwidth, &width,
		XtNheight, &height,
		NULL);

	XWarpPointer (XtDisplay (pane), None, XtWindow (pane),
		    0, 0, 0, 0, width/2, height/2);
}

/*************************************************************************
*
*	Les actions et fonctions de base qui apparaissent dans le menu.
*
*************************************************************************/

/*
 *	Change de fonte.
 */

void setFontFunction (Widget w, char* fontname)
{
	FontInfo*	fp;
	WixWindow*	data;
	int			lnw;

	if (!strcmp (fontname, "default"))
		fp = default_font;
	else
		fp = GetFont (XtDisplay (w), fontname, False);

	if (fp == NULL)
		return;

	XtVaSetValues (w, XtNfont, fp, NULL);
	XtVaGetValues (w,
		XtNlineNumberWidth,	&lnw,
		XtNuserData,		&data,
		NULL);
	XtVaSetValues (data -> ruler, XtNleftOffset, 15 + lnw, NULL);
	/*
	 *	Ne changer les hints que s'il n'y a qu'un seul pane.
	 */
	if (!data -> shadow)
		setHints (data);
}

/*
 *	Change les tabulations.
 */

static void setTabCont (WixWindow* data)
{
	int			cols [32];
	int			tab_equiv;
	int			i;
	FontInfo*	fp;

	tab_equiv = atoi (data -> fs_answer);
	if (tab_equiv < 2 || tab_equiv > 20) {
		setTabAction (data -> active, NULL, NULL, NULL);
		return;
	}

	XtVaGetValues (data -> active, XtNfont, &fp, NULL);
	for (i = 0; i < (int) XtNumber (cols); i++)
		cols [i] = (i+1) * tab_equiv * fp -> f_info -> max_bounds.width;
	XtVaSetValues (data -> ruler,
		XtNscale,	fp -> f_info -> max_bounds.width,
		NULL);
	RulerSetTabstops (data -> ruler, cols, XtNumber (cols));
	XtVaSetValues (data -> active, XtNtabsize, tab_equiv, NULL);
	WixSetTabstops (data -> active, cols);
}

void setTabAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	char		tc [16];
	int			tab_equiv;
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	XtVaGetValues (data -> active, XtNtabsize, &tab_equiv, NULL);
	sprintf (tc, "%d", tab_equiv);

	data -> what_after = setTabCont;
	popQueryStringNoWait (data, "Tab equivalent [2 - 20]", tc, False);
}

/*
 *	Evalue une expression lisp.
 */

static void lispEvalCont (WixWindow* data)
{
	char		tc [2048];
	extern void	_SetCurrentWix (LispExpr);
	LispExpr	exp;

	_SetCurrentWix (data -> active);	/* affecter le symbole 'wix' au buffer courant */
	exp = lisp_read_string (data -> fs_answer);
	exp = lisp_eval (exp);
	lisp_print (exp, tc);
	popAlert (data, tc, False);
}

void lispEvalAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);

	data -> what_after = lispEvalCont;
	popQueryStringNoWait (data, "Lisp evaluation", NULL, False);
}

/*
 *	Recharge un fichier lisp.
 */

static void loadLispFileCont (WixWindow* data)
{
	switch (lisp_load_file (data -> fs_answer)) {
		case 0:
			popAlert (data, loadFailedMessage, False);
			return;

		case 1:
			popAlert (data, "Error during lisp file loading.", False);
			return;

		default:
			Message (data -> active, "Loaded.");
	}
}

void loadLispFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);

	data -> what_after = loadLispFileCont;
	popQueryStringNoWait (data, "Load lisp file", NULL, False);
}

/*
 *	Change la sensibilit  la casse dans les recherches.
 */

void ToggleCaseSensitivityAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	Boolean	cs;

	XtVaGetValues (w, XtNcaseSensitive, &cs, NULL);
	XtVaSetValues (w, XtNcaseSensitive, !cs, NULL);
	Message (w, cs ? "Case insensitive." : "Case sensitive.");
}

/*
 *	Imprime une chaine en baqueslachifiant les bizarreries.
 */

static void print_escape (FILE* fp, char* s)
{
	fprintf (fp, "\"");
	for (;; s++) switch (*s) {
		case 0: fprintf (fp, "\""); return;
		case '\n': fprintf (fp, "\\n"); continue;
		case '\t': fprintf (fp, "\\t"); continue;
		case '\"': fprintf (fp, "\\\""); continue;
		case '\\': fprintf (fp, "\\\\"); continue;
		default:   fprintf (fp, "%c", *s); continue;
	}
}

/*
 *	Decompile une macro.
 */

static Boolean macro_to_lisp (char* filename)
{
	Command*	p;
	CommandProc	fun;
	CommandProc	prevfun = NULL;	/* n'importe quoi de different de InsertChar */
	int			n;
	char		string [1024];
	char*		cp = NULL;
	FILE*		fp;

	fp = fopen (filename, "w");
	if (fp == NULL)
		return False;

	fprintf (fp, "(defun default_macro (w n)\n");
	for (p = DefaultMacroBuffer; (fun = p -> function) != NULL; p++) {
		/*
		 *	compacter les suites de InsertChar en un
		 *	seul InsertString
		 */
		if (fun == InsertChar) {
			if (prevfun != InsertChar) cp = string;
			for (n = p -> count; --n >= 0;)
				*cp++ = (char) ((unsigned long) p -> arg & 0xff);
			prevfun = InsertChar;
			continue;
		}
		if (prevfun == InsertChar) {
			*cp = 0;
			fprintf (fp, "	(insert-string w ");
			print_escape (fp, string);
			fprintf (fp, ")\n");
			prevfun = fun;
		}
		/*
		 *	Trouver le nom de la fonction ainsi que
		 *	le type d'argument.
		 */
		cp = GetFunctionByAddr (p -> function);
		fprintf (fp, "	(%s w ", cp);
		if (p -> arg)
			print_escape (fp, p -> arg);
		fprintf (fp, " %d)\n", p -> count);
	}

	if (prevfun == InsertChar) {	/* le dernier coup */
		*cp = 0;
		fprintf (fp, "	(insert-string w ");
		print_escape (fp, string);
		fprintf (fp, ")\n");
	}
	fprintf (fp, ")\n");
	fclose (fp);
	return True;
}

/*
 *	Sauve la macro.
 */

static void saveMacroCont (WixWindow* data)
{
	if (macro_to_lisp (data -> fs_answer))
		Message (data -> active, "Macro saved.");
}

void SaveMacroAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = saveMacroCont;
	popQueryStringNoWait (data, "Save lisp macro", NULL, False);
}

void SetTitle (WixWindow* data, int m)
{
	char		tc [256];
	char		tmp [256];
	char*		dirname;
	char*		basename;
	static int	homeLen = 0;;

	if (!homeLen)
		homeLen = strlen (HomeDirectory);

	if (!strncmp (HomeDirectory, data -> filename, homeLen))
		sprintf (tc, "~%s", data -> filename + homeLen);
	else
		sprintf (tc, "%s", data -> filename);

	if ((basename = rindex (tc, '/')) != 0) {
		*basename++ = 0;
		dirname = tc;
	} else {
		basename = tc;
		dirname = "";
	}
	sprintf (tmp, m ? "* %s [%s]" : "%s [%s]", basename, dirname);
#if 1
	XtVaSetValues (data -> frame,
					XtNtitle,	tmp,
					NULL);
#endif
	XtVaSetValues (data -> toplevel,
					XtNtitle,		tmp,
					XtNiconName,	basename,
					NULL);
}

/*
 *	Load file
 */

static void rm_checkpoint_file (WixWindow* data)
{
	char	tc [256];

	sprintf (tc, "%s##ckp", data -> filename);
	unlink (tc);
}

Boolean loadFileFunction (Widget w, char* filename)
{
	WixWindow*	data;
	struct stat	b;

	XtVaGetValues (w, XtNuserData, &data, NULL);

	stat (filename, &b);
	if (S_ISDIR (b.st_mode))
		return False;

	if (WixLoadFile (w, filename) == False)
		return False;

	data -> date = b.st_mtime;
	data -> checkpoint = 0;	/* RAZ sauvegarde periodique */
	data -> undo_ptr = -1;	/* On ne peut pas (pour l'instant) dfaire un chargement */
	data -> prev_reason = -1;
	data -> undo_number = -1;
	rm_checkpoint_file (data);
	/* mettre le directory courant */
	if (filename [0] == '/') {
		strcpy (data -> directory, filename);
		rindex (data -> directory, '/') [0] = 0;
	}
	strcpy (data -> filename, filename);
	switchMode (SuffixToMode (data), data);
	WixRedraw (w);
	if (data -> shadow) {
		Widget	other = data -> active == data -> wix
							? data -> shadow
							: data -> wix;

		WixLoadFile (other, filename);
		WixRedraw (other);
	}
	SetTitle (data, 0);
	return True;
}

static void loadFileCont (WixWindow* data)
{
	WixWindow*	wp;
	char*		filename;

	filename = AbsPath (data -> directory, data -> fs_answer);
	for (wp = _head; wp; wp = wp -> next) {
		if (!wp -> free && !strcmp (wp -> filename, filename)) {
			popAlert (data, reloadMessage, True);
			WaitForResponse (data);
			if (data -> cancelled) {
				XMapRaised (XtDisplay (wp -> toplevel), XtWindow (wp -> toplevel));
				warpPointer (wp);
				return;
			} else
				break;
		}
	}
	if (loadFileFunction (data -> active, filename) == False)
		popAlert (data, loadFailedMessage, False);
}

void loadFileAsk (WixWindow* data)
{
	ClearRegion (data -> active);	/* effacer la region */
	data -> what_after = loadFileCont;
	popFileSelector (data, loadFileMessage, False);
}

void loadFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		m;

	XtVaGetValues (w,
		XtNbufferModified,	&m,
		XtNuserData,		&data,
		NULL);

	if (m) {
		data -> what_after = loadFileAsk;
		popAlert (data, discardChangesMessage, True);
	} else
		loadFileAsk (data);
}

static void __loadFileCont (Widget wix, char* reponse)
{
	WixWindow*	data;

	/* rcuprer nos infos */
	XtVaGetValues (XtParent (XtParent (wix)),
					XtNuserData,	&data,
					NULL);
	XtVaSetValues (data -> menubar, XtNsuffixes, 0, NULL);
	if (reponse != 0) {
		data -> fs_answer = reponse;
		loadFileCont (data);
	}
}

void FastLoadFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	SetConfirmAction (data -> menubar, __loadFileCont);
	XtVaSetValues (data -> menubar,
			XtNsuffixes,	wx_modes [data -> modeNumber].m_suf,
			NULL);
	PopMenubarDialog (data -> menubar, XtParent (data -> active),
					  " Load File ", DM_FileSelectionRead, data -> directory);
}

/*
 *	Recharge le fichier en edition.
 */

static void reloadFileCont (WixWindow* data)
{
	ClearRegion (data -> active);	/* effacer la region */
	if (loadFileFunction (data -> active, data -> filename) == False)
		popAlert (data, loadFailedMessage, False);
}

void reloadFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		m;

	XtVaGetValues (w,
		XtNbufferModified,	&m,
		XtNuserData,		&data,
		NULL);

	if (m) {
		data -> what_after = reloadFileCont;
		popAlert (data, discardChangesMessage, True);
	} else
		reloadFileCont (data);
}

/*
 *	Save As
 */

static void writeFileCont (WixWindow* data)
{
	char*		filename;
	struct stat	b;
	int			len;

	filename = data -> fs_answer;
	if ((len = WixSaveFile (data -> active, filename)) < 0) {
		popAlert (data, writeFailedMessage, False);
		return;
	}
	/*
	 *	remettre la pendule a l'heure.
	 */
	stat (filename, &b);
	data -> date = b.st_mtime;
	data -> checkpoint = 0;		/* RAZ sauvegarde periodique */
	rm_checkpoint_file (data);
	WixRedraw (data -> active);	/* pour bufferModifiedcallback */
	strcpy (data -> filename, filename);
	SetTitle (data, 0);
	showErrors (data -> active, T_WRITTEN, len);
}

static void writeFileAsk (WixWindow* data)
{
	char*	filename;

	filename = data -> fs_answer;
	if (access (filename, 0) == 0) {
		data -> what_after = writeFileCont;
		popAlert (data, fileExistsMessage, True);
	} else
		writeFileCont (data);
}

void writeFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = writeFileAsk;
	popFileSelector (data, writeFileMessage, False);
}

/*
 *	La mme utilisant la bote de dialogue rapide.
 */

static void __writeFileAsk (Widget wix, char* filename)
{
	if (filename != 0) {
		WixWindow*	data;
		/* rcuprer nos infos */
		XtVaGetValues (XtParent (XtParent (wix)),
						XtNuserData,	&data,
						NULL);
		data -> fs_answer = AbsPath (data -> directory, filename);
		/* faire faire le travail */
		writeFileAsk (data);
	}
}

void FastWriteFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	SetConfirmAction (data -> menubar, __writeFileAsk);
	PopMenubarDialog (data -> menubar, XtParent (data -> active),
					  writeFileMessage, DM_FileSelectionCreate, data -> directory);
}

/*
 *	Save file
 */

static void saveFileCont (WixWindow* data)
{
	struct stat	b;
	int			len;

	if ((len = WixSaveFile (data -> active, data -> filename)) < 0) {
		popAlert (data, writeFailedMessage, False);
		return;
	}
	/*
	 *	remettre la pendule a l'heure.
	 */
	stat (data -> filename, &b);
	data -> date = b.st_mtime;
	data -> checkpoint = 0;		/* RAZ sauvegarde periodique */
	rm_checkpoint_file (data);
	WixRedraw (data -> active);	/* pour bufferModifiedcallback */
	showErrors (data -> active, T_WRITTEN, len);
}

void saveFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	struct stat	b;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	/*
	 *	S'il n'y a pas de nom, simuler un 'save as'
	 */
	if (data -> filename [0] == 0) {
		writeFileAction (w, event, params, num_params);
		return;
	}
	/*
	 *	regarder si quelqu'un a modifie le fichier dans notre dos!
	 */
	if (stat (data -> filename, &b) >= 0 && data -> date != 0 && b.st_mtime != data -> date) {
		data -> what_after = saveFileCont;
		popAlert (data, fileModifiedMessage, True);
	} else
		saveFileCont (data);
}

/*
 *	Insert File
 */

static void insertFileAsk (WixWindow* data)
{
	char*	filename = AbsPath (data -> directory, data -> fs_answer);

	ClearRegion (data -> active);	/* effacer la region */
	if (WixInsertFile (data -> active, filename) == False) {
		popAlert (data, loadFailedMessage, False);
		return;
	}
	WixRedraw (data -> active);
}

void insertFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = insertFileAsk;
	popFileSelector (data, insertFileMessage, False);
}

static void __insertFileAsk (Widget wix, char* filename)
{
	WixWindow*	data;

	/* rcuprer nos infos */
	XtVaGetValues (XtParent (XtParent (wix)),
					XtNuserData,	&data,
					NULL);
	XtVaSetValues (data -> menubar, XtNsuffixes, 0, NULL);
	if (filename == 0)
		return;

	/* effacer la region */
	ClearRegion (data -> active);
	/* rcuprer le nom absolu du fichier */
	filename = AbsPath (data -> directory, filename);
	if (WixInsertFile (data -> active, filename) == False) {
		popAlert (data, loadFailedMessage, False);
		return;
	}
	WixRedraw (data -> active);
}

void FastInsertFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	XtVaSetValues (data -> menubar,
			XtNsuffixes,	wx_modes [data -> modeNumber].m_suf,
			NULL);
	SetConfirmAction (data -> menubar, __insertFileAsk);
	PopMenubarDialog (data -> menubar, XtParent (data -> active),
					  " Insert File ", DM_FileSelectionRead, data -> directory);
}

/*
 *	Clear buffer
 */

static void clearBufferCont1 (WixWindow* data)
{
	ClearRegion (data -> active);	/* effacer la region */
	(void) WixLoadFile (data -> active, "/dev/null");
	data -> date = 0;
	data -> checkpoint = 0;	/* RAZ sauvegarde periodique */
	data -> undo_ptr = -1;	/* On ne peut pas (pour l'instant) dfaire un chargement */
	data -> prev_reason = -1;
	data -> undo_number = -1;
	rm_checkpoint_file (data);
	data -> filename [0] = 0;	/* pas de fichier courant */
	WixRedraw (data -> active);
	if (data -> shadow) {
		Widget	other = data -> active == data -> wix
							? data -> shadow
							: data -> wix;

		WixLoadFile (other, "/dev/null");
		WixRedraw (other);
	}
	XtVaSetValues (data -> toplevel,
					XtNtitle,		"New File",
					XtNiconName,	"New File",
					NULL);
}

void clearBufferAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		m;

	XtVaGetValues (w,
		XtNbufferModified,	&m,
		XtNuserData,		&data,
		NULL);

	if (m) {
		data -> what_after = clearBufferCont1;
		popAlert (data, discardChangesMessage, True);
	} else
		clearBufferCont1 (data);
}

/*
 *	Delete File
 */

void deleteFileAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	char		tc [256];

	XtVaGetValues (w, XtNuserData, &data, NULL);

	sprintf (tc, "Do you really want to delete\n%s ?", data -> filename);
	popAlert (data, tc, True);
	WaitForResponse (data);
	if (data -> cancelled)
		return;

	unlink (data -> filename);
}

/*
 *	Change Directory
 */

static void changeDirectoryAsk (WixWindow* data)
{
	strcpy (data -> directory, data -> fs_answer);
}

void changeDirectoryAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = changeDirectoryAsk;
	popFileSelector (data, changeDirectoryMessage, True);
}

/*
 *	Delete window.
 */

static void deleteWindowCont (WixWindow* data)
{
	Window		root = RootWindowOfScreen (XtScreen (data -> toplevel));
	Display*	dpy = XtDisplay (data -> toplevel);

	data -> free = True;				/*  rutiliser */
	DestroyFSWindow (data -> menubar);	/* virer la fentre des noms de fichiers */
	XtUnrealizeWidget (data -> toplevel);	/* virer la fenetre */
	XtVaSetValues (data -> toplevel,	/* raz de la geometrie */
			XtNx,		0,
			XtNy,		0,
			XtNwidth,	0,
			XtNheight,	0,
			NULL);
	if (*data -> client_pid)	/* prevenir qu'on est parti */
		XChangeProperty (dpy, root, A_CLIENT, XA_STRING, 8, PropModeReplace,
				(unsigned char*) data -> client_pid, strlen (data -> client_pid));
	*data -> client_pid = 0;
	_SetSnapshot (".wx_state", False);
	XSync (dpy, False);
}

void deleteWindowAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	Boolean		m;
	WixWindow*	data;

	XtVaGetValues (w,
		XtNbufferModified,	&m,
		XtNuserData,		&data,
		NULL);

	if (m) {
		data -> what_after = deleteWindowCont;
		popAlert (data, discardChangesMessage, True);
	} else
		deleteWindowCont (data);
}

/*
 *	Rgle la taille de la fentre.
 */

static void setHints (WixWindow* data)
{
	FontInfo*	fp;
	Dimension	wi, he;
	Dimension	width;
	Dimension	height;
	int			cw, ch;
	int			lnw, st, ib;
#ifdef DEBUG
	static FILE*	_fp = 0;

	if (_fp == 0)
		_fp = fopen ("core/wx.msg", "a");
#endif
	XtVaGetValues (data -> wix,
			XtNfont,	&fp,
			XtNwidth,	&wi,
			XtNheight,	&he,
			XtNshadowThickness,	&st,
			XtNinnerBorder,		&ib,
			XtNlineNumberWidth,	&lnw,
			NULL);

	ch = fp -> f_height;
	cw = Width ('x', fp);

	XtVaGetValues (data -> toplevel,
			XtNheight,	&height,
			XtNwidth,	&width,
			NULL);
#ifdef DEBUG
	fprintf (_fp, "geometry = %dx%d\n", width, height);
	if (height >= 3000 || width >= 3000) {
		fprintf (_fp, "correcting...\n");
		height = ch * 24 + 2 * (st+ib);
		width = cw * 80 + lnw + 2 * (st+ib);
	}
	fflush (_fp);
#endif
	XtVaSetValues (data -> toplevel,
			XtNbaseHeight,	height - (int)he + 2 * (st+ib),
			XtNbaseWidth,	width + lnw - (int)wi + 2 * (st+ib),
			XtNheightInc,	ch,
			XtNwidthInc,	cw,
			NULL);
}

void createWindowCallback (Widget w, XtPointer client, XtPointer call)
{
	char*		directory = (char*) client;
	Widget		top = theWindow -> toplevel;
	Dimension	width;
	Dimension	height;
	Boolean		iconic;

	if (def_width != 0)
		XtVaSetValues (top,
			XtNwidth, def_width,
			XtNheight, def_height,
			NULL);
	XtRealizeWidget (top);
	if (def_width == 0 && def_height == 0) {
		XtVaGetValues (top,
			XtNwidth,	&width,
			XtNheight,	&height,
			NULL);
		def_width = width;
		def_height = height;
	}

	setHints (theWindow);
	if (XtVaGetValues (top, XtNiconic, &iconic, NULL), iconic)
	{
		XWMHints	hints;

		hints.flags = StateHint;
		hints.initial_state = IconicState;
		XSetWMHints (XtDisplay (top), XtWindow (top), & hints);
	}
	XtMapWidget (top);

	theWindow -> free = False;
	theWindow = newWindow (directory);
	_SetSnapshot (".wx_state", False);
}

void createWindowAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Dimension	width, height;

	XtVaGetValues (wix, XtNuserData, &data, NULL);

	strcpy (theWindow -> directory, data -> directory);
	theWindow -> free = False;

	XtVaGetValues (data -> toplevel,
			XtNwidth,	&width,
			XtNheight,	&height,
			NULL);
	XtVaSetValues (theWindow -> toplevel,
			XtNwidth,	width,
			XtNheight,	height,
			NULL);

	XtRealizeWidget (theWindow -> toplevel);
	setHints (theWindow);

	XtMapWidget (theWindow -> toplevel);

	theWindow = newWindow (HomeDirectory);
	_SetSnapshot (".wx_state", False);
}

/*
 *	Coupe la fenetre en deux.
 */

void splitWindowAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	char*		string;
	Boolean		m;

	XtVaGetValues (wix,
					XtNbufferModified, &m,
					XtNuserData, &data,
					NULL);

	if (data -> shadow)
		return;		/* deja coupee */

	data -> shadow = XtCreateScrolledWix (data -> pane, "edit", NULL, 0);
	switchMode (data -> modeNumber, data);	/* met aussi les bonnes translations */
	WixRegisterMessage (data -> shadow, showErrors);

	string = WixCopyBuffer (wix);

	XtVaSetValues (data -> shadow,
					XtNuserData, data,
					XtNvalue, string,
					NULL);

	XtVaSetValues (data -> shadow, XtNbufferModified, m, NULL);
	XtFree (string);

	WixSetInsertPosition (data -> shadow, WixGetInsertPosition (wix));
	WixRedraw (data -> shadow);

	XtAddCallback (data -> wix, XtNverifyModifyCallback, verifyCallback, (XtPointer) data -> shadow);
	XtAddCallback (data -> shadow, XtNverifyModifyCallback, verifyCallback, (XtPointer) data -> wix);
	XtAddCallback (data -> shadow, XtNbufferModifiedCallback, bufferModifiedCallback, (XtPointer) XtParent (data -> wix));

	XtManageChild (XtParent (data -> shadow));
	XtSetKeyboardFocus (XtParent (data -> shadow), data -> shadow);
}

/*
 *	Degage le deuxieme pane.
 */

void removePaneAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (wix, XtNuserData, &data, NULL);

	if (!data -> shadow)
		return;		/* ya rien  retirer */

	XtRemoveCallback (data -> wix, XtNverifyModifyCallback, verifyCallback, (XtPointer) data -> shadow);
	XtRemoveCallback (data -> shadow, XtNbufferModifiedCallback, bufferModifiedCallback, (XtPointer) XtParent (data -> wix));
	XtDestroyWidget (XtParent (data -> shadow));
	data -> shadow = NULL;
	data -> active = data -> wix;
	XtSetKeyboardFocus (XtParent (data -> wix), data -> wix);
}

/*
 *	Pour indiquer le pane actif.
 */

void selectPaneAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);

	if (data -> active != w) {
		/* effacer le cadre */
		XtVaSetValues (data -> active,
			XtNframeThickness, 0,
			NULL);
		data -> active = w;
		XtSetKeyboardFocus (data -> pane, XtParent (w));
		XtVaSetValues (w,
			XtNfullCursor, True,
			XtNframeThickness, 2,
			NULL);
		StartBlink (w);
	}
}

/*
 *	(Des)Habille une fenetre.
 */

void StripWindow (WixWindow* data)
{
	Widget	sw = XtParent (data -> wix);
	Widget	other = data -> shadow != NULL
						? data -> shadow
						: data -> wix2 != NULL
							? data -> wix2
							: NULL;
	data -> stripped = True;
	XtVaSetValues (data -> menubar,
			XtNhidden,		True,
			XtNmnemonicOn,	False,
			NULL);
	XtVaSetValues (data -> pane, XtNmargin, 0, NULL);
	XtVaSetValues (data -> wix, XtNshadowThickness, 0, NULL);
	XtUnmanageChild (data -> menubar);
	XtUnmanageChild (SwHSB (sw));
	XtUnmanageChild (SwVSB (sw));
	if (other) {
		XtVaSetValues (other, XtNshadowThickness, 0, NULL);
		XtUnmanageChild (SwHSB (XtParent (other)));
		XtUnmanageChild (SwVSB (XtParent (other)));
	}
}

void DressWindow (WixWindow* data)
{
	Widget	sw = XtParent (data -> wix);
	Widget	other = data -> shadow != NULL
						? data -> shadow
						: data -> wix2 != NULL
							? data -> wix2
							: NULL;

	data -> stripped = False;
	XtVaSetValues (data -> pane, XtNmargin, 2, NULL);
	XtVaSetValues (data -> wix, XtNshadowThickness, 2, NULL);
	XtManageChild (SwVSB (sw));
	XtManageChild (SwHSB (sw));
	XtManageChild (data -> menubar);
	XtVaSetValues (data -> menubar,
			XtNmnemonicOn,	True,
			XtNhidden,		False,
			 NULL);
	if (other) {
		XtVaSetValues (other, XtNshadowThickness, 2, NULL);
		XtManageChild (SwHSB (XtParent (other)));
		XtManageChild (SwVSB (XtParent (other)));
	}
}

void stripAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	if (w != data -> wix && w != data -> wix2)
		return;

	if (data -> stripped)
		DressWindow (data);
	else
		StripWindow (data);

	/* setHints (data);	/* a a tendance  rduire la taille du shell?!!?? */
}

/*
 *	Met la fentre courante en tte de liste.
 */

static void putOnTop (WixWindow* data)
{
	WixWindow*	p;
	WixWindow*	prec;

	if (_head == data)
		return;		/* dj en tte */

	for (p = _head; p != data; p = p -> next)
		prec = p;

	p = _head;
	_head = data;
	prec -> next = data -> next;
	data -> next = p;
}

/*
 *	Pour circuler dans les fenetres
 */

static int nextWindowFunction (Widget wix, int count, void* arg)
{
	XEvent*		event = (XEvent*) arg;
	WixWindow*	data;
	WixWindow*	wp;

	SaveCommand (wix, nextWindowFunction, arg, False);

	XtVaGetValues (wix, XtNuserData, &data, NULL);
	for (;;) {
		/*
		 *	Trouver la fentre dans la liste.
		 */
		for (wp = _head; ; wp = wp -> next)
			if (wp == data) {
				do
					wp = wp -> next ? wp -> next : _head;
					while (wp -> free);
				if (wp == data)
					return 0;		/* il n'y a qu'une fentre */
				break;
			}
		/*
		 *	La mettre au premier plan.
		 */
		XMapRaised (XtDisplay (wp -> toplevel), XtWindow (wp -> toplevel));
		warpPointer (wp);
		data = wp;
		for (;;) {
			XEvent			ev;
			KeySym			keysym;
			XtAppContext	app = XtWidgetToApplicationContext (data -> pane);

			while (XtAppPeekEvent (app, &ev) == False)
				;
			switch (ev.type) {
				case Expose:
				case GraphicsExpose:
					XtAppNextEvent (app, &ev);
					XtDispatchEvent (&ev);
					continue;

				case KeyRelease:
					XtAppNextEvent (app, &ev);
					XLookupString (&ev.xkey, NULL, 0, &keysym, 0);
					if (keysym == XK_Control_L || keysym == XK_Control_R ||
						keysym == XK_Meta_L || keysym == XK_Meta_R ||
						keysym == XK_Shift_L || keysym == XK_Shift_R)
					{
						putOnTop (data);
						return 1;
					}
					continue;

				case KeyPress:
					if (ev.xkey.keycode == ((XKeyEvent*) event) -> keycode) {
						XtAppNextEvent (app, &ev);
						break;
					}
					putOnTop (data);
					return 1;

				default:
					XtAppNextEvent (app, &ev);
					continue;
			}
			break;
		}
	}
}

void nextWindowAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	nextWindowFunction (wix, 1, event);
}

/*
 *	Start macro
 */

void startMacroAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	Message (w, macroLearnMessage);
	start_macro_learn (w);
}

/*
 *	Stop macro
 */

void stopMacroAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	Message (w, macroLearnedMessage);
	stop_macro_learn (w);
}

/*
 *	Action d'execution de macro nommee.
 */

static int exec_lisp_command (Widget w, int cnt, void* arg)
{
	WixWindow*	data;
	char*		name = (char*) arg;
	char		msg [1024];

	XtVaGetValues (w, XtNuserData, &data, NULL);
	ClearRegion (w);	/* effacer la region */
	msg [0] = 0;
	lisp_eval_lambda (msg, name, (void*) w, cnt);	/* fonction lisp */
	if (msg [0]) {
		data -> what_after = NULL;
		popAlert (data, msg, False);
	}
	else
		WixRedraw (w);
	SaveCommand (w, exec_lisp_command, name, True);
	return 1;
}

void executeNamedMacroAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	extern int	num_arg;

	if (*num_params == 1)
		exec_lisp_command (w, num_arg, params [0]);
}

/*
 *	Pour terminer le serveur.
 */

void quitCont (WixWindow* data)
{
	WixWindow*	wp;

	if (!debugMode)
		XDeleteProperty (XtDisplay (data -> active),
			DefaultRootWindow (XtDisplay (data -> active)),
			A_HERE);
	XSync (XtDisplay (data -> active), True);
	_SetSnapshot (".wx_state", False);
	exit (0);
}

static void quitSaveCont (WixWindow* data)
{
	WixWindow*	wp;

#if 1
	XtPopdown (XtParent (XtParent (data -> query_string)));	/* fermer la boite */
#else
	XtPopdown (XtParent (data -> query_string));	/* fermer la boite */
#endif
	for (wp = _head; wp; wp = wp -> next) {
		Boolean	m;

		if (!wp -> free && (XtVaGetValues (wp -> wix, XtNbufferModified, &m, NULL), m))
			WixSaveFile (wp -> wix, wp -> filename);
	}
	quitCont (data);
}

static void quitCloseCont (WixWindow* data)
{
#if 1
	XtPopdown (XtParent (XtParent (data -> query_string)));	/* fermer la boite */
#else
	XtPopdown (XtParent (data -> query_string));	/* fermer la boite */
#endif
	if (!debugMode)
		XDeleteProperty (XtDisplay (data -> active),
			DefaultRootWindow (XtDisplay (data -> active)),
			A_HERE);
	XSync (XtDisplay (data -> active), True);
	unlink (".wx_state");	/* virer le snapshot */
	exit (0);
}

void quitAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	WixWindow*	wp;
	Boolean		m;
	static char	msg [] = "There are modified but not saved buffers.\nDo you really want to quit the WiX server?";
	/*
	 *	Vrifier si des buffers sont encore modifis.
	 */
	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = quitCont;
	for (wp = _head; wp; wp = wp -> next) {
		if (wp -> free)
			continue;

		if (XtVaGetValues (wp -> wix, XtNbufferModified, &m, NULL), m)
		{
			data -> apply_after = quitSaveCont;
			PopBye (data, msg, False, "Exit", "Cancel", "Save and Exit");
			return;
		}
	}
	data -> apply_after = quitCloseCont;
	PopBye (data, quitMessage, False, "Exit", "Cancel", "Close and Exit");
}

/*
 *	Ou sommes nous?
 */

void lineCounterAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	char	tc [64];
	int		lineno;
	int		nolines;

	XtVaGetValues (w,
		XtNcurrentLineNumber,	&lineno,
		XtNbufferLineNumber,	&nolines,
		NULL);

	sprintf (tc, "Line: %d Total: %d", lineno+1, nolines+1);
	Message (w, tc);
}

void bufferModifiedCallback (Widget wix, XtPointer client, XtPointer call)
{
	Widget		modif = (Widget) client;
	int			m = (int) (long) call;
	WixWindow*	data;
	Pixmap		pix;
	Boolean		ro;
	UndoRedoCallbackStruct	urcb;

	XtVaGetValues (wix,
		XtNuserData, &data,
		XtNreadOnly, &ro,
		NULL);

	if (m == 0 && !data -> undo_mode) {
		urcb.reason = UNMODIFIED_MARK;
		UndoRedoCallback (wix, (XtPointer) data, (XtPointer) &urcb);
	}

	if (!ro) {
		pix = m ? modifPixmap : (Pixmap) 0;
		XtVaSetValues (modif, XtNpixmap, pix, NULL);
		SetTitle (data, m);
	}
}

void modifCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data;
	int			m;
	Boolean		ro;
	UndoRedoCallbackStruct	urcb;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	XtVaGetValues (data -> active,
		XtNbufferModified, &m,
		XtNreadOnly, &ro,
		NULL);

	if (!m || ro)
		return;

	if (! data -> undo_mode) {
		urcb.reason = UNMODIFIED_MARK;
		UndoRedoCallback (w, (XtPointer) data, (XtPointer) &urcb);
	}

	XtVaSetValues (XtParent (data -> active), XtNpixmap, NULL, NULL);
	XtVaSetValues (data -> active, XtNbufferModified, False, NULL);
	if (data -> shadow) {
		Widget	other = data -> active == data -> wix
							? data -> shadow
							: data -> wix;

		XtVaSetValues (other, XtNbufferModified, False, NULL);
	}
	SetTitle (data, 0);
}

/*
 *	Retire la marque  'buffer modifi'
 */

void UnmodifyBufferAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	modifCallback (w, NULL, NULL);
}

/*
 *	Passe en mode read-only et lycee de Versailles
 */

void toggleEditableAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		m;
	Boolean		ro;

	XtVaGetValues (wix,
		XtNuserData, &data,
		XtNbufferModified, &m,
		XtNreadOnly, &ro,
		NULL);

	if (ro) {
		ro = False;
		XtVaSetValues (XtParent (wix),
				XtNpixmap, m ? modifPixmap : (Pixmap)NULL,
				NULL);
	} else {
		ro = True;
		XtVaSetValues (XtParent (wix),
				XtNpixmap, m ? modifPixmap : lockPixmap,
				NULL);
	}

	if (data -> shadow)
		XtVaSetValues (data -> shadow, XtNreadOnly, ro, NULL);
	XtVaSetValues (wix, XtNreadOnly, ro, NULL);
}

/*
 *	Montre ou cache les numeros de ligne.
 */

void ShowLineNumberAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		sln;
	int			lnw;

	XtVaGetValues (wix,
		XtNshowLineNumber,	&sln,
		XtNuserData,		&data,
		NULL);

	XtVaSetValues (wix, XtNshowLineNumber, !sln, NULL);
	if (data -> shadow) {
		Widget	other = wix == data -> wix
							? data -> shadow
							: wix;

		XtVaSetValues (other, XtNshowLineNumber, !sln, NULL);
	}
	XtVaGetValues (wix, XtNlineNumberWidth, &lnw, NULL);
	XtVaSetValues (data -> ruler, XtNleftOffset, 15 + lnw, NULL);
}

/*
 *	Auto wrap.
 */

void AutoWrapAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		wrap;

	XtVaGetValues (wix,
		XtNwrap,		&wrap,
		XtNuserData,	&data,
		NULL);

	XtVaSetValues (wix, XtNwrap, !wrap, NULL);
	if (data -> shadow)
		XtVaSetValues (data -> shadow, XtNwrap, !wrap, NULL);
}

/*
 *	Affiche les caractres non ascii en octal
 */

void ToggleOctalDisplayAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	Boolean		bin_edit;

	XtVaGetValues (wix,
		XtNbinaryEdit,	&bin_edit,
		XtNuserData,	&data,
		NULL);

	XtVaSetValues (wix, XtNbinaryEdit, !bin_edit, NULL);
	if (data -> shadow)
		XtVaSetValues (data -> shadow, XtNbinaryEdit, !bin_edit, NULL);
}

/*
 *	Montre la rgle de placement des taquets de tabulation.
 */

void showRulerAction (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (wix, XtNuserData, &data, NULL);

	if (XtIsManaged (data -> ruler))
		XtUnmanageChild (data -> ruler);
	else
		XtManageChild (data -> ruler);
}
