/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/*************************************************************************
*
*	La boite de diagnostics
*
*************************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/PushButton.h>
#include <Xpa/ScrolledWin.h>
#include <Xpa/PanedWin.h>
#include <Xpa/TextList.h>
#include <Xpa/Wix.h>

#include "wix.h"

extern char* AbsPath (char*, char*);	/* import de FSelector.c */

void ClearDiagBox (WixWindow* data)
{
	int	count;

	XtVaGetValues (data -> diag_list,
					XtNuserData, &count,
					NULL);

	while (count > 0) {
		WixDeleteMark (data -> active, (void*) (long) (count + 1024));
		-- count;
	}
	XtVaSetValues (data -> diag_list,
					XtNuserData, 0,
					NULL);
	WixClearBuffer (data -> diag_list);
}

/*
 *	La callback pour sortir.
 */

static void diag_done (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;

	XtUnmanageChild (XtParent (data -> diag_list));
	ClearDiagBox (data);
}

static void diag_select (Widget list, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;
	int			l;

	XtVaGetValues (list, XtNcurrentLineNumber, &l, NULL);

	l += 1;	/* les lignes sont comptees a partir de 0 */
	WixGotoMark (data -> active, (void*) (long) (l + 1024));
	WixRedraw (data -> active);
}

/*
 *	Initialise la boite.
 */

void CreateDiagBox (Widget parent, WixWindow* data)
{
	Arg			args [16];
	Cardinal	nargs;
	Dimension	height, wheight;
	int			offset;

	nargs = 0;
	XtSetArg (args [nargs], XtNtranslations, listTrans); nargs++;
	XtSetArg (args [nargs], XtNcursor, ArrowCurs); nargs++;
	XtSetArg (args [nargs], XtNnoCursor, True); nargs++;
	XtSetArg (args [nargs], XtNlineNum, 4); nargs++;
	data -> diag_list = XtCreateScrolledWix (parent, "list", args, nargs);

	XtVaGetValues (XtParent (data -> diag_list),
		XtNheight, &height,
		XtNoffset, &offset,
		NULL);
	XtVaGetValues (data -> diag_list,
		XtNheight, &wheight,
		NULL);

	XtVaSetValues (XtParent (data -> diag_list),
		XtNlabel, "Compilation Diagnostics",
		XtNpixmap, OkPixmap,
		NULL);

	XtAddCallback (data -> diag_list, XtNactivateCallback, diag_select, data);
	XtAddCallback (XtParent (data -> diag_list), XtNselect, diag_done, data);
}

void PopDiagBox (WixWindow* data)
{
	BeginningOfBuffer (data -> diag_list, 1, NULL);
	WixRedraw (data -> diag_list);
	WixGotoMark (data -> active, (void*) 1001);
	WixRedraw (data -> active);
	XtManageChild (XtParent (data -> diag_list));
}

/*************************************************************************
*
*	La sous fentre des tags
*
*************************************************************************/

static WixWindow* edit_file (WixWindow* data, char* name)
{
	WixWindow*	wp;
	char*		filename = AbsPath (data -> directory, name);

	for (wp = _head; wp; wp = wp -> next) {
		if (!wp -> free && !strcmp (wp -> filename, filename)) {
			/* Trouv */
			BeginningOfBuffer (wp -> active, 1, NULL);
			RemoveCursor (wp -> active);
			XMapRaised (XtDisplay (wp -> toplevel), XtWindow (wp -> toplevel));
			return wp;
		}
	}
	wp = theWindow;		/* la fentre qui sera cre  l'instruction suivante */
	createWindowCallback (data -> active, data -> directory, NULL);
	(void) loadFileFunction (wp -> active, filename);
	return wp;
}

/*
 *	La callback de slection.
 */

static void tag_select (Widget list, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;
	FILE*		fp;
	int			nl;
	char		buffer [256];
	char*		p;
	char*		q;
	char*		filename;
	char*		pattern;

	/* lire la nl-ime ligne du fichier des tags */
	p = AbsPath (data -> directory, "tags");
	fp = fopen (p, "r");
	if (fp != NULL) {
		for (nl = (long) call + 1; --nl >= 0;)
			fgets (buffer, sizeof (buffer), fp);
		fclose (fp);
		/* chercher le nom du fichier source */
		for (filename = buffer; *filename++ != '\t'; )
			;
		for (pattern = filename; *pattern++ != '\t';)
			;
		pattern [-1] = 0;
		/* chercher le pattern */
		for (pattern += 2, p = pattern; *p++ != '\n';)
			;
		data = edit_file (data, filename);
		if (pattern [-2] != '/') {
			/* numro de ligne */
			p [1] = 0;
			nl = atoi (pattern-2);
			(void) WixGotoLine (data -> active, nl);
		} else {
			/* chane de caractres */
			p [-3] = 0;
			/* supprimer les \ du pattern */
			if ((p = strchr (pattern, '\\')) != NULL) {
				for (q = p; *p;) {
					if (*p == '\\')
						++p;
					*q++ = *p++;
				}
				*q = 0;
			}
			WixForwardSearch (data -> active, pattern, 1);
			BeginningOfLine (data -> active, 1, NULL);
		}
		WixRedraw (data -> active);
	}
}

/*
 *	Initialise la boite.
 */

static void createTagBox (WixWindow* data)
{
	Widget		top;
	Arg			args [16];
	Cardinal	nargs;
	/*
	 *	Creer le shell.
	 */
	top = XtVaCreatePopupShell ("tags-shell", topLevelShellWidgetClass, ghost,
								XtNiconName,			"Function List",
								XtNallowShellResize,	True,
								NULL);
	nargs = 0;
	XtSetArg (args [nargs], XtNlineNum, 25); nargs++;
	data -> tag_list = XpaCreateScrolledList (top, "tag-list", args, nargs);
	XtAddCallback (data -> tag_list, XtNselect, tag_select, data);
	XtManageChild (XtParent (data -> tag_list));
}

static void popTagBox (WixWindow* data)
{
	FILE*	fp;
	char*	p;
	char	buffer [256];
	char**	tl;
	int		nitem;
	int		c;
	static char	tc [64];

	p = wx_modes [data -> modeNumber].m_tagcmd;
	if (p == NULL) {
		sprintf (tc, "No tag command defined\nfor the '%s' mode.",
					wx_modes [data -> modeNumber].m_name);
		popAlert (data, tc, False);
		return;
	}
	sprintf (buffer, "cd %s; %s >/dev/null 2>&1", data -> directory, p);
	system (buffer);

	p = AbsPath (data -> directory, "tags");
	fp = fopen (p, "r");
	if (fp == NULL)
		return;
	/* compter le nombre de lignes */
	for (nitem = 1; (c = getc (fp)) != EOF;)
		if (c == '\n')
			++nitem;

	rewind (fp);
	/* rcuprer la mmoire de l'ancienne liste */
	XtVaGetValues (data -> tag_list,
					XtNitemList, &tl,
					NULL);
	if (tl != NULL) {
		char**	l;

		for (l = tl; *l != NULL; ++l)
			XtFree (*l);
		XtFree ((char*)tl);
	}
	/* allouer la mmoire ncessaire */
	tl = (char**) XtMalloc (nitem * sizeof (char*));
	/* afficher le nombre de fonctions */
	sprintf (tc, "%d functions.", nitem-1);
	p = XtNewString (tc);
	XtVaSetValues (XtParent (data -> tag_list),
					XtNlabel, p,
					NULL);
	/* afficher le directory (dans la barre de titre) */
	if (!strncmp (HomeDirectory, data -> directory, c = strlen (HomeDirectory)))
		sprintf (buffer, "~%s", data -> directory + c);
	else
		sprintf (buffer, "%s", data -> directory);
	XtVaSetValues (XtParent (XtParent (data -> tag_list)),
					XtNtitle,	buffer,
					NULL);
	for (nitem = 0; fgets (buffer, sizeof (buffer), fp); ++nitem) {
		for (p = buffer; *p++ != '\t';)
			;
		p [-1] = 0;
		tl [nitem] = XtNewString (buffer);
	}
	tl [nitem] = NULL;
	XtVaSetValues (data -> tag_list,
					XtNitemList, tl,
					NULL);
	fclose (fp);
	XtPopup (XtParent (XtParent (data -> tag_list)), XtGrabNone);
	XtMapWidget (XtParent (XtParent (data -> tag_list)));
}

/*
 *	Change la commande de production des tags pour le mode courant.
 */

static void setTagCommandCont (WixWindow* data)
{
	if (wx_modes [data -> modeNumber].m_tagcmd)
		XtFree (wx_modes [data -> modeNumber].m_tagcmd);

	wx_modes [data -> modeNumber].m_tagcmd = XtNewString (data -> fs_answer);
}

void SetTagCommandAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	static char	tc [64];

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = setTagCommandCont;
	sprintf (tc, "Tag command for mode\n%s", wx_modes [data -> modeNumber].m_name);
	popQueryStringNoWait (data, tc, wx_modes [data -> modeNumber].m_tagcmd, False);
}

/*
 *	L'action qui fait apparatre la boite des tags.
 */

void ShowTagsAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	WixWindow*	wp;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	if (!data -> tag_list) {
		/* reprendre la tag_list d'une fentre du mme directory */
		for (wp = _head; wp; wp = wp -> next)
			if (!wp -> free && wp -> tag_list && !strcmp (wp -> directory, data -> directory)) {
				/* Trouv */
				popTagBox (wp);
				return;
			}
		createTagBox (data);
	}
	popTagBox (data);
}
