/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/************************************************************************
*
*	Les boites de dialogues
*
*************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#if SYS_DIR
#include	<sys/dir.h>
#else
#include	<dirent.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/Query.h>
#include <Xpa/Icon.h>
#include <Xpa/FSelector.h>
#include <Xpa/PushButton.h>
#include <Xpa/PanedWin.h>
#include <Xpa/MenuBar.h>
#include <Xpa/Wix.h>
#include <Xpa/Row.h>
#if 1
#include <Xpa/Frame.h>
#endif

#include "wix.h"

/*
 *	WM_DELETE_WINDOW handler.
 */

static void wm_delete_callback (Widget w)
{
	Widget*		wl;
#define dbox wl[0]

	XtVaGetValues (w, XtNchildren, &wl, NULL);
#if 1
	XtVaGetValues (wl [0], XtNchildren, &wl, NULL);
#endif
	if (XtClass (dbox) == queryWidgetClass)
		FakeClick (QueryGetChild (dbox, CANCEL_BUTTON));
	else if (XtClass (dbox) == fileSelectorWidgetClass)
		FakeClick (FileSelectorGetChild (dbox, FS_CANCEL));
#undef dbox
}

/*
 *	Les callbacks des boutons ok et cancel.
 */

static void okCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*		data = (WixWindow*) client;
	String			rep = (String) call;
	void			(*cont)();
	XtAppContext	app = XtWidgetToApplicationContext (w);

#if 1
	XtPopdown (XtParent (XtParent (w)));
#else
	XtPopdown (XtParent (w));
#endif
	data -> popped = False;
	data -> cancelled = False;
	if (rep)
		data -> fs_answer = rep;	/* Ne pas bouziller fs_answer s'il n'y a pas de reponse */

	XSync (XtDisplay (w), False);
	while (XtAppPending (app))
		XtAppProcessEvent (app, XtIMAll);

	data -> apply_after = NULL;
	if (data -> what_after) {
		cont = data -> what_after;
		data -> what_after = NULL;
		(*cont) (data);
	}
}

static void applyCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*		data = (WixWindow*) client;
	String			rep = (String) call;
	void			(*cont)();
	XtAppContext	app = XtWidgetToApplicationContext (w);
	Boolean			toggle;

	/*
	 *	Si c'est pas le apply d'une query box ou si c'est un toggle button laisser tomber.
	 */
	if (w != data -> query_string ||
		(XtVaGetValues (QueryGetChild (w, APPLY_BUTTON), XtNtoggle, &toggle, NULL),
		 toggle))
	{
		return;
	}
	/*
	 *	Ok, faire son boulot.
	 */
	data -> popped = False;			/* pas exactement vrai, mais a permet de */
									/* rester dans WaitForResponse avec la boite */
									/* toujours l ! */
	data -> cancelled = False;
	if (rep)
		data -> fs_answer = rep;	/* Ne pas bouziller fs_answer s'il n'y a pas de reponse */

	XSync (XtDisplay (w), False);
	while (XtAppPending (app))
		XtAppProcessEvent (app, XtIMAll);

	data -> what_after = NULL;
	if (data -> apply_after) {
		cont = data -> apply_after;
		data -> apply_after = NULL;
		(*cont) (data);
	}
}

static void cancelCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;

#if 1
	XtPopdown (XtParent (XtParent (w)));
#else
	XtPopdown (XtParent (w));
#endif
	data -> popped = False;
	data -> cancelled = True;
	data -> what_after = NULL;
	data -> apply_after = NULL;
}

/*
 *	Resensibilise la fenetre.
 */

static void popdownCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;

	if (data -> pane)
		XtSetSensitive (data -> pane, True);
}

/*
 *	Cree le FileSelector (dans le but d'avoir les icones deja chargees pour
 *	augmenter les chances d'utiliser les couleurs disponibles).
 */

void _CreateFileSelector (WixWindow* data)
{
#if 1
	Widget	frame;
#endif
	Widget	shell;

	shell = XtVaCreatePopupShell ("popup-fsel", transientShellWidgetClass,
				data -> toplevel,
				XtNwindowGroup,		XtWindow (data -> toplevel),
				XtNtransientFor,	data -> toplevel,
				XtNallowShellResize,True,
				NULL);
	SetDeleteWindow (shell, wm_delete_callback);
	XtAddCallback (shell, XtNpopdownCallback, popdownCallback, data);

#if 1
	frame = XtCreateManagedWidget ("frame", frameWidgetClass, shell, NULL, 0);
	data -> fs = XtCreateWidget ("fsel", fileSelectorWidgetClass, frame, NULL, 0);
#else
	data -> fs = XtCreateWidget ("fsel", fileSelectorWidgetClass, shell, NULL, 0);
#endif
	XtVaSetValues (data -> fs, XtNuserData, data, NULL);
	XtVaSetValues (FileSelectorGetChild (data -> fs, FS_ICON),
				XtNnoDrag,	True,
				NULL);
	XtAddCallback (data -> fs, XtNokCallback, okCallback, data);
	XtAddCallback (data -> fs, XtNcancelCallback, cancelCallback, data);
}

/*
 *	Fait apparaitre le selecteur de fichiers.
 */

void popFileSelector (WixWindow* data, char* title, Boolean dirOnly)
{
	Position	x, y;
	Dimension	width, height;

	if (!data -> fs)
		_CreateFileSelector (data);

	XtSetSensitive (data -> pane, False);
	XtVaGetValues (data -> toplevel,
		XtNx,	&x,
		XtNy,	&y,
		XtNwidth,	&width,
		XtNheight,	&height,
		NULL);
#if 1
	XtVaSetValues (XtParent (XtParent (data -> fs)),
#else
	XtVaSetValues (XtParent (data -> fs),
#endif
		XtNx,	x+20,
		XtNy,	y+20,
		XtNwidth,	width-40,
		XtNheight,	height-40,
		NULL);
#if 1
	XtPopup (XtParent (XtParent (data -> fs)), XtGrabNone);
#else
	XtPopup (XtParent (data -> fs), XtGrabNone);
#endif
	PopFileSelector (data -> fs, title, data -> filename, data -> directory, dirOnly);
}

/*
 *	Centre la boite autour de la souris et insensibilise la fenetre.
 *	si data -> pane est NULL, centre la boite dans l'cran.
 */

static void popupCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;
	int			mode;
	Dimension	bw, bh;
	Position	fx, fy;
	Dimension	fw, fh;
	int			sx, sy;

	if (data -> pane) {
		XtTranslateCoords (data -> pane, 0, 0, &fx, &fy);
		XtVaGetValues (data -> pane,
			XtNwidth,	&fw,
			XtNheight,	&fh,
			NULL);

		sx = fx + (fw / 2);
		sy = fy + (fh / 2);
	} else {
		sx = WidthOfScreen (XtScreen (w)) / 2;
		sy = HeightOfScreen (XtScreen (w)) / 2;
	}
	XWarpPointer (XtDisplay (w), None, RootWindowOfScreen (XtScreen (w)),
			0, 0, 0, 0,
			sx, sy);
	XtVaGetValues (data -> query_string,
		XtNprompt,	&mode,
		XtNwidth,	&bw,
		XtNheight,	&bh,
		NULL);
	XtVaSetValues (w, XtNx, sx - bw/2, XtNy, sy - bh/2, NULL);
	if (data  -> pane)
		XtSetSensitive (data -> pane, False);
	data -> popped = True;
	if (mode > 0)
		XtSetKeyboardFocus (data -> query_string, QueryGetChild (data -> query_string, EDIT));
}

/*
 *	Cree la boite de recherches & remplacements.
 */

void CreateQueryString (WixWindow* data)
{
#if 1
	Widget	frame;
#endif
	Widget	w;
	String	name = "WiX update";

	if (!data -> query_string) {
		w = XtVaCreatePopupShell ("query_popup", transientShellWidgetClass, data -> toplevel,
			XtNwindowGroup,		XtWindow (data -> toplevel),
			XtNtransientFor,	data -> toplevel,
			XtNallowShellResize,	True,
			NULL);
		SetDeleteWindow (w, wm_delete_callback);
#if 1
		frame = XtCreateManagedWidget ("frame", frameWidgetClass, w, NULL, 0);
		data -> query_string = XtVaCreateManagedWidget ("query", queryWidgetClass, frame,
#else
		data -> query_string = XtVaCreateManagedWidget ("query", queryWidgetClass, w,
#endif
			XtNuserData,	data,
			XtNprompt,	2,
			XtNapplyLabel,	"Use Regexp",
			NULL);
		XtSetKeyboardFocus (w, data -> query_string);
		XtAddCallback (w, XtNpopupCallback, popupCallback, data);
		XtAddCallback (w, XtNpopdownCallback, popdownCallback, data);
		XtAddCallback (data -> query_string, XtNokCallback, okCallback, data);
		XtAddCallback (data -> query_string, XtNapplyCallback, applyCallback, data);
		XtAddCallback (data -> query_string, XtNcancelCallback, cancelCallback, data);
		XtVaSetValues (QueryGetChild (data -> query_string, APPLY_BUTTON),
				XtNtoggle, True, NULL);
	}
	XtVaGetValues (data -> toplevel, XtNiconName, &name, NULL);
#if 1
	XtVaSetValues (XtParent (XtParent (data -> query_string)),
#else
	XtVaSetValues (XtParent (data -> query_string),
#endif
		XtNtitle,	name,
		NULL);
}

/*
 *	Attend une reponse.
 */

void WaitForResponse (WixWindow* data)
{
	XtAppContext	app = XtWidgetToApplicationContext (data -> toplevel);

	for (data -> popped = True; data -> popped;)
		XtAppProcessEvent (app, XtIMAll);
}

/*
 *	Fait apparaitre la boite de string.
 */

void popQueryStringNoWait (WixWindow* data, char* title, char* content, Boolean apply)
{
	Widget	w;

	CreateQueryString (data);

	XtManageChild (QueryGetChild (data -> query_string, CANCEL_BUTTON));
	w = QueryGetChild (data -> query_string, APPLY_BUTTON);
	XtVaSetValues (w,
					XtNset,		False,
					XtNtoggle,	True,
					NULL);
	if (apply && !XtIsManaged (w))		XtManageChild (w);
	else if (!apply && XtIsManaged (w))	XtUnmanageChild (w);

	XtVaSetValues (data -> query_string,
		XtNokLabel,			"Ok",
		XtNcancelLabel,		"Cancel",
		XtNprompt,			1,
		XtNdefaultButton,	OK_BUTTON,
		XtNapplyLabel,		"Use Regexp",
		XtNpixmap,			NULL,
		XtNtitle,			title,
		XtNlabel1,			NULL,
		XtNlabel2,			NULL,
		NULL);

#if 1
	XtVaSetValues (XtParent (data -> query_string),
			XtNtitle,	title,
			NULL);
#endif
	/*
	 *	Mettre le contenu par defaut.
	 */
	w = QueryGetChild (data -> query_string, EDIT);
	if (!content)	content = "";
	WixReplace (w, 0, -1, content, -1);
	SelectBuffer (w);
#if 1
	XtPopup (XtParent (XtParent (data -> query_string)), XtGrabNone);
#else
	XtPopup (XtParent (data -> query_string), XtGrabNone);
#endif
}

/*
 *	Fait apparaitre la boite de remplacement.
 */

void popQueryReplace (WixWindow* data, char* title, char* old, char* new)
{
	Widget	w;

	CreateQueryString (data);
	/*
	 *	Montrer les boutons cancel && regexpr.
	 */
	w = QueryGetChild (data -> query_string, APPLY_BUTTON);
	XtVaSetValues (w,
					XtNset,		False,
					XtNtoggle,	True,
					NULL);
	XtManageChild (w);
	XtManageChild (QueryGetChild (data -> query_string, CANCEL_BUTTON));
	XtVaSetValues (data -> query_string,
			XtNokLabel,			"Ok",
			XtNcancelLabel,		"Cancel",
			XtNapplyLabel,		"Use Regexp",
			XtNpixmap,			NULL,
			XtNprompt,			2,
			XtNdefaultButton,	OK_BUTTON,
			XtNtitle,			title,
			XtNlabel1,			"Search for:",
			XtNlabel2,			"Replace by:",
			NULL);
#if 1
	XtVaSetValues (XtParent (data -> query_string),
			XtNtitle,	title,
			NULL);
#endif
	/*
	 *	Mettre le contenu par defaut.
	 */
	w = QueryGetChild (data -> query_string, EDIT);
	if (old == NULL) old = "";
	WixReplace (w, 0, -1, old, -1);
	SelectBuffer (w);
	w = QueryGetChild (data -> query_string, AUX_EDIT);
	if (new == NULL) new = "";
	WixReplace (w, 0, -1, new, -1);
	SelectBuffer (w);
#if 1
	XtPopup (XtParent (XtParent (data -> query_string)), XtGrabNone);
#else
	XtPopup (XtParent (data -> query_string), XtGrabNone);
#endif
}

/*
 *	Fait apparaitre la boite yes/no.
 */

void popAlert (WixWindow* data, char* title, Boolean info)
{
	Widget	apply, cancel;

	CreateQueryString (data);

	apply = QueryGetChild (data -> query_string, APPLY_BUTTON);
	cancel = QueryGetChild (data -> query_string, CANCEL_BUTTON);

	if (XtIsManaged (apply))		XtUnmanageChild (apply);
	if (info && !XtIsManaged (cancel))	XtManageChild (cancel);
	else if (!info && XtIsManaged (cancel))	XtUnmanageChild (cancel);

	XtVaSetValues (data -> query_string,
		XtNokLabel,			info ? "Yes" : "Done",
		XtNcancelLabel,		"No",
		XtNpixmap,			info ? QuestionPixmap : ErrorPixmap,
		XtNprompt,			0,
		XtNdefaultButton,	info ? CANCEL_BUTTON : OK_BUTTON,
		XtNtitle,			title,
		XtNlabel1,			NULL,
		XtNlabel2,			NULL,
		NULL);

#if 1
	XtPopup (XtParent (XtParent (data -> query_string)), XtGrabNone);
#else
	XtPopup (XtParent (data -> query_string), XtGrabNone);
#endif
}

/*
 *	Fait apparaitre la boite yes/no/maybe
 */

void PopBye (WixWindow* data, char* title, Boolean ok_is_default,
				char* ok_label, char* cancel_label, char* apply_label)
{
	Widget	apply, cancel;

	CreateQueryString (data);

	apply = QueryGetChild (data -> query_string, APPLY_BUTTON);
	cancel = QueryGetChild (data -> query_string, CANCEL_BUTTON);

	XtVaSetValues (apply, XtNtoggle, False, NULL);
	XtManageChild (apply);
	XtManageChild (cancel);

	XtVaSetValues (data -> query_string,
		XtNokLabel,			ok_label,
		XtNcancelLabel,		cancel_label,
		XtNapplyLabel,		apply_label,
		XtNpixmap,			QuestionPixmap,
		XtNprompt,			0,
		XtNdefaultButton,	ok_is_default ? OK_BUTTON : CANCEL_BUTTON,
		XtNtitle,			title,
		XtNlabel1,			NULL,
		XtNlabel2,			NULL,
		NULL);

#if 1
	XtPopup (XtParent (XtParent (data -> query_string)), XtGrabNone);
#else
	XtPopup (XtParent (data -> query_string), XtGrabNone);
#endif
}
