/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/*************************************************************************
*
*	La boite d'aide.
*
*************************************************************************/

#include <stdio.h>
#include <sys/utsname.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/xpm.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/MenuBar.h>
#include <Xpa/ScrolledWin.h>
#include <Xpa/PushButton.h>
#include <Xpa/PanedWin.h>
#include <Xpa/Wix.h>
#include <Xpa/Frame.h>

#include "features.h"
#include "wix.h"
#include "logo.h"
#include "logo_bg.h"

extern Pixmap Jpeg2Pixmap (Screen* screen, char* data, int length, int *pw, int *ph);

extern WixCommandRec	AppActionsList [];
extern int				AppActionsNumber;
extern WixCommandRec	WixActionsList [];
extern int				WixActionsNumber;

static Widget	helpWidget;
static String	helpTab [64];
static String*	menuItems [] = { helpTab, helpTab+50, NULL };
static char		prefix [32];

/*
 *	Retrouve une action a partir de son adresse.
 */

static Boolean get_action_by_addr (XtActionProc proc, char** string, char** help_string)
{
	int	i;
	/*
	 *	Les fonctions de l'application.
	 */
	for (i = 0; i < AppActionsNumber; i++)
		if (AppActionsList [i].proc == proc) {
			*string = AppActionsList [i].string;
			*help_string = AppActionsList [i].help_string;
			return True;
		}
	/*
	 *	les fonctions du widget.
	 */
	for (i = 0; i < WixActionsNumber; i++)
		if (WixActionsList [i].proc == proc) {
			*string = WixActionsList [i].string;
			*help_string = WixActionsList [i].help_string;
			return True;
		}
	/*
	 *	pas trouvee.
	 */
	return False;
}

/*
 *	Fabrique la wall-chart.
 */

static char* pr_key (int c, char* p, Boolean f)
{
	switch (c) {
		case 033:	bcopy ("Esc-", p, 5);
				if (!f) {
					p [3] = 0;
					return p+3;
				}
				return p+4;
		case 0177:	bcopy ("Del", p, 4);	return p+3;
		case '\r':	bcopy ("CR", p, 3);	return p+2;
		case '\t':	bcopy ("Tab", p, 4);	return p+3;
		case K_UP:	bcopy ("UP", p, 3);	return p+2;
		case K_DOWN:	bcopy ("DOWN", p, 5);	return p+4;
		case K_LEFT:	bcopy ("LEFT", p, 5);	return p+4;
		case K_RIGHT:	bcopy ("RIGHT", p, 6);	return p+5;
		case ' ':	bcopy ("Space", p, 6);	return p+5;
		default:
			if (c < ' ') {
				*p++ = '^';
				*p++ = c + '@';
			} else
				*p++ = c;
			*p = 0;
			return p;
	}
}

static void print_alias (Widget	w, alias* ap)
{
	char	tc [1024];

	for (; ap -> a_src; ap++) {
		sprintf (tc, "%-15s %s\n", ap -> a_src, ap -> a_dst);
		WixInsert (w, tc);
	}
}

static void print_translations (Widget w, char* trans)
{
	char	tc [256];
	char*	p;

	while (*trans) {
		while (*trans == '\t' || *trans == ' ') trans++;
		for (p = tc; *trans && (*p++ = *trans++) != '\n';) ;
		*p = 0;
		WixInsert (w, tc);
	}
}

static void print_keytable (Widget w, Trans* r, char* pref)
{
	XtActionProc	prev_proc = NULL;
	int				i;
	char			tc [1024];

	for (i = 0; i < K_NUMBER; i++) {
		if (r [i].prefix) {
			print_keytable (w, r [i].prefix, pr_key (i, pref, True));
			continue;
		}
		if (r [i].proc == insert_char || r [i].proc == prev_proc)
			continue;
		if (r [i].proc) {
			String	name;
			String	doc;

			prev_proc = r [i].proc;
			pr_key (i, pref, False);
			get_action_by_addr (r [i].proc, &name, &doc);
			if (!doc) doc = "No documentation for this command.";
			WixSetFont (w, KEYWORD_COLOR);
			sprintf (tc, "%s\t", prefix);	WixInsert (w, tc);
			WixSetFont (w, COMMENT_COLOR);
			sprintf (tc, "%s\n", name);	WixInsert (w, tc);
			WixSetFont (w, STANDARD_COLOR);
			sprintf (tc, "\t%s\n\n", doc);	WixInsert (w, tc);
		}
	}
}

/*
 *	Afficher ma photo.
 */

static int		me_width, me_height;
static Pixmap	me_bg;
static Pixel	help_bg;
static Pixmap	me_pixmap;
static Boolean	with_me = True;

static void show_me (Widget w, XtPointer client, XEvent* event, Boolean* cont)
{
	Pixmap			me_pixmap = (Pixmap) client;
	XExposeEvent*	e = (XExposeEvent*) event;
	XRectangle		r;
	int				y;
	char**			p;
	FontInfo*		fp;
	FontInfo*		vfp;
	Pixel			pix;
	Pixel			kwpix;
	extern char		__version [];
	char			tc [128];
	static GC		gc = NULL;
	static char*	signature [] = {
			"Laboratoire de Recherche en Informatique",
			"Universit Paris-Sud Orsay.",
			"web: http://www.lri.fr/~pa",
			"email: pa@lri.fr",
			NULL
	};
	static struct utsname	u;

	if (u.sysname [0] == 0)		/* sur quel systme sommes nous */
		uname (&u);

	if (!with_me)
		return;		/* sans moi */

	if (!gc)
		gc = XCreateGC (XtDisplay (w), XtWindow (w), 0, NULL);

	r.x = e -> x;
	r.y = e -> y;
	r.width = e -> width;
	r.height = e -> height;
	XSetClipRectangles (XtDisplay (w), gc, 0, 0, &r, 1, Unsorted);

	XtVaGetValues (w,
		XtNforeground,		&pix,
		XtNkeywordColor,	&kwpix,
		XtNcppFont,			&fp,
		XtNcommentFont,		&vfp,
		NULL);

	if (me_pixmap)
		XCopyArea (XtDisplay (w), me_pixmap, XtWindow (w),
			gc,
			0, 8, me_width, me_height-8,
			2, 4);
	/*
	 *	Afficher le message.
	 */
	XSetForeground (XtDisplay (w), gc, WhitePixelOfScreen (XtScreen (w)));
	XSetFont (XtDisplay (w), gc, fp -> f_info -> fid);
	y = 20 + 300 + fp -> f_info -> ascent;
	for (p = signature; *p; p++, y += fp -> f_height)
		XDrawString (XtDisplay (w), XtWindow (w), gc,
				10, y,
				*p, strlen (*p));
	/*
	 *	Afficher la version
	 */
	XSetForeground (XtDisplay (w), gc, kwpix);
	XSetFont (XtDisplay (w), gc, vfp -> f_info -> fid);
	sprintf (tc, "WiX Version: %s (%s %s)", __version, u.sysname, u.release);
	XDrawString (XtDisplay (w), XtWindow (w), gc,
					20, 20 + vfp -> f_info -> ascent,
					tc, strlen (tc));
}

/*
 *	Affiche la liste des commandes.
 */

static void show_commands (void)
{
	char	tc [1024];
	int		i;
	char*	s;

	WixClearBuffer (helpWidget);
	XtVaSetValues (helpWidget,
		XtNreadOnly,	False,
		XtNwrap,	True,
		NULL);

	WixInsert (helpWidget, "\t\t");
	WixSetFont (helpWidget, CPP_COLOR);
	WixInsert (helpWidget, "WiX Application actions\n");
	WixSetFont (helpWidget, STANDARD_COLOR);
	WixInsert (helpWidget, "\n");
	/*
	 *	Les actions de l'application.
	 */
	for (i = 0; i < AppActionsNumber; i++) {
		WixSetFont (helpWidget, COMMENT_COLOR);
		sprintf (tc, "%s\n", AppActionsList [i].string);
		WixInsert (helpWidget, tc);
		WixSetFont (helpWidget, STANDARD_COLOR);
		s = AppActionsList [i].help_string;
		sprintf (tc, "\t%s\n\n", s ? s : "No info available.");
		WixInsert (helpWidget, tc);
	}
	/*
	 *	les actions du widget.
	 */
	WixInsert (helpWidget, "\t\t");
	WixSetFont (helpWidget, CPP_COLOR);
	WixInsert (helpWidget, "WiX Widget actions\n");
	WixSetFont (helpWidget, STANDARD_COLOR);
	WixInsert (helpWidget, "\n");
	for (i = 0; i < WixActionsNumber; i++) {
		/* ne pas citer les actions sans doc */
		if (!WixActionsList[i].help_string)
			continue;

		WixSetFont (helpWidget, COMMENT_COLOR);
		sprintf (tc, "%s\n", WixActionsList [i].string);
		WixInsert (helpWidget, tc);
		WixSetFont (helpWidget, STANDARD_COLOR);
		sprintf (tc, "\t%s\n\n", WixActionsList [i].help_string);
		WixInsert (helpWidget, tc);
	}

	WixSetFont (helpWidget, STANDARD_COLOR);
	BeginningOfBuffer (helpWidget, 1, NULL);
	XSetWindowBackground (XtDisplay (helpWidget), XtWindow (helpWidget), help_bg);
	XClearWindow (XtDisplay (helpWidget), XtWindow (helpWidget));
	WixRedraw (helpWidget);
	XtVaSetValues (helpWidget,
		XtNreadOnly,	True,
		XtNwrap,	False,
		NULL);
}

/*
 *	Le menu de selection des modes.
 */

static void menu_select (Widget w, XtPointer client, XtPointer call)
{
	Widget		top = (Widget) client;
	int			num = (int) (long) call;
	int			item;
	WxMode*		mp;
	static int	prev_doc = -1;
	static char	msg [32];
	Display*	dpy = XtDisplay (w);
	Window		win = XtWindow (helpWidget);

	item = num & 0xffff;
	num = (num >> 16) & 0xffff;

	SwSetLabel (XtParent (helpWidget), NULL);	/* effacer l'ventuel ancien message */
	if (num == 1)
		switch (item) {
			case 2:
				/*
				 *	Afficher ma photo.
				 */
				XtUnmapWidget (SwHSB (XtParent (helpWidget)));	/* sans scroll horizontal */
				with_me = True;
				prev_doc = -1;
				SwSetTitle (XtParent (helpWidget), NULL);
				WixClearBuffer (helpWidget);
				WixInsert (helpWidget, "\n");
				WixRedraw (helpWidget);
				XSetWindowBackgroundPixmap (dpy, win, me_bg);
				XClearArea (dpy, win, 0, 0, 0, 0, True);
				return;

			case 1:
				/*
				 *	Donner la liste des actions.
				 */
				with_me = False;	/* Retirer ma photo! */
				prev_doc = -2;
				SwSetTitle (XtParent (helpWidget), "WiX actions");
				XtMapWidget (SwHSB (XtParent (helpWidget)));
				show_commands ();

			default:
				return;
		}
	/*
	 *	Fermer la fenetre. (les numeros sont comptes a partir
	 *	de la fin car on ne connait pas a priori le nombre
	 *	de modes)
	 */
	if (helpTab [item+1] == NULL) {
		XtUnmapWidget (top);
		return;
	}
	/*
	 *	Sauve la doc
	 */
	if (helpTab [item+2] == NULL) {
		if (msg [0] == 0)
			strcpy (msg, "Saved as: ~/");

		switch (prev_doc) {
			case -1:
				return;

			case -2:
				strcpy (msg + 12, "commands.doc");
				break;

			default:
				if (0 <= prev_doc && prev_doc < wx_numModes) {
					sprintf (msg + 12, "%s.doc", wx_modes [prev_doc]);
					break;
				}
				return;
		}
		if (WixSaveFile (helpWidget, msg + 12) > 0)
			SwSetLabel (XtParent (helpWidget), msg);
		return;
	}
	/*
	 *	Doc sur un mode particulier.
	 */
	XtMapWidget (SwHSB (XtParent (helpWidget)));
	with_me = False;	/*	Retirer ma photo! */

	WixClearBuffer (helpWidget);
	prev_doc = item-1;
	mp = & wx_modes [item-1];
	SwSetTitle (XtParent (helpWidget), mp -> m_name);
	XtVaSetValues (helpWidget,
		XtNreadOnly,	False,
		XtNwrap,		False,
		NULL);

	WixInsert (helpWidget, "\t\t");
	WixSetFont (helpWidget, CPP_COLOR);
	WixInsert (helpWidget, "Translation Table\n");
	WixSetFont (helpWidget, STANDARD_COLOR);
	WixInsert (helpWidget, "\n");
	if (mp -> m_string_trans)
		print_translations (helpWidget, mp -> m_string_trans);
	else
		WixInsert (helpWidget, "translations table empty");

	WixInsert (helpWidget, "\n\n\t\t");
	WixSetFont (helpWidget, CPP_COLOR);
	WixInsert (helpWidget, "Alias Table ");
	WixSetFont (helpWidget, STANDARD_COLOR);
	WixInsert (helpWidget, "\n\n");

	if (mp -> m_alias)
		print_alias (helpWidget, mp -> m_alias);
	else
		WixInsert (helpWidget, "Alias table empty.\n");

	WixInsert (helpWidget, "\n");
	WixSetFont (helpWidget, CPP_COLOR);
	WixInsert (helpWidget, "\t\tKeymap.\n\n");
	WixSetFont (helpWidget, STANDARD_COLOR);
	XtVaSetValues (helpWidget, XtNwrap, True, NULL);
	print_keytable (helpWidget, mp -> m_keytable, prefix);

	BeginningOfBuffer (helpWidget, 1, NULL);
	XSetWindowBackground (dpy, win, help_bg);
	XClearWindow (dpy, win);
	WixRedraw (helpWidget);
	XtVaSetValues (helpWidget, XtNreadOnly, True, NULL);
}

/*
 *	Cree la boite de help.
 */

void HelpAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	Widget	top, form, menubar;
	Widget	frame;
	int		i, n;
	int		junk;
	Arg		arglist [16];

	if (helpWidget) {
		SwSetLabel (XtParent (helpWidget), NULL);
		XMapRaised (
		    XtDisplay (helpWidget),
		    XtWindow (XtParent (XtParent (XtParent (XtParent (helpWidget))))));
		return;
	}
	/*
	 *	Creer le menu des modes.
	 */
	helpTab [0] = "Modes";
	for (i = 1; i <= wx_numModes; i++)
		helpTab [i] = wx_modes [i-1].m_name;
	helpTab [i++] = "=";	/* un separateur */
	helpTab [i++] = "Save";
	helpTab [i++] = "Close";
	helpTab [i] = NULL;
	/*
	 *	Creer le menu des commandes.
	 */
	helpTab [50] = "General";
	helpTab [51] = "Commands";
	helpTab [52] = "Author";
	helpTab [53] = NULL;
	/*
	 *	Creer le shell.
	 */
	top = XtVaCreatePopupShell ("help", topLevelShellWidgetClass, ghost,
									XtNtitle, "Wix Help",
									XtNiconName, "Wix Help",
									XtNminHeight, 486,
									XtNminWidth, 617,
									XtNmaxHeight, 800,
									NULL);
	/*
	 *	Creer le frame et le form.
	 */
	frame = XtCreateManagedWidget ("help-form", frameWidgetClass, top, NULL, 0);
	form = XtCreateManagedWidget ("help-form", panedWindowWidgetClass, frame, NULL, 0);
	/*
	 *	Creer la barre de menu.
	 */
	menubar = XtVaCreateManagedWidget ("menubar", menuBarWidgetClass, form,
			XtNresizable,	False,
			XtNitem,	menuItems,
			NULL);
	/*
	 *	Me charger.
	 */
	me_pixmap = Jpeg2Pixmap (XtScreen (top),
					logo_jpg, sizeof (logo_jpg),
					&me_width, &me_height);
	me_bg = Jpeg2Pixmap (XtScreen (top),
					logo_bg_jpg, sizeof (logo_bg_jpg),
					&junk, &junk);
	/*
	 *	Creer le wix
	 */
	n = 0;
	XtSetArg (arglist [n], XtNbackgroundPixmap, me_bg); n++;
	XtSetArg (arglist [n], XtNnoCursor, True); n++;
	XtSetArg (arglist [n], XtNtranslations, HelpTrans); n++;
	XtSetArg (arglist [n], XtNshadowThickness, 0); n++;
	helpWidget = XtCreateScrolledWix (form, "help", arglist, n);
	/* rcuprer la couleur du fond */
	XtVaGetValues (helpWidget, XtNbackground, &help_bg, NULL);
	/* le handler pour le logo */
	XtAddEventHandler (helpWidget, ExposureMask, False, show_me, (XtPointer) me_pixmap);
	/*
	 *	Les ressources de contraintes.
	 */
	XtManageChild (XtParent (helpWidget));
	XtAddCallback (menubar, XtNselect, menu_select, top);

	XtSetKeyboardFocus (top, form);
	XtSetKeyboardFocus (form, XtParent (helpWidget));
	XtSetKeyboardFocus (XtParent (helpWidget), helpWidget);

	XtRealizeWidget (top);
	XtUnmapWidget (SwHSB (XtParent (helpWidget)));
	XtMapWidget (top);
}

/*
 *	Fait disparaitre la doc sur le fichier de configuration.
 */

static void unmap (Widget w, XtPointer client, XtPointer call)
{
	XtUnmapWidget ((Widget) client);
}

/*
 *	Fait apparaitre la doc sur le fichier de configuration.
 */

void HelpConfigAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	static Widget	configWidget = NULL;
	Widget			form, dismiss;
	KeyWord*		help_keywords;
	Arg				arglist [16];
	char			keywords [8192];
	int				n;
	extern char		__def_config [];

	if (configWidget) {
		XMapRaised (
		    XtDisplay (configWidget),
		    XtWindow (configWidget));
		return;
	}
	/*
	 *	Creer le shell.
	 */
	configWidget = XtVaCreatePopupShell ("configHelp", topLevelShellWidgetClass, ghost,
						XtNtitle, "WiX Help (Configuration file)",
						XtNiconName, "Wix Help",
						NULL);
	/*
	 *	Creer le form.
	 */
	form = XtCreateManagedWidget ("help-form", panedWindowWidgetClass, configWidget, NULL, 0);
	/*
	 *	Creer le bouton pour sortir.
	 */
	dismiss = XtVaCreateManagedWidget ("dismiss", pushButtonWidgetClass, form,
								XtNtipMessage,	"Close this help box",
								XtNresizable,	False,
								NULL);
	XtAddCallback (dismiss, XtNrelease, unmap, (XtPointer) configWidget);
	/*
	 *	On copie la chaine dfinissant les mots clefs dans une variable locale
	 *	car 'CvtStringToKeyWordTable' remplace les '\n' par des '\0' et les
	 *	chaines constantes son non-modifiables ...
	 */
	strcpy (keywords,
				"KeywordColor	CppColor\n"
				"LibraryColor		CppColor\n"
				"CppColor		CppColor\n"
				"ConstantColor	CppColor\n"
				"ClassColor		CppColor\n"
				"StringColor		CppColor\n"
				"StandardColor	CppColor\n"
				"defmenu	KeywordColor\n"
				"menu			KeywordColor\n"
				"fonts			KeywordColor\n"
				"modes			KeywordColor\n"
				"defmode			KeywordColor\n"
				"include			KeywordColor\n"
				"button:			KeywordColor\n"
				"suffix:			KeywordColor\n"
				"magic:			KeywordColor\n"
				"string_delimitor:	KeywordColor\n"
				"delimitors:		KeywordColor\n"
				"begin_keyword:	KeywordColor\n"
				"end_keyword:		KeywordColor\n"
				"comment_start:	KeywordColor\n"
				"comment_stop:	KeywordColor\n"
				"line_comment:	KeywordColor\n"
				"abbrev:			KeywordColor\n"
				"tag_cmd:			KeywordColor\n"
				"auto_wrap:		KeywordColor\n"
				"errors:			KeywordColor\n"
				"keywords:		KeywordColor\n"
				"print:			KeywordColor\n"
				"compile:			KeywordColor\n"
				"paragraph_delimitor:	KeywordColor\n"
				"background:		KeywordColor\n"
				"translations:			KeywordColor\n"
				"keytable:			KeywordColor\n"
				"standard_font:	KeywordColor\n"
				"standard_color:	KeywordColor\n"
				"keyword_font:	KeywordColor\n"
				"keyword_color:	KeywordColor\n"
				"comment_font:	KeywordColor\n"
				"comment_color:	KeywordColor\n"
				"comment_submode:	KeywordColor\n"
				"constant_font:	KeywordColor\n"
				"constant_color:	KeywordColor\n"
				"class_font:		KeywordColor\n"
				"class_color:		KeywordColor\n"
				"library_font:			KeywordColor\n"
				"library_color:			KeywordColor\n"
				"cpp_font:		KeywordColor\n"
				"cpp_color:		KeywordColor\n"
				"string_font:			KeywordColor\n"
				"string_color:			KeywordColor\n"
				"string_submode:			KeywordColor\n"
				"alias:			KeywordColor\n"
				"auto-wrap-mode	LibraryColor\n"
				"adjust-paragraph	LibraryColor\n"
				"change-mode-behavior	LibraryColor\n"
				"backward-search		LibraryColor\n"
				"fast-backward-search	LibraryColor\n"
				"begin-defun			LibraryColor\n"
				"change-directory		LibraryColor\n"
				"clear-buffer			LibraryColor\n"
				"compile-buffer		LibraryColor\n"
				"create-window		LibraryColor\n"
				"delete-file			LibraryColor\n"
				"delete-window		LibraryColor\n"
				"end-defun			LibraryColor\n"
				"eval-lisp-line		LibraryColor\n"
				"exec-command		LibraryColor\n"
				"exec-named-macro	LibraryColor\n"
				"fast-forward-search	LibraryColor\n"
				"forward-search		LibraryColor\n"
				"incremental-search	LibraryColor\n"
				"indent-lisp-function	LibraryColor\n"
				"indent-lisp-line		LibraryColor\n"
				"insert-file			LibraryColor\n"
				"fast-insert-file		LibraryColor\n"
				"line-counter			LibraryColor\n"
				"lisp-new-line		LibraryColor\n"
				"fast-load-file		LibraryColor\n"
				"load-file			LibraryColor\n"
				"load-lisp-file		LibraryColor\n"
				"load-same-file		LibraryColor\n"
				"next-window		LibraryColor\n"
				"pop-help-box		LibraryColor\n"
				"pop-mode-menu		LibraryColor\n"
				"pop-kill-stack		LibraryColor\n"
				"pop-window-menu		LibraryColor\n"
				"print-buffer			LibraryColor\n"
				"query-replace		LibraryColor\n"
				"quit-the-editor		LibraryColor\n"
				"restart-server		LibraryColor\n"
				"save-file			LibraryColor\n"
				"save-lisp-macro		LibraryColor\n"
				"select-pane			LibraryColor\n"
				"set-compile-command	LibraryColor\n"
				"set-tag-command	LibraryColor\n"
				"set-tab-equiv		LibraryColor\n"
				"show-tags			LibraryColor\n"
				"show-kill-stack		LibraryColor\n"
				"show-line-number	LibraryColor\n"
				"show-ruler			LibraryColor\n"
				"split-window		LibraryColor\n"
				"start-macro-learn	LibraryColor\n"
				"strip-window		LibraryColor\n"
				"stop-macro-learn	LibraryColor\n"
				"toggle-octal-display	LibraryColor\n"
				"toggle-editable		LibraryColor\n"
				"undo				LibraryColor\n"
				"unmodify-buffer		LibraryColor\n"
				"unsplit-window		LibraryColor\n"
				"write-file			LibraryColor\n"
				"wxrc-help			LibraryColor\n"
				"indent-line-c		LibraryColor\n"
				"backward-char		LibraryColor\n"
				"backward-line		LibraryColor\n"
				"backward-page		LibraryColor\n"
				"backward-word		LibraryColor\n"
				"c-beginning-of-function	LibraryColor\n"
				"c-end-of-function	LibraryColor\n"
				"beginning-of-buffer	LibraryColor\n"
				"beginning-of-line	LibraryColor\n"
				"capitalize-word		LibraryColor\n"
				"copy-end-line		LibraryColor\n"
				"copy-region			LibraryColor\n"
				"curline-top			LibraryColor\n"
				"delete-char			LibraryColor\n"
				"delete-end-line		LibraryColor\n"
				"delete-line			LibraryColor\n"
				"delete-word			LibraryColor\n"
				"end-of-buffer		LibraryColor\n"
				"end-of-line			LibraryColor\n"
				"exchange-point-mark	LibraryColor\n"
				"exec-macro			LibraryColor\n"
				"expand-alias		LibraryColor\n"
				"expand-from-alias	LibraryColor\n"
				"extend-adjust		LibraryColor\n"
				"extend-end			LibraryColor\n"
				"flasher				LibraryColor\n"
				"forward-char		LibraryColor\n"
				"forward-line		LibraryColor\n"
				"forward-page		LibraryColor\n"
				"forward-word		LibraryColor\n"
				"goto-line			LibraryColor\n"
				"incremental-yank	LibraryColor\n"
				"insert-accent		LibraryColor\n"
				"insert-char			LibraryColor\n"
				"insert-octal		LibraryColor\n"
				"lower-case-word		LibraryColor\n"
				"match-closing-brace	LibraryColor\n"
				"match-closing-par	LibraryColor\n"
				"match-opening-brace	LibraryColor\n"
				"match-opening-par	LibraryColor\n"
				"multiplier			LibraryColor\n"
				"new-line			LibraryColor\n"
				"new-line-indent		LibraryColor\n"
				"new-line-indent-c	LibraryColor\n"
				"noop				LibraryColor\n"
				"open-line			LibraryColor\n"
				"open-space			LibraryColor\n"
				"repeat-command		LibraryColor\n"
				"replace-string		LibraryColor\n"
				"rubout-char			LibraryColor\n"
				"rubout-word			LibraryColor\n"
				"scroll-down			LibraryColor\n"
				"scroll-left			LibraryColor\n"
				"scroll-right		LibraryColor\n"
				"scroll-up			LibraryColor\n"
				"scroll-window		LibraryColor\n"
				"select-line			LibraryColor\n"
				"select-word			LibraryColor\n"
				"selection-start		LibraryColor\n"
				"set-mark			LibraryColor\n"
				"stuff				LibraryColor\n"
				"show-line-number	LibraryColor\n"
				"switch-font			LibraryColor\n"
				"tabulation			LibraryColor\n"
				"twiddle-chars		LibraryColor\n"
				"upper-case-word		LibraryColor\n"
				"wipe-region			LibraryColor\n"
				"yank				LibraryColor\n");

	help_keywords = CvtStringToKeyWordTable (keywords, " \t\n\\(");
	/*
	 *	Creer le wix
	 */
	n = 0;
	XtSetArg (arglist [n], XtNnoCursor, True); n++;
	XtSetArg (arglist [n], XtNlineComment, "# "); n++;
	XtSetArg (arglist [n], XtNlineNum, 24); n++;
	XtSetArg (arglist [n], XtNcolumnNumber, 80); n++;
	XtSetArg (arglist [n], XtNtranslations, HelpTrans); n++;
	XtSetArg (arglist [n], XtNkeywordTable, help_keywords); n++;
	XtSetArg (arglist [n], XtNdelimitors, " \t\n\\("); n++;
	w = XtCreateScrolledWix (form, "confWix", arglist, n);
	WixInsert (w, __def_config);
	BeginningOfBuffer (w, 1, NULL);
	WixRedraw (w);
	/*
	 *	Les ressources de contraintes.
	 */
	XtVaSetValues (XtParent (w),
			XtNlabel,		"Configuration file (~/.wxrc)",
			NULL);
	XtManageChild (XtParent (w));

	XtSetKeyboardFocus (configWidget, form);
	XtSetKeyboardFocus (form, XtParent (w));
	XtSetKeyboardFocus (XtParent (w), w);

	XtRealizeWidget (configWidget);
	XtMapWidget (configWidget);
}

/*
 *	Fait apparaitre la doc sur les nouveauts
 */

void HelpNewFeaturesAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	static Widget	featuresWidget = NULL;
	Widget			pane;
	KeyWord*		help_keywords;
	Arg				arglist [16];
	char			keywords [2048];
	int				n;

	if (featuresWidget) {
		XMapRaised (XtDisplay (featuresWidget), XtWindow (featuresWidget));
		return;
	}
	/*
	 *	Crer le shell.
	 */
	featuresWidget = XtVaCreatePopupShell ("configHelp", topLevelShellWidgetClass, ghost,
											XtNtitle, 		"New Features",
											XtNiconName,	"New Features",
											NULL);
	/*
	 *	Crer le pane.
	 */
	pane = XtCreateManagedWidget ("help-form", panedWindowWidgetClass, featuresWidget, NULL, 0);
	/*
	 *	Crer le bouton pour sortir.
	 */
	XtAddCallback (
		XtVaCreateManagedWidget ("dismiss", pushButtonWidgetClass, pane,
										XtNtipMessage,	"Close this help box",
										XtNresizable,	False,
										NULL),
		XtNrelease, unmap, (XtPointer) featuresWidget);
	/*
	 *	On copie la chane dfinissant les mots clefs dans une variable locale
	 *	car 'CvtStringToKeyWordTable' remplace les '\n' par des '\0' et les
	 *	chanes constantes son non-modifiables ...
	 */
	strcpy (keywords,
				"pop-mode-menu			CppColor\n"
				"pop-window-menu		CppColor\n"
				"quit-the-editor			CppColor\n"
				"save-file					CppColor\n"
				"show-tags				CppColor\n"
				"expand-alias				CppColor\n"
				"expand-from-alias		CppColor\n"
				"delete-beginning-of-line		CppColor\n"
				"incremental-search		CppColor\n"
				"fast-backward-search		CppColor\n"
				"fast-forward-search		CppColor\n"
				"fast-insert-file			CppColor\n"
				"fast-load-file				CppColor\n"
				"tabulation				CppColor\n"
				"toggle-octal-display		CppColor\n"
				"insert-accent				CppColor\n"
				"replace-string			CppColor\n"
				"change-mode-behavior	CppColor\n"
				"c-beginning-of-function	CppColor\n"
				"c-end-of-function			CppColor\n"
				"center-curline			CppColor\n"
				"adjust-paragraph			CppColor\n"
				"include					CppColor\n"
				"begin_keyword			KeywordColor\n"
				"end_keyword				KeywordColor\n"
				"line_comment			KeywordColor\n"
				"paragraph_delimitor		KeywordColor\n"
				"keyword_font				KeywordColor\n"
				"keyword_color			KeywordColor\n"
				"cpp_font					KeywordColor\n"
				"cpp_color				KeywordColor\n"
				"constant_font				KeywordColor\n"
				"constant_color			KeywordColor\n"
				"class_font				KeywordColor\n"
				"class_color				KeywordColor\n"
				"library_font				KeywordColor\n"
				"library_color				KeywordColor\n"
				"string_font				KeywordColor\n"
				"string_color				KeywordColor\n"
				"startup					KeywordColor\n"
				"submode				KeywordColor\n"
				"printf					LibraryColor\n"
				"NULL					LibraryColor\n"
				"#include					CppColor\n"
				"New					LibraryColor\n"
				"Features					LibraryColor\n"
				"Exit						LibraryColor\n"
				"Misc					LibraryColor\n"
				"Meta					LibraryColor\n"
				"Escape					LibraryColor\n"
				"Shift					LibraryColor\n"
				"Ctrl						LibraryColor\n"
				"Server					LibraryColor\n"
				"inside					LibraryColor\n"
				"outside					LibraryColor\n"
				"WiX					CommentColor\n");

	help_keywords = CvtStringToKeyWordTable (keywords, " \t\"\n\\<>(~");
	/*
	 *	Crer le wix
	 */
	n = 0;
	XtSetArg (arglist [n], XtNnoCursor, True); n++;
	XtSetArg (arglist [n], XtNlineComment, "Version"); n++;
	XtSetArg (arglist [n], XtNdelimitors, " \t\"\n\\<>(~"); n++;
	XtSetArg (arglist [n], XtNlineNum, 24); n++;
	XtSetArg (arglist [n], XtNcolumnNumber, 84); n++;
	XtSetArg (arglist [n], XtNtranslations, HelpTrans); n++;
	XtSetArg (arglist [n], XtNkeywordTable, help_keywords); n++;
	w = XtCreateScrolledWix (pane, "changesWix", arglist, n);
	WixInsert (w, __new_features);
	BeginningOfBuffer (w, 1, NULL);
	WixRedraw (w);
	XtVaSetValues (XtParent (w),
					XtNlabel,	"Changes since previous versions",
					NULL);
	XtManageChild (XtParent (w));

	XtSetKeyboardFocus (featuresWidget, pane);
	XtSetKeyboardFocus (pane, XtParent (w));
	XtSetKeyboardFocus (XtParent (w), w);

	XtRealizeWidget (featuresWidget);
	XtMapWidget (featuresWidget);
}
