#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <X11/Intrinsic.h>
#include <jpeglib.h>
#include <jerror.h>

/*
 * jmemsrc.c
 *
 *	Copyright Patrick Amar 1998.
 */

/* this is not a core library module, so it doesn't define JPEG_INTERNALS */

/* Expanded data source object for stdio input */

typedef struct {
	struct jpeg_source_mgr
			pub;		/* public fields */
	JOCTET*	buffer;		/* start of buffer */
	int		length;		/* buffer length */
} my_source_mgr;

typedef my_source_mgr * my_src_ptr;

#define INPUT_BUF_SIZE  4096	/* choose an efficiently fread'able size */


/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

METHODDEF(void)
init_source (j_decompress_ptr cinfo)
{
	/* rien  faire */
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 */

METHODDEF(boolean)
fill_input_buffer (j_decompress_ptr cinfo)
{
  my_src_ptr	src = (my_src_ptr) cinfo->src;

  src->pub.next_input_byte = src -> buffer;
  src->pub.bytes_in_buffer = src -> length;

  return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

METHODDEF(void)
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
	my_src_ptr	src = (my_src_ptr) cinfo->src;

	if (num_bytes > 0) {
		src->pub.next_input_byte += (size_t) num_bytes;
		src->pub.bytes_in_buffer -= (size_t) num_bytes;
	}
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */


/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF(void)
term_source (j_decompress_ptr cinfo)
{
  /* no work necessary here */
}


/*
 * Prepare for input from a stdio stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing decompression.
 */

GLOBAL(void)
jpeg_mem_src (j_decompress_ptr cinfo, char* data, int length)
{
	my_src_ptr	src;

	if (cinfo->src == NULL) {	/* first time for this JPEG object? */
		cinfo->src = (struct jpeg_source_mgr *)
			(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
											sizeof(my_source_mgr));
		src = (my_src_ptr) cinfo->src;
		src->buffer = (JOCTET *) data;
	}

	src = (my_src_ptr) cinfo->src;
	src->pub.init_source = init_source;
	src->pub.fill_input_buffer = fill_input_buffer;
	src->pub.skip_input_data = skip_input_data;
	src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
	src->pub.term_source = term_source;
	src->buffer = (JOCTET*) data;
	src->length = length;
	src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
	src->pub.next_input_byte = NULL; /* until buffer loaded */
}


typedef struct my_error_mgr {
	struct jpeg_error_mgr pub;	/* "public" fields */
	jmp_buf setjmp_buffer;		/* for return to caller */
} my_error_mgr;

/*
 * Here's the routine that will replace the standard error_exit method:
 */

static void my_error_exit (j_common_ptr cinfo)
{
	/* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
	my_error_mgr* myerr = (my_error_mgr*) cinfo->err;
	 (*cinfo->err->output_message) (cinfo);
	longjmp (myerr->setjmp_buffer, 1);
}

/*
 * Sample routine for JPEG decompression.  We assume that the source file name
 * is passed in.  We want to return 1 on success, 0 on error.
 */

/* Le cube de couleurs */

#define	nred	 (guess[theCube].r)
#define	ngreen	 (guess[theCube].g)
#define	nblue	 (guess[theCube].b)

static struct cube {
	int	r, g, b;
} guess [] = {
	{ 4, 8, 4 },
	{ 4, 7, 4 },
	{ 4, 9, 3 },
	{ 4, 8, 3 },
	{ 2, 2, 2 },
};

static unsigned char	reds [256],
						greens [256],
						blues [256];
static unsigned char*	colormap [3] = { reds, greens, blues };
static unsigned long	c_index [256];

static int		theCube = -1;
Screen*			TheScreen;	/* la structure screen qui nous permet de tout savoir */

/*
 *	Initialiser le cube de couleurs (appellee qu'en cas de visuel PseudoColor).
 */

static void initColorCube (void)
{
	XColor		xc;
	XColor		colors [256];
	Colormap	cmap = DefaultColormapOfScreen (TheScreen);
	Visual*		visual = DefaultVisualOfScreen (TheScreen);
	Display*	dpy = DisplayOfScreen (TheScreen);
	int			r,g,b;
	int			i;
	double a;

	if (theCube >= 0)
		return;		/* deja fait */

	for (theCube = 0; theCube < XtNumber (guess) - 1; theCube++)
		if (XAllocColorCells (dpy, cmap, False, NULL, 0, c_index, nred*ngreen*nblue))
		{
			i = 0;
			for (r = 0; r < nred; r++)
				for (g = 0; g < ngreen; g++)
					for (b = 0; b < nblue; b++) {
						colors [i].red = r * (65535/ (nred-1));
						colors [i].green = g * (65535/ (ngreen-1));
						colors [i].blue = b * (65535/ (nblue-1));
						colors [i].pixel = c_index [i];
						colors [i].flags = DoRed|DoGreen|DoBlue;
						/* les couleurs sur 8 bits */
						reds [i] = colors [i].red /256;
						greens [i] = colors [i].green /256;
						blues [i] = colors [i].blue /256;
						i++;
					}
			XStoreColors (dpy, cmap, colors, nred*ngreen*nblue);
			return;
		}
	/*
	 *	Prendre un cube read-only
	 */
	xc.flags = DoRed|DoGreen|DoBlue;
	for (; theCube < XtNumber (guess);) {
		i = 0;
		for (r = 0; r < nred; r++)
			for (g = 0; g < ngreen; g++)
				for (b = 0; b < nblue; b++) {
					xc.red = r * (65535/ (nred-1));
					xc.green = g * (65535/ (ngreen-1));
					xc.blue = b * (65535/ (nblue-1));
					if (!XAllocColor (dpy, cmap, &xc)) {
						XFreeColors (dpy, cmap, c_index, i, 0);
						goto retry;
					}
					c_index [i] = xc.pixel;
					reds [i] = xc.red /256;
					greens [i] = xc.green /256;
					blues [i] = xc.blue /256;
					i++;
				}
		return;
		retry:
			theCube++;
	}
}

static unsigned char* read_JPEG_data (char* data, int length, Boolean do_quantize, int* pw, int* ph)
{
	unsigned char*					rgbpic;
	struct jpeg_decompress_struct	cinfo;
	struct my_error_mgr				jerr;
	JSAMPROW						rowptr [1];
	int								width;
	int								height;
	int								bperpix;

	/* Step 1: allocate and initialize JPEG decompression object */
	cinfo.err = jpeg_std_error (&jerr.pub);
	jerr.pub.error_exit = my_error_exit;
	if (setjmp (jerr.setjmp_buffer)) {
		jpeg_destroy_decompress (&cinfo);
		return (unsigned char*) 0;
	}
	/* Now we can initialize the JPEG decompression object. */
	jpeg_create_decompress (&cinfo);
	/* Step 2: specify data source (eg, a memory block) */
	jpeg_mem_src (&cinfo, data, length);
	/* Step 3: read file parameters with jpeg_read_header () */
	 (void) jpeg_read_header (&cinfo, TRUE);
	/* Step 4: set parameters for decompression */
	*pw = width = cinfo.image_width;
	*ph = height = cinfo.image_height;
	if (do_quantize) {
		cinfo.actual_number_of_colors = nred*ngreen*nblue;
		cinfo.colormap = colormap;
		cinfo.quantize_colors = TRUE;
		cinfo.dct_method = JDCT_FLOAT;
	}
	/* Step 5: Start decompressor */
	 (void) jpeg_start_decompress (&cinfo);

	bperpix = cinfo.output_components;
	/* allocate rgb matrix */
	rgbpic = (unsigned char*) XtMalloc (width * height * bperpix);
	/* Step 6:	while (scan lines remain to be read) */
	/*			jpeg_read_scanlines (...); */
	while (cinfo.output_scanline < cinfo.output_height) {
		rowptr[0] = (JSAMPROW) &rgbpic[cinfo.output_scanline * width * bperpix];
		 (void) jpeg_read_scanlines (&cinfo, rowptr, (JDIMENSION) 1);
	}
	/* Step 7: Finish decompression */
	 (void) jpeg_finish_decompress (&cinfo);
	/* Step 8: Release JPEG decompression object */
	jpeg_destroy_decompress (&cinfo);
	return rgbpic;
}

static int highbit (unsigned long ul)
{
	/* returns position of highest set bit in 'ul' as an integer (0-31), or -1 if none */
	int				i;
	unsigned long	hb;

	hb = 0x80;  hb = hb << 24;   /* hb = 0x80000000UL */
	for (i=31; ((ul & hb) == 0) && i>=0;  i--, ul<<=1)
		;
	return i;
}

/*
 *	Fait une XImage a partir d'une matrice RGB.
 */

XImage* rgbpic2XImage (XImage* xim, unsigned char* rgbpic, int width, int height)
{
	Visual*			visual = DefaultVisualOfScreen (TheScreen);
	Display*		dpy = DisplayOfScreen (TheScreen);
	int				depth = DefaultDepthOfScreen (TheScreen);
	int				i,j;
	unsigned char	*lip, *ip, *pp;
	unsigned long	r, g, b, rmask, gmask, bmask, xcol;
	int				rshift, gshift, bshift, bperpix, bperline, border, cshift;

	/*
	 * ColorMapped Visual:  PseudoColor
	 */
	if (visual->class == PseudoColor || visual -> class == GrayScale) {
		int	imWIDE;
		int	nullCount;

		nullCount = (4 - (width % 4)) & 0x03;  /* # of padding bytes per line */
	    imWIDE = width + nullCount;

		if (xim == (XImage*) 0) {
			xim = XCreateImage (dpy, visual, depth, ZPixmap, 0,
								(char*)0,  width,  height, 32, imWIDE);
			xim -> data = XtMalloc (imWIDE * height);
		}
		/* Now create the image data - pad each scanline as necessary */
		pp = rgbpic;
		ip = (unsigned char*) xim -> data;
		for (i = 0, pp = rgbpic; i < height; i++) {
			for (j = 0; j < width; j++)
				*ip++ = c_index [*pp++];
			for (j = 0; j < nullCount; j++)
				*ip++ = 0;
		}
		return xim;
	}
	if (visual->class != TrueColor)
		return (XImage*) 0;
	/*
	 * Non-ColorMapped Visual:  TrueColor
	 */
	/* compute various shifting constants that we'll need... */

    rmask = visual->red_mask;
    gmask = visual->green_mask;
    bmask = visual->blue_mask;

    rshift = 7 - highbit (rmask);
    gshift = 7 - highbit (gmask);
    bshift = 7 - highbit (bmask);
    cshift = 7 - highbit ((unsigned long)255);

	if (xim == (XImage*) 0) {
	    xim = XCreateImage (dpy, visual, depth, ZPixmap, 0, NULL,
							width,  height, 32, 0);
		xim -> data = XtMalloc ((size_t) (height * xim->bytes_per_line));
	}
    bperline = xim->bytes_per_line;
    bperpix  = xim->bits_per_pixel;
    border   = xim->byte_order;

    lip = (unsigned char*) xim -> data;
	pp = rgbpic;
	for (i=0; i<height; i++, lip+=bperline) {
		for (j=0, ip=lip; j<width; j++) {
			r = *pp++;  g = *pp++;  b = *pp++;
			/* shift the bits around */
			if (rshift<0)	r = r << (-rshift);
			else			r = r >> rshift;
	
			if (gshift<0)	g = g << (-gshift);
			else			g = g >> gshift;
	
			if (bshift<0)	b = b << (-bshift);
			else			b = b >> bshift;
			xcol = (r & rmask) | (g & gmask) | (b & bmask);

			switch (bperpix) {
			case 32:
				if (border == MSBFirst) {
					*ip++ = (xcol>>24) & 0xff;
					*ip++ = (xcol>>16) & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ =  xcol      & 0xff;
				} else {	/* LSBFirst */
					*ip++ =  xcol      & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ = (xcol>>16) & 0xff;
					*ip++ = (xcol>>24) & 0xff;
				}
				break;

			case 24:
				if (border == MSBFirst) {
					*ip++ = (xcol>>16) & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ =  xcol      & 0xff;
				} else {	/* LSBFirst */
					*ip++ =  xcol      & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ = (xcol>>16) & 0xff;
				}
				break;

			case 16:
				if (border == MSBFirst) {
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ =  xcol      & 0xff;
				} else {	/* LSBFirst */
					*ip++ =  xcol      & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
				}
				break;
			}
		}
	}
	return xim;
}

Pixmap Jpeg2Pixmap (Screen* screen, char* data, int length, int *pw, int *ph)
{
	Visual*			visual;
	Boolean			do_quantize;
	unsigned char*	rgbpic;
	Pixmap			pixmap;
	Display*		dpy;
	XImage*			xim = NULL;
	int				width, height;

	TheScreen = screen;
	dpy = DisplayOfScreen (TheScreen);
	visual = DefaultVisualOfScreen (TheScreen);
	do_quantize = (visual -> class == PseudoColor);
	/* Initialiser le cube de couleur */
	if (do_quantize && theCube == -1)
		initColorCube ();
	/* lire les donnes */
	rgbpic = read_JPEG_data (data, length, do_quantize, &width, &height);
	if (!rgbpic)
		return (Pixmap) 0;
	/* Faire une XImage */
	xim = rgbpic2XImage (xim, rgbpic, width, height);
	XtFree ((char*) rgbpic);
	pixmap = XCreatePixmap (dpy, RootWindowOfScreen (TheScreen),
							width, height, DefaultDepthOfScreen (TheScreen));
	XPutImage (dpy, pixmap, DefaultGCOfScreen (TheScreen), xim,
				0, 0, 0, 0, width, height);
	XFlush (dpy);
	XtFree (xim -> data);
	*pw = width;
	*ph = height;
	return pixmap;
}
