/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include "lisp.h"

#include <string.h>

#ifndef bcopy
#	define index(s,c) strchr((s),(c))
#	define rindex(s,c) strrchr((s),(c))
#	define bcopy(s,d,n) memcpy((d),(s),(n))
#	define bcmp(a,b,n) memcmp((a),(b),(n))
#	define bzero(a,n) memset((a),0,(n))
#endif

#if !STAND_ALONE
/*
 *	Import de Wix.h
 */

#define HILITED	16
#define STANDARD_COLOR		0
#define COMMENT_COLOR		1
#define CPP_COLOR			2
#define KEYWORD_COLOR		3
#define LIBRARY_COLOR		4
#define STRING_COLOR		5
#define CurrentCharType(w) ((int)((*_GetAttCursorEnd(w)) & ~HILITED))

static lisp	sym_wix;

extern char* XtName (void*);

extern char*	pa_re_exec ();
extern char*	pa_re_comp ();
extern int	re_backward_search ();
extern int	re_forward_search ();
extern int	CBeginningOfFunction ();
extern int	CEndOfFunction ();
extern int	WixForwardSearch ();
extern int	WixBackwardSearch ();
extern int	DeleteChar ();
extern int	WixInsert ();
extern int	ForwardLine ();
extern int	BackwardLine ();
extern int	ForwardChar ();
extern int	BackwardWord ();
extern int	ForwardWord ();
extern int	BackwardChar ();
extern int	BeginningOfLine ();
extern int	EndOfLine ();
extern int	BeginningOfBuffer ();
extern int	EndOfBuffer ();
extern void	Message ();
extern int	WixGetColumn ();
extern int	WixGetWidth ();
extern char*	_GetCursorEnd ();
extern char*	_GetBufferEnd ();
extern char*	_GetAttCursorEnd ();
extern int	EnableRedisplay ();
extern int	WixSetMark ();
extern int	WixGotoMark ();
extern int	WixDeleteMark ();
extern int	WixCreateMark ();
extern int	find_open ();
extern int	find_close ();
extern int 	DeleteEndLine ();
#endif

/*
 *	Les fichiers.
 */
static FILE*	_files [NOFILE];
/*
 *	Les cellules de liste.
 */
static cell*	free_list;
static char*	_mark;
static cell*	_heap;
static int	heap_size = HEAP_SIZE;
/*
 *	L'oblist.
 */
static sym*	_obarray;
static int	obarray_size = OBARRAY_SIZE;
/*
 *	Les chaines de caracteres.
 */
static char*	_string;
static int	free_string = 0;
static int	string_size = STRING_SIZE;
/*
 *	Le reste.
 */
static stack_frame*	frame_p = NULL;
static lisp	sym_tag_error;
static lisp	sym_quote;
static lisp	sym_eof;
static lisp*	stack_org;
static int	hnil;
/*
 *	Les messages d'erreur.
 */
#if !STAND_ALONE
static lisp	widget_expected;
#endif
static lisp	string_expected;
static lisp	number_expected;
static lisp	file_expected;
static lisp	file_closed;
static lisp	list_expected;
static lisp	symbol_expected;
static lisp	out_of_cons;
static lisp	dot_error;
static lisp	bad_arglist;
static lisp	unbound_symbol;
static lisp	undefined_function;
static lisp	no_catch;
static lisp	pname_too_long;
static lisp	missing_quote;
static lisp	out_of_symbol;

/*
 *	Les autres symboles
 */

static lisp	internal_comma_atsign;
static lisp	internal_comma_dot;
static lisp	internal_comma;
static lisp	internal_backquote;

static void	init_symbols ();
static void	error ();
static lisp	throw ();
static lisp	rfirst ();

/*
 *	Ouvre un fichier.
 */

static lisp file_open (lisp name, lisp lmode)
{
	FILE*	fp;
	char*	mode = (char*) lmode;

	if (!stringp (name))
		error (string_expected, name);

	if (!stringp (mode))
		error (string_expected, mode);

	if (*mode != 'r' && *mode != 'w')
		return NIL;

	fp = fopen (name, mode);
	if (!fp)
		return NIL;
	_files [fileno (fp)] = fp;
	return (lisp) &_files [fileno (fp)];
}

/*
 *	Ferme un fichier.
 */

static lisp file_close (lisp f)
{
	int	i;
	FILE**	fpp = (FILE**) f;

	if (!filep (f)) {
		error (file_expected, f);
		return NIL;
	}
	for (i = 0; i < NOFILE; i++)
		if (*fpp == _files [i]) {
			fclose (*fpp);
			_files [i] = NULL;
			return NIL;
		}
	error (file_closed, f);
	return NIL;
}

/*
 *	Cree une string lisp a partir d'une string C.
 */

static lisp make_string (char* s)
{
	int	i;

	for (i = free_string; (_string [free_string++] = *s++) != 0;)
		;
	return (lisp)& _string [i];
}

/*
 *	Fonction de hash-code pour les symboles.
 */

static int hash (char* name)
{
	int	c;
	int	h;

	for (h = 0; (c = *name++) != 0;)
		h = ((h * 17) ^ c) % obarray_size;
	return h;
}

/*
 *	Alloue la memoire pour les objets.
 *	et calcule le hashcode de nil.
 */

static void alloc_space (void)
{
	_heap = (cell*) malloc (heap_size * sizeof (cell));
	_obarray = (sym*) malloc (obarray_size * sizeof (sym));
	/* Zroifier le tableau (on le suppose dans 'intern') */
	bzero ((char *) _obarray, obarray_size * sizeof (sym));
	_string = (char*) malloc (string_size);
	hnil = hash ("nil");
}

/*
 *	Initialise la liste des doublets libres.
 */

static void init_cons (void)
{
	int	i;

	free_list = NIL;
	for (i = heap_size; --i >= 0;) {
		CDR (&_heap[i]) = (lisp) free_list;
		free_list = &_heap[i];
	}
}

/*
 *	Balaye le tas pour recuperer les morceaux.
 */

static int sweep (void)
{
	int	i;
	int	n = 0;

	free_list = NIL;
	for (i = heap_size; --i >= 0;) {
		if (_mark [i])
			continue;

		CDR (&_heap[i]) = (lisp) free_list;
		free_list = &_heap[i];
		n++;
	}
	return n;
}

/*
 *	Marque une liste.
 */

static void mark_list (cell* l)
{
	while (cellp (l)) {
		if (_mark [l - _heap])
			return;

		_mark [l - _heap] = 1;
		mark_list (CAR (l));
		l = (cell*) CDR (l);
	}
}

/*
 *	Marque la liste des objets.
 */

static void mark_sym (void)
{
	int	i;

	for (i = 0; i < obarray_size; i++)
		if (_obarray [i].pname)
			mark_list (_obarray [i].value);
}

/*
 *	Marque la pile.
 */

static void mark_stack (void)
{
	lisp	stack_end;
	lisp*	p;

	for (p = &stack_end; p < stack_org; p++)
		mark_list (*p);
}

/*
 *	Le garbage collector.
 */

static int gc (lisp a, lisp d)
{
	static int	nbgc = 0;
	int		n;

	_mark = (char*) malloc (heap_size);
	bzero (_mark, heap_size);

	mark_list (a);	/* sauver le car */
	mark_list (d);	/* et le cdr du doublet */
	mark_sym ();	/* marquer la cval */
	mark_stack ();	/* ne pas oublier la pile */

	n = sweep ();	/* recuperer les petits */
	free (_mark);
	nbgc ++;	/* un gc de plus */
	if (n == 0)
		error (out_of_cons, UNDEF);

	return n;
}

static lisp ugc (void)
{
	int	n;

	n = gc (NIL, NIL);
	return (lisp) numify (n);
}

/*
 *	Alloue un nouveau doublet.
 */

static lisp cons (lisp a, lisp d)
{
	register cell*	fr;

	if (free_list == NIL)
		gc (a, d);

	fr = free_list;
	free_list = (cell*) CDR (fr);
	CAR (fr) = a;
	CDR (fr) = d;
	return (lisp) fr;
}

/*
 *	Imprime un objet.
 */

static char*	__outp;

static void my_printf (FILE* fp, char* fmt, void* a1, void* a2, void* a3)
{
	if (fp) {
		fprintf (fp, fmt, a1, a2, a3);
		return;
	}
	sprintf (__outp, fmt, a1, a2, a3);
	__outp += strlen (__outp);
}

static lisp probj (lisp l, lisp file)
{
	FILE*	fp = (file != NIL) ? *(FILE**)file : NULL;

	if (l == NIL)
		my_printf (fp, "()", NULL, NULL, NULL);
	else if (l == TRUE)
		my_printf (fp, "t", NULL, NULL, NULL);
	else if (symbolp (l))
		my_printf (fp, "%s", PNAME (l), NULL, NULL);
	else if (numbp (l))
		my_printf (fp, "%d", (void*) denum (l), NULL, NULL);
	else if (stringp (l))
		my_printf (fp, "\"%s\"", (char*) l, NULL, NULL);
	else if (filep (l)) {
		if (*(FILE**)l)
			my_printf (fp, "<FILE>%d", (void*) (long) fileno (*(FILE**)l), NULL, NULL);
		else
			my_printf (fp, "<FILE_CLOSED>", NULL, NULL, NULL);
	} else if (cellp (l)) {
		switch ((long)CAR (l)) {
		case SUBR0:
		case SUBR1:
		case SUBR2:
		case SUBR3:
		case FSUBR:
			my_printf (fp, "<SUBR:0x%x>", CDR (l), NULL, NULL);
			break;
		case LEXPR:
			my_printf (fp, "<LEXPR>", NULL, NULL, NULL);
			goto def;
		case LAMBDA:
			my_printf (fp, "<LAMBDA>", NULL, NULL, NULL);
			goto def;
		case MACRO:
			my_printf (fp, "<MACRO>", NULL, NULL, NULL);
		def:
			l = CDR (l);
		default:
			my_printf (fp, "(", NULL, NULL, NULL);
			probj (CAR (l), file);
			l = CDR (l);
			while (cellp (l)) {
				my_printf (fp, " ", NULL, NULL, NULL);
				probj (CAR (l), file);
				l = CDR (l);
			}
			if (l != NIL) {
				my_printf (fp, " . ", NULL, NULL, NULL);
				probj (l, file);
			}
			my_printf (fp, ")", NULL, NULL, NULL);
		}
	}
#if !STAND_ALONE
	else if ((((unsigned long)l) & 3) == 0 && *(char**)l == (char*)l)
		my_printf (fp, "<Widget>%s", XtName (l), NULL, NULL);
#endif
	else
		my_printf (fp, "<Unknown>%x", l, NULL, NULL);
	return NIL;
}

/*
 *	fonction externe pour imprimer dans une chaine C.
 */

void lisp_print (lisp exp, char* p)
{
	__outp = p;
	probj (exp, NIL);
}

/*
 *	Vide les buffers du fichier 'file'
 *	et force un LF.
 */

static lisp terpri (lisp file)
{
	FILE*	fp = (file != NIL) ? *(FILE**)file : stdout;

	fprintf (fp, "\n");
	fflush (fp);
	return NIL;
}

/*
 *	La fonction d'erreur.
 */

static void error (lisp msg, lisp l)
{
	if (l != (lisp)UNDEF)
		msg = cons (msg, l);
	(void) throw (sym_tag_error, msg);
}

/*
 *	Interne un symbole.
 */

static lisp intern (char* name, lisp value)
{
	int	h = hash (name);
	int	i;

	if (h == hnil)
		return NIL;

	for (i = obarray_size; --i >= 0;) {
		if (_obarray [h].pname == NULL) {
			_obarray [h].pname = name;
			_obarray [h].value = value;
			return (lisp)&_obarray[h];
		}
		if (!strcmp (_obarray [h].pname, name))
			return (lisp)&_obarray[h];
		h += 41 % obarray_size;
	}
	error (out_of_symbol, UNDEF);
	return NIL;
}

/*
 *	Masque un symbole.
 */

static lisp remove_symbol (lisp l)
{
	for (; cellp (l); l = CDR (l)) {
		if (!symbolp (CAR (l)))
			error (symbol_expected, CAR (l));
		PNAME (CAR (l)) = "nil";	/* inatteignable */
	}
	return NIL;
}

/*
 *	Retourne la liste des symboles.
 */

static lisp oblist (void)
{
	int		i;
	lisp	l = NIL;

	for (i = 0; i < obarray_size; i++)
		if (_obarray [i].pname)
			l = cons (&_obarray [i], l);
	return l;
}

/*
 *	Analyseur lexical.
 */

static char*	__p = "";

#define get_char(fp) ((fp) ? getc(fp) : (*__p == 0 ? EOF : *__p++))
#define unget_char(c,fp) ((fp) ? ungetc(c,fp) : (*--__p = c))

static lisp lex (lisp file, lisp eof)
{
	int	i;
	int	c;
	int	nm;
	FILE*	fp = (file == NIL) ? NULL : *(FILE**)file;

loop:
	switch (c = get_char (fp)) {
	case ';':
		while ('\n' != get_char (fp)) ;
	case ' ':
	case '\n':
	case '\t':
		goto loop;

	case EOF:
		return eof;

	case '(':
	case ')':
	case '\'':
	case '.':
		return (lisp) (long) c;

	case '`':
		return cons (internal_backquote, rfirst (file, eof));

	case ',':
		c = get_char (fp);
		switch (c) {
		case '@':
			return cons (internal_comma_atsign, cons (rfirst (file, eof), NIL));
		case '.':
			return cons (internal_comma_dot, cons (rfirst (file, eof), NIL));
		}
		unget_char (c, fp);
		file = cons (internal_comma, cons (rfirst (file, eof), NIL));
		return file;

	case '\"':
		for (i = free_string; (_string[i] = get_char (fp)) != '\"'; i++) {
			if (_string[i] == '\\' && get_char (fp) == 'n')
				_string [i] = '\n';
			if (i >= string_size)
				error (missing_quote, UNDEF);
		}
		_string [i] = 0;
		c = free_string;
		free_string = i+1;
		return &_string [c];

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
		nm = 1;
		goto def;

	case '+': case '-':
		i = get_char (fp); unget_char (i, fp);
		nm = (i >= '0' && i <= '9');
		goto def;

	default:
		nm = 0;
	def:	i = free_string+1;
		for (_string [i-1] = c; i < string_size; i++)
			switch (c = get_char (fp)) {
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				_string [i] = c;
				continue;

			default:
				nm = 0;
				_string [i] = c;
				continue;
			case EOF:
			case '(':
			case ')':
			case '\'':
			case '\"':
			case '.':
			case ' ':
			case '\n':
			case '\t':
				_string [i] = 0;
				unget_char (c, fp);
				if (nm)
					return (lisp) numify (atoi (&_string[free_string]));
				c = free_string;
				free_string = i+1;
				return intern (&_string[c], UNDEF);
			}
		error (pname_too_long, UNDEF);
		return NULL;	/* non atteint */
	}
}

/*
 *	Analyseur syntaxique.
 */

static lisp rnext (lisp file, lisp eof)
{
	lisp	l;

	l = lex (file, eof);
	if (l == eof)
		return eof;

	switch ((long)l) {
		case ')':
			return NIL;
		case '.':
			l = rfirst (file, eof);
			if ((long)lex (file, eof) != ')')
				error (dot_error, UNDEF);
			return l;
		case '\'':
			l = cons (sym_quote, cons (rfirst (file, eof), NIL));
			break;
		case '(':
			l = rnext (file, eof);
			break;
	}
	return cons (l, rnext (file, eof));
}

static lisp rfirst (lisp file, lisp eof)
{
	lisp	l;

	for (;;) {
		l = lex (file, eof);
		switch ((long)l) {
			case '\'':
				return cons (sym_quote, cons (rfirst (file,eof), NIL));
			case '.':
				error (dot_error, UNDEF);
			case ')':
				continue;
			case '(':
				return rnext (file,eof);
			default:
				return l;
		}
	}
}

/*
 *	La fonction lisp read.
 */

static lisp uread (lisp file, lisp eof)
{
	if (file == NIL)
		return rfirst ((lisp)&_files [fileno (stdin)], eof);

	if (filep (file))
		return rfirst (file, eof);

	error (file_expected, file);
	return NIL;
}

/*
 *	readstring.
 */

lisp lisp_read_string (char* p)
{
	__p = p;
	return rfirst (NIL, NIL);
}

static lisp read_string (lisp s)
{
	if (stringp (s)) {
		__p = (char*) s;
		return rfirst (NIL, NIL);
	}
	error (string_expected, s);
	return NIL;
}

/*
 *	La fonction quote.
 */

static lisp quote (lisp l)
{
	return CAR (l);
}

/*
 *	Les lambda expressions.
 */

static lisp lambda (lisp l)
{
	if (!cellp (CAR (l)))
		error (bad_arglist, CAR (l));
	return cons ((void*) LAMBDA, l);
}

/*
 *	Les lexpr.
 */

static lisp lexpr (lisp l)
{
	if (!cellp (CAR (l)))
		error (bad_arglist, CAR (l));
	return cons ((void*) LEXPR, l);
}

/*
 *	Les macro expressions.
 */

static lisp macro (lisp l)
{
	if (!cellp (CAR (l)))
		error (bad_arglist, CAR (l));
	return cons ((void*) MACRO, l);
}

/*
 *	L'affectation.
 */

static lisp set (lisp var, lisp value)
{
	if (!symbolp (var))
		error (symbol_expected, var);
	VALUE (var) = value;
	return value;
}

/*
 *	L'evaluateur.
 */

typedef lisp (*pfun) ();

lisp lisp_eval (lisp l)
{
	lisp		f, fv;
	int			i;
	lisp		a1 = NIL;
	lisp		a2;
	lisp		a3;
	lisp		call;
	stack_frame	frame;

tail_rec:
	/*
	 *	Tester d'abord le cas le plus frequent.
	 */
	if (symbolp (l)) {
		if (VALUE (l) == UNDEF)
			error (unbound_symbol, l);
		else
			return VALUE (l);
	}
	/* les objets auto-evalues: nil, nombres, strings, files */
	if (!cellp (l))
		return l;
	/*
	 *	((subr . addr)			arg1 ... argn)
	 *	((lambda (f1 ... fn) . body)	arg1 ... argn)
	 */
	call = l;	/* la forme a evaluer */
	f = CAR (l);	/* la fonction */
	l = CDR (l);	/* la liste d'arguments */

	if (symbolp (f))
		f = VALUE (f);
	else if (cellp (f) && ((long)CAR(f) > MACRO || (long)CAR(f) < SUBR0))
		f = lisp_eval (f);

	if (!cellp (f))
		error (undefined_function, CAR (call));

	fv = CDR (f);
	switch ((long)CAR (f)) {
		case SUBR0:
			return (*(pfun)fv) ();

		case SUBR1:
			a1 = cellp (l) ? lisp_eval (CAR (l)) : NIL;
			return (*(pfun)fv) (a1);

		case SUBR2:
			if (cellp (l)) {
				a1 = lisp_eval (CAR (l)); l = CDR (l);
				a2 = cellp (l) ? lisp_eval (CAR (l)) : NIL;
			} else {
				a1 = NIL;
				a2 = NIL;
			}
			return (*(pfun)fv) (a1, a2);

		case SUBR3:
			if (cellp (l)) {
				a1 = lisp_eval (CAR (l)); l = CDR (l);
				if (cellp (l)) {
					a2 = lisp_eval (CAR (l)); l = CDR (l);
					a3 = cellp (l) ? lisp_eval (CAR (l)) : NIL;
				} else {
					a2 = NIL;
					a3 = NIL;
				}
			} else {
				a1 = NIL;
				a2 = NIL;
				a3 = NIL;
			}
			return (*(pfun)fv) (a1, a2, a3);

		case FSUBR:
			return (*(pfun)fv) (l);

		case MACRO:
			/*
			 *	Empiler l'ancienne valeur du parametre
			 *	et effectuer la liaison.
			 */
			frame.next = frame_p;
			frame.type = MACRO;
			frame.nb = 1;
			frame.param [0] = CAR (CAR (fv));
			frame.value [0] = VALUE (CAR (CAR (fv)));
			frame_p = & frame;
			VALUE (frame.param [0]) = call;
			/*
			 *	Evaluer le corps de la fonction.
			 */
			for (l = CDR (fv); cellp (l); l = CDR (l))
				a1 = lisp_eval (CAR (l));
			/*
			 *	Restaurer l'ancienne liaison.
			 */
			VALUE (frame.param [0]) = frame.value [0];
			frame_p = frame.next;
			l = a1;
			goto tail_rec;

		case LEXPR:
			/*
			 *	Empiler l'ancienne valeur du parametre
			 *	et effectuer la liaison.
			 */
			a2 = a1 = cons (lisp_eval (CAR (l)), NIL);
			for (l = CDR (l); cellp (l); l = CDR (l)) {
				CDR (a1) = cons (lisp_eval (CAR (l)), NIL);
				a1 = CDR (a1);
			}
			frame.next = frame_p;
			frame.type = LEXPR;
			frame.nb = 1;
			frame.param [0] = CAR (CAR (fv));
			frame.value [0] = VALUE (CAR (CAR (fv)));
			frame_p = & frame;
			VALUE (frame.param [0]) = a2;
			/*
			 *	Evaluer le corps de la fonction.
			 */
			for (l = CDR (fv); cellp (l); l = CDR (l))
				a1 = lisp_eval (CAR (l));
			/*
			 *	Restaurer l'ancienne liaison.
			 */
			VALUE (frame.param [0]) = frame.value [0];
			frame_p = frame.next;
			return a1;

		case LAMBDA:
			/*
			 *	Evaluer les arguments actuels.
			 */
			frame.next = frame_p;
			frame.type = LAMBDA;
			for (i = 0; cellp (l); l = CDR (l))
				frame.value [i++] = lisp_eval (CAR (l));
			frame.nb = i;
			frame_p = & frame;
			/*
			 *	Empiler les anciennes valeurs.
			 */
			for (i = 0, l = CAR (fv); cellp (l); l = CDR (l)) {
				frame.param [i] = CAR (l);
				a1 = VALUE (CAR (l));
				VALUE (CAR (l)) = (i < frame.nb) ? frame.value [i] : NIL;
				frame.value [i++] = a1;
			}
			frame.nb = i;	/* nombre de parametres formels */
			/*
			 *	Evaluer le corps de la fonction.
			 */
			for (l = CDR (fv); cellp (l); l = CDR (l))
				a1 = lisp_eval (CAR (l));
			/*
			 *	Restaurer les anciennes liaisons.
			 */
			for (i = 0; i < frame.nb; i++)
				VALUE (frame.param [i]) = frame.value [i];
			frame_p = frame.next;
			return a1;

		default:
			error (undefined_function, CAR (call));
			return NIL;
	}
}

/*
 *	L'affectation (bis).
 */

static lisp setq (lisp l)
{
	if (l == NIL)
		error (bad_arglist, UNDEF);
	return set (CAR (l), CDR (l) ? lisp_eval (CAR (CDR (l))) : NIL);
}

#if !STAND_ALONE

void _SetCurrentWix (lisp value)
{
	VALUE (sym_wix) = value;
}

/*
 *	Evaluer un appel d'expr.
 */

void lisp_eval_lambda (char* msg, char* name, void* w, int n)
{
	lisp		fv;
	lisp		l;
	stack_frame	frame;
	catch_frame	cframe;

	cframe.tag = sym_tag_error;	/* tag pour error */
	cframe.type = CATCH;		/* type de bloc */
	cframe.next = (catch_frame*) frame_p;
	frame_p = (stack_frame*)& cframe;

	if (setjmp (cframe.jbuf)) {
		if (cellp (cframe.value)) {
			__outp = msg;
			sprintf (__outp, "%s", (char*) CAR (cframe.value));
			__outp += strlen (__outp);
			if (CDR (cframe.value) != UNDEF) {
				sprintf (__outp, ": ");
				__outp += strlen (__outp);
				probj (CDR (cframe.value), NIL);
			}
		} else
			sprintf (__outp, "%s", (char*) cframe.value);
		frame_p = (stack_frame*)cframe.next;	/* bloc initial */
		return;
	}

	fv = intern (name, NIL);
	fv = VALUE (fv);
	if (fv == NIL)
		return;		/* fonction non liee */
	fv = CDR (fv);	/* la lambda expression */

	frame.next = frame_p;
	frame.type = LAMBDA;
	frame.nb = 2;
	frame.param [0] = CAR (CAR (fv));
	frame.value [0] = VALUE (frame.param [0]);
	frame.param [1] = CAR (CDR (CAR (fv)));
	frame.value [1] = VALUE (frame.param [1]);
	frame_p = & frame;
	VALUE (frame.param [0]) = (lisp) w;
	VALUE (frame.param [1]) = (lisp) numify (n);
	/*
	 *	Evaluer le corps de la fonction.
	 */
	for (l = CDR (fv); cellp (l); l = CDR (l))
		(void) lisp_eval (CAR (l));
	/*
	 *	Restaurer l'ancienne liaison.
	 */
	VALUE (frame.param [0]) = frame.value [0];
	VALUE (frame.param [1]) = frame.value [1];
	frame_p = frame.next;
}

#endif

/*
 *	Retourne la liste des elements de l evalues.
 */

static lisp list (lisp l)
{
	lisp	a1;
	lisp	r;

	r = a1 = cons (lisp_eval (CAR (l)), NIL);
	for (l = CDR (l); cellp (l); l = CDR (l)) {
		CDR (a1) = cons (lisp_eval (CAR (l)), NIL);
		a1 = CDR (a1);
	}
	return r;
}

/*
 *	(*catch 'tag . body) [FSUBR]
 */

static lisp catch (lisp l)
{
	catch_frame	frame;
	lisp		r = NIL;

	frame.tag = lisp_eval (CAR (l));	/* sauver le tag */
	frame.type = CATCH;		/* type de bloc */
	frame.next = (catch_frame*) frame_p;
	frame_p = (stack_frame*)& frame;

	if (setjmp (frame.jbuf)) {			/* il y a eu un throw */
		frame_p = (stack_frame*)frame.next;	/* delier le bloc et */
		return frame.value;			/* retourner la valeur */
	}
	/*
	 *	Evaluer le corps du catch.
	 */
	for (l = CDR (l); cellp (l); l = CDR (l))
		r = lisp_eval (CAR (l));
	frame_p = (stack_frame*)frame.next;	/* delier le bloc */
	return r;				/* valeur du corps */
}

/*
 *	(*throw 'tag 'value) [SUBR2]
 */

static lisp throw (lisp tag, lisp value)
{
	stack_frame*	p;
	stack_frame*	q;
	int				i;

	for (p = frame_p; p; p = p -> next)
		if (p -> type == CATCH && ((catch_frame*)p) -> tag == tag) {
			for (q = frame_p; q != p; q = q -> next)
				switch (q -> type) {
				case LAMBDA:
				case LEXPR:
				case MACRO:
					for (i = 0; i < q -> nb; i++)
						VALUE (q -> param [i]) = q -> value [i];
				}
			((catch_frame*)p) -> value = value;
			longjmp (((catch_frame*)p) -> jbuf, 1);
		}
	return NIL;
}

/*
 *	La boucle print, eval, read.
 */

#if STAND_ALONE

static lisp toplevel (void)
{
	lisp		l;
	catch_frame	frame;

	frame.tag = sym_tag_error;	/* tag pour error */
	frame.type = CATCH;		/* type de bloc */
	frame.next = (catch_frame*) frame_p;
	frame_p = (stack_frame*)& frame;

	if (setjmp (frame.jbuf)) {
		if (cellp (frame.value)) {
			fprintf (stderr, "%s: ", CAR (frame.value));
			if (CDR (frame.value) != UNDEF)
				probj (CDR (frame.value), &_files [fileno (stderr)]);
			fprintf (stderr, "\n");
		} else
			fprintf (stderr, "%s\n", frame.value);
		frame_p = (stack_frame*)&frame;	/* bloc initial */
	}
	for (;;) {
		printf ("-> "); fflush (stdout);
		l = rfirst (&_files [fileno (stdin)], sym_eof);
		if (l == sym_eof) {
			printf ("Bye\n");
			exit (0);
		}
		l = lisp_eval (l);
		printf ("= ");
		probj (l, &_files [fileno (stdout)]);
		printf ("\n");
	}
}

#endif

/*
 *	Load initial du startup lisp.
 */

int lisp_load_file (char* file)
{
	lisp		l;
	catch_frame	frame;
	FILE*		fp;

	frame.tag = sym_tag_error;	/* tag pour error */
	frame.type = CATCH;		/* type de bloc */
	frame.next = (catch_frame*) frame_p;
	frame_p = (stack_frame*)& frame;

	fp = fopen (file, "r");
	if (!fp)
		return 0;

	_files [fileno (fp)] = fp;

	if (setjmp (frame.jbuf)) {
		fclose (fp);
		return 1;
	}
	for (;;) {
		l = rfirst (&_files [fileno (fp)], sym_eof);
		if (l == sym_eof) {
			fclose (fp);
			return 2;
		}
		(void) lisp_eval (l);
	}
}

static void startup (void)
{
	extern char HomeDirectory [];
	char	tc [256];
	lisp	l;

	__p = _startup_lisp;
	while ((l = rfirst (NIL, sym_eof)) != sym_eof)
		(void) lisp_eval (l);
#if 123
	sprintf (tc, "%s/.wx_startup.lisp", HomeDirectory);
	(void) lisp_load_file (tc);
#endif
}

void lisp_initialize (void)
{
	lisp	stk;

	stack_org = &stk;

	_files [fileno (stdin)] = stdin;
	_files [fileno (stdout)] = stdout;
	_files [fileno (stderr)] = stderr;

	alloc_space ();
	init_cons ();

#if !STAND_ALONE
	widget_expected		= make_string ("Widget expected");
#endif
	number_expected		= make_string ("Number expected");
	string_expected		= make_string ("String expected");
	file_expected		= make_string ("File expected");
	list_expected		= make_string ("List expected");
	symbol_expected		= make_string ("Symbol expected");
	out_of_cons		= make_string ("Out of cons cells");
	dot_error		= make_string ("Dot error");
	bad_arglist		= make_string ("Bad arglist");
	unbound_symbol		= make_string ("Unbound symbol");
	undefined_function	= make_string ("Undefined function");
	no_catch		= make_string ("No *catch found with this tag");
	pname_too_long		= make_string ("Pname too long");
	missing_quote		= make_string ("Missing \"");
	out_of_symbol		= make_string ("No room for a new symbol");

	init_symbols ();
	startup ();
}

#if STAND_ALONE
int main (int argc, char** argv)
{
	lisp_initialize ();
	printf ("PLisp Version 1.1\n");
	toplevel ();
}
#endif

/*
 *	les fonctions.
 */

static lisp consp (lisp l)
{
	return (lisp) (long) cellp (l);
}

static lisp car (lisp x)
{
	if (!x)
		return NIL;
	CHECK_CELL (x);
	return CAR (x);
}

static lisp cdr (lisp x)
{
	if (!x)
		return NIL;
	CHECK_CELL (x);
	return CDR (x);
}

static lisp rplaca (lisp x, lisp y)
{
	CHECK_CELL (x);
	return CAR (x) = y;
}

static lisp rplacd (lisp x, lisp y)
{
	CHECK_CELL (x);
	return CDR (x) = y;
}

static lisp rplacb (lisp x, lisp y)
{
	CHECK_CELL (x);
	CHECK_CELL (y);
	CAR (x) = CAR (y);
	CDR (x) = CDR (y);
	return x;
}

static lisp lessp (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) (long) (denum (x) < denum (y));
}

static lisp less_eqp (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) (long) (denum (x) <= denum (y));
}

static lisp greaterp (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) (long) (denum (x) > denum (y));
}

static lisp greater_eqp (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) (long) (denum (x) >= denum (y));
}

static lisp eq (lisp x, lisp y)
{
	return (lisp) (long) (x == y);
}

static lisp neq (lisp x, lisp y)
{
	return (lisp) (long) (x != y);
}

static lisp sum (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) numify (denum (x) + denum (y));
}

static lisp diff (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) numify (denum (x) - denum (y));
}

static lisp product (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) numify (denum (x) * denum (y));
}

static lisp quotient (lisp x, lisp y)
{
	CHECK_NUMBER (x);
	CHECK_NUMBER (y);
	return (lisp) numify (denum (x) / denum (y));
}

static lisp progn (lisp l)
{
	lisp	r;

	for (r = NIL; cellp (l); l = CDR (l))
		r = lisp_eval (CAR (l));
	return r;
}

static lisp _if (lisp l)
{
	return (lisp_eval (CAR (l)) != NIL)
		? lisp_eval (CAR (CDR (l)))
		: progn (CDR (CDR (l)));
}

static lisp _while (lisp l)
{
	while (lisp_eval (CAR (l)) != NIL)
		progn (CDR (l));
	return NIL;
}

static lisp _do_while (lisp l)
{
	do {
		progn (CDR (l));
	} while (lisp_eval (CAR (l)) != NIL);
	return NIL;
}

/*
 *	les conditionnelles and & or
 */

static lisp and (lisp l)
{
	lisp	r;

	for (r = NIL; cellp (l); l = CDR (l))
		if (! (r = lisp_eval (CAR (l))))
			return NIL;
	return r;
}

static lisp or (lisp l)
{
	lisp	r;

	for (; cellp (l); l = CDR (l))
		if ((r = lisp_eval (CAR (l))) != NULL)
			return r;
	return NIL;
}

#if !STAND_ALONE
static lisp c_match_closing (lisp w, lisp l, lisp r)
{
	return (lisp) (long) find_close (w, denum (l), denum (r), 1);
}

static lisp c_match_opening (lisp w, lisp l, lisp r)
{
	return (lisp) (long) find_open (w, denum (l), denum (r), 1);
}

static lisp match_closing (lisp w, lisp l, lisp r)
{
	return (lisp) (long) find_close (w, denum (l), denum (r), 0);
}

static lisp match_opening (lisp w, lisp l, lisp r)
{
	return (lisp) (long) find_open (w, denum (l), denum (r), 0);
}

static lisp forward_search_yanked (lisp w, lisp c)
{
	char*	pattern;
	long	len = 0;
	lisp	ret;

	extern char* GetKilled (int n, long* len);

	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);

	pattern = GetKilled (denum (c), &len);
	if (pattern == 0)
		return NIL;

	ret = (lisp) (long) WixForwardSearch (w, pattern, 1);
	XtFree (pattern);
	return ret;
}

static lisp forward_search (lisp w, lisp s, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) WixForwardSearch (w, s, denum (c));
}

static lisp backward_search (lisp w, lisp s, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) WixBackwardSearch (w, s, denum (c));
}

static lisp backward_re_search (lisp w, lisp s)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);

	if (!stringp (s))
		error (string_expected, s);

	return (lisp) (long) re_backward_search (w, s);
}

static lisp forward_re_search (lisp w, lisp s)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);

	if (!stringp (s))
		error (string_expected, s);

	return (lisp) (long) re_forward_search (w, s);
}

static lisp forward_char (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) ForwardChar (w, denum (c));
}

static lisp backward_char (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) BackwardChar (w, denum (c));
}

static lisp forward_word (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) ForwardWord (w, denum (c));
}

static lisp backward_word (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) BackwardWord (w, denum (c));
}

static lisp forward_line (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) ForwardLine (w, denum (c));
}

static lisp backward_line (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) BackwardLine (w, denum (c));
}

static lisp delete_char (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) DeleteChar (w, denum (c));
}

static lisp yank (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) (long) Yank (w, denum (c), 0);
}

static lisp display_message (lisp w, lisp c)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	Message (w, c);
	return NIL;
}

static lisp get_column (lisp w)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) numify (WixGetColumn (w));
}

static lisp get_width (lisp w)
{
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);
	return (lisp) numify (WixGetWidth (w));
}

static lisp is_keyword (lisp w) {
	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);

	return (lisp) (long) (CurrentCharType (w) == KEYWORD_COLOR);
}

static lisp looking_at (lisp w, lisp s, lisp n)
{
	int	len;

	if (!stringp (s))
		error (string_expected, s);

	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);

	len = (n != NIL) ? denum (n) : strlen (s);
	return (lisp) (long) (0 == bcmp (s, _GetCursorEnd (w), len));
}

static lisp looking_at_re (lisp w, lisp s)
{
	char*	cp;
	char*	end;
	char*	begp;
	char*	ok;

	if (!stringp (s))
		error (string_expected, s);

	if ((((unsigned long)w) & 3) || *(lisp*)w != w)
		error (widget_expected, w);

	if (pa_re_comp (s))
		error (make_string ("Invalid re."), w);

	for (cp = end = _GetCursorEnd (w); *cp != '\n'; ++cp)
		;
	*cp = 0;
	ok = pa_re_exec (end, &begp);
	*cp = '\n';
	return ok ? TRUE : NIL;
}

static lisp at_end_of_buffer (lisp w)
{
	return (lisp) (long) (_GetCursorEnd (w) == _GetBufferEnd (w));
}
#endif

static void init_symbols (void)
{
   sym_quote		= intern ("quote", cons ((void*)FSUBR, quote));
sym_tag_error		= intern ("%ERROR", NIL);
      sym_eof		= intern ("eof", NIL); VALUE (sym_eof) = sym_eof;
internal_comma_atsign	= intern ("internal-comma-atsign", NIL);
internal_comma_dot	= intern ("internal-comma-dot", NIL);
internal_comma		= intern ("internal-comma", NIL);
internal_backquote	= intern ("internal-backquote", NIL);

	(void) intern ("t", TRUE);
	(void) intern ("stdin",	 & _files [fileno (stdin)]);
	(void) intern ("stdout", & _files [fileno (stdout)]);
	(void) intern ("stderr", & _files [fileno (stderr)]);
	(void) intern ("open",	 cons ((void*)SUBR2, file_open));
	(void) intern ("close",	 cons ((void*)SUBR1, file_close));
	(void) intern ("remob",	 cons ((void*)FSUBR, remove_symbol));
	(void) intern ("gc",	 cons ((void*)SUBR0, ugc));
	(void) intern ("cons",	 cons ((void*)SUBR2, cons));
	(void) intern ("list",	 cons ((void*)FSUBR, list));
	(void) intern ("set",	 cons ((void*)SUBR2, set));
	(void) intern ("setq",	 cons ((void*)FSUBR, setq));
	(void) intern ("eval",	 cons ((void*)SUBR1, lisp_eval));
	(void) intern ("print",	 cons ((void*)SUBR2, probj));
	(void) intern ("terpri", cons ((void*)SUBR1, terpri));
	(void) intern ("readstring",	cons ((void*)SUBR1, read_string));
	(void) intern ("read",	 cons ((void*)SUBR2, uread));
	(void) intern ("oblist", cons ((void*)SUBR0, oblist));
	(void) intern ("lambda", cons ((void*)FSUBR, lambda));
	(void) intern ("macro",	 cons ((void*)FSUBR, macro));
	(void) intern ("lexpr",	 cons ((void*)FSUBR, lexpr));
	(void) intern ("*catch", cons ((void*)FSUBR, catch));
	(void) intern ("*throw", cons ((void*)SUBR2, throw));
	(void) intern ("rplaca", cons ((void*)SUBR2, rplaca));
	(void) intern ("rplacd", cons ((void*)SUBR2, rplacd));
	(void) intern ("rplacb", cons ((void*)SUBR2, rplacb));
	(void) intern ("consp", cons ((void*)SUBR1, consp));
	(void) intern ("car", cons ((void*)SUBR1, car));
	(void) intern ("cdr", cons ((void*)SUBR1, cdr));
	(void) intern ("<", cons ((void*)SUBR2, lessp));
	(void) intern (">", cons ((void*)SUBR2, greaterp));
	(void) intern ("<=", cons ((void*)SUBR2, less_eqp));
	(void) intern (">=", cons ((void*)SUBR2, greater_eqp));
	(void) intern ("=", cons ((void*)SUBR2, eq));
	(void) intern ("not", cons ((void*)SUBR2, eq));
	(void) intern ("null", cons ((void*)SUBR2, eq));
	(void) intern ("eq", cons ((void*)SUBR2, eq));
	(void) intern ("neq", cons ((void*)SUBR2, neq));
	(void) intern ("!=", cons ((void*)SUBR2, neq));
	(void) intern ("+", cons ((void*)SUBR2, sum));
	(void) intern ("-", cons ((void*)SUBR2, diff));
	(void) intern ("*", cons ((void*)SUBR2, product));
	(void) intern ("/", cons ((void*)SUBR2, quotient));
	(void) intern ("and", cons ((void*)FSUBR, and));
	(void) intern ("or", cons ((void*)FSUBR, or));
	(void) intern ("if", cons ((void*)FSUBR, _if));
	(void) intern ("progn", cons ((void*)FSUBR, progn));
	(void) intern ("while", cons ((void*)FSUBR, _while));
	(void) intern ("dowhile", cons ((void*)FSUBR, _do_while));
#if !STAND_ALONE
sym_wix = intern ("wix", TRUE);
	(void) intern ("c-match-closing", cons ((void*)SUBR3, c_match_closing));
	(void) intern ("c-match-opening", cons ((void*)SUBR3, c_match_opening));
	(void) intern ("match-closing", cons ((void*)SUBR3, match_closing));
	(void) intern ("match-opening", cons ((void*)SUBR3, match_opening));
	(void) intern ("forward-search-yanked", cons ((void*)SUBR2, forward_search_yanked));
	(void) intern ("forward-search", cons ((void*)SUBR3, forward_search));
	(void) intern ("forward-re-search", cons ((void*)SUBR2, forward_re_search));
	(void) intern ("backward-search", cons ((void*)SUBR3, backward_search));
	(void) intern ("backward-re-search", cons ((void*)SUBR2, backward_re_search));
	(void) intern ("delete-char", cons ((void*)SUBR2, delete_char));
	(void) intern ("insert-string", cons ((void*)SUBR2, (lisp) WixInsert));
	(void) intern ("forward-line", cons ((void*)SUBR2, forward_line));
	(void) intern ("backward-line", cons ((void*)SUBR2, backward_line));
	(void) intern ("forward-char", cons ((void*)SUBR2, forward_char));
	(void) intern ("backward-char", cons ((void*)SUBR2, backward_char));
	(void) intern ("forward-word", cons ((void*)SUBR2, forward_word));
	(void) intern ("backward-word", cons ((void*)SUBR2, backward_word));
	(void) intern ("c-beginning-of-function", cons ((void*)SUBR1, (lisp) CBeginningOfFunction));
	(void) intern ("c-end-of-function", cons ((void*)SUBR1, (lisp) CEndOfFunction));
	(void) intern ("beginning-of-line", cons ((void*)SUBR1, (lisp) BeginningOfLine));
	(void) intern ("end-of-line", cons ((void*)SUBR1, (lisp) EndOfLine));
	(void) intern ("delete-end-line", cons ((void*)SUBR1, (lisp) DeleteEndLine));
	(void) intern ("beginning-of-buffer", cons ((void*)SUBR1, (lisp) BeginningOfBuffer));
	(void) intern ("end-of-buffer", cons ((void*)SUBR1, (lisp) EndOfBuffer));
	(void) intern ("at-end-of-buffer", cons ((void*)SUBR1, at_end_of_buffer));
	(void) intern ("display-message", cons ((void*)SUBR2, display_message));
	(void) intern ("yank", cons ((void*)SUBR2, yank));
	(void) intern ("get-column", cons ((void*)SUBR1, get_column));
	(void) intern ("get-width", cons ((void*)SUBR1, get_width));
	(void) intern ("is-keyword", cons ((void*)SUBR1, is_keyword));
	(void) intern ("looking-at", cons ((void*)SUBR3, looking_at));
	(void) intern ("looking-at-re", cons ((void*)SUBR2, looking_at_re));
	(void) intern ("enable-redisplay", cons ((void*)SUBR3, (lisp) EnableRedisplay));
	(void) intern ("set-mark", cons ((void*)SUBR2, (lisp) WixSetMark));
	(void) intern ("goto-mark", cons ((void*)SUBR2, (lisp) WixGotoMark));
	(void) intern ("delete-mark", cons ((void*)SUBR2, (lisp) WixDeleteMark));
	(void) intern ("create-mark", cons ((void*)SUBR2, (lisp) WixCreateMark));
#endif
}
