/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>

#if defined (SYSV) || defined (__linux__)
#include <string.h>

#define bcopy(s,d,n) memcpy((d),(s),(n))
#define bcmp(a,b,n) memcmp((a),(b),(n))
#define bzero(a,n) memset((a),0,(n))
#endif

#define STACK_SIZE	32
#define NOFILE		32

typedef void*	lisp;	/* objet lisp generique */

typedef struct _stack_frame {
	struct _stack_frame*	next;				/* bloc suivant */
	int						type;				/* type de bloc */
	int						nb;					/* nombre de liaisons */
	lisp					param [STACK_SIZE];	/* symbole */
	lisp					value [STACK_SIZE];	/* ancienne valeur */
} stack_frame;

typedef struct _catch_frame {
	struct _catch_frame*	next;	/* bloc suivant */
	int						type;	/* type de bloc */
	jmp_buf					jbuf;
	lisp					tag;
	lisp					value;
} catch_frame;

typedef struct {
	lisp	kar;
	lisp	kdr;
} cell;

typedef struct {
	lisp	value;
	char*	pname;
} sym;

#define CAR(x) (((cell*)(x))->kar)
#define CDR(x) (((cell*)(x))->kdr)

#define VALUE(x) (((sym*)(x))->value)
#define PNAME(x) (((sym*)(x))->pname)

#define FTYPE(x) (((lispfun*)(x))->type)
#define FVAL(x) (((lispfun*)(x))->f)

#define cellp(x) (((cell*)(x))>=_heap && ((cell*)(x))<_heap+heap_size)
#define symbolp(x) (((sym*)(x))>=_obarray && ((sym*)(x))<_obarray+obarray_size)
#define stringp(x) (((char*)(x))>=_string && ((char*)(x))<_string+string_size)
#define numbp(x) (((long)(x))<0)
#define filep(x) (((FILE**)(x))>=_files && ((FILE**)(x))<_files+NOFILE)

#define CHECK_NUMBER(x) if (!numbp (x)) error (number_expected, x)
#define CHECK_CELL(x) if (!cellp (x) || ((long)CAR(x) <= MACRO && (long)CAR(x) >= SUBR0)) error (list_expected, x)

#define numify(x) (((unsigned long)(x))|0x80000000)
#define denum(x) ((((long)(x))<<1)>>1)

#define NIL		0
#define TRUE		((lisp)1)
#define UNDEF		((lisp)2)

#define SUBR0	100
#define SUBR1	101
#define SUBR2	102
#define SUBR3	103
#define FSUBR	104
#define LEXPR	105
#define LAMBDA	106
#define MACRO	107
#define CATCH	108

#define HEAP_SIZE	4096
#define OBARRAY_SIZE	2048
#define STRING_SIZE	8192

/*
 *	le startup
 */

static char _startup_lisp [] = "\
(set 'defmac (macro (l) \n\
        (list 'set \n\
                (list 'quote (car (cdr l))) \n\
                (cons 'macro (cdr (cdr l)))))) \n\
 \n\
(defmac defun (l) \n\
        (list 'set \n\
                (list 'quote (car (cdr l))) \n\
                (cons 'lambda (cdr (cdr l)))))) \n\
 \n\
(defun error (l what) (*throw '%ERROR (cons l what))) \n\
 \n\
(defun load (l) \n\
	((lambda (f) \n\
		(*catch '%ERROR \n\
			(setq f (open l \"r\")) \n\
			(while (neq '**eof** (eval (read f '**eof**)))) \n\
			(close f)) \n\
		l) nil)) \n\
 \n\
(defmac cadr (l) (rplacb l (list 'car (list 'cdr (car (cdr l)))))) \n\
(defmac caar (l) (rplacb l (list 'car (list 'car (cadr l))))) \n\
(defmac cdar (l) (rplacb l (list 'cdr (list 'car (cadr l))))) \n\
(defmac cddr (l) (rplacb l (list 'cdr (list 'cdr (cadr l))))) \n\
(defmac caddr (l) (rplacb l (list ' car (list 'cdr (list 'cdr (cadr l)))))) \n\
(defmac cdddr (l) (rplacb l (list ' cdr (list 'cdr (list 'cdr (cadr l)))))) \n\
 \n\
(defun cond-convert (l) \n\
	(if (null l) nil \n\
	    (if (= (caar l) t) \n\
			(if (cdr (cdar l)) \n\
			    (cons 'progn (cdar l)) \n\
			    (car (cdar l))) \n\
		(list	'if \n\
			(caar l) \n\
			(if (cdr (cdar l)) \n\
			    (cons 'progn (cdar l)) \n\
			    (car (cdar l))) \n\
			(cond-convert (cdr l)))))) \n\
 \n\
(defmac cond (l) (rplacb l (cond-convert (cdr l)))) \n\
 \n\
(defun length (l) (if l (+ 1 (length (cdr l))) 0)) \n\
 \n\
(defun match-open-par (w n) (match-opening w 40 41)) \n\
(defun match-close-par (w n) (match-closing w 40 41)) \n\
 \n\
(defun next-function (w n) \n\
	   (if (c-end-of-function w) \n\
		   nil \n\
		   (c-match-closing w 123 125)) \n\
	   (if (null (c-match-closing w 123 125)) \n\
		   (display-message w \"No next function.\")) \n\
	   (backward-char w 1) \n\
	   (c-beginning-of-function w)) \n\
 \n\
(defun previous-function (w n) \n\
	   (c-beginning-of-function w) \n\
	   (if (null (backward-search w \"}\" 1)) \n\
		   (display-message w \"No previous function.\")) \n\
	   (c-beginning-of-function w)) \n\
)\n\
(remob w n l what f **eof**) \n\
";
