/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/resource.h>
#if defined (SYSV) || defined (__linux__)
#include <sys/utsname.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Xmu/Editres.h>
#include <X11/xpm.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/Query.h>
#include <Xpa/Icon.h>
#include <Xpa/FSelector.h>
#include <Xpa/PanedWin.h>
#include <Xpa/MenuBar.h>
#include <Xpa/ScrolledWin.h>
#include <Xpa/Scrollbar.h>
#include <Xpa/PushButton.h>
#include <Xpa/Ruler.h>
#include <Xpa/Row.h>
#include <Xpa/Wix.h>
#include <Xpa/Frame.h>
#include <Xpa/PopupMenu.h>

#include "wix.h"
#include "client.h"
#include "doc.h"

typedef struct {
	int	checkpointRate;
	Boolean	restoreState;
	String	configFile;
	Boolean	stripWindow;
	int	ncol;
	int	nline;
} AppData, *AppDataPtr;

extern int _Xdebug;

static AppData	app_data;

static Widget	mode_popup;		/* le menu spcifique  chaque mode */
static Widget	window_popup;	/* le menu des fentres */

static	char	listTranslations [] =
	"<Btn3Down>:		scroll-window()\n"
	"<Btn3Motion>:		scroll-window()\n"
	"<Btn3Up>:			scroll-window()\n"
	"<Btn1Down>:		select-line() activate()\n"
	"Ctrl<Btn4Down>:		scroll-up(8)\n"
	"Ctrl<Btn5Down>:		scroll-down(8)\n"
	"Shift<Btn4Down>:	scroll-up(4)\n"
	"Shift<Btn5Down>:	scroll-down(4)\n"
	"<Btn4Down>:		scroll-up(2)\n"
	"<Btn5Down>:		scroll-down(2)\n"
	"<Key>Prior:			backward-page()\n"
	"<Key>Next:			forward-page()\n"
	"<Key>space:			forward-page()\n"
	"<Key>BackSpace:		backward-page()\n"
	"<Key>Down:			scroll-down()\n"
	"<Key>Up:			scroll-up()";

static char	helpTranslations [] =
	"Ctrl<Btn4Down>:		scroll-up(8)\n"
	"Ctrl<Btn5Down>:		scroll-down(8)\n"
	"Shift<Btn4Down>:	scroll-up(4)\n"
	"Shift<Btn5Down>:	scroll-down(4)\n"
	"<Btn4Down>:		scroll-up(2)\n"
	"<Btn5Down>:		scroll-down(2)\n"
	"<Key>Prior:			backward-page()\n"
	"<Key>Next:			forward-page()\n"
	"<Key>space:			forward-page()\n"
	"<Key>BackSpace:		backward-page()\n"
	"<Key>Left:			scroll-left()\n"
	"<Key>Right:			scroll-right()\n"
	"<Key>Down:			scroll-up()\n"
	"<Key>Up:			scroll-down()";

XtTranslations	listTrans;	/* table de translations des wix-list */
XtTranslations	HelpTrans;	/* table de translations des wix-help */
Cursor		WatchCurs;
Cursor		TextCurs;
Cursor		ArrowCurs;
Boolean		debugMode = False;
Atom		A_HERE;
Atom		A_CLIENT;
FontInfo*	default_font;
WixWindow*	theWindow;
Pixmap		QuestionPixmap;
Pixmap		ErrorPixmap;
Pixmap		OkPixmap;
Pixmap		modifPixmap;
Pixmap		lockPixmap;
Widget		ghost;

/*
 *	Relance le serveur.
 */

static void getSnapshot (char* file)
{
	char	tc [256];
	FILE*	fp;
	int		x, y;
	int		width, height;
	int		pos;
	int		map_state;	/* IsUnmapped, IsUnviewable, IsViewable */
extern int	def_width;
extern int	def_height;

	fp = fopen (file, "r");
	if (!fp)
		return;

	(void*) fgets (tc, sizeof (tc) - 1, fp);
	if (!bcmp (tc, "pattern=", 8)) {
		for (x=0; tc [x+8] != '\n'; ++x)
			prev_searched_string [x] = tc [x+8];
		prev_searched_string [x] = 0;	/* virer le \n de fin */
		while (8 == fscanf (fp, "%s %s %d %d %d %d %d %d",
					theWindow -> directory,
					theWindow -> filename,
					&width, &height, &x, &y, &pos, &map_state))
		{
			if (def_width == 0 && def_height == 0) {
				Dimension	wi, he;

				XtRealizeWidget (theWindow -> toplevel);
				XtVaGetValues (theWindow -> toplevel,
					XtNwidth,	&wi,
					XtNheight,	&he,
					NULL);
				def_width = wi;
				def_height = he;
				XtUnrealizeWidget (theWindow -> toplevel);
			}

			loadFileFunction (theWindow -> wix, theWindow -> filename);
			sprintf (tc, "%dx%d+%d+%d", width, height, x, y);
			WixSetInsertPosition (theWindow -> wix, pos);
			WixRedraw (theWindow -> wix);
			XtVaSetValues (theWindow -> toplevel,
					XtNgeometry,	tc,
					XtNiconic,		map_state == IsUnmapped,
					NULL);
			createWindowCallback (NULL, HomeDirectory, NULL);
		}
	}
	fclose (fp);
	unlink (file);
}

/*
 *	Cre le fichier de snapshot.
 */

Boolean _SetSnapshot (String file, Boolean modp)
{
	WixWindow*			data;
	Boolean				m;
	int					pos;
	Position			x, y;
	Dimension			width, height;
	FILE*				fp;
	XWindowAttributes	xwa;

	/* regarder s'il y a des buffers  sauver avant toute chose */
	for (data = _head; data != NULL; data = data -> next) {
		/* se dinteresser des fichiers inexistants (et des fentres non utilises) */
		if (data -> free || access (data -> filename, 0) != 0)
			continue;

		if (modp && (XtVaGetValues (data -> wix, XtNbufferModified, &m, NULL), m))
		{
			/* le message vaut tous les commetaires ! */
			static char msg [] = "There are modified but not saved\n"
								 "buffers. You must save or unmodify\n"
								 "them before restarting the WiX server.";
			data -> what_after = NULL;
			popAlert (data, msg, False);
			WaitForResponse (data);
			return False;
		}
	}
	/* ok, on peut y aller: crer le snapshot avec au moins la chane  chercher */
	fp = fopen (file, "w");
	if (fp == NULL)
		return True;	/* bizarre, c'est le home-dir de l'utilisateur ! */

	fprintf (fp, "pattern=%s\n", prev_searched_string);
	/* parcourir la liste des fentres */
	for (data = _head; data; data = data -> next) {
		Window			root;
		Window			wm_win;
		Window			*children;
		unsigned int	n_children;
		int				xoffset;
		int				yoffset;
		/* se dinteresser des fichiers inexistants (et des fentres non utilises) */
		if (data -> free || access (data -> filename, 0) != 0)
			continue;
		/* Pour viter le dcalage progressif des fentres  chaque */
		/* restart, on cherche les offsets par rapport au dcor du wm */
		XQueryTree (
				XtDisplay (data -> toplevel),
				XtWindow (data -> toplevel),
				&root, &wm_win, &children, &n_children);
		if (n_children > 0)
			XFree (children);	/* inutilis */
		XGetWindowAttributes (
			XtDisplay (data -> toplevel),
			wm_win,
			&xwa);
		xoffset = xwa.x;	/* offsets dus au dcor */
		yoffset = xwa.y;	/* du window manager */
		/* La fentre est-elle iconifie? */
		XGetWindowAttributes (
				XtDisplay (data -> toplevel),
				XtWindow (data -> toplevel),
				&xwa);
		XtVaGetValues (data -> toplevel,
			XtNx,	&x,
			XtNy,	&y,
			XtNwidth, &width,
			XtNheight, &height,
			NULL);
		pos = WixGetInsertPosition (data -> wix);
		fprintf (fp, "%s %s %d %d %d %d %d %d\n",
				data -> directory,
				data -> filename,
				width, height, x - xoffset, y - yoffset, pos, xwa.map_state);
	}
	fclose (fp);	/* tout est ok: fermer le snapshot */
	return True;
}

/*
 *	Redmarre l'diteur (pour rinitialiser les ressources et/ou rinterpreter le startup).
 */

extern char	ProgName [256];	/* de menu.c */

static void restartServerAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	int	i;

	if (! _SetSnapshot (".wx_state", True))
		return;

	for (i = 3; i < 32; i++)
		(void) close (i);

	execlp (ProgName,
			"wxd", "-display", DisplayString (XtDisplay (w)), "-r",
			NULL);
}

/*
 *	Cration des pixmaps pour les boites de dialogues, et autres...
 */

static void createPixmaps (Widget w)
{
#define ok_width 16
#define ok_height 16
static char ok_bits[] = {
   0xe0, 0x03, 0xf8, 0x0f, 0xfc, 0x1f, 0xe6, 0x2c, 0xc2, 0x2c, 0x99, 0x64,
   0x99, 0x70, 0x99, 0x78, 0x99, 0x70, 0x99, 0x64, 0xc2, 0x2c, 0xe6, 0x2c,
   0xfc, 0x1f, 0xf8, 0x0f, 0xe0, 0x03, 0x00, 0x00};
#define lock_width 16
#define lock_height 16
	static char lock_bits[] = {
	   0xe0, 0x03, 0xf8, 0x0f, 0x1c, 0x1c, 0x0c, 0x18, 0x06, 0x30, 0x06, 0x30,
	   0xff, 0x7f, 0x7f, 0x7f, 0x3f, 0x7e, 0x7f, 0x7f, 0x3f, 0x7e, 0x3f, 0x7e,
	   0x3f, 0x7e, 0x3f, 0x7e, 0xff, 0x7f, 0x00, 0x00};
#define modif_width 16
#define modif_height 16
	static char modif_bits[] = {
	   0x00, 0x00, 0xe0, 0x03, 0xf8, 0x0f, 0xfc, 0x1f, 0xee, 0x3b, 0xce, 0x39,
	   0xaf, 0x7a, 0x6f, 0x7b, 0x6f, 0x7b, 0xef, 0x7b, 0xef, 0x7b, 0xee, 0x3b,
	   0xee, 0x3b, 0xfc, 0x1f, 0xf8, 0x0f, 0xe0, 0x03};
#define	question_width	22
#define	question_height	22
	static char question_bits [] = {
		0xf0, 0x3f, 0x00, 0x58, 0x55, 0x00, 0xac, 0xaa, 0x00, 0xd6, 0x5f, 0x01,
		0xea, 0xbf, 0x02, 0xf6, 0x7f, 0x01, 0xea, 0xba, 0x02, 0xf6, 0x7d, 0x05,
		0xea, 0xba, 0x0a, 0x56, 0x7d, 0x15, 0xaa, 0xbe, 0x1e, 0x56, 0x5f, 0x01,
		0xac, 0xaf, 0x02, 0x58, 0x57, 0x01, 0xb0, 0xaf, 0x00, 0x60, 0x55, 0x01,
		0xa0, 0xaa, 0x00, 0x60, 0x17, 0x00, 0xa0, 0x2f, 0x00, 0x60, 0x17, 0x00,
		0xb0, 0x2a, 0x00, 0x50, 0x55, 0x00
	};
#define error_width 20
#define error_height 20
	static char error_bits[] = {
		0x00, 0x00, 0x00, 0xc0, 0x0f, 0x00, 0xf0, 0x3a, 0x00, 0x58, 0x55, 0x00,
		0x2c, 0xa0, 0x00, 0x56, 0x40, 0x01, 0xaa, 0x80, 0x02, 0x46, 0x81, 0x01,
		0x8a, 0x82, 0x02, 0x06, 0x85, 0x01, 0x0a, 0x8a, 0x02, 0x06, 0x94, 0x01,
		0x0a, 0xe8, 0x02, 0x14, 0x50, 0x01, 0x28, 0xb0, 0x00, 0xd0, 0x5f, 0x00,
		0xa0, 0x2a, 0x00, 0x40, 0x15, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};
/* XPM */
static char *ptrouge[] = {
/* width height num_colors chars_per_pixel */
"    14    14       16            1",
/* colors */
". c #ff8cd6", "# c #ff317b", "a c #ff105a", "b c #d60842", "c s none", "d c #ce637b",
"e c #ce426b", "f c #c67b94", "g c #ad0031", "h c #9c4a63", "i c #9c0029", "j c #8c314a",
"k c #731031", "l c #5a0018", "m c #4a0018", "n c #4a0010",
/* pixels */
"cccccccccccccc",
"cccccccccccccc",
"ccccfebbefcccc",
"cccdbaaabbdccc",
"ccfbaa#abbifcc",
"cceaa..abgikcc",
"ccbaa#abgiilcc",
"ccbbbbbgiillcc",
"cceggggiilmkcc",
"ccfiiiiklnmhcc",
"cccdlllmmmjccc",
"ccccfkllkhcccc",
"cccccccccccccc",
"cccccccccccccc"
};

	Window			root = RootWindowOfScreen (XtScreen (w));
	Display*		dpy = XtDisplay (w);
    XpmAttributes	xpmatt;
    XpmColorSymbol	xpmcol;

    xpmcol.name = "none";
    xpmcol.value = NULL;
	XtVaGetValues (w,
					XtNbackground, &xpmcol.pixel,
					NULL);
    xpmatt.colorsymbols = &xpmcol;
    xpmatt.numsymbols = 1;
    xpmatt.valuemask = XpmColorSymbols;
	if (XpmSuccess != XpmCreatePixmapFromData (dpy, root, ptrouge, &modifPixmap, NULL, &xpmatt))
		modifPixmap = XCreateBitmapFromData (dpy, root, modif_bits, modif_width, modif_height);

	QuestionPixmap = XCreateBitmapFromData (dpy, root, question_bits, question_width, question_height);
	ErrorPixmap = XCreateBitmapFromData (dpy, root, error_bits, error_width, error_height);
	OkPixmap = XCreateBitmapFromData (dpy, root, ok_bits, ok_width, ok_height);
	lockPixmap = XCreateBitmapFromData (dpy, root, lock_bits, lock_width, lock_height);
}

/*
 *	Fabrique la barre de menus.
 */

static void menu_post (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;
	int			item = (int) (long) call;
	int			n;
	Boolean		boolvar;
	FontInfo*	fp;

	if (MenuProc [item][0] == (XtActionProc) switchMode)
	{
		for (n = 1; n <= wx_numModes; n++)
			MenuTitle [item][n][0] = 1;
		MenuTitle [item][data -> modeNumber+1][0] = 2;
		return;
	}
	/*
	 *	Desactiver le toggle button.
	 */
	n = (int) (long) MenuProc [item] [0];
	if (n >= 100 && n <= 150) {
		int	def_fnt = -1;
		Boolean	found = False;

		XtVaGetValues (data -> active, XtNfont, &fp, NULL);
		for (n -= 102; n > 0; --n) {
			if (MenuTitle [item][n][0] == '-')
				continue;	/* laisser les separateurs */
			if (MenuTitle [item][n][0] != 1 && MenuTitle [item][n][0] != 2) {
				int	l = strlen (MenuTitle [item][n]);
				String	s = XtMalloc (l + 2);

				bcopy (MenuTitle [item][n], s+1, l+1);
				s [0] = 1;
				XtFree (MenuTitle [item][n]);
				MenuTitle [item][n] = s;
			}
			if (!strcmp ((char*) MenuProc [item][n], "default")) {
				def_fnt = n;
				continue;
			}
			if (fp == GetFont (XtDisplay (w),
					   (String)MenuProc [item][n],
					   True))
			{
				MenuTitle [item][n][0] = 2;
				found = True;
			} else
				MenuTitle [item][n][0] = 1;
		}
		/* fonte par defaut */
		if (def_fnt != -1)
			MenuTitle [item][def_fnt][0] = found ? 1 : 2;
		return;
	}
	for (; n > 0; --n) {
		if (MenuProc [item] [n] == ToggleOctalDisplayAction) {
			if (MenuTitle [item][n][0] != 3 && MenuTitle [item][n][0] != 4) {
				int	l = strlen (MenuTitle [item][n]);
				String	s = XtMalloc (l + 2);

				bcopy (MenuTitle [item][n], s+1, l+1);
				s [0] = 3;
				XtFree (MenuTitle [item][n]);
				MenuTitle [item][n] = s;
			}
			XtVaGetValues (data -> active, XtNbinaryEdit, &boolvar, NULL);
			MenuTitle [item][n][0] = boolvar ? 4 : 3;
			continue;
		}
		if (MenuProc [item] [n] == AutoWrapAction) {
			if (MenuTitle [item][n][0] != 3 && MenuTitle [item][n][0] != 4) {
				int	l = strlen (MenuTitle [item][n]);
				String	s = XtMalloc (l + 2);

				bcopy (MenuTitle [item][n], s+1, l+1);
				s [0] = 3;
				XtFree (MenuTitle [item][n]);
				MenuTitle [item][n] = s;
			}
			XtVaGetValues (data -> active, XtNwrap, &boolvar, NULL);
			MenuTitle [item][n][0] = boolvar ? 4 : 3;
			continue;
		}
		if (MenuProc [item] [n] == ShowLineNumberAction) {
			if (MenuTitle [item][n][0] != 3 && MenuTitle [item][n][0] != 4) {
				int	l = strlen (MenuTitle [item][n]);
				String	s = XtMalloc (l + 2);

				bcopy (MenuTitle [item][n], s+1, l+1);
				s [0] = 3;
				XtFree (MenuTitle [item][n]);
				MenuTitle [item][n] = s;
			}
			XtVaGetValues (data -> active, XtNshowLineNumber, &boolvar, NULL);
			MenuTitle [item][n][0] = boolvar ? 4 : 3;
			continue;
		}
		if (MenuProc [item] [n] == showRulerAction) {
			if (MenuTitle [item][n][0] != 3 && MenuTitle [item][n][0] != 4) {
				int	l = strlen (MenuTitle [item][n]);
				String	s = XtMalloc (l + 2);

				bcopy (MenuTitle [item][n], s+1, l+1);
				s [0] = 3;
				XtFree (MenuTitle [item][n]);
				MenuTitle [item][n] = s;
			}
			MenuTitle [item][n][0] = XtIsManaged (data -> ruler) ? 4 : 3;
			continue;
		}
		if (MenuProc [item] [n] == stripAction) {
			if (MenuTitle [item][n][0] != 3 && MenuTitle [item][n][0] != 4) {
				int	l = strlen (MenuTitle [item][n]);
				String	s = XtMalloc (l + 2);

				bcopy (MenuTitle [item][n], s+1, l+1);
				s [0] = 3;
				XtFree (MenuTitle [item][n]);
				MenuTitle [item][n] = s;
			}
			MenuTitle [item][n][0] = XtIsManaged (data -> menubar) ? 3 : 4;
			continue;
		}
	}
}

static void menu_select (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*		data = (WixWindow*) client;
	int				num = (int) (long) call;
	Cardinal		zero = 0;
	XEvent			event;
	int				item;
	XtActionProc	proc;

	item = num & 0xffff;
	num = (num >> 16) & 0xffff;

	if (MenuProc [num][0] == (XtActionProc) switchMode)
	{
		switchMode (item-1, data);
		return;
	}
	if ((long) MenuProc [num][0] >= 100 && (long) MenuProc [num][0] <= 150) {
		setFontFunction (data -> active, (String) MenuProc [num][item]);
		return;
	}
	if ((proc = MenuProc [num][item]) != NULL)
		(*proc) (data -> active, &event, NULL, &zero);
}

/*
 *	Appelle l'action associee a un bouton de commande.
 */

void commandCallback (Widget b, XtPointer client, XtPointer call)
{
	int				bn = (int) (long) client;
	Cardinal		zero = 0;
	XEvent			event;
	WixWindow*		data;
	XtActionProc	proc;

	XtVaGetValues (b, XtNuserData, &data, NULL);
	event.type = 0;
	proc = wx_modes [data -> modeNumber].m_buttons [bn].proc;

	if (proc)
		(*proc) (data -> active, &event, NULL, &zero);
}

/*
 *	Callback pour la coherence d'un wix et de son ombre.
 */

void verifyCallback (Widget w, XtPointer client, XtPointer call)
{
	static Boolean			locked = False;
	Widget					shadow = (Widget) client;
	VerifyCallbackStruct*	cb = (VerifyCallbackStruct*) call;
	long					pos;

	if (!locked) {
		locked = True;
		pos = WixGetInsertPosition (shadow);
		if (cb -> reason == INSERT_STRING) {
			WixReplace (shadow, cb -> pos, cb -> pos, cb -> string, cb -> length);
			if (cb -> pos < pos)
				pos += cb -> length;
		} else {
			WixReplace (shadow, cb -> pos, cb -> pos + cb -> length, NULL, 0);
			if (cb -> pos + cb -> length <= pos)
				pos -= cb -> length;
			else if (cb -> pos <= pos)
				pos = cb -> pos;
		}
		WixSetInsertPosition (shadow, pos);
		WixRedraw (shadow);
		locked = False;
	}
}

/*
 *	Callback pour la sauvegarde periodique.
 */

static void checkpointCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;
	char		tc [256];

	if (++data -> checkpoint > app_data.checkpointRate) {
		data -> checkpoint = 0;
		sprintf (tc, "%s##ckp", data -> filename);
		sprintf (tc, "Checkpointed: %d chars.",
				WixSaveFile (data -> active, tc));
		Message (data -> active, tc);
	}
}

/*
 *	Cree une nouvelle fenetre d'edition.
 */

static void rulerScroll (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;

	if (XtIsManaged (data -> ruler))
		RulerScroll (data -> ruler, WixGetHorScroll (data -> active));
}

/*
 *	WM_DELETE_WINDOW handler.
 */

static void wm_delete_callback (Widget w)
{
	Widget*		wl;
	WixWindow*	data = NULL;

	XtVaGetValues (w, XtNchildren, &wl, NULL);
	XtVaGetValues (wl [0], XtNuserData, &data, NULL);
	deleteWindowAction (data -> wix, NULL, NULL, NULL);
}

/*
 *	Rponses aux boutons du cadre de la fentre.
 */

static void close_callback (Widget w, XtPointer cd, XtPointer cl)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	deleteWindowAction (data -> wix, NULL, NULL, NULL);
}

static void iconify_callback (Widget w, XtPointer cd, XtPointer cl)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	XIconifyWindow (XtDisplay (data -> toplevel),
					XtWindow (data -> toplevel),
					XScreenNumberOfScreen (XtScreen (w)));
}

static void maximize_callback (Widget w, XtPointer cd, XtPointer cl)
{
	WixWindow*		data;
	XWindowChanges	xwc;

	xwc.height = HeightOfScreen (XtScreen (w)) - 4;
	xwc.y = 2;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	XReconfigureWMWindow (XtDisplay (data -> toplevel),
						  XtWindow (data -> toplevel),
						  XScreenNumberOfScreen (XtScreen (w)),
						  CWY|CWHeight, &xwc);
}

static void changeTabs (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;
	int*		cols = (int*) call;

	WixSetTabstops (data -> active, cols);
}

static void windowPopupAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	static char*	titles [64];
	WixWindow*		data;
	char*			p;
	char*			t;
	int				i = 0;

	titles [i++] = "\7Windows";
	titles [i++] = "=";
	for (data = _head; data != NULL; data = data -> next) {
		if (data -> free)
			continue;

		t = ((p = rindex (data -> filename, '/')) != 0)
				? ++p
				: data -> filename;
		if (t [0] == 0)
			t = "New File";
		/*
		 *	HACK:	on recopie le basename  la fin du tableau pour
		 *			pouvoir mettre juste avant l'indicateur de toggle button
		 */
		p = data -> filename + sizeof (data -> filename) - strlen (t) - 1;
		strcpy (p, t);
		*--p = (data -> wix == w) ? 2 : 1;
		titles [i++] = p;
		if (i >= 64-1)
			break;
	}
	titles [i] = NULL;

	XtVaSetValues (window_popup,	/* bidon: juste pour forcer un recalcul */
		XtNshadowThickness, 0,		/* de la taille du menu (comme on ne */
		NULL);						/* change pas le pointeur, title, il ne s'en */
	XtVaSetValues (window_popup,	/* rend pas compte). */
		XtNitem, 			titles,
		XtNshadowThickness, 1,
		NULL);

	PopMenu (w, (XtPointer) window_popup, (XtPointer) event);
}

static void window_menu_select (Widget w, XtPointer client, XtPointer call)
{
	int			item = (int) (long) call;
	int			i = 2;
	WixWindow*	data;
	/*
	 *	Chercher la fentre.
	 */
	for (data = _head; True; data = data -> next) {
		if (data -> free)
			continue;
		if (++i > item)
			break;
	}
	/*
	 *	La mettre au premier plan.
	 */
	XMapRaised (XtDisplay (data -> toplevel), XtWindow (data -> toplevel));
	warpPointer (data);
}

static void modePopupAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	if (wx_modes [data -> modeNumber].m_menu_names [0] == 0)
		return;

	XtVaSetValues (mode_popup,
		XtNitem,			wx_modes [data -> modeNumber].m_menu_names,
		XtNshadowThickness,	1,
		XtNuserData,		data,
		NULL);
	PopMenu (w, (XtPointer) mode_popup, (XtPointer) event);
}

static void mode_menu_select (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*		data;
	int				item = (int) (long) call;
	Cardinal		zero = 0;
	XEvent			event;
	XtActionProc	proc;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	if ((proc = (wx_modes [data -> modeNumber].m_menu_actions + 1) [item]) != NULL)
		(*proc) (data -> active, &event, NULL, &zero);
}

/*
 *	Cre un nouvelle fentre d'dition.
 */

WixWindow* newWindow (char* directory)
{
	static Boolean	default_show_line_number;
	Arg			arg [10];
	int			n = 0;
	int			i;
	static int	default_tab_equiv = 0;
	static int	fno = 0;
	int			cols [32];
	int			tab_equiv;
	int			lnw;
	Dimension	height;
	Boolean		sln;
	Widget		hsb;
	WixWindow*	data;
	FontInfo*	fp;
#if !ORIGINAL
	Widget		hrow;
#endif

	for (data = _head; data; data = data -> next)
		if (data -> free) {
			WixClearBuffer (data -> wix);
			removePaneAction (data -> wix, NULL, NULL, NULL);
			WixRedraw (data -> wix);

			if ((data -> stripped = app_data.stripWindow) == True)
				StripWindow (data);
			else
				DressWindow (data);

			XtVaSetValues (data -> wix,
				XtNtabsize,		default_tab_equiv,
				XtNshowLineNumber,	default_show_line_number,
				XtNfont,		default_font,
				NULL);
			XtVaGetValues (data -> wix,
				XtNlineNumberWidth,	&lnw,
				NULL);
			XtUnmanageChild (data -> ruler);
			fp = default_font;
			for (i = 0; i < (int) XtNumber (cols); i++)
				cols [i] = (i+1) * default_tab_equiv * fp -> f_info -> max_bounds.width;
			XtVaSetValues (data -> ruler,
				XtNscale,	fp -> f_info -> max_bounds.width,
				XtNleftOffset,	15 + lnw,
				NULL);
			RulerSetTabstops (data -> ruler, cols, XtNumber (cols));

			Message (data -> wix, NULL);
			data -> date = 0;
			data -> undo_mode = False;
			data -> undo_ptr = -1;
			data -> undo_number = -1;
			data -> prev_reason = -1;
			data -> filename [0] = 0;
			strcpy (data -> directory, directory);
			XtSetSensitive (data -> pane, True);

			XtVaSetValues (data -> frame,
							XtNtitle, "New File",
							NULL);

			XtVaSetValues (data -> toplevel,
					XtNtitle,		"New File",
					XtNiconName,	"New File",
					XtNiconic,		False,
					NULL);

			XtVaSetValues (XtParent (data -> wix),
					XtNpixmap, NULL,
					NULL);

			switchMode (0, data);	/* mode fondamental */
			return data;
		}
	/*
	 *	Le shell.
	 */
	data = (WixWindow*) XtMalloc (sizeof (WixWindow));
	/*
	 *	chainer la structure en tete de liste.
	 */
	data -> next = _head; _head = data;
	/*
	 *	initialiser le fichier et le directory courant.
	 */
	data -> filename [0] = 0;
	strcpy (data -> directory, directory);

	data -> toplevel = XtVaCreatePopupShell ("wixToplevel", applicationShellWidgetClass, ghost,
				XtNtitle, 		"New File",
				XtNiconName,	"New File",
				NULL);
	data -> free = True;
	data -> undo_mode = False;
	data -> undo_ptr = -1;
	data -> undo_number = -1;
	data -> prev_reason = -1;
	SetDeleteWindow (data -> toplevel, wm_delete_callback);
	XtAddEventHandler (data -> toplevel, 0, True, _XEditResCheckMessages, NULL);
	/*
	 *	Le cadre
	 */
	data -> frame = XtVaCreateWidget ("frame", frameWidgetClass, data -> toplevel,
											XtNuserData, data,
											XtNtitle, "New File",
											NULL);
	XtAddCallback (data -> frame, XtNcloseCallback, close_callback, NULL);
	XtAddCallback (data -> frame, XtNminCallback, iconify_callback, NULL);
	XtAddCallback (data -> frame, XtNmaxCallback, maximize_callback, NULL);
	/*
	 *	La paned window.
	 */
	data -> pane = XtVaCreateWidget ("pane", panedWindowWidgetClass, data -> frame,
					XtNuserData, data,
					NULL);
	/*
	 *	La barre de menus.
	 */
	data -> menubar = XtVaCreateWidget ("menubar", menuBarWidgetClass, data -> pane,
				XtNresizable,	False,
				XtNitem,	MenuTitle,
				XtNhelp,	True,
				NULL);
	XtAddCallback (data -> menubar, XtNselect, menu_select, (XtPointer) data);
	XtAddCallback (data -> menubar, XtNrelease, menu_post, (XtPointer) data);
	/*
	 *	les jolis boutons.
	 */
#if ORIGINAL
	data -> button_row = XtVaCreateWidget ("button_row", rowWidgetClass, data -> pane,
				XtNshadowThickness, 0,
				XtNresizable, False,
				NULL);
#endif
	for (i = 0; i < (int) XtNumber (data -> buttons); i++)
		data -> buttons [i] = NULL;
	/*
	 *	la regle.
	 */
	data -> ruler = XtVaCreateWidget ("ruler", rulerWidgetClass, data -> pane,
				XtNuserData,	data,
				XtNheight,	36,
				XtNresizable,	False,
				NULL);
	XtAddCallback (data -> ruler, XtNselect, changeTabs, (XtPointer) data);
#if !ORIGINAL
	hrow = XtVaCreateManagedWidget ("horiz_row", rowWidgetClass, data -> pane,
				XtNshadowThickness, 0,
				XtNresizable, True,
				NULL);
	data -> button_row = XtVaCreateWidget ("button_row", rowWidgetClass, hrow,
				XtNshadowThickness, 0,
				XtNresizable, False,
				XtNorientation, XpaVERTICAL,
				NULL);
#endif
	/*
	 *	Le wix.
	 */
	XtSetArg (arg [n], XtNlineNum, app_data.nline), n++;
	XtSetArg (arg [n], XtNcolumnNumber, app_data.ncol), n++;
	XtSetArg (arg [n], XtNuserData, data), n++;
	XtSetArg (arg [n], XtNkeyTable, wx_modes [0].m_keytable), n++;
	XtSetArg (arg [n], XtNtranslations, wx_modes [0].m_trans), n++;

#if ORIGINAL
	data -> wix = XtCreateScrolledWix (data -> pane, "edit", arg, n);
#else
	data -> wix = XtCreateScrolledWix (hrow, "edit", arg, n);
	XtVaSetValues (hrow,
					XtNtoResize,	XtParent (data -> wix),
					NULL);
#endif
	XtAddCallback (data -> wix, XtNundoRedoCallback, UndoRedoCallback, (XtPointer) data);
	XtAddCallback (data -> wix, XtNverifyModifyCallback, checkpointCallback, (XtPointer) data);
	XtAddCallback (data -> wix, XtNhScrollPositionCallback, rulerScroll, (XtPointer) data);
	XtVaGetValues (XtParent (data -> wix), XtNhScrollBar, &hsb, NULL);
	XtAddCallback (hsb, XtNsliderMoved, rulerScroll, (XtPointer) data);
	XtVaSetValues (XtParent (data -> wix), XtNuserData, data, NULL);
	/*
	 *	Mettre les tabulations sur la regle.
	 */
	XtVaGetValues (data -> wix,
		XtNtabsize,		&tab_equiv,
		XtNfont,		&fp,
		XtNshowLineNumber,	&sln,
		XtNlineNumberWidth,	&lnw,
		NULL);
	for (i = 0; i < (int) XtNumber (cols); i++)
		cols [i] = (i+1) * tab_equiv * fp -> f_info -> max_bounds.width;
	XtVaSetValues (data -> ruler,
		XtNscale,	fp -> f_info -> max_bounds.width,
		XtNleftOffset,	15 + lnw,
		NULL);
	RulerSetTabstops (data -> ruler, cols, XtNumber (cols));

	XtAddCallback (data -> wix, XtNbufferModifiedCallback, bufferModifiedCallback, (XtPointer) XtParent (data -> wix));
	XtAddCallback (data -> wix, XtNcontextCallback, ContextCallback, (XtPointer) data);
	XtAddCallback (XtParent (data -> wix), XtNselect, modifCallback, NULL);
	if (wx_modes -> m_background == (Pixel)-1) {
		XtVaGetValues (data -> wix,
				XtNtabsize,			&default_tab_equiv,
				XtNfont,			&default_font,
				XtNshowLineNumber,	&default_show_line_number,
				XtNbackground,		&wx_modes -> m_background,
				XtNcommentFont,		&wx_modes -> m_comment_font,
				XtNcppFont,			&wx_modes -> m_cpp_font,
				XtNstringFont,		&wx_modes -> m_string_font,
				XtNlibraryFont,		&wx_modes -> m_library_font,
				XtNconstantFont,	&wx_modes -> m_constant_font,
				XtNclassFont,		&wx_modes -> m_class_font,
				XtNkeywordFont,		&wx_modes -> m_keyword_font,
				XtNcommentColor,	&wx_modes -> m_comment_color,
				XtNcppColor,		&wx_modes -> m_cpp_color,
				XtNstringColor,		&wx_modes -> m_string_color,
				XtNlibraryColor,	&wx_modes -> m_library_color,
				XtNconstantColor,	&wx_modes -> m_constant_color,
				XtNclassColor,		&wx_modes -> m_class_color,
				XtNkeywordColor,	&wx_modes -> m_keyword_color,
				NULL);
		wx_modes -> m_fonts = default_font;
	}
	data -> shadow = NULL;
	data -> wix2 = NULL;
	data -> active = data -> wix;	/* un buffer */
	switchMode (0, data);	/* mode fondamental */

	WixRegisterMessage (data -> wix, showErrors);
	WixSetTabstops (data -> wix, cols);
	data -> stripped = app_data.stripWindow;
	if (data -> stripped)
		StripWindow (data);
	else
		DressWindow (data);
	XtManageChild (XtParent (data -> wix));
	XtManageChild (data -> pane);
	XtManageChild (data -> frame);

	XtSetKeyboardFocus (data -> toplevel, data -> frame);
	XtSetKeyboardFocus (data -> frame, data -> pane);
	XtSetKeyboardFocus (data -> pane, XtParent (data -> wix));
	XtSetKeyboardFocus (XtParent (data -> wix), data -> wix);


	data -> date = 0;
	data -> diag_list = NULL;
	data -> tag_list = NULL;
	data -> fs = NULL;
	data -> query_string = NULL;
	return data;
}

#if 0
static void newPane (char* directory, Widget parent)
{
	static Boolean	default_show_line_number;
	Arg			arg [10];
	int			n = 0;
	int			i;
	static int	default_tab_equiv = 0;
	static int	fno = 0;
	int			cols [32];
	int			tab_equiv;
	int			lnw;
	Dimension	height;
	Boolean		sln;
	Widget		hsb;
	WixWindow*	data;
	FontInfo*	fp;
#if !ORIGINAL
	Widget		hrow;
#endif
	/*
	 *	les diffrents flags.
	 */
	data = (WixWindow*) XtMalloc (sizeof (WixWindow));
	data -> filename [0] = 0;	/* initialiser le fichier et le directory courant */
	strcpy (data -> directory, directory);
	data -> free = True;
	data -> undo_mode = False;
	data -> undo_ptr = -1;
	data -> undo_number = -1;
	data -> prev_reason = -1;

	data -> toplevel = parent;
	/*
	 *	La paned window.
	 */
	data -> pane = XtVaCreateWidget ("pane", panedWindowWidgetClass, parent,
					XtNuserData, data,
					NULL);
	/*
	 *	La barre de menus.
	 */
	data -> menubar = XtVaCreateWidget ("menubar", menuBarWidgetClass, data -> pane,
				XtNresizable,	False,
				XtNitem,	MenuTitle,
				XtNhelp,	True,
				NULL);
	XtAddCallback (data -> menubar, XtNselect, menu_select, (XtPointer) data);
	XtAddCallback (data -> menubar, XtNrelease, menu_post, (XtPointer) data);
	/*
	 *	La fentre de dialogue pour experts.
	 */
	XtVaGetValues (data -> menubar, XtNheight, &height, NULL);
	data -> exp_row = XtVaCreateWidget ("expert_row", rowWidgetClass, data -> pane,
										XtNshadowThickness,	1,
										XtNresizable,		False,
										XtNpadding,			1,
										XtNheight,			height,
										NULL);
	/*
	 *	les jolis boutons.
	 */
#if ORIGINAL
	data -> button_row = XtVaCreateWidget ("button_row", rowWidgetClass, data -> pane,
				XtNshadowThickness, 0,
				XtNresizable, False,
				NULL);
#endif
	for (i = 0; i < (int) XtNumber (data -> buttons); i++)
		data -> buttons [i] = NULL;
	/*
	 *	la regle.
	 */
	data -> ruler = XtVaCreateWidget ("ruler", rulerWidgetClass, data -> pane,
				XtNuserData,	data,
				XtNheight,	36,
				XtNresizable,	False,
				NULL);
	XtAddCallback (data -> ruler, XtNselect, changeTabs, (XtPointer) data);
#if !ORIGINAL
	hrow = XtVaCreateManagedWidget ("horiz_row", rowWidgetClass, data -> pane,
				XtNshadowThickness, 0,
				XtNresizable, True,
				NULL);
	data -> button_row = XtVaCreateWidget ("button_row", rowWidgetClass, hrow,
				XtNshadowThickness, 0,
				XtNresizable, False,
				XtNorientation, XpaVERTICAL,
				NULL);
#endif
	/*
	 *	Le wix.
	 */
	XtSetArg (arg [n], XtNlineNum, app_data.nline), n++;
	XtSetArg (arg [n], XtNcolumnNumber, app_data.ncol), n++;
	XtSetArg (arg [n], XtNuserData, data), n++;
	XtSetArg (arg [n], XtNkeyTable, wx_modes [0].m_keytable), n++;
	XtSetArg (arg [n], XtNtranslations, wx_modes [0].m_trans), n++;

#if ORIGINAL
	data -> wix = XtCreateScrolledWix (data -> pane, "edit", arg, n);
#else
	data -> wix = XtCreateScrolledWix (hrow, "edit", arg, n);
	XtVaSetValues (hrow,
					XtNtoResize,	XtParent (data -> wix),
					NULL);
#endif
	XtAddCallback (data -> wix, XtNundoRedoCallback, UndoRedoCallback, (XtPointer) data);
	XtAddCallback (data -> wix, XtNverifyModifyCallback, checkpointCallback, (XtPointer) data);
	XtAddCallback (data -> wix, XtNhScrollPositionCallback, rulerScroll, (XtPointer) data);
	XtVaGetValues (XtParent (data -> wix), XtNhScrollBar, &hsb, NULL);
	XtAddCallback (hsb, XtNsliderMoved, rulerScroll, (XtPointer) data);
	XtVaSetValues (XtParent (data -> wix), XtNuserData, data, NULL);
	/*
	 *	Mettre les tabulations sur la regle.
	 */
	XtVaGetValues (data -> wix,
		XtNtabsize,		&tab_equiv,
		XtNfont,		&fp,
		XtNshowLineNumber,	&sln,
		XtNlineNumberWidth,	&lnw,
		NULL);
	for (i = 0; i < (int) XtNumber (cols); i++)
		cols [i] = (i+1) * tab_equiv * fp -> f_info -> max_bounds.width;
	XtVaSetValues (data -> ruler,
		XtNscale,	fp -> f_info -> max_bounds.width,
		XtNleftOffset,	15 + lnw,
		NULL);
	RulerSetTabstops (data -> ruler, cols, XtNumber (cols));

	XtAddCallback (data -> wix, XtNbufferModifiedCallback, bufferModifiedCallback, (XtPointer) XtParent (data -> wix));
	XtAddCallback (data -> wix, XtNcontextCallback, ContextCallback, (XtPointer) data);
	XtAddCallback (XtParent (data -> wix), XtNselect, modifCallback, NULL);
	if (wx_modes -> m_background == (Pixel)-1) {
		XtVaGetValues (data -> wix,
				XtNtabsize,			&default_tab_equiv,
				XtNfont,			&default_font,
				XtNshowLineNumber,	&default_show_line_number,
				XtNbackground,		&wx_modes -> m_background,
				XtNcommentFont,		&wx_modes -> m_comment_font,
				XtNcppFont,			&wx_modes -> m_cpp_font,
				XtNlibraryFont,		&wx_modes -> m_library_font,
				XtNconstantFont,	&wx_modes -> m_constant_font,
				XtNclassFont,		&wx_modes -> m_class_font,
				XtNkeywordFont,		&wx_modes -> m_keyword_font,
				XtNcommentColor,	&wx_modes -> m_comment_color,
				XtNcppColor,		&wx_modes -> m_cpp_color,
				XtNstringColor,		&wx_modes -> m_string_color,
				XtNlibraryColor,	&wx_modes -> m_library_color,
				XtNconstantColor,	&wx_modes -> m_constant_color,
				XtNclassColor,		&wx_modes -> m_class_color,
				XtNkeywordColor,	&wx_modes -> m_keyword_color,
				NULL);
		wx_modes -> m_fonts = default_font;
	}
	data -> shadow = NULL;
	switchMode (0, data);	/* mode fondamental */

	WixRegisterMessage (data -> wix, showErrors);
	WixSetTabstops (data -> wix, cols);
	data -> stripped = app_data.stripWindow;
	if (data -> stripped)
		StripWindow (data);
	else
		DressWindow (data);
	XtManageChild (XtParent (data -> wix));
	XtManageChild (data -> pane);

	XtSetKeyboardFocus (data -> pane, XtParent (data -> wix));
	XtSetKeyboardFocus (XtParent (data -> wix), data -> wix);

	data -> active = data -> wix;

	data -> date = 0;
	data -> diag_list = NULL;
	data -> tag_list = NULL;
	data -> fs = NULL;
	data -> exp_label = NULL;
	data -> exp_wix = NULL;
	data -> list_win = (Window) 0;
	data -> query_string = NULL;
}

void newWindow2 (char* directory)
{
	Widget	shell;
	Widget	pane;
	/*
	 *	Le shell.
	 */
	shell = XtVaCreatePopupShell ("wixToplevel", applicationShellWidgetClass, ghost,
				XtNtitle, 		"New File",
				XtNiconName,	"New File",
				NULL);
	SetDeleteWindow (shell, wm_delete_callback);
	/*
	 *	La paned window.
	 */
	pane = XtCreateWidget ("pane", panedWindowWidgetClass, shell, NULL, 0);
	newPane (directory, pane);
	newPane (directory, pane);
	XtManageChild (pane);
	XtPopup (shell, XtGrabNone);
}
#else
/*
 *	Coupe la fenetre en deux.
 */

void splitWindow2Action (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	char*		string;
	Boolean		m;

	XtVaGetValues (wix, XtNuserData, &data, NULL);

	if (data -> shadow || data -> wix2)
		return;		/* dj coupe */

	data -> wix2 = XtCreateScrolledWix (data -> pane, "edit2", NULL, 0);
	data -> active = data -> wix2;
	switchMode (data -> modeNumber, data);	/* met aussi les bonnes translations */
	WixRegisterMessage (data -> wix2, showErrors);
	XtVaSetValues (data -> wix2, XtNuserData, data, NULL);
	XtManageChild (XtParent (data -> wix2));
	XtSetKeyboardFocus (XtParent (data -> wix2), data -> wix2);
	XtSetKeyboardFocus (data -> pane, XtParent (data -> wix2));
}

/*
 *	Degage le deuxieme pane.
 */

void removePane2Action (Widget wix, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (wix, XtNuserData, &data, NULL);

	if (!data -> wix2)
		return;		/* ya rien  retirer */

	XtDestroyWidget (XtParent (data -> wix2));
	data -> wix2 = NULL;
	data -> active = data -> wix;
	XtSetKeyboardFocus (XtParent (data -> wix), data -> wix);
}

#endif

WixCommandRec	AppActionsList [] = {
#if 1
	{ "split-window2", splitWindow2Action, NULL, _help_8 },
	{ "unsplit-window2", removePane2Action, NULL, _help_9 },
#endif
	{ "adjust-paragraph", AdjustParagraphAction, NULL, "Adjust the current paragraph (TeX mode)." },
	{ "auto-wrap-mode", AutoWrapAction, NULL, _help_28 },
	{ "fast-backward-search", FastBackwardSearchAction, _backwardSearch, _help_12 },
	{ "backward-search", BackwardSearchAction, _backwardSearch, _help_12 },
	{ "begin-defun", BeginDefunAction, NULL, _help_35 },
	{ "change-directory", changeDirectoryAction, NULL, _help_6 },
	{ "change-mode-behavior", ChangeSubmodeAction, NULL, "Change the translations and alias table of the current mode using the ones of the specified mode." },
	{ "clear-buffer", clearBufferAction, NULL, "Erase the current buffer." },
	{ "compile-buffer", compileBufferAction, NULL, _help_21 },
	{ "create-window", createWindowAction, NULL, _help_7 },
	{ "delete-file", deleteFileAction, NULL, "Unlink the specified file." },
	{ "delete-window", deleteWindowAction, NULL, _help_5 },
	{ "end-defun", EndDefunAction, NULL, _help_36 },
	{ "eval-lisp-line", lispEvalAction, NULL, _help_32 },
	{ "exec-command", execCommandAction, NULL, _help_20 },
	{ "exec-named-macro", executeNamedMacroAction, NULL, _help_30 },
	{ "fast-forward-search", FastForwardSearchAction, _forwardSearch, _help_11 },
	{ "forward-search", ForwardSearchAction, _forwardSearch, _help_11 },
	{ "incremental-search", IncrementalSearchAction, NULL, "Search forward for a pattern in an incremental way: each change made to the pattern (insertion or deletion) triggers a new search from the position in the buffer where the search command was issued. The ^N (^P) keys search for the next (previous) occurrence of the pattern." },
	{ "indent-lisp-function", IndentLispFunctionAction, NULL, _help_38 },
	{ "indent-lisp-line", IndentLispLineAction, NULL, _help_37 },
	{ "insert-file", insertFileAction, NULL, _help_2 },
	{ "fast-insert-file", FastInsertFileAction, NULL, _help_2 },
	{ "line-counter", lineCounterAction, NULL, _help_18 },
	{ "lisp-new-line", NewLineIndentLispAction, NULL, _help_39 },
	{ "fast-load-file", FastLoadFileAction, NULL, _help_1 },
	{ "load-file", loadFileAction, NULL, _help_1 },
	{ "load-lisp-file", loadLispFileAction, NULL, _help_33 },
	{ "load-same-file", reloadFileAction, NULL, _help_24 },
	{ "next-window", nextWindowAction, NULL, _help_22 },
	{ "pop-help-box", HelpAction, NULL, _help_13 },
	{ "pop-mode-menu", modePopupAction, NULL, "Pop the menu specific to the current mode." },
	{ "pop-window-menu", windowPopupAction, NULL, "Navigate through the WiX windows." },
	{ "print-buffer", printBufferAction, NULL, "Print the file edited in the window (it is automatically saved) using the current mode print command." },
	{ "query-replace", queryReplaceAction, NULL, _help_14 },
	{ "quit-the-editor", quitAction, NULL, _help_17 },
	{ "restart-server", restartServerAction, NULL, _help_29 },
	{ "save-file", saveFileAction, NULL, _help_4 },
	{ "save-lisp-macro", SaveMacroAction, NULL, _help_34 },
	{ "select-pane", selectPaneAction, NULL, _help_10 },
	{ "set-tag-command", SetTagCommandAction, NULL, "Set the Unix command used to create a tag file for the current project." },
	{ "set-compile-command", setCompileCommandAction, NULL, _help_31 },
	{ "set-tab-equiv", setTabAction, NULL, _help_26 },
	{ "show-tags", ShowTagsAction, NULL, "show tags." },
	{ "show-kill-stack", showKillStackAction, NULL, _help_23 },
	{ "show-line-number", ShowLineNumberAction, NULL, _help_27 },
	{ "show-ruler", showRulerAction, NULL, _help_25 },
	{ "split-window", splitWindowAction, NULL, _help_8 },
	{ "start-macro-learn", startMacroAction, NULL, _help_15 },
	{ "strip-window", stripAction, NULL, "Strip (Dress) the window." },
	{ "stop-macro-learn", stopMacroAction, NULL, _help_16 },
	{ "toggle-case-sensitivity", ToggleCaseSensitivityAction, NULL, "Toggle the case and accent sensitivity of the search commands." },
	{ "toggle-editable", toggleEditableAction, NULL, _help_19 },
	{ "toggle-octal-display", ToggleOctalDisplayAction, NULL, "Toggle the display of the non ascii characters between font glyph and \\xxx representation." },
	{ "undo", UndoAction, NULL, "Undo the preceding commands." },
	{ "unmodify-buffer", UnmodifyBufferAction, NULL, "Clear the modified flag of the buffer." },
	{ "unsplit-window", removePaneAction, NULL, _help_9 },
	{ "fast-write-file", FastWriteFileAction, NULL, _help_3 },
	{ "write-file", writeFileAction, NULL, _help_3 },
	{ "wxrc-help", HelpConfigAction, NULL, "Show a commented .wxrc startup file." },
};
int	AppActionsNumber = XtNumber (AppActionsList);

static void propertyHandler (Display* dpy, Window root, XPropertyEvent* ev)
{
	extern void		SetTitle (WixWindow*, int);
	WixWindow*		data;
	Atom			type;
	unsigned long	nitems, left;
	int				format;
	Widget			w;
	remote*			r;
	char*			filename;

	if (XGetWindowProperty (dpy, root, ev -> atom, 0, 384, True, XA_STRING, &type, &format, &nitems, &left, (unsigned char**)&r) != Success || type != XA_STRING)
		return;

	/* Tester si le fichier n'est pas dja dit */
	if (r -> file [0]) {
		filename = AbsPath (r -> directory, r -> file);
		for (data = _head; data; data = data -> next) {
			if (data -> free)
				continue;
			if (!strcmp (filename, data -> filename)) {
					XMapRaised (XtDisplay (data -> toplevel), XtWindow (data -> toplevel));
					warpPointer (data);
					goto reste;
			}
		}
	}

	strcpy (theWindow -> directory, r -> directory);
	strcpy (theWindow -> filename, r -> file);

	w = theWindow -> wix;
	if (*r -> geometry)
		XtVaSetValues (theWindow -> toplevel,
						XtNgeometry,	r -> geometry,
						NULL);
	/*
	 *	Prevenir wx quand on ferme la fenetre.
	 */
	if (*r -> client_pid)
		strcpy (theWindow -> client_pid, r -> client_pid);
	/*
	 *	Choisir la fonte.
	 */
	if (*r -> font)
		setFontFunction (w, r -> font);
	/*
	 *	Afficher le selecteur de fichier.
	 */
	if (!strcmp (r -> file, "+fsel"))
	{
		data = theWindow;
		createWindowCallback (NULL, HomeDirectory, NULL);
		XtFree ((char*) r);
		loadFileAsk (data);
		return;
	}
	/*
	 *	Charger le fichier.
	 */
	if (r -> file [0] && !loadFileFunction (w, filename)) {
		strcpy (theWindow -> filename, filename);
		switchMode (SuffixToMode (theWindow), theWindow);
		SetTitle (theWindow, 0);
	}
	createWindowCallback (NULL, HomeDirectory, NULL);

	/* nouvelle interface avec liste de fichiers  diter */
reste:
	if (nitems == sizeof (remote)) {
		int		i, n;
		char	files [20][32];

		n = sscanf (r -> files, "%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
					files [0], files [1], files [2], files [3], files [4],
					files [5], files [6], files [7], files [8], files [9],
					files [10], files [11], files [12], files [13], files [14],
					files [15], files [16], files [17], files [18], files [19]);
		for (i = 0; i < n; ++i) {
			filename = AbsPath (r -> directory, files [i]);
			for (data = _head; data != 0; data = data -> next) {
				if (data -> free)
					continue;
				if (!strcmp (filename, data -> filename)) {
					XMapRaised (XtDisplay (data -> toplevel), XtWindow (data -> toplevel));
					break;
				}
			}
			if (data != 0)
				continue;

			strcpy (theWindow -> directory, r -> directory);
			strcpy (theWindow -> filename, files [i]);

			if (!loadFileFunction (theWindow -> wix, filename)) {
				strcpy (theWindow -> filename, filename);
				switchMode (SuffixToMode (theWindow), theWindow);
				SetTitle (theWindow, 0);
			}
			createWindowCallback (NULL, HomeDirectory, NULL);
		}
	}
	XtFree ((char*) r);
}

/*
 *	Fait taire la toolkit
 */

static void quiet (char* msg)
{
	return;	/* does nothing at all */
}

/*
 *	Traite les erreurs fatales (termine 'proprement' le serveur)
 */

static int fatal_error_handler (Display* dpy)
{
fprintf (stderr, "fatal_error_handler\n");
	_SetSnapshot (".wx_state", False);
	exit (0);
}

/*
 *	Fabrique le nom de fichier temporaire.
 */

static void makeSaveName (char* name, char* full)
{
	static Boolean	first = True;
	static char		sv_dir [] = ".wx_recover/";

	if (first) {
		first = False;
		mkdir (sv_dir, 0777);	/* creer le directory de sauvegarde */
	}
	strcpy (full, sv_dir);
	/* s'assurer qu'il y a un slash au debut */
	strcpy (full + sizeof (sv_dir) - 1 - (*name == '/'), name);
	/* traduire les slashes par des points d'exclamation */
	for (full += sizeof (sv_dir); *full; full++)
		if (*full == '/')
			*full = '!';
}

static void trap_handler (int signo)
{
	WixWindow*	data;
	Boolean		m;
	int			n;
	char		tc [256];

	for (data = _head; data; data = data -> next) {
		if (data -> free)
			continue;
		XtVaGetValues (data -> wix, XtNbufferModified, &m, NULL);
		if (!m)
			continue;
		makeSaveName (data -> filename, tc);
		n = WixSaveFile (data -> wix, tc);
		fprintf (stderr, "Save %d chars in: ~/%s\n", n, tc);
	}
#if 1
	/* si c'est moi, faire un core dump pour savoir ce qu'il s'est pass */
	if (100 == getuid ()) {
		struct passwd*	pwd;

		pwd = getpwuid (100);
		if (pwd != NULL && !strcmp ("pa", pwd -> pw_name)) {
			struct rlimit	rlim;

			rlim.rlim_cur = RLIM_INFINITY;
			rlim.rlim_max = RLIM_INFINITY;
			/* ne pas/plus limiter la taille d'un core dump */
			setrlimit (RLIMIT_CORE, &rlim);
			/* sauver la config actuelle */
			_SetSnapshot (".wx_state", False);
			/* aller dans /tmp et faire le coredump */
			chdir ("/tmp");
			kill (getpid (), SIGQUIT);
			exit (1);	/* non atteint */
		}
	} else
#endif
	quitCont (theWindow);	/* termine le process */
}

static char usage [] = "Usage:\n\
\n\
    wxd	[-help] [-h]		   print out this message\n\
	[-fs] [-fontsize] size	   size of the fons (default 14 points)\n\
	[-g] [-geometry] +-X+-Y	   location of the copyright button\n\
	[-ncol num]		   number of text columns\n\
	[-nline num]		   number of text lines\n\
	[-display display_name]	   X server to contact\n\
	[-f \033[4mfile\033[m]		   use \033[4mfile\033[m as startup file instead of ~/.wxrc\n\
	[-r]			   restore all the windows\n\
	[+r]			   start without any opened window\n\
	[-ckp \033[4mnumber\033[m]		   checkpoint every \033[4mnumber\033[m commands\n\
	[toolkit options ...]	   any standard toolkit options\n\
";

int main (int argc, char** argv)
{
	Widget			toplevel, copyright;
	XtActionsRec*	actions;
	register int	i;
	char			tc [64], host [32];
	Window			root;
	Atom			A_FILE;
	XtAppContext	app;
	Display*		dpy;
	String			display_name = NULL;
	int				depth = 1;
	int				fontsize = 14;
	char**			fbp;
	char*			cp;
	char*			fallback [256];
	char			fb_buffer [4096];
	static XrmOptionDescRec options [] = {
		{ "-ckp", "*checkpoint",   XrmoptionSepArg, NULL },
#if 0
		{ "-g" ,  "WIX.geometry",   XrmoptionSepArg, NULL },
		{ "-geometry", "WIX.geometry", XrmoptionSepArg, NULL },
#else
		{ "-g" ,  "*WixCopyright.geometry",   XrmoptionSepArg, NULL },
		{ "-geometry", "*WixCopyright.geometry", XrmoptionSepArg, NULL },
#endif
		{ "-ncol",   "*columnNumber",   XrmoptionSepArg, NULL },
		{ "-nline",   "*lineNumber",   XrmoptionSepArg, NULL },
		{ "-f",   "*configFile",   XrmoptionSepArg, NULL },
		{ "-r",   "*restoreState", XrmoptionNoArg, "True" },
		{ "+r",   "*restoreState", XrmoptionNoArg, "False" },
		{ "-strip",   "*stripWindow", XrmoptionNoArg, "True" },
		{ "+strip",   "*stripWindow", XrmoptionNoArg, "False" },
	};
	static XtResource resources [] = {
		{ "columnNumber", "ColumnNumber", XtRInt, sizeof (int),
		  XtOffset (AppDataPtr, ncol), XtRImmediate, (XtPointer) 80
		},
		{ "lineNumber", "LineNumber", XtRInt, sizeof (int),
		  XtOffset (AppDataPtr, nline), XtRImmediate, (XtPointer) 24
		},
		{ "checkpoint", "Checkpoint", XtRInt, sizeof (int),
		  XtOffset (AppDataPtr, checkpointRate), XtRImmediate, (XtPointer) 1234567890
		},
		{ "restoreState", "RestoreState", XtRBoolean, sizeof (Boolean),
		  XtOffset (AppDataPtr, restoreState), XtRImmediate, (XtPointer) True
		},
		{ "configFile", "ConfigFile", XtRString, sizeof (String),
		  XtOffset (AppDataPtr, configFile), XtRImmediate, (XtPointer) NULL
		},
		{ "stripWindow", "StripWindow", XtRBoolean, sizeof (Boolean),
		  XtOffset (AppDataPtr, stripWindow), XtRImmediate, (XtPointer) False
		},
	};
	static String bw_fallback [] = {
		"*WixCopyright.geometry:	-0-0",
		"*WixCopyright.PushButton.tipMessage:		Create a new WiX window",
		"*WixCopyright*Ok.tipMessage:		Replace your startup file",
		"*WixCopyright*Apply.tipMessage:	Pop the new features box",
		"*WixCopyright*Cancel.tipMessage:	Keep your startup file",
		"*Scrollbar.sliderColor:		black",
		"*Scrollbar.width:		12",
		"*Scrollbar.height:		12",
		"*Scrollbar.shadowThickness:	1",
		"*shadowThickness:		2",
		"*borderWidth:			0",
		"*wixToplevel.borderWidth:	2",
		"*Ruler.leftOffset:		15",
		"*Ruler.hotColor:		black",
		"*font:					*new*bold-r*%d*iso8859-1",
		"*Wix.font:				*-courier-bold-r-normal--%d-*iso8859-1",
		"*Wix.keywordFont:		*-courier-bold-r-normal--%d-*iso8859-1",
		"*Wix.cppFont:			*-courier-bold-o-normal--%d-*iso8859-1",
		"*Wix.stringFont:			*-courier-medium-r-normal--%d-*iso8859-1",
		"*Wix.commentFont:		*-new*-bold-r-normal--%d-*iso8859-1",
		"*Wix.constantFont:		*-courier-bold-r-normal--%d-*iso8859-1",
		"*Wix.classFont:			*-courier-medium-r-normal--%d-*iso8859-1",
		"*Wix.commentColor:		black",
		"*Wix.cppColor:			black",
		"*Wix.stringColor:		black",
		"*Wix.keywordColor:		black",
		"*help.help-form*Wix.font:	*-times-bold-r-normal--%d-*iso8859-1",
		"*background:		white",
		"*foreground:		black",
		"*Icon.font:			*new*bold-r-normal*%d*iso8859-1",
		"*menubar.shadowThickness:	1",
		"*oblong:		true",
		"*menubar.margin:		1",
		"*menubar*background:	black",
		"*menubar*foreground:	white",
		"*menubar.acceleratorColor:	white",
		"*edit.showLineNumber:	True",
		"*edit.lineNumberColor:	black",
		"*confWix.boldFont:		*new*bold-i*%d*iso8859-1",
		"*configHelp*dismiss.label:		Close",
		"*ScrolledWindow.labelColor:	black",
		"*FileSelector.noObject:	true",
		NULL
	};
	static String color_fallback [] = {
		/* la bote de copyright */
		"*WixCopyright.geometry:			-0-0",
		"*WixCopyright.PushButton.tipMessage:		Create a new WiX window",
		"*WixCopyright*Ok.tipMessage:		Replace your startup file",
		"*WixCopyright*Apply.tipMessage:	Pop the new features box",
		"*WixCopyright*Cancel.tipMessage:	Keep your startup file",
		"*WixCopyright*Query*background:	#eee",
		"*WixCopyright*Query*foreground:	#008",
		/* les Wix en gnral */
		"*Wix.font:						*-courier-medium-r-normal--%d-*iso8859-1",
		"*Wix.keywordFont:				*-courier-medium-r-normal--%d-*iso8859-1",
		"*Wix.libraryFont:					*-courier-medium-r-normal--%d-*iso8859-1",
		"*Wix.stringFont:					*helvetica-medium-r-normal--%d*iso8859-1",
		"*Wix.cppFont:					*-courier-medium-r-normal--%d-*iso8859-1",
		"*Wix.commentFont:				*helvetica-medium-r-normal--%d*iso8859-1",
		"*Wix.constantFont:				*-courier-medium-r-normal--%d-*iso8859-1",
		"*Wix.classFont:					*-courier-bold-r-normal--%d-*iso8859-1",
		"*Wix.tabsize:						4",
		"*Wix.foreground:					black",
		"*Wix.background:				#e0e0e0",
		/* la fentre d'dition */
		"*edit.showLineNumber:			True",
		"*edit.lineNumberColor:			maroon",
		"*edit.background:				#fffaf0",
		"*edit.shadowThickness:			0",
		"*edit.font:						*-courier-medium-r-normal--%d-*iso8859-1",
		"*edit.keywordFont:				*-courier-medium-r-normal--%d-*iso8859-1",
		"*edit.libraryFont:					*-courier-medium-r-normal--%d-*iso8859-1",
		"*edit.stringFont:					*helvetica-medium-r-normal--%d*iso8859-1",
		"*edit.cppFont:					*-courier-medium-r-normal--%d-*iso8859-1",
		"*edit.commentFont:				*helvetica-medium-r-normal--%d*iso8859-1",
		"*edit.constantFont:				*-courier-medium-r-normal--%d-*iso8859-1",
		"*edit.classFont:					*-courier-bold-r-normal--%d-*iso8859-1",
		"*edit.commentColor:				#00f",
		"*edit.cppColor:					#0b5",
		"*edit.constantColor:				#0ac",
		"*edit.classColor:					black",
		"*edit.keywordColor:				brown",
		"*edit.libraryColor:				#a4a",
		"*edit.stringColor:					#974",
		/* boite des tags */
		"*TextList.background:				wheat",
		"*TextList.selectionColor:			yellow",
		"*TextList.foreground:				black",
		/* frames */
		"*Frame.borderSize: 				8",
		"*Frame.font:						*-courier-bold-r-normal--%d-*iso8859-1",
		"*Frame.foreground:				#088",
		"*Frame.hilite:					cyan",
		/* boite d'aide gnrale */
		"*help.help-form*Wix.commentFont:	*new*bold-r-normal--%d*iso8859-1",
		"*help.help-form*Wix.cppColor:		red",
		"*help.help-form*Wix.cppFont:		*-new*-bold-i-normal--24-*iso8859-1",
		"*help.help-form*Wix.background: 	antiquewhite",
		/* boite aide sur les nouvelles possibilits */
		"*changesWix.commentFont:		*new*bold-i-normal--%d*iso8859-1",
		"*changesWix.keywordFont:			*-courier-bold-r-normal--%d*iso8859-1",
		"*changesWix.font:				*new*bold-r-normal--%d*iso8859-1",
		"*changesWix.background:			antiquewhite",
		/* boite aide sur la configuration */
		"*confWix.commentFont:			*new*medium-i-normal--%d*iso8859-1",
		"*confWix.keywordFont:			*new*medium-r-normal--%d*iso8859-1",
		"*confWix.background:				antiquewhite",
		"*configHelp*dismiss.oblong:		true",
		"*configHelp*dismiss.label:			Close",
		"*configHelp*dismiss.foreground: 	yellow",
		/* gnralits */
		"*tipFont:						8x13",
		"*font:							*new*medium-r-normal--%d*iso8859-1",
		"*shadowThickness:				2",
		"*borderWidth:					0",
		"*highlightColor:					red",
		"*background:					#aa976d",
		"*foreground:						white",
		"*wixToplevel.borderWidth:			2",
		/* les scrollbars */
		"*Scrollbar.width:					12",
		"*Scrollbar.height:				12",
		"*Scrollbar.shadowThickness:		1",
		"*ScrolledWindow.labelColor:		yellow",
		/* les taquets de tabulation */
		"*Ruler.leftOffset:					15",
		"*Ruler.hotColor:					red",
		/* la barre de menus */
		"*menubar.shadowThickness:		1",
		"*menubar.oblong:				true",
		"*menubar.margin:				1",
		"*menubar*foreground:			white",
		"*menubar.acceleratorColor:		yellow",
		/* les menus popup */
		"*PopupMenu.shadowThickness:		1",
		"*PopupMenu.oblong:				true",
		"*PopupMenu.margin:				1",
		"*PopupMenu.foreground:			white",
		"*PopupMenu.acceleratorColor:		yellow",
		/* les botes de dialogue */
		"*Query*oblong:					true",
		"*Query*foreground:				white",
		"*Query*Wix.foreground:			black",
		"*Query.shadowThickness:			0",
		/* le slecteur de fichiers */
		"*FileSelector*oblong:				true",
		"*FileSelector.noObject:			true",
		/* la bote  icones */
		"*Icon.foreground:				maroon",
		"*Icon.background:				antiquewhite",
		"*Icon.font:						*helvetica-medium-r-normal--%d*iso8859-1",
		NULL
	};

	if (argc == 1 || strcmp (argv [1], "-debug")) {
		extern char __version [];	/* import de menu.c */

		fprintf (stderr,
			"  WiX version %s, Copyright (C) 1989 - 2010 by Patrick Amar (pa@lri.fr)\n"
			"  WiX comes with ABSOLUTELY NO WARRANTY.\n"
			"  This is free software, and you are welcome to redistribute it\n"
			"  under certain conditions; for details see the General Public License.\n",
			__version);
		if (fork ())
			exit (0);
		setpgrp ();	/* pour ne pas se faire tuer par */
					/* un SIGINT ou autre ... */
	} else {
		debugMode = True;
		--argc;
		argv [1] = argv [0];
		argv++;
	}
	/*
	 *	Sur quel type d'ecran sommes nous?
	 */
	for (i = 1; i < argc; i++) {
		if (argv [i][1] == 'h' &&
			(!strcmp (argv [i], "-help") || !strcmp (argv [i], "-h")))
		{
			fprintf (stderr, usage);
			exit (0);
			break;
		}
		if (argv [i][1] == 'd' &&
			(!strcmp (argv [i], "-display") || !strcmp (argv [i], "-d")))
		{
			display_name = argv [i+1];
			break;
		}
		if (argv [i][1] == 'f' &&
			(!strcmp (argv [i], "-fontsize") || !strcmp (argv [i], "-fs")))
		{
			fontsize = atoi (argv [i+1]);
			break;
		}
	}

	if ((dpy = XOpenDisplay (display_name)) != NULL) {
		depth = DefaultDepthOfScreen (DefaultScreenOfDisplay (dpy));
		XCloseDisplay (dpy);
	}
#if defined (SYSV) || defined (__linux__)
	{
		struct utsname	uts_name;
		uname (&uts_name);

		strcpy (host, uts_name.nodename);
	}
#else
	gethostname (host, sizeof (host));
#endif
	/*
	 *	Fabriquer les bonnes ressources de fallback a partir du
	 *	type d'ecran (profondeur) et de la taille des fontes.
	 */
	fbp = (depth == 1) ? bw_fallback : color_fallback;
	cp = fb_buffer;
	for (i = 0; fbp [i]; i++) {
		fallback [i] = cp;
		sprintf (cp, fbp [i], fontsize);
		cp += strlen (cp) + 1;
	}
	fallback [i] = NULL;
	/* initialiser le tout */
	ghost = XtVaAppInitialize (&app, "WIX", options, XtNumber (options),
					&argc, argv, fallback, NULL);
	XtGetApplicationResources (ghost, &app_data, resources, XtNumber (resources), NULL, 0);
	dpy = XtDisplay (ghost);
#if 0
	if (debugMode) {
		XSynchronize (dpy, True);
	}
#endif
	toplevel = XtVaCreatePopupShell ("WixCopyright", overrideShellWidgetClass, ghost, NULL);
	XtAppSetWarningHandler (app, quiet);
	XSetIOErrorHandler (fatal_error_handler);

	root = DefaultRootWindow (dpy);
	A_FILE = XInternAtom (dpy, "WIX_FILE", False);
	A_HERE = XInternAtom (dpy, "WIX_SERVER", False);
	A_CLIENT = XInternAtom (dpy, "WIX_CLIENT", False);
	WatchCurs = XCreateFontCursor (dpy, XC_watch);
	TextCurs = XCreateFontCursor (dpy, XC_xterm);
	ArrowCurs = XCreateFontCursor (dpy, XC_top_left_arrow);

	listTrans = XtParseTranslationTable (listTranslations);
	HelpTrans = XtParseTranslationTable (helpTranslations);

	/*
	 *	Dupliquer la table d'actions
	 */
	actions = (XtActionsRec*) XtMalloc (AppActionsNumber * sizeof (XtActionsRec));
	for (i = 0; i < AppActionsNumber; i++) {
		actions [i].string = AppActionsList [i].string;
		actions [i].proc = AppActionsList [i].proc;
	}
	XtAppAddActions (app, actions, i);
	/*
	 *	Charger le fichier de configuration.
	 */
	strcpy (HomeDirectory, getenv ("HOME"));
	if (app_data.configFile)
		strcpy (tc, app_data.configFile);
	else
		sprintf (tc, "%s/.wxrc", HomeDirectory);
	lisp_initialize ();	/* doit tre avant Restart */
	Restart (tc, toplevel, argv [0]);	/* nous laisse dans notre home directory */
	sprintf (tc, "WiX(%s) \251 Patrick Amar @ LRI.", host);
	copyright = XtVaCreateManagedWidget (tc, pushButtonWidgetClass, toplevel,
						XtNlabel, tc,
						NULL);
	XtAddCallback (copyright, XtNrelease, createWindowCallback, HomeDirectory);
	/* Crer le menu popup de navigation dans les fentres */
	window_popup = CreatePopupMenu (ghost, "window-menu", NULL, 0);
	XtAddCallback (window_popup, XtNselect, window_menu_select, NULL);
	/* Crer le menu popup spcifique  chaque mode */
	mode_popup = CreatePopupMenu (ghost, "mode-menu", NULL, 0);
	XtAddCallback (mode_popup, XtNselect, mode_menu_select, NULL);
	/* Crer la premire fentre */
	theWindow = newWindow (HomeDirectory);
	createPixmaps (XtParent (theWindow -> wix));
#ifdef PRE_LOAD_ICONS
	_CreateFileSelector (theWindow);	/* pour avoir les icones prcharges */
#endif
	/*
	 *	Prevenir que l'on est l.
	 */
#ifndef NO_PROP
	XChangeProperty (dpy, root, A_HERE, XA_STRING, 8, PropModeReplace,
						(unsigned char*) host, strlen (host));
#endif
	/*
	 *	Attraper les SIGFPE, SIGBUS, SIGSEGV & SIGTERM.
	 */
	(void) signal (SIGBUS, trap_handler);
	(void) signal (SIGSEGV, trap_handler);
	(void) signal (SIGFPE, trap_handler);
	(void) signal (SIGTERM, trap_handler);
	/*
	 *	Boucle d'evenements.
	 */
	XSelectInput (dpy, root, PropertyChangeMask);
	if (app_data.restoreState)
		getSnapshot (".wx_state");	/* recharger le snapshot */
	XtRealizeWidget (toplevel);
	XtMapWidget (toplevel);

	for (;;) {
		XEvent	ev;

		XtAppNextEvent (app, &ev);
		if (ev.type == PropertyNotify &&
		    ev.xproperty.window == root &&
		    ev.xproperty.atom == A_FILE &&
		    ev.xproperty.state == PropertyNewValue)
			propertyHandler (dpy, root, (XPropertyEvent*) &ev);
		else
			XtDispatchEvent (&ev);
	}
}
