/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <utime.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/xpm.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/ScrolledWin.h>
#include <Xpa/Query.h>
#include <Xpa/PushButton.h>
#include <Xpa/MenuBar.h>
#include <Xpa/Wix.h>
#include <Xpa/abspath.h>
#include <sys/mman.h>

#include "wix.h"
#include "config.h"

char __version [] = "5.46";

/*************************************************************************
*
*	Tout pour les modes et les menus.
*
*************************************************************************/

#define skipSpace(cp)	while (*cp == ' ' || *cp == '\t') cp ++
#define findSpace(cp)	while (*cp != ' ' && *cp != '\t') cp ++
#define findSpaceOrNewline(cp)	while (*cp != ' ' && *cp != '\t' && *cp != '\n') cp ++
#define endLine(cp)	{ while (*cp != '\n') cp++; line++; }

WxMode		wx_modes [64];

int				wx_numModes = 1;
String**		MenuTitle;
XtActionProc**	MenuProc;

extern WixCommandRec	AppActionsList [];
extern int				AppActionsNumber;
extern WixCommandRec	WixActionsList [];
extern int				WixActionsNumber;

static XtActionProc		get_action_proc (char*);
static Trans*			stdKeyTable;
	   XtTranslations	stdTranslations;

/*
 *	Initialise le mode fondamental.
 */

static char defaultKeyTable [] =
	"^X^X	exchange-point-mark\n"
	"^X^B	toggle-octal-display\n"
	"^X^F	fast-load-file\n"
	"^X^P	scroll-up\n"
	"^X^N	scroll-down\n"
	"^X^Z	quit-the-editor\n"
	"^X=	line-counter\n"
	"^X#	insert-line-number\n"
	"^X^C	delete-window\n"
	"^X^D	change-directory\n"
	"^X^H	scroll-left\n"
	"^X^J	scroll-right\n"
	"^X^K	clear-buffer\n"
	"^X^R	load-file\n"
	"^X^V	load-same-file\n"
	"^X^W	write-file\n"
	"^X^I	insert-file\n"
	"^X^L	load-lisp-file\n"
	"^X^S	save-file\n"
	"^X^U	undo\n"
	"^X(		start-macro-learn\n"
	"^X)		stop-macro-learn\n"
	"^X^E	exec-macro\n"
	"^Z^T	toggle-editable\n"
	"^Z^C	toggle-case-sensitivity\n"
	"^Zk	show-kill-stack\n"
	"^ZK	pop-kill-stack\n"
	"^Z^H	strip-window\n"
	"^Z^L	show-line-number\n"
	"^Z^Z	restart-server\n"
	"^Z^W	create-window\n"
	"^Z^N	next-window\n"
	"^Z^R	show-ruler\n"
	"^Z^S	split-window\n"
	"^Z^U	unsplit-window\n"
	"^@		set-mark\n"
	"^A		beginning-of-line\n"
	"^B		backward-char\n"
	"^C		repeat-command\n"
	"^D		delete-char\n"
	"^E		end-of-line\n"
	"^F		forward-char\n"
	"^G		goto-line\n"
	"^H		rubout-char\n"
	"^I		tabulation\n"
	"^J		fast-forward-search\n"
	"^K		delete-end-line\n"
	"^L		center-curline\n"
	"^M		new-line\n"
	"^N		forward-line\n"
	"^O		open-space\n"
	"^P		backward-line\n"
	"^Q		insert-octal\n"
	"^R		delete-line\n"
	"^S		forward-search\n"
	"^T		twiddle-chars\n"
	"^U		multiplier\n"
	"^V		forward-page\n"
	"$E$D	forward-page\n"
	"^W		wipe-region\n"
	"^Y		yank\n"
	"^?		delete-char\n"
	"$E~	unmodify-buffer\n"
	"$Eb		backward-word\n"
	"$Ec		capitalize-word\n"
	"$Ed		delete-word\n"
	"$Ee		end-word\n"
	"$Ef		forward-word\n"
	"$Ek		copy-end-line\n"
	"$El		lower-case-word\n"
	"$Eo		open-line\n"
	"$Eu		upper-case-word\n"
	"$Ev		backward-page\n"
	"$E$U	backward-page\n"
	"$Ew	copy-region\n"
	"$Ey		incremental-yank\n"
	"$E^?	delete-word\n"
	"$E^H	rubout-word\n"
	"$E^I	indent-line-c\n"
	"$E1		multiplier\n"
	"$E2		multiplier\n"
	"$E3		multiplier\n"
	"$E4		multiplier\n"
	"$E5		multiplier\n"
	"$E6		multiplier\n"
	"$E7		multiplier\n"
	"$E8		multiplier\n"
	"$E9		multiplier\n"
	"$E<	beginning-of-buffer\n"
	"$E>	end-of-buffer\n"
	"$E%	query-replace\n"
	"$E$E	pop-help-box\n"
	"$L		backward-char\n"
	"$R		forward-char\n"
	"$U		backward-line\n"
	"$D		forward-line\n"
	"$E^R	curline-top\n"
	"$Es		forward-search\n"
	"$Er		backward-search\n";

static char defaultTranslations [] =
	"<Btn1Down>:		select-pane() selection-start()\n"
	"<Btn1Motion>:		extend-adjust()\n"
	"<Btn1Up>:			extend-end()\n"
	"<Btn2Down>:		stuff()\n"
	"~Shift<Btn3Down>:	scroll-window()\n"
	"~Shift<Btn3Motion>:	scroll-window()\n"
	"~Shift<Btn3Up>:		scroll-window()\n"
	"Ctrl<Btn4Down>:		scroll-down(8)\n"
	"Ctrl<Btn5Down>:		scroll-up(8)\n"
	"Shift<Btn4Down>:	scroll-down(4)\n"
	"Shift<Btn5Down>:	scroll-up(4)\n"
	"<Btn4Down>:		scroll-down(2)\n"
	"<Btn5Down>:		scroll-up(2)\n"
	"Ctrl<Key>Prior:		scroll-up()\n"
	"Ctrl<Key>Next:		scroll-down()\n"
	"<Key>Insert:			open-space()\n"
	"<Key>Home:			beginning-of-line()\n"
	"<Key>End:			end-of-line()\n"
	"<Key>Prior:			backward-page()\n"
	"<Key>Next:			forward-page()\n"
	"<Key>:				keyboard-action()";

static void initFundamental (void) {
	static Boolean	first = True;

	if (first) {
		first = False;

		WixSetActionProc (get_action_proc);	/* choisir les bonnes actions */
		stdKeyTable = CvtStringToKeyTable (defaultKeyTable);
		stdTranslations = XtParseTranslationTable (defaultTranslations);
	}
	bzero ((char*)wx_modes, sizeof (wx_modes));
	wx_modes -> m_name = "Fundamental";
	wx_modes -> m_background = -1;
	wx_modes -> m_comment_color = -1;
	wx_modes -> m_cpp_color = -1;
	wx_modes -> m_string_color = -1;
	wx_modes -> m_library_color = -1;
	wx_modes -> m_constant_color = -1;
	wx_modes -> m_class_color = -1;
	wx_modes -> m_keyword_color = -1;
	wx_modes -> m_str_delim = '\"';
	wx_modes -> m_keytable = stdKeyTable;
	wx_modes -> m_delim = " \t\n;:'`~}()[]=+-|!@*$%^&\\?>.<,\"";
	wx_modes -> m_wrap = False;
}

static Boolean ReMatch (char* re, char* s)
{
	char	tc [256];
	char*	p1 = tc;
	char*	p2 = tc;

	for (;;) {
		switch (*p1++ = *re++) {
			case 0:
				return FilenameMatch (p2, s);

			case ' ':
				/* sauter les espaces contigus */
				for (p1 [-1] = 0; *p1 == ' '; ++p1)
					;
				if (FilenameMatch (p2, s))
					return True;
				p2 = p1;
		}
	}
}

/*
 *	Recherche une expression rgulire (retourne la distance par
 *	rapport au dbut de la chane si a marche, ou -1 sinon)
 */

static int re_search (char* pattern, char* text, char* end)
{
	char*	begp;
	char*	start = text;
	char*	ok;
	char*	p;

	if (pa_re_comp (pattern))
		return -1;	/* invalid re */
	/*
	 *	Chercher, en remplacant momentanement les
	 *	\n par un \0 pour que chaque ligne soit
	 *	une string C.
	 */
	while (text < end) {
		for (p = text; *p != '\n'; p++);	/* trouver la fin de ligne */
		*p = 0;				/* mettre un \0 */
		ok = pa_re_exec (text, &begp);
		*p = '\n';
		if (ok != 0)
			return ok - start;
		text = p+1;			/* ligne suivante */
	}
	return -1;
}

/*
 *	Retourne le numro du mode correspondant au nom du fichier.
 *	( ce stade, le fichier est dj charg).
 */

int SuffixToMode (WixWindow* data)
{
	char*			p;
	unsigned char*	buffer = _GetCursorEnd (data -> wix);
	int				length = 1024;	/* ne regarder que sur le premier Koctet */
	int				i, n;
	int				dist;
	int				md;

	i = _GetBufferEnd (data -> wix) - buffer;
	if (length > i)
		length = i;
	/* regarder le suffixe en premier */
	for (i = 1; i < wx_numModes; i++)
		if ((p = wx_modes [i].m_suf) != 0 && ReMatch (p, data -> filename))
			return i;

	/* puis le mot magique si le suffixe ne va pas */
	dist = 2000;	/* un vision locale de l'infini (> 1024) */
	md = 0;			/* mode fondamental par dfaut */
	for (i = 1; i < wx_numModes; i++) {
		if ((p = wx_modes [i].m_magic) != 0 &&
			 (n = re_search (p, (char*)buffer, (char*) buffer + length)) != -1 &&
			 n < dist)
		{
			dist = n;	/* plus prs */
			md = i;		/* dans ce mode */
		}
	}
	return md;
}

/*
 *	Affiche un message dans la ligne de statut.
 */

void Message (Widget wix, char* msg)
{
	WixWindow*	data;

	XtVaGetValues (wix, XtNuserData, &data, NULL);

	if (msg == NULL) {
		if (data -> msg_cleared)
			return;
		data -> msg_cleared = True;
		data -> msg [0] = 0;
		SwSetLabel (XtParent (wix), NULL);
	} else {
		data -> msg_cleared = False;
		strncpy (data -> msg, msg, sizeof (data -> msg) - 1);
		SwSetLabel (XtParent (wix), data -> msg);
	}
}

/*
 *	Charge un fichier d'alias.
 */

static char* get_file (char* file, int* len)
{
	int		fd;
	char*	buffer;

	if (!file) {
		*len = 0;
		return NULL;
	}

	if ((fd = open (file, 0)) < 0) {
		fprintf (stderr, "Cannot read abbrev file %s.\n", file);
		*len = 0;
		return NULL;
	}

	*len = lseek (fd, 0L, 2);
	lseek (fd, 0L, 0);
	buffer = (char*) mmap (0, (size_t) *len, PROT_READ, MAP_SHARED, fd, 0);
	close (fd);
	return buffer;
}

/*
 *	Retrouve une action a partir de son nom.
 */

static XtActionProc get_action_proc (char* name)
{
	register int	i;

	/*
	 *	Les actions de l'application.
	 */
	for (i = 0; i < AppActionsNumber; i++)
		if (!strcmp (AppActionsList [i].string, name))
			return AppActionsList [i].proc;
	/*
	 *	les actions du widget.
	 */
	for (i = 0; i < WixActionsNumber; i++)
		if (!strcmp (WixActionsList [i].string, name))
			return WixActionsList [i].proc;
	/*
	 *	pas trouvee.
	 */
	return NULL;
}

/*
 *	Retrouve une fonction a partir de son adresse.
 */

char* GetFunctionByAddr (CommandProc fun)
{
	int	i;
	/*
	 *	Les functions de l'application.
	 */
	for (i = 0; i < AppActionsNumber; i++)
		if (AppActionsList [i].function == fun)
			return AppActionsList [i].string;
	/*
	 *	les functions du widget.
	 */
	for (i = 0; i < WixActionsNumber; i++)
		if (WixActionsList [i].function == fun)
			return WixActionsList [i].string;
	/*
	 *	pas trouve.
	 */
	return NULL;
}

static int	line;
static int	errors = 0;

/*
 *	Dfinition de menus.
 */

static int getMenu (String* p, char* title, String* names, String* functions)
{
	char*	cp = *p;
	int		itemNumber;

	endLine (cp); cp++;					/* aller au debut de la ligne */
	skipSpace (cp);						/* sur le premier car non blanc */
	for (itemNumber = 1; *cp != '}'; itemNumber++) {
		if (*cp == '-') {
			names [itemNumber] = "-";	/* sparateur simple */
			functions [itemNumber] = "sep";
		} else if (*cp == '=') {
			names [itemNumber] = "=";	/* sparateur double */
			functions [itemNumber] = "sep";
		} else {
			names [itemNumber] = cp;
			while (*cp != ',') cp ++;	/* chercher la prochaine virgule */
			*cp++ = 0;					/* terminer le nom de l'item */
			skipSpace (cp);				/* avancer sur le nom de l'action */
			functions [itemNumber] = cp;
		}
		endLine (cp);					/*  la fin du nom */
		*cp++ = 0;						/* aller au debut de la ligne suivante */
		skipSpace (cp);					/* sur le premier caractre non blanc */
	}
	names [0] = title;
	functions [0] = (String) (long) (itemNumber - 1);
	*p = cp;
	return itemNumber+1;
}

/*
 *	Copie un label de menu en vitant les tabulations multiples.
 */

static char* copy_label (char* s)
{
	char*	buffer = XtMalloc (strlen (s) + 1);
	char*	d = buffer;

	while (*d++ = *s) {
		if (*s++ == '\t')
			while (*s == '\t')
				s++;
	}
	return buffer;
}

#undef	MAX_KEYWORDS
#undef	COMMENT_COLOR
#undef	CPP_COLOR
#undef	KEYWORD_COLOR
#undef	LIBRARY_COLOR
#undef	STRING_COLOR
#undef	CONSTANT_COLOR
#undef	CLASS_COLOR

#define MAX_KEYWORDS	64

typedef struct {
	char*	keyword;
	char*	value;
	int		length;
} ConfigKeyWord;

static ConfigKeyWord	keyword_table [MAX_KEYWORDS];

/*
 *	Fonction de hash-code pour les mots clefs.
 */

static int hash (char* name, int n)
{
	int	h;

	for (h = 11; --n >= 0;)
		h = ((h * 17) ^ (int)(*name++)) % MAX_KEYWORDS;

	if (h < 0)
		h += MAX_KEYWORDS;
	return h;
}

/*
 *	Recherche un mot clef dans la table; si 'value' est != GET le met dans la table.
 */

#define GET	(char*) 1

static int intern_keyword (char* name, int length, char* value)
{
	int	h = hash (name, length);
	int	i;

	for (i = MAX_KEYWORDS; --i >= 0;) {
		if (keyword_table [h].keyword == NULL) {
			if (value == GET)
				return -1;	/* pas trouv */
			/* on le met dans la table */
			keyword_table [h].keyword = name;	/* pas besoin de rallouer */
			keyword_table [h].length = length;
			keyword_table [h].value = value;	/* non plus: on pointe directement */
			return h;							/* dans le buffer */
		}

		if (keyword_table [h].length == length && !strcmp (keyword_table [h].keyword, name))
		{
			if (value != GET)
				keyword_table [h].value = value;
			return h;
		}

		h = (h + 41) % MAX_KEYWORDS;	/* collision */
	}
	return -2;	/* plus de place dans la table */
}

/*
 *	Initialise la table des mots clefs.
 */

static char* keywords [] = {
#define ABBREV				0
	"abbrev:",
#define ALIAS				1
	"alias:",
#define AUTO_WRAP			2
	"auto_wrap:",
#define BACKGROUND			3
	"background:",
#define BEGIN_KEYWORD		4
	"begin_keyword:",
#define COMMENT_COLOR		5
	"comment_color:",
#define COMMENT_FONT		6
	"comment_font:",
#define COMMENT_START		7
	"comment_start:",
#define COMMENT_STOP		8
	"comment_stop:",
#define COMMENT_SUBMODE		9
	"comment_submode:",
#define COMPILE				10
	"compile:",
#define CPP_COLOR			11
	"cpp_color:",
#define CPP_FONT			12
	"cpp_font:",
#define DELIMITORS			13
	"delimitors:",
#define END_KEYWORD			14
	"end_keyword:",
#define ERRORS				15
	"errors:",
#define KEYTABLE			16
	"keytable:",
#define KEYWORD_COLOR		17
	"keyword_color:",
#define KEYWORD_FONT		18
	"keyword_font:",
#define KEYWORDS			19
	"keywords:",
#define LIBRARY_COLOR		20
	"library_color:",
#define LIBRARY_FONT		21
	"library_font:",
#define LINE_COMMENT		22
	"line_comment:",
#define MAGIC				23
	"magic:",
#define PARAGRAPH_DELIMITOR	24
	"paragraph_delimitor:",
#define PRINT				25
	"print:",
#define STANDARD_FONT		26
	"standard_font:",
#define STRING_COLOR		27
	"string_color:",
#define STRING_DELIMITOR	28
	"string_delimitor:",
#define STRING_FONT			29
	"string_font:",
#define STRING_SUBMODE		30
	"string_submode:",
#define SUFFIX				31
	"suffix:",
#define TAG_CMD				32
	"tag_cmd:",
#define TRANSLATIONS		33
	"translations:",
#define VERSION				34
	"version:",
#define CONSTANT_FONT		35
	"constant_font:",
#define CONSTANT_COLOR		36
	"constant_color:",
#define CLASS_FONT			37
	"class_font:",
#define CLASS_COLOR			38
	"class_color:",
};

static void init_keywords () {
	int	i;

	for (i = 0; i < XtNumber (keywords); ++i) 
		(void) intern_keyword (keywords [i], strlen (keywords [i]), 0);
}

/*
 *	Retourne la valeur du mot clef 'kw' (NULL si pas mentionn dans le fichier).
 */

static char* get_keyword (int k)
{
	char*	kw = keywords [k];
	/* intern_keyword trouve forcment 'kw' */
	return keyword_table [intern_keyword (kw, strlen (kw), GET)].value;
}

static FontInfo* get_font (Display* dpy, int k)
{
	char*	kw = get_keyword (k);

	return kw ? GetFont (dpy, kw, False) : NULL;
}


static Pixel get_color (Display* dpy, Colormap cmap, int k)
{
	char*	kw = get_keyword (k);
	XColor	xc;

	/* intern_keyword trouve forcment 'kw' */
	if (kw && XParseColor (dpy, cmap, kw, &xc) && XAllocColor (dpy, cmap, &xc))
		return xc.pixel;
	return -1;
}

static char* get_alloc_keyword (int k)
{
	char*	kw = get_keyword (k);
	/* intern_keyword trouve forcment 'kw' */
	return kw ? XtNewString (kw) : NULL;
}

/*
 *	Inspecte le fichier de configuration  la recherche de mots clefs valides;
 *	met la valeur dans la table et retoune vrai si ok, faux sinon.
 */

static Boolean getValue (char** p)
{
	char*	cp = *p;
	char*	keyword = cp;
	char*	value;
	int		length;

	findSpace (cp);
	*cp++ = 0;					/* terminer le mot clef */
	length = cp - keyword - 1;	/* sa longueur */

	skipSpace (cp);	/* sauter les espaces prcdant la valeur */
	value = cp;		/* la valeur */
	for (;; cp++)
		if (*cp == '\n') {
			line ++;
			if (cp [-1] == '\\')
			{
				if (cp [-2] == '\\') {
					--cp;
					break;	/* \\ en fin de ligne ==> \ */
				} else {
					cp [-1] = '\n';
					*cp = ' ';	/* \ en fin de ligne ==> \n espace */
					continue;
				}
			}
			break;
		}
		else if (*cp == 'n' && cp [-1] == '\\' && cp [1] == '\n') {
			cp [-1] = '\n';
			cp [0] = 0;
		}
	*cp = 0;	/* terminer la valeur */
	*p = cp;	/* passage par rfrence */
	length = intern_keyword (keyword, length, GET);
	if (length < 0)
		return False;

	keyword_table [length].value = value;
	return True;
}

static Boolean getBlock (String* p, char* keyword, String* pname)
{
	char*	cp = *p;
	int		len = strlen (keyword);

	if (bcmp (keyword, cp, len))
		return False;

	cp += len;			/* sauter le mot clef */
	skipSpace (cp);		/* et les espaces suivants */
	if (pname != NULL) {
		/* bloc nomm */
		*pname = cp;	/* le nom du bloc */
		findSpace (cp);	/* aller au bout */
		*cp++ = 0;		/* terminer le nom */
		skipSpace (cp);	/* sauter les espaces suivants */
	}
	*p = cp;
	return True;
}

/*
 *	Affiche les erreurs de syntaxe.
 */

static void synerr (char* file_name, char* str)
{
	errors ++;
	fprintf (stderr, "%s:%d\t%s\n", file_name, line, str);
}

#define INIT_VARS() \
	errors = 0; \
	for (i = 0; i < (int) XtNumber (button_pix); i++) button_pix [i] = NULL; \
	for (i = 0; i < (int) XtNumber (button_act); i++) button_act [i] = NULL; \
	for (i = 0; i < (int) XtNumber (button_tip); i++) button_tip [i] = NULL; \
	bn = 0; \
	specmenu_number = 0

/*
 *	Met les menus dans de la mmoire statique.
 */

static void commit_menus (XtActionProc** functions, int mnb, String** names)
{
	static String		helpTitle [5] = { "Help", "Commands", "Configuration", "New features", NULL };
	static XtActionProc	helpActions [5] = { 0, HelpAction, HelpConfigAction, HelpNewFeaturesAction, 0 };
	int		i;
	int		j;
	int		n;

	MenuTitle = (String**) XtMalloc ((mnb+3) * sizeof (String*));
	MenuProc = (XtActionProc**) XtMalloc ((mnb+3) * sizeof (XtActionProc*));
	for (i = 0; i < mnb; i++) {
		if ((char*)functions [i][0] == (char*)switchMode)
			n = wx_numModes;
		else
			n = (int) (long) functions [i][0];
		if (n >= 100 && n <= 150)
			n -= 102;	/* menu fontes */
		MenuTitle [i] = (String*) XtMalloc ((n+2) * sizeof (String));
		MenuProc [i] = (XtActionProc*) XtMalloc ((n+2) * sizeof (XtActionProc));
		/* titre du menu */
		MenuTitle [i][0] = XtNewString (names [i][0]);
		/* nombre d'items du menu */
		MenuProc [i][0] = functions [i][0];
		for (j = 1; j <= n; j++) {
			/* menu des modes */
			if ((void*)functions [i][0] == (void*)switchMode) {
				MenuTitle [i][j] = XtMalloc (2 + strlen (wx_modes [j-1].m_name));
				strcpy (MenuTitle [i][j] + 1, wx_modes [j-1].m_name);
				MenuTitle [i][j][0] = 1;
				continue;
			} else
			/* menu des fontes */
			if ((long)functions [i][0] >= 100 && (long)functions [i][0] <= 150) {
				MenuProc [i][j] = (XtActionProc) XtNewString ((char*)functions [i][j]);
				MenuTitle [i][j] = XtNewString (names [i][j]);
				continue;
			}
			/* menu standard */
			MenuTitle [i][j] = copy_label (names [i][j]);
			if (* names [i][j] == '-')
				MenuProc [i][j] = NULL;		/* sparateur (donc pas de fonction associe) */
			else
				MenuProc [i][j] = get_action_proc ((char*) functions [i][j]);
		}
		MenuProc [i][j] = NULL;
		MenuTitle [i][j] = NULL;
	}
	/* mettre ici le menu d'aide (doit tre  la fin) */
	MenuProc [i] = helpActions;
	MenuTitle [i] = helpTitle;
	/* la fin de la barre de menus */
	MenuProc [i+1] = NULL;
	MenuTitle [i+1] = NULL;
}

/*
 *	Recharge le fichier de configuration par defaut.
 */

static void _restart (int fd, char* _wxrc_name, Widget w)
{
	WxMode*	mp;
	int		i, n;
	int		_errors;
	int		bn;
	int		state;
	char*	name;
	char*	cp;
	char*	p;
	char	buffer [128000];
	char	buffer_aux [64000];
#define	STANDARD	0
#define	IN_MODE		1
#define	IN_MENU		2

#define EMPTY() (p_stack == stack)
#define PUSH(x,y) (p_stack -> file_name = _wxrc_name, p_stack -> cp = (x), p_stack -> line = (y), ++p_stack)
#define POP(x,y) (--p_stack, _wxrc_name = p_stack -> file_name, (x) = p_stack -> cp, (y) = p_stack -> line)
	struct Stack {
		char*	file_name;	/* nom du fichier inclu */
		char*	cp;			/* pointeur dans le buffer o se trouve le 'include' */
		int		line;		/* numro de la ligne du 'include' */
	}		stack [10],
			*p_stack = stack;

	String	button_pix [10], button_act [10], button_tip [10];
	int		menuNumber;
	String*	menuName [64];
	String	tstrings [64 * 20];
	XtActionProc*	menuFunction [64];
	XtActionProc	tprocs [64 * 20];

	String	specmenu_name [20];		/* les labels du menu spcifique du mode */
	String	specmenu_funs [20];		/* les actions du menu spcifique du mode */
	int		specmenu_number;		/* le nombre d'items du menu */

	Colormap	colormap = DefaultColormapOfScreen (XtScreen (w));
	Display*	dpy = XtDisplay (w);

retry:
	initFundamental ();	/* initialiser le mode fondamental */
	_errors = 0;
	menuNumber = 0;
	specmenu_number = 0;
	bn = 0;
	/*
	 *	Chercher le fichier de configuration.
	 *	sinon utiliser le defaut.
	 */
	switch (fd) {
	default:
		i = read (fd, buffer, sizeof (buffer) - 1);
		buffer [i] = 0;
		close (fd);
		if (i < sizeof (buffer) - 1)
			break;
		fprintf (stderr, "\nConfiguration file too large.");
		/* fall through */
	case -2:
		fprintf (stderr, "\nUsing builtin configuration file.\n");
		sprintf (buffer, "version: %s\n%s", __version, __def_config);
		break;
		
	case -1:
		fprintf (stderr, "\nConfiguration file created (~/.wxrc).\n");
		sprintf (buffer, "%s/.wxrc", HomeDirectory);
		fd = creat (buffer, 0644);
		sprintf (buffer, "version: %s\n%s", __version, __def_config);
		if (fd < 0)
			fprintf (stderr, "Cannot create %s, sorry...\n", buffer);
		else {
			/* crire la config par dfaut */
			write (fd, buffer, strlen (buffer));
			close (fd);
		}
		break;
	}
	cp = buffer;
	/*
	 *	Initialiser les tableaux pour les menus.
	 */
	menuName [0] = tstrings;
	menuFunction [0] = tprocs;

	wx_numModes = 1;
	/*
	 *	Interpreter le fichier.
	 */
	line = 0;
	INIT_VARS ();
	init_keywords ();
	state = STANDARD;
more:
	for (; *cp; cp++)
	switch (*cp) {
		/*
		 *	Sauter les espaces initiaux ou ligne vide.
		 */
		case ' ':
		case '\t':
			continue;
		case '\n':
			/* Commentaires en dbut de ligne */
			if (cp [1] == '#') {
				cp++;
				endLine (cp);
				--cp;	/* ringurgiter le \n */
				continue;
			}
			line ++;
			continue;
		/*
		 *	Commentaires en cours de ligne.
		 */
		case '!':
		case '#':
		nextline:
			endLine (cp);
			continue;

		default:
			/* inclusion de fichier (dans ou hors d'un mode) */
			if (!bcmp ("include", cp, 7)) {
				static char	iname [128];
				int			fd;
				int			cnt;

				cp += 7;	/* sauter le mot clef */
				skipSpace (cp);	/* les espaces suivants */
				for (p = iname; *cp != '\n'; )
					*p++ = *cp++;
				*p = 0;
				/* pas d'inclusion imbrique (pour l'instant) */
				if (!EMPTY ()) {
					fprintf (stderr, "%s:%d\tNested inclusion not allowed!\n", _wxrc_name, line);
					goto nextline;
				}
				/* fichier non lisible */
				if ((fd = open (iname, 0)) < 0) {
					fprintf (stderr, "%s:%d\tCannot include: %s\n", _wxrc_name, line, iname);
					goto nextline;
				}
				PUSH (cp, line);	/* empiler le contexte actuel */
				_wxrc_name = iname;
				line = 0;			/* dbut du fichier inclus */
				cp = buffer_aux - 1;	/* il ya un cp++ dans la boucle! */
				cnt = read (fd, buffer_aux, sizeof (buffer_aux) - 1);
				buffer_aux [cnt] = '\n';	/* s'assurer qu'un \n soit  la fin */
				buffer_aux [cnt+1] = 0;
				close (fd);
				continue;
			}
			/* directives autorises qu'en dehors d'un mode */
			if (state == STANDARD) {
				/* ignorer le numro de version (dj trait) */
				if (!bcmp ("version:", cp, 8))
					goto nextline;
				/*
				 *	Chargement des macros.
				 */
				if (!bcmp ("loadmacro", cp, 9)) {
					cp += 9;	/* sauter le mot clef */
					skipSpace (cp);	/* les espaces suivants */
					for (name = cp; *cp != '\n'; cp++) ;
					*cp = 0;
					if (2 != lisp_load_file (name))
						fprintf (stderr, "Cannot load macrofile: %s\n", name);
					*cp = '\n';
					goto nextline;
				}
				/*
				 *	Definition de menus.
				 */
				if (getBlock (&cp, "fonts", &name)) {
					if (*cp == '{') {
						i = getMenu (&cp, name, menuName [menuNumber], (String*) menuFunction [menuNumber]);
						menuFunction [menuNumber][0] = (XtActionProc) (long) (100 + i);
						/*
						 *	preparer le coup suivant.
						 */
						menuName [menuNumber+1] = menuName [menuNumber] + i;
						menuFunction [menuNumber+1] = menuFunction [menuNumber] + i;
						menuNumber ++;
					} else
						synerr (_wxrc_name, "Missing {");
					goto nextline;
				}
				if (getBlock (&cp, "defmenu", &name)) {
					if (*cp == '{') {
						i = getMenu (&cp, name, menuName [menuNumber], (String*) menuFunction [menuNumber]);
						/*
						 *	preparer le coup suivant.
						 */
						menuName [menuNumber+1] = menuName [menuNumber] + i;
						menuFunction [menuNumber+1] = menuFunction [menuNumber] + i;
						menuNumber ++;
					} else if (!bcmp (cp, "modes", 5)) {
						menuName [menuNumber][0] = name;
						menuFunction [menuNumber][0] = (XtActionProc) switchMode;
						/*
						 *	preparer le coup suivant.
						 */
						menuName [menuNumber+1] = menuName [menuNumber] + 2;
						menuFunction [menuNumber+1] = menuFunction [menuNumber] + 2;
						menuNumber ++;
					} else
						synerr (_wxrc_name, "Missing {");
					goto nextline;
				}
				/*
				 *	Definition de modes.
				 */
				if (getBlock (&cp, "defmode", &name)) {
					if (*cp == '{')
						state = IN_MODE;
					else
						synerr (_wxrc_name, "Missing {");
					goto nextline;
				}
				/*
				 *	Erreur!
				 */
				synerr (_wxrc_name, "only 'version', 'defmode', 'include', 'defmenu', 'loadmacro', or 'fonts' can be specified.");
				goto nextline;
			} else if (state == IN_MODE) {
				/*
				 *	Menu spcifique du mode.
				 */
				if (getBlock (&cp, "menu", NULL)) {
					if (*cp == '{')
						specmenu_number = getMenu (&cp, name, specmenu_name, specmenu_funs);
					else
						synerr (_wxrc_name, "Missing {");
					goto nextline;
				}
				/*
				 *	Barre de boutons
				 */
				if (!bcmp ("button:", cp, 7)) {
					if (bn >= (int) XtNumber (button_pix)) {
						synerr (_wxrc_name, "No more than 10 buttons per mode.");
						goto nextline;
					}
					cp += 7;
					skipSpace (cp);
					button_pix [bn] = cp;
					findSpace (cp);
					*cp++ = 0;
					skipSpace (cp);
					button_act [bn] = cp;
					findSpaceOrNewline (cp);
					if (*cp != '\n') {
						*cp++ = 0;
						skipSpace (cp);
						button_tip [bn] = cp;
					}
					++bn;
					endLine (cp);
					*cp = 0;
					continue;
				}
				/*
				 *	Divers switches.
				 */
				p = cp;
				if (!getValue (&cp)) {
					static Boolean	first = True;

					if (first) {
						first = False;
						n = 0;
						fprintf (stderr, "Keywords supported in a mode definition:\n\n");
						for (i = 0; i < XtNumber (keywords); ++i)
							fprintf (stderr, "%-20s%s", keywords [i],
													(++n % 3) == 0 ? "\n" : " ");
						if ((n % 3) != 0)
							fprintf (stderr, "\n\n");
					}
					fprintf (stderr, "%s:%d\tbad switch '%.30s' (mode '%s')\n",
									_wxrc_name, line, p, name);
				}
				continue;
			}
		/*
		 *	Fin de menu ou de mode.
		 */
		case '}':
			switch (state) {
			case STANDARD:
				synerr (_wxrc_name, "Unexpected }");
				break;
			case IN_MODE:
				_errors += errors;	/* accumuler les erreurs! */
				if (_errors == 0) {
					if (!strcmp (name, "Fundamental"))
						mp = wx_modes;
					else
						mp = & wx_modes [wx_numModes++];
					mp -> m_name = XtNewString (name);
					mp -> m_tagcmd = get_alloc_keyword (TAG_CMD);
					mp -> m_beg_keyword = get_alloc_keyword (BEGIN_KEYWORD);
					mp -> m_end_keyword = get_alloc_keyword (END_KEYWORD);
					mp -> m_start_comment = get_alloc_keyword (COMMENT_START);
					mp -> m_stop_comment = get_alloc_keyword (COMMENT_STOP);
					mp -> m_line_comment = get_alloc_keyword (LINE_COMMENT);

					p = get_keyword (STRING_DELIMITOR); mp -> m_str_delim = p ? p [0] : '\"';
					mp -> m_suf = get_alloc_keyword (SUFFIX);
					mp -> m_magic = get_alloc_keyword (MAGIC);

					p = get_keyword (AUTO_WRAP);
					mp -> m_wrap = p ? (0 == strcmp (p, "true")) : False;

					mp -> m_fonts = get_font (dpy, STANDARD_FONT);
					mp -> m_comment_font = get_font (dpy, COMMENT_FONT);
					mp -> m_cpp_font = get_font (dpy, CPP_FONT);
					mp -> m_string_font = get_font (dpy, STRING_FONT);
					mp -> m_library_font = get_font (dpy, LIBRARY_FONT);
					mp -> m_constant_font = get_font (dpy, CONSTANT_FONT);
					mp -> m_class_font = get_font (dpy, CLASS_FONT);
					mp -> m_keyword_font = get_font (dpy, KEYWORD_FONT);

					mp -> m_abbrev = get_file (get_keyword (ABBREV), &mp -> m_abbrev_len);

					p = get_alloc_keyword (STRING_SUBMODE); mp -> m_string_submode = p ? (long) p : -1;
					p = get_alloc_keyword (COMMENT_SUBMODE); mp -> m_comment_submode = p ? (long) p : -1;

					mp -> m_comment_color = get_color (dpy, colormap, COMMENT_COLOR);
					mp -> m_string_color = get_color (dpy, colormap, STRING_COLOR);
					mp -> m_library_color = get_color (dpy, colormap, LIBRARY_COLOR);
					mp -> m_constant_color = get_color (dpy, colormap, CONSTANT_COLOR);
					mp -> m_class_color = get_color (dpy, colormap, CLASS_COLOR);
					mp -> m_cpp_color = get_color (dpy, colormap, CPP_COLOR);
					mp -> m_keyword_color = get_color (dpy, colormap, KEYWORD_COLOR);
					mp -> m_background = get_color (dpy, colormap, BACKGROUND);

					mp -> m_par_delim = get_alloc_keyword (PARAGRAPH_DELIMITOR);
					mp -> m_print = get_alloc_keyword (PRINT);
					mp -> m_compile = get_alloc_keyword (COMPILE);
					mp -> m_errors = get_alloc_keyword (ERRORS);
					/*
					 *	Ajouter l'espace, la tabulation et le line-feed aux dlimiteurs.
					 */
					if (p = get_keyword (DELIMITORS)) {
						char tc [256];
						sprintf (tc, " \t\n%s", p);
						mp -> m_delim = XtNewString (tc);
					} else
						mp -> m_delim = XtNewString (" \t\n;:'`~}()[]=+-|!@*$%^&\\?>.<,\"");
					/* doit tre aprs les dlimiteurs pour les passer  CvtStringToKeyTable */
					mp -> m_keywords = CvtStringToKeyWordTable (
										get_keyword (KEYWORDS), mp -> m_delim);
					p = get_keyword (ALIAS); mp -> m_alias = p ? CvtStringToAlias (p) : NULL;
					mp -> m_string_trans = get_alloc_keyword (TRANSLATIONS);
					mp -> m_trans = mp -> m_string_trans
										? XtParseTranslationTable (mp -> m_string_trans)
										: NULL;
					/*
					 *	Les liaisons du mode 'Fundamental' servent de base aux
					 *	liaisons dans les autres modes (i.e. dfinir une liaison dans
					 *	le mode Fundamental la dfinit dans tous les autres modes).
					 */
					p = get_keyword (KEYTABLE);
					if (mp == wx_modes)		/* c'est le mode Fundamental */
						mp -> m_keytable = p
							? OverrideTranslation (CopyKeyTable (stdKeyTable), p)
							: CopyKeyTable (stdKeyTable);
					else					/* c'est un autre mode */
						mp -> m_keytable = p
							? OverrideTranslation (CopyKeyTable (wx_modes -> m_keytable), p)
							: CopyKeyTable (wx_modes -> m_keytable);
					/*
					 *	Le menu spcifique du mode.
					 */
					if (specmenu_number > 0) {
						mp -> m_menu_names [0] = XtMalloc (2 + strlen (mp -> m_name));
						mp -> m_menu_names [0][0] = 7;	/* titre du menu popup */
						strcpy (mp -> m_menu_names [0] + 1, mp -> m_name);
						mp -> m_menu_names [1] = "=";	/* un sparateur */
						for (i = 1; i < specmenu_number - 1; i++) {
							mp -> m_menu_names [i+1] = copy_label (specmenu_name [i]);
							mp -> m_menu_actions [i+2] = get_action_proc (specmenu_funs [i]);
						}
						mp -> m_menu_names [i+1] = NULL;
						mp -> m_menu_actions [i+2] = (XtActionProc) 0;
					} else
						mp -> m_menu_names [0] = NULL;	/* pas de menu du mode */
					/*
					 *	Les boutons de commande du mode.
					 */
					if (bn > 0) {
						for (i = 0; i < bn; i++) {
							Display*	dpy = XtDisplay (w);
							Screen*		screen = XtScreen (w);
							int			junk;
							unsigned int	ujunk;

							if (XpmSuccess != XpmReadPixmapFile (dpy,
										 RootWindowOfScreen (screen),
										 button_pix [i],
										 &mp -> m_buttons [i].pixmap,
										 NULL, NULL) &&
							    BitmapSuccess != XReadBitmapFile (dpy,
										 RootWindowOfScreen (screen),
										 button_pix [i],
										 &ujunk, &ujunk,
										 &mp -> m_buttons [i].pixmap,
										 &junk, &junk))
							{
								mp -> m_buttons [i].pixmap = (Pixmap)0;
								mp -> m_buttons [i].label = XtNewString (button_pix [i]);
							}
							mp -> m_buttons [i].proc = get_action_proc (button_act [i]);
							if (button_tip [i])
								mp -> m_buttons [i].tip = XtNewString (button_tip [i]);
						}
					}
				}
				INIT_VARS ();
				init_keywords ();
				break;
			}
			state = STANDARD;
			goto nextline;
		}
	/* si on est dans un fichier inclus */
	if (!EMPTY ()) {
		POP (cp, line);	/* restaurer le contexte */
		goto more;		/* et continuer avec le fichier original */
	}
	/*
	 *	Si on a trouv trop d'erreurs utiliser la configuration par dfaut.
	 */
	if (_errors > 0) {
		fd = -2;	/* utiliser la configuration par dfaut */
		goto retry;
	}
	/* mettre les numros des modes  la place de leur nom dans les 'submodes' */
	for (i = 0; i < wx_numModes; ++i) {
		char*	sm;
		mp = & wx_modes [i];
		/* sous mode dans une string */
		sm = (char*) (long) mp -> m_string_submode;
		mp -> m_string_submode = (mp -> m_string_submode == -1)
									? i	/* pas de sous mode => le mode lui mme */
									: _GetModeNumber (sm);
		if (mp -> m_string_submode == -1) {
			fprintf (stderr, "Unknown submode name: %s (mode: %s)\n", sm, mp -> m_name);
			mp -> m_string_submode = i;
		}
		/* sous mode dans un commentaire */
		sm = (char*) (long) mp -> m_comment_submode;
		mp -> m_comment_submode = (mp -> m_comment_submode == -1)
									? i	/* pas de sous mode => le mode lui mme */
									: _GetModeNumber (sm);
		if (mp -> m_comment_submode == -1) {
			fprintf (stderr, "Unknown submode name: %s (mode: %s)\n", sm, mp -> m_name);
			mp -> m_comment_submode = i;
		}
	}
	/* valider les menus */
	commit_menus (menuFunction, menuNumber, menuName);
}

/*
 *	Retourne le pathname complet du fichier.
 */

char	ProgName [256];

char* get_executable_path (char* name)
{
	Boolean		done = False;
	char		tc [256];
	char		path [2048];
	char*		p;
	char*		dir;

	if (*name != '/') {
		if (!(p = getenv ("PATH")))
			return NULL;

		strncpy (path, p, sizeof (path));
		for (dir = p = path; !done; p++) {
			if (*p == 0)
				done = True;
			else if (*p != ':')
				continue;

			*p = 0;
			if (dir [0] == 0)
				dir = ".";		/* si le PATH commence par : */

			sprintf (tc, "%s/%s", dir, name);
			if (0 == access (tc, X_OK))
				break;
			dir = p + 1;
		}
		if (done)
			return NULL;		/* pas trouv */
		name = tc;
	}
	if (name [0] != '/' && (dir = getenv ("cwd")) != NULL)
		strcpy (ProgName, AbsPath (dir, name));
	else
		strcpy (ProgName, name);
	return ProgName;
}

static char update_message [] = 
"This is a new version of WiX (%s), your startup file        \n\
may be obsolete (%s).\n\
\n\
You can:\n\
    Keep your startup file with the \"Keep\" button\n\
    (or with the ^G key),\n\
\n\
    Get an up to date startup file (the old one being\n\
    renamed \"~/.wxrc.old\") with the \"Replace\" button\n\
    (or with the Return key),\n\
\n\
    See the new features of this version with the\n\
    \"New Features\" button.";

/*
 *	Affiche la boite des nouveauts.
 */

static void showNewFeatures (WixWindow* data)
{
	data -> popped = True;	/* ne PAS virer la boite tout de suite! */
	data -> apply_after = showNewFeatures;
	HelpNewFeaturesAction ((Widget) NULL, (XEvent*) NULL, (String*) NULL, (Cardinal*) NULL);
	XRaiseWindow (XtDisplay (data -> query_string),
				  XtWindow (XtParent (data -> query_string)));
}

void Restart (char* file, Widget w, char* progname)
{
	int			fd;
	WixWindow	data;
	Widget		apply;
	char		version [128];
	char		buffer [128000];
	Boolean		ok = True;	/* bonne version */
	int			offset;

	(void) get_executable_path (progname);
	fd = open (file, 0);	/* rester dans le directory courant pour chercher le fichier */
							/* de config (option -f) */
	chdir (HomeDirectory);	/* aller chez soi aprs avoir rcupr la date du programme */
	if (fd >= 0) {
		buffer [read (fd, buffer, 1024)] = 0;
		if (1 == sscanf (buffer, "version: %s", &version)) {
			if (strcmp (version, __version)) {
				int	wv, wr;
				int	cv, cr;

				if (2 == sscanf (__version, "%d.%d", &wv, &wr) &&
					2 == sscanf (version, "%d.%d", &cv, &cr))
				{
					wv = wv * 1000 + wr;
					cv = cv * 1000 + cr;
					if (wv > cv)		/* prog plus rcent que config */
						ok = False;
					else if (cv > wv)	/* config plus rcente que prog */
						fprintf (stderr,
							"The current WiX version (%s) is older than the .wxrc version (%s).\n"
							"Some functions may be unknown.\n",
							__version, version);
				}
				else
					ok = False;		/* mauvais format de version */
			}
			/* pour rcrire un bon numro de version */
			for (offset = 0; buffer [offset++] != '\n'; )
				;
		} else {
			strcpy (version, "pre-5.40");
			offset = 0;				/* il n'y a mme pas de numro de version */
			ok = False;				/* donc on le rcrira au dbut */
		}
		close (fd);
	}
	if (ok) {
		fd = open (file, 0);	/* Chercher le fichier de configuration. */
		_restart (fd, file, w);	/* -1: create new one, -2: use builtin, otherwise ok */
	} else {
		char	tc [1024];

		/* faire une fausse structure WixWindow */
		data.toplevel = w;
		data.pane = NULL;
		data.what_after = NULL;
		data.apply_after = showNewFeatures;
		data.query_string = NULL;
		sprintf (tc, update_message, __version, version);
		/* Crer et afficher la boite de dialogue qui propose de changer le fichier de config */
		PopBye (&data, tc, True, "Replace", "Keep", "New Features");
		WaitForResponse (&data);
		XtDestroyWidget (XtParent (data.query_string));	/* a ferme aussi la boite !! */
		if (data.cancelled == False) {
			rename (file, ".wxrc.old");
			_restart (-1, file, w);	/* -1: create new one, -2: use builtin, otherwise ok */
		} else {
			int	length;

			fd = open (file, 0);
			length = read (fd, buffer, sizeof (buffer));
			close (fd);
			fd = creat (file, 0644);
			if (fd >= 0) {
				sprintf (tc, "version: %s\n", __version);
				write (fd, tc, strlen (tc));			/* crire le numro de version */
				write (fd, buffer + offset, length - offset); /* et le reste du fichier de config */
				close (fd);
			} else
				fprintf (stderr, "Wix: cannot update the .wxrc version number.\n");
			fd = open (file, 0);
			_restart (fd, file, w);	/* -1: create new one, -2: use builtin, otherwise ok */
		}
	}
}
