/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/PushButton.h>
#include <Xpa/Wix.h>
#include "wix.h"

/*
 *	Tout pour le mode Lisp.
 */

#define LP '('
#define RP ')'
#define DQUOTE '\"'

/*
 *	BeginDefunAction: trouve un debut de definition de fonction lisp.
 */

void BeginDefunAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	unsigned char*	e;

	e = _GetCursorEnd (w);
	
	if (_GetCursorBeg (w) [-1] == '\n' && !bcmp ("(def", e, 4))
		return;	/* on est deja au debut */

	if (WixBackwardSearch (w, "\n(def", 1))
		ForwardChar (w, 1, NULL);
	else if (!bcmp ("(def", _GetBufferBeg (w), 4))
		BeginningOfBuffer (w, 1, NULL);

	WixRedraw (w);
}

/*
 *	EndDefunAction: va a la fin de la fonction.
 */

void EndDefunAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	EndOfLine (w, 1, NULL);						/* aller a la fin */
	BeginDefunAction (w, NULL, NULL, NULL);		/* trouver le defun */
	find_close (w, '(', ')', False);			/* aller a la fin */
	WixRedraw (w);
}

/*
 *	bk_list: recherche la premiere lpar non matchee.
 */

static Boolean _def;

static Boolean bk_list (Widget w)
{
	for (;;) {
		if (!BackwardChar (w, 1, NULL))
			return False;	/* perdu */

		if (!_def)
			return True;

		if (_GetCursorEnd (w) [0] == DQUOTE)
			do {
				if (!BackwardChar (w, 1, NULL))
					return False;	/* encore perdu */
			} while (_GetCursorEnd (w) [0] != DQUOTE);

		if (_GetCursorEnd (w) [0] == LP) {
			_def = _GetCursorBeg (w) [-1] != '\n';
			return True;
		}
		if (_GetCursorEnd (w) [0] == RP)
			(void) bk_list (w);
	}
}

/*
 *	IndentLispLineAction: indente correctement une ligne.
 */

void IndentLispLineAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	unsigned char*	p;
	unsigned char*	e;
	int				ind;
	int				tab_equiv;
	FontInfo*		f;

	EnableRedisplay (w, False, True);
	/* aller en debut de ligne */
	BeginningOfLine (w, 1, NULL);
	/* effacer les espaces */
	for (p = _GetCursorEnd (w); *p == ' ' || *p == '\t'; p++)
		;
	removeChar (w, p - _GetCursorEnd (w), False);
	WixCreateMark (w, (void*) 125);		/* pour revenir inserer les espaces */

	_def = True;			/* pour les \n(de  */
	if (!bk_list (w)) {		/* la sexpr precedente */
		_def = True;
		WixGotoMark (w, (void*) 125);
		WixDeleteMark (w, (void*) 125);
		EnableRedisplay (w, True, True);
		return;
	}

	XtVaGetValues (w,
		XtNtabsize,	&tab_equiv,
		XtNfont,	&f,
		NULL);

	if (!bcmp("(def", _GetCursorEnd (w), 4) ||
	    !bcmp("(set", _GetCursorEnd (w), 4))
	{
		for (p = _GetCursorEnd (w); *p != ' ' && *p != '\t'; p++) ;
		ForwardChar (w, p - _GetCursorEnd (w), NULL);
		ind = WixGetColumn (w) + f -> f_width [' '];
	} else {
		ind = WixGetColumn (w) + 3 * f -> f_width ['x'];
		e = _GetBufferEnd (w);
		p = _GetCursorEnd (w);
		while (++p < e && *p != '\n' && *p != LP) ;
		ForwardChar (w, p - _GetCursorEnd (w), NULL);
		if (*p != '\n')
			ind = WixGetColumn (w);
	}
	_def = True;
	WixGotoMark (w, (void*) 125);
	WixDeleteMark (w, (void*) 125);
	InsertChar (w, ind / (tab_equiv * f -> f_width ['x']), (void*) '\t');
	InsertChar (w, (ind % (tab_equiv * f -> f_width ['x'])) / f -> f_width [' '], (void*) ' ');
	EnableRedisplay (w, True, True);
}

/*
 *	IndentLispFunctionAction: indente un fonction lisp.
 */

void IndentLispFunctionAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	int	fl, ll;

	WixCreateMark (w, (void*) 126);
	BeginDefunAction (w, NULL, NULL, NULL);		/* debut de la fonction */
	XtVaGetValues (w, XtNcurrentLineNumber, &fl, NULL);
	find_close (w, '(', ')', False);			/* aller a la fin */
	XtVaGetValues (w, XtNcurrentLineNumber, &ll, NULL);
	find_open (w, '(', ')', False);				/* aller au debut */
	for (ll -= fl; ll >= 0; --ll) {
		IndentLispLineAction (w, NULL, NULL, NULL);
		if (!ForwardLine (w, 1, NULL))
			break;
	}
	WixGotoMark (w, (void*) 126);
	WixDeleteMark (w, (void*) 126);
	WixRedraw (w);
}

/*
 *	Le retour chariot en mode lisp.
 */

void NewLineIndentLispAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	NewLine (w, 1, NULL);
	IndentLispLineAction (w, NULL, NULL, NULL);
}

/*
 *	Tout pour le mode TeX
 */

static int		wrap_column;

/*
 *	Cherche si on trouve le dlimiteur de commentaire dans la ligne
 */

static Boolean is_commented (char* end, char* lcmt, int lcsize)
{
	if (lcmt)
		for (; *end != '\n' && *end != 0; ++end)
			if (end [-1] != '\\' && *end == *lcmt && !bcmp (end+1, lcmt+1, lcsize-1))
				return True;
	return False;
}

/*
 *	Va au dbut du paragraphe courant.
 */

static void goto_beginning_of_paragraph (WixWindow* data)
{
	Widget	w = data -> active;
	char*	cur_beg = (char*) _GetCursorBeg (w);
	char*	buf_beg = (char*) _GetBufferBeg (w);
	char*	begp;
	int		ok;
	char*	cp;
	char*	p;
	Boolean	delim = (wx_modes [data -> modeNumber].m_par_delim != NULL);
	char*	lcmt = wx_modes [data -> modeNumber].m_line_comment;
	int		lcsize = strlen (lcmt);

	if (delim)
		delim = (pa_re_comp (wx_modes [data -> modeNumber].m_par_delim) == NULL);
	/*
	 *	Commencer par la ligne courante.
	 */
	cp = cur_beg; *cp = 0;
	while (*--cp != '\n') ;		/* dbut de la ligne courante */
	/* on est dja sur un dbut de paragraphe (ligne prcdente vide) */
	if (cp [1] == 0 && cp [-1] == '\n')
		return;
	/* dbut de paragraphe indiqu par le dlimiteur */
	if ((delim && pa_re_exec (cp+1, &begp)) || is_commented (cp+1, lcmt, lcsize))
	{
		BackwardChar (w, cur_beg - (cp+1), NULL);
		ForwardLine (w, 1, 0);
		return;
	}
	/*
	 *	Chercher, en remplacant momentanment les
	 *	\n par un \0 pour que chaque ligne soit une string C.
	 */
	while (cp > buf_beg) {
		p = cp;
		*p = 0;
		while (*--cp != '\n');	/* trouver la fin de la prcdente ligne */
		ok = cp [1] == 0 ||
			(delim && pa_re_exec (cp+1, &begp) != NULL) ||
			is_commented (cp+1, lcmt, lcsize);
		*p = '\n';
		if (ok) {
			BackwardChar (w, cur_beg - (cp+1), NULL);
			ForwardLine (w, 1, 0);
			return;
		}
	}
	BeginningOfBuffer (w, 1, 0);
}

/*
 *	Sommes nous  la fin d'un paragraphe ?
 */

static Boolean at_end_of_paragraph (WixWindow* data, char* end)
{
	int		ok;
	char*	begp;
	char*	cp;
	Widget	w = data -> active;
	Boolean	delim = (wx_modes [data -> modeNumber].m_par_delim != NULL);
	char*	lcmt = wx_modes [data -> modeNumber].m_line_comment;
	int		lcsize = strlen (lcmt);

	if (end >= (char*) _GetBufferEnd (w))
		return True;	/* fin de buffer donc de paragraphe */

	if (delim)
		delim = (pa_re_comp (wx_modes [data -> modeNumber].m_par_delim) == NULL);
	/* comme on est appell depuis une fin de ligne, on cherche au dbut de la suivante */
	for (cp = ++end; *cp != '\n';)
		++cp;

	cp [0] = 0;			/* (fin de ligne suivante) pour re_exec */
	ok = end [0] == 0 ||
		(delim && pa_re_exec (end, &begp) != NULL) ||
		is_commented (end, lcmt, lcsize);
	cp [0] = '\n';		/* remettre la fin de ligne */

	if (ok || lcsize == 0)
		return ok;

	/* chercher si on trouve le dlimiteur de commentaire en cours de route */
	return is_commented (end, lcmt, lcsize);
}

/*
 *	Sommes nous sur la ligne qui spare deux paragraphes?
 */

static Boolean between_paragraphs (WixWindow* data)
{
	Widget	w = data -> active;
	char*	e = (char*) _GetCursorEnd (w);
	char*	b = (char*) _GetCursorBeg (w);

	/* aller  la fin de la ligne prcdente */
	while ((*--b = *--e) != '\n')
		;
	/* et voir si on est en fin de paragraphe */
	return at_end_of_paragraph (data, e);
}

/*
 *	Retourne la valeur de l'indentation de la ligne courante.
 */

static void get_line_indent (Widget w, int* ntabs, int* nspaces)
{
	unsigned char*	end = _GetCursorEnd (w);
	unsigned char*	beg = _GetCursorBeg (w);
	int				t = 0;
	int				s = 0;

	/* aller au dbut de la ligne si on y est pas dj */
	if (beg [-1] != '\n') {
		while ((*--end = *--beg) != '\n')
			;
		++end;
	}
	for (;;)
		switch (*end++) {
			case '\t':	++t; continue;
			case ' ':	++s; continue;
			default:
				*ntabs = t;
				*nspaces = s;
				return;
		}
}

/*
 *	Ajuste un paragraphe.
 */

void AdjustParagraphAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	int			from, to;
	int			lnob, lnoa;
	int			nspaces, ntabs;
	char*		cp;
	FontInfo*	fp;
	static char	my_mark [] = "my_mark";

	XtVaGetValues (w,
		XtNuserData,	&data,
		XtNfont,		&fp,
		NULL);

	if (between_paragraphs (data))		/* il n'y a donc rien  faire */
		return;

	(void) EnableRedisplay (w, False, True);
	(void) WixCreateMark (w, my_mark);
	wrap_column = WixGetWidth (w) - Width ('x', fp);	/* provisoire */
	goto_beginning_of_paragraph (data);		/* va au dbut du paragraphe */
	from = WixGetY (w);						/* numro de la ligne dans la fentre */
	lnob = WixGetNumLines (w);				/* nombre de lignes du buffer */
	get_line_indent (w, &ntabs, &nspaces);	/* rcuprer l'indentation de la premire */
											/* ligne du paragraphe */
	/* ajuste toutes les lignes jusqu' la finde paragraphe */
	for (;;) {
		while (EndOfLine (w, 1, 0) && WixGetColumn (w) < wrap_column &&
				!at_end_of_paragraph (data, (char*) _GetCursorEnd (w)))
		{
			/* recolle la ligne courante avec la suivante */
			/* en sparant par un seul espace (vire les espaces en trop) */
			InsertChar (w, 1, (void*) ' ');
			for (cp = (char*) _GetCursorEnd (w);*++cp == ' ' || *cp == '\t'; )
				;
			removeChar (w, cp - (char*) _GetCursorEnd (w), False);
		}
		if (WixGetColumn (w) < wrap_column)
			break;	/* on a justifi tout le paragraphe */
		/* reculer jusqu'au dernier espace qui entre dans la largeur impose */
		do {
			for (cp = (char*) _GetCursorBeg (w);*--cp != ' ' && *cp != '\t'; )
				;
			while (*cp == ' ' || *cp == '\t')
				--cp;
			++cp;
			BackwardChar (w, (char*) _GetCursorBeg (w) - cp, 0);
		} while (WixGetColumn (w) >= wrap_column);
		removeChar (w, 1, False);	/* remplacer l'espace par une nouvelle */
		NewLine (w, 1, NULL);
		if (ntabs > 0)				/* ligne indente comme la premire */
			InsertChar (w, ntabs, (void*) '\t');
		if (nspaces > 0)
			InsertChar (w, nspaces, (void*) ' ');
	}
	to = WixGetY (w);				/* numro de la ligne dans la fentre de la fin de paragraphe */
	lnoa = WixGetNumLines (w);		/* nombre de lignes du buffer aprs repliage */
	WixGotoMark (w, my_mark);
	WixDeleteMark (w, my_mark);
	if (WixGetHorizontalScroll (w) != 0) {
		WixSetHorizontalScroll (w, 0);	/* s'assurer qu'on est pas scroll  gauche */
		WixSetRefreshArea (w, 0, 1000);
	} else
	if (lnoa == lnob)				/* s'il y a le mme nombre de lignes */
		WixSetRefreshArea (w, from, to);	/*on ne raffiche que le paragraphe */
	(void) EnableRedisplay (w, True, True);
}

/*
 *	Change de sous mode.
 */

static void changeSubmode (int submode, WixWindow* data)
{
	WxMode*	smp = & wx_modes [submode];
	char*	p;
	char	tc [256];

	if (submode == data -> submodeNumber)
		return;		/* on y est dj */

	data -> submodeNumber = submode;
	/* afficher le nom du mode et du sous mode si ncessaire */
	if (data -> wix2) {
		p = rindex (data -> filename, '/');
		p = p ? p+1 : data -> filename;
		sprintf (tc, "%s [%s]", wx_modes [data -> modeNumber].m_name, p);

		if (submode == data -> modeNumber)
			SwSetTitle (XtParent (data -> active), XtNewString (tc));
		else {
			sprintf (data -> mode_name, "(%s) %s", smp -> m_name, tc);
			SwSetTitle (XtParent (data -> active), data -> mode_name);
		}
	} else {
		if (submode == data -> modeNumber)
			SwSetTitle (XtParent (data -> active), wx_modes [data -> modeNumber].m_name);
		else {
			sprintf (data -> mode_name, "(%s) %s", smp -> m_name, wx_modes [data -> modeNumber].m_name);
			SwSetTitle (XtParent (data -> active), data -> mode_name);
		}
	}
	/* ne modifier que les attributs du sous mode */
	XtVaSetValues (data -> active,
		XtNkeyTable,			smp -> m_keytable,
		XtNtranslations,		stdTranslations,
		XtNaliasTable,			smp -> m_alias,
		XtNaliasBuffer,			smp -> m_abbrev,
		XtNaliasBufferLength,	smp -> m_abbrev_len,
		NULL);
	if (wx_modes -> m_trans)	/* ajouter les translations du mode Fundamental */
		XtOverrideTranslations (data -> active, wx_modes -> m_trans);
	if (smp -> m_trans)			/* puis celles du mode choisi */
		XtOverrideTranslations (data -> active, smp -> m_trans);
	/* pareil pour l'ventuelle deuxime fentre */
	if (data -> shadow) {
		Widget	other = data -> active == data -> wix
							? data -> shadow
							: data -> wix;

		XtVaSetValues (other,
			XtNkeyTable,			smp -> m_keytable,
			XtNtranslations,		stdTranslations,
			XtNaliasTable,			smp -> m_alias,
			XtNaliasBuffer,			smp -> m_abbrev,
			XtNaliasBufferLength,	smp -> m_abbrev_len,
			NULL);
		if (wx_modes -> m_trans)	/* ajouter les translations du mode Fundamental */
			XtOverrideTranslations (other, wx_modes -> m_trans);
		if (smp -> m_trans)			/* puis celles du mode choisi */
			XtOverrideTranslations (other, smp -> m_trans);
	}
}

/*
 *	Retourne le numro du mode,  partir de son nom.
 */

int _GetModeNumber (char* name)
{
	int	n;

	for (n = 0; n < wx_numModes; ++n)
		if (! strcmp (name, wx_modes [n].m_name))
			return n;
	return -1;
}

/*
 *	L'action de changement de sous mode (explicite).
 */

void ChangeSubmodeAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;
	int			n;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	/* restaure le comportement du mode */
	if (* num_params == 0) {
		changeSubmode (data -> modeNumber, data);
		return;
	}
	n = _GetModeNumber (params [0]);
	if (n >= 0)
		changeSubmode (n, data);
	else
		Message (w, "Unknown mode name.");
}

void ContextCallback (Widget w, XtPointer client, XtPointer call)
{
	WixWindow*	data = (WixWindow*) client;

	switch ((long)call) {
		case COMMENT_COLOR:
			changeSubmode (wx_modes [data -> modeNumber].m_comment_submode, data);
			return;

		case STRING_COLOR:
			changeSubmode (wx_modes [data -> modeNumber].m_string_submode, data);
			return;

		default:
			changeSubmode (data -> modeNumber, data);
			return;
	}
}

/*
 *	Change de mode.
 */

void switchMode (int modeNumber, WixWindow* data)
{
	WxMode*	mp = & wx_modes [modeNumber];
	int		i;
	char*	p;
	char	tc [256];

	/*
	 *	Indiquer le nom du mode courant.
	 */
	data -> modeNumber = modeNumber;	/* numro du mode */
	data -> submodeNumber = modeNumber;	/* le sous mode est le mme */
	if (data -> wix2) {
		p = rindex (data -> filename, '/');
		p = p ? p+1 : data -> filename;
		sprintf (tc, "%s [%s]", mp -> m_name, p);
		SwSetTitle (XtParent (data -> active), XtNewString (tc));
	} else {
		SwSetTitle (XtParent (data -> active), mp -> m_name);
	}
	/*
	 *	Remettre les translations par defaut,
	 *	les couleurs, la table d'alias, et rajouter les
	 *	translations propres au nouveau mode.
	 */
	if (mp -> m_background == (Pixel)-1)	mp -> m_background = wx_modes -> m_background;
	if (mp -> m_keyword_color == (Pixel)-1)	mp -> m_keyword_color = wx_modes -> m_keyword_color;
	if (mp -> m_library_color == (Pixel)-1)	mp -> m_library_color = wx_modes -> m_library_color;
	if (mp -> m_constant_color == (Pixel)-1)	mp -> m_constant_color = wx_modes -> m_constant_color;
	if (mp -> m_class_color == (Pixel)-1)	mp -> m_class_color = wx_modes -> m_class_color;
	if (mp -> m_cpp_color == (Pixel)-1)		mp -> m_cpp_color = wx_modes -> m_cpp_color;
	if (mp -> m_string_color == (Pixel)-1)	mp -> m_string_color = wx_modes -> m_string_color;
	if (mp -> m_comment_color == (Pixel)-1)	mp -> m_comment_color = wx_modes -> m_comment_color;

	if (!mp -> m_fonts)			mp -> m_fonts = wx_modes -> m_fonts;
	if (!mp -> m_comment_font)	mp -> m_comment_font = wx_modes -> m_comment_font;
	if (!mp -> m_library_font)	mp -> m_library_font = wx_modes -> m_library_font;
	if (!mp -> m_constant_font)	mp -> m_constant_font = wx_modes -> m_constant_font;
	if (!mp -> m_class_font)	mp -> m_class_font = wx_modes -> m_class_font;
	if (!mp -> m_cpp_font)		mp -> m_cpp_font = wx_modes -> m_cpp_font;
	if (!mp -> m_string_font)	mp -> m_string_font = wx_modes -> m_string_font;
	if (!mp -> m_keyword_font)	mp -> m_keyword_font = wx_modes -> m_keyword_font;

	if (DefaultDepthOfScreen (XtScreen (data -> active)) > 1) {
		XtVaSetValues (data -> active,
			XtNbackground,		mp -> m_background,
			XtNcommentColor,	mp -> m_comment_color,
			XtNstringColor,		mp -> m_string_color,
			XtNcppColor,		mp -> m_cpp_color,
			XtNlibraryColor,	mp -> m_library_color,
			XtNconstantColor,	mp -> m_constant_color,
			XtNclassColor,		mp -> m_class_color,
			XtNkeywordColor,	mp -> m_keyword_color,
			NULL);
		if (data -> shadow)
			XtVaSetValues (data -> shadow,
				XtNbackground,		mp -> m_background,
				XtNcommentColor,	mp -> m_comment_color,
				XtNstringColor,		mp -> m_string_color,
				XtNcppColor,		mp -> m_cpp_color,
				XtNlibraryColor,	mp -> m_library_color,
				XtNconstantColor,	mp -> m_constant_color,
				XtNclassColor,		mp -> m_class_color,
				XtNkeywordColor,	mp -> m_keyword_color,
				NULL);
	}

	if (mp -> m_delim == NULL)
		mp -> m_delim = wx_modes -> m_delim;
	XtVaSetValues (data -> active,
		XtNbeginKeyword,		mp -> m_beg_keyword,
		XtNendKeyword,			mp -> m_end_keyword,
		XtNstartComment,		mp -> m_start_comment,
		XtNstopComment, 		mp -> m_stop_comment,
		XtNlineComment, 		mp -> m_line_comment,
		XtNkeywordTable,		mp -> m_keywords,
		XtNdelimitors,			mp -> m_delim,
		XtNstringDelimitor,		mp -> m_str_delim,
		XtNkeyTable,			mp -> m_keytable,
		XtNtranslations,		stdTranslations,
		XtNaliasTable,			mp -> m_alias,
		XtNaliasBuffer,			mp -> m_abbrev,
		XtNaliasBufferLength,	mp -> m_abbrev_len,
		XtNfont,				mp -> m_fonts,
		XtNcommentFont,			mp -> m_comment_font,
		XtNcppFont,				mp -> m_cpp_font,
		XtNstringFont,			mp -> m_string_font,
		XtNlibraryFont,			mp -> m_library_font,
		XtNconstantFont,		mp -> m_constant_font,
		XtNclassFont,			mp -> m_class_font,
		XtNkeywordFont,			mp -> m_keyword_font,
		XtNwrap,				mp -> m_wrap,
		NULL);
	/* Rajouter les translations du mode Fundamental */
	if (wx_modes -> m_trans)
		XtOverrideTranslations (data -> active, wx_modes -> m_trans);
	/* Puis celles du mode choisi */
	if (mp -> m_trans)
		XtOverrideTranslations (data -> active, mp -> m_trans);

	if (data -> shadow) {
		Widget	other = data -> active == data -> wix
							? data -> shadow
							: data -> wix;

		XtVaSetValues (other,
			XtNkeyTable,			mp -> m_keytable,
			XtNbeginKeyword,		mp -> m_beg_keyword,
			XtNendKeyword,			mp -> m_end_keyword,
			XtNstartComment,		mp -> m_start_comment,
			XtNstopComment,			mp -> m_stop_comment,
			XtNlineComment, 		mp -> m_line_comment,
			XtNkeywordTable,		mp -> m_keywords,
			XtNdelimitors,			mp -> m_delim,
			XtNstringDelimitor,		mp -> m_str_delim,
			XtNtranslations,		stdTranslations,
			XtNaliasTable,			mp -> m_alias,
			XtNaliasBuffer,			mp -> m_abbrev,
			XtNaliasBufferLength,	mp -> m_abbrev_len,
			XtNfont,				mp -> m_fonts,
			XtNcommentFont,			mp -> m_comment_font,
			XtNcppFont,				mp -> m_cpp_font,
			XtNstringFont,			mp -> m_string_font,
			XtNlibraryFont,			mp -> m_library_font,
			XtNconstantFont,		mp -> m_constant_font,
			XtNclassFont,			mp -> m_class_font,
			XtNkeywordFont,			mp -> m_keyword_font,
			XtNwrap,				mp -> m_wrap,
			NULL);
		/* Rajouter les translations du mode Fundamental */
		if (wx_modes -> m_trans)
			XtOverrideTranslations (other, wx_modes -> m_trans);
		/* Puis celles du mode choisi */
		if (mp -> m_trans)
			XtOverrideTranslations (other, mp -> m_trans);
	}
	/*
	 *	Mapper les bons boutons de commande.
	 */
	for (i = 0; mp -> m_buttons [i].pixmap || mp -> m_buttons [i].label; i++) {
		static char*	bname [] = { "b1", "b2", "b3", "b4", "b5",
									 "b6", "b7", "b8", "b9", "b10" };
		if (data -> buttons [i] == NULL) {
			data -> buttons [i] = XtVaCreateWidget (bname [i], pushButtonWidgetClass,
									data -> button_row,
									XtNuserData, data,
									XtNpixmap, mp -> m_buttons [i].pixmap,
									XtNlabel, mp -> m_buttons [i].label,
									XtNtipMessage, mp -> m_buttons [i].tip,
									NULL);
			XtAddCallback (data -> buttons [i], XtNrelease, commandCallback, (XtPointer) (long) i);
		} else {
			XtVaSetValues (data -> buttons [i],
				XtNpixmap, mp -> m_buttons [i].pixmap,
				XtNlabel, mp -> m_buttons [i].label,
				XtNtipMessage, mp -> m_buttons [i].tip,
				NULL);
		}
	}
	if (i > 0) {
		XtUnmanageChildren (data -> buttons+i, 10-i);
		XtManageChildren (data -> buttons, i);
		XtManageChild (data -> button_row);
	} else
		XtUnmanageChild (data -> button_row);
}
