/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/*
 * routines to do regular expression matching
 */

#define	NULL	((void*)0)

#define	CBRA	1
#define	CCHR	2
#define	CDOT	4
#define	CCL		6
#define	NCCL	8
#define	CDOL	10
#define	CEOF	11
#define	CKET	12
#define	CBACK	18

#define	CSTAR	01

#define	ESIZE	512
#define	NBRA	9
#define	REPLACE_SIZE 80
#define	comerr(msg) {expbuf [0] = 0; numbra = 0; return(msg); }

static int		numbra;
static char		expbuf [ESIZE];
static char*	braslist [NBRA];
static char*	braelist [NBRA];
static int		circf;
static char		replace_buf [REPLACE_SIZE+1];


char*	_compsub (char*, char*);
char*	_dosub (char*, char*);
char*	pa_re_comp (char*);
char*	pa_re_exec (char*, char**);

static int backref (int i, char* lp)
{
	char*	bp;

	bp = braslist [i];
	while (*bp++ == *lp++)
		if (bp >= braelist [i])
			return 1;
	return 0;
}

static int cclass (char* set, char c, int af)
{
	int	n;

	if (c == 0)
		return 0;
	n = *set++;
	while (--n != 0)
		if (*set++ == c)
			return af;
	return !af;
}

/* 
 * try to match the next thing in the dfa
 */

static char* advance (char* lp, char* ep)
{
	char*	curlp;
	char*	rv;
	int		ct, i;

	for (;;)
		switch (*ep++) {

		case CCHR:
			if (*ep++ == *lp++)
				continue;
			return NULL;

		case CDOT:
			if (*lp++)
				continue;
			return NULL;

		case CDOL:
			if (*lp == '\0')
				continue;
			return NULL;

		case CEOF:
			return lp;

		case CCL:
			if (cclass(ep, *lp++, 1)) {
				ep += *ep;
				continue;
			}
			return NULL;

		case NCCL:
			if (cclass(ep, *lp++, 0)) {
				ep += *ep;
				continue;
			}
			return NULL;

		case CBRA:
			braslist [(int) *ep++] = lp;
			continue;

		case CKET:
			braelist [(int) *ep++] = lp;
			continue;

		case CBACK:
			if (braelist [i = *ep++] == 0)
				return NULL;

			if (backref(i, lp)) {
				lp += braelist [i] - braslist [i];
				continue;
			}
			return NULL;

		case CBACK|CSTAR:
			if (braelist [i = *ep++] == 0)
				return NULL;
			curlp = lp;
			ct = braelist [i] - braslist [i];
			while (backref(i, lp))
				lp += ct;
			while (lp >= curlp) {
				if ((rv = advance(lp, ep)) != NULL)
					return rv;
				lp -= ct;
			}
			continue;

		case CDOT|CSTAR:
			curlp = lp;
			while (*lp++)
				;
			goto star;

		case CCHR|CSTAR:
			curlp = lp;
			while (*lp++ == *ep)
				;
			ep++;
			goto star;

		case CCL|CSTAR:
		case NCCL|CSTAR:
			curlp = lp;
			while (cclass(ep, *lp++, ep [-1] == (CCL|CSTAR)))
				;
			ep += *ep;
		star:
			do {
				lp--;
				if ((rv = advance(lp, ep)) != NULL)
					return rv;
			} while (lp > curlp);
			return NULL;

		default:
			return NULL;
		}
}

/*
 * compile un remplacement: s/old/new/
 */

char* _compsub (char* old, char* new)
{
	int		c;
	char*	p;
	char*	pc;

	if ((p = pa_re_comp (old)) != NULL)
		return p;	/* malformed re */
	p = replace_buf;
	for (pc = new; (c = *pc++) != 0;) {
		if (c=='\\')
			c = *pc++ | 0200;
		*p++ = c;
		if (p >= & replace_buf [REPLACE_SIZE])
			return "re too long.";
	}
	*p++ = 0;
	return NULL;	/* ok */
}

static char* place (char* sp, char* l1, char* l2)
{
	while (l1 < l2)
		*sp++ = *l1++;
	return sp;
}

/*
 *	Retourne la chaine remplaante.
 */

char* _dosub (char* loc1, char* loc2)
{
	static char	genbuf [REPLACE_SIZE+1];
	char*		sp;
	char*		rp;
	int			c;

	sp = genbuf;
	rp = replace_buf;
	while ((c = *rp++ & 0377) != 0) {
		if (c=='&') {
			sp = place (sp, loc1, loc2);
			continue;
		} else if (c&0200 && (c &= 0177) >='1' && c < numbra+'1') {
			sp = place (sp, braslist [c-'1'], braelist [c-'1']);
			continue;
		}
		*sp++ = c&0177;
	}
	*sp = 0;
	return genbuf;
}

/*
 * compile the regular expression argument into a dfa
 */

char* pa_re_comp (char* sp)
{
	static char*	retoolong = "Regular expression too long";
	int				c;
	char*			ep = expbuf;
	int				cclcnt;
	char*			lastep = 0;
	char			bracket [NBRA];
	char*			bracketp = bracket;

	numbra = 0;
	if (sp == 0 || *sp == '\0') {
		if (*ep == 0)
			return "No previous regular expression";
		return NULL;
	}
	if (*sp == '^') {
		circf = 1;
		sp++;
	}
	else
		circf = 0;
	for (;;) {
		if (ep >= &expbuf [ESIZE])
			comerr(retoolong);
		if ((c = *sp++) == '\0') {
			if (bracketp != bracket)
				comerr("unmatched \\(");
			*ep++ = CEOF;
			*ep++ = 0;
			return NULL;
		}
		if (c != '*')
			lastep = ep;
		switch (c) {

		case '.':
			*ep++ = CDOT;
			continue;

		case '*':
			if (lastep == 0 || *lastep == CBRA || *lastep == CKET)
				goto defchar;
			*lastep |= CSTAR;
			continue;

		case '$':
			if (*sp != '\0')
				goto defchar;
			*ep++ = CDOL;
			continue;

		case '[':
			*ep++ = CCL;
			*ep++ = 0;
			cclcnt = 1;
			if ((c = *sp++) == '^') {
				c = *sp++;
				ep [-2] = NCCL;
			}
			do {
				if (c == '\0')
					comerr("missing ]");
				if (c == '-' && ep [-1] != 0) {
					if ((c = *sp++) == ']') {
						*ep++ = '-';
						cclcnt++;
						break;
					}
					while (ep [-1] < c) {
						*ep = ep [-1] + 1;
						ep++;
						cclcnt++;
						if (ep >= &expbuf [ESIZE])
							comerr(retoolong);
					}
				}
				*ep++ = c;
				cclcnt++;
				if (ep >= &expbuf [ESIZE])
					comerr(retoolong);
			} while ((c = *sp++) != ']');
			lastep [1] = cclcnt;
			continue;

		case '\\':
			if ((c = *sp++) == '(') {
				if (numbra >= NBRA)
					comerr("too many \\(\\) pairs");
				*bracketp++ = numbra;
				*ep++ = CBRA;
				*ep++ = numbra++;
				continue;
			}
			if (c == ')') {
				if (bracketp <= bracket)
					comerr("unmatched \\)");
				*ep++ = CKET;
				*ep++ = *--bracketp;
				continue;
			}
			if (c >= '1' && c < ('1' + NBRA)) {
				*ep++ = CBACK;
				*ep++ = c - '1';
				continue;
			}
			*ep++ = CCHR;
			*ep++ = c;
			continue;

		defchar:
		default:
			*ep++ = CCHR;
			*ep++ = c;
		}
	}
}

/* 
 * match the argument string against the compiled re
 * returns 0 if no match, a pointer to the end of the
 * string matched else.
 */

char* pa_re_exec (char* p1, char** begp)
{
	char*	p2 = expbuf;
	int		c;
	char*	rv;

	for (c = 0; c < NBRA; c++) {
		braslist [c] = 0;
		braelist [c] = 0;
	}
	if (circf) {
		*begp = p1;
		return advance (p1, p2);
	}
	/*
	 * fast check for first character
	 */
	if (*p2 == CCHR) {
		c = p2 [1];
		do {
			if (*p1 != c)
				continue;
			if ((rv = advance(p1, p2)) != NULL) {
				*begp = p1;
				return rv;
			}
		} while (*p1++);
		return NULL;
	}
	/*
	 * regular algorithm
	 */
	do
		if ((rv = advance (p1, p2)) != NULL) {
			*begp = p1;
			return rv;
		} while (*p1++);
	return NULL;
}
