/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/StringDefs.h>

#include <Xpa/Primitive.h>
#include <Xpa/Container.h>
#include <Xpa/PushButton.h>
#include <Xpa/Wix.h>
#include <Xpa/Query.h>
#include <Xpa/MenuBar.h>
#include <Xpa/ScrolledWin.h>

#include "wix.h"

char			prev_searched_string [256];

static char*	ForwardSearchMessage = "Forward Search";
static char*	BackwardSearchMessage = "Backward Search";

static char*	oldPattern = NULL;
static char*	newPattern = NULL;

/*
 *	Recherche en avant une expression rgulire.
 */

Boolean re_forward_search (Widget w, char* pattern)
{
	char*	cur_end = (char*) _GetCursorEnd (w);
	char*	buf_end = (char*) _GetBufferEnd (w);
	char*	begp;
	char*	ok;
	char*	cp;
	char*	p;

	if (pa_re_comp (pattern))
		return False;	/* invalid re */
	/*
	 *	Chercher, en remplacant momentanement les
	 *	\n par un \0 pour que chaque ligne soit
	 *	une string C.
	 */
 	for (cp = cur_end; cp < buf_end; ) {
		for (p = cp; *p != '\n'; p++);	/* trouver la fin de ligne */
		*p = 0;				/* mettre un \0 */
		ok = pa_re_exec (cp, &begp);
		*p = '\n';
		if (ok) {
			ForwardChar (w, ok - cur_end, NULL);
			return True;
		}
		cp = p+1;			/* ligne suivante */
	}
	return False;
}

/*
 *	Recherche en arriere une expression reguliere.
 */

Boolean re_backward_search (Widget w, char* pattern)
{
	char*	cur_beg = (char*) _GetCursorBeg (w);
	char*	buf_beg = (char*) _GetBufferBeg (w);
	char*	begp;
	char*	ok;
	char*	cp;
	char*	p;

	if (pa_re_comp (pattern))
		return False;	/* invalid re */
	/*
	 *	Commencer par la ligne courante.
	 */
	cp = cur_beg; *cp = 0;
	while (*--cp != '\n') ;		/* debut de la ligne courante */
	if (pa_re_exec (cp+1, &begp)) {
		BackwardChar (w, cur_beg - begp, NULL);
		return True;		/* bingo */
	}
	/*
	 *	Chercher, en remplacant momentanement les
	 *	\n par un \0 pour que chaque ligne soit
	 *	une string C.
	 */
	while (cp > buf_beg) {
		p = cp;
		*p = 0;
		while (*--cp != '\n');	/* trouver la fin de la precedente ligne */
		ok = pa_re_exec (cp+1, &begp);
		*p = '\n';
		if (ok) {
			BackwardChar (w, cur_beg - begp, NULL);
			return True;
		}
	}
	return False;
}

/*
 *	Les actions: recherche en avant.
 */

static int _reForwardSearchFunction (Widget w, int count, void* arg)
{
	char*	pattern = (char*) arg;

	ClearRegion (w);	/* effacer la region */
	if (!re_forward_search (w, pattern)) {
		XBell (XtDisplay (w), 100);
		return 0;
	}
	return 1;
}

/*
 *	Les actions: recherche en arriere.
 */

static int _reBackwardSearchFunction (Widget w, int count, void* arg)
{
	char*	pattern = (char*) arg;

	ClearRegion (w);	/* effacer la region */
	if (!re_backward_search (w, pattern)) {
		XBell (XtDisplay (w), 100);
		return 0;
	}
	return 1;
}

/*
 *	Forward Search.
 */

int _forwardSearch (Widget w, int count, void* arg)
{
	unsigned char*	pattern = (unsigned char*) arg;
	unsigned char*	ce = _GetCursorEnd (w);
	unsigned char*	cp;

	ClearRegion (w);	/* effacer la region */
	cp = _WixForwardSearch (w, ce, pattern, 1);
	if (cp != NULL) {
		WixUnhighlightAll (w);
		(void) ForwardChar (w, cp - ce, NULL);
		WixHighlight (w, - strlen (pattern));
		return 1;
	}
	XBell (XtDisplay (w), 100);
	return 0;
}

static void forwardSearchCont (WixWindow* data)
{
	String	pattern;
	Boolean	use_regexp = False;

	pattern = data -> fs_answer;
	if (data -> query_string)	/* elle peut ne pas exister si on est expert ... */
		XtVaGetValues (QueryGetChild (data -> query_string, APPLY_BUTTON),
			XtNset,	&use_regexp,
			NULL);
	if (use_regexp) {
		_reForwardSearchFunction (data -> active, 1, pattern);
		SaveCommand (data -> active, _reForwardSearchFunction, pattern, True);
	} else {
		_forwardSearch (data -> active, 1, pattern);
		SaveCommand (data -> active, _forwardSearch, pattern, True);
	}
	strcpy (prev_searched_string, pattern);
	WixRedraw (data -> active);
}

void ForwardSearchAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	if (* num_params == 1) {
		char*	s;
		char*	d;
		char	tc [128];

		for (d = tc, s = params [0];;) {
			switch (*d++ = *s++) {
				case 0:
					break;
				case '\\':
					if (*s == 'n') {
						d [-1] = '\n';
						s++;
					}
				default:
					continue;
			}
			break;
		}
		_forwardSearch (w, 1, tc);
		WixRedraw (w);
		return;
	}
	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = forwardSearchCont;
	popQueryStringNoWait (data, ForwardSearchMessage, prev_searched_string, True);
}

static void __re_forwardSearchCont (Widget wix, char* reponse)
{
	WixWindow*	data;
	String	pattern;

	if (reponse == 0)
		return;

	/* rcuprer nos infos */
	XtVaGetValues (XtParent (XtParent (wix)),
					XtNuserData,	&data,
					NULL);
	data -> fs_answer = reponse;
	pattern = reponse;
	_reForwardSearchFunction (data -> active, 1, pattern);
	SaveCommand (data -> active, _reForwardSearchFunction, pattern, True);
	strcpy (prev_searched_string, pattern);
	WixRedraw (data -> active);
}

void FastForwardSearchAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	SetConfirmAction (data -> menubar, __re_forwardSearchCont);
	PopMenubarDialog (data -> menubar, XtParent (data -> active),
					  "Regexpr Search Forward", DM_GetString,
					  prev_searched_string);
}

/*
 *	Backward Search.
 */

int _backwardSearch (Widget w, int count, void* arg)
{
	unsigned char*	pattern = (unsigned char*) arg;
	unsigned char*	ce = _GetCursorBeg (w);
	unsigned char*	cp;

	ClearRegion (w);	/* effacer la region */
	cp = _WixBackwardSearch (w,  ce - 1, pattern, 1);
	if (cp != NULL) {
		WixUnhighlightAll (w);
		(void) BackwardChar (w, ce - cp, NULL);
		WixHighlight (w, strlen (pattern));
		return 1;
	}
	XBell (XtDisplay (w), 100);
	return 0;
}

static void backwardSearchCont (WixWindow* data)
{
	String	pattern;
	Boolean	use_regexp = False;

	pattern = data -> fs_answer;
	if (data -> query_string)	/* elle peut ne pas exister si on est expert ... */
		XtVaGetValues (QueryGetChild (data -> query_string, APPLY_BUTTON),
			XtNset,	&use_regexp,
			NULL);
	if (use_regexp) {
		_reBackwardSearchFunction (data -> active, 1, pattern);
		SaveCommand (data -> active, _reBackwardSearchFunction, pattern, True);
	} else {
		_backwardSearch (data -> active, 1, pattern);
		SaveCommand (data -> active, _backwardSearch, pattern, True);
	}
	strcpy (prev_searched_string, pattern);
	WixRedraw (data -> active);
}

static void __backwardSearchCont (Widget wix, char* reponse)
{
	WixWindow*	data;

	if (reponse == 0)
		return;

	/* rcuprer nos infos */
	XtVaGetValues (XtParent (XtParent (wix)),
					XtNuserData,	&data,
					NULL);
	data -> fs_answer = reponse;
	backwardSearchCont (data);
}

void FastBackwardSearchAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	SetConfirmAction (data -> menubar, __backwardSearchCont);
	PopMenubarDialog (data -> menubar, XtParent (data -> active),
					  " Backward Search ", DM_GetString,
					  prev_searched_string);
}

void BackwardSearchAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	if (* num_params == 1) {
		char*	s;
		char*	d;
		char	tc [128];

		for (d = tc, s = params [0];;) {
			switch (*d++ = *s++) {
				case 0:
					break;
				case '\\':
					if (*s == 'n') {
						d [-1] = '\n';
						s++;
					}
				default:
					continue;
			}
			break;
		}
		_backwardSearch (w, 1, tc);
		WixRedraw (w);
		return;
	}
	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> what_after = backwardSearchCont;
	popQueryStringNoWait (data, BackwardSearchMessage, prev_searched_string, True);
}

static Boolean forw_re_sub (Widget w, char* old, char* new, char** pnew, int* plen, int* pnl)
{
	char*	p;
	char*	cu_e = (char*) _GetCursorEnd (w);
	char*	end = (char*) _GetBufferEnd (w);
	char*	nl_addr;
	char*	cp;
	char*	begp;

	if (_compsub (old, new))
		return False;

 	for (p = cu_e; p < end; p = nl_addr + 1) {
		for (nl_addr = p; *nl_addr != '\n'; nl_addr++);
		*nl_addr = 0;
		cp = pa_re_exec (p, &begp);
		*nl_addr = '\n';
		if (cp) {
			*pnew = _dosub (begp, cp);
			ForwardChar (w, cp - cu_e, NULL);
			*pnl = strlen (*pnew);
			*plen = cp - begp;
			return True;
		}
	}
	return False;
}

static void _process_events (WixWindow* data)
{
	XtAppContext	app = XtWidgetToApplicationContext (data -> active);

	while (XtAppPending (app))
		XtAppProcessEvent (app, XtIMAll);
}

static void queryReplaceFunction (WixWindow* data)
{
	long		marks [15000];
	XEvent		ev;
	Widget		widget;
	Window		win;
	int			oldLength;
	int			newLength;
	Boolean		prev_case_sensitive;
	Boolean		undo_or_done;
	Boolean		do_exit;
	Boolean		use_regexp;
	char		c;
	int			pos;
	char		tc [64];
	char*		pnew;
	int			nrep = 0;
	static char	msg [] =
		"Space: skip   Return: change   ^R: replace all   ^U: undo   ^G: exit   ^L: last replace";

	if (oldPattern) XtFree (oldPattern);
	if (newPattern) XtFree (newPattern);

	oldPattern = WixCopyBuffer (QueryGetChild (data -> query_string, EDIT));
	newPattern = WixCopyBuffer (QueryGetChild (data -> query_string, AUX_EDIT));

	XtVaGetValues (QueryGetChild (data -> query_string, APPLY_BUTTON),
		XtNset,	&use_regexp,
		NULL);

	if (use_regexp) {
		if (!forw_re_sub (data -> active, oldPattern, newPattern, &pnew, &oldLength, &newLength))
		{
			XBell (XtDisplay (data -> active), 100);
			return;
		}
	} else {
		XtVaGetValues (data -> active,
			XtNcaseSensitive, &prev_case_sensitive,
			NULL);
		/* rendre la recherche sensible  la casse */
		XtVaSetValues (data -> active,
			XtNcaseSensitive, True,
			NULL);
		if (!WixForwardSearch (data -> active, (unsigned char*) oldPattern, 1)) {
			XBell (XtDisplay (data -> active), 100);
			XtVaSetValues (data -> active,
				XtNcaseSensitive, prev_case_sensitive,
				NULL);
			return;
		}
		oldLength = strlen (oldPattern);
		newLength = strlen (newPattern);
	}

	/* Retirer momentanment la scrollbar du bas */
	if (XtIsManaged (data -> menubar))
		XtUnmanageChild (SwHSB (XtParent (data -> wix)));

	Message (data -> active, msg);

	win = XtWindow (data -> active);
	data -> qr_MarkTab = marks;	/* pour undo */
	data -> qr_MarkNum = 0;

	undo_or_done = False;
	for (do_exit = False;;) {
		WixRedraw (data -> active);
	noredraw:
		for (;;) {
			XtAppContext	app = XtWidgetToApplicationContext (data -> pane);

			XtAppNextEvent (app, &ev);
			widget = XtWindowToWidget (XtDisplay (data -> active), ev.xany.window);
			while (widget) {
				if (widget == data -> toplevel)
					goto dispatch;
				widget = XtParent (widget);
			}
			XtDispatchEvent (&ev);
		}
	dispatch:
		switch (ev.type) {
		case KeyPress:
			if (XLookupString ((XKeyEvent *) &ev, &c, 1, 0, 0)) {
				switch (c) {
					case 'r'&037:
						if (undo_or_done) {
							XBell (XtDisplay (data -> active), 100);
							continue;
						}
						XtSetSensitive (data -> pane, False);
						if (use_regexp)
							do {
								pos = WixGetInsertPosition (data -> active) - oldLength;
								WixReplace (data -> active, pos, pos + oldLength, pnew, newLength);
								_process_events (data);
								nrep ++;
							} while (forw_re_sub (data -> active, oldPattern, newPattern, &pnew, &oldLength, &newLength));
						else
							do {
								pos = WixGetInsertPosition (data -> active) - oldLength;
								if (data -> qr_MarkNum < XtNumber (marks))
									data -> qr_MarkTab [data -> qr_MarkNum ++] = pos;
								WixReplace (data -> active, pos, pos + oldLength, newPattern, newLength);
								_process_events (data);
							} while (WixForwardSearch (data -> active, (unsigned char*) oldPattern, 1));
						XtSetSensitive (data -> pane, True);
						goto undo;

					case 'u'&037:
						if (data -> qr_MarkNum > 0 && data -> qr_MarkNum < XtNumber (marks)) {
							pos = data -> qr_MarkTab [--data -> qr_MarkNum];
							WixReplace (data -> active, pos, pos + newLength, oldPattern, oldLength);
							data -> qr_MarkTab [data -> qr_MarkNum] = pos;
							undo_or_done = False;
							Message (data -> active, msg);
							continue;
						}
					default:
						XBell (XtDisplay (data -> active), 100);
						continue;

					case ' ':
						if (undo_or_done) {
							XBell (XtDisplay (data -> active), 100);
							continue;
						}
						break;

					case 'l'&037:
						do_exit = True;

					case '\r':
						if (undo_or_done) {
							XBell (XtDisplay (data -> active), 100);
							continue;
						}
						pos = WixGetInsertPosition (data -> active) - oldLength;
						if (use_regexp) {
							WixReplace (data -> active, pos, pos + oldLength, pnew, newLength);
							nrep ++;
						} else {
							data -> qr_MarkTab [data -> qr_MarkNum++] = pos;
							WixReplace (data -> active, pos, pos + oldLength, newPattern, newLength);
						}
						if (!do_exit)
							break;

					case 'g'&037:
						goto done;
				}
				if ((use_regexp && !forw_re_sub (data -> active, oldPattern, newPattern, &pnew, &oldLength, &newLength)) ||
				    (!use_regexp && !WixForwardSearch (data -> active, (unsigned char*) oldPattern, 1)))
				{
				undo:	Message (data -> active, use_regexp ? "^G exit" : "^G exit ^U undo");
					undo_or_done = True;
					continue;

				done:	sprintf (tc, "%d patterns replaced.", use_regexp ? nrep : data -> qr_MarkNum);
					/* Remettre la scrollbar du bas */
					if (XtIsManaged (data -> menubar))
						XtManageChild (SwHSB (XtParent (data -> wix)));
					Message (data -> active, tc);
					WixRedraw (data -> active);
					XtVaSetValues (data -> active,
						XtNcaseSensitive, prev_case_sensitive,
						NULL);
					return;
				}
				continue;
			}
			continue;

		case ButtonPress:
		case ButtonRelease:
		case MotionNotify:
			goto noredraw;

		default:
			XtDispatchEvent (&ev);
			goto noredraw;
		}
	}
}

void queryReplaceAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*	data;

	XtVaGetValues (w, XtNuserData, &data, NULL);

	popQueryReplace (data, "Search/Replace", oldPattern, newPattern);
	data -> what_after = queryReplaceFunction;
}

/*
 *	Recherche incrmentale.
 */

static void incrementalSearchConfirm (Widget wix, char* reponse)
{
	WixWindow*	data;

	if (reponse == 0)
		return;			/* abandon */

	XtVaGetValues (XtParent (XtParent (wix)),
					XtNuserData,	&data,
					NULL);
	if (data -> qr_MarkTab == NULL) {
		/* on a juste tap return pour rechercher la mme chaine qu'avant */
		_forwardSearch (data -> active, 1, (void*) prev_searched_string);
		WixRedraw (data -> active);
	} else {
		data -> qr_MarkTab = NULL;
		strcpy (prev_searched_string, reponse);
	}
	/* pour ^C et les macros */
	SaveCommand (data -> active, _forwardSearch, prev_searched_string, True);
}

static void incrementalSearchNavigate (Widget wix, XtPointer call)
{
	WixWindow*		data;
	XKeyEvent*		e = (XKeyEvent*) call;
	unsigned char*	cp;
	unsigned char*	ce;
	char			c;
	char*			p;
	int				l;

	if (0 == XLookupString (e, &c, 1, NULL, 0))
		return;

	XtVaGetValues (XtParent (XtParent (wix)),
					XtNuserData,	&data,
					NULL);

	p = WixCopyBuffer (wix);
	switch (c) {
		/* aller  l'occurence suivante */
		case 'n'&037:
			_forwardSearch (data -> active, 1, (void*) p);
			break;
		/* aller  l'occurence prcdente */
		case 'p'&037:
			l = strlen (p);
			if (cp = _WixBackwardSearch (data -> active, ce = _GetCursorBeg (data -> active) - l - 1, p, 1)) {
				WixUnhighlightAll (data -> active);
				(void) BackwardChar (data -> active, ce - cp + 1, NULL);
				WixHighlight (data -> active, -l);
			} else
				XBell (XtDisplay (wix), 100);
			break;
	}
	data -> qr_MarkTab = (long*) 1;	/* on est pass par l */
	XtFree (p);						/* on n'en a plus besoin */
	WixRedraw (data -> active);									/* rafficher  et */
	data -> qr_MarkNum = WixGetInsertPosition (data -> active); /* repartir d'ici */
}

static void incrementalSearch (Widget wix, XtPointer call)
{
	VerifyCallbackStruct* cb = (VerifyCallbackStruct*) call;
	WixWindow*		data;
	int				ici;
	char			pattern [128];
	char*			p;
	int				length;
	unsigned char*	ce;
	unsigned char*	cp;

	/* rcuprer nos infos */
	XtVaGetValues (XtParent (XtParent (wix)),
					XtNuserData,	&data,
					NULL);
	/* reconstiuer le chaine saisie */
	p = WixCopyBuffer (wix);
	length = strlen (p);
	pattern [0] = 0;
	strncpy (pattern, p, cb -> pos);
	if (cb -> reason == INSERT_STRING) {
		strncpy (pattern+cb -> pos, cb -> string, cb -> length);
		strcpy (pattern+cb -> pos + cb -> length, p+cb -> pos);
	} else
		strcpy (pattern+cb -> pos, p + cb -> pos + cb -> length);
	XtFree (p);		/* on n'en a plus besoin */

	data -> qr_MarkTab = (long*) 1;	/* on est pass par l */
	ici = WixGetInsertPosition (data -> active);
	WixUnhighlight (data -> active, -length);
	EnableRedisplay (data -> active, False, False);
	WixSetInsertPosition (data -> active, data -> qr_MarkNum);
	if (pattern [0] != 0 && !_forwardSearch (data -> active, 1, (void*) pattern)) {
		WixSetInsertPosition (data -> active, ici);
		WixHighlight (data -> active, -length);
		cb -> doit = False;
	}
	EnableRedisplay (data -> active, True, False);
}

void IncrementalSearchAction (Widget w, XEvent* event, String* params, Cardinal* num_params)
{
	WixWindow*		data;

	XtVaGetValues (w, XtNuserData, &data, NULL);
	data -> qr_MarkNum = WixGetInsertPosition (w);
	data -> qr_MarkTab = NULL;
	SetConfirmAction (data -> menubar, incrementalSearchConfirm);
	SetDefaultAction (data -> menubar, incrementalSearchNavigate);
	SetAutoExpand (data -> menubar, incrementalSearch);
	PopMenubarDialog (data -> menubar, XtParent (data -> active),
					  " Incremental Search ", DM_GetStringIncremental,
					  prev_searched_string);
}
