/*****************************************************************

  Copyright (c) 1989 - 2009 by Patrick Amar - pa@lri.fr

    http://www.lri.fr/~pa

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

*****************************************************************/

/*
 *	Ce qui dfinit une fentre wix.
 */

typedef struct _Data {
	struct _Data*	next;			/* ces structures sont chaines */
	Boolean			free;			/* fentre disponible */
	Widget			toplevel;		/* le shell */
	Widget			frame;			/* le beau cadre */
	Widget			menubar;		/* la barre de menus */
	Widget			pane;			/* la paned window */
	Widget			wix;			/* le wix */
	Widget			wix2;			/* l'autre (sur un autre fichier) */
	Widget			shadow;			/* son ombre */
	Widget			active;			/* la fentre active */
	char			mode_name [64];	/* buffer pour le nom du mode */
	char			filename [256];	/* le nom du fichier en dition */
	long			date;			/* sa date de dernire modification */
	char			directory [256];/* le directory courant */
	char			client_pid [8];	/* numro du process client */
	int				checkpoint;		/* o en est la sauvegarde priodique */
	int				modeNumber;		/* le numro du mode courant */
	int				submodeNumber;	/* le numro du sous mode */
	Boolean			msg_cleared;	/* le prcdent message est effac */
	char			msg [128];		/* la chaine pour les messages */
	Boolean			stripped;		/* la fentre est sans menu ni scrollbars */
	/* tout le undo/redo */
	Boolean			undo_mode;		/* en cours d'undo/redo */
	char			prev_reason;	/* la prcdente commande */
	long			undo_ptr;		/* o on en est dans le fichier d'undo/redo */
	int				undo_number;	/* nombre de commandes  dfaire */
	/* les flags pour les boites de dialogue */
	Boolean			popped;			/* une boite est actuellement affiche */
	Boolean			cancelled;		/* on a annul la commande */
	/* le file selector */
	Widget			fs;
	String			fs_answer;
	/* la boite de string */
	Widget			query_string;
	void			(*what_after) (struct _Data*);
	void			(*apply_after) (struct _Data*);
	long*			qr_MarkTab;
	int				qr_MarkNum;
	/* la boite de diagnostics */
	Widget			diag_list;
	int				diag_pid;
	char			diag_file [32];	/* le fichier de diagnostics */
	/* la boite de tags */
	Widget			tag_list;
	/* les boutons de commande */
	Widget			button_row;		/* le conteneur des boutons */
	Widget			buttons [10];	/* dix devraient suffire */
	Widget			ruler;			/* la regle pour les tabulations */
} WixWindow;

/*
 *	Les boutons de la barre d'outils d'un mode.
 */

typedef struct {
	Pixmap			pixmap;	/*  l'image */
	String			label;	/* le nom */
	String			tip;	/* la doc */
	XtActionProc	proc;
} ModeAction;

/*
 *	Structure dfinissant un mode.
 */

typedef struct {
	String			m_name;				/* le nom du mode */
	String			m_suf;				/* les suffixes */
	String			m_magic;			/* le mot magique pour dterminer le type du fichier */
	alias*			m_alias;			/* la table d'alias */
	XtTranslations	m_trans;			/* celle de translation */
	String			m_string_trans;		/* au format string */
	String			m_compile;			/* la commande de compilation */
	Trans*			m_keytable;			/* la table des clefs */
	Pixel			m_background;		/* la couleur du fond */
	String			m_delim;			/* les delimiteurs */
	String			m_par_delim;		/* dlimiteur de paragraphe */
	String			m_beg_keyword;		/* dbut de mot clef */
	String			m_end_keyword;		/* fin de mot clef */
	String			m_start_comment;	/* debut de commentaire */
	String			m_stop_comment;		/* fin de commentaire */
	String			m_line_comment;		/* commentaire sur une ligne */
	char			m_str_delim;		/* dlimiteur de chaine */
	KeyWord*		m_keywords;			/* la table des mots clefs */
	FontInfo*		m_fonts;			/* les fontes */
	FontInfo*		m_comment_font;		/* des commentaires */
	FontInfo*		m_cpp_font;			/* des directives pour le pre-processeur */
	FontInfo*		m_library_font;		/* des fonctions de bibliothques */
	FontInfo*		m_constant_font;	/* des constantes */
	FontInfo*		m_class_font;		/* des noms de classe */
	FontInfo*		m_string_font;		/* fonte des chaines de caractres */
	FontInfo*		m_keyword_font;		/* des mots clefs */
	Pixel			m_comment_color;	/* couleur des commentaires */
	Pixel			m_cpp_color;		/* couleur des directives pour le pre-processeur */
	Pixel			m_library_color;	/* couleur des fonctions de bibliothques */
	Pixel			m_string_color;		/* couleur des chaines de caractres */
	Pixel			m_constant_color;	/* couleur des constantes */
	Pixel			m_class_color;		/* couleur des noms de classe */
	Pixel			m_keyword_color;	/* couleur de mots clefs */
	int				m_comment_submode;	/* sous mode dans un commentaire */
	int				m_string_submode;	/* sous mode dans une chaine */
	char*			m_abbrev;			/* buffer des abbreviations */
	int				m_abbrev_len;		/* sa longueur */
	ModeAction		m_buttons [10];		/* les boutons specifiques */
	String			m_errors;			/* regexp pour detecter les erreurs */
	String			m_print;			/* la commande d'impression */
	String			m_menu_names [20];	/* le menu popup */
	XtActionProc	m_menu_actions [20];
	Boolean			m_wrap;				/* wrap par dfaut dans ce mode */
	char*			m_tagcmd;			/* la commande pour crer les tags */
} WxMode;

extern WxMode		wx_modes [];
extern int			wx_numModes;
extern String**		MenuTitle;			/* menu.c */
extern XtActionProc**	MenuProc;		/* menu.c */
extern XtTranslations	stdTranslations;/* menu.c */

extern char			prev_searched_string [];	/* search.c */

/*
 *	Fonctions exportes par:	dialog.c
 */
extern void		PopBye (WixWindow*, char*, Boolean, char*, char*, char*);
extern void		CreateQueryString (WixWindow*);
extern void		popQueryStringNoWait (WixWindow*, char*, char*, Boolean);
extern void		popQueryReplace (WixWindow*, char*, char*, char*);
extern void		popAlert (WixWindow*, char*, Boolean);
extern void		WaitForResponse (WixWindow*);
extern void		popFileSelector (WixWindow*, char*, Boolean);
extern void		_CreateFileSelector (WixWindow*);
/*
 *	Fonctions exportes par:	diag.c
 */
extern void		ClearDiagBox (WixWindow*);
extern void		CreateDiagBox (Widget, WixWindow*);
extern void		PopDiagBox (WixWindow*);
extern void		ShowTagsAction (Widget, XEvent*, String*, Cardinal*);
extern void		SetTagCommandAction (Widget, XEvent*, String*, Cardinal*);
/*
 *	Fonctions exportes par:	regexpr.c
 */
extern char*	_compsub (char*, char*);
extern char*	_dosub (char*, char*);
extern char*	pa_re_comp (char*);
extern char*	pa_re_exec (char*, char**);
/*
 *	Fonctions exportes par:	search.c
 */
extern void		IncrementalSearchAction (Widget, XEvent*, String*, Cardinal*);
extern void		FastForwardSearchAction (Widget, XEvent*, String*, Cardinal*);
extern void		FastBackwardSearchAction (Widget, XEvent*, String*, Cardinal*);
extern void		ForwardSearchAction (Widget, XEvent*, String*, Cardinal*);
extern void		BackwardSearchAction (Widget, XEvent*, String*, Cardinal*);
extern void		queryReplaceAction (Widget, XEvent*, String*, Cardinal*);
extern int		_forwardSearch (Widget, int, void*);
extern int		_backwardSearch (Widget, int, void*);
/*
 *	Fonctions exportes par:	help.c
 */
extern void		HelpAction (Widget, XEvent*, String*, Cardinal*);
extern void		HelpConfigAction (Widget, XEvent*, String*, Cardinal*);
extern void		HelpNewFeaturesAction (Widget, XEvent*, String*, Cardinal*);
/*
 *	Fonctions exportes par:	undo.c
 */
extern void		UndoAction (Widget, XEvent*, String*, Cardinal*);
extern void		UndoRedoCallback (Widget, XtPointer, XtPointer);
/*
 *	Fonctions exportes par:	mode.c
 */
extern int		_GetModeNumber (char*);
extern void		ContextCallback (Widget, XtPointer, XtPointer);
extern void		AdjustParagraphAction (Widget, XEvent*, String*, Cardinal*);
extern void		ChangeSubmodeAction (Widget, XEvent*, String*, Cardinal*);
extern void		BeginDefunAction (Widget, XEvent*, String*, Cardinal*);
extern void		EndDefunAction (Widget, XEvent*, String*, Cardinal*);
extern void		IndentLispLineAction (Widget, XEvent*, String*, Cardinal*);
extern void		IndentLispFunctionAction (Widget, XEvent*, String*, Cardinal*);
extern void		NewLineIndentLispAction (Widget, XEvent*, String*, Cardinal*);
/*
 *	Fonctions exportes par:	lisp.c
 */
typedef	void*	LispExpr;

extern int		lisp_load_file (char*);
extern void		lisp_print (LispExpr, char*);
extern void		lisp_eval_lambda (char*, char*, void*, int);
extern LispExpr	lisp_eval (LispExpr);
extern LispExpr	lisp_read_string (char*);
extern void		lisp_initialize (void);
/*
 *	Fonctions exportes par:	menu.c
 */
extern int		SuffixToMode (WixWindow*);
extern void		Message (Widget, char*);
extern void		switchMode (int, WixWindow*);
extern char*	GetFunctionByAddr (CommandProc);
extern void		Restart (char*, Widget, char*);
/*
 *	Fonctions extern par:	cmd.c
 */
extern char		HomeDirectory [];
extern			WixWindow*	_head;

extern void		WatchCursor (WixWindow*);
extern void		TextCursor (WixWindow*);
extern void		warpPointer (WixWindow*);
extern void		setFontFunction (Widget, char*);
extern Boolean	loadFileFunction (Widget, char*);
extern void		loadFileAsk (WixWindow*);
extern void		createWindowCallback (Widget, XtPointer, XtPointer);
extern void		bufferModifiedCallback (Widget, XtPointer, XtPointer);
extern void		modifCallback (Widget, XtPointer, XtPointer);
extern void		quitCont (WixWindow*);
extern void		StripWindow (WixWindow*);
extern void		DressWindow (WixWindow*);

extern void		ToggleCaseSensitivityAction (Widget, XEvent*, String*, Cardinal*);
extern void		setTabAction (Widget, XEvent*, String*, Cardinal*);
extern void		lispEvalAction (Widget, XEvent*, String*, Cardinal*);
extern void		loadLispFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		SaveMacroAction (Widget, XEvent*, String*, Cardinal*);
extern void		loadFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		FastLoadFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		reloadFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		writeFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		FastWriteFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		saveFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		insertFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		FastInsertFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		clearBufferAction (Widget, XEvent*, String*, Cardinal*);
extern void		deleteFileAction (Widget, XEvent*, String*, Cardinal*);
extern void		changeDirectoryAction (Widget, XEvent*, String*, Cardinal*);
extern void		deleteWindowAction (Widget, XEvent*, String*, Cardinal*);
extern void		createWindowAction (Widget, XEvent*, String*, Cardinal*);
extern void		splitWindowAction (Widget, XEvent*, String*, Cardinal*);
extern void		removePaneAction (Widget, XEvent*, String*, Cardinal*);
extern void		selectPaneAction (Widget, XEvent*, String*, Cardinal*);
extern void		stripAction (Widget, XEvent*, String*, Cardinal*);
extern void		nextWindowAction (Widget, XEvent*, String*, Cardinal*);
extern void		startMacroAction (Widget, XEvent*, String*, Cardinal*);
extern void		stopMacroAction (Widget, XEvent*, String*, Cardinal*);
extern void		executeNamedMacroAction (Widget, XEvent*, String*, Cardinal*);
extern void		quitAction (Widget, XEvent*, String*, Cardinal*);
extern void		lineCounterAction (Widget, XEvent*, String*, Cardinal*);
extern void		UnmodifyBufferAction (Widget, XEvent*, String*, Cardinal*);
extern void		toggleEditableAction (Widget, XEvent*, String*, Cardinal*);
extern void		ShowLineNumberAction (Widget, XEvent*, String*, Cardinal*);
extern void		AutoWrapAction (Widget, XEvent*, String*, Cardinal*);
extern void		ToggleOctalDisplayAction (Widget, XEvent*, String*, Cardinal*);
extern void		showRulerAction (Widget, XEvent*, String*, Cardinal*);
/*
 *	Fonctions exportes par:	errors.c
 */
extern void		showErrors (Widget, int, int);
extern void		execCommandAction (Widget, XEvent*, String*, Cardinal*);
extern void		setCompileCommandAction (Widget, XEvent*, String*, Cardinal*);
extern void		compileBufferAction (Widget, XEvent*, String*, Cardinal*);
extern void		printBufferAction (Widget, XEvent*, String*, Cardinal*);
extern void		showKillStackAction (Widget, XEvent*, String*, Cardinal*);
/*
 *	Fonctions exportes par:	main.c
 */
extern XtTranslations	HelpTrans;
extern XtTranslations	listTrans;
extern Cursor		WatchCurs;
extern Cursor		TextCurs;
extern Cursor		ArrowCurs;
extern Boolean		debugMode;
extern Atom			A_HERE;
extern Atom			A_CLIENT;
extern FontInfo*	default_font;
extern WixWindow*	theWindow;
extern Pixmap		QuestionPixmap;
extern Pixmap		ErrorPixmap;
extern Pixmap		OkPixmap;
extern Pixmap		modifPixmap;
extern Pixmap		lockPixmap;
extern Widget		ghost;

extern void			commandCallback (Widget, XtPointer, XtPointer);
extern void			verifyCallback (Widget, XtPointer, XtPointer);
extern WixWindow*	newWindow (char*);
extern Boolean		_SetSnapshot (char*, Boolean);
