#include <stdio.h>
#include <setjmp.h>
#include <X11/Intrinsic.h>
#include "Wxd/JPEG/jpeglib.h"
#include "jpg_gif.h"

typedef struct my_error_mgr {
	struct jpeg_error_mgr pub;	/* "public" fields */
	jmp_buf setjmp_buffer;		/* for return to caller */
} my_error_mgr;

/*
 * Here's the routine that will replace the standard error_exit method:
 */

static void my_error_exit (j_common_ptr cinfo)
{
	/* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
	my_error_mgr* myerr = (my_error_mgr*) cinfo->err;
	 (*cinfo->err->output_message) (cinfo);
	longjmp (myerr->setjmp_buffer, 1);
}

/*
 * Sample routine for JPEG decompression.  We assume that the source file name
 * is passed in.  We want to return 1 on success, 0 on error.
 */

/* Le cube de couleurs */

struct cube guess [] = {
	{ 5, 10, 4 },
	{ 5, 9, 4 },
	{ 5, 8, 4 },
	{ 4, 9, 4 },
	{ 4, 8, 4 },
	{ 4, 7, 4 },
	{ 4, 9, 3 },
	{ 4, 8, 3 },
	{ 2, 2, 2 },
};

static unsigned char	reds [256],
						greens [256],
						blues [256];
unsigned char*			dither_colormap [3] = { reds, greens, blues };
static unsigned long	c_index [256];

Screen*		TheScreen;
int			theCube = -1;

/*
 *	Initialiser le cube de couleurs (appele qu'en cas de visuel PseudoColor).
 */

void initColorCube (Pixel bg)
{
	XColor		xc;
	XColor		colors [256];
	Colormap	cmap = DefaultColormapOfScreen (TheScreen);
	Visual*		visual = DefaultVisualOfScreen (TheScreen);
	Display*	dpy = DisplayOfScreen (TheScreen);
	int			r,g,b;
	int			i;

	if (theCube >= 0)
		return;		/* deja fait */

	for (theCube = 0; theCube < XtNumber (guess) - 1; theCube++)
		if (XAllocColorCells (dpy, cmap, False, NULL, 0, c_index, nred*ngreen*nblue))
		{
			i = 0;
			for (r = 0; r < nred; r++)
				for (g = 0; g < ngreen; g++)
					for (b = 0; b < nblue; b++) {
						colors [i].red = r * (65535/ (nred-1));
						colors [i].green = g * (65535/ (ngreen-1));
						colors [i].blue = b * (65535/ (nblue-1));
						colors [i].pixel = c_index [i];
						colors [i].flags = DoRed|DoGreen|DoBlue;
						/* les couleurs sur 8 bits */
						reds [i] = colors [i].red /256;
						greens [i] = colors [i].green /256;
						blues [i] = colors [i].blue /256;
						i++;
					}
			/* changer le blanc en couleur de fond. */
			colors [i-1].pixel = bg;
			XQueryColor (dpy, cmap, &colors [i-1]);
			colors [i-1].pixel = c_index [i-1];
			XStoreColors (dpy, cmap, colors, nred*ngreen*nblue);
			return;
		}
	/*
	 *	Prendre un cube read-only
	 */
	xc.flags = DoRed|DoGreen|DoBlue;
	for (; theCube < XtNumber (guess);) {
		i = 0;
		for (r = 0; r < nred; r++)
			for (g = 0; g < ngreen; g++)
				for (b = 0; b < nblue; b++) {
					xc.red = r * (65535/ (nred-1));
					xc.green = g * (65535/ (ngreen-1));
					xc.blue = b * (65535/ (nblue-1));
					if (!XAllocColor (dpy, cmap, &xc)) {
						XFreeColors (dpy, cmap, c_index, i, 0);
						goto retry;
					}
					c_index [i] = xc.pixel;
					reds [i] = xc.red /256;
					greens [i] = xc.green /256;
					blues [i] = xc.blue /256;
					i++;
				}
		if (theCube > 0)
			fprintf (stderr, "Using a %dx%dx%d read-only color cube.\n", nred, ngreen, nblue);
		return;
		retry:
			theCube++;
	}
}

static unsigned char* read_JPEG_file (char* filename, Boolean get_size, Boolean do_quantize, Boolean grey_scale, int* pw, int* ph)
{
	unsigned char*					rgbpic;
	struct jpeg_decompress_struct	cinfo;
	struct my_error_mgr				jerr;
	FILE*							infile;
	JSAMPROW						rowptr [1];
	int								width;
	int								height;
	int								bperpix;

	if ((infile = fopen (filename, "rb")) == NULL)
		return 0;

	/* Step 1: allocate and initialize JPEG decompression object */
	cinfo.err = jpeg_std_error (&jerr.pub);
	jerr.pub.error_exit = my_error_exit;
	if (setjmp (jerr.setjmp_buffer)) {
		jpeg_destroy_decompress (&cinfo);
		fclose (infile);
		return 0;
	}
	/* Now we can initialize the JPEG decompression object. */
	jpeg_create_decompress (&cinfo);
	/* Step 2: specify data source (eg, a file) */
	jpeg_stdio_src (&cinfo, infile);
	/* Step 3: read file parameters with jpeg_read_header () */
	 (void) jpeg_read_header (&cinfo, TRUE);
	/* Step 4: set parameters for decompression */
	*pw = width = cinfo.image_width;
	*ph = height = cinfo.image_height;
	if (get_size) {
		jpeg_destroy_decompress (&cinfo);
		fclose (infile);
		return (unsigned char*) 1;
	}
	if (do_quantize) {
		cinfo.actual_number_of_colors = nred*ngreen*nblue;
		cinfo.colormap = dither_colormap;
		cinfo.quantize_colors = TRUE;
		cinfo.dct_method = JDCT_FLOAT;
	} else if (grey_scale) {
		cinfo.out_color_space = JCS_GRAYSCALE;
	}
	/* Step 5: Start decompressor */
	 (void) jpeg_start_decompress (&cinfo);

	bperpix = cinfo.output_components;
	/* allocate rgb matrix */
	rgbpic = (unsigned char*) XtMalloc (width * height * bperpix);
	/* Step 6:	while (scan lines remain to be read) */
	/*			jpeg_read_scanlines (...); */
	while (cinfo.output_scanline < cinfo.output_height) {
		rowptr[0] = (JSAMPROW) &rgbpic[cinfo.output_scanline * width * bperpix];
		 (void) jpeg_read_scanlines (&cinfo, rowptr, (JDIMENSION) 1);
	}
	/* Step 7: Finish decompression */
	 (void) jpeg_finish_decompress (&cinfo);
	/* Step 8: Release JPEG decompression object */
	jpeg_destroy_decompress (&cinfo);
	fclose (infile);
	return rgbpic;
}

static int highbit (unsigned long ul)
{
	/* returns position of highest set bit in 'ul' as an integer (0-31), or -1 if none */
	int				i;
	unsigned long	hb;

	hb = 0x80;  hb = hb << 24;   /* hb = 0x80000000UL */
	for (i=31; ((ul & hb) == 0) && i>=0;  i--, ul<<=1)
		;
	return i;
}

/*
 *	Fait une XImage a partir d'une matrice RGB.
 */

XImage *rgbpic2XImage (unsigned char* rgbpic, int width, int height, Pixel bg)
{
	XColor			xc;
	XImage*			xim;
	Visual*			visual = DefaultVisualOfScreen (TheScreen);
	Display*		dpy = DisplayOfScreen (TheScreen);
	int				depth = DefaultDepthOfScreen (TheScreen);
	int				i,j;
	unsigned char	*lip, *ip, *pp;
	unsigned long	r, g, b, rmask, gmask, bmask, xcol;
	int				rshift, gshift, bshift, bperpix, bperline, border, cshift;

	xim = (XImage *) NULL;
	/*
	 * ColorMapped Visual:  PseudoColor
	 */
	if (visual->class == PseudoColor || visual -> class == GrayScale) {
		int	imWIDE;
		int	nullCount;

		nullCount = (4 - (width % 4)) & 0x03;  /* # of padding bytes per line */
	    imWIDE = width + nullCount;

		xim = XCreateImage (dpy, visual, depth, ZPixmap, 0,
							(char*)0,  width,  height, 32, imWIDE);
		xim -> data = XtMalloc (imWIDE * height);
		/* Now create the image data - pad each scanline as necessary */
		pp = rgbpic;
		ip = (unsigned char*) xim -> data;
		for (i = 0, pp = rgbpic; i < height; i++) {
			for (j = 0; j < width; j++)
				*ip++ = c_index [*pp++];
			for (j = 0; j < nullCount; j++)
				*ip++ = 0;
		}
		return xim;
	}
	if (visual->class != TrueColor)
		return xim;
	/*
	 * Non-ColorMapped Visual:  TrueColor
	 */
	xc.pixel = bg;
	xc.flags = DoRed|DoGreen|DoBlue;
	XQueryColor (dpy, DefaultColormapOfScreen (TheScreen), &xc);
	/* compute various shifting constants that we'll need... */

    rmask = visual->red_mask;
    gmask = visual->green_mask;
    bmask = visual->blue_mask;

    rshift = 7 - highbit (rmask);
    gshift = 7 - highbit (gmask);
    bshift = 7 - highbit (bmask);
    cshift = 7 - highbit ((unsigned long)255);

    xim = XCreateImage (dpy, visual, depth, ZPixmap, 0, NULL,
						width,  height, 32, 0);
    bperline = xim->bytes_per_line;
    bperpix  = xim->bits_per_pixel;
    border   = xim->byte_order;

	xim -> data = XtMalloc ((size_t) (height * bperline));
    lip = (unsigned char*) xim -> data;
	pp = rgbpic;
	for (i=0; i<height; i++, lip+=bperline) {
		for (j=0, ip=lip; j<width; j++) {
			r = *pp++;  g = *pp++;  b = *pp++;
			/* substituer la couleur de fond au blanc ... */
			if (r*g*b > 249*249*249) {
				r = xc.red >> 8;
				g = xc.green >> 8;
				b = xc.blue >> 8;
			}
			/* shift the bits around */
			if (rshift<0)	r = r << (-rshift);
			else			r = r >> rshift;
	
			if (gshift<0)	g = g << (-gshift);
			else			g = g >> gshift;
	
			if (bshift<0)	b = b << (-bshift);
			else			b = b >> bshift;
			xcol = (r & rmask) | (g & gmask) | (b & bmask);

			switch (bperpix) {
			case 32:
				if (border == MSBFirst) {
					*ip++ = (xcol>>24) & 0xff;
					*ip++ = (xcol>>16) & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ =  xcol      & 0xff;
				} else {	/* LSBFirst */
					*ip++ =  xcol      & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ = (xcol>>16) & 0xff;
					*ip++ = (xcol>>24) & 0xff;
				}
				break;

			case 24:
				if (border == MSBFirst) {
					*ip++ = (xcol>>16) & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ =  xcol      & 0xff;
				} else {	/* LSBFirst */
					*ip++ =  xcol      & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ = (xcol>>16) & 0xff;
				}
				break;

			case 16:
				if (border == MSBFirst) {
					*ip++ = (xcol>>8)  & 0xff;
					*ip++ =  xcol      & 0xff;
				} else {	/* LSBFirst */
					*ip++ =  xcol      & 0xff;
					*ip++ = (xcol>>8)  & 0xff;
				}
				break;
			}
		}
	}
	return xim;
}

unsigned char*	Jpeg2ps (char* filename, int* pwidth, int* pheight)
{
	return read_JPEG_file (filename, False, False, False, pwidth, pheight);
}

XImage*	Jpeg2XImage (char* filename, Pixel bg, int* w, int* h)
{
	Visual*			visual = DefaultVisualOfScreen (TheScreen);
	Boolean			do_quantize = (visual -> class == PseudoColor);
	unsigned char*	rgbpic;
	Pixmap			pixmap;
	XImage*			xim;
	int				width, height;
	Cache*			p;
	static Cache*	cache = (Cache*)0;

	/* chercher d'abord dans le cache */
	for (p = cache; p; p = p -> next)
		if (filename [0] == p -> name [0] && !strcmp (filename+1, p -> name+1)) {
			if (w != 0) {
				*w = p -> width;
				*h = p -> height;
			}
			return p -> image;
		}
	/* Initialiser le cube de couleur */
	if (do_quantize && w == 0)
		initColorCube (bg);
	/* lire le fichier */
	rgbpic = read_JPEG_file (filename, w != 0, do_quantize, False, &width, &height);
	if (!rgbpic)
		return (XImage*)0;
	/* Juste la geometrie?? */
	if (w != 0) {
		*w = width;
		*h = height;
		return (XImage*)0;
	}
	/* Faire une XImage */
	xim = rgbpic2XImage (rgbpic, width, height, bg);
	XtFree ((char*) rgbpic);
	/* Mettre la nouvelle image dans le cache */
	p = (Cache*) XtMalloc (sizeof (Cache));
	p -> image = xim;
	p -> name = filename;
	p -> width = width;
	p -> height = height;
	p -> next = cache;
	cache = p;
	/* et voila */
	return p -> image;
}
